
#define SGS_INTERNAL_STRINGTABLES
#define SGS_NO_STDLIB 1



#ifndef SGS_INT_H_INCLUDED
#define SGS_INT_H_INCLUDED



#ifndef SGSCRIPT_H_INCLUDED
#define SGSCRIPT_H_INCLUDED


#define SGS_VERSION_MAJOR 1
#define SGS_VERSION_MINOR 4
#define SGS_VERSION_INCR  1
#define SGS_VERSION "1.4.1"

#define SGS_VERSION_OFFSET 8
#define SGS_VERSION_INT ( ( ( ( SGS_VERSION_MAJOR << SGS_VERSION_OFFSET ) | \
        SGS_VERSION_MINOR ) << SGS_VERSION_OFFSET ) | SGS_VERSION_INCR )


#ifdef __cplusplus
extern "C" {
#endif

#include <stddef.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>


/******
* CONFIGURATION *
     ==========*/

/* global debug switch */
#ifndef SGS_DEBUG
#  ifdef _DEBUG
#    define SGS_DEBUG 1
#  else
#    define SGS_DEBUG 0
#  endif
#endif

/* compile without standard libraries */
#ifndef SGS_NO_STDLIB
#  define SGS_NO_STDLIB 0
#endif

/* stage output dumps */
#ifndef SGS_DEBUG_DATA
#  define SGS_DEBUG_DATA 0
#endif

/* function entry/exit tagging */
#ifndef SGS_DEBUG_FLOW
#  define SGS_DEBUG_FLOW 0
#endif

/* instruction logging */
#ifndef SGS_DEBUG_INSTR
#  define SGS_DEBUG_INSTR 0
#endif

/* stack dump before execution of each instruction */
#ifndef SGS_DEBUG_STATE
#  define SGS_DEBUG_STATE 0
#endif

/* break-if */
#ifndef SGS_DEBUG_VALIDATE
#  define SGS_DEBUG_VALIDATE 1
#endif

/* extensive buffer overflow tests */
#ifndef SGS_DEBUG_EXTRA
#  define SGS_DEBUG_EXTRA 0
#endif

/* checks for having more memory allocated with..
	..the context memory functions than freed
	very fast, runs only in sgs_DestroyEngine */
#ifndef SGS_DEBUG_LEAKS
#  define SGS_DEBUG_LEAKS 1
#endif


/* profiling */
#ifndef SGS_DUMP_BYTECODE
#  define SGS_DUMP_BYTECODE 0
#endif

#ifndef SGS_PERFEVENT
#  define SGS_PERFEVENT( x )
#endif


/* SGScript error messages */
#define SGS_ERRMSG_CALLSTACKLIMIT "max. call stack size reached"


/* bytes reserved initially for token storage
	used to prevent multiple reallocations in the beginning */
#define SGS_TOKENLIST_PREALLOC 1024

/* maximum length for a string to be included in the string table */
#define SGS_STRINGTABLE_MAXLEN 0x7fffffff

/* max. length of the call stack */
#define SGS_MAX_CALL_STACK_SIZE 200

/* the size of the stack buffer for temporary vsprintf'ed text
	if predicted text length is longer than that, the memory
	is allocated using the context allocator */
#define SGS_OUTPUT_STACKBUF_SIZE 128

/* postfix for function names in C functions
	EXTRABYTES should be set to the length of EXTRASTRING */
#define SGS_MSG_EXTRASTRING "(): "
#define SGS_MSG_EXTRABYTES 4

/* object pool settings
	- size of object pool (max. number of items)
	- max. appended memory size for pool usage */
#define SGS_OBJPOOL_SIZE 64
#define SGS_OBJPOOL_MAX_APPMEM 256

/* stack frame pool settings */
#define SGS_STACKFRAMEPOOL_SIZE 512

/* inclusion system settings:
	- name of entry point function to look for
	- the default include path */
#define SGS_LIB_ENTRY_POINT "sgscript_main"
#define SGS_INCLUDE_FILEPATHS( PFX ) \
	PFX "?;" /*exact name */ \
	PFX "?" SGS_MODULE_EXT ";" /* library without prefix */ \
	PFX "lib?" SGS_MODULE_EXT ";" /* library with prefix */ \
	PFX "?.sgc;" /* compiled SGScript code */ \
	PFX "?.sgs" /* SGScript source code */
#define SGS_INCLUDE_PATH \
	SGS_INCLUDE_FILEPATHS( "|/" ) ";" /* in calling file's directory */ \
	SGS_INCLUDE_FILEPATHS( "" ) ";" /* in current directory */ \
	SGS_INCLUDE_FILEPATHS( "@/" ) /* in process directory */


/*****
* X-PLATFORM *
    ========*/

#ifdef _MSC_VER
#  pragma warning( disable: 4996 )
#  include <stdint.h>
#  if _MSC_VER >= 1700
#    include <inttypes.h>
#  else
#    define PRId32 "d"
#    define PRId64 "lld"
#  endif
#  define SGS_INLINE
#  define snprintf _snprintf
#  define SGS_VSPRINTF_LEN( str, args ) _vscprintf( str, args )
#else
#  include <inttypes.h>
#  define SGS_INLINE inline
#  define SGS_VSPRINTF_LEN( str, args ) vsnprintf( NULL, 0, str, args )
#endif

#define SGS_UNUSED( x ) (void)(x)

/* ideally it would be possible to specify how the pointer got its alignment ..
.. or for the compiler to detect that; now it just throws pointless warnings...
On top of that, OSX Clang seems to align sgs_Variable to 4 bytes even though ..
.. the structure contains a union that contains pointers and a double. */
#define SGS_ASSUME_ALIGNED( x, t ) ((t*)(void*)(x))
#define SGS_ASSUME_ALIGNED_CONST( x, t ) ((const t*)(const void*)(x))

#ifdef __cplusplus
#  define SGS_DECLARE extern
#  define SGS_CLINK extern "C"
#else
#  define SGS_DECLARE static
#  define SGS_CLINK
#endif


#ifdef SGS_COMPILE_MODULE
#  define BUILDING_SGS 1
#endif

#if defined( _WIN32 ) || defined( __CYGWIN__ )
#  define SGS_DLL_IMPORT __declspec(dllimport)
#  define SGS_DLL_EXPORT __declspec(dllexport)
#else
#  if __GNUC__ >= 4
#    define SGS_DLL_IMPORT __attribute__((visibility("default")))
#    define SGS_DLL_EXPORT __attribute__((visibility("default")))
#  else
#    define SGS_DLL_IMPORT
#    define SGS_DLL_EXPORT
#  endif
#endif

#if SGS_DLL
#  if BUILDING_SGS
#    define SGS_APIFUNC SGS_DLL_EXPORT
#  else
#    define SGS_APIFUNC SGS_DLL_IMPORT
#  endif
#  define SGS_IF_DLL( dll, nodll ) dll
#else
#  define SGS_APIFUNC
#  define SGS_IF_DLL( dll, nodll ) nodll
#endif

#ifdef SGS_COMPILE_MODULE
#  define SGS_MODULE_ENTRY_POINT( fname ) \
	SGS_CLINK SGS_APIFUNC int sgscript_main( SGS_CTX ){ return fname( C ); }
#else
#  define SGS_MODULE_ENTRY_POINT( fname )
#endif

#if defined(WINAPI_FAMILY) && (WINAPI_FAMILY == WINAPI_FAMILY_PC_APP || WINAPI_FAMILY == WINAPI_FAMILY_PHONE_APP)
#  define SGS_WINAPP 1
#endif

#ifdef SGS_USE_FILESYSTEM
#  include <sys/types.h>
#  include <sys/stat.h>

#  ifdef _WIN32
#    if SGS_WINAPP
#      include <windows.h>
#    else
#      include <direct.h>
#      define getcwd _getcwd
#      define mkdir _mkdir
#      define rmdir _rmdir
#      define stat _stat

#    endif
#  else
#    include <unistd.h>
#    include <dirent.h>
#  endif
#endif

#ifdef _WIN32
#  define SGS_MODULE_EXT ".dll"
#else
#  define SGS_MODULE_EXT ".so"
#endif

/* basic platform info */
#if _WIN32 || __WIN32__ || __WINDOWS__
#  define SGS_OS_TYPE "Windows"
#elif __APPLE__
#  define SGS_OS_TYPE "Mac OS X"
#elif __ANDROID__
#  define SGS_OS_TYPE "Android"
#elif __linux || __linux__
#  define SGS_OS_TYPE "Linux"
#elif __unix || __unix__
#  define SGS_OS_TYPE "Unix"
#elif __posix
#  define SGS_OS_TYPE "POSIX"
#else
#  define SGS_OS_TYPE "Unknown"
#endif

/* simplified architecture info */
#if defined __arm__ || defined __thumb__ || defined _M_ARM || defined _M_ARMT
#  define SGS_ARCH_ARM 1
#endif
#if defined i386 || defined __i386 || defined __i386 || defined _M_IX86 || defined _X86_ || defined __X86__
#  define SGS_ARCH_X86 1
#endif
#if defined __amd64__ || defined __amd64 || defined __x86_64__ || defined __x86_64 || defined _M_X64 || defined _M_AMD64
#  define SGS_ARCH_AMD64 1
#endif
#if defined __ia64__ || defined _IA64 || defined __IA64__ || defined _M_IA64
#  define SGS_ARCH_IA64 1
#endif
#if defined __powerpc || defined __powerpc__ || defined __powerpc64__ || \
	defined __ppc__ || defined __ppc64 || defined _M_PPC || defined _XENON
#  define SGS_ARCH_PPC 1
#endif


/* http://stackoverflow.com/a/2103095/1648140 */

enum
{
    O32_LITTLE_ENDIAN = 0x03020100ul,
    O32_BIG_ENDIAN = 0x00010203ul
};

static const union { unsigned char bytes[4]; uint32_t value; } o32_host_order =
    { { 0, 1, 2, 3 } };

#define O32_HOST_ORDER (o32_host_order.value)


#ifdef __cplusplus
extern "C" {
#endif

#define SGS_XPC_NOFILE -1
#define SGS_XPC_NOPROC -2
#define SGS_XPC_NOTSUP -3
#define SGS_XPC_NOTLIB -4
#define SGS_XPC_LDFAIL -5
int sgsXPC_GetProcAddress( const char* file, const char* proc, void** out );

char* sgsXPC_GetCurrentDirectory();
int sgsXPC_SetCurrentDirectory( char* path );
char* sgsXPC_GetModuleFileName();

#ifdef __cplusplus
}
#endif


/***
* SYSTEM *
  ======*/

/* API self-doc helpers */
#define SGSZERO   int     /* always 0 */
#define SGSONE    int     /* always 1 */
#define SGSRESULT int     /* output code */
#define SGSBOOL   int     /* true (1) / false (0) */


/* Output codes */
#define SGS_SUCCESS  0  /* success (no errors) */
#define SGS_ENOTFND -1  /* item was not found */
#define SGS_ECOMP   -2  /* compile error */
#define SGS_ENOTSUP -4  /* not supported */
#define SGS_EBOUNDS -5  /* index out of bounds */
#define SGS_EINVAL  -6  /* invalid value was passed */
#define SGS_EINPROC -7  /* process was interrupted */


/* Code tests */
#define SGS_FAILED( e ) ((e)<0)
#define SGS_SUCCEEDED( e ) ((e)>=0)


/* Accessible / transferable data */
typedef int32_t sgs_Bool;
typedef int64_t sgs_Int;
typedef double  sgs_Real;
typedef int32_t sgs_SizeVal;
typedef int32_t sgs_StkIdx;
typedef struct  sgs_Context sgs_Context;
typedef struct  sgs_Variable sgs_Variable;
typedef struct  sgs_StackFrame sgs_StackFrame;
typedef int (*sgs_CFunc) ( sgs_Context* );

#define sgs_Integer sgs_Int
#define sgs_Float sgs_Real


/* Memory management */
typedef void* (*sgs_MemFunc)
(
	void* /* userdata */,
	void* /* curmem */,
	size_t /* numbytes */
);

/*
	unlike realloc, if size = 0, this function MUST return NULL
*/
static void* sgs_DefaultMemFunc( void* ud, void* ptr, size_t size )
{
	SGS_UNUSED( ud );
	if( ptr && size ) return realloc( ptr, size );
	else if( size )   return malloc( size );
	if( ptr )         free( ptr );
	return NULL;
}


/* Main output */
typedef void (*sgs_OutputFunc) (
	void* /* userdata */,
	sgs_Context* /* ctx / SGS_CTX */,
	const void* /* ptr */,
	size_t /* size */
);


/* Debug output */
#define SGS_INFO    100  /* information about potential issues and state of the system */
#define SGS_WARNING 200  /* non-fatal problems */
#define SGS_ERROR   300  /* fatal problems */
#define SGS_APIERR  330  /* API usage errors */
#define SGS_INTERR  360  /* internal/integral/interesting errors */

typedef void (*sgs_MsgFunc) (
	void* /* data */,
	sgs_Context* /* ctx / SGS_CTX */,
	int /* type */,
	const char* /* message */
);


/* Event Hook */
/* additional info can be found in the last stack frame */
#define SGS_HOOK_ENTER 1 /* entered a function */
#define SGS_HOOK_EXIT  2 /* exited a function */
#define SGS_HOOK_STEP  3 /* about to execute an instruction */
#define SGS_HOOK_PAUSE 4 /* paused execution */
#define SGS_HOOK_CONT  5 /* continuing execution */
#define SGS_HOOK_CREAT 6 /* context created */
#define SGS_HOOK_CFORK 7 /* context forked */
#define SGS_HOOK_CFREE 8 /* context freed */
typedef void (*sgs_HookFunc) (
	void* /* userdata */,
	sgs_Context* /* ctx / SGS_CTX */,
	int /* event_id */
);


/* Script file system */
typedef struct sgs_ScriptFSData
{
	void* userhandle;
	const char* filename;
	void* output;
	size_t size;
}
sgs_ScriptFSData;
#define SGS_SFS_FILE_EXISTS 1
#define SGS_SFS_FILE_OPEN 2
#define SGS_SFS_FILE_READ 3
#define SGS_SFS_FILE_CLOSE 4
typedef SGSRESULT (*sgs_ScriptFSFunc) (
	void* /* userdata */,
	sgs_Context* /* ctx / SGS_CTX */,
	int /* op */,
	sgs_ScriptFSData* /* data */
);


/* for state/query functions */
#define SGS_TRUE 1
#define SGS_FALSE 0
#define SGS_QUERY -1
#define SGS_SERIALIZE_DEFAULT 0


/* Statistics / debugging */
#define SGS_STAT_VERSION      0
#define SGS_STAT_STATECOUNT   1
#define SGS_STAT_OBJCOUNT     2
#define SGS_STAT_MEMSIZE      3
#define SGS_STAT_NUMALLOCS    4
#define SGS_STAT_NUMFREES     5
#define SGS_STAT_NUMBLOCKS    6
#define SGS_STAT_DUMP_STACK   10
#define SGS_STAT_DUMP_GLOBALS 11
#define SGS_STAT_DUMP_OBJECTS 12
#define SGS_STAT_DUMP_FRAMES  13
#define SGS_STAT_DUMP_STATS   14
#define SGS_STAT_DUMP_SYMBOLS 15
#define SGS_STAT_DUMP_RSRC    16
#define SGS_STAT_XDUMP_STACK  20


/* Virtual machine control */
#define SGS_CNTL_STATE      1
#define SGS_CNTL_GET_STATE  2
#define SGS_CNTL_MINLEV     3
#define SGS_CNTL_GET_MINLEV 4
#define SGS_CNTL_APILEV     5
#define SGS_CNTL_GET_APILEV 6
#define SGS_CNTL_ERRNO      7
#define SGS_CNTL_SET_ERRNO  8
#define SGS_CNTL_GET_ERRNO  9
#define SGS_CNTL_ERRSUP     10
#define SGS_CNTL_GET_ERRSUP 11
#define SGS_CNTL_NUMRETVALS 13
#define SGS_CNTL_GET_PAUSED 14
#define SGS_CNTL_GET_ABORT  15


/* Context internals */
/* - variable types */
#define SGS_VT_NULL   0
#define SGS_VT_BOOL   1
#define SGS_VT_INT    2
#define SGS_VT_REAL   3
#define SGS_VT_STRING 4
#define SGS_VT_FUNC   5
#define SGS_VT_CFUNC  6
#define SGS_VT_OBJECT 7
#define SGS_VT_PTR    8
#define SGS_VT_THREAD 9
#define SGS_VT__COUNT 10

/* - object data */
typedef struct sgs_ObjData sgs_VarObj;
typedef int (*sgs_ObjCallback) ( sgs_Context*, sgs_VarObj* );

typedef int (*sgs_OC_Self) ( sgs_Context*, sgs_VarObj* );
typedef int (*sgs_OC_SlPr) ( sgs_Context*, sgs_VarObj*, int );

typedef struct sgs_ObjInterface
{
	const char* name;
	
	sgs_OC_Self destruct;
	sgs_OC_Self gcmark;
	
	sgs_OC_Self getindex;
	sgs_OC_Self setindex;
	
	sgs_OC_SlPr convert;
	sgs_OC_Self serialize;
	sgs_OC_SlPr dump;
	sgs_OC_SlPr getnext;
	
	sgs_OC_Self call;
	sgs_OC_Self expr;
}
sgs_ObjInterface;

struct sgs_ObjData
{
	sgs_SizeVal refcount;
	uint32_t appsize;
	uint8_t redblue;    /* red or blue? mark & sweep */
	uint8_t mm_enable;  /* use metamethods? */
	uint8_t is_iface;   /* whether is also in the interface table */
	uint8_t in_setindex; /* whether already running custom setindex code */
	void* data;         /* should have offset=8 with packing alignment>=8 */
	sgs_ObjInterface* iface;
	sgs_VarObj* prev;   /* pointer to previous GC object */
	sgs_VarObj* next;   /* pointer to next GC object */
	sgs_VarObj* metaobj; /* pointer to meta-object */
};

typedef struct sgs_iStr sgs_iStr;
struct sgs_iStr
{
	sgs_SizeVal refcount;
	uint32_t size;
	uint32_t hash;
};
#define sgs_str_cstr( pstr ) (((char*)(pstr))+sizeof(sgs_iStr))
#define sgs_str_c_cstr( pstr ) (((const char*)(pstr))+sizeof(sgs_iStr))
#define sgs_var_cstr( var ) sgs_str_cstr( (var)->data.S )

typedef struct sgs_iFunc sgs_iFunc;
typedef union sgs_VarData
{
	sgs_SizeVal* pRC;
	sgs_Bool     B;
	sgs_Int      I;
	sgs_Real     R;
	sgs_iStr*    S;
	sgs_iFunc*   F;
	sgs_CFunc    C;
	sgs_VarObj*  O;
	void*        P;
	sgs_Context* T;
}
sgs_VarData;

struct sgs_Variable
{
	uint32_t    type;
	sgs_VarData data;
};


/* parameter flags / special values */
#define SGS_GETNEXT_KEY   0x01
#define SGS_GETNEXT_VALUE 0x02

#define SGS_EOP_ADD       0
#define SGS_EOP_SUB       1
#define SGS_EOP_MUL       2
#define SGS_EOP_DIV       3
#define SGS_EOP_MOD       4
#define SGS_EOP_COMPARE   5
#define SGS_EOP_NEGATE    6

#define SGS_CONVOP_CLONE  0x10000
#define SGS_CONVOP_TOITER 0x30000


/* Engine context */
#define SGS_CTX sgs_Context* C


SGS_APIFUNC
sgs_Context*
sgs_CreateEngineExt
(
	sgs_MemFunc memfunc,
	void* mfuserdata
);
static SGS_INLINE sgs_Context* sgs_CreateEngine()
	{ return sgs_CreateEngineExt( sgs_DefaultMemFunc, NULL ); }

SGS_APIFUNC void sgs_DestroyEngine( SGS_CTX );
SGS_APIFUNC sgs_Context* sgs_RootContext( SGS_CTX );
SGS_APIFUNC sgs_Context* sgs_ForkState( SGS_CTX, int copystate );
SGS_APIFUNC void sgs_ReleaseState( SGS_CTX );
SGS_APIFUNC SGSBOOL sgs_PauseState( SGS_CTX );
SGS_APIFUNC SGSBOOL sgs_ResumeStateRet( SGS_CTX, int args, int* outrvc );
SGS_APIFUNC SGSBOOL sgs_ResumeStateExp( SGS_CTX, int args, int expect );
#define sgs_ResumeState( C ) sgs_ResumeStateExp( C, 0, 0 )
SGS_APIFUNC void sgs_CreateSubthread( sgs_Context* T, SGS_CTX,
	sgs_Variable* out, sgs_StkIdx args, int gotthis );
SGS_APIFUNC int sgs_ProcessSubthreads( SGS_CTX, sgs_Real dt );
SGS_APIFUNC void sgs_EndOn( SGS_CTX, sgs_Variable ev, int enable );


#define SGS_CODE_ER 0 /* error codes */
#define SGS_CODE_VT 1 /* variable type */
#define SGS_CODE_OP 2 /* VM instruction */
SGS_APIFUNC const char* sgs_CodeString( int type, int val );


/* Core systems */
#define SGSOUTPUTFN_DEFAULT ((sgs_OutputFunc)-1)
SGS_APIFUNC void sgs_GetOutputFunc( SGS_CTX, sgs_OutputFunc* outf, void** outc );
SGS_APIFUNC void sgs_SetOutputFunc( SGS_CTX, sgs_OutputFunc func, void* ctx );
SGS_APIFUNC void sgs_Write( SGS_CTX, const void* ptr, size_t size );
SGS_APIFUNC SGSBOOL sgs_Writef( SGS_CTX, const char* what, ... );

SGS_APIFUNC void sgs_GetErrOutputFunc( SGS_CTX, sgs_OutputFunc* outf, void** outc );
SGS_APIFUNC void sgs_SetErrOutputFunc( SGS_CTX, sgs_OutputFunc func, void* ctx );
SGS_APIFUNC void sgs_ErrWrite( SGS_CTX, const void* ptr, size_t size );
SGS_APIFUNC SGSBOOL sgs_ErrWritef( SGS_CTX, const char* what, ... );

#define SGSMSGFN_DEFAULT ((sgs_MsgFunc)-1)
#define SGSMSGFN_DEFAULT_NOABORT ((sgs_MsgFunc)-2)
SGS_APIFUNC void sgs_GetMsgFunc( SGS_CTX, sgs_MsgFunc* outf, void** outc );
SGS_APIFUNC void sgs_SetMsgFunc( SGS_CTX, sgs_MsgFunc func, void* ctx );
SGS_APIFUNC SGSZERO sgs_Msg( SGS_CTX, int type, const char* what, ... );

typedef int (*sgs_ErrorOutputFunc) ( void*, const char*, ... );
#define SGS_ERRORINFO_STACK 0x01
#define SGS_ERRORINFO_ERROR 0x02
#define SGS_ERRORINFO_FULL (SGS_ERRORINFO_STACK | SGS_ERRORINFO_ERROR)
SGS_APIFUNC void sgs_WriteErrorInfo( SGS_CTX, int flags,
	sgs_ErrorOutputFunc func, void* ctx, int type, const char* msg );
SGS_APIFUNC void sgs_PushErrorInfo( SGS_CTX, int flags, int type, const char* msg );

SGS_APIFUNC SGSBOOL sgs_GetHookFunc( SGS_CTX, sgs_HookFunc* outf, void** outc );
SGS_APIFUNC void sgs_SetHookFunc( SGS_CTX, sgs_HookFunc func, void* ctx );

SGS_APIFUNC SGSBOOL sgs_GetScriptFSFunc( SGS_CTX, sgs_ScriptFSFunc* outf, void** outc );
SGS_APIFUNC void sgs_SetScriptFSFunc( SGS_CTX, sgs_ScriptFSFunc func, void* ctx );

SGS_APIFUNC void* sgs_Memory( SGS_CTX, void* ptr, size_t size );
#define sgs_Malloc( C, size ) sgs_Memory( C, NULL, size )
#define sgs_Free( C, ptr ) sgs_Memory( C, ptr, 0 )
#define sgs_Realloc sgs_Memory
/* assumes SGS_CTX: */
#define sgs_Alloc( what ) (what*) sgs_Malloc( C, sizeof(what) )
#define sgs_Alloc_n( what, cnt ) (what*) sgs_Malloc( C, sizeof(what) * cnt )
#define sgs_Alloc_a( what, app ) (what*) sgs_Malloc( C, sizeof(what) + app )
#define sgs_Dealloc( ptr ) sgs_Free( C, ptr )


SGS_APIFUNC SGSRESULT sgs_EvalBuffer( SGS_CTX, const char* buf, size_t size );
SGS_APIFUNC SGSRESULT sgs_EvalFile( SGS_CTX, const char* file );
SGS_APIFUNC SGSRESULT sgs_AdjustStack( SGS_CTX, int expected, int ret );
SGS_APIFUNC SGSBOOL sgs_IncludeExt( SGS_CTX, const char* name, const char* searchpath );
SGS_APIFUNC SGSRESULT sgs_Compile( SGS_CTX, const char* buf, size_t size, char** outbuf, size_t* outsize );

SGS_APIFUNC SGSRESULT sgs_DumpCompiled( SGS_CTX, const char* buf, size_t size );
SGS_APIFUNC SGSBOOL sgs_Abort( SGS_CTX );
SGS_APIFUNC ptrdiff_t sgs_Stat( SGS_CTX, int type );
SGS_APIFUNC int32_t sgs_Cntl( SGS_CTX, int what, int32_t val );

SGS_APIFUNC void sgs_StackFrameInfo( SGS_CTX, sgs_StackFrame* frame, const char** name, const char** file, int* line );
SGS_APIFUNC sgs_StackFrame* sgs_GetFramePtr( SGS_CTX, sgs_StackFrame* from, int bwd );


#ifndef SGS_STRINGLENGTHFUNC
#define SGS_STRINGLENGTHFUNC strlen
#endif
#define SGS_STRLITBUF( lit ) lit, sizeof(lit) - 1

#define sgs_ExecBuffer( C, buf, sz ) sgs_AdjustStack( C, 0, sgs_EvalBuffer( C, buf, sz ) )
#define sgs_ExecString( C, str ) sgs_ExecBuffer( C, str, SGS_STRINGLENGTHFUNC( str ) )
#define sgs_EvalString( C, str ) sgs_EvalBuffer( C, str, SGS_STRINGLENGTHFUNC( str ) )
#define sgs_ExecFile( C, str ) sgs_AdjustStack( C, 0, sgs_EvalFile( C, str ) )
#define sgs_Include( C, str ) sgs_IncludeExt( C, str, NULL )
#define sgs_WriteStr( C, str ) sgs_Write( C, str, SGS_STRINGLENGTHFUNC( str ) )
#define sgs_ErrWriteStr( C, str ) sgs_ErrWrite( C, str, SGS_STRINGLENGTHFUNC( str ) )


/* Additional libraries */

#if !SGS_NO_STDLIB
SGS_APIFUNC void sgs_LoadLib_Fmt( SGS_CTX );
SGS_APIFUNC void sgs_LoadLib_IO( SGS_CTX );
SGS_APIFUNC void sgs_LoadLib_Math( SGS_CTX );
SGS_APIFUNC void sgs_LoadLib_OS( SGS_CTX );
SGS_APIFUNC void sgs_LoadLib_RE( SGS_CTX );
SGS_APIFUNC void sgs_LoadLib_String( SGS_CTX );
#endif

#define SGS_RC_END() { NULL, NULL }

SGS_APIFUNC void sgs_RegSymbol( SGS_CTX, const char* prefix, const char* name, sgs_Variable var );
SGS_APIFUNC SGSBOOL sgs_GetSymbol( SGS_CTX, sgs_Variable var, sgs_Variable* out );

typedef struct sgs_RegFuncConst
{
	const char* name;
	sgs_CFunc value;
}
sgs_RegFuncConst;
SGS_APIFUNC void sgs_RegFuncConstsExt( SGS_CTX, const sgs_RegFuncConst* list, int size, const char* prefix );
#define sgs_RegFuncConsts( C, list, size ) sgs_RegFuncConstsExt( C, list, size, "" );
SGS_APIFUNC void sgs_StoreFuncConsts( SGS_CTX, sgs_Variable var, const sgs_RegFuncConst* list, int size, const char* prefix );

typedef struct sgs_RegIntConst
{
	const char* name;
	sgs_Int value;
}
sgs_RegIntConst;
SGS_APIFUNC void sgs_RegIntConsts( SGS_CTX, const sgs_RegIntConst* list, int size );
SGS_APIFUNC void sgs_StoreIntConsts( SGS_CTX, sgs_Variable var, const sgs_RegIntConst* list, int size );

typedef struct sgs_RegRealConst
{
	const char* name;
	sgs_Real value;
}
sgs_RegRealConst;
SGS_APIFUNC void sgs_RegRealConsts( SGS_CTX, const sgs_RegRealConst* list, int size );
SGS_APIFUNC void sgs_StoreRealConsts( SGS_CTX, sgs_Variable var, const sgs_RegRealConst* list, int size );

SGS_APIFUNC SGSBOOL sgs_RegisterType( SGS_CTX, const char* name, sgs_ObjInterface* iface );
SGS_APIFUNC SGSBOOL sgs_UnregisterType( SGS_CTX, const char* name );
SGS_APIFUNC sgs_ObjInterface* sgs_FindType( SGS_CTX, const char* name );

SGS_APIFUNC SGSONE sgs_PushInterface( SGS_CTX, sgs_CFunc igfn );
SGS_APIFUNC void sgs_InitInterface( SGS_CTX, sgs_Variable* var, sgs_CFunc igfn );


/* The core interface */

/*
	OBJECT CREATION
*/

static SGS_INLINE sgs_Variable sgs_MakeNull()
{
	sgs_Variable out; out.type = SGS_VT_NULL; return out;
}
static SGS_INLINE sgs_Variable sgs_MakeBool( sgs_Bool v )
{
	sgs_Variable out; out.type = SGS_VT_BOOL; out.data.B = v ? 1 : 0; return out;
}
static SGS_INLINE sgs_Variable sgs_MakeInt( sgs_Int v )
{
	sgs_Variable out; out.type = SGS_VT_INT; out.data.I = v; return out;
}
static SGS_INLINE sgs_Variable sgs_MakeReal( sgs_Real v )
{
	sgs_Variable out; out.type = SGS_VT_REAL; out.data.R = v; return out;
}
static SGS_INLINE sgs_Variable sgs_MakeCFunc( sgs_CFunc v )
{
	sgs_Variable out; out.type = SGS_VT_CFUNC; out.data.C = v; return out;
}
static SGS_INLINE sgs_Variable sgs_MakeObjPtrNoRef( sgs_VarObj* o )
{
	sgs_Variable out; out.type = SGS_VT_OBJECT; out.data.O = o; return out;
}
static SGS_INLINE sgs_Variable sgs_MakePtr( void* v )
{
	sgs_Variable out; out.type = SGS_VT_PTR; out.data.P = v; return out;
}

SGS_APIFUNC void sgs_InitStringBuf( SGS_CTX, sgs_Variable* out, const char* str, sgs_SizeVal size );
SGS_APIFUNC void sgs_InitString( SGS_CTX, sgs_Variable* out, const char* str );
SGS_APIFUNC void sgs_InitObjectPtr( sgs_Variable* out, sgs_VarObj* obj );
SGS_APIFUNC void sgs_InitThreadPtr( sgs_Variable* out, sgs_Context* T );

/* pushed to stack if out = null */
SGS_APIFUNC SGSONE sgs_CreateObject( SGS_CTX, sgs_Variable* out, void* data, sgs_ObjInterface* iface );
SGS_APIFUNC void* sgs_CreateObjectIPA( SGS_CTX, sgs_Variable* out, uint32_t added, sgs_ObjInterface* iface );
SGS_APIFUNC SGSONE sgs_CreateArray( SGS_CTX, sgs_Variable* out, sgs_SizeVal numitems );
SGS_APIFUNC SGSONE sgs_CreateDict( SGS_CTX, sgs_Variable* out, sgs_SizeVal numitems );
SGS_APIFUNC SGSONE sgs_CreateMap( SGS_CTX, sgs_Variable* out, sgs_SizeVal numitems );
SGS_APIFUNC SGSONE sgs_CreateEvent( SGS_CTX, sgs_Variable* out );
SGS_APIFUNC SGSONE sgs_CreatePooledEventBuf( SGS_CTX, sgs_Variable* out, sgs_Variable dict, const char* str, sgs_SizeVal size );
#define sgs_CreatePooledEvent( C, out, dict, str ) sgs_CreatePooledEventBuf( C, out, dict, str, SGS_STRINGLENGTHFUNC( str ) )

/*
	STACK & SUB-ITEMS
*/
#define sgs_PushNull( C ) sgs_PushNulls( C, 1 )
SGS_APIFUNC SGSONE sgs_PushNulls( SGS_CTX, int count );
SGS_APIFUNC SGSONE sgs_PushBool( SGS_CTX, sgs_Bool value );
SGS_APIFUNC SGSONE sgs_PushInt( SGS_CTX, sgs_Int value );
SGS_APIFUNC SGSONE sgs_PushReal( SGS_CTX, sgs_Real value );
SGS_APIFUNC SGSONE sgs_PushStringBuf( SGS_CTX, const char* str, sgs_SizeVal size );
SGS_APIFUNC SGSONE sgs_PushString( SGS_CTX, const char* str );
SGS_APIFUNC SGSONE sgs_PushCFunc( SGS_CTX, sgs_CFunc func );
SGS_APIFUNC SGSONE sgs_PushObjectPtr( SGS_CTX, sgs_VarObj* obj );
SGS_APIFUNC SGSONE sgs_PushThreadPtr( SGS_CTX, sgs_Context* T );
SGS_APIFUNC SGSONE sgs_PushPtr( SGS_CTX, void* ptr );

SGS_APIFUNC SGSONE sgs_PushVariable( SGS_CTX, sgs_Variable var );
#define sgs_PushItem( C, item ) sgs_PushVariable( C, sgs_StackItem( C, item ) )
SGS_APIFUNC void sgs_StoreVariable( SGS_CTX, sgs_Variable* out );
SGS_APIFUNC void sgs_SetStackItem( SGS_CTX, sgs_StkIdx item, sgs_Variable val );
SGS_APIFUNC void sgs_InsertVariable( SGS_CTX, sgs_StkIdx pos, sgs_Variable val );

/* string generation */
SGS_APIFUNC char* sgs_PushStringAlloc( SGS_CTX, sgs_SizeVal size );
SGS_APIFUNC char* sgs_InitStringAlloc( SGS_CTX, sgs_Variable* var, sgs_SizeVal size );
SGS_APIFUNC void sgs_FinalizeStringAlloc( SGS_CTX, sgs_StkIdx item );
SGS_APIFUNC void sgs_FinalizeStringAllocP( SGS_CTX, sgs_Variable* var );

/* almost-raw access */
SGS_APIFUNC SGSBOOL sgs_GetIndex( SGS_CTX, sgs_Variable obj, sgs_Variable idx, sgs_Variable* out, int isprop );
SGS_APIFUNC SGSBOOL sgs_SetIndex( SGS_CTX, sgs_Variable obj, sgs_Variable idx, sgs_Variable val, int isprop );
SGS_APIFUNC SGSBOOL sgs_PushIndex( SGS_CTX, sgs_Variable obj, sgs_Variable idx, int isprop );

/* special case sub-variable access */
SGS_APIFUNC SGSBOOL sgs_PushProperty( SGS_CTX, sgs_Variable obj, const char* name );
SGS_APIFUNC SGSBOOL sgs_SetProperty( SGS_CTX, sgs_Variable obj, const char* name, sgs_Variable val );
SGS_APIFUNC SGSBOOL sgs_PushNumIndex( SGS_CTX, sgs_Variable obj, sgs_Int idx );
SGS_APIFUNC SGSBOOL sgs_SetNumIndex( SGS_CTX, sgs_Variable obj, sgs_Int idx, sgs_Variable val );

/* raw & simplified global access */
SGS_APIFUNC SGSBOOL sgs_GetGlobal( SGS_CTX, sgs_Variable idx, sgs_Variable* out );
SGS_APIFUNC SGSBOOL sgs_SetGlobal( SGS_CTX, sgs_Variable idx, sgs_Variable val );
SGS_APIFUNC SGSBOOL sgs_PushGlobalByName( SGS_CTX, const char* name );
SGS_APIFUNC SGSBOOL sgs_GetGlobalByName( SGS_CTX, const char* name, sgs_Variable* out );
SGS_APIFUNC void sgs_SetGlobalByName( SGS_CTX, const char* name, sgs_Variable val );

#define SGS_REG_ROOT 0
#define SGS_REG_SYM 1
#define SGS_REG_INC 2
SGS_APIFUNC sgs_Variable sgs_Registry( SGS_CTX, int subtype );

SGS_APIFUNC void sgs_GetEnv( SGS_CTX, sgs_Variable* out );
SGS_APIFUNC void sgs_SetEnv( SGS_CTX, sgs_Variable var );
SGS_APIFUNC void sgs_PushEnv( SGS_CTX );

SGS_APIFUNC SGSBOOL sgs_PushPath( SGS_CTX, sgs_Variable var, const char* path, ... );
SGS_APIFUNC SGSBOOL sgs_StorePath( SGS_CTX, sgs_Variable var, sgs_Variable val, const char* path, ... );

SGS_APIFUNC int sgs_ArgErrorExt( SGS_CTX, int argid, int gotthis, const char* expect, const char* expfx );
SGS_APIFUNC int sgs_ArgError( SGS_CTX, int argid, int gotthis, int expect, int is_strict );

#define sgs_FuncArgError( C, argid, expect, strict ) sgs_ArgError( C, argid, 0, expect, strict )
#define sgs_MethodArgError( C, argid, expect, strict ) sgs_ArgError( C, argid, 1, expect, strict )

typedef int (*sgs_ArgCheckFunc)
(
	sgs_Context* /* ctx / SGS_CTX */,
	int /* argid */,
	va_list* /* args */,
	int /* flags */
);
#define SGS_LOADARG_STRICT 0x01
#define SGS_LOADARG_WRITE 0x02
#define SGS_LOADARG_OPTIONAL 0x04
#define SGS_LOADARG_INTSIGN 0x08
#define SGS_LOADARG_INTRANGE 0x10
#define SGS_LOADARG_INTCLAMP 0x20

SGS_APIFUNC SGSBOOL sgs_LoadArgsExtVA( SGS_CTX, int from, const char* cmd, va_list* args );
SGS_APIFUNC SGSBOOL sgs_LoadArgsExt( SGS_CTX, int from, const char* cmd, ... );
SGS_APIFUNC SGSBOOL sgs_LoadArgs( SGS_CTX, const char* cmd, ... );
SGS_APIFUNC SGSBOOL sgs_ParseMethod( SGS_CTX, sgs_ObjInterface* iface, void** ptrout,
	const char* name );
#define SGS_PARSE_METHOD( C, iface, ptr, objname, methodname ) \
	sgs_ParseMethod( C, iface, (void**) &ptr, #objname "." #methodname )

SGS_APIFUNC int sgs_ArgCheck_Object( SGS_CTX, int argid, va_list* args, int flags );

SGS_APIFUNC void sgs_Pop( SGS_CTX, sgs_StkIdx count );
SGS_APIFUNC void sgs_PopSkip( SGS_CTX, sgs_StkIdx count, sgs_StkIdx skip );

SGS_APIFUNC sgs_StkIdx sgs_StackSize( SGS_CTX );
SGS_APIFUNC void sgs_SetStackSize( SGS_CTX, sgs_StkIdx size );
SGS_APIFUNC void sgs_SetDeltaSize( SGS_CTX, sgs_StkIdx diff );
SGS_APIFUNC sgs_StkIdx sgs_AbsIndex( SGS_CTX, sgs_StkIdx item );
SGS_APIFUNC SGSBOOL sgs_IsValidIndex( SGS_CTX, sgs_StkIdx item );
SGS_APIFUNC sgs_Variable sgs_OptStackItem( SGS_CTX, sgs_StkIdx item );
SGS_APIFUNC sgs_Variable sgs_StackItem( SGS_CTX, sgs_StkIdx item );
SGS_APIFUNC SGSBOOL sgs_GetStackItem( SGS_CTX, sgs_StkIdx item, sgs_Variable* out );
SGS_APIFUNC uint32_t sgs_ItemType( SGS_CTX, sgs_StkIdx item );

SGS_APIFUNC void sgs_Assign( SGS_CTX, sgs_Variable* var_to, sgs_Variable* var_from );
SGS_APIFUNC void sgs_ArithOp( SGS_CTX, sgs_Variable* out, sgs_Variable* A, sgs_Variable* B, int op );
SGS_APIFUNC void sgs_IncDec( SGS_CTX, sgs_Variable* out, sgs_Variable* A, int inc );

/*
	OPERATIONS
*/
SGS_APIFUNC int sgs_XFCall( SGS_CTX, int args, int gotthis );
#define sgs_XCall( C, args ) sgs_XFCall( C, args, 0 )
#define sgs_XThisCall( C, args ) sgs_XFCall( C, args, 1 )
#define sgs_FCall( C, args, expect, gotthis ) sgs_AdjustStack( C, expect, sgs_XFCall( C, args, gotthis ) )
#define sgs_Call( C, args, expect ) sgs_FCall( C, args, expect, 0 )
#define sgs_ThisCall( C, args, expect ) sgs_FCall( C, args, expect, 1 )
SGS_APIFUNC SGSBOOL sgs_GlobalCall( SGS_CTX, const char* name, int args, int expect );
SGS_APIFUNC void sgs_TypeOf( SGS_CTX, sgs_Variable var );
SGS_APIFUNC void sgs_DumpVar( SGS_CTX, sgs_Variable var, int maxdepth );
SGS_APIFUNC void sgs_GCExecute( SGS_CTX );

SGS_APIFUNC const char* sgs_DebugDumpVarExt( SGS_CTX, sgs_Variable var, int maxdepth );
#define sgs_DebugDumpVar( C, var ) sgs_DebugDumpVarExt( C, var, 5 )
#define sgs_DebugPrintVar( C, var ) sgs_DebugDumpVarExt( C, var, -1 )

SGS_APIFUNC void sgs_PadString( SGS_CTX );
SGS_APIFUNC void sgs_ToPrintSafeString( SGS_CTX );
SGS_APIFUNC void sgs_StringConcat( SGS_CTX, sgs_StkIdx args );
SGS_APIFUNC void sgs_CloneItem( SGS_CTX, sgs_Variable var );

SGS_APIFUNC SGSBOOL sgs_IsArray( sgs_Variable var );
SGS_APIFUNC sgs_SizeVal sgs_ArraySize( sgs_Variable var );
SGS_APIFUNC void sgs_ArrayPush( SGS_CTX, sgs_Variable var, sgs_StkIdx count );
SGS_APIFUNC void sgs_ArrayPop( SGS_CTX, sgs_Variable var, sgs_StkIdx count, SGSBOOL ret );
SGS_APIFUNC void sgs_ArrayErase( SGS_CTX, sgs_Variable var, sgs_StkIdx at, sgs_StkIdx count );
SGS_APIFUNC sgs_SizeVal sgs_ArrayFind( SGS_CTX, sgs_Variable var, sgs_Variable what );
SGS_APIFUNC sgs_SizeVal sgs_ArrayRemove( SGS_CTX, sgs_Variable var, sgs_Variable what, SGSBOOL all );
SGS_APIFUNC SGSBOOL sgs_IsDict( sgs_Variable var );
SGS_APIFUNC SGSBOOL sgs_IsMap( sgs_Variable var );
SGS_APIFUNC SGSBOOL sgs_Unset( SGS_CTX, sgs_Variable var, sgs_Variable key );
SGS_APIFUNC SGSBOOL sgs_EventState( SGS_CTX, sgs_Variable evt, int state );

SGS_APIFUNC void sgs_SerializeExt( SGS_CTX, sgs_Variable var, int mode );
SGS_APIFUNC void sgs_SerializeObject( SGS_CTX, sgs_StkIdx args, const char* func );
SGS_APIFUNC void sgs_SerializeObjIndex( SGS_CTX, sgs_Variable key, sgs_Variable val, int isprop );
SGS_APIFUNC SGSBOOL sgs_UnserializeExt( SGS_CTX, sgs_Variable var, int mode );
#define sgs_Serialize( C, var ) sgs_SerializeExt( C, var, SGS_SERIALIZE_DEFAULT )
#define sgs_Unserialize( C, var ) sgs_UnserializeExt( C, var, SGS_SERIALIZE_DEFAULT )

SGS_APIFUNC void sgs_SerializeSGSON( SGS_CTX, sgs_Variable var, const char* tab );
SGS_APIFUNC void sgs_UnserializeSGSONExt( SGS_CTX, const char* str, size_t size );
#define sgs_UnserializeSGSON( C, str ) sgs_UnserializeSGSONExt( C, str, SGS_STRINGLENGTHFUNC( str ) )

SGS_APIFUNC int sgs_Compare( SGS_CTX, sgs_Variable* v1, sgs_Variable* v2 );
SGS_APIFUNC SGSBOOL sgs_EqualTypes( sgs_Variable* v1, sgs_Variable* v2 );

/*
	CONVERSION / RETRIEVAL
*/
/* pointer versions */
SGS_APIFUNC sgs_Bool sgs_GetBoolP( SGS_CTX, sgs_Variable* var );
SGS_APIFUNC sgs_Int sgs_GetIntP( SGS_CTX, sgs_Variable* var );
SGS_APIFUNC sgs_Real sgs_GetRealP( SGS_CTX, sgs_Variable* var );
SGS_APIFUNC void* sgs_GetPtrP( SGS_CTX, sgs_Variable* var );

SGS_APIFUNC char* sgs_ToStringBufP( SGS_CTX, sgs_Variable* var, sgs_SizeVal* outsize );
#define sgs_ToStringP( ctx, var ) sgs_ToStringBufP( ctx, var, NULL )
SGS_APIFUNC char* sgs_ToStringBufFastP( SGS_CTX, sgs_Variable* var, sgs_SizeVal* outsize );
#define sgs_ToStringFastP( ctx, var ) sgs_ToStringBufFast( ctx, var, NULL )

SGS_APIFUNC SGSBOOL sgs_IsObjectP( sgs_Variable* var, sgs_ObjInterface* iface );
#define sgs_IsTypeP( C, var, name ) sgs_IsObjectP( var, sgs_FindType( C, name ) )
SGS_APIFUNC SGSBOOL sgs_IsCallableP( sgs_Variable* var );
SGS_APIFUNC SGSBOOL sgs_ParseStringP( SGS_CTX, sgs_Variable* var, char** out, sgs_SizeVal* size );

/* index versions */
SGS_APIFUNC sgs_Bool sgs_GetBool( SGS_CTX, sgs_StkIdx item );
SGS_APIFUNC sgs_Int sgs_GetInt( SGS_CTX, sgs_StkIdx item );
SGS_APIFUNC sgs_Real sgs_GetReal( SGS_CTX, sgs_StkIdx item );
SGS_APIFUNC void* sgs_GetPtr( SGS_CTX, sgs_StkIdx item );

SGS_APIFUNC char* sgs_ToStringBuf( SGS_CTX, sgs_StkIdx item, sgs_SizeVal* outsize );
#define sgs_ToString( ctx, item ) sgs_ToStringBuf( ctx, item, NULL )
SGS_APIFUNC char* sgs_ToStringBufFast( SGS_CTX, sgs_StkIdx item, sgs_SizeVal* outsize );
#define sgs_ToStringFast( ctx, item ) sgs_ToStringBufFast( ctx, item, NULL )

SGS_APIFUNC SGSBOOL sgs_IsObject( SGS_CTX, sgs_StkIdx item, sgs_ObjInterface* iface );
#define sgs_IsType( C, item, name ) sgs_IsObject( C, item, sgs_FindType( C, name ) )
SGS_APIFUNC SGSBOOL sgs_IsCallable( SGS_CTX, sgs_StkIdx item );
SGS_APIFUNC SGSBOOL sgs_ParseBool( SGS_CTX, sgs_StkIdx item, sgs_Bool* out );
SGS_APIFUNC SGSBOOL sgs_ParseInt( SGS_CTX, sgs_StkIdx item, sgs_Int* out );
SGS_APIFUNC SGSBOOL sgs_ParseReal( SGS_CTX, sgs_StkIdx item, sgs_Real* out );
SGS_APIFUNC SGSBOOL sgs_ParseString( SGS_CTX, sgs_StkIdx item, char** out, sgs_SizeVal* size );
SGS_APIFUNC SGSBOOL sgs_ParseObjectPtr( SGS_CTX, sgs_StkIdx item,
	sgs_ObjInterface* iface, sgs_VarObj** out, int strict );
SGS_APIFUNC SGSBOOL sgs_ParsePtr( SGS_CTX, sgs_StkIdx item, void** out );

/* global versions */
SGS_APIFUNC sgs_Bool sgs_GlobalBool( SGS_CTX, const char* name );
SGS_APIFUNC sgs_Int sgs_GlobalInt( SGS_CTX, const char* name );
SGS_APIFUNC sgs_Real sgs_GlobalReal( SGS_CTX, const char* name );
SGS_APIFUNC char* sgs_GlobalStringBuf( SGS_CTX, const char* name, sgs_SizeVal* outsize );
#define sgs_GlobalString( C, name ) sgs_GlobalStringBuf( C, name, NULL )

/* iterator interface fns */
SGS_APIFUNC SGSBOOL sgs_PushIterator( SGS_CTX, sgs_Variable var );
SGS_APIFUNC SGSBOOL sgs_GetIterator( SGS_CTX, sgs_Variable var, sgs_Variable* out );
SGS_APIFUNC SGSBOOL sgs_IterAdvance( SGS_CTX, sgs_Variable var );
SGS_APIFUNC void sgs_IterPushData( SGS_CTX, sgs_Variable var, int key, int value );
SGS_APIFUNC void sgs_IterGetData( SGS_CTX, sgs_Variable var, sgs_Variable* key, sgs_Variable* value );

SGS_APIFUNC SGSBOOL sgs_IsNumericString( const char* str, sgs_SizeVal size );

/*
	EXTENSION UTILITIES
*/
SGS_APIFUNC SGSBOOL sgs_Method( SGS_CTX );
SGS_APIFUNC SGSBOOL sgs_HideThis( SGS_CTX );
SGS_APIFUNC SGSBOOL sgs_ForceHideThis( SGS_CTX );
SGS_APIFUNC int sgs_ObjectArg( SGS_CTX );

SGS_APIFUNC void sgs_Acquire( SGS_CTX, sgs_Variable* var );
SGS_APIFUNC void sgs_AcquireArray( SGS_CTX, sgs_Variable* var, sgs_SizeVal count );
SGS_APIFUNC void sgs_Release( SGS_CTX, sgs_Variable* var );
SGS_APIFUNC void sgs_ReleaseArray( SGS_CTX, sgs_Variable* var, sgs_SizeVal count );
SGS_APIFUNC void sgs_GCMark( SGS_CTX, sgs_Variable* var );
SGS_APIFUNC void sgs_GCMarkArray( SGS_CTX, sgs_Variable* var, sgs_SizeVal count );
SGS_APIFUNC void sgs_ObjAcquire( SGS_CTX, sgs_VarObj* obj );
SGS_APIFUNC void sgs_ObjRelease( SGS_CTX, sgs_VarObj* obj );
SGS_APIFUNC void sgs_ObjGCMark( SGS_CTX, sgs_VarObj* obj );
SGS_APIFUNC void sgs_ObjAssign( SGS_CTX, sgs_VarObj** dest, sgs_VarObj* src );
SGS_APIFUNC void sgs_ObjCallDtor( SGS_CTX, sgs_VarObj* obj );
SGS_APIFUNC void sgs_ObjSetMetaObj( SGS_CTX, sgs_VarObj* obj, sgs_VarObj* metaobj );
SGS_APIFUNC sgs_VarObj* sgs_ObjGetMetaObj( sgs_VarObj* obj );
SGS_APIFUNC void sgs_ObjSetMetaMethodEnable( sgs_VarObj* obj, SGSBOOL enable );
SGS_APIFUNC SGSBOOL sgs_ObjGetMetaMethodEnable( sgs_VarObj* obj );

SGS_APIFUNC char* sgs_GetStringPtrP( sgs_Variable* var );
SGS_APIFUNC sgs_SizeVal sgs_GetStringSizeP( sgs_Variable* var );
SGS_APIFUNC sgs_VarObj* sgs_GetObjectStructP( sgs_Variable* var );
SGS_APIFUNC void* sgs_GetObjectDataP( sgs_Variable* var );
SGS_APIFUNC sgs_ObjInterface* sgs_GetObjectIfaceP( sgs_Variable* var );
SGS_APIFUNC void sgs_SetObjectDataP( sgs_Variable* var, void* data );
SGS_APIFUNC void sgs_SetObjectIfaceP( sgs_Variable* var, sgs_ObjInterface* iface );

SGS_APIFUNC char* sgs_GetStringPtr( SGS_CTX, sgs_StkIdx item );
SGS_APIFUNC sgs_SizeVal sgs_GetStringSize( SGS_CTX, sgs_StkIdx item );
SGS_APIFUNC sgs_VarObj* sgs_GetObjectStruct( SGS_CTX, sgs_StkIdx item );
SGS_APIFUNC void* sgs_GetObjectData( SGS_CTX, sgs_StkIdx item );
SGS_APIFUNC sgs_ObjInterface* sgs_GetObjectIface( SGS_CTX, sgs_StkIdx item );
SGS_APIFUNC void sgs_SetObjectData( SGS_CTX, sgs_StkIdx item, void* data );
SGS_APIFUNC void sgs_SetObjectIface( SGS_CTX, sgs_StkIdx item, sgs_ObjInterface* iface );

SGS_APIFUNC int sgs_HasFuncName( SGS_CTX );
SGS_APIFUNC void sgs_FuncName( SGS_CTX, const char* fnliteral );

#define SGSFN( x ) sgs_FuncName( C, x )
#define SGSBASEFN( x ) if( !sgs_HasFuncName( C ) ) sgs_FuncName( C, x )

static SGS_INLINE int sgs_Errno( SGS_CTX, int clear )
{
	sgs_Cntl( C, SGS_CNTL_ERRNO, clear );
	return clear;
}
#define sgs_SetErrno( C, err ) sgs_Cntl( C, SGS_CNTL_SET_ERRNO, err )
#define sgs_GetLastErrno( C ) sgs_Cntl( C, SGS_CNTL_GET_ERRNO, 0 )


/* predefined output / messaging functions */

static SGS_INLINE void sgs_StdOutputFunc( void* userdata, SGS_CTX, const void* ptr, size_t size )
{
	SGS_UNUSED( C );
	fwrite( ptr, 1, size, (FILE*) userdata );
}

static SGS_INLINE void sgs_StdMsgFunc_NoAbort( void* ctx, SGS_CTX, int type, const char* msg )
{
	SGS_UNUSED( ctx );
	sgs_WriteErrorInfo( C, SGS_ERRORINFO_FULL, (sgs_ErrorOutputFunc) sgs_ErrWritef, C, type, msg );
}

static SGS_INLINE void sgs_StdMsgFunc( void* ctx, SGS_CTX, int type, const char* msg )
{
	sgs_StdMsgFunc_NoAbort( ctx, C, type, msg );
	if( type >= SGS_ERROR )
		sgs_Abort( C );
}

static SGS_INLINE SGSRESULT sgs_StdScriptFSFunc( void* ctx, SGS_CTX, int op, sgs_ScriptFSData* data )
{
	SGS_UNUSED( ctx );
	SGS_UNUSED( C );
	switch( op )
	{
	case SGS_SFS_FILE_EXISTS:
		{
			FILE* f = fopen( data->filename, "rb" );
			if( f )
				fclose( f );
			return f ? SGS_SUCCESS : SGS_ENOTFND;
		}
	case SGS_SFS_FILE_OPEN:
		{
			long sz;
			FILE* f = fopen( data->filename, "rb" );
			if( f == NULL )
				return SGS_ENOTFND;
			fseek( f, 0, SEEK_END );
			sz = ftell( f );
			if( sz < 0 )
			{
				fclose( f );
				return SGS_EINPROC;
			}
			data->userhandle = f;
			data->size = (size_t) sz;
			return SGS_SUCCESS;
		}
	case SGS_SFS_FILE_READ:
		fseek( (FILE*) data->userhandle, 0, SEEK_SET );
		return fread( data->output, 1, data->size, (FILE*) data->userhandle ) == data->size ? SGS_SUCCESS : SGS_EINPROC;
	case SGS_SFS_FILE_CLOSE:
		fclose( (FILE*) data->userhandle );
		data->userhandle = NULL;
		return SGS_SUCCESS;
	}
	return SGS_ENOTSUP;
}


/* utility macros */
#define SGS_MODULE_CHECK_VERSION( C ) if( SGS_VERSION_INT != sgs_Stat( C, SGS_STAT_VERSION ) ) \
	{ sgs_Msg( C, SGS_ERROR, "SGScript version mismatch: module compiled for %06X, loaded in %06X", \
	(int) SGS_VERSION_INT, (int) sgs_Stat( C, SGS_STAT_VERSION ) ); return SGS_ENOTSUP; }

#define SGS_RETURN_THIS( C ) sgs_Method( C ); sgs_SetStackSize( C, 1 ); return 1;

#define SGS_ARGS_OBJFUNC SGS_CTX, sgs_VarObj* obj
#define SGS_ARGS_GETINDEXFUNC SGS_CTX, sgs_VarObj* obj
#define SGS_ARGS_SETINDEXFUNC SGS_CTX, sgs_VarObj* obj
#define SGS_BEGIN_INDEXFUNC char* str; if( sgs_ParseString( C, 0, &str, NULL ) ){
#define SGS_END_INDEXFUNC } return SGS_ENOTFND;
#define SGS_CASE( name ) if( !strcmp( str, name ) )


#ifdef __cplusplus
}
#endif

#endif /* SGSCRIPT_H_INCLUDED */


#ifndef SGS_UTIL_H_INCLUDED
#define SGS_UTIL_H_INCLUDED

#ifdef __cplusplus
extern "C" {
#endif




/* useful shortcut definitions */
#define SGS_MAX( a, b ) ((a)>(b)?(a):(b))
#define SGS_MIN( a, b ) ((a)<(b)?(a):(b))
#define SGS_ARRAY_SIZE( a ) (sizeof(a)/sizeof(a[0]))
#define SGS_HAS_FLAG( val, flag ) (((val)&(flag))==(flag))
#define SGS_HAS_ANY_FLAG( val, flag ) (((val)&(flag))!=0)

#define SGS_GLUE(a,b) __SGS_GLUE(a,b)
#define __SGS_GLUE(a,b) a ## b
#define SGS_CASSERT(expr, msg) typedef char SGS_GLUE (compiler_verify_, msg) [(expr) ? (+1) : (-1)]


typedef int16_t sgs_LineNum;

SGS_APIFUNC void sgs_BreakIfFunc( const char* code, const char* file, int line );
#if SGS_DEBUG && SGS_DEBUG_VALIDATE
#  define sgs_BreakIf( expr ) { if( (expr) != 0 ){ sgs_BreakIfFunc( #expr, __FILE__, __LINE__ ); } }
#else
#  define sgs_BreakIf( expr )
#endif


/* text/character ops */
int sgs_isoneof( int chr, const char* str );
SGS_APIFUNC int sgs_hexchar( int c );
SGS_APIFUNC int sgs_gethex( int c );
int sgs_tolower( int c );
#define sgs_decchar( c ) ((c) >= '0' && (c) <= '9')
#define sgs_getdec( c ) ((c) - '0')
#define sgs_octchar( c ) ((c) >= '0' && (c) <= '7')
#define sgs_getoct( c ) ((c) - '0')
#define sgs_binchar( c ) ((c) == '0' || (c) == '1')
#define sgs_getbin( c ) ((c) - '0')
#define sgs_isalpha( c ) (((c) >= 'a' && (c) <= 'z') || ((c) >= 'A' && (c) <= 'Z' ))
#define sgs_isalnum( c ) (((c) >= 'a' && (c) <= 'z') || \
	((c) >= 'A' && (c) <= 'Z') || ((c) >= '0' && (c) <= '9'))
#define sgs_isdigit( c ) ((c) >= '0' && (c) <= '9')
#define sgs_isgraph( c ) ((c) >= 0x21 && (c) <= 0x7E)


#define SGS_AS_( tgt, ptr, wat ) do{ memcpy( &(tgt), (ptr), sizeof(wat) ); }while(0)
#define SGS_AS_INT8( tgt, ptr ) SGS_AS_( tgt, ptr, int8_t )
#define SGS_AS_UINT8( tgt, ptr ) SGS_AS_( tgt, ptr, uint8_t )
#define SGS_AS_INT16( tgt, ptr ) SGS_AS_( tgt, ptr, int16_t )
#define SGS_AS_UINT16( tgt, ptr ) SGS_AS_( tgt, ptr, uint16_t )
#define SGS_AS_INT32( tgt, ptr ) SGS_AS_( tgt, ptr, int32_t )
#define SGS_AS_UINT32( tgt, ptr ) SGS_AS_( tgt, ptr, uint32_t )
#define SGS_AS_INT64( tgt, ptr ) SGS_AS_( tgt, ptr, int64_t )
#define SGS_AS_UINT64( tgt, ptr ) SGS_AS_( tgt, ptr, uint64_t )
#define SGS_AS_FLOAT( tgt, ptr ) SGS_AS_( tgt, ptr, float )
#define SGS_AS_DOUBLE( tgt, ptr ) SGS_AS_( tgt, ptr, double )

#define SGS_AS_INTEGER( tgt, ptr ) SGS_AS_( tgt, ptr, sgs_Int )
#define SGS_AS_REAL( tgt, ptr ) SGS_AS_( tgt, ptr, sgs_Real )


/* flow/data debugging */
#if SGS_DEBUG && SGS_DEBUG_FLOW
#  define SGS_FN_HIT( what ) \
	printf( "Hit \"%s\" line %d in function \"%s\"\n", what, __LINE__, __FUNCTION__ );
#  define SGS_FN_ENTER \
	printf( "Entering a function from \"%s\" at line %d\n", __FUNCTION__, __LINE__ );
#  define SGS_FN_BEGIN \
	printf( "Inside \"%s\"\n", __FUNCTION__ );
#  define SGS_FN_END \
	printf( "Out of \"%s\" at line %d\n", __FUNCTION__, __LINE__ );
#else
#  define SGS_FN_HIT( what )
#  define SGS_FN_ENTER
#  define SGS_FN_BEGIN
#  define SGS_FN_END
#endif

SGS_APIFUNC void sgs_print_safe( FILE* fp, const char* buf, size_t size );


/* string buffer */
typedef
struct _sgs_MemBuf
{
	char*  ptr;
	size_t size;
	size_t mem;
}
sgs_MemBuf;


/* data buffer */
SGS_APIFUNC sgs_MemBuf sgs_membuf_create( void );
SGS_APIFUNC void sgs_membuf_destroy( sgs_MemBuf* sb, SGS_CTX );
SGS_APIFUNC sgs_MemBuf sgs_membuf_partial( char* ch, size_t size );
SGS_APIFUNC void sgs_membuf_reserve( sgs_MemBuf* mb, SGS_CTX, size_t size );
SGS_APIFUNC void sgs_membuf_resize( sgs_MemBuf* mb, SGS_CTX, size_t size );
SGS_APIFUNC void sgs_membuf_resize_opt( sgs_MemBuf* mb, SGS_CTX, size_t size );
SGS_APIFUNC void sgs_membuf_insbuf( sgs_MemBuf* mb, SGS_CTX, size_t pos, const void* buf, size_t size );
SGS_APIFUNC void sgs_membuf_erase( sgs_MemBuf* mb, size_t from, size_t to );
SGS_APIFUNC void sgs_membuf_appbuf( sgs_MemBuf* mb, SGS_CTX, const void* buf, size_t size );
static SGS_INLINE void sgs_membuf_setstr( sgs_MemBuf* mb, SGS_CTX, const char* str )
	{ mb->size = 0; sgs_membuf_appbuf( mb, C, str, strlen( str ) + 1 ); mb->size--; }
static SGS_INLINE void sgs_membuf_setstrbuf( sgs_MemBuf* mb, SGS_CTX, const char* str, size_t size )
	{ sgs_membuf_reserve( mb, C, size + 1 ); mb->size = 0;
		sgs_membuf_appbuf( mb, C, str, size ); mb->ptr[ mb->size ] = 0; }
static SGS_INLINE void sgs_membuf_appchr( sgs_MemBuf* mb, SGS_CTX, char chr )
	{ sgs_membuf_appbuf( mb, C, &chr, 1 ); }


/* hashing functions */
typedef uint32_t sgs_Hash;
SGS_APIFUNC sgs_Hash sgs_HashFunc( const char* str, size_t size );
SGS_APIFUNC sgs_Hash sgs_HashVar( const sgs_Variable* v );


/* hash table */
typedef sgs_SizeVal sgs_VHTIdx;

#define SGS_VHTIDX_EMPTY -1
#define SGS_VHTIDX_REMOVED -2

typedef
struct _sgs_VHTVar
{
	sgs_Variable key;
	sgs_Variable val;
	sgs_Hash     hash;
}
sgs_VHTVar;

typedef
struct _sgs_VHTable
{
	sgs_VHTIdx* pairs;
	sgs_VHTVar* vars;
	sgs_VHTIdx  pair_mem;
	sgs_VHTIdx  var_mem;
	sgs_VHTIdx  size;
	sgs_VHTIdx  num_rem;
}
sgs_VHTable;

typedef struct sgs_VHTStats
{
	/* in */
	int print;
	int print_ubmp; /* usage bitmap */
	int print_cols;
	/* out */
	unsigned buckets;
	unsigned used;
	unsigned empty;
	unsigned removed;
	unsigned collisions;
	unsigned worst_probe_length;
	float avg_probe_length;
}
sgs_VHTStats;

SGS_APIFUNC void sgs_vht_init( sgs_VHTable* T, SGS_CTX, int32_t initial_pair_mem, int32_t initial_var_mem );
SGS_APIFUNC void sgs_vht_free( sgs_VHTable* T, SGS_CTX );
SGS_APIFUNC void sgs_vht_analyze( sgs_VHTable* T, sgs_VHTStats* io );
SGS_APIFUNC sgs_VHTIdx sgs_vht_pair_id( sgs_VHTable* T, sgs_Variable* K, sgs_Hash hash );
SGS_APIFUNC sgs_VHTVar* sgs_vht_get( sgs_VHTable* T, sgs_Variable* K );
SGS_APIFUNC sgs_VHTVar* sgs_vht_get_str( sgs_VHTable* T, const char* str, uint32_t size, sgs_Hash hash );
SGS_APIFUNC sgs_VHTVar* sgs_vht_set( sgs_VHTable* T, SGS_CTX, sgs_Variable* K, sgs_Variable* V );
SGS_APIFUNC void sgs_vht_unset( sgs_VHTable* T, SGS_CTX, sgs_Variable* K );

#define sgs_vht_size( T ) ((T)->size)


SGS_APIFUNC double sgs_GetTime();


/* returns 0 on failure, 1/2 on integer/real */
SGS_APIFUNC int sgs_util_strtonum( const char** at, const char* end, sgs_Int* outi, sgs_Real* outf );
SGS_APIFUNC sgs_Int sgs_util_atoi( const char* str, size_t len );
SGS_APIFUNC sgs_Real sgs_util_atof( const char* str, size_t len );


SGS_APIFUNC void sgs_quicksort( void *array, size_t length, size_t size,
	int(*compare)(const void *, const void *, void*), void* userdata);

/*
	UNICODE helper functions
	- utf8_decode: returns number of bytes parsed (negated if input was invalid)
	- utf8_encode: returns number of bytes written (up to 4, make sure there's space)
*/
#define SGS_UNICODE_INVCHAR 0xfffd
#define SGS_UNICODE_INVCHAR_STR "\xef\xbf\xbd"
#define SGS_UNICODE_INVCHAR_LEN 3
SGS_APIFUNC int sgs_utf8_decode( char* buf, size_t size, uint32_t* outchar );
SGS_APIFUNC int sgs_utf8_encode( uint32_t ch, char* out );


#ifdef __cplusplus
}
#endif

#endif /* SGS_UTIL_H_INCLUDED */



/*
	L I M I T S
	
	! some limits may cancel or reduce others, ..
	.. these are kept for the purposes of knowing the size of variables ..
	.. to use in various systems
	
	- number of constants in a function: 0 - 65535
	- number of instructions in a function: 0 - 65535
	- string size: 0 - 2^31-1
	- max. jump distance: 32767 instructions
	- max. loop depth: 65535 nested loops
		(no need to test for it since instruction count ..
		.. effectively limits this further)
	- bytecode size limit: 14 - 2^32-1
	- source size limit: 0 - 2^31-1
	- variable size limit: string size + overhead (<=256 bytes)
	- item type value limits: 8 bits
	
	- identifier size: 0 - 255
	- argument count: 0 - 255
	- returned value count: 0 - 255
	- total closure count in one function: 0 - 255
	- temporary variable count (incl. args): 0 - 255
	- useful line count limits in source file: 0 - 32767
	
	- stack size: 0 - 2^31-1
	- closure stack size: 0 - 2^31-1
	- array size: 0 - 2^31-1
	- hash table size: 0 - 2^31-1
*/


#ifdef __cplusplus
extern "C" {
#endif



/*
	Token stream
	<1 byte: type> <additional data> <2 bytes: line number>
*/
/*     special          id */
#define SGS_ST_NULL	    '\0'
#define SGS_ST_RBRKL    '('
#define SGS_ST_RBRKR    ')'
#define SGS_ST_SBRKL    '['
#define SGS_ST_SBRKR    ']'
#define SGS_ST_CBRKL    '{'
#define SGS_ST_CBRKR    '}'
#define SGS_ST_ARGSEP   ','
#define SGS_ST_STSEP    ';'
#define SGS_ST_PICKSEP  ':'
/*     other            id    additional data */
#define SGS_ST_IDENT    'N' /* 1 byte (string size), N bytes (string), not null-terminated */
#define SGS_ST_KEYWORD  'K' /* same as IDENT */
#define SGS_ST_NUMREAL  'R' /* 8 bytes (double) */
#define SGS_ST_NUMINT   'I' /* 8 bytes (int64) */
#define SGS_ST_STRING   'S' /* 4 bytes (string size), N bytes (string), not null-terminated */
/*     operators        id    type  */
#define SGS_ST_OP_SEQ   200 /* ===  */
#define SGS_ST_OP_SNEQ  201 /* !==  */
#define SGS_ST_OP_EQ    202 /* ==   */
#define SGS_ST_OP_NEQ   203 /* !=   */
#define SGS_ST_OP_LEQ   204 /* <=   */
#define SGS_ST_OP_GEQ   205 /* >=   */
#define SGS_ST_OP_LESS  206 /* <    */
#define SGS_ST_OP_GRTR  207 /* >    */
#define SGS_ST_OP_RWCMP 208 /* <=>  */
#define SGS_ST_OP_ADDEQ 209 /* +=   */
#define SGS_ST_OP_SUBEQ 210 /* -=   */
#define SGS_ST_OP_MULEQ 211 /* *=   */
#define SGS_ST_OP_DIVEQ 212 /* /=   */
#define SGS_ST_OP_MODEQ 213 /* %=   */
#define SGS_ST_OP_ANDEQ 214 /* &=   */
#define SGS_ST_OP_OREQ  215 /* |=   */
#define SGS_ST_OP_XOREQ 216 /* ^=   */
#define SGS_ST_OP_LSHEQ 217 /* <<=  */
#define SGS_ST_OP_RSHEQ 218 /* >>=  */
#define SGS_ST_OP_BLAEQ 219 /* &&=  */
#define SGS_ST_OP_BLOEQ 220 /* ||=  */
#define SGS_ST_OP_NLOEQ 221 /* ??=  */
#define SGS_ST_OP_CATEQ 222 /* $=   */
#define SGS_ST_OP_SET   223 /* =    */
#define SGS_ST_OP_ERSUP 224 /* @    */
#define SGS_ST_OP_BLAND 225 /* &&   */
#define SGS_ST_OP_BLOR  226 /* ||   */
#define SGS_ST_OP_NLOR  227 /* ??   */
#define SGS_ST_OP_ADD   228 /* +    */
#define SGS_ST_OP_SUB   229 /* -    */
#define SGS_ST_OP_MUL   230 /* *    */
#define SGS_ST_OP_DIV   231 /* /    */
#define SGS_ST_OP_MOD   232 /* %    */
#define SGS_ST_OP_AND   233 /* &    */
#define SGS_ST_OP_OR    234 /* |    */
#define SGS_ST_OP_XOR   235 /* ^    */
#define SGS_ST_OP_LSH   236 /* <<   */
#define SGS_ST_OP_RSH   237 /* >>   */
#define SGS_ST_OP_MMBR  238 /* .    */
#define SGS_ST_OP_CAT   239 /* $    */
#define SGS_ST_OP_NOT   240 /* !    */
#define SGS_ST_OP_INV   241 /* ~    */
#define SGS_ST_OP_INC   242 /* ++   */
#define SGS_ST_OP_DEC   243 /* --   */

#define SGS_ST_ISOP( chr )      ( (chr) >= 200 && (chr) <= 243 )
#define SGS_ST_OP_UNARY( chr )  ( (chr) == SGS_ST_OP_ERSUP || (chr) == SGS_ST_OP_ADD || \
	(chr) == SGS_ST_OP_SUB || ( (chr) >= SGS_ST_OP_NOT && (chr) <= SGS_ST_OP_DEC ) )
#define SGS_ST_OP_BINARY( chr ) ( (chr) >= 200 && (chr) <= 239 && (chr) != 224 )
#define SGS_ST_OP_ASSIGN( chr ) ( (chr) >= 209 && (chr) <= 223 )
#define SGS_ST_OP_BINMUL( chr ) ( (chr) == SGS_ST_OP_MUL || (chr) == SGS_ST_OP_DIV || (chr) == SGS_ST_OP_MOD )
#define SGS_ST_OP_BINADD( chr ) ( (chr) == SGS_ST_OP_ADD || (chr) == SGS_ST_OP_SUB )
#define SGS_ST_OP_BINOPS( chr ) ( (chr) >= SGS_ST_OP_AND && (chr) <= SGS_ST_OP_RSH )
#define SGS_ST_OP_COMP( chr )   ( (chr) >= 200 && (chr) <= 208 )
#define SGS_ST_OP_BOOL( chr )   ( (chr) == SGS_ST_OP_BLAEQ || (chr) == SGS_ST_OP_BLOEQ || \
	(chr) == SGS_ST_OP_BLAND || (chr) == SGS_ST_OP_BLOR )
#define SGS_ST_OP_FNN( chr )    ( (chr) == SGS_ST_OP_NLOEQ || (chr) == SGS_ST_OP_NLOR )

#define SGS_ST_ISSPEC( chr )    sgs_isoneof( (chr), "()[]{},;:" )

#define SGS_ST_READINT( tgt, pos )   SGS_AS_INT32( tgt, pos )
#define SGS_ST_READLN( tgt, pos )    SGS_AS_( tgt, pos, sgs_LineNum )


typedef unsigned char sgs_TokenType;
typedef unsigned char* sgs_TokenList;

SGS_APIFUNC int sgsT_IsKeyword( sgs_TokenList tok, const char* text );
SGS_APIFUNC int sgsT_IsIdent( sgs_TokenList tok, const char* text );

SGS_APIFUNC sgs_TokenList sgsT_Gen( SGS_CTX, const char* code, size_t length );
SGS_APIFUNC void          sgsT_Free( SGS_CTX, sgs_TokenList tlist );
SGS_APIFUNC sgs_TokenList sgsT_Next( sgs_TokenList tok );
SGS_APIFUNC sgs_LineNum   sgsT_LineNum( sgs_TokenList tok );

SGS_APIFUNC size_t sgsT_ListSize( sgs_TokenList tlist );
SGS_APIFUNC size_t sgsT_ListMemSize( sgs_TokenList tlist );

SGS_APIFUNC void sgsT_TokenString( SGS_CTX, sgs_MemBuf* out, sgs_TokenList tlist, sgs_TokenList tend, int xs );

SGS_APIFUNC void sgsT_DumpToken( sgs_TokenList tok );
SGS_APIFUNC void sgsT_DumpList( sgs_TokenList tlist, sgs_TokenList tend );


/*
	Function tree
*/

#define SGS_SFT_NULL    0
/* data */
#define SGS_SFT_IDENT   1
#define SGS_SFT_KEYWORD 2
#define SGS_SFT_CONST   3
/* expression parts */
#define SGS_SFT_OPER    4
#define SGS_SFT_OPER_P  5 /* post-op (inc/dec) */
#define SGS_SFT_FCALL   6
#define SGS_SFT_INDEX   7
#define SGS_SFT_MIDXSET 8
#define SGS_SFT_MPROPSET 9
/* statement data */
#define SGS_SFT_ARGMT   10
#define SGS_SFT_ARGLIST 11
#define SGS_SFT_VARLIST 12
#define SGS_SFT_GVLIST  13
#define SGS_SFT_USELIST 14
#define SGS_SFT_EXPLIST 15
#define SGS_SFT_ARRLIST 16
#define SGS_SFT_DCTLIST 17
#define SGS_SFT_MAPLIST 18
#define SGS_SFT_RETURN  19
/* statement types */
#define SGS_SFT_BLOCK   21
#define SGS_SFT_IFELSE  22
#define SGS_SFT_WHILE   23
#define SGS_SFT_DOWHILE 24
#define SGS_SFT_FOR     25
#define SGS_SFT_FOREACH 26
#define SGS_SFT_BREAK   27
#define SGS_SFT_CONT    28
#define SGS_SFT_DEFER   29
#define SGS_SFT_FUNC    30
#define SGS_SFT_CLASS   31
#define SGS_SFT_CLSINH  32
#define SGS_SFT_CLSINC  33
#define SGS_SFT_CLSGLOB 34
#define SGS_SFT_CLSPFX  35
#define SGS_SFT_NEWCALL 36 /* new */
/* threading addons */
#define SGS_SFT_THRCALL 40 /* thread */
#define SGS_SFT_STHCALL 41 /* subthread */
#define SGS_SFT_HEAPBIT 255

typedef struct sgs_FTNode sgs_FTNode;
struct sgs_FTNode
{
	sgs_TokenList token;
	sgs_FTNode*   next;
	sgs_FTNode*   child;
	int           type;
};

SGS_APIFUNC void sgsFT_Destroy( SGS_CTX, sgs_FTNode* tree );

SGS_APIFUNC sgs_FTNode* sgsFT_Compile( SGS_CTX, sgs_TokenList tlist );
SGS_APIFUNC void sgsFT_Dump( sgs_FTNode* tree );



/* - bytecode generator */
SGS_APIFUNC sgs_iFunc* sgsBC_Generate( SGS_CTX, sgs_FTNode* tree );
SGS_APIFUNC void sgsBC_DumpEx( const char* constptr, size_t constsize,
	const char* codeptr, size_t codesize );


/*
	Serialized bytecode
*/
SGS_APIFUNC int sgsBC_Func2Buf( SGS_CTX, sgs_iFunc* func, sgs_MemBuf* outbuf );

/* assumes headers have already been validated (except size) but are still in the buffer */
SGS_APIFUNC const char* sgsBC_Buf2Func( SGS_CTX, const char* fn,
	const char* buf, size_t size, sgs_Variable* outfunc );

/* validates header size and bytes one by one (except last flag byte)
-- will return header_size on success and failed byte position on failure */
SGS_APIFUNC int sgsBC_ValidateHeader( const char* buf, size_t size );
#define SGS_HEADER_SIZE 14
#define SGS_MIN_BC_SIZE 8 + SGS_HEADER_SIZE
#define SGSBC_FLAG_LITTLE_ENDIAN 0x01


/*
	Virtual Machine
*/

#define SGS_CONSTVAR( x ) ((x)&0x100)
#define SGS_CONSTENC( x ) ((x)|0x100)
#define SGS_CONSTDEC( x ) ((x)&0xff)

#define SGS_INT_ERRSUP_INC 1
#define SGS_INT_ERRSUP_DEC 2
#define SGS_INT_RESET_WAIT_TIMER 3

typedef enum sgs_Instruction
{
	SGS_SI_NOP = 0,

	SGS_SI_PUSH,     /* (B:src)                 push register/constant */
	SGS_SI_INT,      /* (C:type)                invoke a system state change */

	SGS_SI_RET1,     /* (C:src)                 return one value (see next op) */
	SGS_SI_RETN,     /* (A:N)                   exit current frame of execution, preserve N output arguments */
	SGS_SI_JUMP,     /* (E:off)                 add to instruction pointer */
	SGS_SI_JMPT,     /* (C:src, E:off)          jump (add to instr.ptr.) if true */
	SGS_SI_JMPF,     /* (C:src, E:off)          jump (add to instr.ptr.) if false */
	SGS_SI_JMPN,     /* (C:src, E:off)          jump (add to instr.ptr.) if null */
	SGS_SI_CALL,     /* (A:exp, B:from, C:to)   call a variable */

	SGS_SI_FORPREP,  /* (A:out, B:src)          retrieves the iterator to work the object */
	SGS_SI_FORLOAD,  /* (A:iter, B:oky, C:ovl)  retrieves current key & value from iterator */
	SGS_SI_FORJUMP,  /* (C:iter, E:off)         advances and jumps if next key exists */

	SGS_SI_LOADCONST,/* (C:out, E:off)          load a constant to a register */
	SGS_SI_GETVAR,   /* (A:out, B:name)         <varname> => <value> */
	SGS_SI_SETVAR,   /* (B:name, C:src)         <varname> <value> => set <value> to <varname> */
	SGS_SI_GETPROP,  /* (A:out, B:var, C:name)  <var> <prop> => <var> */
	SGS_SI_SETPROP,  /* (A:var, B:name, C:src)  <var> <prop> <value> => set a <prop> of <var> to <value> */
	SGS_SI_GETINDEX, /* -- || -- */
	SGS_SI_SETINDEX, /* -- || -- */
	SGS_SI_ARRPUSH,  /* (A:var, C:src)          push <src> into <var> */

	/* closures */
	SGS_SI_CLSRINFO, /* (metadata)              indices of closure variables to import */
	SGS_SI_MAKECLSR, /* (A:out, B:from, C:N)    creates a closure object containing closure vars */
	SGS_SI_GETCLSR,  /* (A:out, B:which)        loads data from the specified closure */
	SGS_SI_SETCLSR,  /* (B:which, C:src)        stores data to the specified closure */

	/* operators */
	/*
		A: (A:out, B:s1, C:s2)
		B: (A:out, B:s1)
	*/
	SGS_SI_SET,      /* B */
	SGS_SI_MCONCAT,  /* (A:out, B:N) */
	SGS_SI_CONCAT,   /* A */
	SGS_SI_NEGATE,   /* B */
	SGS_SI_BOOL_INV,
	SGS_SI_INVERT,

	SGS_SI_INC,      /* B */
	SGS_SI_DEC,
	SGS_SI_ADD,      /* A */
	SGS_SI_SUB,
	SGS_SI_MUL,
	SGS_SI_DIV,
	SGS_SI_MOD,

	SGS_SI_AND,      /* A */
	SGS_SI_OR,
	SGS_SI_XOR,
	SGS_SI_LSH,
	SGS_SI_RSH,

	SGS_SI_SEQ,      /* A */
	SGS_SI_SNEQ,
	SGS_SI_EQ,
	SGS_SI_NEQ,
	SGS_SI_LT,
	SGS_SI_GTE,
	SGS_SI_GT,
	SGS_SI_LTE,
	SGS_SI_RAWCMP,

	/* specials */
	SGS_SI_ARRAY,    /* (C:out, E:args) */
	SGS_SI_DICT,     /* (C:out) */
	SGS_SI_MAP,      /* (C:out) */
	SGS_SI_CLASS,    /* (A:out, B:name, C:inhname) */
	SGS_SI_NEW,      /* (B:out/class, C:lastarg) */
	SGS_SI_RSYM,     /* (B:name, C:var)         performs dual registration to symbol table */
	SGS_SI_COTRT,    /* (A:to, B:from)          sets A to true if `from` is finished */
	SGS_SI_COTRF,    /* (A:to, B:from)          sets A to false if `from` is not finished */
	SGS_SI_COABORT,  /* (C:arg, E:from)         if arg = true, look for COTR* and abort threads */
	SGS_SI_YLDJMP,   /* (C:arg, E:off)          if arg = false, yield and jump */

	SGS_SI_COUNT
}
sgs_Instruction;


typedef uint32_t sgs_instr_t;

/*
	instruction data: 32 bits
	OOOOOOAAAAAAAABBBBBBBBBCCCCCCCCC
	OOOOOOEEEEEEEEEEEEEEEEECCCCCCCCC
*/

#define SGS_INSTR_SIZE    4

#define SGS_INSTR_OFF_OP  0
#define SGS_INSTR_OFF_A   6
#define SGS_INSTR_OFF_B   14
#define SGS_INSTR_OFF_C   23
#define SGS_INSTR_OFF_E   6

#define SGS_INSTR_MASK_OP 0x003f /* OP: 6 bits */
#define SGS_INSTR_MASK_A  0x00ff /* A:  8 bits */
#define SGS_INSTR_MASK_B  0x01ff /* B:  9 bits */
#define SGS_INSTR_MASK_C  0x01ff /* C:  9 bits */
#define SGS_INSTR_MASK_E  0x0001ffff /* E: 17 bits */

#define SGS_INSTR_GET_OP( x )  ((int) (((x) >> SGS_INSTR_OFF_OP) & SGS_INSTR_MASK_OP))
#define SGS_INSTR_GET_A( x )   ((int) (((x) >> SGS_INSTR_OFF_A ) & SGS_INSTR_MASK_A ))
#define SGS_INSTR_GET_B( x )   ((int) (((x) >> SGS_INSTR_OFF_B ) & SGS_INSTR_MASK_B ))
#define SGS_INSTR_GET_C( x )   ((int) (((x) >> SGS_INSTR_OFF_C ) & SGS_INSTR_MASK_C ))
#define SGS_INSTR_GET_E( x )   ((int) (((x) >> SGS_INSTR_OFF_E ) & SGS_INSTR_MASK_E ))

#define SGS_INSTR_MAKE_OP( x ) ((sgs_instr_t) (((x) & SGS_INSTR_MASK_OP) << SGS_INSTR_OFF_OP))
#define SGS_INSTR_MAKE_A( x )  ((sgs_instr_t) (((x) & SGS_INSTR_MASK_A ) << SGS_INSTR_OFF_A ))
#define SGS_INSTR_MAKE_B( x )  ((sgs_instr_t) (((x) & SGS_INSTR_MASK_B ) << SGS_INSTR_OFF_B ))
#define SGS_INSTR_MAKE_C( x )  ((sgs_instr_t) (((x) & SGS_INSTR_MASK_C ) << SGS_INSTR_OFF_C ))
#define SGS_INSTR_MAKE_E( x )  ((sgs_instr_t) (((x) & SGS_INSTR_MASK_E ) << SGS_INSTR_OFF_E ))

#define SGS_INSTR_MAKE( op, a, b, c ) \
	( SGS_INSTR_MAKE_OP( op ) | SGS_INSTR_MAKE_A( a ) | SGS_INSTR_MAKE_B( b ) | SGS_INSTR_MAKE_C( c ) )
#define SGS_INSTR_MAKE_EX( op, ex, c ) \
	( SGS_INSTR_MAKE_OP( op ) | SGS_INSTR_MAKE_E( ex ) | SGS_INSTR_MAKE_C( c ) )
#define SGS_INSTR_RECOMB_E( a, b ) ( ( ( b ) << 8 ) | ( a ) )


struct sgs_iFunc
{
	int32_t refcount;
	uint32_t size;
	uint32_t instr_off;
	uint8_t gotthis;
	uint8_t numargs;
	uint8_t numtmp;
	uint8_t numclsr;
	uint8_t inclsr;
	sgs_LineNum linenum;
	sgs_LineNum* lineinfo;
	sgs_iStr* sfuncname;
	sgs_iStr* sfilename;
};

SGS_CASSERT( sizeof(sgs_Variable) % 4 == 0, variable_object_chaining_issue );
SGS_CASSERT( sizeof(sgs_iFunc) % 4 == 0, ifunc_object_chaining_issue );
SGS_CASSERT( sizeof(sgs_iStr) % 4 == 0, istr_object_storage_compat_issue );

#define sgs_func_size( pfn ) (pfn)->size
#define sgs_func_instr_off( pfn ) (pfn)->instr_off
#define sgs_func_instr_count( pfn ) ((sgs_func_size(pfn) - sgs_func_instr_off(pfn)) / sizeof(sgs_instr_t))
#define sgs_func_const_count( pfn ) (sgs_func_instr_off(pfn) / sizeof(sgs_Variable))
#define sgs_func_consts( pfn )   (SGS_ASSUME_ALIGNED(((sgs_iFunc*)(pfn))+1,sgs_Variable))
#define sgs_func_bytecode( pfn ) ((sgs_instr_t*)(sgs_func_consts(pfn)+sgs_func_instr_off(pfn)/sizeof(sgs_Variable)))
#define sgs_func_lineinfo( pfn ) ((pfn)->lineinfo)
#define sgs_func_c_consts( pfn )   (SGS_ASSUME_ALIGNED_CONST(((const sgs_iFunc*)(pfn))+1,sgs_Variable))
#define sgs_func_c_bytecode( pfn ) \
	((const sgs_instr_t*)(sgs_func_c_consts(pfn) + sgs_func_instr_off(pfn) / sizeof(sgs_Variable)))


typedef size_t sgs_clsrcount_t;
typedef struct sgs_Closure sgs_Closure;
struct sgs_Closure
{
	int32_t refcount;
	sgs_Variable var;
};


/* VM interface */
void sgsVM_VarCreateString( SGS_CTX, sgs_Variable* out, const char* str, sgs_SizeVal len );
void sgsVM_VarDestroyObject( SGS_CTX, sgs_VarObj* O );
void sgsVM_DestroyVar( SGS_CTX, sgs_Variable* p );

#if 1
#define IS_REFTYPE( type ) ((1<<(type))&0x2b0)
#else
#define IS_REFTYPE( type ) ( type == SGS_VT_STRING || type == SGS_VT_FUNC \
	|| type == SGS_VT_OBJECT || type == SGS_VT_THREAD )
#endif
#define VAR_ACQUIRE( pvar ) { \
	if( IS_REFTYPE( (pvar)->type ) ) ++*(pvar)->data.pRC; }
#define VAR_RELEASE( pvar ) { \
	if( IS_REFTYPE( (pvar)->type ) ){ \
		--*(pvar)->data.pRC; \
		if( (*(pvar)->data.pRC) <= 0 ){ sgsVM_DestroyVar( C, pvar ); } \
	} \
	(pvar)->type = SGS_VT_NULL; }

#define ptr_add( a, b ) (((char*)(a)) + (b))
#define ptr_sub( a, b ) ((char*)(a) - (char*)(b))
#define ptr_dec( o, x ) *(char**)&(o) -= (x)

#define _STACK_PREPARE ptrdiff_t _stksz = 0;
#define _STACK_PROTECT _stksz = C->stack_off - C->stack_base; C->stack_off = C->stack_top;
#define _STACK_PROTECT_SKIP( n ) do{ _stksz = C->stack_off - C->stack_base; \
	C->stack_off = C->stack_top - (n); }while(0)
#define _STACK_UNPROTECT fstk_pop( C, SGS_STACKFRAMESIZE ); C->stack_off = C->stack_base + _stksz;
#define _STACK_UNPROTECT_SKIP( n ) do{ sgs_StkIdx __n = (n); \
	fstk_clean( C, C->stack_off, stk_ptop( C, -__n ) ); \
	C->stack_off = C->stack_base + _stksz; }while(0)

#define p_setvar_leave( dstp, srcp ){ \
	VAR_RELEASE( dstp ); \
	*(dstp) = *(srcp); }
#define p_setvar( dstp, srcp ){ \
	p_setvar_leave( dstp, srcp ); \
	VAR_ACQUIRE( dstp ); }
#define p_setvar_race( dstp, srcp ){ \
	sgs_Variable old = *(dstp); \
	*(dstp) = *(srcp); \
	VAR_ACQUIRE( dstp ); \
	VAR_RELEASE( &old ); }
#define p_obj( p ) ((p)->data.O)
#define p_objdata( p ) (p_obj( p )->data)
#define p_str( p ) ((p)->data.S)
#define p_cstr( p ) sgs_str_cstr( p_str( p ) )
#define p_strsz( p ) ((sgs_SizeVal)p_str( p )->size)
#define p_initnull( p ) {(p)->type = SGS_VT_NULL;}
#define p_initvar( dstp, srcp ){ \
	*(dstp) = *(srcp); \
	VAR_ACQUIRE( dstp ); }

#define stk_poff( C, off ) ((C)->stack_off + (off))
#define stk_ptop( C, off ) ((C)->stack_top + (off))
#define stk_gettop( C ) stk_ptop( C, -1 )
#define stk_size( C ) ((sgs_StkIdx)((C)->stack_top - (C)->stack_off))
#define stk_topindex( C, off ) ((sgs_StkIdx)((off) + stk_size(C)))
#define stk_absindex( C, off ) ((sgs_StkIdx)((off) >= 0 ? (off) : (off) + stk_size(C)))

#define DEBUG_STACK 0
void fstk_resize( SGS_CTX, size_t nsz );
#if DEBUG_STACK
#  define stk_makespace( C, num ){ \
	size_t _reqsz = (size_t) ( ( C->stack_top - C->stack_base ) + (num) ); \
	sgs_BreakIf( (num) < 0 ); \
	sgs_BreakIf( _reqsz < (num) ); /* overflow test */ \
	fstk_resize( C, _reqsz ); }
#else
#  define stk_makespace( C, num ){ \
	size_t _reqsz = (size_t) ( ( C->stack_top - C->stack_base ) + (num) ); \
	sgs_BreakIf( (num) < 0 ); \
	sgs_BreakIf( _reqsz < (size_t)(num) ); /* overflow test */ \
	if( _reqsz > (C)->stack_mem ) fstk_resize( (C), _reqsz ); }
#endif

#define stk_setlvar( C, off, srcp ){ \
	sgs_Variable* dstp = stk_poff( C, off ); \
	p_setvar( dstp, srcp ); }
#define stk_setlvar_leave( C, off, srcp ){ \
	sgs_Variable* dstp = stk_poff( C, off ); \
	p_setvar_leave( dstp, srcp ); }
#define stk_upush( C, vp ){ \
	*(C)->stack_top = *(vp); \
	VAR_ACQUIRE( (C)->stack_top ); \
	(C)->stack_top++; }
#define stk_upush_leave( C, vp ) \
	*(C)->stack_top++ = *(vp)
#define stk_push( C, vp ){ \
	stk_makespace( C, 1 ); \
	stk_upush( C, vp ); }
#define stk_push_leave( C, vp ){ \
	stk_makespace( C, 1 ); \
	stk_upush_leave( C, vp ); }
#define stk_push_null( C ){ \
	stk_makespace( C, 1 ); \
	((C)->stack_top++)->type = SGS_VT_NULL; }
#define stk_push2( C, vp1, vp2 ){ \
	stk_makespace( C, 2 ); \
	C->stack_top[ 0 ] = *(vp1); \
	C->stack_top[ 1 ] = *(vp2); \
	VAR_ACQUIRE( &C->stack_top[ 0 ] ); \
	VAR_ACQUIRE( &C->stack_top[ 1 ] ); \
	C->stack_top += 2; }
#define stk_push3( C, vp1, vp2, vp3 ){ \
	stk_makespace( C, 3 ); \
	C->stack_top[ 0 ] = *(vp1); \
	C->stack_top[ 1 ] = *(vp2); \
	C->stack_top[ 2 ] = *(vp3); \
	VAR_ACQUIRE( &C->stack_top[ 0 ] ); \
	VAR_ACQUIRE( &C->stack_top[ 1 ] ); \
	VAR_ACQUIRE( &C->stack_top[ 2 ] ); \
	C->stack_top += 3; }
#define stk_mpush( C, vp, cnt ){ \
	stk_makespace( C, cnt ); \
	fstk_umpush( C, vp, cnt ); }
#define stk_popskip( C, num, skip ){ \
	if( (num) > 0 ){ \
		sgs_Variable* off = C->stack_top - skip; \
		sgs_Variable* ptr = off - num; \
		fstk_clean( C, ptr, off ); \
	} }
#define stk_popto( C, to ){ \
	sgs_Variable* top = C->stack_top; \
	while( top > to ){ \
		top--; \
		VAR_RELEASE( top ); } \
	C->stack_top = to; }
#define stk_pop1( C ){ \
	sgs_BreakIf( C->stack_top == C->stack_off ); \
	C->stack_top--; \
	VAR_RELEASE( C->stack_top ); }
#define stk_pop1nr( C ){ \
	sgs_BreakIf( C->stack_top == C->stack_off ); \
	C->stack_top--; }

void fstk_push( SGS_CTX, sgs_Variable* vp );
void fstk_push_leave( SGS_CTX, sgs_Variable* vp );
void fstk_push2( SGS_CTX, sgs_Variable* vp1, sgs_Variable* vp2 );
void fstk_push_null( SGS_CTX );
void fstk_umpush( SGS_CTX, sgs_Variable* vp, sgs_SizeVal cnt );
void fstk_clean( SGS_CTX, sgs_Variable* from, sgs_Variable* to );
void fstk_pop( SGS_CTX, sgs_StkIdx num );
void fstk_pop1( SGS_CTX );

size_t sgsVM_VarSize( const sgs_Variable* var );
void sgsVM_VarDump( const sgs_Variable* var );

void sgsVM_StackDump( SGS_CTX );

int sgsVM_PushStackFrame( SGS_CTX, sgs_Variable* func );
void sgsVM_PushClosures( SGS_CTX, sgs_Closure** cls, int num );


void sgs_SerializeInt_V3( SGS_CTX, sgs_Variable var, const char* tab, sgs_SizeVal tablen );
const char* sgson_parse( SGS_CTX, sgs_MemBuf* stack, const char* buf, sgs_SizeVal size );


sgs_Variable* sgsVM_VarMake_Dict();

int sgsVM_RegStdLibs( SGS_CTX );

int sgs_specfn_apply( SGS_CTX );


/*
	Context handling
*/

/* register / constant position */
typedef struct sgs_FuncCtx sgs_FuncCtx;

typedef struct sgs_ObjPoolItem
{
	sgs_VarObj* obj;
	uint32_t appsize;
}
sgs_ObjPoolItem;

#define SGS_SF_METHOD  0x01
#define SGS_SF_HASTHIS 0x02
#define SGS_SF_ABORTED 0x04
#define SGS_SF_REENTER 0x08
#define SGS_SF_PAUSED  0x10

struct sgs_StackFrame
{
	sgs_Variable*   func;
	const uint32_t* iptr;
#if SGS_DEBUG && SGS_DEBUG_VALIDATE
	const uint32_t* code;
	const uint32_t* iend;
#endif
	sgs_Variable*   cptr;
	sgs_Closure**   clsrlist;
	sgs_VarObj*     clsrref;
	const char*     nfname;
	sgs_StackFrame* prev;
	sgs_StackFrame* next;
	sgs_StkIdx argbeg;
	sgs_StkIdx stkoff;
#if SGS_DEBUG && SGS_DEBUG_VALIDATE
	int32_t constcount;
#endif
	int32_t errsup;
	uint8_t argcount;
	uint8_t flags;
#if SGS_DEBUG && SGS_DEBUG_VALIDATE
	uint8_t clsrcount;
#endif
};
#define SGS_SF_ARG_COUNT( sf ) \
	((sf)->argcount + !!SGS_HAS_ANY_FLAG( (sf)->flags, SGS_SF_METHOD ))
#define SGS_SF_ARGC_EXPECTED( sf ) \
	((sf)->func->type == SGS_VT_FUNC ? (sf)->func->data.F->numargs : 0)

typedef struct sgs_ShCtx sgs_ShCtx;
#define SGS_SHCTX sgs_ShCtx* S
#define SGS_SHCTX_USE SGS_SHCTX = C->shared
struct sgs_ShCtx
{
	uint32_t      version;
	sgs_Context*  ctx_root;
	sgs_Context*  state_list;
	int32_t       statecount;
	
	/* script file system */
	sgs_ScriptFSFunc sfs_fn;
	void*         sfs_ctx;
	
	/* output */
	sgs_OutputFunc output_fn; /* output function */
	void*         output_ctx; /* output context */
	sgs_OutputFunc erroutput_fn; /* error output function */
	void*         erroutput_ctx; /* error output context */
	
	/* memory */
	sgs_MemFunc   memfunc;
	void*         mfuserdata;
	size_t        memsize;
	size_t        numallocs;
	size_t        numfrees;
	size_t        numblocks;
	
	/* > object info */
	sgs_VarObj*   objs;
	int32_t       objcount;
	/* >> object GC */
	uint8_t       redblue;
	uint16_t      gcrun;
	/* >> object pool */
	sgs_ObjPoolItem* objpool_data;
	int32_t       objpool_size;
	
	/* more pools */
	sgs_Context* ctx_pool;
	sgs_StackFrame* sf_pool;
	int ctx_pool_size;
	int sf_pool_size;
	
	/* tables / cache */
	sgs_VHTable   typetable; /* type interface table */
	sgs_VHTable   stringtable; /* string constant caching hash table */
	sgs_VHTable   ifacetable; /* interface generator => object table */
	
	/* cached interface objects */
	sgs_VarObj*   array_iface;
	
	/* > _R (global registry) */
	sgs_VarObj*   _R;
	sgs_VarObj*   _SYM;
	sgs_VarObj*   _INC;
};

/* Virtual machine state */
#define SGS_HAS_ERRORS          0x00010000
#define SGS_MUST_STOP          (0x00020000 | SGS_HAS_ERRORS)
#define SGS_STATE_LASTFUNCPAUSE 0x0002U
#define SGS_STATE_DESTROYING    0x0010U
#define SGS_STATE_LASTFUNCABORT 0x0020U
#define SGS_STATE_INSIDE_API    0x0040U
#define SGS_STATE_COROSTART     0x0080U /* function is pushed to stack */

struct sgs_Context
{
	int32_t       refcount;
	sgs_ShCtx*    shared;
	sgs_Context*  prev;
	sgs_Context*  next;
	
	void*         serialize_state; /* current serialization state */
	
	/* info output */
	int32_t       minlev;
	int32_t       apilev;
	sgs_MsgFunc   msg_fn;  /* messaging function */
	void*         msg_ctx; /* messaging context */
	int           last_errno;
	int           object_arg;
	
	/* hook */
	sgs_HookFunc  hook_fn;
	void*         hook_ctx;
	
	/* compilation */
	uint32_t      state;
	sgs_FuncCtx*  fctx;      /* ByteCodeGen */
	const char*   filename;  /* filename of currently compiled code */
	
	/* virtual machine */
	/* > coop micro-threading */
	sgs_Context*  parent; /* owning (parent) context */
	sgs_Context*  subthreads; /* owned contexts */
	sgs_Context*  st_next; /* next owned context of parent */
	sgs_Real      st_timeout;
	sgs_VarObj*   _E; /* end events */
	sgs_Real      wait_timer; /* sync/race */
	sgs_Real      tm_accum; /* delta time accumulator for yield return */
	
	/* > main stack */
	sgs_Variable* stack_base;
	uint32_t      stack_mem;
	sgs_Variable* stack_off;
	sgs_Variable* stack_top;
	
	/* > stack frame info */
	sgs_StackFrame* sf_first;
	sgs_StackFrame* sf_last;
	int           sf_count;
	int           num_last_returned;
	
	/* > _G (global variable dictionary) */
	sgs_VarObj*   _G;
};

#define SGS_STACKFRAMESIZE ((sgs_StkIdx)(C->stack_top - C->stack_off))
#define SGS_STACK_PRESERVE( C, ptr ) ((char*)(ptr) - (char*)C->stack_base)
#define SGS_STACK_RESTORE( C, off ) SGS_ASSUME_ALIGNED((char*)C->stack_base + (off),sgs_Variable)


#ifdef SGS_INTERNAL_STRINGTABLES

static const char* sgs_ErrNames[] =
{
	"SUCCESS", "ENOTFND", "ECOMP", "<UNUSED>",
	"ENOTSUP", "EBOUNDS", "EINVAL", "EINPROC"
};

static const char* sgs_VarNames[] =
{
	"null",
	"bool",
	"int",
	"real",
	"string",
	"function",
	"C function",
	"object",
	"ptr",
	"thread",
};

static const char* sgs_OpNames[] =
{
	"nop", "push", "int",
	"ret1", "return", "jump", "jump_if_true", "jump_if_false", "jump_if_null", "call",
	"for_prep", "for_load", "for_jump", "loadconst", "getvar", "setvar",
	"getprop", "setprop", "getindex", "setindex", "arrpush",
	"clsrinfo", "makeclsr", "getclsr", "setclsr",
	"set", "mconcat", "concat", "negate", "bool_inv", "invert",
	"inc", "dec", "add", "sub", "mul", "div", "mod",
	"and", "or", "xor", "lsh", "rsh",
	"seq", "sneq", "eq", "neq", "lt", "gte", "gt", "lte", "rawcmp",
	"array", "dict", "map", "class", "new", "rsym", "cotrt", "cotrf", "coabort", "yldjmp",
};

#endif


sgs_Context* sgsCTX_ForkState( SGS_CTX, int copystate );
void sgsCTX_FreeState( SGS_CTX );
sgs_StackFrame* sgsCTX_AllocFrame( SGS_CTX );

#define sgsCTX_FreeFrame( F ) { \
	SGS_SHCTX_USE; \
	if( S->sf_pool_size >= SGS_STACKFRAMEPOOL_SIZE ){ \
		sgs_Dealloc( F ); \
	} else { \
		F->next = S->sf_pool; \
		S->sf_pool = F; \
		S->sf_pool_size++; \
	} \
}

void sgsSTD_PostInit( SGS_CTX );
SGSBOOL sgsSTD_MakeArray( SGS_CTX, sgs_Variable* out, sgs_SizeVal cnt );
SGSBOOL sgsSTD_MakeDict( SGS_CTX, sgs_Variable* out, sgs_SizeVal cnt );
SGSBOOL sgsSTD_MakeMap( SGS_CTX, sgs_Variable* out, sgs_SizeVal cnt );
sgs_Closure** sgsSTD_MakeClosure( SGS_CTX, sgs_Variable* out, sgs_Variable* func, size_t clc );
void sgsSTD_ThreadsFree( SGS_CTX );
void sgsSTD_ThreadsGC( SGS_CTX );
void sgsSTD_RegistryInit( SGS_CTX );
void sgsSTD_RegistryFree( SGS_CTX );
void sgsSTD_RegistryGC( SGS_SHCTX );
void sgsSTD_RegistryIter( SGS_CTX, int subtype, sgs_VHTVar** outp, sgs_VHTVar** outpend );
void sgsSTD_GlobalInit( SGS_CTX );
void sgsSTD_GlobalFree( SGS_CTX );
SGSBOOL sgsSTD_GlobalGet( SGS_CTX, sgs_Variable* out, sgs_Variable* idx );
SGSBOOL sgsSTD_GlobalSet( SGS_CTX, sgs_Variable* idx, sgs_Variable* val );
void sgsSTD_GlobalGC( SGS_CTX );
void sgsSTD_GlobalIter( SGS_CTX, sgs_VHTVar** outp, sgs_VHTVar** outpend );



/* STRUCTS */

typedef struct sgsstd_array_header_s
{
	sgs_SizeVal size;
	sgs_SizeVal mem;
	sgs_Variable* data;
}
sgsstd_array_header_t;

typedef struct sgsstd_array_iter_s
{
	sgs_Variable ref;
	sgs_SizeVal size;
	sgs_SizeVal off;
}
sgsstd_array_iter_t;

extern SGS_APIFUNC sgs_ObjInterface sgsstd_array_iface[1];
extern SGS_APIFUNC sgs_ObjInterface sgsstd_array_iter_iface[1];
extern SGS_APIFUNC sgs_ObjInterface sgsstd_dict_iface[1];
extern SGS_APIFUNC sgs_ObjInterface sgsstd_dict_iter_iface[1];
extern SGS_APIFUNC sgs_ObjInterface sgsstd_map_iface[1];
extern SGS_APIFUNC sgs_ObjInterface sgsstd_closure_iface[1];
extern SGS_APIFUNC sgs_ObjInterface sgsstd_event_iface[1];

void sgsstd_array_insert_p( SGS_CTX, sgsstd_array_header_t* hdr, sgs_SizeVal pos, sgs_Variable* var );




#ifdef __cplusplus
}
#endif


#endif /* SGS_INT_H_INCLUDED */



typedef int32_t sgs_rcpos_t;
#define rcpos_t sgs_rcpos_t

typedef struct sgs_BreakInfo sgs_BreakInfo;
struct sgs_BreakInfo
{
	sgs_BreakInfo* next;
	uint32_t jdoff;  /* jump data offset */
	uint16_t numlp;  /* which loop */
	uint8_t  iscont; /* is a "continue"? */
};

#ifndef SGS_MAX_DEFERRED_BLOCKS
#  define SGS_MAX_DEFERRED_BLOCKS 256
#endif

typedef struct sgs_BlockInfo sgs_BlockInfo;
struct sgs_BlockInfo
{
	sgs_BlockInfo* parent;
	size_t defer_start;
};

typedef struct sgs_LoopInfo sgs_LoopInfo;
struct sgs_LoopInfo
{
	sgs_LoopInfo* parent;
	sgs_BlockInfo* block;
};

typedef struct sgs_CompFunc
{
	sgs_MemBuf consts;
	sgs_MemBuf code;
	sgs_MemBuf lnbuf;
	uint8_t gotthis; /* bool */
	uint8_t numargs; /* guaranteed to be <= 255 by a test in `preparse_arglist` */
	uint8_t numtmp; /* reg. count (0-255, incl. numargs) - numargs (0-255) */
	uint8_t numclsr;
	uint8_t inclsr;
}
sgs_CompFunc;

struct sgs_FuncCtx
{
	int32_t func;
	sgs_rcpos_t regs, lastreg;
	sgs_MemBuf vars;
	sgs_MemBuf gvars;
	sgs_MemBuf clsr;
	int inclsr, outclsr, syncdepth;
	int32_t loops;
	sgs_BreakInfo* binfo;
	sgs_BlockInfo* blocks;
	sgs_LoopInfo* loopinfo;
	sgs_FTNode* defers[ SGS_MAX_DEFERRED_BLOCKS ];
	size_t num_defers;
	sgs_CompFunc cfunc;
};


#define over_limit( x, lim ) ((x)>(lim)||(x)<(-lim))


/* register allocation */

static SGS_INLINE rcpos_t comp_reg_alloc( SGS_CTX )
{
	rcpos_t out = C->fctx->regs++;
	if( out > 0xff )
	{
		C->state |= SGS_HAS_ERRORS | SGS_MUST_STOP;
		sgs_Msg( C, SGS_ERROR, "Max. register count exceeded" );
	}
	return out;
}

static SGS_INLINE rcpos_t comp_reg_alloc_n( SGS_CTX, int n )
{
	rcpos_t out;
	sgs_BreakIf( n < 1 );
	out = comp_reg_alloc( C );
	n--;
	while( n --> 0 )
	{
		if( SGS_HAS_FLAG( C->state, SGS_MUST_STOP ) )
			break;
		comp_reg_alloc( C );
	}
	return out;
}

static SGS_INLINE void comp_reg_unwind( SGS_CTX, rcpos_t pos )
{
	if( C->fctx->regs > C->fctx->lastreg )
		C->fctx->lastreg = C->fctx->regs;
	C->fctx->regs = pos;
}


static void fctx_binfo_add( SGS_CTX, sgs_FuncCtx* fctx,
	uint32_t ioff, uint16_t loop, uint8_t iscont )
{
	sgs_BreakInfo* binfo = sgs_Alloc( sgs_BreakInfo );
	binfo->jdoff = ioff;
	binfo->numlp = loop;
	binfo->iscont = iscont;
	binfo->next = fctx->binfo;
	fctx->binfo = binfo;
}

static void fctx_binfo_rem( SGS_CTX, sgs_FuncCtx* fctx, sgs_BreakInfo* prev )
{
	sgs_BreakInfo* pn;
	if( prev )
	{
		pn = prev->next;
		prev->next = prev->next->next;
		sgs_Dealloc( pn );
	}
	else
	{
		pn = fctx->binfo;
		fctx->binfo = fctx->binfo->next;
		sgs_Dealloc( pn );
	}
}

static void fctx_block_push( sgs_FuncCtx* fctx, sgs_BlockInfo* bdata, sgs_LoopInfo* ldata )
{
	bdata->defer_start = fctx->num_defers;
	bdata->parent = fctx->blocks;
	fctx->blocks = bdata;
	if( ldata )
	{
		ldata->block = bdata;
		ldata->parent = fctx->loopinfo;
		fctx->loopinfo = ldata;
	}
}

static void fctx_block_pop( sgs_FuncCtx* fctx, sgs_BlockInfo* bdata, sgs_LoopInfo* ldata )
{
	sgs_BreakIf( bdata != fctx->blocks );
	fctx->num_defers = bdata->defer_start;
	fctx->blocks = bdata->parent;
	if( ldata )
	{
		sgs_BreakIf( ldata != fctx->loopinfo );
		fctx->loopinfo = ldata->parent;
	}
}

static void fctx_defer_add( SGS_CTX, sgs_FTNode* stmt )
{
	sgs_FuncCtx* fctx = C->fctx;
	if( fctx->num_defers >= SGS_MAX_DEFERRED_BLOCKS )
	{
		C->state |= SGS_HAS_ERRORS | SGS_MUST_STOP;
		sgs_Msg( C, SGS_ERROR, "exceeded deferred block limit" );
	}
	else
		fctx->defers[ fctx->num_defers++ ] = stmt;
}


static sgs_FuncCtx* fctx_create( SGS_CTX )
{
	sgs_FuncCtx* fctx = sgs_Alloc( sgs_FuncCtx );
	fctx->func = SGS_TRUE;
	fctx->regs = 0;
	fctx->lastreg = -1;
	fctx->vars = sgs_membuf_create();
	fctx->gvars = sgs_membuf_create();
	fctx->clsr = sgs_membuf_create();
	fctx->inclsr = 0;
	fctx->outclsr = 0;
	fctx->syncdepth = 0;
	fctx->loops = 0;
	fctx->binfo = NULL;
	fctx->blocks = NULL;
	fctx->loopinfo = NULL;
	fctx->num_defers = 0;
	sgs_membuf_appbuf( &fctx->gvars, C, "_G=", 3 );
	
	fctx->cfunc.consts = sgs_membuf_create();
	fctx->cfunc.code = sgs_membuf_create();
	fctx->cfunc.lnbuf = sgs_membuf_create();
	fctx->cfunc.gotthis = SGS_FALSE;
	fctx->cfunc.numargs = 0;
	fctx->cfunc.numtmp = 0;
	fctx->cfunc.numclsr = 0;
	fctx->cfunc.inclsr = 0;
	
	return fctx;
}

static void fctx_destroy( SGS_CTX, sgs_FuncCtx* fctx )
{
	sgs_CompFunc* func = &fctx->cfunc;
	sgs_Variable* vbeg = SGS_ASSUME_ALIGNED( func->consts.ptr, sgs_Variable );
	sgs_Variable* vend = SGS_ASSUME_ALIGNED( func->consts.ptr + func->consts.size, sgs_Variable );
	sgs_Variable* var = vbeg;
	while( var < vend )
	{
		sgs_Release( C, var );
		var++;
	}

	sgs_membuf_destroy( &func->code, C );
	sgs_membuf_destroy( &func->consts, C );
	sgs_membuf_destroy( &func->lnbuf, C );
	
	while( fctx->binfo )
		fctx_binfo_rem( C, fctx, NULL );
	sgs_membuf_destroy( &fctx->vars, C );
	sgs_membuf_destroy( &fctx->gvars, C );
	sgs_membuf_destroy( &fctx->clsr, C );
	sgs_Dealloc( fctx );
}

#if SGS_DUMP_BYTECODE || ( SGS_DEBUG && SGS_DEBUG_DATA )
static void fctx_dumpvarlist( sgs_MemBuf* mb )
{
	char* p = mb->ptr, *pend = mb->ptr + mb->size;
	while( p < pend )
	{
		if( *p == '=' )
		{
			if( p < pend - 1 )
				printf( ", " );
		}
		else
			putchar( *p );
		p++;
	}
}
static void fctx_dump( sgs_FuncCtx* fctx )
{
	printf( "Type: %s\n", fctx->func ? "Function" : "Main code" );
	printf( "Globals: " ); fctx_dumpvarlist( &fctx->gvars ); printf( "\n" );
	printf( "Variables: " ); fctx_dumpvarlist( &fctx->vars ); printf( "\n" );
	printf( "Closures in=%d out=%d : ", fctx->inclsr, fctx->outclsr );
	fctx_dumpvarlist( &fctx->clsr ); printf( "\n" );
}
#endif


static void dump_rcpos( int arg )
{
	char rc = SGS_CONSTVAR( arg ) ? 'C' : 'R';
	arg = SGS_CONSTDEC( arg );
	printf( "%c%d", rc, arg );
}
static void dump_opcode_a( const char* name, sgs_instr_t I )
{
	printf( "%s R%" PRId32" <= ", name, SGS_INSTR_GET_A( I ) );
	dump_rcpos( SGS_INSTR_GET_B( I ) );
	printf( ", " );
	dump_rcpos( SGS_INSTR_GET_C( I ) );
}
static void dump_opcode_b( const char* name, sgs_instr_t I )
{
	printf( "%s R%" PRId32" <= ", name, SGS_INSTR_GET_A( I ) );
	dump_rcpos( SGS_INSTR_GET_B( I ) );
}
static void dump_opcode_b1( const char* name, sgs_instr_t I )
{
	printf( "%s ", name );
	dump_rcpos( SGS_INSTR_GET_B( I ) );
	printf( " <= " );
	dump_rcpos( SGS_INSTR_GET_C( I ) );
}
static void dump_opcode( const sgs_instr_t* ptr, size_t count )
{
	const sgs_instr_t* pend = ptr + count;
	const sgs_instr_t* pbeg = ptr;
	while( ptr < pend )
	{
		sgs_instr_t I = *ptr++;
		int op = SGS_INSTR_GET_OP( I ), argA = SGS_INSTR_GET_A( I ),
			argB = SGS_INSTR_GET_B( I ), argC = SGS_INSTR_GET_C( I ),
			argE = SGS_INSTR_GET_E( I );

		printf( "\t%04d |  ", (int)( ptr - pbeg - 1 ) );

		switch( op )
		{
#define DOP_A( wat ) case SGS_SI_##wat: dump_opcode_a( #wat, I ); break;
#define DOP_B( wat ) case SGS_SI_##wat: dump_opcode_b( #wat, I ); break;
		case SGS_SI_NOP: printf( "NOP   " ); break;

		case SGS_SI_PUSH: printf( "PUSH " ); dump_rcpos( argB ); break;
		case SGS_SI_INT: printf( "INT %d", argC ); break;

		case SGS_SI_RET1: printf( "RET_1 <= " ); dump_rcpos( argC ); break;
		case SGS_SI_RETN: printf( "RETURN %d", argA ); break;
		case SGS_SI_JUMP: printf( "JUMP %d", (int) (int16_t) argE ); break;
		case SGS_SI_JMPT: printf( "JMP_T " ); dump_rcpos( argC );
			printf( ", %d", (int) (int16_t) argE ); break;
		case SGS_SI_JMPF: printf( "JMP_F " ); dump_rcpos( argC );
			printf( ", %d", (int) (int16_t) argE ); break;
		case SGS_SI_JMPN: printf( "JMP_N " ); dump_rcpos( argC );
			printf( ", %d", (int) (int16_t) argE ); break;
		case SGS_SI_CALL: printf( "CALL args: %d - %d expect: %d%s",
			argB & 0xff, argC, argA, ( argB & 0x100 ) ? ", method" : "" );
			break;

		case SGS_SI_FORPREP: printf( "FOR_PREP " ); dump_rcpos( argA );
			printf( " <= " ); dump_rcpos( argB ); break;
		case SGS_SI_FORLOAD: printf( "FOR_LOAD " ); dump_rcpos( argA );
			printf( " => " ); dump_rcpos( argB );
			printf( ", " ); dump_rcpos( argC ); break;
		case SGS_SI_FORJUMP: printf( "FOR_JUMP " ); dump_rcpos( argC );
			printf( ", %d", (int) (int16_t) argE ); break;

		case SGS_SI_LOADCONST: printf( "LOADCONST " ); dump_rcpos( argC );
			printf( " <= C%d", argE ); break;

		DOP_B( GETVAR );
		case SGS_SI_SETVAR: dump_opcode_b1( "SETVAR", I ); break;
		DOP_A( GETPROP );
		DOP_A( SETPROP );
		DOP_A( GETINDEX );
		DOP_A( SETINDEX );
		case SGS_SI_ARRPUSH: printf( "ARR_PUSH R%d <= ", argA );
			dump_rcpos( argC ); break;

		case SGS_SI_CLSRINFO: printf( "CLSR_INFO %d %d %d", argA, argB, argC ); break;
		case SGS_SI_MAKECLSR: printf( "MAKE_CLSR " ); dump_rcpos( argA );
			printf( " <= " ); dump_rcpos( argB );
			printf( " [%d]", argC ); break;
		case SGS_SI_GETCLSR: printf( "GET_CLSR " ); dump_rcpos( argA );
			printf( " <= CL%d", argB ); break;
		case SGS_SI_SETCLSR: printf( "SET_CLSR CL%d <= ", argB );
			dump_rcpos( argC ); break;

		DOP_B( SET );
		case SGS_SI_MCONCAT: printf( "MCONCAT " ); dump_rcpos( argA );
			printf( " [%d]", argB ); break;
		DOP_A( CONCAT );
		DOP_B( NEGATE );
		DOP_B( BOOL_INV );
		DOP_B( INVERT );

		DOP_B( INC );
		DOP_B( DEC );
		DOP_A( ADD );
		DOP_A( SUB );
		DOP_A( MUL );
		DOP_A( DIV );
		DOP_A( MOD );

		DOP_A( AND );
		DOP_A( OR );
		DOP_A( XOR );
		DOP_A( LSH );
		DOP_A( RSH );

		DOP_A( SEQ );
		DOP_A( EQ );
		DOP_A( LT );
		DOP_A( LTE );

		DOP_A( SNEQ );
		DOP_A( NEQ );
		DOP_A( GT );
		DOP_A( GTE );
		
		DOP_A( RAWCMP );

		case SGS_SI_ARRAY:
			printf( "ARRAY args:%d output:", argE );
			dump_rcpos( argC ); break;
		case SGS_SI_DICT:
			printf( "DICT output:" );
			dump_rcpos( argC ); break;
		case SGS_SI_CLASS:
			printf( "CLASS output:%d", argA );
			printf( ", name:" ); dump_rcpos( argB );
			printf( ", inhname:%s", argC == argA ? "<none>" : "" );
			if( argC != argA ) dump_rcpos( argC );
			break;
		case SGS_SI_NEW:
			printf( "NEW output:%d", argA );
			printf( ", class:" ); dump_rcpos( argB );
			break;
		case SGS_SI_RSYM:
			printf( "RSYM name:" ); dump_rcpos( argB );
			printf( " value:" ); dump_rcpos( argC ); break;
			
		DOP_B( COTRT );
		DOP_B( COTRF );
		case SGS_SI_COABORT: printf( "CO_ABORT %d", (int) (int16_t) argE ); break;
		case SGS_SI_YLDJMP: printf( "YIELD_JUMP %d", (int) (int16_t) argE ); break;
			
#undef DOP_A
#undef DOP_B
		default:
			printf( "<error> \t\t(op=%d A=%d B=%d C=%d E=%d)",
				op, argA, argB, argC, argE ); break;
		}
		printf( "\n" );
	}
}


static int find_var( sgs_MemBuf* S, const char* str, unsigned len )
{
	int at = 0;
	const char* ptr = S->ptr;
	const char* pend = ptr + S->size;
	const char* estr = str + len;
	
	while( ptr < pend )
	{
		/* compare ptr .. '=' with str .. str + len */
		const char* cstr = str;
		while( ptr < pend && *ptr != '=' && cstr < estr )
		{
			if( *ptr != *cstr )
				break;
			ptr++;
			cstr++;
		}
		if( ptr < pend && *ptr == '=' && cstr == estr )
			return at;
		
		/* advance to next variable */
		at++;
		while( ptr < pend && *ptr != '=' )
			ptr++;
		ptr++; /* skip '=' */
	}
	return -1;
}

static int find_nth_var( sgs_MemBuf* S, int which, char** outstr, unsigned* outlen )
{
	char* ptr = S->ptr;
	char* pend = ptr + S->size;
	while( which > 0 )
	{
		while( ptr < pend && *ptr != '=' )
			ptr++;
		ptr++;
		which--;
	}
	if( ptr >= pend )
		return 0;
	*outstr = ptr;
	while( ptr < pend && *ptr != '=' )
		ptr++;
	/* WP: ptr always bigger or equal to *outstr, difference cannot exceed 255 */
	*outlen = (unsigned) ( ptr - *outstr );
	return 1;
}

static int add_var( sgs_MemBuf* S, SGS_CTX, char* str, unsigned len )
{
	int pos = find_var( S, str, len );
	if( pos < 0 )
	{
		sgs_membuf_appbuf( S, C, str, len );
		sgs_membuf_appchr( S, C, '=' );
		return SGS_TRUE;
	}
	return SGS_FALSE;
}

#define find_varT( S, tok ) \
	find_var( S, (char*) (tok) + 2, tok[1] )
#define add_varT( S, C, tok ) \
	add_var( S, C, (char*) (tok) + 2, tok[1] )

static int preadd_thisvar( sgs_MemBuf* S, SGS_CTX )
{
	int pos = find_var( S, "this", 4 );
	if( pos < 0 )
	{
		sgs_membuf_insbuf( S, C, 0, "this=", 5 );
		return SGS_TRUE;
	}
	return SGS_FALSE;
}


/* simplifies writing code */
#define SGS_FNTCMP_ARGS SGS_CTX, sgs_CompFunc* func, sgs_FTNode* node
static void add_instr( SGS_FNTCMP_ARGS, sgs_instr_t I )
{
	sgs_LineNum ln = sgsT_LineNum( node->token );
	sgs_membuf_appbuf( &func->lnbuf, C, &ln, sizeof( ln ) );
	sgs_membuf_appbuf( &func->code, C, &I, sizeof( I ) );
}
#define INSTR_N( i, n ) add_instr( C, func, n, i )
#define INSTR( i )      INSTR_N( i, node )
#define INSTR_WRITE( op, a, b, c ) INSTR( SGS_INSTR_MAKE( op, a, b, c ) )
#define INSTR_WRITE_EX( op, ex, c ) INSTR( SGS_INSTR_MAKE_EX( op, ex, c ) )
#define INSTR_WRITE_PCH() INSTR_WRITE( 63, 0, 0, 0 )

#define QPRINT( str ) sgs_Msg( C, SGS_ERROR, "[line %d] " str, sgsT_LineNum( node->token ) )
#define QINTERR( id ) sgs_Msg( C, SGS_INTERR, "[line %d] error SGSINT%d [%s:%d,v" SGS_VERSION "]", \
	sgsT_LineNum( node->token ), id, __FILE__, __LINE__ )



static int preparse_varlists( SGS_FNTCMP_ARGS );

static int preparse_varlist( SGS_FNTCMP_ARGS )
{
	int ret = SGS_TRUE;
	node = node->child;
	while( node )
	{
		if( node->type != SGS_SFT_IDENT && node->type != SGS_SFT_KEYWORD && node->type != SGS_SFT_ARGMT )
			goto cont; /* compatibility with explists */
		if( find_varT( &C->fctx->clsr, node->token ) >= 0 )
			goto cont; /* closure */
		if( find_varT( &C->fctx->gvars, node->token ) >= 0 )
		{
			QPRINT( "Variable storage redefined: global -> local" );
			return SGS_FALSE;
		}
		if( add_varT( &C->fctx->vars, C, node->token ) )
			comp_reg_alloc( C );
		if( node->child )
			ret &= preparse_varlists( C, func, node );
cont:
		node = node->next;
	}
	return ret;
}

static int register_gv( SGS_FNTCMP_ARGS )
{
	if( find_varT( &C->fctx->clsr, node->token ) >= 0 )
	{
		QPRINT( "Variable storage redefined: closure -> global" );
		return SGS_FALSE;
	}
	if( find_varT( &C->fctx->vars, node->token ) >= 0 )
	{
		QPRINT( "Variable storage redefined: local -> global" );
		return SGS_FALSE;
	}
	add_varT( &C->fctx->gvars, C, node->token );
	return SGS_TRUE;
}

static int preparse_gvlist( SGS_FNTCMP_ARGS )
{
	int ret = SGS_TRUE;
	node = node->child;
	while( node )
	{
		if( !register_gv( C, func, node ) )
			return SGS_FALSE;
		if( node->child )
			ret &= preparse_varlists( C, func, node );
		node = node->next;
	}
	return ret;
}

static int preparse_varlists( SGS_FNTCMP_ARGS )
{
	int ret = 1;
	while( node )
	{
		if( node->type == SGS_SFT_VARLIST )
			ret &= preparse_varlist( C, func, node );
		else if( node->type == SGS_SFT_GVLIST )
			ret &= preparse_gvlist( C, func, node );
		else if( node->type == SGS_SFT_CLASS )
		{
			sgs_FTNode* ch = node->child;
			ret &= register_gv( C, func, ch );
			ch = ch->next;
			while( ch )
			{
				ret &= preparse_varlists( C, func, ch );
				ch = ch->next;
			}
		}
		else if( node->type == SGS_SFT_KEYWORD && node->token && sgsT_IsKeyword( node->token, "this" ) )
		{
			func->gotthis = SGS_TRUE;
			if( preadd_thisvar( &C->fctx->vars, C ) )
				comp_reg_alloc( C );
		}
		else if( node->type == SGS_SFT_OPER )
		{
			if( SGS_ST_OP_ASSIGN( *node->token ) && node->child )
			{
				if( node->child->type == SGS_SFT_IDENT )
				{
					/* add_var calls find_var internally but - GVARS vs VARS - note the difference */
					if( find_varT( &C->fctx->gvars, node->child->token ) == -1 &&
						find_varT( &C->fctx->clsr, node->child->token ) == -1 &&
						add_varT( &C->fctx->vars, C, node->child->token ) )
						comp_reg_alloc( C );
				}
				if( node->child->type == SGS_SFT_EXPLIST )
					ret &= preparse_varlist( C, func, node->child );
			}
			ret &= preparse_varlists( C, func, node->child );
		}
		else if( node->type == SGS_SFT_FOREACH )
		{
			if( find_varT( &C->fctx->gvars, node->token ) >= 0 )
			{
				QPRINT( "Variable storage redefined (foreach key variable cannot be global): global -> local" );
				ret = SGS_FALSE;
			}
			else
			{
				if( node->child->type != SGS_SFT_NULL && 
					add_varT( &C->fctx->vars, C, node->child->token ) )
					comp_reg_alloc( C );
				if( node->child->next->type != SGS_SFT_NULL && 
					add_varT( &C->fctx->vars, C, node->child->next->token ) )
					comp_reg_alloc( C );
			}

			ret &= preparse_varlists( C, func, node->child->next );
		}
		else if( node->type == SGS_SFT_FUNC )
		{
			sgs_FTNode* N = node->child->next->next->next;
			if( N && N->type == SGS_SFT_IDENT )
			{
				if( find_varT( &C->fctx->gvars, N->token ) == -1 && /* if the variable hasn't been .. */
					find_varT( &C->fctx->clsr, N->token ) == -1 && /* .. created before */
					/* if it was successfully added */
					add_varT( C->fctx->func ? &C->fctx->vars : &C->fctx->gvars, C, N->token ) &&
					C->fctx->func ) /* and if it was added as a local variable */
					comp_reg_alloc( C ); /* add a register for it */
			}
		}
		else if( node->child )
			ret &= preparse_varlists( C, func, node->child );
		node = node->next;
	}
	return ret;
}

static int preparse_closures( SGS_FNTCMP_ARGS, int decl )
{
	int ret;
	node = node->child;
	while( node )
	{
		ret = add_varT( &C->fctx->clsr, C, node->token );
		if( !ret && decl )
		{
			QPRINT( "Cannot redeclare used variables with the same name" );
			return 0;
		}
		if( ret )
		{
			C->fctx->outclsr++;
			if( decl )
				C->fctx->inclsr++;
		}
		node = node->next;
	}
	return 1;
}

static int preparse_clsrlists( SGS_FNTCMP_ARGS )
{
	int ret = 1;
	while( node )
	{
		if( node->type == SGS_SFT_FUNC )
			ret &= preparse_closures( C, func, node->child->next, 0 );
		else if( node->child )
			ret &= preparse_clsrlists( C, func, node->child );
		node = node->next;
	}
	return ret;
}

static int preparse_arglist( SGS_FNTCMP_ARGS )
{
	node = node->child;
	while( node )
	{
		if( func->numargs == 255 )
		{
			QPRINT( "Argument count exceeded (max. 255 arguments)" );
			return 0;
		}
		if( !add_var( &C->fctx->vars, C, (char*) node->token + 2, node->token[ 1 ] ) )
		{
			QPRINT( "Cannot redeclare arguments with the same name" );
			return 0;
		}
		comp_reg_alloc( C );
		func->numargs++;
		node = node->next;
	}
	return 1;
}

static int preparse_funcorder( SGS_FNTCMP_ARGS )
{
	sgs_FTNode* sub = node->child, *psub = NULL, *nsub;
	if( sub )
	{
		psub = sub;
		sub = sub->next; /* skip first item as it cannot be swapped (with itself) */
	}
	while( sub )
	{
		if( sub->type == SGS_SFT_FUNC /* function */ &&
			!sub->child->next->child /* no closures */ &&
			sub->child->next->next->next /* has a name .. */ &&
			sub->child->next->next->next->type == SGS_SFT_IDENT /* .. that is simple */ )
		{
			/* move to front */
			nsub = sub->next;
			if( psub )
				psub->next = nsub;
			sub->next = node->child;
			node->child = sub;
			sub = nsub;
		}
		else
		{
			psub = sub;
			sub = sub->next;
		}
	}
	return 1;
}


#define add_const_HDR \
	sgs_Variable* vbeg = SGS_ASSUME_ALIGNED( func->consts.ptr, sgs_Variable ); \
	sgs_Variable* vend = SGS_ASSUME_ALIGNED( func->consts.ptr + func->consts.size, sgs_Variable ); \
	sgs_Variable* var = vbeg; \
	sgs_Variable nvar;

static rcpos_t add_const_null( SGS_CTX, sgs_CompFunc* func )
{
	add_const_HDR;
	while( var < vend )
	{
		if( var->type == SGS_VT_NULL )
			return (rcpos_t) ( var - vbeg ); /* WP: const limit */
		var++;
	}

	nvar.type = SGS_VT_NULL;
	sgs_membuf_appbuf( &func->consts, C, &nvar, sizeof( nvar ) );
	return (rcpos_t) ( vend - vbeg ); /* WP: const limit */
}

static rcpos_t add_const_b( SGS_CTX, sgs_CompFunc* func, sgs_Bool bval )
{
	add_const_HDR;
	while( var < vend )
	{
		if( var->type == SGS_VT_BOOL && var->data.B == bval )
			return (rcpos_t) ( var - vbeg ); /* WP: const limit */
		var++;
	}

	nvar.type = SGS_VT_BOOL;
	nvar.data.B = bval;
	sgs_membuf_appbuf( &func->consts, C, &nvar, sizeof( nvar ) );
	return (rcpos_t) ( vend - vbeg ); /* WP: const limit */
}

static rcpos_t add_const_i( SGS_CTX, sgs_CompFunc* func, sgs_Int ival )
{
	add_const_HDR;
	while( var < vend )
	{
		if( var->type == SGS_VT_INT && var->data.I == ival )
			return (rcpos_t) ( var - vbeg ); /* WP: const limit */
		var++;
	}

	nvar.type = SGS_VT_INT;
	nvar.data.I = ival;
	sgs_membuf_appbuf( &func->consts, C, &nvar, sizeof( nvar ) );
	return (rcpos_t) ( vend - vbeg ); /* WP: const limit */
}

static rcpos_t add_const_r( SGS_CTX, sgs_CompFunc* func, sgs_Real rval )
{
	add_const_HDR;
	while( var < vend )
	{
		if( var->type == SGS_VT_REAL && var->data.R == rval )
			return (rcpos_t) ( var - vbeg ); /* WP: const limit */
		var++;
	}

	nvar.type = SGS_VT_REAL;
	nvar.data.R = rval;
	sgs_membuf_appbuf( &func->consts, C, &nvar, sizeof( nvar ) );
	return (rcpos_t) ( vend - vbeg ); /* WP: const limit */
}

static rcpos_t add_const_s( SGS_CTX, sgs_CompFunc* func, uint32_t len, const char* str )
{
	add_const_HDR;
	while( var < vend )
	{
		if( var->type == SGS_VT_STRING && var->data.S->size == len
			&& memcmp( sgs_var_cstr( var ), str, len ) == 0 )
			return (rcpos_t) ( var - vbeg ); /* WP: const limit */
		var++;
	}
	
	sgs_BreakIf( len > 0x7fffffff );

	sgsVM_VarCreateString( C, &nvar, str, (int32_t) len );
	sgs_membuf_appbuf( &func->consts, C, &nvar, sizeof( nvar ) );
	return (rcpos_t) ( vend - vbeg ); /* WP: const limit */
}

sgs_iFunc* sgsBC_ConvertFunc( SGS_CTX, sgs_CompFunc* nf,
	const char* funcname, size_t fnsize, sgs_LineNum lnum )
{
	sgs_Variable strvar;
	sgs_iFunc* F = sgs_Alloc_a( sgs_iFunc, nf->consts.size + nf->code.size );

	F->refcount = 1;
	/* WP: const/instruction limits */
	F->size = (uint32_t) ( nf->consts.size + nf->code.size );
	F->instr_off = (uint32_t) nf->consts.size;
	F->gotthis = nf->gotthis;
	F->numargs = nf->numargs;
	F->numtmp = nf->numtmp;
	F->numclsr = nf->numclsr;
	F->inclsr = nf->inclsr;

	{
		size_t lnc = nf->lnbuf.size / sizeof( sgs_LineNum );
		F->lineinfo = sgs_Alloc_n( sgs_LineNum, lnc );
		memcpy( F->lineinfo, nf->lnbuf.ptr, nf->lnbuf.size );
	}
	/* WP: string limit */
	sgsVM_VarCreateString( C, &strvar, funcname, (sgs_SizeVal) fnsize );
	F->sfuncname = strvar.data.S;
	F->linenum = lnum;
	/* WP: string limit */
	if( C->filename )
		sgsVM_VarCreateString( C, &strvar, C->filename, (sgs_SizeVal) strlen( C->filename ) );
	else
		sgsVM_VarCreateString( C, &strvar, "", 0 );
	F->sfilename = strvar.data.S;
	
	memcpy( sgs_func_consts( F ), nf->consts.ptr, nf->consts.size );
	memcpy( sgs_func_bytecode( F ), nf->code.ptr, nf->code.size );
	
	/* transfer ownership by making it look like constants don't exist here anymore */
	nf->consts.size = 0;
	
	return F;
}

static rcpos_t add_const_f( SGS_CTX, sgs_CompFunc* func, sgs_CompFunc* nf,
	const char* funcname, size_t fnsize, sgs_LineNum lnum )
{
	sgs_Variable nvar;
	rcpos_t pos;
	sgs_iFunc* F = sgsBC_ConvertFunc( C, nf, funcname, fnsize, lnum );
	
	/* WP: const limit */
	pos = (rcpos_t) ( func->consts.size / sizeof( nvar ) );
	nvar.type = SGS_VT_FUNC;
	nvar.data.F = F;
	sgs_membuf_appbuf( &func->consts, C, &nvar, sizeof( nvar ) );
	return pos;
}

static int op_pick_opcode( int oper, int binary )
{
	if( !binary )
	{
		if( oper == SGS_ST_OP_ADD ) return 0;
		if( oper == SGS_ST_OP_SUB )	return SGS_SI_NEGATE;
		if( oper == SGS_ST_OP_NOT ) return SGS_SI_BOOL_INV;
		if( oper == SGS_ST_OP_INV ) return SGS_SI_INVERT;
	}

	switch( oper )
	{
	case SGS_ST_OP_ADD: case SGS_ST_OP_ADDEQ: return SGS_SI_ADD;
	case SGS_ST_OP_SUB: case SGS_ST_OP_SUBEQ: return SGS_SI_SUB;
	case SGS_ST_OP_MUL: case SGS_ST_OP_MULEQ: return SGS_SI_MUL;
	case SGS_ST_OP_DIV: case SGS_ST_OP_DIVEQ: return SGS_SI_DIV;
	case SGS_ST_OP_MOD: case SGS_ST_OP_MODEQ: return SGS_SI_MOD;

	case SGS_ST_OP_AND: case SGS_ST_OP_ANDEQ: return SGS_SI_AND;
	case SGS_ST_OP_OR: case SGS_ST_OP_OREQ: return SGS_SI_OR;
	case SGS_ST_OP_XOR: case SGS_ST_OP_XOREQ: return SGS_SI_XOR;
	case SGS_ST_OP_LSH: case SGS_ST_OP_LSHEQ: return SGS_SI_LSH;
	case SGS_ST_OP_RSH: case SGS_ST_OP_RSHEQ: return SGS_SI_RSH;

	case SGS_ST_OP_CAT: case SGS_ST_OP_CATEQ: return SGS_SI_CONCAT;

	case SGS_ST_OP_SEQ: return SGS_SI_SEQ;
	case SGS_ST_OP_SNEQ: return SGS_SI_SNEQ;
	case SGS_ST_OP_EQ: return SGS_SI_EQ;
	case SGS_ST_OP_NEQ: return SGS_SI_NEQ;
	case SGS_ST_OP_LESS: return SGS_SI_LT;
	case SGS_ST_OP_LEQ: return SGS_SI_LTE;
	case SGS_ST_OP_GRTR: return SGS_SI_GT;
	case SGS_ST_OP_GEQ: return SGS_SI_GTE;
	case SGS_ST_OP_RWCMP: return SGS_SI_RAWCMP;

	default: return 0;
	}
}


static SGSBOOL compile_node( SGS_FNTCMP_ARGS );
static SGSBOOL compile_node_r( SGS_FNTCMP_ARGS, rcpos_t* out );
static SGSBOOL compile_node_w( SGS_FNTCMP_ARGS, rcpos_t src );
static SGSBOOL compile_node_rrw( SGS_FNTCMP_ARGS, rcpos_t dst );
static SGSBOOL compile_oper( SGS_FNTCMP_ARGS, rcpos_t* arg, int out, int expect );


static void compile_defers( SGS_CTX, sgs_CompFunc* func, sgs_BlockInfo* until )
{
	size_t i, end = until ? until->defer_start : 0;
	for( i = C->fctx->num_defers; i > end; )
	{
		--i;
		compile_node( C, func, C->fctx->defers[ i ] );
	}
}

#define BLOCK_BEGIN { sgs_BlockInfo binfo; fctx_block_push( C->fctx, &binfo, NULL );
#define BLOCK_END compile_defers( C, func, &binfo ); fctx_block_pop( C->fctx, &binfo, NULL ); }
#define LOOP_BEGIN { sgs_BlockInfo binfo; sgs_LoopInfo linfo; fctx_block_push( C->fctx, &binfo, &linfo );
#define LOOP_END compile_defers( C, func, &binfo ); fctx_block_pop( C->fctx, &binfo, &linfo ); }




static rcpos_t const_maybeload( SGS_FNTCMP_ARGS, rcpos_t cid )
{
	if( cid > 65535 )
	{
		QPRINT( "Maximum number of constants exceeded" );
		C->state |= SGS_MUST_STOP;
		return 0;
	}
	if( cid < 256 )
		return SGS_CONSTENC( cid );
	else
	{
		rcpos_t out = comp_reg_alloc( C );
		INSTR_WRITE_EX( SGS_SI_LOADCONST, cid, out );
		return out;
	}
}

#define BC_CONSTENC( cid ) const_maybeload( C, func, node, cid )


static void compile_ident( SGS_FNTCMP_ARGS, rcpos_t* out )
{
	rcpos_t pos = add_const_s( C, func, node->token[ 1 ], (const char*) node->token + 2 );
	*out = BC_CONSTENC( pos );
}


static SGSBOOL compile_ident_r( SGS_FNTCMP_ARGS, rcpos_t* out )
{
	rcpos_t pos;
	if( sgsT_IsKeyword( node->token, "null" ) )
	{
		pos = add_const_null( C, func );
		*out = BC_CONSTENC( pos );
		return 1;
	}
	if( sgsT_IsKeyword( node->token, "true" ) )
	{
		pos = add_const_b( C, func, SGS_TRUE );
		*out = BC_CONSTENC( pos );
		return 1;
	}
	if( sgsT_IsKeyword( node->token, "false" ) )
	{
		pos = add_const_b( C, func, SGS_FALSE );
		*out = BC_CONSTENC( pos );
		return 1;
	}
	if( *node->token == SGS_ST_KEYWORD )
	{
		if( sgsT_IsKeyword( node->token, "this" ) )
		{
			if( func->gotthis )
			{
				*out = 0;
				return 1;
			}
			else
			{
				QPRINT( "This function is not a method, cannot use 'this'" );
				return 0;
			}
		}
		QPRINT( "Cannot read from specified keyword" );
		return 0;
	}

	/* closures */
	if( ( pos = find_var( &C->fctx->clsr, (char*) node->token + 2, node->token[1] ) ) >= 0 )
	{
		*out = comp_reg_alloc( C );
		INSTR_WRITE( SGS_SI_GETCLSR, *out, pos, 0 );
		return 1;
	}

	if( C->fctx->func )
	{
		rcpos_t gpos = find_var( &C->fctx->gvars, (char*) node->token + 2, node->token[ 1 ] );
		if( gpos >= 0 )
			pos = -1;
		else
		{
			pos = find_var( &C->fctx->vars, (char*) node->token + 2, node->token[ 1 ] );
			if( pos < 0 )
				pos = -1; /* read from globals by default */
		}
	}
	else
	{
		pos = find_var( &C->fctx->vars, (char*) node->token + 2, node->token[ 1 ] );
	}

	if( pos >= 0 )
	{
		*out = pos;
	}
	else
	{
		*out = comp_reg_alloc( C );
		compile_ident( C, func, node, &pos );
		INSTR_WRITE( SGS_SI_GETVAR, *out, pos, 0 );
	}
	return 1;
}

static SGSBOOL compile_ident_w( SGS_FNTCMP_ARGS, rcpos_t src )
{
	rcpos_t pos;
	if( *node->token == SGS_ST_KEYWORD )
	{
		QPRINT( "Cannot write to reserved keywords" );
		return 0;
	}

	if( ( pos = find_var( &C->fctx->clsr, (char*) node->token + 2, node->token[1] ) ) >= 0 )
	{
		INSTR_WRITE( SGS_SI_SETCLSR, 0, pos, src );
		return 1;
	}

	if( C->fctx->func )
	{
		rcpos_t gpos = find_var( &C->fctx->gvars, (char*) node->token + 2, node->token[ 1 ] );
		if( gpos >= 0 )
			pos = -1;
		else
		{
			add_var( &C->fctx->vars, C, (char*) node->token + 2, node->token[ 1 ] );
			pos = find_var( &C->fctx->vars, (char*) node->token + 2, node->token[ 1 ] );
		}
	}
	else
	{
		pos = find_var( &C->fctx->vars, (char*) node->token + 2, node->token[ 1 ] );
	}

	if( pos >= 0 )
	{
		/* optimization */
		if( pos != src )
		{
			INSTR_WRITE( SGS_SI_SET, pos, src, 0 );
		}
	}
	else
	{
		compile_ident( C, func, node, &pos );
		INSTR_WRITE( SGS_SI_SETVAR, 0, pos, src );
	}
	return 1;
}

static SGSBOOL compile_const( SGS_FNTCMP_ARGS, rcpos_t* opos )
{
	if( *node->token == SGS_ST_NUMINT )
	{
		sgs_Int val;
		SGS_AS_INTEGER( val, node->token + 1 );
		*opos = BC_CONSTENC( add_const_i( C, func, val ) );
	}
	else if( *node->token == SGS_ST_NUMREAL )
	{
		sgs_Real val;
		SGS_AS_REAL( val, node->token + 1 );
		*opos = BC_CONSTENC( add_const_r( C, func, val ) );
	}
	else if( *node->token == SGS_ST_STRING )
	{
		uint32_t val;
		SGS_AS_UINT32( val, node->token + 1 );
		*opos = BC_CONSTENC( add_const_s( C, func, val, (const char*) node->token + 5 ) );
	}
	else
	{
		QINTERR( 1011 );
		return 0;
	}
	return 1;
}


static SGSBOOL compile_regcopy( SGS_FNTCMP_ARGS, size_t from, rcpos_t srcpos, rcpos_t dstpos )
{
	INSTR_WRITE( SGS_SI_SET, dstpos, srcpos, 0 );
	return 1;
}

static SGSBOOL compile_fcall( SGS_FNTCMP_ARGS, rcpos_t* out, int expect )
{
	/* IF (ternary-like) */
	if( sgsT_IsKeyword( node->child->token, "if" ) )
	{
		sgs_FTNode* n = node->child->next->child;
		int argc = 0;
		size_t csz1, csz2, csz3;
		rcpos_t exprpos = -1, srcpos = -1, retpos = -1;
		if( expect > 1 )
		{
			QPRINT( "'if' pseudo-function cannot be used as input for expression writes with multiple outputs" );
			return 0;
		}
		while( n )
		{
			argc++;
			n = n->next;
		}
		if( argc != 3 )
		{
			QPRINT( "'if' pseudo-function requires exactly 3 arguments" );
			return 0;
		}
		
		if( expect )
			retpos = comp_reg_alloc( C );
		
		n = node->child->next->child;
		SGS_FN_ENTER;
		if( !compile_node_r( C, func, n, &exprpos ) || exprpos < 0 ) return 0;
		
		n = n->next;
		INSTR_WRITE_PCH();
		csz1 = func->code.size;
		if( !compile_node_r( C, func, n, &srcpos ) ||
			!compile_regcopy( C, func, n, csz1, srcpos, retpos ) ) return 0;
		
		n = n->next;
		INSTR_WRITE_PCH();
		csz2 = func->code.size;
		if( !compile_node_r( C, func, n, &srcpos ) ||
			!compile_regcopy( C, func, n, csz2, srcpos, retpos ) ) return 0;
		
		csz3 = func->code.size;
		
		INSTR_WRITE_EX( SGS_SI_NOP, 0, 0 ); /* harmful optimization prevention hack */
		{
			uint32_t instr1, instr2;
			instr1 = SGS_INSTR_MAKE_EX( SGS_SI_JMPF, ( csz2 - csz1 ) / SGS_INSTR_SIZE, exprpos );
			instr2 = SGS_INSTR_MAKE_EX( SGS_SI_JUMP, ( csz3 - csz2 ) / SGS_INSTR_SIZE, 0 );
			memcpy( func->code.ptr + csz1 - 4, &instr1, sizeof(instr1) );
			memcpy( func->code.ptr + csz2 - 4, &instr2, sizeof(instr2) );
		}
		
		*out = retpos;
		return 1;
	}
	/* SYNC/RACE (wait until all threads have finished/at least one of threads has finished) */
	else if( sgsT_IsKeyword( node->child->token, "sync" ) || sgsT_IsKeyword( node->child->token, "race" ) )
	{
		int i, argc = 0;
		size_t csz1;
		rcpos_t boolpos, srcpos;
		sgs_FTNode* n = node->child->next->child;
		int israce = sgsT_IsKeyword( node->child->token, "race" );
		
		if( expect > 1 )
		{
			QPRINT( "'sync' pseudo-function cannot be used as input for expression writes with multiple outputs" );
			return 0;
		}
		while( n )
		{
			argc++;
			n = n->next;
		}
		if( argc < 1 )
		{
			QPRINT( "'sync' pseudo-function requires [1;255] arguments" );
			return 0;
		}
		
		C->fctx->syncdepth++;
		if( C->fctx->syncdepth == 1 )
		{
			INSTR_WRITE( SGS_SI_INT, 0, 0, SGS_INT_RESET_WAIT_TIMER );
		}
		
		csz1 = func->code.size; /* the jump-back spot */
		
		boolpos = comp_reg_alloc( C );
		/* initialize bool to 0 if race and 1 if sync */
		INSTR_WRITE( SGS_SI_SET, boolpos, BC_CONSTENC( add_const_b( C, func, !israce ) ), 0 );
		
		i = 0;
		n = node->child->next->child;
		while( n )
		{
			if( !compile_node_r( C, func, n, &srcpos ) )
			{
				C->fctx->syncdepth--;
				return 0;
			}
			/* set bool to 1 if finished in race or 0 if not finished in sync */
			INSTR_WRITE( israce ? SGS_SI_COTRT : SGS_SI_COTRF, boolpos, srcpos, 0 );
			i++;
			n = n->next;
		}
		
		if( israce )
		{
			INSTR_WRITE_EX( SGS_SI_COABORT, ( csz1 - func->code.size ) / SGS_INSTR_SIZE - 1, boolpos );
		}
		
		if( C->fctx->syncdepth == 1 )
		{
			INSTR_WRITE_EX( SGS_SI_YLDJMP, ( csz1 - func->code.size ) / SGS_INSTR_SIZE - 1, boolpos );
		}
		C->fctx->syncdepth--;
		
		if( out )
			*out = boolpos;
		comp_reg_unwind( C, boolpos + expect );
		return 1;
	}
	else
	{
		sgs_FTNode* n;
		int i = 0, gotthis = 0, regc = 0, pfxfuncmode = 0, isthreadfunc;
		rcpos_t argpos, funcpos, fnargpos, objpos, argend;
		
		if( node->type == SGS_SFT_THRCALL ) pfxfuncmode = 1;
		if( node->type == SGS_SFT_STHCALL ) pfxfuncmode = 2;
		if( node->type == SGS_SFT_NEWCALL ) pfxfuncmode = 3;
		isthreadfunc = pfxfuncmode == 1 || pfxfuncmode == 2;
		
		/* count the required number of registers */
		if( isthreadfunc || ( node->child->type == SGS_SFT_OPER && /* thread always has 'this' */
			( *node->child->token == SGS_ST_OP_MMBR || *node->child->token == SGS_ST_OP_NOT ) ) )
		{
			gotthis = 1;
			regc++;
		}
		n = node->child->next->child;
		while( n )
		{
			regc++;
			n = n->next;
		}
		regc++; /* function register */
		if( isthreadfunc )
			regc++; /* 'thread_create'/'subthread_create' function register */
		argpos = comp_reg_alloc_n( C, regc > expect ? regc : expect );
		funcpos = argpos;
		if( isthreadfunc )
		{
			const char* key = pfxfuncmode == 1 ? "thread_create" : "subthread_create";
			rcpos_t strpos = add_const_s( C, func, (uint32_t) strlen( key ), key );
			INSTR_WRITE( SGS_SI_GETVAR, argpos, BC_CONSTENC( strpos ), 0 );
			gotthis = 0;
			funcpos++;
		}
		
		argend = argpos + regc - 1;
		objpos = funcpos + 1;
		fnargpos = funcpos + ( gotthis || isthreadfunc ? 2 : 1 );
		
		/* return register positions for expected data */
		if( out )
			*out = argpos;
		
		/* load function (for properties, object too) */
		if( node->child->type == SGS_SFT_OPER &&
			( *node->child->token == SGS_ST_OP_MMBR || *node->child->token == SGS_ST_OP_NOT ) )
		{
			sgs_FTNode* ncc = node->child->child;
			rcpos_t proppos = -1;
			SGS_FN_ENTER;
			if( !compile_node_rrw( C, func, ncc, objpos ) ) return 0; /* read object */
			if( *node->child->token == SGS_ST_OP_MMBR )
			{
				if( ncc->next->type == SGS_SFT_IDENT )
				{
					/* make string property key constant */
					compile_ident( C, func, ncc->next, &proppos );
				}
				else
				{
					SGS_FN_ENTER;
					/* load property key */
					if( !compile_node_r( C, func, ncc->next, &proppos ) ) return 0;
				}
				/* function as property of object */
				INSTR_WRITE( SGS_SI_GETPROP, funcpos, objpos, proppos );
			}
			else
			{
				SGS_FN_ENTER;
				/* function from own variable */
				if( !compile_node_rrw( C, func, ncc->next, funcpos ) ) return 0;
			}
		}
		else
		{
			if( objpos != fnargpos )
			{
				/* object does not apply, insert null at position */
				rcpos_t nullpos = add_const_null( C, func );
				INSTR_WRITE_EX( SGS_SI_LOADCONST, nullpos, objpos );
			}
			SGS_FN_ENTER;
			/* function from own variable */
			if( !compile_node_rrw( C, func, node->child, funcpos ) ) return 0;
		}
		
		/* load arguments */
		i = 0;
		{
			/* passing objects where the call is formed appropriately */
			n = node->child->next->child;
			while( n )
			{
				SGS_FN_ENTER;
				if( !compile_node_rrw( C, func, n, fnargpos + i ) ) return 0;
				i++;
				n = n->next;
			}
		}
		
		/* compile call */
		if( pfxfuncmode == 3 )
		{
			/* new <class/ctor>(..) */
			INSTR_WRITE( SGS_SI_NEW, 0, argpos, argend );
		}
		else
		{
			INSTR_WRITE( SGS_SI_CALL, expect, gotthis ? argpos | 0x100 : argpos, argend );
		}
		
		comp_reg_unwind( C, argpos + expect );

		return 1;
	}
}

static SGSBOOL compile_index_r( SGS_FNTCMP_ARGS, rcpos_t* out )
{
	rcpos_t var, name, opos = comp_reg_alloc( C );
	rcpos_t regpos = C->fctx->regs;
	SGS_FN_ENTER;
	if( !compile_node_r( C, func, node->child, &var ) ) return 0;
	SGS_FN_ENTER;
	if( !compile_node_r( C, func, node->child->next, &name ) ) return 0;
	INSTR_WRITE( SGS_SI_GETINDEX, opos, var, name );
	comp_reg_unwind( C, regpos );
	if( out )
		*out = opos;
	return 1;
}

static SGSBOOL compile_index_w( SGS_FNTCMP_ARGS, rcpos_t src )
{
	rcpos_t var, name;
	rcpos_t regpos = C->fctx->regs;
	SGS_FN_ENTER;
	if( !compile_node_r( C, func, node->child, &var ) ) return 0;
	SGS_FN_ENTER;
	if( !compile_node_r( C, func, node->child->next, &name ) ) return 0;
	if( SGS_CONSTVAR( var ) )
	{
		QPRINT( "Cannot set indexed value of a constant" );
		return 0;
	}
	INSTR_WRITE( SGS_SI_SETINDEX, var, name, src );
	comp_reg_unwind( C, regpos );
	return 1;
}

static SGSBOOL compile_clspfx_w( SGS_FNTCMP_ARGS, rcpos_t src )
{
	rcpos_t var, name;
	rcpos_t regpos = C->fctx->regs;
	SGS_FN_ENTER;
	if( !compile_node_r( C, func, node, &var ) ) return 0;
	SGS_FN_ENTER;
	compile_ident( C, func, node->child, &name );
	INSTR_WRITE( SGS_SI_SETINDEX, var, name, src );
	comp_reg_unwind( C, regpos );
	return 1;
}

static SGSBOOL compile_midxset( SGS_FNTCMP_ARGS, rcpos_t* out, int isprop )
{
	sgs_FTNode* mapi;
	rcpos_t var, name, src;
	rcpos_t regpos = C->fctx->regs, regpos2;
	SGS_FN_ENTER;
	if( !compile_node_r( C, func, node->child, &var ) ) return 0;
	regpos2 = C->fctx->regs;
	mapi = node->child->next->child;
	while( mapi )
	{
		if( *mapi->token == SGS_ST_STRING )
		{
			compile_const( C, func, mapi, &name );
		}
		else
		{
			compile_ident( C, func, mapi, &name );
		}
		mapi = mapi->next;
		
		SGS_FN_ENTER;
		if( !compile_node_r( C, func, mapi, &src ) ) return 0;
		mapi = mapi->next;
		
		INSTR_WRITE( isprop ? SGS_SI_SETPROP : SGS_SI_SETINDEX, var, name, src );
		comp_reg_unwind( C, regpos2 );
	}
	if( out )
		*out = var;
	else
		comp_reg_unwind( C, regpos );
	return 1;
}


static SGSBOOL try_optimize_last_instr_out( SGS_FNTCMP_ARGS, size_t ioff, rcpos_t* out )
{
	rcpos_t pos = -1;
	
	SGS_FN_BEGIN;
	SGS_UNUSED( C );
	
	if( ( node->type != SGS_SFT_IDENT && node->type != SGS_SFT_ARGMT ) || *node->token != SGS_ST_IDENT )
		goto cannot;
	
	/* moved offset 4 to other side of equation to prevent unsigned underflow */
	if( ioff + 4 > func->code.size )
		goto cannot;
	
	ioff = func->code.size - 4;
	
	/* check if closure variable */
	pos = find_varT( &C->fctx->clsr, node->token );
	if( pos >= 0 )
		goto cannot;
	
	/* find the variable output register */
	if( C->fctx->func )
	{
		rcpos_t gpos = find_varT( &C->fctx->gvars, node->token );
		if( gpos >= 0 )
			pos = -1;
		else
		{
			add_varT( &C->fctx->vars, C, node->token );
			pos = find_varT( &C->fctx->vars, node->token );
		}
	}
	else
	{
		pos = find_varT( &C->fctx->vars, node->token );
	}
	
	/* global variable */
	if( pos < 0 )
		goto cannot;
	
	{
		sgs_instr_t I;
		SGS_AS_( I, func->code.ptr + ioff, sgs_instr_t );
		int op = SGS_INSTR_GET_OP( I ), argB = SGS_INSTR_GET_B( I ), argC = SGS_INSTR_GET_C( I );
		switch( op )
		{
		case SGS_SI_GETVAR: case SGS_SI_GETPROP: case SGS_SI_GETINDEX:
		case SGS_SI_MAKECLSR: case SGS_SI_GETCLSR:
		case SGS_SI_SET: case SGS_SI_CONCAT:
		case SGS_SI_NEGATE: case SGS_SI_BOOL_INV: case SGS_SI_INVERT:
		case SGS_SI_ADD: case SGS_SI_SUB: case SGS_SI_MUL: case SGS_SI_DIV: case SGS_SI_MOD:
		case SGS_SI_AND: case SGS_SI_OR: case SGS_SI_XOR: case SGS_SI_LSH: case SGS_SI_RSH:
		case SGS_SI_SEQ: case SGS_SI_EQ: case SGS_SI_LT: case SGS_SI_LTE:
		case SGS_SI_SNEQ: case SGS_SI_NEQ: case SGS_SI_GT: case SGS_SI_GTE: case SGS_SI_RAWCMP:
			{
				char* dummy0 = NULL;
				unsigned dummy1 = 0;
				if( find_nth_var( &C->fctx->vars, SGS_INSTR_GET_A( I ), &dummy0, &dummy1 ) )
					goto cannot;
			}
			I = SGS_INSTR_MAKE( op, pos, argB, argC );
			memcpy( func->code.ptr + ioff, &I, sizeof(I) );
			if( out )
				*out = pos;
			break;
		case SGS_SI_ARRAY: case SGS_SI_DICT: case SGS_SI_MAP:
			{
				int argE = SGS_INSTR_GET_E( I );
				char* dummy0 = NULL;
				unsigned dummy1 = 0;
				if( find_nth_var( &C->fctx->vars, SGS_INSTR_GET_A( I ), &dummy0, &dummy1 ) )
					goto cannot;
				I = SGS_INSTR_MAKE_EX( op, argE, pos );
				memcpy( func->code.ptr + ioff, &I, sizeof(I) );
				if( out )
					*out = pos;
			}
			break;
		default:
			goto cannot;
		}
	}
	
	SGS_FN_END;
	return 1;
	
cannot:
	SGS_FN_END;
	return 0;
}

static SGSBOOL try_optimize_set_op( SGS_CTX, sgs_CompFunc* func, size_t ioff, rcpos_t ireg )
{
	SGS_FN_BEGIN;
	SGS_UNUSED( C );
	
	/* moved offset 4 to other side of equation to prevent unsigned underflow */
	if( ioff + 4 > func->code.size )
		goto cannot;
	
	ioff = func->code.size - 4;
	
	{
		sgs_instr_t I;
		SGS_AS_( I, func->code.ptr + ioff, sgs_instr_t );
		int op = SGS_INSTR_GET_OP( I ), argB = SGS_INSTR_GET_B( I ), argC = SGS_INSTR_GET_C( I );
		switch( op )
		{
		case SGS_SI_GETVAR: case SGS_SI_GETPROP: case SGS_SI_GETINDEX:
		case SGS_SI_MAKECLSR: case SGS_SI_GETCLSR:
		case SGS_SI_SET: case SGS_SI_CONCAT:
		case SGS_SI_NEGATE: case SGS_SI_BOOL_INV: case SGS_SI_INVERT:
		case SGS_SI_ADD: case SGS_SI_SUB: case SGS_SI_MUL: case SGS_SI_DIV: case SGS_SI_MOD:
		case SGS_SI_AND: case SGS_SI_OR: case SGS_SI_XOR: case SGS_SI_LSH: case SGS_SI_RSH:
		case SGS_SI_SEQ: case SGS_SI_EQ: case SGS_SI_LT: case SGS_SI_LTE:
		case SGS_SI_SNEQ: case SGS_SI_NEQ: case SGS_SI_GT: case SGS_SI_GTE: case SGS_SI_RAWCMP:
			{
				char* dummy0 = NULL;
				unsigned dummy1 = 0;
				if( find_nth_var( &C->fctx->vars, SGS_INSTR_GET_A( I ), &dummy0, &dummy1 ) )
					goto cannot;
			}
			I = SGS_INSTR_MAKE( op, ireg, argB, argC );
			memcpy( func->code.ptr + ioff, &I, sizeof(I) );
			break;
		case SGS_SI_ARRAY: case SGS_SI_DICT: case SGS_SI_MAP:
			{
				int argE = SGS_INSTR_GET_E( I );
				char* dummy0 = NULL;
				unsigned dummy1 = 0;
				if( find_nth_var( &C->fctx->vars, SGS_INSTR_GET_C( I ), &dummy0, &dummy1 ) )
					goto cannot;
				I = SGS_INSTR_MAKE_EX( op, argE, ireg );
				memcpy( func->code.ptr + ioff, &I, sizeof(I) );
			}
			break;
		default:
			goto cannot;
		}
	}
	
	SGS_FN_END;
	return 1;
	
cannot:
	SGS_FN_END;
	return 0;
}

static SGSBOOL compile_node_rrw( SGS_FNTCMP_ARGS, rcpos_t dst )
{
	rcpos_t ireg = -1, bkup = C->fctx->regs;
	size_t newcodestart = func->code.size;
	
	SGS_FN_ENTER;
	if( !compile_node_r( C, func, node, &ireg ) ) return 0;
	
	SGS_FN_ENTER;
	if( !try_optimize_set_op( C, func, newcodestart, dst ) )
	{
		/* just set the contents */
		SGS_FN_ENTER;
		INSTR_WRITE( SGS_SI_SET, dst, ireg, 0 );
	}
	
	comp_reg_unwind( C, bkup );
	
	return 1;
}

static int compile_mconcat( SGS_FNTCMP_ARGS, rcpos_t* arg )
{
	int numch = 0;
	sgs_FTNode* cur = node->child;
	rcpos_t ireg, oreg = comp_reg_alloc( C );
	if( C->state & SGS_MUST_STOP )
		return 0;
	
	/* get source data registers */
	while( cur )
	{
		int32_t bkup = C->fctx->regs;
		
		SGS_FN_ENTER;
		if( !compile_node_r( C, func, cur, &ireg ) )
			return 0;
		INSTR_WRITE( SGS_SI_PUSH, 0, ireg, 0 );
		numch++;
		cur = cur->next;
		
		comp_reg_unwind( C, bkup );
	}
	
	INSTR_WRITE( SGS_SI_MCONCAT, oreg, numch, 0 );
	
	if( arg )
		*arg = oreg;
	return 1;
}

static SGSBOOL compile_oper( SGS_FNTCMP_ARGS, rcpos_t* arg, int out, int expect )
{
	int assign = SGS_ST_OP_ASSIGN( *node->token );
	SGS_FN_BEGIN;
	
	/* Error suppression op */
	if( *node->token == SGS_ST_OP_ERSUP )
	{
		size_t csz;
		INSTR_WRITE( SGS_SI_INT, 0, 0, SGS_INT_ERRSUP_INC );
		csz = func->code.size;
		
		if( out && expect )
		{
			SGS_FN_ENTER;
			if( !compile_node_r( C, func, node->child, arg ) ) goto fail;
		}
		else
		{
			SGS_FN_ENTER;
			if( !compile_node( C, func, node->child ) ) goto fail;
		}
		
		if( func->code.size > csz )
		{
			/* write INT ERRSUP_DEC if any code was written */
			INSTR_WRITE( SGS_SI_INT, 0, 0, SGS_INT_ERRSUP_DEC );
		}
		else
		{
			/* otherwise, remove the already written IN ERRSUP_INC */
			func->code.size -= 4;
		}
		return 1;
	}
	/* Boolean ops */
	else if( SGS_ST_OP_BOOL( *node->token ) || SGS_ST_OP_FNN( *node->token ) )
	{
		int jin;
		rcpos_t ireg1, ireg2, oreg = 0;
		size_t csz, csz2;
		
		if( !assign )
			oreg = comp_reg_alloc( C );
		if( C->state & SGS_MUST_STOP )
			goto fail;
		
		/* get source data register */
		SGS_FN_ENTER;
		if( !compile_node_r( C, func, node->child, &ireg1 ) ) goto fail;
		
		/* write cond. jump */
		jin = ( *node->token == SGS_ST_OP_BLAND || *node->token == SGS_ST_OP_BLAEQ ) ? SGS_SI_JMPT :
			( ( *node->token == SGS_ST_OP_BLOR || *node->token == SGS_ST_OP_BLOEQ ) ? SGS_SI_JMPF : SGS_SI_JMPN );
		INSTR_WRITE_PCH();
		csz = func->code.size;
		
		/* compile write of value 1 */
		if( assign )
		{
			SGS_FN_ENTER;
			if( !compile_node_w( C, func, node->child, ireg1 ) ) goto fail;
		}
		else
		{
			INSTR_WRITE( SGS_SI_SET, oreg, ireg1, 0 );
		}
		
		INSTR_WRITE_PCH();
		csz2 = func->code.size;
		
		/* fix-up jump 1 */
		{
			sgs_instr_t instr;
			/* WP: instruction limit */
			ptrdiff_t jmp_off = (ptrdiff_t) ( func->code.size - csz ) / SGS_INSTR_SIZE;
			instr = SGS_INSTR_MAKE_EX( jin, jmp_off, ireg1 );
			memcpy( func->code.ptr + csz - 4, &instr, sizeof(instr) );
		}
		
		/* get source data register 2 */
		SGS_FN_ENTER;
		if( !compile_node_r( C, func, node->child->next, &ireg2 ) ) goto fail;
		
		/* compile write of value 2 */
		if( assign )
		{
			SGS_FN_ENTER;
			if( !compile_node_w( C, func, node->child, ireg2 ) ) goto fail;
		}
		else
		{
			INSTR_WRITE( SGS_SI_SET, oreg, ireg2, 0 );
		}
		
		INSTR_WRITE( SGS_SI_NOP, 0, 0, 0 );
		
		/* fix-up jump 2 */
		{
			sgs_instr_t instr;
			/* WP: instruction limit */
			ptrdiff_t jmp_off = (ptrdiff_t) ( func->code.size - csz2 ) / SGS_INSTR_SIZE;
			instr = SGS_INSTR_MAKE_EX( SGS_SI_JUMP, jmp_off, 0 );
			memcpy( func->code.ptr + csz2 - 4, &instr, sizeof(instr) );
		}
		
		/* re-read from assignments */
		if( arg )
		{
			if( assign )
			{
				SGS_FN_ENTER;
				if( !compile_node_r( C, func, node->child, arg ) ) goto fail;
			}
			else
				*arg = oreg;
		}
	}
	else
	/* Increment / decrement */
	if( *node->token == SGS_ST_OP_INC || *node->token == SGS_ST_OP_DEC )
	{
		rcpos_t ireg, oreg;
		
		/* register with input data */
		SGS_FN_ENTER;
		if( !compile_node_r( C, func, node->child, &ireg ) ) goto fail;
		
		/* output register selection */
		oreg = expect && node->type == SGS_SFT_OPER_P ? comp_reg_alloc( C ) : ireg;
		if( C->state & SGS_MUST_STOP )
			goto fail;
		if( oreg != ireg )
		{
			INSTR_WRITE( SGS_SI_SET, oreg, ireg, 0 );
		}
		
		/* check for errors if this operator generates output */
		if( expect )
		{
			if( expect != 1 )
			{
				QPRINT( "Too many expected outputs for operator" );
				goto fail;
			}
		}
		
		/* write bytecode */
		INSTR_WRITE( *node->token == SGS_ST_OP_INC ? SGS_SI_INC : SGS_SI_DEC, ireg, ireg, 0 );
		
		if( arg )
			*arg = oreg;
		
		/* compile writeback */
		SGS_FN_ENTER;
		if( !compile_node_w( C, func, node->child, ireg ) ) goto fail;
	}
	/* Assignment */
	else if( assign )
	{
		/* 1 operand */
		if( *node->token == SGS_ST_OP_SET )
		{
			rcpos_t ireg;
			size_t isb = func->code.size;
			
			if( node->child->type == SGS_SFT_EXPLIST )
			{
				sgs_FTNode* n;
				int i, xpct = 0;
				int32_t bkup;
				rcpos_t freg;
				if( node->child->next->type != SGS_SFT_FCALL )
				{
					QPRINT( "Expression writes only allowed with function call reads" );
					goto fail;
				}
				
				/* multiwrite */
				n = node->child->child;
				while( n )
				{
					xpct++;
					n = n->next;
				}
				
				if( !compile_fcall( C, func, node->child->next, &freg, xpct ) ) goto fail;
				
				bkup = C->fctx->regs;
				n = node->child->child;
				for( i = 0; i < xpct; ++i )
				{
					SGS_FN_ENTER;
					if( !compile_node_w( C, func, n, freg + i ) ) goto fail;
					
					comp_reg_unwind( C, bkup );
					
					n = n->next;
				}
			}
			else
			{
				/* get source data register */
				SGS_FN_ENTER;
				if( !compile_node_r( C, func, node->child->next, &ireg ) ) goto fail;
				
				SGS_FN_ENTER;
				if( !try_optimize_last_instr_out( C, func, node->child, isb, arg ) )
				{
					/* just set the contents */
					SGS_FN_ENTER;
					if( !compile_node_w( C, func, node->child, ireg ) ) goto fail;
				}
				
				if( arg )
				{
					SGS_FN_ENTER;
					if( !compile_node_r( C, func, node->child, arg ) ) goto fail;
				}
			}
		}
		/* 3+ operands (MCONCAT only) */
		else if( *node->token == SGS_ST_OP_CATEQ && node->child &&
			node->child->next && node->child->next->next )
		{
			rcpos_t oreg;
			SGS_FN_ENTER;
			if( !compile_mconcat( C, func, node, &oreg ) )
				goto fail;
			if( arg )
				*arg = oreg;
			
			/* compile write */
			SGS_FN_ENTER;
			if( !compile_node_w( C, func, node->child, oreg ) ) goto fail;
		}
		/* 2 operands */
		else
		{
			int op;
			size_t isb = func->code.size;
			rcpos_t ireg1, ireg2, oreg = comp_reg_alloc( C );
			if( C->state & SGS_MUST_STOP )
				goto fail;
			
			if( !node->child || !node->child->next )
			{
				QINTERR( 1012 );
				goto fail;
			}
			
			/* get source data registers */
			SGS_FN_ENTER;
			if( !compile_node_r( C, func, node->child, &ireg1 ) ) goto fail;
			SGS_FN_ENTER;
			if( !compile_node_r( C, func, node->child->next, &ireg2 ) ) goto fail;
			
			/* compile op */
			op = op_pick_opcode( *node->token, 1 );
			INSTR_WRITE( op, oreg, ireg1, ireg2 );
			
			SGS_FN_ENTER;
			if( !try_optimize_last_instr_out( C, func, node->child, isb, arg ) )
			{
				/* just set the contents */
				SGS_FN_ENTER;
				if( !compile_node_w( C, func, node->child, oreg ) ) goto fail;
			}
			
			if( arg )
			{
				SGS_FN_ENTER;
				if( !compile_node_r( C, func, node->child, arg ) ) goto fail;
			}
		}
	}
	/* Any other */
	else
	{
		rcpos_t ireg1, ireg2, oreg;
		
		if( expect > 1 )
		{
			QPRINT( "Too many expected outputs for operator" );
			goto fail;
		}
		
		if( /* no operands, unary used as binary, binary used as unary */
			( !node->child ) ||
			( SGS_ST_OP_UNARY( *node->token ) && !SGS_ST_OP_BINARY( *node->token ) && node->child->next ) ||
			( !SGS_ST_OP_UNARY( *node->token ) && SGS_ST_OP_BINARY( *node->token ) && !node->child->next )
		)
		{
			QPRINT( "Invalid expression" );
			goto fail;
		}
		
		if( *node->token == SGS_ST_OP_MMBR )
		{
			/* oreg points to output register if "out", source register otherwise */
			if( out )
			{
				oreg = comp_reg_alloc( C );
				if( C->state & SGS_MUST_STOP )
					goto fail;
				if( arg )
					*arg = oreg;
			}
			else
				oreg = *arg;
			
			/* get source data registers */
			SGS_FN_ENTER;
			if( !compile_node_r( C, func, node->child, &ireg1 ) ) goto fail;
			
			if( node->child->next->type == SGS_SFT_IDENT )
				compile_ident( C, func, node->child->next, &ireg2 );
			else
			{
				SGS_FN_ENTER;
				if( !compile_node_r( C, func, node->child->next, &ireg2 ) ) goto fail;
			}
			
			/* compile op */
			if( out )
				INSTR_WRITE( SGS_SI_GETPROP, oreg, ireg1, ireg2 );
			else
			{
				if( SGS_CONSTVAR( ireg1 ) )
				{
					QPRINT( "Cannot set property of a constant" );
					goto fail;
				}
				INSTR_WRITE( SGS_SI_SETPROP, ireg1, ireg2, oreg );
			}
		}
		/* 3+ operands (MCONCAT only) */
		else if( *node->token == SGS_ST_OP_CAT && node->child &&
			node->child->next && node->child->next->next )
		{
			SGS_FN_ENTER;
			if( !compile_mconcat( C, func, node, arg ) )
				goto fail;
		}
		else
		{
			int op;
			
			oreg = comp_reg_alloc( C );
			if( C->state & SGS_MUST_STOP )
				goto fail;
			
			/* get source data registers */
			SGS_FN_ENTER;
			if( !compile_node_r( C, func, node->child, &ireg1 ) ) goto fail;
			if( node->child->next )
			{
				SGS_FN_ENTER;
				if( !compile_node_r( C, func, node->child->next, &ireg2 ) ) goto fail;
			}
			
			if( arg )
				*arg = oreg;
			
			/* compile op */
			op = op_pick_opcode( *node->token, !!node->child->next );
			if( !op )
				INSTR_WRITE( SGS_SI_SET, oreg, ireg1, 0 );
			else if( node->child->next )
				INSTR_WRITE( op, oreg, ireg1, ireg2 );
			else
				INSTR_WRITE( op, oreg, ireg1, 0 );
		}
	}
	
	SGS_FN_END;
	return 1;
	
fail:
	C->state |= SGS_HAS_ERRORS;
	SGS_FN_END;
	return 0;
}


static SGSBOOL compile_breaks( SGS_FNTCMP_ARGS, uint8_t iscont )
{
	sgs_BreakInfo* binfo = C->fctx->binfo, *prev = NULL;
	while( binfo )
	{
		if( binfo->numlp == C->fctx->loops && binfo->iscont == iscont )
		{
			/* WP: jump limit */
			ptrdiff_t off = (ptrdiff_t) ( func->code.size - binfo->jdoff ) / SGS_INSTR_SIZE - 1;
			if( over_limit( off, 32767 ) )
			{
				QPRINT( "Max. jump limit exceeded (32767 instructions) @ break/continue; reduce size of loops" );
				return 0;
			}
			sgs_instr_t instr = SGS_INSTR_MAKE_EX( SGS_SI_JUMP, off, 0 );
			memcpy( func->code.ptr + binfo->jdoff, &instr, sizeof(instr) );
			binfo = binfo->next;
			fctx_binfo_rem( C, C->fctx, prev );
		}
		else
		{
			prev = binfo;
			binfo = binfo->next;
		}
	}
	return 1;
}




static void rpts( sgs_MemBuf* out, SGS_CTX, sgs_FTNode* root )
{
	switch( root->type )
	{
	case SGS_SFT_IDENT:
		sgs_membuf_appbuf( out, C, root->token + 2, root->token[1] );
		break;
	case SGS_SFT_CLSPFX:
		sgs_membuf_appbuf( out, C, root->token + 2, root->token[1] );
		sgs_membuf_appchr( out, C, '.' );
		rpts( out, C, root->child );
		break;
	case SGS_SFT_OPER:
		switch( *root->token )
		{
		case SGS_ST_OP_MMBR:
			rpts( out, C, root->child );
			sgs_membuf_appchr( out, C, '.' );
			rpts( out, C, root->child->next );
			break;
		}
		break;
	}
}


static void prefix_bytecode( SGS_CTX, sgs_CompFunc* func, int args )
{
	sgs_MemBuf ncode = sgs_membuf_create();
	sgs_MemBuf nlnbuf = sgs_membuf_create();
	
	if( C->fctx->outclsr > C->fctx->inclsr )
	{
		int i;
		sgs_instr_t I;
		uint16_t ln = 0;

		for( i = 0; i < args; ++i )
		{
			char* varstr = NULL;
			unsigned varstrlen;
			int result, which;
			result = find_nth_var( &C->fctx->vars, i, &varstr, &varstrlen );
			sgs_BreakIf( !result );
			if( !result )
				continue;
			which = find_var( &C->fctx->clsr, varstr, varstrlen );
			if( which < 0 )
				continue;
			I = SGS_INSTR_MAKE( SGS_SI_SETCLSR, 0, which, i );
			sgs_membuf_appbuf( &ncode, C, &I, sizeof( I ) );
			sgs_membuf_appbuf( &nlnbuf, C, &ln, sizeof( ln ) );
		}
	}

	sgs_membuf_appbuf( &ncode, C, func->code.ptr, func->code.size );
	sgs_membuf_appbuf( &nlnbuf, C, func->lnbuf.ptr, func->lnbuf.size );
	
	/* append one extra return instruction to avoid VM bounds check */
	{
		sgs_instr_t I = SGS_INSTR_MAKE( SGS_SI_RETN, 0, 0, 0 );
		uint16_t ln = 0;
		sgs_membuf_appbuf( &ncode, C, &I, sizeof( I ) );
		sgs_membuf_appbuf( &nlnbuf, C, &ln, sizeof( ln ) );
	}
	
	sgs_membuf_destroy( &func->code, C );
	sgs_membuf_destroy( &func->lnbuf, C );
	
	/* WP: both lastreg and args cannot exceed 255, lastreg includes args */
	func->numtmp = (uint8_t) ( C->fctx->lastreg - args );
	func->code = ncode;
	func->lnbuf = nlnbuf;
}


static int compile_fn_base( SGS_FNTCMP_ARGS, int args )
{
	BLOCK_BEGIN;
	
	SGS_FN_ENTER;
	if( !preparse_clsrlists( C, func, node ) ) return 0;
	
	SGS_FN_ENTER;
	if( !preparse_varlists( C, func, node ) ) return 0;
	args += func->gotthis;
	
	SGS_FN_ENTER;
	if( !preparse_funcorder( C, func, node ) ) return 0;
	
	SGS_FN_ENTER;
	if( !compile_node( C, func, node ) ) return 0;
	
	comp_reg_unwind( C, 0 );
	
	if( C->fctx->lastreg > 0xff )
	{
		QPRINT( "Max. register count exceeded" );
		return 0;
	}
	if( C->fctx->outclsr > 0xff )
	{
		QPRINT( "Max. closure count exceeded" );
		return 0;
	}
	
	BLOCK_END;
	
	prefix_bytecode( C, func, args );
	/* WP: closure limit */
	func->numclsr = (uint8_t) C->fctx->outclsr;
	func->inclsr = (uint8_t) C->fctx->inclsr;
	
#if SGS_DUMP_BYTECODE || ( SGS_DEBUG && SGS_DEBUG_DATA )
	fctx_dump( C->fctx );
	sgsBC_Dump( func );
#endif
	
	return 1;
}

static SGSBOOL compile_func( SGS_FNTCMP_ARGS, rcpos_t* out )
{
	sgs_FuncCtx* fctx = fctx_create( C ), *bkfctx = C->fctx;
	sgs_CompFunc* nf = &fctx->cfunc;

	sgs_FTNode* n_arglist = node->child;
	sgs_FTNode* n_uselist = n_arglist->next;
	sgs_FTNode* n_body = n_uselist->next;
	sgs_FTNode* n_name = n_body->next;

	/* pre-context-change closure-apply */
	SGS_FN_ENTER;
	if( !preparse_closures( C, func, n_uselist, 0 ) ) { goto fail; }

	C->fctx = fctx;

	SGS_FN_ENTER;
	if( !preparse_closures( C, nf, n_uselist, 1 ) ) { goto fail; }

	SGS_FN_ENTER;
	if( !preparse_arglist( C, nf, n_arglist ) ) { goto fail; }
	
	if( !compile_fn_base( C, nf, n_body, fctx->regs ) ) goto fail;
	
	C->fctx = bkfctx;

	{
		sgs_MemBuf ffn = sgs_membuf_create();
		if( n_name )
			rpts( &ffn, C, n_name );
		*out = BC_CONSTENC( add_const_f( C, func, nf, ffn.ptr, ffn.size, sgsT_LineNum( node->token ) ) );
		sgs_membuf_destroy( &ffn, C );
		
		if( fctx->outclsr > 0 )
		{
			int i;
			rcpos_t ro = comp_reg_alloc( C );
			INSTR_WRITE( SGS_SI_MAKECLSR, ro, *out, fctx->inclsr );
			if( fctx->inclsr )
			{
				sgs_FTNode* uli = n_uselist->child;
				for( i = 0; i < fctx->inclsr; i += 3 )
				{
					int a, b = 0, c = 0;
					a = find_varT( &bkfctx->clsr, uli->token );
					uli = uli->next;
					if( uli )
					{
						b = find_varT( &bkfctx->clsr, uli->token );
						uli = uli->next;
						if( uli )
						{
							c = find_varT( &bkfctx->clsr, uli->token );
							uli = uli->next;
						}
					}
					INSTR_WRITE( SGS_SI_CLSRINFO, a, b, c );
				}
			}
			*out = ro;
		}
	}
	fctx_destroy( C, fctx );
	return 1;

fail:
	C->fctx = bkfctx;
	fctx_destroy( C, fctx );
	C->state |= SGS_HAS_ERRORS;
	return 0;
}


static SGSBOOL compile_node_w( SGS_FNTCMP_ARGS, rcpos_t src )
{
	SGS_FN_BEGIN;
	switch( node->type )
	{
	case SGS_SFT_IDENT:
	case SGS_SFT_KEYWORD:
		SGS_FN_HIT( "W_IDENT" );
		if( !compile_ident_w( C, func, node, src ) ) goto fail;
		break;

	case SGS_SFT_CONST:
		SGS_FN_HIT( "W_CONST" );
		QPRINT( "Cannot write to constants" );
		goto fail;
	case SGS_SFT_FUNC:
		SGS_FN_HIT( "W_FUNC" );
		QPRINT( "Cannot write to constants" );
		goto fail;
	case SGS_SFT_ARRLIST:
		SGS_FN_HIT( "W_ARRLIST" );
		QPRINT( "Cannot write to constants" );
		goto fail;
	case SGS_SFT_DCTLIST:
		SGS_FN_HIT( "W_DCTLIST" );
		QPRINT( "Cannot write to constants" );
		goto fail;
	case SGS_SFT_MAPLIST:
		SGS_FN_HIT( "W_MAPLIST" );
		QPRINT( "Cannot write to constants" );
		goto fail;

	case SGS_SFT_OPER:
	case SGS_SFT_OPER_P:
		SGS_FN_HIT( "W_OPER" );
		if( !compile_oper( C, func, node, &src, 0, 1 ) ) goto fail;
		break;

	case SGS_SFT_FCALL:
	case SGS_SFT_THRCALL:
	case SGS_SFT_STHCALL:
	case SGS_SFT_NEWCALL:
		SGS_FN_HIT( "W_FCALL" );
		if( !compile_fcall( C, func, node, NULL, 0 ) ) goto fail;
		break;

	case SGS_SFT_INDEX:
		SGS_FN_HIT( "W_INDEX" );
		if( !compile_index_w( C, func, node, src ) ) goto fail;
		break;
		
	case SGS_SFT_MIDXSET:
		SGS_FN_HIT( "MIDXSET" );
		QPRINT( "Cannot write to multi-index-set expression" );
		break;
		
	case SGS_SFT_MPROPSET:
		SGS_FN_HIT( "MPROPSET" );
		QPRINT( "Cannot write to multi-property-set expression" );
		break;
		
	case SGS_SFT_EXPLIST:
		SGS_FN_HIT( "W_EXPLIST" );
		QPRINT( "Expression writes only allowed with function call reads" );
		goto fail;
		
	case SGS_SFT_CLSPFX:
		SGS_FN_HIT( "W_CLSPFX" );
		if( !compile_clspfx_w( C, func, node, src ) ) goto fail;
		break;

	default:
		QINTERR( 1003 );
		goto fail;
	}
	SGS_FN_END;
	return 1;

fail:
	C->state |= SGS_HAS_ERRORS;
	SGS_FN_END;
	return 0;
}
static SGSBOOL compile_node_r( SGS_FNTCMP_ARGS, rcpos_t* out )
{
	SGS_FN_BEGIN;
	switch( node->type )
	{
	case SGS_SFT_IDENT:
	case SGS_SFT_KEYWORD:
	case SGS_SFT_CLSPFX:
		SGS_FN_HIT( "R_IDENT" );
		if( !compile_ident_r( C, func, node, out ) ) goto fail;
		break;

	case SGS_SFT_CONST:
		SGS_FN_HIT( "R_CONST" );
		if( !compile_const( C, func, node, out ) ) goto fail;
		break;
	case SGS_SFT_FUNC:
		SGS_FN_HIT( "R_FUNC" );
		if( !compile_func( C, func, node, out ) ) goto fail;
		break;
	case SGS_SFT_ARRLIST:
		SGS_FN_HIT( "R_ARRLIST" );
		{
			rcpos_t pos = 0;
			int args = 0, off = 0;
			sgs_FTNode* n = node->child;
			while( n )
			{
				args++;
				n = n->next;
			}
			pos = comp_reg_alloc( C );
			INSTR_WRITE_EX( SGS_SI_ARRAY, args, pos );
			n = node->child;
			while( n )
			{
				rcpos_t bkup = C->fctx->regs, vpos = 0;
				SGS_FN_ENTER;
				if( !compile_node_r( C, func, n, &vpos ) )
					goto fail;
				INSTR_WRITE( SGS_SI_ARRPUSH, pos, 0, vpos );
				comp_reg_unwind( C, bkup );
				off++;
				n = n->next;
			}
			*out = pos;
		}
		break;
	case SGS_SFT_DCTLIST:
	case SGS_SFT_MAPLIST:
		SGS_FN_HIT( "R_(DCT|MAP)LIST" );
		{
			int args = 0;
			sgs_FTNode* n = node->child;
			rcpos_t pos = comp_reg_alloc( C ), kpos, vpos, bkup;
			bkup = C->fctx->regs;
			INSTR_WRITE_EX( node->type == SGS_SFT_DCTLIST ? SGS_SI_DICT : SGS_SI_MAP, 0, pos );
			while( n )
			{
				if( args % 2 == 0 )
				{
					kpos = 0;
					if( n->type != SGS_SFT_ARGMT )
					{
						SGS_FN_ENTER;
						if( !compile_node_r( C, func, n, &kpos ) )
							goto fail;
					}
					else if( *n->token == SGS_ST_STRING )
						compile_const( C, func, n, &kpos );
					else
						compile_ident( C, func, n, &kpos );
				}
				else
				{
					vpos = 0;
					SGS_FN_ENTER;
					if( !compile_node_r( C, func, n, &vpos ) )
						goto fail;
				}
				if( args % 2 == 1 )
				{
					INSTR_WRITE( SGS_SI_SETINDEX, pos, kpos, vpos );
					comp_reg_unwind( C, bkup );
				}
				args++;
				n = n->next;
			}
			sgs_BreakIf( args % 2 != 0 );
			*out = pos;
		}
		break;

	case SGS_SFT_OPER:
	case SGS_SFT_OPER_P:
		SGS_FN_HIT( "R_OPER" );
		if( !compile_oper( C, func, node, out, 1, 1 ) ) goto fail;
		break;

	case SGS_SFT_FCALL:
	case SGS_SFT_THRCALL:
	case SGS_SFT_STHCALL:
	case SGS_SFT_NEWCALL:
		SGS_FN_HIT( "R_FCALL" );
		if( !compile_fcall( C, func, node, out, 1 ) ) goto fail;
		break;

	case SGS_SFT_INDEX:
		SGS_FN_HIT( "R_INDEX" );
		if( !compile_index_r( C, func, node, out ) ) goto fail;
		break;
		
	case SGS_SFT_MIDXSET:
		SGS_FN_HIT( "MIDXSET" );
		if( !compile_midxset( C, func, node, out, 0 ) ) goto fail;
		break;
		
	case SGS_SFT_MPROPSET:
		SGS_FN_HIT( "MPROPSET" );
		if( !compile_midxset( C, func, node, out, 1 ) ) goto fail;
		break;
		
	case SGS_SFT_EXPLIST:
		SGS_FN_HIT( "R_EXPLIST" );
		{
			sgs_FTNode* n = node->child;
			if( !n )
			{
				QPRINT( "Empty expression found" );
				goto fail;
			}
			while( n )
			{
				SGS_FN_ENTER;
				if( !compile_node_r( C, func, n, out ) )
					goto fail;
				n = n->next;
			}
		}
		break;

	default:
		QINTERR( 1002 );
		goto fail;
	}
	SGS_FN_END;
	return 1;

fail:
	SGS_FN_END;
	return 0;
}

static SGSBOOL compile_for_explist( SGS_FNTCMP_ARGS, rcpos_t* out )
{
	sgs_FTNode* n;

	SGS_FN_BEGIN;

	if( node->type != SGS_SFT_EXPLIST )
	{
		QINTERR( 1004 );
		goto fail;
	}

	SGS_FN_HIT( "Rs_EXPLIST" );

	n = node->child;
	while( n )
	{
		SGS_FN_ENTER;
		if( !compile_node_r( C, func, n, out ) )
			goto fail;
		n = n->next;
	}

	SGS_FN_END;
	return 1;

fail:
	SGS_FN_END;
	return 0;
}

static SGSBOOL compile_node( SGS_FNTCMP_ARGS )
{
	SGS_FN_BEGIN;

	switch( node->type )
	{
	/* ignore these items if they're leading in statements */
	case SGS_SFT_IDENT:
	case SGS_SFT_KEYWORD:
	case SGS_SFT_CONST:
	case SGS_SFT_ARRLIST:
	case SGS_SFT_DCTLIST:
	case SGS_SFT_MAPLIST:
		break;

	case SGS_SFT_OPER:
	case SGS_SFT_OPER_P:
		SGS_FN_HIT( "OPERATOR" );
		if( !compile_oper( C, func, node, NULL, 1, 0 ) ) goto fail;
		break;

	case SGS_SFT_INDEX:
		SGS_FN_HIT( "INDEX" );
		if( !compile_index_r( C, func, node, NULL ) ) goto fail;
		break;
		
	case SGS_SFT_MIDXSET:
		SGS_FN_HIT( "MIDXSET" );
		if( !compile_midxset( C, func, node, NULL, 0 ) ) goto fail;
		break;
		
	case SGS_SFT_MPROPSET:
		SGS_FN_HIT( "MPROPSET" );
		if( !compile_midxset( C, func, node, NULL, 1 ) ) goto fail;
		break;
		
	case SGS_SFT_FCALL:
	case SGS_SFT_THRCALL:
	case SGS_SFT_STHCALL:
	case SGS_SFT_NEWCALL:
		SGS_FN_HIT( "FCALL" );
		if( !compile_fcall( C, func, node, NULL, 0 ) ) goto fail;
		break;

	case SGS_SFT_EXPLIST:
		SGS_FN_HIT( "EXPLIST" );
		{
			sgs_FTNode* n = node->child;
			while( n )
			{
				SGS_FN_ENTER;
				if( !compile_node( C, func, n ) )
					goto fail;
				n = n->next;
			}
		}
		break;

	case SGS_SFT_RETURN:
		SGS_FN_HIT( "RETURN" );
		compile_defers( C, func, 0 );
		{
			rcpos_t regstate = C->fctx->regs;
			sgs_FTNode* n = node->child;
			if( n && n->next == NULL )
			{
				/* one value */
				rcpos_t arg = 0;
				SGS_FN_ENTER;
				if( !compile_node_r( C, func, n, &arg ) ) goto fail;
				INSTR_WRITE( SGS_SI_RET1, 0, 0, arg );
			}
			else
			{
				int num = 0;
				while( n )
				{
					rcpos_t arg = 0;
					SGS_FN_ENTER;
					if( !compile_node_r( C, func, n, &arg ) ) goto fail;
					INSTR_WRITE( SGS_SI_PUSH, 0, arg, 0 );
					n = n->next;
					num++;
				}
				INSTR_WRITE( SGS_SI_RETN, num, 0, 0 );
			}
			comp_reg_unwind( C, regstate );
		}
		break;

	case SGS_SFT_BLOCK:
		SGS_FN_HIT( "BLOCK" );
		node = node->child;
		BLOCK_BEGIN;
		while( node )
		{
			rcpos_t regstate = C->fctx->regs;
			SGS_FN_ENTER;
			if( !compile_node( C, func, node ) ) goto fail;
			node = node->next;
			comp_reg_unwind( C, regstate );
		}
		BLOCK_END;
		break;

	case SGS_SFT_IFELSE:
		SGS_FN_HIT( "IF/ELSE" );
		{
			rcpos_t arg = 0;
			rcpos_t regstate = C->fctx->regs;
			SGS_FN_ENTER;
			if( !compile_node_r( C, func, node->child, &arg ) ) goto fail;
			comp_reg_unwind( C, regstate );
			INSTR_WRITE_PCH();
			{
				size_t jp1, jp2 = 0, jp3 = 0;
				jp1 = func->code.size;

				regstate = C->fctx->regs;
				SGS_FN_ENTER;
				if( !compile_node( C, func, node->child->next ) ) goto fail;
				comp_reg_unwind( C, regstate );

				if( node->child->next->next )
				{
					INSTR_WRITE_PCH();
					jp2 = func->code.size;
					{
						sgs_instr_t instr;
						/* WP: jump limit */
						ptrdiff_t jmp_off = (ptrdiff_t) ( jp2 - jp1 ) / SGS_INSTR_SIZE;
						if( over_limit( jmp_off, 32767 ) )
						{
							QPRINT( "Max. jump limit exceeded (32767 instructions) @ if/else; reduce size of construct" );
							goto fail;
						}
						instr = SGS_INSTR_MAKE_EX( SGS_SI_JMPF, jmp_off, arg );
						memcpy( func->code.ptr + jp1 - 4, &instr, sizeof(instr) );
					}

					regstate = C->fctx->regs;
					SGS_FN_ENTER;
					if( !compile_node( C, func, node->child->next->next ) ) goto fail;
					jp3 = func->code.size;
					{
						sgs_instr_t instr;
						/* WP: jump limit */
						ptrdiff_t jmp_off = (ptrdiff_t) ( jp3 - jp2 ) / SGS_INSTR_SIZE;
						if( over_limit( jmp_off, 32767 ) )
						{
							QPRINT( "Max. jump limit exceeded (32767 instructions) @ if/else; reduce size of construct" );
							goto fail;
						}
						instr = SGS_INSTR_MAKE_EX( SGS_SI_JUMP, jmp_off, 0 );
						memcpy( func->code.ptr + jp2 - 4, &instr, sizeof(instr) );
					}
					comp_reg_unwind( C, regstate );
				}
				else
				{
					sgs_instr_t instr;
					/* WP: jump limit */
					ptrdiff_t jmp_off = (ptrdiff_t) ( func->code.size - jp1 ) / SGS_INSTR_SIZE;
					if( over_limit( jmp_off, 32767 ) )
					{
						QPRINT( "Max. jump limit exceeded (32767 instructions) @ if/else; reduce size of construct" );
						goto fail;
					}
					instr = SGS_INSTR_MAKE_EX( SGS_SI_JMPF, jmp_off, arg );
					memcpy( func->code.ptr + jp1 - 4, &instr, sizeof(instr) );
				}
			}
		}
		break;

	case SGS_SFT_WHILE:
		SGS_FN_HIT( "WHILE" );
		{
			size_t codesize;
			rcpos_t arg = -1;
			rcpos_t regstate = C->fctx->regs;
			C->fctx->loops++;
			codesize = func->code.size;
			SGS_FN_ENTER;
			if( !compile_node_r( C, func, node->child, &arg ) ) goto fail; /* test */
			comp_reg_unwind( C, regstate );
			INSTR_WRITE_PCH();
			{
				ptrdiff_t off;
				size_t jp1, jp2 = 0;
				jp1 = func->code.size;

				LOOP_BEGIN;
				regstate = C->fctx->regs;
				SGS_FN_ENTER;
				if( !compile_node( C, func, node->child->next ) ) goto fail; /* while */
				comp_reg_unwind( C, regstate );
				LOOP_END;

				if( !compile_breaks( C, func, node, 1 ) )
					goto fail;

				jp2 = func->code.size;
				/* WP: jump limit */
				off = (ptrdiff_t) ( codesize - jp2 ) / SGS_INSTR_SIZE - 1;
				if( over_limit( off, 32767 ) )
				{
					QPRINT( "Max. jump limit exceeded (32767 instructions) @ while; reduce size of loop" );
					goto fail;
				}
				INSTR_WRITE_EX( SGS_SI_JUMP, off, 0 );
				{
					sgs_instr_t instr;
					instr = SGS_INSTR_MAKE_EX( SGS_SI_JMPF, ( jp2 - jp1 ) / SGS_INSTR_SIZE + 1, arg );
					memcpy( func->code.ptr + jp1 - 4, &instr, sizeof(instr) );
				}
			}
			if( !compile_breaks( C, func, node, 0 ) )
				goto fail;
			C->fctx->loops--;
		}
		break;

	case SGS_SFT_DOWHILE:
		SGS_FN_HIT( "DO/WHILE" );
		{
			size_t codesize;
			rcpos_t regstate = C->fctx->regs;
			rcpos_t arg = -1;
			ptrdiff_t off;
			C->fctx->loops++;
			codesize = func->code.size;
			{
				LOOP_BEGIN;
				SGS_FN_ENTER;
				if( !compile_node( C, func, node->child->next ) ) goto fail; /* while */
				comp_reg_unwind( C, regstate );
				LOOP_END;

				if( !compile_breaks( C, func, node, 1 ) )
					goto fail;
			}
			SGS_FN_ENTER;
			if( !compile_node_r( C, func, node->child, &arg ) ) goto fail; /* test */
			comp_reg_unwind( C, regstate );
			/* WP: jump limit */
			off = (ptrdiff_t) ( codesize - func->code.size ) / SGS_INSTR_SIZE - 1;
			if( over_limit( off, 32767 ) )
			{
				QPRINT( "Max. jump limit exceeded (32767 instructions) @ do/while; reduce size of loop" );
				goto fail;
			}
			INSTR_WRITE_EX( SGS_SI_JMPT, off, arg );
			if( !compile_breaks( C, func, node, 0 ) )
				goto fail;
			C->fctx->loops--;
		}
		break;

	case SGS_SFT_FOR:
		SGS_FN_HIT( "FOR" );
		{
			size_t codesize;
			rcpos_t regstate = C->fctx->regs;
			rcpos_t arg = -1;
			C->fctx->loops++;
			SGS_FN_ENTER;
			if( !compile_node( C, func, node->child ) ) goto fail; /* init */
			comp_reg_unwind( C, regstate );
			codesize = func->code.size;
			SGS_FN_ENTER;
			if( !compile_for_explist( C, func, node->child->next, &arg ) ) goto fail; /* test */
			comp_reg_unwind( C, regstate );
			if( arg != -1 )
			{
				INSTR_WRITE_PCH();
			}
			{
				ptrdiff_t off;
				size_t jp1, jp2 = 0;
				jp1 = func->code.size;

				LOOP_BEGIN;
				SGS_FN_ENTER;
				if( !compile_node( C, func, node->child->next->next->next ) ) goto fail; /* block */
				comp_reg_unwind( C, regstate );
				LOOP_END;

				if( !compile_breaks( C, func, node, 1 ) )
					goto fail;
				SGS_FN_ENTER;
				if( !compile_node( C, func, node->child->next->next ) ) goto fail; /* incr */
				comp_reg_unwind( C, regstate );

				jp2 = func->code.size;
				/* WP: jump limit */
				off = (ptrdiff_t) ( codesize - jp2 ) / SGS_INSTR_SIZE - 1;
				if( over_limit( off, 32767 ) )
				{
					QPRINT( "Max. jump limit exceeded (32767 instructions) @ for; reduce size of loop" );
					goto fail;
				}
				INSTR_WRITE_EX( SGS_SI_JUMP, off, 0 );
				if( arg != -1 )
				{
					sgs_instr_t instr;
					instr = SGS_INSTR_MAKE_EX( SGS_SI_JMPF, ( jp2 - jp1 ) / SGS_INSTR_SIZE + 1, arg );
					memcpy( func->code.ptr + jp1 - 4, &instr, sizeof(instr) );
				}
			}
			if( !compile_breaks( C, func, node, 0 ) )
				goto fail;
			C->fctx->loops--;
		}
		break;
		
	case SGS_SFT_FOREACH:
		SGS_FN_HIT( "FOREACH" );
		{
			size_t codesize, jp1, jp2;
			rcpos_t var, iter, key = -1, val = -1;
			rcpos_t regstate, regstate2;
			regstate2 = C->fctx->regs;
			
			/* init */
			var = -1;
			SGS_FN_ENTER;
			if( !compile_node_r( C, func, node->child->next->next, &var ) ) goto fail; /* get variable */
			
			iter = comp_reg_alloc( C );
			if( node->child->type != SGS_SFT_NULL ) key = comp_reg_alloc( C );
			if( node->child->next->type != SGS_SFT_NULL ) val = comp_reg_alloc( C );
			regstate = C->fctx->regs;
			C->fctx->loops++;
			
			INSTR_WRITE( SGS_SI_FORPREP, iter, var, 0 );
			comp_reg_unwind( C, regstate );
			
			/* iterate */
			codesize = func->code.size;
			INSTR_WRITE_PCH();
			jp1 = func->code.size;
			INSTR_WRITE( SGS_SI_FORLOAD, iter, key, val );
			
			{
				ptrdiff_t off = 0;
				
				/* write to key variable */
				if( node->child->type != SGS_SFT_NULL && !compile_ident_w( C, func, node->child, key ) ) goto fail;
				
				/* write to value variable */
				if( node->child->next->type != SGS_SFT_NULL && !compile_ident_w( C, func, node->child->next, val ) ) goto fail;
				
				LOOP_BEGIN;
				SGS_FN_ENTER;
				if( !compile_node( C, func, node->child->next->next->next ) ) goto fail; /* block */
				comp_reg_unwind( C, regstate );
				LOOP_END;
				
				if( !compile_breaks( C, func, node, 1 ) )
					goto fail;
				
				jp2 = func->code.size;
				/* WP: jump limit */
				off = (ptrdiff_t) ( codesize - jp2 ) / SGS_INSTR_SIZE - 1;
				if( over_limit( off, 32767 ) )
				{
					QPRINT( "Max. jump limit exceeded (32767 instructions) @ foreach; reduce size of loop" );
					goto fail;
				}
				INSTR_WRITE_EX( SGS_SI_JUMP, off, 0 );
				{
					sgs_instr_t instr;
					instr = SGS_INSTR_MAKE_EX( SGS_SI_FORJUMP, ( func->code.size - jp1 ) / SGS_INSTR_SIZE, iter );
					memcpy( func->code.ptr + jp1 - 4, &instr, sizeof(instr) );
				}
			}
			
			if( !compile_breaks( C, func, node, 0 ) )
				goto fail;
			C->fctx->loops--;
			comp_reg_unwind( C, regstate2 );
		}
		break;

	case SGS_SFT_BREAK:
		SGS_FN_HIT( "BREAK" );
		{
			sgs_TokenList tl = sgsT_Next( node->token );
			int32_t blev = 1;
			if( *tl == SGS_ST_NUMINT )
			{
				sgs_Int tint;
				SGS_AS_INTEGER( tint, tl + 1 );
				if( tint < 1 || tint > 0xffff )
				{
					QPRINT( "Invalid break level" );
					goto fail;
				}
				blev = (int32_t) tint;
			}
			if( blev > C->fctx->loops )
			{
				if( C->fctx->loops )
					QPRINT( "Break level too high" );
				else
					QPRINT( "Attempted to break while not in a loop" );
				goto fail;
			}
			sgs_LoopInfo* loop = C->fctx->loopinfo;
			{
				int lev = blev;
				while( --lev > 0 )
					loop = loop->parent;
			}
			compile_defers( C, func, loop->block );
			/* WP: instruction limit, max loop depth */
			fctx_binfo_add( C, C->fctx, (uint32_t) func->code.size, (uint16_t)( C->fctx->loops + 1 - blev ), SGS_FALSE );
			INSTR_WRITE_PCH();
		}
		break;

	case SGS_SFT_CONT:
		SGS_FN_HIT( "CONTINUE" );
		{
			sgs_TokenList tl = sgsT_Next( node->token );
			int32_t blev = 1;
			if( *tl == SGS_ST_NUMINT )
			{
				sgs_Int tint;
				SGS_AS_INTEGER( tint, tl + 1 );
				if( tint < 1 || tint > 0xffff )
				{
					QPRINT( "Invalid continue level" );
					goto fail;
				}
				blev = (int32_t) tint;
			}
			if( blev > C->fctx->loops )
			{
				if( C->fctx->loops )
					QPRINT( "Continue level too high" );
				else
					QPRINT( "Attempted to continue while not in a loop" );
				goto fail;
			}
			sgs_LoopInfo* loop = C->fctx->loopinfo;
			{
				int lev = blev;
				while( --lev > 0 )
					loop = loop->parent;
			}
			compile_defers( C, func, loop->block );
			/* WP: instruction limit, max loop depth */
			fctx_binfo_add( C, C->fctx, (uint32_t) func->code.size, (uint16_t)( C->fctx->loops + 1 - blev ), SGS_TRUE );
			INSTR_WRITE_PCH();
		}
		break;
	
	case SGS_SFT_DEFER:
		SGS_FN_HIT( "DEFER" );
		fctx_defer_add( C, node->child );
		break;
	
	case SGS_SFT_FUNC:
		SGS_FN_HIT( "FUNC" );
		{
			sgs_FTNode* n_name;
			rcpos_t pos;
			SGS_FN_ENTER;
			if( !compile_func( C, func, node, &pos ) ) goto fail;
			n_name = node->child->next->next->next;

			if( n_name )
			{
				SGS_FN_ENTER;
				if( !compile_node_w( C, func, n_name, pos ) ) goto fail;
				
				// symbol registration
				if( C->fctx->func == SGS_FALSE || n_name->type == SGS_SFT_CLSPFX )
				{
					rcpos_t r_name;
					sgs_MemBuf ffn = sgs_membuf_create();
					rpts( &ffn, C, n_name );
					r_name = add_const_s( C, func, (uint32_t) ffn.size, ffn.ptr );
					sgs_membuf_destroy( &ffn, C );
					INSTR_WRITE( SGS_SI_RSYM, 0, BC_CONSTENC( r_name ), pos );
				}
			}
		}
		break;
		
	case SGS_SFT_CLASS:
		SGS_FN_HIT( "CLASS" );
		{
			sgs_FTNode* name = node->child;
			sgs_FTNode* it = name->next;
			rcpos_t regstate = C->fctx->regs;
			
			/* create class */
			rcpos_t clsvar, r_name, r_inhname, vname, vsrc;
			clsvar = comp_reg_alloc( C );
			compile_ident( C, func, name, &r_name );
			r_inhname = clsvar; /* r_inhname is 'empty' if it equals clsvar (class cannot inherit itself) */
			if( it && it->type == SGS_SFT_CLSINH )
			{
				compile_ident( C, func, it, &r_inhname );
				it = it->next;
			}
			INSTR_WRITE( SGS_SI_CLASS, clsvar, r_name, r_inhname );
			compile_ident_w( C, func, name, clsvar );
			
			while( it )
			{
				if( it->type == SGS_SFT_CLSGLOB )
				{
					/* class globals -- convert to 'setindex' */
					sgs_FTNode* vn = it->child;
					while( vn )
					{
						compile_ident( C, func, vn, &vname );
						if( vn->child )
						{
							SGS_FN_ENTER;
							if( !compile_node_r( C, func, vn->child, &vsrc ) )
								goto fail;
						}
						else
						{
							vsrc = add_const_null( C, func );
						}
						INSTR_WRITE( SGS_SI_SETINDEX, clsvar, vname, vsrc );
						
						vn = vn->next;
					}
				}
				else if( it->type == SGS_SFT_FUNC )
				{
					compile_node( C, func, it );
				}
				else
				{
					QINTERR( 1005 );
					goto fail;
				}
				it = it->next;
			}
			
			comp_reg_unwind( C, regstate );
		}
		break;

	case SGS_SFT_VARLIST:
	case SGS_SFT_GVLIST:
		SGS_FN_HIT( node->type == SGS_SFT_VARLIST ? "VARLIST" : "GLOBALVARLIST" );
		{
			rcpos_t regstate = C->fctx->regs;
			sgs_FTNode* pp = node->child;
			while( pp )
			{
				if( pp->child )
				{
					rcpos_t arg = -1;
					size_t lastsize = func->code.size;
					if( !compile_node_r( C, func, pp->child, &arg ) ) goto fail;
					if( !pp->token || *pp->token != SGS_ST_IDENT ) goto fail;
					if( node->type != SGS_SFT_VARLIST || !try_optimize_last_instr_out( C, func, pp, lastsize, NULL ) )
					{
						compile_ident_w( C, func, pp, arg );
					}
					comp_reg_unwind( C, regstate );
				}
				pp = pp->next;
			}
		}
		break;

	default:
		QINTERR( 1001 );
		goto fail;
	}

	SGS_FN_END;
	return 1;

fail:
	SGS_FN_END;
	return 0;
}


sgs_iFunc* sgsBC_Generate( SGS_CTX, sgs_FTNode* tree )
{
	sgs_FuncCtx* fctx = fctx_create( C );
	fctx->func = SGS_FALSE;
	C->fctx = fctx;
	
	if( !compile_fn_base( C, &fctx->cfunc, tree, 0 ) || ( C->state & SGS_HAS_ERRORS ) )
		goto fail;
	
	C->fctx = NULL;
	{
		sgs_iFunc* outfn = sgsBC_ConvertFunc( C, &fctx->cfunc, "<main>", 6, 0 );
		fctx_destroy( C, fctx );
		return outfn;
	}

fail:
	C->fctx = NULL;
	fctx_destroy( C, fctx );
	C->state |= SGS_HAS_ERRORS;
	return NULL;
}

void sgsBC_Dump( sgs_CompFunc* func )
{
	printf( "function (this=%s args=%d tmp=%d clsr=%d inclsr=%d)\n",
		func->gotthis ? "Y" : "n", func->numargs, func->numtmp, func->numclsr, func->inclsr );
	sgsBC_DumpEx( func->consts.ptr, func->consts.size, func->code.ptr, func->code.size );
}

void sgsBC_DumpEx( const char* constptr, size_t constsize,
	const char* codeptr, size_t codesize )
{
	const sgs_Variable* vbeg = SGS_ASSUME_ALIGNED_CONST( constptr, sgs_Variable );
	const sgs_Variable* vend = SGS_ASSUME_ALIGNED_CONST( constptr + constsize, sgs_Variable );
	const sgs_Variable* var = vbeg;

	printf( "{\n" );
	printf( "> constants:\n" );
	while( var < vend )
	{
		printf( "%4d = ", (int) ( var - vbeg ) );
		sgsVM_VarDump( var );
		printf( "\n" );
		var++;
	}
	printf( "> code:\n" );
	dump_opcode( SGS_ASSUME_ALIGNED_CONST( codeptr, sgs_instr_t ), codesize / sizeof( sgs_instr_t ) );
	printf( "}\n" );
}


#undef rcpos_t
#undef SGS_FNTCMP_ARGS
#undef INSTR_N
#undef INSTR
#undef INSTR_WRITE
#undef INSTR_WRITE_EX
#undef INSTR_WRITE_PCH
#undef QPRINT


#include <errno.h>

#define SGS_INTERNAL_STRINGTABLES





#if SGS_DEBUG && SGS_DEBUG_FLOW
#  define DBGINFO( text ) sgs_Msg( C, SGS_INFO, text )
#else
#  define DBGINFO( text )
#endif


static const char* g_varnames[] = { "null", "bool", "int", "real", "string", "func", "cfunc", "obj", "ptr" };


static void* int_memory( SGS_SHCTX, void* ptr, size_t size )
{
	void* p;
	if( size )
	{
		size += 16;
		S->memsize += size;
		S->numallocs++;
		S->numblocks++;
	}
	if( ptr )
	{
		ptr = ((char*)ptr) - 16;
		S->memsize -= *(size_t*)ptr;
		S->numfrees++;
		S->numblocks--;
	}
	p = S->memfunc( S->mfuserdata, ptr, size );
	if( p )
	{
		memcpy( p, &size, sizeof(size_t) );
		p = ((char*)p) + 16;
	}
	return p;
}


sgs_Context* sgs_CreateEngineExt( sgs_MemFunc memfunc, void* mfuserdata )
{
	///
	// SHARED STATE
	SGS_CTX;
	sgs_ShCtx* S = (sgs_ShCtx*) memfunc( mfuserdata, NULL, sizeof( sgs_ShCtx ) );
	
	S->version = SGS_VERSION_INT;
	S->sfs_fn = sgs_StdScriptFSFunc;
	S->sfs_ctx = NULL;
	
	S->output_fn = sgs_StdOutputFunc;
	S->output_ctx = stdout;
	S->erroutput_fn = sgs_StdOutputFunc;
	S->erroutput_ctx = stderr;
	
	S->memfunc = memfunc;
	S->mfuserdata = mfuserdata;
	S->memsize = sizeof( sgs_ShCtx );
	S->numallocs = 1;
	S->numblocks = 1;
	S->numfrees = 0;
	
	S->objs = NULL;
	S->objcount = 0;
	S->redblue = 0;
	S->gcrun = SGS_FALSE;
	S->objpool_size = 0;
	S->ctx_pool = NULL;
	S->ctx_pool_size = 0;
	S->sf_pool = NULL;
	S->sf_pool_size = 0;
	
	S->array_iface = NULL;
	S->_R = NULL;
	S->_SYM = NULL;
	S->_INC = NULL;
	
	///
	// CONTEXT
	C = (sgs_Context*) int_memory( S, NULL, sizeof( sgs_Context ) );
	
	S->ctx_root = C;
	S->state_list = C;
	S->statecount = 1;
	C->refcount = 1;
	C->shared = S;
	C->prev = NULL;
	C->next = NULL;
	
	C->serialize_state = NULL;
	C->minlev = SGS_INFO;
	C->apilev = SGS_ERROR + 1;
	C->msg_fn = sgs_StdMsgFunc;
	C->msg_ctx = stderr;
	C->last_errno = 0;
	C->object_arg = 0;
	C->hook_fn = NULL;
	C->hook_ctx = NULL;
	
	C->state = 0;
	C->fctx = NULL;
	C->filename = NULL;
	
	C->parent = NULL;
	C->subthreads = NULL;
	C->st_next = NULL;
	C->st_timeout = 0;
	C->_E = NULL;
	C->wait_timer = 0;
	C->tm_accum = 0;
	
	C->stack_mem = 32;
	C->stack_base = sgs_Alloc_n( sgs_Variable, C->stack_mem );
	C->stack_off = C->stack_base;
	C->stack_top = C->stack_base;
	
	C->sf_first = NULL;
	C->sf_last = NULL;
	C->sf_count = 0;
	C->num_last_returned = 0;
	
	// SHARED/FINALIZE
#if SGS_OBJPOOL_SIZE > 0
	S->objpool_data = sgs_Alloc_n( sgs_ObjPoolItem, SGS_OBJPOOL_SIZE );
#else
	S->objpool_data = NULL;
#endif
	sgs_vht_init( &S->typetable, C, 32, 32 );
	sgs_vht_init( &S->stringtable, C, 256, 256 );
	sgs_vht_init( &S->ifacetable, C, 64, 64 );
	// ---
	
	sgsSTD_RegistryInit( C );
	sgsSTD_GlobalInit( C );
	sgsSTD_PostInit( C );
	return C;
}


static void ctx_safedestroy( SGS_CTX )
{
	if( C->state & SGS_STATE_DESTROYING )
	{
		return;
	}
	C->state |= SGS_STATE_DESTROYING;
	
	if( C->hook_fn )
	{
		C->hook_fn( C->hook_ctx, C, SGS_HOOK_CFREE );
		C->hook_fn = NULL;
	}
	
	/* clear the stack */
	if( C->stack_base )
	{
		while( C->stack_base != C->stack_top )
		{
			C->stack_top--;
			sgs_Release( C, C->stack_top );
		}
		sgs_Dealloc( C->stack_base );
		C->stack_base = NULL;
		C->stack_off = NULL;
		C->stack_top = NULL;
		C->stack_mem = 0;
	}
	
	sgsSTD_GlobalFree( C );
	sgsSTD_ThreadsFree( C );
	
	/* free the call stack */
	if( C->sf_first )
	{
		sgs_StackFrame* sf = C->sf_first, *sfn;
		while( sf )
		{
			/* sf->func is on stack */
			sfn = sf->next;
			sgsCTX_FreeFrame( sf );
			sf = sfn;
		}
		C->sf_first = NULL;
		C->sf_last = NULL;
		C->sf_count = 0;
	}
}

static void ctx_destroy( SGS_CTX )
{
	SGS_SHCTX_USE;
	
	sgs_BreakIf( C->refcount > 0 );
	
	// LAST ONE CLEANS UP
	if( C->prev == NULL && C->next == NULL )
	{
		sgsSTD_RegistryFree( C );
		sgs_ObjRelease( C, S->array_iface );
		S->array_iface = NULL;
		
		sgs_GCExecute( C );
		sgs_BreakIf( S->objs || S->objcount );
		
		sgs_vht_free( &S->ifacetable, C ); /* table should be empty here */
		sgs_vht_free( &S->typetable, C );
		sgs_VHTVar* p = S->stringtable.vars;
		sgs_VHTVar* pend = p + S->stringtable.size;
		int numerrs = 0;
		while( p < pend )
		{
			sgs_iStr* unfreed_string = p->key.data.S;
#if SGS_DEBUG && SGS_DEBUG_VALIDATE
			if( unfreed_string->refcount > 0 )
			{
				fprintf( stderr, "unfreed string: (rc=%d)[%d]\"%.*s\"\n",
					(int) unfreed_string->refcount, (int) unfreed_string->size,
					(int) unfreed_string->size, sgs_str_cstr( unfreed_string ) );
			}
			else if( unfreed_string->refcount < 0 )
			{
				fprintf( stderr, "overfreed string: (rc=%d)[%d]",
					(int) unfreed_string->refcount, (int) unfreed_string->size );
			}
#endif
			numerrs += unfreed_string->refcount != 0;
			sgs_BreakIf( unfreed_string->refcount < 0 && "memory error" );
			sgs_Dealloc( unfreed_string );
			p++;
		}
		sgs_BreakIf( numerrs > 0 && "string not freed" );
		sgs_vht_free( &S->stringtable, C );
	}
	// ----
	
	if( S->state_list == C ) S->state_list = C->next;
	if( C->prev ) C->prev->next = C->next;
	if( C->next ) C->next->prev = C->prev;
	S->statecount--;
	
	// free the context
	if( C->prev == NULL && C->next == NULL )
	{
		int_memory( S, C, 0 );
	}
	else
	{
		C->next = S->ctx_pool;
		S->ctx_pool = C;
	}
}

static void shctx_destroy( SGS_SHCTX )
{
#if SGS_OBJPOOL_SIZE > 0
	{
		int32_t i;
		for( i = 0; i < S->objpool_size; ++i )
			int_memory( S, S->objpool_data[ i ].obj, 0 );
		int_memory( S, S->objpool_data, 0 );
	}
#endif
	
	while( S->sf_pool )
	{
		sgs_StackFrame* sfn = S->sf_pool->next;
		int_memory( S, S->sf_pool, 0 );
		S->sf_pool = sfn;
	}
	
	while( S->ctx_pool )
	{
		sgs_Context* ctxn = S->ctx_pool->next;
		int_memory( S, S->ctx_pool, 0 );
		S->ctx_pool = ctxn;
	}
	
#ifdef SGS_DEBUG_LEAKS
	sgs_BreakIf( S->memsize > sizeof( sgs_ShCtx ) &&
		"not all resources have been freed" );
	sgs_BreakIf( S->memsize < sizeof( sgs_ShCtx ) &&
		"some resorces may have been freed more than once" );
#endif
	
	S->memfunc( S->mfuserdata, S, 0 );
}

void sgs_DestroyEngine( SGS_CTX )
{
	sgs_ShCtx* S = C->shared;
	sgs_BreakIf( C != S->ctx_root );
	C->refcount--;
	
	sgsSTD_RegistryFree( C );
	
	while( S->state_list )
	{
		int numfreed = 0;
		sgs_Context* cur = S->state_list;
		C->refcount++; /* prevent freeing of engine context before others */
		while( cur != NULL )
		{
			cur->refcount++; /* prevent self-free */
			ctx_safedestroy( cur );
			sgs_BreakIf( cur->refcount < 1 );
			cur->refcount--;
			cur = cur->next;
		}
		sgs_GCExecute( C );
		C->refcount--; /* allow freeing the engine context again now */
		cur = S->state_list;
		while( cur != NULL )
		{
			if( cur->refcount <= 0 )
			{
				ctx_destroy( cur );
				numfreed++;
				break;
			}
			else
			{
				cur = cur->next;
			}
		}
		
		sgs_BreakIf( numfreed == 0 && "some user-made states have not been freed" );
		if( numfreed == 0 )
			break;
	}
	
	while( S->state_list )
	{
		sgs_Context* cur = S->state_list;
		cur->refcount++; /* prevent self-free */
		ctx_safedestroy( cur );
		sgs_BreakIf( cur->refcount < 1 );
		cur->refcount--;
		ctx_destroy( cur );
	}
	
	shctx_destroy( S );
}


sgs_StackFrame* sgsCTX_AllocFrame( SGS_CTX )
{
	SGS_SHCTX_USE;
	
	if( S->sf_pool )
	{
		sgs_StackFrame* ret = S->sf_pool;
		S->sf_pool = ret->next;
		S->sf_pool_size--;
		return ret;
	}
	
	return sgs_Alloc( sgs_StackFrame );
}

static void copy_append_frame( SGS_CTX, sgs_StackFrame* sf )
{
	sgs_StackFrame* nsf = sgsCTX_AllocFrame( C );
	memcpy( nsf, sf, sizeof(*nsf) );
	sgs_Acquire( C, nsf->func );
	
	nsf->next = NULL;
	if( C->sf_last )
	{
		nsf->prev = C->sf_last;
		nsf->prev->next = nsf;
	}
}

static sgs_Context* sgs__alloc_ctx( SGS_CTX )
{
	SGS_SHCTX_USE;
	
	if( S->ctx_pool )
	{
		sgs_Context* ret = S->ctx_pool;
		S->ctx_pool = ret->next;
		return ret;
	}
	
	return sgs_Alloc( sgs_Context );
}

sgs_Context* sgsCTX_ForkState( SGS_CTX, int copystate )
{
	SGS_SHCTX_USE;
	sgs_Context* NC = sgs__alloc_ctx( C );
	memcpy( NC, C, sizeof(*NC) );
	NC->refcount = 0;
	NC->hook_fn = NULL; /* not shareable */
	NC->hook_ctx = NULL; /* not shareable */
	NC->parent = NULL; /* not shareable */
	NC->subthreads = NULL; /* not shareable */
	NC->st_next = NULL; /* not shareable */
	NC->st_timeout = 0; /* not shareable */
	NC->_E = NULL; /* not shareable */
	NC->wait_timer = 0;
	
	// realloc / acquire
	if( copystate == SGS_FALSE )
	{
		NC->stack_mem = 32;
	}
	
	NC->stack_base = sgs_Alloc_n( sgs_Variable, NC->stack_mem );
	
	if( copystate == SGS_FALSE )
	{
		NC->stack_top = NC->stack_off = NC->stack_base;
	}
	else
	{
		// WP: assuming top > base
		memcpy( NC->stack_base, C->stack_base, sizeof(sgs_Variable) * (size_t)( C->stack_top - C->stack_base ) );
		NC->stack_off = NC->stack_base + ( C->stack_off - C->stack_base );
		NC->stack_top = NC->stack_base + ( C->stack_top - C->stack_base );
		{
			sgs_Variable* vp = NC->stack_base, *vpend = NC->stack_top;
			while( vp != vpend )
				sgs_Acquire( C, vp++ );
		}
	}
	
	// - global var. dict.
	sgs_ObjAcquire( C, NC->_G );
	
	// - call stack
	NC->sf_first = NULL;
	NC->sf_last = NULL;
	NC->sf_count = 0;
	if( copystate )
	{
		sgs_StackFrame* sf = C->sf_first;
		while( sf != C->sf_last )
		{
			copy_append_frame( NC, sf );
			sf = sf->next;
		}
	}
	
	// add state to list
	NC->prev = NULL;
	NC->next = S->state_list;
	if( NC->next )
		NC->next->prev = NC;
	S->state_list = NC;
	S->statecount++;
	
	if( C->hook_fn )
		C->hook_fn( C->hook_ctx, NC, copystate ? SGS_HOOK_CFORK : SGS_HOOK_CREAT );
	
	return NC;
}

static void _sgs_dumprsrc( SGS_SHCTX );

void sgsCTX_FreeState( SGS_CTX )
{
	SGS_SHCTX_USE;
	sgs_BreakIf( C->refcount < 0 );
	C->refcount++; /* prevent self-free */
	ctx_safedestroy( C );
	sgs_BreakIf( C->refcount < 1 );
	C->refcount--;
	ctx_destroy( C );
	
	if( S->state_list == NULL )
		shctx_destroy( S );
	else if( S->ctx_root == C && S->statecount > 1 )
	{
		_sgs_dumprsrc( S );
		sgs_BreakIf( "trying to free root context before others" );
		return;
	}
}

sgs_Context* sgs_RootContext( SGS_CTX )
{
	if( C == NULL )
		return NULL;
	else
	{
		SGS_SHCTX_USE;
		return S->ctx_root;
	}
}

sgs_Context* sgs_ForkState( SGS_CTX, int copystate )
{
	sgs_Context* F = sgsCTX_ForkState( C, copystate );
	F->refcount = 1;
	return F;
}

void sgs_ReleaseState( SGS_CTX )
{
	C->refcount--;
	if( C->refcount <= 0 )
		sgsCTX_FreeState( C );
}

SGSBOOL sgs_PauseState( SGS_CTX )
{
	sgs_StackFrame* sf = C->sf_last;
	if( sf == NULL )
		return SGS_FALSE; /* nothing to pause */
	if( sf->flags & SGS_SF_PAUSED )
		return SGS_FALSE; /* already paused, but possibly not at the expected location */
	
	if( sf && !sf->iptr )
		sf = sf->prev; /* should be able to use this inside a C function */
	if( !sf || !sf->iptr )
		return SGS_FALSE; /* not in SGS function */
	{
		sgs_StackFrame* tsf = sf->prev;
		while( tsf && tsf->iptr )
			tsf = tsf->prev;
		if( tsf )
			return SGS_FALSE; /* cannot have any C functions in the middle */
	}
	
	sf = C->sf_last;
	if( !sf->iptr )
		sf = sf->prev;
	sf->flags |= SGS_SF_PAUSED;
	
	return SGS_TRUE;
}

SGSBOOL sgs_ResumeStateExp( SGS_CTX, int args, int expect )
{
	int rvc = 0;
	int ret = sgs_ResumeStateRet( C, args, &rvc );
	if( ret )
		sgs_SetDeltaSize( C, expect - rvc );
	return ret;
}


const char* sgs_CodeString( int type, int val )
{
	if( type == SGS_CODE_ER )
	{
		if( val < SGS_EINPROC )
			return "UNKNOWN ERROR";
		if( SGS_SUCCEEDED( val ) )
			val = 0;
		return sgs_ErrNames[  -  val ];
	}
	else if( type == SGS_CODE_VT )
	{
		if( val < 0 || val >= SGS_VT__COUNT )
			return NULL;
		return sgs_VarNames[ val ];
	}
	else if( type == SGS_CODE_OP )
	{
		if( val < 0 || val >= SGS_SI_COUNT )
			return NULL;
		return sgs_OpNames[ val ];
	}
	else
		return NULL;
}


void sgs_GetOutputFunc( SGS_CTX, sgs_OutputFunc* outf, void** outc )
{
	SGS_SHCTX_USE;
	*outf = S->output_fn;
	*outc = S->output_ctx;
}

void sgs_SetOutputFunc( SGS_CTX, sgs_OutputFunc func, void* ctx )
{
	SGS_SHCTX_USE;
	sgs_BreakIf( func == NULL );
	if( func == SGSOUTPUTFN_DEFAULT )
		func = sgs_StdOutputFunc;
	if( !ctx && func == sgs_StdOutputFunc )
		ctx = stdout;
	S->output_fn = func;
	S->output_ctx = ctx;
}

void sgs_Write( SGS_CTX, const void* ptr, size_t size )
{
	SGS_SHCTX_USE;
	S->output_fn( S->output_ctx, C, ptr, size );
}

SGSBOOL sgs_Writef( SGS_CTX, const char* what, ... )
{
	char buf[ SGS_OUTPUT_STACKBUF_SIZE ];
	sgs_MemBuf info = sgs_membuf_create();
	int cnt;
	va_list args;
	char* ptr = buf;
	
	va_start( args, what );
	cnt = SGS_VSPRINTF_LEN( what, args );
	va_end( args );
	
	if( cnt < 0 )
		return SGS_FALSE;
	
	if( cnt >= SGS_OUTPUT_STACKBUF_SIZE )
	{
		/* WP: false alarm */
		sgs_membuf_resize( &info, C, (size_t) cnt + 1 );
		ptr = info.ptr;
	}
	
	va_start( args, what );
	vsprintf( ptr, what, args );
	va_end( args );
	ptr[ cnt ] = 0;
	
	sgs_WriteStr( C, ptr );
	
	sgs_membuf_destroy( &info, C );
	return SGS_TRUE;
}

void sgs_GetErrOutputFunc( SGS_CTX, sgs_OutputFunc* outf, void** outc )
{
	SGS_SHCTX_USE;
	*outf = S->erroutput_fn;
	*outc = S->erroutput_ctx;
}

void sgs_SetErrOutputFunc( SGS_CTX, sgs_OutputFunc func, void* ctx )
{
	SGS_SHCTX_USE;
	sgs_BreakIf( func == NULL );
	if( func == SGSOUTPUTFN_DEFAULT )
		func = sgs_StdOutputFunc;
	if( !ctx && func == sgs_StdOutputFunc )
		ctx = stderr;
	S->erroutput_fn = func;
	S->erroutput_ctx = ctx;
}

void sgs_ErrWrite( SGS_CTX, const void* ptr, size_t size )
{
	SGS_SHCTX_USE;
	S->erroutput_fn( S->erroutput_ctx, C, ptr, size );
}

SGSBOOL sgs_ErrWritef( SGS_CTX, const char* what, ... )
{
	char buf[ SGS_OUTPUT_STACKBUF_SIZE ];
	sgs_MemBuf info = sgs_membuf_create();
	int cnt;
	va_list args;
	char* ptr = buf;
	
	va_start( args, what );
	cnt = SGS_VSPRINTF_LEN( what, args );
	va_end( args );
	
	if( cnt < 0 )
		return SGS_FALSE;
	
	if( cnt >= SGS_OUTPUT_STACKBUF_SIZE )
	{
		/* WP: false alarm */
		sgs_membuf_resize( &info, C, (size_t) cnt + 1 );
		ptr = info.ptr;
	}
	
	va_start( args, what );
	vsprintf( ptr, what, args );
	va_end( args );
	ptr[ cnt ] = 0;
	
	sgs_ErrWriteStr( C, ptr );
	
	sgs_membuf_destroy( &info, C );
	return SGS_TRUE;
}


void sgs_GetMsgFunc( SGS_CTX, sgs_MsgFunc* outf, void** outc )
{
	*outf = C->msg_fn;
	*outc = C->msg_ctx;
}

void sgs_SetMsgFunc( SGS_CTX, sgs_MsgFunc func, void* ctx )
{
	sgs_BreakIf( func == NULL );
	if( func == SGSMSGFN_DEFAULT )
		func = sgs_StdMsgFunc;
	else if( func == SGSMSGFN_DEFAULT_NOABORT )
		func = sgs_StdMsgFunc_NoAbort;
	if( !ctx && ( func == sgs_StdMsgFunc || func == sgs_StdMsgFunc_NoAbort ) )
		ctx = NULL;
	C->msg_fn = func;
	C->msg_ctx = ctx;
}

SGSZERO sgs_Msg( SGS_CTX, int type, const char* what, ... )
{
	char buf[ SGS_OUTPUT_STACKBUF_SIZE ];
	sgs_MemBuf info = sgs_membuf_create();
	int off = 0, cnt = 0, slen = 0;
	va_list args;
	char* ptr = buf;
	
	/* error level filter */
	if( type < ( C->state & SGS_STATE_INSIDE_API ? C->apilev : C->minlev ) ) return 0;
	/* error suppression */
	if( C->sf_last && C->sf_last->errsup > 0 ) return 0;
	
	va_start( args, what );
	cnt = SGS_VSPRINTF_LEN( what, args );
	va_end( args );
	
	sgs_BreakIf( cnt < 0 );
	if( cnt < 0 )
	{
		C->msg_fn( C->msg_ctx, C, SGS_ERROR, "sgs_Msg ERROR: failed to print the message" );
		return 0;
	}
	
	if( C->sf_last && C->sf_last->nfname )
	{
		/* WP: it is expected that native functions ..
		.. will not set names of more than 2GB length */
		slen = (int) strlen( C->sf_last->nfname );
		off = slen + SGS_MSG_EXTRABYTES;
		cnt += off;
	}

	if( cnt >= SGS_OUTPUT_STACKBUF_SIZE )
	{
		/* WP: cnt is never negative */
		sgs_membuf_resize( &info, C, (size_t) cnt + 1 );
		ptr = info.ptr;
	}
	
	if( C->sf_last && C->sf_last->nfname )
	{
		/* WP: slen is generated from size_t, thus is never negative */
		memcpy( ptr, C->sf_last->nfname, (size_t) slen );
		memcpy( ptr + slen, SGS_MSG_EXTRASTRING, SGS_MSG_EXTRABYTES );
	}

	va_start( args, what );
	vsprintf( ptr + off, what, args );
	va_end( args );
	ptr[ cnt ] = 0;

	C->msg_fn( C->msg_ctx, C, type, ptr );

	sgs_membuf_destroy( &info, C );
	
	return 0;
}


void sgs_WriteErrorInfo( SGS_CTX, int flags, sgs_ErrorOutputFunc func, void* ctx, int type, const char* msg )
{
	if( flags & SGS_ERRORINFO_STACK )
	{
		sgs_StackFrame* p = sgs_GetFramePtr( C, NULL, SGS_FALSE );
		SGS_UNUSED( ctx );
		while( p != NULL )
		{
			const char* file, *name;
			int ln;
			if( !p->next && !p->iptr )
				break;
			sgs_StackFrameInfo( C, p, &name, &file, &ln );
			if( ln )
				func( ctx, "- \"%s\" in %s, line %d\n", name, file, ln );
			else
				func( ctx, "- \"%s\" in %s\n", name, file );
			p = p->next;
		}
	}
	if( flags & SGS_ERRORINFO_ERROR )
	{
		static const char* errpfxs[ 3 ] = { "Info", "Warning", "Error" };
		type = type / 100 - 1;
		if( type < 0 ) type = 0;
		if( type > 2 ) type = 2;
		func( ctx, "%s: %s\n", errpfxs[ type ], msg );
	}
}

static void serialize_output_func( void* ud, SGS_CTX, const void* ptr, size_t datasize )
{
	sgs_MemBuf* B = (sgs_MemBuf*) ud;
	sgs_membuf_appbuf( B, C, ptr, datasize );
}

void sgs_PushErrorInfo( SGS_CTX, int flags, int type, const char* msg )
{
	SGS_SHCTX_USE;
	sgs_OutputFunc oldfn = S->output_fn;
	void* oldctx = S->output_ctx;
	
	sgs_MemBuf B = sgs_membuf_create();
	S->output_fn = serialize_output_func;
	S->output_ctx = &B;
	
	sgs_WriteErrorInfo( C, flags, (sgs_ErrorOutputFunc) sgs_Writef, C, type, msg );
	
	/* WP: hopefully the error messages will not be more than 2GB long */
	sgs_PushStringBuf( C, B.ptr, (sgs_SizeVal) B.size );
	
	sgs_membuf_destroy( &B, C );
	S->output_fn = oldfn;
	S->output_ctx = oldctx;
}


SGSBOOL sgs_GetHookFunc( SGS_CTX, sgs_HookFunc* outf, void** outc )
{
	if( C->hook_fn )
	{
		*outf = C->hook_fn;
		*outc = C->hook_ctx;
		return 1;
	}
	return 0;
}

void sgs_SetHookFunc( SGS_CTX, sgs_HookFunc func, void* ctx )
{
	C->hook_fn = func;
	C->hook_ctx = ctx;
}


SGSBOOL sgs_GetScriptFSFunc( SGS_CTX, sgs_ScriptFSFunc* outf, void** outc )
{
	if( C->shared->sfs_fn )
	{
		*outf = C->shared->sfs_fn;
		*outc = C->shared->sfs_ctx;
		return 1;
	}
	return 0;
}

void sgs_SetScriptFSFunc( SGS_CTX, sgs_ScriptFSFunc func, void* ctx )
{
	C->shared->sfs_fn = func;
	C->shared->sfs_ctx = ctx;
}


void* sgs_Memory( SGS_CTX, void* ptr, size_t size )
{
	SGS_SHCTX_USE;
	return int_memory( S, ptr, size );
}


static SGSRESULT ctx_push_function( SGS_CTX, const char* buf, size_t size )
{
	if( size > 0x7fffffff )
		return SGS_EINVAL;
	if( sgsBC_ValidateHeader( buf, size ) >= SGS_HEADER_SIZE )
	{
		/* header of buffer recognized as compiled source */
		const char* ret;
		sgs_Variable funcvar;
		ret = sgsBC_Buf2Func( C, C->filename ? C->filename : "", buf, size, &funcvar );
		if( ret )
		{
			sgs_Msg( C, SGS_ERROR, "Failed to read bytecode file (%s)", ret );
			return SGS_EINVAL;
		}
		fstk_push_leave( C, &funcvar );
		return SGS_SUCCESS;
	}
	else
	{
		/* not a bytecode file, try parsing as source */
		sgs_iFunc* func = NULL;
		sgs_TokenList tlist = NULL;
		sgs_FTNode* ftree = NULL;

		C->state = 0;

		tlist = sgsT_Gen( C, buf, size );
		if( !tlist || C->state & SGS_HAS_ERRORS )
			goto error;
#if SGS_DEBUG && SGS_DEBUG_DATA
		sgsT_DumpList( tlist, NULL );
#endif
		
		ftree = sgsFT_Compile( C, tlist );
		if( !ftree || C->state & SGS_HAS_ERRORS )
			goto error;
#if SGS_DEBUG && SGS_DEBUG_DATA
		sgsFT_Dump( ftree->child );
#endif
		
		func = sgsBC_Generate( C, ftree->child );
		if( !func || C->state & SGS_HAS_ERRORS )
			goto error;
		sgsFT_Destroy( C, ftree ); ftree = NULL;
		sgsT_Free( C, tlist ); tlist = NULL;
		
		/* push the function */
		{
			sgs_Variable funcvar;
			funcvar.type = SGS_VT_FUNC;
			funcvar.data.F = func;
			fstk_push_leave( C, &funcvar );
		}
		return SGS_SUCCESS;
		
error:
		if( ftree ) sgsFT_Destroy( C, ftree );
		if( tlist ) sgsT_Free( C, tlist );
		return SGS_ECOMP;
	}
}

SGSRESULT sgs_EvalBuffer( SGS_CTX, const char* buf, size_t size )
{
	int rr = ctx_push_function( C, buf, size );
	if( SGS_FAILED( rr ) )
		return rr;
	return sgs_XCall( C, 0 );
}

SGSRESULT sgs_EvalFile( SGS_CTX, const char* file )
{
	int ret;
	size_t ulen;
	char* data;
	const char* ofn;
	unsigned char magic[4];
	sgs_ScriptFSData fsd = {0};
	SGS_SHCTX_USE;
	DBGINFO( "sgs_EvalFile called!" );
	
	fsd.filename = file;
	ret = S->sfs_fn( S->sfs_ctx, C, SGS_SFS_FILE_OPEN, &fsd );
	if( SGS_FAILED( ret ) )
	{
		return ret == SGS_ENOTFND ? ret : SGS_EINPROC;
	}
	ulen = fsd.size;
	
	if( ulen > 4 )
	{
		fsd.output = magic;
		fsd.size = 4;
		ret = S->sfs_fn( S->sfs_ctx, C, SGS_SFS_FILE_READ, &fsd );
		if( SGS_FAILED( ret ) )
		{
			S->sfs_fn( S->sfs_ctx, C, SGS_SFS_FILE_CLOSE, &fsd );
			return SGS_EINPROC;
		}
		
		if( ( magic[0] == 0x7f && magic[1] == 'E' && magic[2] == 'L' && magic[3] == 'F' ) /* ELF binary */
		 || ( magic[0] == 'M' && magic[1] == 'Z' ) /* DOS binary / DOS stub for PE binary */
		 || ( magic[0] == 0xCA && magic[1] == 0xFE && magic[2] == 0xBA && magic[3] == 0xBE ) /* Mach-O binary 1 */
		 || ( magic[0] == 0xCE && magic[1] == 0xFA && magic[2] == 0xED && magic[3] == 0xFE ) /* Mach-O binary 2 */
		 || ( magic[0] == 0xCF && magic[1] == 0xFA && magic[2] == 0xED && magic[3] == 0xFE ) /* Mach-O binary 3 */
		 )
		{
			S->sfs_fn( S->sfs_ctx, C, SGS_SFS_FILE_CLOSE, &fsd );
			return SGS_ENOTSUP;
		}
	}
	
	data = sgs_Alloc_n( char, ulen );
	fsd.output = data;
	fsd.size = ulen;
	ret = S->sfs_fn( S->sfs_ctx, C, SGS_SFS_FILE_READ, &fsd );
	S->sfs_fn( S->sfs_ctx, C, SGS_SFS_FILE_CLOSE, &fsd );
	if( SGS_FAILED( ret ) )
	{
		sgs_Dealloc( data );
		return SGS_EINPROC;
	}
	
	ofn = C->filename;
	C->filename = file;
	ret = sgs_EvalBuffer( C, data, ulen );
	C->filename = ofn;
	
	sgs_Dealloc( data );
	return ret;
}

SGSRESULT sgs_Compile( SGS_CTX, const char* buf, size_t size, char** outbuf, size_t* outsize )
{
	int rr = ctx_push_function( C, buf, size );
	if( SGS_FAILED( rr ) )
		return rr;
	
	sgs_MemBuf mb;
	mb = sgs_membuf_create();
	if( !sgsBC_Func2Buf( C, (C->stack_top-1)->data.F, &mb ) )
	{
		sgs_membuf_destroy( &mb, C );
		return SGS_EINPROC;
	}
	*outbuf = mb.ptr;
	*outsize = mb.size;
	
	sgs_Pop( C, 1 );
	return SGS_SUCCESS;
}


static void _fndump( sgs_iFunc* F );
static void _recfndump( const char* constptr, size_t constsize,
	const char* codeptr, size_t codesize, int gt, int args, int tmp, int clsr, int inclsr )
{
	const sgs_Variable* var = SGS_ASSUME_ALIGNED_CONST( constptr, sgs_Variable );
	const sgs_Variable* vend = SGS_ASSUME_ALIGNED_CONST( constptr + constsize, sgs_Variable );
	while( var < vend )
	{
		if( var->type == SGS_VT_FUNC )
			_fndump( var->data.F );
		var++;
	}
	printf( "\nFUNC: type=%s args=%d tmp=%d closures=%d in.clsr=%d\n",
		gt ? "method" : "function", args, tmp, clsr, inclsr );
	sgsBC_DumpEx( constptr, constsize, codeptr, codesize );
}

static void _fndump( sgs_iFunc* F )
{
	_recfndump( (const char*) sgs_func_consts( F ), sgs_func_instr_off( F ),
		(const char*) sgs_func_bytecode( F ),
		sgs_func_size( F ) - sgs_func_instr_off( F ),
		F->gotthis, F->numargs, F->numtmp, F->numclsr, F->inclsr );
}

SGSRESULT sgs_DumpCompiled( SGS_CTX, const char* buf, size_t size )
{
	int rr = ctx_push_function( C, buf, size );
	if( SGS_FAILED( rr ) )
		return rr;
	
	_fndump( stk_gettop( C )->data.F );
	sgs_Pop( C, 1 );
	return SGS_SUCCESS;
}


SGSBOOL sgs_Abort( SGS_CTX )
{
	sgs_StackFrame* sf = C->sf_last;
	if( sf && !sf->iptr )
		sf = sf->prev; /* should be able to use this inside a C function */
	if( !sf || !sf->iptr )
		return SGS_FALSE;
	while( sf && sf->iptr )
	{
		sgs_iFunc* F = sf->func->data.F;
		sf->iptr = sgs_func_bytecode( F ) +
			( sgs_func_instr_count( F ) - 1 ); /* the last RETN instruction */
		sf->flags |= SGS_SF_ABORTED;
		sf = sf->prev;
	}
	return SGS_TRUE;
}



static void ctx_print_safe( SGS_CTX, const char* str, size_t size )
{
	const char* strend = str + size;
	while( str < strend )
	{
		if( *str == ' ' || sgs_isgraph( *str ) )
			sgs_Write( C, str, 1 );
		else
		{
			static const char* hexdigs = "0123456789ABCDEF";
			char buf[ 4 ] = { '\\', 'x', 0, 0 };
			{
				buf[2] = hexdigs[ (*str & 0xf0) >> 4 ];
				buf[3] = hexdigs[ *str & 0xf ];
			}
			sgs_Write( C, buf, 4 );
		}
		str++;
	}
}

static void dumpobj( SGS_CTX, sgs_VarObj* p )
{
	sgs_Writef( C, "OBJECT %p refcount=%d data=%p iface=%p prev=%p next=%p redblue=%s",
		p, p->refcount, p->data, p->iface, p->prev, p->next, p->redblue ? "R" : "B" );
}

/* g_varnames ^^ */

static void dumpvar( SGS_CTX, sgs_Variable* var )
{
	if( var->type > SGS_VT_PTR )
	{
		sgs_Writef( C, "INVALID TYPE %d\n", (int) var->type );
		return;
	}
	/* WP: var->type base type info uses bits 1-8 */
	sgs_Writef( C, "%s (size:%d)", g_varnames[ var->type ], sgsVM_VarSize( var ) );
	switch( var->type )
	{
	case SGS_VT_NULL: break;
	case SGS_VT_BOOL: sgs_Writef( C, " = %s", var->data.B ? "true" : "false" ); break;
	case SGS_VT_INT: sgs_Writef( C, " = %" PRId64, var->data.I ); break;
	case SGS_VT_REAL: sgs_Writef( C, " = %f", var->data.R ); break;
	case SGS_VT_STRING:
		sgs_Writef( C, " [rc:%d] = \"", var->data.S->refcount );
		ctx_print_safe( C, sgs_var_cstr( var ), SGS_MIN( var->data.S->size, 16 ) );
		sgs_Writef( C, var->data.S->size > 16 ? "...\"" : "\"" );
		break;
	case SGS_VT_FUNC:
		sgs_Writef( C, " [rc:%d] '%s'[%d]%s tmp=%d clsr=%d", var->data.F->refcount,
			var->data.F->sfuncname->size ? sgs_str_cstr( var->data.F->sfuncname ) : "<anonymous>",
			(int) var->data.F->numargs, var->data.F->gotthis ? " (method)" : "",
			(int) var->data.F->numtmp, (int) var->data.F->numclsr, (int) var->data.F->inclsr );
		break;
	case SGS_VT_CFUNC: sgs_Writef( C, " = %p", var->data.C ); break;
	case SGS_VT_OBJECT: sgs_Writef( C, " = " ); dumpobj( C, var->data.O ); break;
	case SGS_VT_PTR: sgs_Writef( C, " = %p", var->data.P ); break;
	}
}

static void _sgs_dumprsrc( SGS_SHCTX )
{
	sgs_VHTIdx i;
	sgs_VarObj* obj;
	sgs_VHTable* tbl;
	SGS_CTX = S->state_list;
	sgs_Writef( C, "SYSTEM: %p\n", S );
	sgs_Writef( C, "- version: 0x%X\n", (unsigned) S->version );
	sgs_Writef( C, "- STATE COUNT: %d\n", (int) S->statecount );
	sgs_Writef( C, "- OBJECT COUNT: %d\n", (int) S->objcount );
	sgs_Writef( C, "- TYPE COUNT: %d\n", (int) S->typetable.size );
	sgs_Writef( C, "- STRING COUNT: %d\n", (int) S->stringtable.size );
	sgs_Writef( C, "- INTERFACE GEN. COUNT: %d\n", (int) S->ifacetable.size );
	sgs_Writef( C, "- file system (func=%p, userdata=%p)\n", (void*) S->sfs_fn, S->sfs_ctx );
	sgs_Writef( C, "- memory system (func=%p, userdata=%p)\n", (void*) S->memfunc, S->mfuserdata );
	sgs_Writef( C, "- memory usage: %d B (%.2f KB / %.2f MB)\n",
		(int) S->memsize, ((float) S->memsize) / 1024.0f, ((float) S->memsize) / (1024.0f * 1024.0f) );
	sgs_Writef( C, "- memory blocks (current=%d, added=%d, removed=%d)\n",
		(int) S->numblocks, (int) S->numallocs, (int) S->numfrees );
	sgs_Writef( C, "- pools (objects=%d, contexts=%d, stack frames=%d)\n",
		(int) S->objpool_size, (int) S->ctx_pool_size, (int) S->sf_pool_size );
	sgs_Writef( C, "- REGISTRY (_R=%p, _SYM=%p, _INC=%p)\n", S->_R, S->_SYM, S->_INC );
	while( C )
	{
		sgs_Writef( C, "CONTEXT: %p (rc=%d)\n", C, (int) C->refcount );
		sgs_Writef( C, "- links (prev=%p, next=%p, parent=%p)\n", C->prev, C->next, C->parent );
		sgs_Writef( C, "- state: 0x%X\n", (unsigned) C->state );
		sgs_Writef( C, "- wait timeout: %g\n", C->wait_timer );
		sgs_Writef( C, "- STACK (size=%d, off=%d, mem=%d)\n",
			(int)( C->stack_top - C->stack_base ),
			(int)( C->stack_off - C->stack_base ),
			(int) C->stack_mem );
		sgs_Writef( C, "- call stack frame count: %d\n", (int) C->sf_count );
		if( C->subthreads )
		{
			int count = 0;
			sgs_Context* st;
			for( st = C->subthreads; st != NULL; st = st->st_next )
				count++;
			sgs_Writef( C, "- SUBTHREADS: (%d)\n", count );
			for( st = C->subthreads; st != NULL; st = st->st_next )
			{
				sgs_Writef( C, "- - %p (timeout: %g)\n", st, st->st_timeout );
			}
		}
		else sgs_Writef( C, "- SUBTHREADS: 0\n" );
		if( C->_E )
		{
			tbl = (sgs_VHTable*) C->_E->data;
			sgs_Writef( C, "- END EVENTS: (%d)\n", (int) tbl->size );
		}
		else sgs_Writef( C, "- END EVENTS: 0\n" );
		sgs_Stat( C, SGS_STAT_DUMP_FRAMES );
		sgs_Stat( C, SGS_STAT_DUMP_STACK );
		C = C->next;
	}
	
	obj = S->objs;
	C = S->state_list;
	sgs_Writef( C, "OBJECTS (%d):\n", (int) S->objcount );
	while( obj )
	{
		sgs_Writef( C, "OBJECT: %p (rc=%d, app=%d)\n",
			obj, (int) obj->refcount, (int) obj->appsize );
		sgs_Writef( C, "- misc (mm=%s, iface=%s)\n",
			obj->mm_enable ? "Y" : "N", obj->is_iface ? "Y" : "N" );
		sgs_Writef( C, "- interface: %p\n", obj->iface );
		sgs_Writef( C, "- metaobj: %p\n", obj->metaobj );
		obj = obj->next;
	}
	
	tbl = &S->stringtable;
	sgs_Writef( C, "STRINGS (%d):\n", (int) tbl->size );
	for( i = 0; i < tbl->size; ++i )
	{
		uint32_t j = 0;
		sgs_iStr* str = tbl->vars[ i ].key.data.S;
		sgs_Writef( C, "- [%u]\"", (unsigned) str->size );
		for( j = 0; j < str->size; ++j )
		{
			char c = sgs_str_cstr( str )[ j ] ;
			if( sgs_isgraph( c ) || c == ' ' )
				sgs_Writef( C, "%c", c );
			else
				sgs_Writef( C, "%02X", (unsigned) c );
		}
		sgs_Writef( C, "\"\n" );
	}
}

ptrdiff_t sgs_Stat( SGS_CTX, int type )
{
	SGS_SHCTX_USE;
	switch( type )
	{
	/* WP: not important */
	case SGS_STAT_VERSION: return (ptrdiff_t) S->version;
	case SGS_STAT_STATECOUNT: return (ptrdiff_t) S->statecount;
	case SGS_STAT_OBJCOUNT: return (ptrdiff_t) S->objcount;
	case SGS_STAT_MEMSIZE: return (ptrdiff_t) S->memsize;
	case SGS_STAT_NUMALLOCS: return (ptrdiff_t) S->numallocs;
	case SGS_STAT_NUMFREES: return (ptrdiff_t) S->numfrees;
	case SGS_STAT_NUMBLOCKS: return (ptrdiff_t) S->numblocks;
	case SGS_STAT_DUMP_STACK:
		{
			sgs_Variable* p = C->stack_base;
			sgs_WriteStr( C, "\nVARIABLE -- ---- STACK ---- BASE ----\n" );
			while( p < C->stack_top )
			{
				if( p == C->stack_off )
				{
					sgs_WriteStr( C, "VARIABLE -- ---- STACK ---- OFFSET ----\n" );
				}
				sgs_Writef( C, "VARIABLE %02d ", p - C->stack_base );
				dumpvar( C, p );
				sgs_WriteStr( C, "\n" );
				p++;
			}
			sgs_WriteStr( C, "VARIABLE -- ---- STACK ---- TOP ----\n" );
		}
		return SGS_SUCCESS;
	case SGS_STAT_DUMP_GLOBALS:
		{
			sgs_VHTVar *p, *pend;
			sgsSTD_GlobalIter( C, &p, &pend );
			sgs_WriteStr( C, "\nGLOBAL ---- LIST ---- START ----\n" );
			while( p < pend )
			{
				sgs_iStr* str = p->key.data.S;
				sgs_WriteStr( C, "GLOBAL '" );
				ctx_print_safe( C, sgs_str_cstr( str ), str->size );
				sgs_WriteStr( C, "' = " );
				dumpvar( C, &p->val );
				sgs_WriteStr( C, "\n" );
				p++;
			}
			sgs_WriteStr( C, "GLOBAL ---- LIST ---- END ----\n" );
		}
		return SGS_SUCCESS;
	case SGS_STAT_DUMP_OBJECTS:
		{
			sgs_VarObj* p = S->objs;
			sgs_WriteStr( C, "\nOBJECT ---- LIST ---- START ----\n" );
			while( p )
			{
				dumpobj( C, p );
				sgs_WriteStr( C, "\n" );
				p = p->next;
			}
			sgs_WriteStr( C, "OBJECT ---- LIST ---- END ----\n" );
		}
		return SGS_SUCCESS;
	case SGS_STAT_DUMP_FRAMES:
		{
			sgs_StackFrame* p = sgs_GetFramePtr( C, NULL, SGS_FALSE );
			sgs_WriteStr( C, "\nFRAME ---- LIST ---- START ----\n" );
			while( p != NULL )
			{
				const char* file, *name;
				int ln;
				sgs_StackFrameInfo( C, p, &name, &file, &ln );
				sgs_Writef( C, "FRAME \"%s\" in %s, line %d\n", name, file, ln );
				p = p->next;
			}
			sgs_WriteStr( C, "FRAME ---- LIST ---- END ----\n" );
		}
		return SGS_SUCCESS;
	case SGS_STAT_DUMP_STATS:
		{
			sgs_WriteStr( C, "\nSTATS ---- ---- ----\n" );
			sgs_Writef( C, "# allocs: %d\n", S->numallocs );
			sgs_Writef( C, "# frees: %d\n", S->numfrees );
			sgs_Writef( C, "# mem blocks: %d\n", S->numblocks );
			sgs_Writef( C, "# mem bytes: %d\n", S->memsize );
			sgs_Writef( C, "# objects: %d\n", S->objcount );
			sgs_Writef( C, "GC state: %s\n", S->redblue ? "red" : "blue" );
			sgs_WriteStr( C, "---- ---- ---- -----\n" );
		}
		return SGS_SUCCESS;
	case SGS_STAT_DUMP_RSRC:
		_sgs_dumprsrc( C->shared );
		return SGS_SUCCESS;
	case SGS_STAT_XDUMP_STACK:
		{
			ptrdiff_t i = 0, ssz = C->stack_top - C->stack_base;
			sgs_WriteStr( C, "\nVARIABLE -- ---- STACK ---- BASE ----\n" );
			while( i < ssz )
			{
				sgs_Variable* p = C->stack_base + i;
				if( p == C->stack_off )
				{
					sgs_WriteStr( C, "VARIABLE -- ---- STACK ---- OFFSET ----\n" );
				}
				sgs_Writef( C, "VARIABLE %02d ", (int) i );
				dumpvar( C, p );
				sgs_WriteStr( C, "\n" );
				sgs_DumpVar( C, *p, 6 );
				sgs_WriteStr( C, sgs_ToString( C, -1 ) );
				sgs_Pop( C, 1 );
				sgs_WriteStr( C, "\n" );
				i++;
			}
			sgs_WriteStr( C, "VARIABLE -- ---- STACK ---- TOP ----\n" );
		}
		return SGS_SUCCESS;
	case SGS_STAT_DUMP_SYMBOLS:
		{
			sgs_VHTVar *p, *pend;
			sgsSTD_RegistryIter( C, SGS_REG_SYM, &p, &pend );
			sgs_WriteStr( C, "\nSYMBOL ---- LIST ---- START ----\n" );
			while( p < pend )
			{
				sgs_iStr* str = p->key.data.S;
				if( p->key.type == SGS_VT_STRING )
				{
					sgs_WriteStr( C, "SYMBOL '" );
					ctx_print_safe( C, sgs_str_cstr( str ), str->size );
					sgs_WriteStr( C, "' = " );
					dumpvar( C, &p->val );
					sgs_WriteStr( C, "\n" );
				}
				p++;
			}
			sgs_WriteStr( C, "SYMBOL ---- LIST ---- END ----\n" );
		}
		return SGS_SUCCESS;
	default:
		return SGS_EINVAL;
	}
}

int32_t sgs_Cntl( SGS_CTX, int what, int32_t val )
{
	int32_t x;
	switch( what )
	{
	/* WP: last bit of C->state is not used */
	case SGS_CNTL_STATE: x = (int32_t) C->state; C->state = (uint32_t) val; return x;
	case SGS_CNTL_GET_STATE: return (int32_t) C->state;
	case SGS_CNTL_MINLEV: x = C->minlev; C->minlev = val; return x;
	case SGS_CNTL_GET_MINLEV: return C->minlev;
	case SGS_CNTL_APILEV: x = C->apilev; C->apilev = val; return x;
	case SGS_CNTL_GET_APILEV: return C->apilev;
	case SGS_CNTL_ERRNO: x = C->last_errno; C->last_errno = val ? 0 : errno; return x;
	case SGS_CNTL_SET_ERRNO: x = C->last_errno; C->last_errno = val; return x;
	case SGS_CNTL_GET_ERRNO: return C->last_errno;
	case SGS_CNTL_ERRSUP:
		if( C->sf_last )
		{
			x = C->sf_last->errsup;
			C->sf_last->errsup = (int16_t) val;
			return x;
		}
		return 0;
	case SGS_CNTL_GET_ERRSUP: return C->sf_last ? C->sf_last->errsup : 0;
	case SGS_CNTL_NUMRETVALS: return C->num_last_returned;
	case SGS_CNTL_GET_PAUSED: return C->state & SGS_STATE_LASTFUNCPAUSE ? 1 : 0;
	case SGS_CNTL_GET_ABORT: return C->state & SGS_STATE_LASTFUNCABORT ? 1 : 0;
	}
	return 0;
}

void sgs_StackFrameInfo( SGS_CTX, sgs_StackFrame* frame, const char** name, const char** file, int* line )
{
	int L = 0;
	const char* N = NULL;
	const char* F = NULL;
	sgs_Variable* func = frame->func;
	
	SGS_UNUSED( C );
	
	switch( func->type )
	{
	case SGS_VT_FUNC: {
		sgs_iFunc* iF = frame->func->data.F;
		N = "<anonymous function>";
		if( iF->sfuncname->size )
			N = sgs_str_cstr( iF->sfuncname );
		L = !iF->lineinfo ? 1 :
			iF->lineinfo[ SGS_MAX( frame->iptr - 1 - sgs_func_bytecode( iF ), 0 ) ];
		if( iF->sfilename->size )
			F = sgs_str_cstr( iF->sfilename );
		else
			F = "<buffer>";
	} break;
	case SGS_VT_CFUNC: {
		N = frame->nfname ? frame->nfname : "[C function]";
		F = "[C code]";
	} break;
	case SGS_VT_OBJECT: {
		sgs_VarObj* O = frame->func->data.O;
		N = O->iface->name ? O->iface->name : "<object>";
		F = "[C code]";
	} break;
#if SGS_DEBUG
	case SGS_VT_NULL: N = "<non-callable type>(NULL)"; F = "<?>"; break;
	case SGS_VT_BOOL: N = "<non-callable type>(BOOL)"; F = "<?>"; break;
	case SGS_VT_INT: N = "<non-callable type>(INT)"; F = "<?>"; break;
	case SGS_VT_REAL: N = "<non-callable type>(REAL)"; F = "<?>"; break;
	case SGS_VT_STRING: N = "<non-callable type>(STRING)"; F = "<?>"; break;
	case SGS_VT_PTR: N = "<non-callable type>(PTR)"; F = "<?>"; break;
	case SGS_VT_THREAD: N = "<non-callable type>(THREAD)"; F = "<?>"; break;
#endif
	default: {
		N = "<non-callable type>";
		F = "<buffer>";
	} break;
	}
	if( name ) *name = N;
	if( file ) *file = F;
	if( line ) *line = L;
}

sgs_StackFrame* sgs_GetFramePtr( SGS_CTX, sgs_StackFrame* from, int bwd )
{
	if( from )
		return bwd ? from->prev : from->next;
	return bwd ? C->sf_last : C->sf_first;
}


SGSBOOL sgs_RegisterType( SGS_CTX, const char* name, sgs_ObjInterface* iface )
{
	size_t len;
	sgs_VHTVar* p;
	SGS_SHCTX_USE;
	if( !iface )
	{
		sgs_Msg( C, SGS_APIERR, "sgs_RegisterType: cannot register NULL interface" );
		return SGS_FALSE;
	}
	len = strlen( name );
	/* WP: unimportant */
	p = sgs_vht_get_str( &S->typetable, name, (uint32_t) len, sgs_HashFunc( name, len ) );
	if( p )
		return SGS_FALSE;
	{
		sgs_Variable tmp = sgs_MakePtr( iface );
		sgs_PushStringBuf( C, name, (sgs_SizeVal) len );
		sgs_vht_set( &S->typetable, C, C->stack_top-1, &tmp );
		sgs_Pop( C, 1 );
	}
	return SGS_TRUE;
}

SGSBOOL sgs_UnregisterType( SGS_CTX, const char* name )
{
	SGS_SHCTX_USE;
	size_t len = strlen( name );
	/* WP: unimportant */
	sgs_VHTVar* p = sgs_vht_get_str( &S->typetable, name, (uint32_t) len, sgs_HashFunc( name, len ) );
	if( !p )
		return SGS_FALSE;
	sgs_vht_unset( &S->typetable, C, &p->key );
	return SGS_TRUE;
}

sgs_ObjInterface* sgs_FindType( SGS_CTX, const char* name )
{
	SGS_SHCTX_USE;
	size_t len = strlen( name );
	/* WP: unimportant */
	sgs_VHTVar* p = sgs_vht_get_str( &S->typetable, name, (uint32_t) len, sgs_HashFunc( name, len ) );
	if( p )
		return (sgs_ObjInterface*) p->val.data.P;
	return NULL;
}

SGSONE sgs_PushInterface( SGS_CTX, sgs_CFunc igfn )
{
	sgs_VHTVar* vv;
	sgs_Variable key = sgs_MakeCFunc( igfn );
	SGS_SHCTX_USE;
	
	vv = sgs_vht_get( &S->ifacetable, &key );
	if( vv )
	{
		return sgs_PushVariable( C, vv->val );
	}
	else
	{
		sgs_VarObj* obj;
		sgs_Variable val;
		sgs_StkIdx ssz;
		
		ssz = sgs_StackSize( C );
		sgs_PushVariable( C, key );
		sgs_Call( C, 0, 1 );
		if( sgs_ItemType( C, ssz ) != SGS_VT_OBJECT )
		{
			sgs_Msg( C, SGS_APIERR, "sgs_PushInterface: failed to create the interface" );
			sgs_SetStackSize( C, ssz );
			return sgs_PushNull( C );
		}
		val = sgs_StackItem( C, ssz );
		sgs_vht_set( &S->ifacetable, C, &key, &val );
		obj = sgs_GetObjectStruct( C, ssz );
		obj->is_iface = 1;
		obj->refcount--; /* being in interface table doesn't count */
		return 1;
	}
}

void sgs_InitInterface( SGS_CTX, sgs_Variable* var, sgs_CFunc igfn )
{
	sgs_PushInterface( C, igfn );
	sgs_StoreVariable( C, var );
}


void sgs_FuncName( SGS_CTX, const char* fnliteral )
{
	if( C->sf_last )
		C->sf_last->nfname = fnliteral;
}

int sgs_HasFuncName( SGS_CTX )
{
	return !!C->sf_last && !!C->sf_last->nfname;
}






static int isoneofN( char ch, const char* what, int size )
{
	const char* end = what + size;
	while( what < end )
	{
		if( ch == *what++ )
			return 1;
	}
	return 0;
}


/* debugging */


static void dumpnode( sgs_FTNode* N )
{
	switch( N->type )
	{
	case SGS_SFT_FCALL: printf( "FCALL" ); break;
	case SGS_SFT_INDEX: printf( "INDEX" ); break;
	case SGS_SFT_MIDXSET: printf( "MULTI_INDEX_SET" ); break;
	case SGS_SFT_MPROPSET: printf( "MULTI_PROPERTY_SET" ); break;
	case SGS_SFT_ARGMT: printf( "ARG " ); sgsT_DumpToken( N->token ); break;
	case SGS_SFT_ARGLIST: printf( "ARG_LIST" ); break;
	case SGS_SFT_VARLIST: printf( "VAR_LIST" ); break;
	case SGS_SFT_GVLIST: printf( "GLOBAL_VAR_LIST" ); break;
	case SGS_SFT_USELIST: printf( "USE_LIST" ); break;
	case SGS_SFT_EXPLIST: printf( "EXPR_LIST" ); break;
	case SGS_SFT_ARRLIST: printf( "ARRAY_LIST" ); break;
	case SGS_SFT_DCTLIST: printf( "DICT_LIST" ); break;
	case SGS_SFT_MAPLIST: printf( "MAP_LIST" ); break;
	case SGS_SFT_RETURN: printf( "RETURN" ); break;
	case SGS_SFT_BLOCK: printf( "BLOCK" ); break;
	case SGS_SFT_IFELSE: printf( "IF/ELSE" ); break;
	case SGS_SFT_WHILE: printf( "WHILE" ); break;
	case SGS_SFT_DOWHILE: printf( "DO/WHILE" ); break;
	case SGS_SFT_FOR: printf( "FOR" ); break;
	case SGS_SFT_FOREACH: printf( "FOR_EACH" ); break;
	case SGS_SFT_BREAK: printf( "BREAK" );
		if( *sgsT_Next( N->token ) == SGS_ST_NUMINT )
		{
			sgs_Int val;
			SGS_AS_INTEGER( val, sgsT_Next( N->token ) + 1 );
			printf( " %" PRId64, val );
		}
		break;
	case SGS_SFT_CONT: printf( "CONTINUE" );
		if( *sgsT_Next( N->token ) == SGS_ST_NUMINT )
		{
			sgs_Int val;
			SGS_AS_INTEGER( val, sgsT_Next( N->token ) + 1 );
			printf( " %" PRId64, val );
		}
		break;
	case SGS_SFT_DEFER: printf( "DEFER" ); break;
	case SGS_SFT_FUNC: printf( "FUNC" ); break;
	case SGS_SFT_CLASS: printf( "CLASS" ); break;
	case SGS_SFT_CLSINH: printf( "CLASS_INHERIT" ); break;
	case SGS_SFT_CLSINC: printf( "CLASS_INCLUDE" ); break;
	case SGS_SFT_CLSGLOB: printf( "CLASS_GLOBALS" ); break;
	case SGS_SFT_NEWCALL: printf( "NEW" ); break;
	case SGS_SFT_THRCALL: printf( "THREAD" ); break;
	case SGS_SFT_STHCALL: printf( "SUBTHREAD" ); break;
	default:
		if( N->token ) sgsT_DumpToken( N->token );
		if( N->type == SGS_SFT_OPER_P ) printf( " [post]" );
		break;
	}
}

static void ft_dump( sgs_FTNode* node, int level )
{
	int i;
	sgs_FTNode* N = node;
	if( !node ) return;

	for( i = 0; i < level; ++i ) printf( "  " );
	dumpnode( N );
	printf( "\n" );

	if( node->child )
	{
		for( i = 0; i < level; ++i ) printf( "  " );
		printf( "{\n" );

		ft_dump( node->child, level + 1 );

		for( i = 0; i < level; ++i ) printf( "  " );
		printf( "}\n" );
	}

	ft_dump( node->next, level );
}

void sgsFT_Dump( sgs_FTNode* tree )
{
	ft_dump( tree, 0 );
}


/*
// C O M P I L E R
*/


typedef struct _ftcomp
{
	SGS_CTX;
	sgs_TokenList at;
	sgs_FTNode* heap;
	sgs_FTNode* heapend;
}
FTComp;

#define SFTC FTComp* F
#define SFTRET static sgs_FTNode*
#define SFTC_AT F->at
#define SFTC_NEXT F->at = sgsT_Next( F->at )
#define SFTC_IS( type ) (*F->at == (type))
/* WP: char/uchar conversion */
#define SFTC_IN( buf, sz ) isoneofN( (char) *F->at, buf, sz )
#define SFTC_HASERR ( F->C->state & SGS_HAS_ERRORS )
#define SFTC_SETERR F->C->state |= SGS_HAS_ERRORS
#define SFTC_ISKEY( name ) sgsT_IsKeyword( F->at, name )
#define SFTC_IS_ID( name ) sgsT_IsIdent( F->at, name )
#define SFTC_LINENUM sgsT_LineNum( F->at )
#define SFTC_PRINTERR( what ) sgs_Msg( F->C, SGS_ERROR, "[line %d] " what, SFTC_LINENUM )
#define SFTC_UNEXP sgs_Msg( F->C, SGS_ERROR, "Unexpected end of code", SFTC_LINENUM )


static sgs_FTNode* _make_heap( SGS_CTX )
{
	sgs_FTNode* node = sgs_Alloc_n( sgs_FTNode, 33 );
	node->type = SGS_SFT_HEAPBIT;
	node->token = NULL;
	node->next = NULL;
	node->child = NULL;
	return node;
}

static sgs_FTNode* _make_node( SFTC, int type, sgs_TokenList token, sgs_FTNode* next, sgs_FTNode* child )
{
	sgs_FTNode* node = NULL;
	if( ( F->heapend->type >> 8 ) == 32 )
	{
		F->heapend = F->heapend->next = _make_heap( F->C );
	}
	node = F->heapend + 1 + ( F->heapend->type >> 8 );
	F->heapend->type += ( 1 << 8 );
	
	node->type = type;
	node->token = token;
	node->next = next;
	node->child = child;
	return node;
}
#define make_node( ty, tok, next, ch ) _make_node( F, ty, tok, next, ch )

void sgsFT_Destroy( SGS_CTX, sgs_FTNode* tree )
{
	while( tree )
	{
		sgs_FTNode* next = tree->next;
		/* if( tree->child ) sgsFT_Destroy( C, tree->child ); HEAP CHILD IS NOT A SEPARATE ALLOCATION */
		sgs_Dealloc( tree );
		tree = next;
	}
}
#define SFTC_DESTROY( node ) (void)0 /* sgsFT_Destroy( F->C, node ) */





SFTRET parse_exp( SFTC, char* endtoklist, int etlsize );
SFTRET parse_stmt( SFTC );
SFTRET parse_stmtlist( SFTC, char end );
SFTRET parse_function( SFTC, int inexp, sgs_TokenList namepfx );





SFTRET parse_arg( SFTC, int argid, char end )
{
	sgs_FTNode* node = NULL;
	char toks[ 3 ] = { ',', 0, 0 };
	toks[1] = end;

	SGS_FN_BEGIN;

	if( SFTC_IS(0) )
	{
		SFTC_UNEXP;
		goto fail;
	}

	if( SFTC_IS( SGS_ST_KEYWORD ) )
	{
		SFTC_PRINTERR( "Argument name cannot be a reserved keyword" );
		goto fail;
	}

	if( !SFTC_IS( SGS_ST_IDENT ) && !SFTC_IS( SGS_ST_KEYWORD ) )
	{
		sgs_Msg( F->C, SGS_ERROR, "[line %d] Unexpected token while parsing argument %d", SFTC_LINENUM, argid );
		goto fail;
	}

	node = make_node( SGS_SFT_ARGMT, SFTC_AT, NULL, NULL );
	SFTC_NEXT;

	if( SFTC_IS( SGS_ST_OP_SET ) )
	{
		SFTC_NEXT;
		if( SFTC_IS( end ) || SFTC_IS( ',' ) )
		{
			SFTC_PRINTERR( "Expected initializing expression" );
			goto fail;
		}
		node->child = parse_exp( F, toks, 2 );
		if( !node->child )
			goto fail;
	}

	SGS_FN_END;
	return node;

fail:
	if( node ) SFTC_DESTROY( node );
	SFTC_SETERR;
	SGS_FN_END;
	return 0;
}

/*
	FUNC / parses a list of arguments
	ARGS / context, function tree, token stream
	ERRS / unexpected token, @parse_arg
*/
SFTRET parse_arglist( SFTC, char end )
{
	sgs_FTNode* arglist = make_node( SGS_SFT_ARGLIST, NULL, NULL, NULL );
	sgs_FTNode* curnode = NULL;
	sgs_FTNode* argnode;
	int id = 1;

	SGS_FN_BEGIN;

	for(;;)
	{
		if( SFTC_IS( end ) )
		{
			break;
		}
		else if( SFTC_IS( 0 ) )
		{
			SFTC_UNEXP;
			goto fail;
		}
		else if( id == 1 || SFTC_IS( ',' ) )
		{
			if( id != 1 )
				SFTC_NEXT;
			argnode = parse_arg( F, id, end );
			if( !argnode && F->C->state & SGS_MUST_STOP )
			{
				goto fail;
			}
			if( curnode ) curnode->next = argnode;
			else arglist->child = argnode;
			curnode = argnode;

			id++;
		}
		else
		{
			sgs_Msg( F->C, SGS_ERROR, "[line %d] Expected ',' or '%c'", SFTC_LINENUM, end );
			goto fail;
		}
	}

	SGS_FN_END;
	return arglist;

fail:
	SFTC_SETERR;
	SFTC_DESTROY( arglist );
	return NULL;
}

/*
	FUNC / calculates the weight of the part
	ARGS / context, part of function tree
	ERRS / none
*/
static int part_weight( sgs_FTNode* part, int isfcall, int binary )
{
	sgs_BreakIf( !part );

	if( part->type == SGS_SFT_OPER && SGS_ST_OP_ASSIGN( *part->token ) )
		return 40;

	if( isfcall )
		return 9;

	if( part->type == SGS_SFT_OPER )
	{
		sgs_TokenType type = *part->token;
		if( binary )
		{
			if( SGS_ST_OP_BOOL( type ) )	return 30;
			if( type == SGS_ST_OP_NLOR )    return 29;
			if( type == SGS_ST_OP_RWCMP )   return 28; /* lower split prio .. */
			if( SGS_ST_OP_COMP( type ) )	return 29; /* .. than other comp ops */
			if( type == SGS_ST_OP_CAT )		return 27;
			if( SGS_ST_OP_BINOPS( type ) )	return 26;
			if( SGS_ST_OP_BINADD( type ) )	return 25;
			if( SGS_ST_OP_BINMUL( type ) )	return 24;
			if( type == SGS_ST_OP_NOT )     return 8;
			if( type == SGS_ST_OP_MMBR )	return 7;
			return 11;
		}

		/* unary operators */
		return 10;
	}

	/* everything else */
	return -1;
}



static sgs_LineNum findlinenum( sgs_FTNode* node ) /* local, next, child */
{
	sgs_LineNum ln = -1;

	if( node->token )
		return sgsT_LineNum( node->token );

	ln = findlinenum( node->next );
	if( ln != -1 ) return ln;

	ln = findlinenum( node->child );
	if( ln != -1 ) return ln;

	return -1;
}

static sgs_LineNum predictlinenum( sgs_FTNode* node ) /* next, child, local */
{
	sgs_LineNum ln = -1;

	ln = findlinenum( node->next );
	if( ln != -1 ) return ln;

	ln = predictlinenum( node->child );
	if( ln != -1 ) return ln;

	if( node->token )
		return sgsT_LineNum( node->token );

	return -1;
}


#define PFXFUNC_NONE 0
#define PFXFUNC_THREAD 1
#define PFXFUNC_SUBTHREAD 2
#define PFXFUNC_NEW 3
static int level_exp( SFTC, sgs_FTNode** tree )
{
	sgs_FTNode* node = *tree, *prev = NULL, *mpp = NULL;
	int weight = 0, isfcall, binary, count = 0;
	int pfxfuncmode = PFXFUNC_NONE;

	SGS_FN_BEGIN;
	sgs_BreakIf( !tree );

	if( !*tree )
	{
		SGS_FN_END;
		return 0;
	}
	
	if( sgsT_IsKeyword( node->token, "thread" ) )
	{
		pfxfuncmode = PFXFUNC_THREAD;
		node = node->next;
		(*tree)->next = NULL;
		SFTC_DESTROY( *tree );
		*tree = node;
	}
	if( sgsT_IsKeyword( node->token, "subthread" ) )
	{
		pfxfuncmode = PFXFUNC_SUBTHREAD;
		node = node->next;
		(*tree)->next = NULL;
		SFTC_DESTROY( *tree );
		*tree = node;
	}
	if( sgsT_IsKeyword( node->token, "new" ) )
	{
		pfxfuncmode = PFXFUNC_NEW;
		node = node->next;
		(*tree)->next = NULL;
		SFTC_DESTROY( *tree );
		*tree = node;
	}
	
	/* find the most powerful part (mpp) */
	while( node )
	{
		int leftmostsplit, curwt;
		
		count++;
		
		/* only interested in operators and subexpressions */
		if( node->type != SGS_SFT_OPER && node->type != SGS_SFT_EXPLIST && node->type != SGS_SFT_ARRLIST )
			goto _continue;
		
		/* "fcall" = [..]/(..) at the end, preceded by a data source */
		isfcall = node->type == SGS_SFT_EXPLIST || node->type == SGS_SFT_ARRLIST;
		if( isfcall ) isfcall = !!prev;
		if( isfcall ) isfcall = !node->next || ( node->type == SGS_SFT_ARRLIST && node->next->type == SGS_SFT_DCTLIST );
		if( isfcall ) isfcall = prev->type != SGS_SFT_OPER || !SGS_ST_OP_BINARY( *prev->token );
		
		/* op tests */
		binary = node->type == SGS_SFT_OPER;
		if( binary ) binary = prev && node->next;
		if( binary ) binary = prev->type != SGS_SFT_OPER || 
			*prev->token == SGS_ST_OP_INC || *prev->token == SGS_ST_OP_DEC;
		
		/* weighting */
		curwt = part_weight( node, isfcall, binary );
		leftmostsplit = 
			( node->type == SGS_SFT_OPER && SGS_ST_OP_ASSIGN( *node->token ) ) || /* assignment ops */
			( node->type == SGS_SFT_OPER && !binary ); /* unary ops */
		if( ( leftmostsplit && curwt > weight ) || ( !leftmostsplit && curwt >= weight ) )
		{
			weight = curwt;
			mpp = node;
		}
		
		/* move to next */
_continue:
		prev = node;
		node = node->next;
	}
	
	if( mpp )
	{
		/* function call */
		if( mpp->type == SGS_SFT_EXPLIST || mpp->type == SGS_SFT_ARRLIST )
		{
			int ret1, ret2;
			sgs_TokenList mpp_token = mpp->token;
			sgs_FTNode* se1 = *tree, *se2 = mpp, *se1i = *tree;
			while( se1i )
			{
				if( se1i->next == mpp )
					se1i->next = NULL;
				se1i = se1i->next;
			}
			
			SGS_FN_ENTER;
			ret1 = level_exp( F, &se1 );
			if( !ret1 )
			{
				*tree = NULL;
				if( se1 ) SFTC_DESTROY( se1 );
				if( se2 ) SFTC_DESTROY( se2 );
				SGS_FN_END;
				return 0;
			}
			
			if( mpp->type == SGS_SFT_ARRLIST && !mpp->child && mpp->next && mpp->next->type == SGS_SFT_DCTLIST )
			{
				/* a multiset (index) expression */
				mpp->type = SGS_SFT_MIDXSET;
				mpp->child = se1;
				mpp->child->next = mpp->next;
				mpp->next = NULL;
				*tree = mpp;
				goto retsuccess;
			}
			
			SGS_FN_ENTER;
			ret2 = level_exp( F, &se2 );
			if( !ret2 )
			{
				*tree = NULL;
				if( se1 ) SFTC_DESTROY( se1 );
				if( se2 ) SFTC_DESTROY( se2 );
				SGS_FN_END;
				return 0;
			}
			if( *mpp_token == SGS_ST_SBRKL )
			{
				/* array */
				if( !se2->child || se2->child->next )
				{
					sgs_Msg( F->C, SGS_ERROR, "[line %d] Invalid number of arguments "
						"in an array accessor", sgsT_LineNum( mpp_token ) );
					*tree = NULL;
					if( se1 ) SFTC_DESTROY( se1 );
					/* if( se2 ) */ SFTC_DESTROY( se2 );
					SGS_FN_END;
					return 0;
				}
				se1->next = se2->child;
				se2->child = NULL;
				SFTC_DESTROY( se2 );
				*tree = make_node( SGS_SFT_INDEX, mpp_token, NULL, se1 );
				goto retsuccess;
			}
			se1->next = se2;
			*tree = make_node( SGS_SFT_FCALL, mpp_token, NULL, se1 );
			goto retsuccess;
		}
		
		/* operators */
		if( mpp->type == SGS_SFT_OPER )
		{
			if( mpp == *tree )
				prev = NULL;
			else
			{
				prev = *tree;
				while( prev->next != mpp )
					prev = prev->next;
			}
			
			/* binary operators */
			if( mpp != *tree && mpp->next && 
				( prev->type != SGS_SFT_OPER || *prev->token == SGS_ST_OP_INC || *prev->token == SGS_ST_OP_DEC ) )
			{
				sgs_TokenList mpptoken = mpp->token;
				int ret1, ret2;
				sgs_FTNode* se1 = *tree, *se2 = mpp->next, *se1i = *tree;
				while( se1i )
				{
					if( se1i->next == mpp )
						se1i->next = NULL;
					se1i = se1i->next;
				}
				mpp->next = NULL;
				
				SGS_FN_ENTER;
				ret1 = level_exp( F, &se1 );
				if( !ret1 )
				{
					*tree = NULL;
					if( se1 ) SFTC_DESTROY( se1 );
					if( se2 ) SFTC_DESTROY( se2 );
					SFTC_DESTROY( mpp );
					SGS_FN_END;
					return 0;
				}
				
				if( mpptoken && *mpptoken == SGS_ST_OP_MMBR && !mpp->child && se2->type == SGS_SFT_DCTLIST )
				{
					/* a multiset (property) expression */
					mpp->type = SGS_SFT_MPROPSET;
					mpp->child = se1;
					mpp->child->next = se2;
					mpp->next = NULL;
					*tree = mpp;
					SGS_FN_END;
					goto retsuccess;
				}
				
				SFTC_DESTROY( mpp );
				
				SGS_FN_ENTER;
				ret2 = level_exp( F, &se2 );
				if( !ret2 )
				{
					*tree = NULL;
					if( se1 ) SFTC_DESTROY( se1 );
					if( se2 ) SFTC_DESTROY( se2 );
					SGS_FN_END;
					return 0;
				}
				
				se1->next = se2;
				*tree = make_node( SGS_SFT_OPER, mpptoken, NULL, se1 );
				SGS_FN_END;
				
				if( *mpptoken == SGS_ST_OP_CAT || *mpptoken == SGS_ST_OP_CATEQ )
				{
					/* merge in CAT on first operand (works only with non-assignment op) */
					if( *mpptoken == SGS_ST_OP_CAT && *se1->token == SGS_ST_OP_CAT )
					{
						/* target tree: tree { se1:children, se2 } */
						sgs_FTNode* tmp = se1->child;
						while( tmp->next )
							tmp = tmp->next;
						tmp->next = se2;
						(*tree)->child = se1->child;
						
						se1->child = NULL;
						se1->next = NULL;
						SFTC_DESTROY( se1 );
					}
					/* merge in CAT on second operand */
					if( *se2->token == SGS_ST_OP_CAT )
					{
						/* target tree: tree { children without se2, se2:children } */
						sgs_FTNode* tmp = (*tree)->child;
						while( tmp->next && tmp->next != se2 )
							tmp = tmp->next;
						sgs_BreakIf( tmp->next == NULL );
						tmp->next = se2->child;
						
						se2->child = NULL;
						SFTC_DESTROY( se2 );
					}
				}
				
				goto retsuccess;
			}
			/* unary operators
				must be one of these cases:
				- no tokens before operator
				- is inc|dec and there are tokens either before or after the operator
					- can't have both, can't have neither
			*/
			else if( SGS_ST_OP_UNARY( *mpp->token ) && ( mpp == *tree || 
				( ( *mpp->token == SGS_ST_OP_INC || *mpp->token == SGS_ST_OP_DEC ) &&
				( mpp != *tree ) != ( !!mpp->next ) ) ) )
			{
				int ret1;
				if( !mpp->next )
				{
					sgs_FTNode* pp = *tree;
					if( pp == mpp ) goto fail;

					while( pp->next != mpp )
						pp = pp->next;
					pp->next = NULL;
					mpp->child = *tree;
					*tree = mpp;

					mpp->type = SGS_SFT_OPER_P;
				}
				else
				{
					mpp->child = mpp->next;
					mpp->next = NULL;
				}
				SGS_FN_ENTER;
				ret1 = level_exp( F, &mpp->child );
				if( !ret1 )
				{
					SGS_FN_END;
					return 0;
				}
				goto retsuccess;
			}
			/* problems */
			else goto fail;
		}
	}
	
	if( count <= 1 )
	{
retsuccess:
		switch( pfxfuncmode )
		{
		case PFXFUNC_THREAD:
			if( (*tree)->type != SGS_SFT_FCALL )
			{
				SFTC_PRINTERR( "expected function call after 'thread'" );
				goto fail_no_err;
			}
			(*tree)->type = SGS_SFT_THRCALL;
			break;
		case PFXFUNC_SUBTHREAD:
			if( (*tree)->type != SGS_SFT_FCALL )
			{
				SFTC_PRINTERR( "expected function call after 'subthread'" );
				goto fail_no_err;
			}
			(*tree)->type = SGS_SFT_STHCALL;
			break;
		case PFXFUNC_NEW:
			if( (*tree)->type != SGS_SFT_FCALL ||
				(*tree)->child->type == SGS_SFT_OPER )
			{
				SFTC_PRINTERR( "expected plain function call after 'new'" );
				goto fail_no_err;
			}
			(*tree)->type = SGS_SFT_NEWCALL;
			break;
		default:
			break;
		}
		
		SGS_FN_END;
		return 1;
	}

	/* in case we failed unexpectedly, dump & debug */
	sgs_Msg( F->C, SGS_ERROR, "[line %d] Missing operators or separators", predictlinenum( *tree ) );
	F->C->state |= SGS_HAS_ERRORS;
#if SGS_DEBUG && SGS_DEBUG_DATA
	sgsFT_Dump( *tree );
#endif
	SGS_FN_END;
	return 0;

fail:
	if( mpp == NULL )
		mpp = *tree;
	sgs_Msg( F->C, SGS_ERROR, "[line %d] Invalid expression", mpp ? sgsT_LineNum( mpp->token ) : 0 );
fail_no_err:
#if SGS_DEBUG && SGS_DEBUG_DATA
	sgsFT_Dump( *tree );
#endif
	SGS_FN_END;
	return 0;
}


SFTRET parse_dict( SFTC )
{
	sgs_TokenList startok = SFTC_AT;
	sgs_FTNode* expr = NULL, *fexp = NULL, *cur;
	/* dictionary expression */
	SFTC_NEXT;
	while( !SFTC_IS( '}' ) )
	{
		/* in a dict literal, with the exception of values, keywords = identifers */
		int is_ident = SFTC_IS( SGS_ST_IDENT ) || SFTC_IS( SGS_ST_KEYWORD );
		int is_varkey = SFTC_IS( '[' );
		if( !is_ident && !is_varkey && !SFTC_IS( SGS_ST_STRING ) )
		{
			SFTC_PRINTERR( "expected key identifier, string or '[' in dictionary expression" );
			goto fail;
		}
		
		/* make key */
		if( is_varkey )
		{
			SFTC_NEXT;
			cur = parse_exp( F, "]", 1 );
			if( !cur )
				goto fail;
		}
		else
		{
			cur = make_node( SGS_SFT_ARGMT, SFTC_AT, NULL, NULL );
		}
		SFTC_NEXT;
		
		/* add key to node list */
		if( !fexp )
			expr = fexp = cur;
		else
		{
			expr->next = cur;
			expr = expr->next;
		}
		
		if( !SFTC_IS( SGS_ST_OP_SET ) )
		{
			if( is_ident )
			{
				if( SFTC_IS( ',' ) || SFTC_IS( '}' ) )
				{
					expr->next = make_node( SGS_SFT_IDENT, expr->token, NULL, NULL );
					expr = expr->next;
				}
				else
				{
					SFTC_PRINTERR( "Expected '=', ',' or '}' after dictionary key" );
					goto fail;
				}
			}
			else
			{
				SFTC_PRINTERR( "Expected '=' in dictionary expression "
					"/ missing closing bracket before '{'" );
				goto fail;
			}
		}
		else
		{
			SFTC_NEXT;
			
			expr->next = parse_exp( F, ",}", 2 );
			if( !expr->next )
				goto fail;
			else
				expr = expr->next;
		}
		
		if( SFTC_IS( ',' ) )
			SFTC_NEXT;
	}
	return make_node( SGS_SFT_DCTLIST, startok, NULL, fexp );
	
fail:
	if( fexp )
		SFTC_DESTROY( fexp );
	return NULL;
}


SFTRET parse_exp( SFTC, char* endtoklist, int etlsize )
{
	sgs_FTNode* node, *cur;
	char prev = 0;

	SGS_FN_BEGIN;

	if( SFTC_IS( 0 ) )
	{
		SFTC_UNEXP;
		SFTC_SETERR;
		SGS_FN_END;
		return NULL;
	}

	/* special cases */
	if( SFTC_ISKEY( "var" ) )
	{
		SFTC_NEXT;
		node = parse_arglist( F, endtoklist[ endtoklist[0] == ',' ] );
		if( node )
			node->type = SGS_SFT_VARLIST;
		SGS_FN_END;
		return node;
	}
	if( SFTC_ISKEY( "global" ) )
	{
		SFTC_NEXT;
		node = parse_arglist( F, endtoklist[ endtoklist[0] == ',' ] );
		if( node )
			node->type = SGS_SFT_GVLIST;
		SGS_FN_END;
		return node;
	}

	cur = node = make_node( 0, NULL, NULL, NULL );

	for(;;)
	{
		if( SFTC_IN( endtoklist, etlsize ) )
		{
			break;
		}
		else if( SFTC_IS( 0 ) )
		{
			SFTC_UNEXP;
			goto fail;
		}
		else if( SFTC_IS( SGS_ST_STRING )
			  || SFTC_IS( SGS_ST_NUMINT )
			  || SFTC_IS( SGS_ST_NUMREAL ) )
			cur = cur->next = make_node( SGS_SFT_CONST, SFTC_AT, NULL, NULL );
		else if( SFTC_IS( SGS_ST_IDENT ) )
		{
			if( SFTC_IS_ID( "map" ) && sgsT_Next( SFTC_AT ) && *sgsT_Next( SFTC_AT ) == '{' )
			{
				SFTC_NEXT;
				cur->next = parse_dict( F );
				if( !cur->next )
					goto fail;
				cur = cur->next;
				cur->type = SGS_SFT_MAPLIST;
			}
			else
				cur = cur->next = make_node( SGS_SFT_IDENT, SFTC_AT, NULL, NULL );
		}
		else if( SFTC_IS( SGS_ST_KEYWORD ) )
		{
			if( SFTC_ISKEY( "function" ) )
			{
				cur->next = parse_function( F, 1, NULL );
				if( !cur->next )
					goto fail;
				cur = cur->next;
				continue;
			}
			else
				cur = cur->next = make_node( SGS_SFT_KEYWORD, SFTC_AT, NULL, NULL );
		}
		else if( SGS_ST_ISOP( *SFTC_AT ) )
			cur = cur->next = make_node( SGS_SFT_OPER, SFTC_AT, NULL, NULL );
		else if( SGS_ST_ISSPEC( *SFTC_AT ) )
		{
			/* array accesor / argument list / subexpression */
			if( SFTC_IS( '(' ) || SFTC_IS( '[' ) )
			{
				int isidx = prev == SGS_ST_IDENT || prev == ')' || prev == ']';
				char cend = SFTC_IS( '(' ) ? ')' : ']';
				sgs_FTNode* exprlist = make_node( SFTC_IS( '(' ) ? SGS_SFT_EXPLIST : SGS_SFT_ARRLIST, SFTC_AT, NULL, NULL );
				sgs_FTNode* expr, * curexpr = NULL;
				char endcstr[ 3 ] = { ',', 0, 0 };
				endcstr[1] = cend;

				SFTC_NEXT;
				/* if this is an empty expression (for a function call), do not process it further */
				if( !SFTC_IS( cend ) )
				{
					for(;;)
					{
						/* if not index, extra ',' is allowed */
						if( !isidx && SFTC_IS( cend ) )
						{
							SFTC_NEXT;
							break;
						}

						SGS_FN_ENTER;
						expr = parse_exp( F, endcstr, 2 );
						if( !expr )
						{
							SFTC_DESTROY( exprlist );
							goto fail;
						}

						if( curexpr )
							curexpr->next = expr;
						else
							exprlist->child = expr;
						curexpr = expr;

						if( SFTC_IS( cend ) )
						{
							SFTC_NEXT;
							break;
						}

						SFTC_NEXT;
					}
				}
				else
					SFTC_NEXT;

				cur = cur->next = exprlist;
				continue;
			}
			/* dictionary */
			else if( SFTC_IS( '{' ) )
			{
				cur->next = parse_dict( F );
				if( !cur->next )
					goto fail;
				cur = cur->next;
			}
			else
			{
				sgs_Msg( F->C, SGS_ERROR, "[line %d] Unexpected token '%c' found!", SFTC_LINENUM, *SFTC_AT );
				F->C->state |= SGS_MUST_STOP;
			}
		}
		else
		{
			SFTC_PRINTERR( "INTERNAL ERROR in parse_exp: unknown token found!" );
			F->C->state |= SGS_MUST_STOP;
		}

		if( F->C->state & SGS_MUST_STOP )
		{
			SFTC_DESTROY( node );
			SGS_FN_END;
			return NULL;
		}
		
		/* WP: char/uchar conversion */
		prev = (char) *SFTC_AT;
		SFTC_NEXT;
	}

	cur = node->next;
	{
		node->child = NULL;
		node->next = NULL;
		SFTC_DESTROY( node );
	}
	node = cur;
	if( !node )
	{
		SFTC_PRINTERR( "Empty expression found" );
		goto fail;
	}

	if( !level_exp( F, &node ) )
		goto fail;
	
	SGS_FN_END;
	return node;

fail:
	SFTC_SETERR;
	if( node ) SFTC_DESTROY( node );
	SGS_FN_END;
	return NULL;
}





SFTRET parse_explist( SFTC, char endtok )
{
	sgs_FTNode* explist = make_node( SGS_SFT_EXPLIST, SFTC_AT, NULL, NULL );
	sgs_FTNode* curexp = NULL, *node;
	char endtoklist[] = { ',', 0, 0 };
	endtoklist[1] = endtok;

	SGS_FN_BEGIN;

	for(;;)
	{
		if( SFTC_IS( endtok ) )
		{
			break;
		}
		else if( SFTC_IS( 0 ) )
		{
			SFTC_UNEXP;
			goto fail;
		}
		else if( SFTC_IS( ',' ) || SFTC_AT == explist->token )
		{
			if( SFTC_AT != explist->token )
				SFTC_NEXT;
			node = parse_exp( F, endtoklist, 2 );
			if( !node )
				goto fail;
			if( curexp )
				curexp->next = node;
			else
				explist->child = node;
			curexp = node;
		}
		else
		{
			sgs_Msg( F->C, SGS_ERROR, "[line %d] Expected ',' or '%c'", SFTC_LINENUM, endtok );
			goto fail;
		}
	}

	SGS_FN_END;
	return explist;

fail:
	SFTC_SETERR;
	SFTC_DESTROY( explist );
	SGS_FN_END;
	return NULL;
}

SFTRET parse_if( SFTC )
{
	sgs_FTNode *node = NULL, *nexp = NULL, *nif = NULL, *nelse = NULL;
	sgs_TokenList begin = SFTC_AT;

	SGS_FN_BEGIN;

	SFTC_NEXT;
	if( !SFTC_IS( '(' ) )
	{
		SFTC_PRINTERR( "Expected '(' after 'if'" );
		goto fail;
	}

	SFTC_NEXT;

	nexp = parse_exp( F, ")", 1 );
	if( !nexp ) goto fail;
	SFTC_NEXT;

	nif = parse_stmt( F );
	if( !nif ) goto fail;

	if( SFTC_ISKEY( "else" ) )
	{
		SFTC_NEXT;
		nelse = parse_stmt( F );
		if( !nelse ) goto fail;
	}

	nexp->next = nif;
	nif->next = nelse;
	node = make_node( SGS_SFT_IFELSE, begin, NULL, nexp );

	SGS_FN_END;
	return node;

fail:
	if( nexp ) SFTC_DESTROY( nexp );
	if( nif ) SFTC_DESTROY( nif );
	if( nelse ) SFTC_DESTROY( nelse );
	SFTC_SETERR;
	SGS_FN_END;
	return NULL;
}

SFTRET parse_while( SFTC )
{
	sgs_FTNode *node, *nexp = NULL, *nwhile = NULL;
	sgs_TokenList begin = SFTC_AT;

	SGS_FN_BEGIN;

	SFTC_NEXT;
	if( !SFTC_IS( '(' ) )
	{
		SFTC_PRINTERR( "Expected '(' after 'while'" );
		goto fail;
	}

	SFTC_NEXT;

	nexp = parse_exp( F, ")", 1 );
	if( !nexp ) goto fail;
	SFTC_NEXT;

	nwhile = parse_stmt( F );
	if( !nwhile ) goto fail;

	nexp->next = nwhile;
	node = make_node( SGS_SFT_WHILE, begin, NULL, nexp );

	SGS_FN_END;
	return node;

fail:
	if( nexp ) SFTC_DESTROY( nexp );
	if( nwhile ) SFTC_DESTROY( nwhile );
	SFTC_SETERR;
	SGS_FN_END;
	return NULL;
}

SFTRET parse_dowhile( SFTC )
{
	sgs_FTNode *node, *nexp = NULL, *nwhile = NULL;
	sgs_TokenList begin = SFTC_AT;

	SGS_FN_BEGIN;

	SFTC_NEXT;
	nwhile = parse_stmt( F );
	if( !nwhile ) goto fail;

	if( !SFTC_ISKEY( "while" ) )
	{
		SFTC_PRINTERR( "Expected 'while' after statement in do/while" );
		goto fail;
	}

	SFTC_NEXT;
	if( !SFTC_IS( '(' ) )
	{
		SFTC_PRINTERR( "Expected '(' after 'while'" );
		goto fail;
	}

	SFTC_NEXT;

	nexp = parse_exp( F, ")", 1 );
	if( !nexp ) goto fail;
	SFTC_NEXT;

	nexp->next = nwhile;
	node = make_node( SGS_SFT_DOWHILE, begin, NULL, nexp );

	SGS_FN_END;
	return node;

fail:
	if( nexp ) SFTC_DESTROY( nexp );
	if( nwhile ) SFTC_DESTROY( nwhile );
	SFTC_SETERR;
	SGS_FN_END;
	return NULL;
}

SFTRET parse_for( SFTC )
{
	sgs_FTNode *node, *ninit = NULL, *nexp = NULL, *nincr = NULL, *nwhile = NULL;
	sgs_TokenList begin = SFTC_AT;

	SGS_FN_BEGIN;

	SFTC_NEXT;
	if( !SFTC_IS( '(' ) )
	{
		SFTC_PRINTERR( "Expected '(' after 'for'" );
		goto fail;
	}

	SFTC_NEXT;

	ninit = parse_explist( F, ';' );
	if( !ninit ) goto fail;
	SFTC_NEXT;

	nexp = parse_explist( F, ';' );
	if( !nexp ) goto fail;
	SFTC_NEXT;

	nincr = parse_explist( F, ')' );
	if( !nincr ) goto fail;
	SFTC_NEXT;

	nwhile = parse_stmt( F );
	if( !nwhile ) goto fail;

	ninit->next = nexp;
	nexp->next = nincr;
	nincr->next = nwhile;
	node = make_node( SGS_SFT_FOR, begin, NULL, ninit );

	SGS_FN_END;
	return node;

fail:
	if( ninit ) SFTC_DESTROY( ninit );
	if( nexp ) SFTC_DESTROY( nexp );
	if( nincr ) SFTC_DESTROY( nincr );
	if( nwhile ) SFTC_DESTROY( nwhile );
	SFTC_SETERR;
	SGS_FN_END;
	return NULL;
}

SFTRET parse_foreach( SFTC )
{
	/*
		(x: => null=>ident
		(x,x: => ident=>ident
		(x,: => ident=>null
	*/
	sgs_FTNode *node, *nvar = NULL, *nkey = NULL, *nexp = NULL, *nwhile = NULL;
	sgs_TokenList begin = SFTC_AT;

	SGS_FN_BEGIN;

	SFTC_NEXT;
	if( !SFTC_IS( '(' ) )
	{
		SFTC_PRINTERR( "Expected '(' after 'foreach'" );
		goto fail;
	}

	SFTC_NEXT;
	if( !SFTC_IS( SGS_ST_IDENT ) )
	{
		SFTC_PRINTERR( "Expected identifier after '(' in 'foreach'" );
		goto fail;
	}
	/* (x:e) */
	nkey = make_node( SGS_SFT_NULL, SFTC_AT, NULL, NULL );
	nvar = make_node( SGS_SFT_IDENT, SFTC_AT, NULL, NULL );
	SFTC_NEXT;

	if( !SFTC_IS( ':' ) && !SFTC_IS( ',' ) )
	{
		SFTC_PRINTERR( "Expected ':' or ',' after identifier in 'foreach'" );
		goto fail;
	}

	if( SFTC_IS( ',' ) )
	{
		SFTC_NEXT;
		if( !SFTC_IS( SGS_ST_IDENT ) && !SFTC_IS( ':' ) )
		{
			SFTC_PRINTERR( "Expected identifier or ':' after ',' in 'foreach'" );
			goto fail;
		}

		if( SFTC_IS( SGS_ST_IDENT ) )
		{
			/* (x,x:e) */
			nkey->type = SGS_SFT_IDENT;
			nvar->token = SFTC_AT;
			SFTC_NEXT;
		}
		else
		{
			/* (x,:e) */
			nkey->type = SGS_SFT_IDENT;
			nvar->type = SGS_SFT_NULL;
		}

		if( !SFTC_IS( ':' ) )
		{
			SFTC_PRINTERR( "Expected ':' after identifier #2 or ',' in 'foreach'" );
			goto fail;
		}
	}

	SFTC_NEXT;

	nexp = parse_explist( F, ')' );
	if( !nexp ) goto fail;

	SFTC_NEXT;
	nwhile = parse_stmt( F );
	if( !nwhile ) goto fail;

	nkey->next = nvar;
	nvar->next = nexp;
	nexp->next = nwhile;
	node = make_node( SGS_SFT_FOREACH, begin, NULL, nkey );

	SGS_FN_END;
	return node;

fail:
	if( nvar ) SFTC_DESTROY( nvar );
	if( nkey ) SFTC_DESTROY( nkey );
	if( nexp ) SFTC_DESTROY( nexp );
	if( nwhile ) SFTC_DESTROY( nwhile );
	SGS_FN_END;
	return NULL;
}

SFTRET parse_function( SFTC, int inexp, sgs_TokenList namepfx )
{
	int hasname = 0;
	sgs_FTNode *node, *nname = NULL, *nargs = NULL, *nbody = NULL, *nclos = NULL;
	sgs_TokenList begin = SFTC_AT;
	
	SGS_FN_BEGIN;
	
	SFTC_NEXT;
	if( !inexp && !namepfx )
	{
		if( !SFTC_IS( SGS_ST_IDENT ) )
		{
			SFTC_PRINTERR( "Expected identifier after 'function'" );
			goto fail;
		}
	}
	
	if( SFTC_IS( SGS_ST_IDENT ) )
	{
		hasname = 1;
		nname = make_node( SGS_SFT_IDENT, SFTC_AT, NULL, NULL );
		SFTC_NEXT;
		if( namepfx )
		{
			nname = make_node( SGS_SFT_CLSPFX, namepfx, NULL, nname );
		}
		else if( SFTC_IS( SGS_ST_OP_MMBR ) )
		{
			nname = make_node( SGS_SFT_OPER, SFTC_AT, NULL, nname );
			SFTC_NEXT;
			if( !SFTC_IS( SGS_ST_IDENT ) )
			{
				SFTC_PRINTERR( "Expected identifier after 'function', identifier and '.'" );
				goto fail;
			}
			else
			{
				nname->child->next = make_node( SGS_SFT_IDENT, SFTC_AT, NULL, NULL );
				SFTC_NEXT;
			}
		}
	}
	
	if( !SFTC_IS( '(' ) )
	{
		if( !hasname )
			SFTC_PRINTERR( "Expected '(' after 'function'" );
		else
			SFTC_PRINTERR( "Expected '(' after 'function' and its name" );
		goto fail;
	}
	
	SFTC_NEXT;
	
	nargs = parse_arglist( F, ')' );
	if( !nargs ) goto fail;
	SFTC_NEXT;
	
	if( SFTC_ISKEY( "use" ) )
	{
		/* closure */
		SFTC_NEXT;
		if( !SFTC_IS( '(' ) )
		{
			SFTC_PRINTERR( "Expected '(' after 'use' in 'function'" );
			goto fail;
		}
		SFTC_NEXT;
		nclos = parse_arglist( F, ')' );
		if( !nclos ) goto fail;
		nclos->type = SGS_SFT_USELIST;
		SFTC_NEXT;
	}
	else
		nclos = make_node( SGS_SFT_USELIST, SFTC_AT, NULL, NULL );
	
	if( !SFTC_IS( '{' ) && !SFTC_IS( SGS_ST_OP_SET ) )
	{
		SFTC_PRINTERR( "Expected '{', '=' or 'use'" );
		goto fail;
	}
	
	if( SFTC_IS( SGS_ST_OP_SET ) )
	{
		SFTC_NEXT;
		nbody = parse_explist( F, ';' );
		if( !nbody ) goto fail;
		nbody->type = SGS_SFT_RETURN;
		SFTC_NEXT;
	}
	else
	{
		nbody = parse_stmt( F );
		if( !nbody ) goto fail;
	}
	
	nargs->next = nclos;
	nclos->next = nbody;
	nbody->next = nname;
	node = make_node( SGS_SFT_FUNC, begin, NULL, nargs );
	
	SGS_FN_END;
	return node;
	
fail:
	if( nname ) SFTC_DESTROY( nname );
	if( nargs ) SFTC_DESTROY( nargs );
	if( nclos ) SFTC_DESTROY( nclos );
	if( nbody ) SFTC_DESTROY( nbody );
	SFTC_SETERR;
	SGS_FN_END;
	return NULL;
}

SFTRET parse_class( SFTC )
{
	sgs_FTNode *node, **nit;
		
	node = make_node( SGS_SFT_CLASS, SFTC_AT, NULL, NULL );
	nit = &node->child;
	
	if( !SFTC_IS( SGS_ST_IDENT ) || SFTC_IS_ID( "class" ) )
	{
		SFTC_PRINTERR( "Expected identifier after 'class'" );
		goto fail;
	}
	*nit = make_node( SGS_SFT_IDENT, SFTC_AT, NULL, NULL );
	nit = &(*nit)->next;
	SFTC_NEXT;
	
	if( SFTC_IS( ':' ) )
	{
		SFTC_NEXT;
		if( !SFTC_IS( SGS_ST_IDENT ) )
		{
			SFTC_PRINTERR( "Expected identifier after ':' in class" );
			goto fail;
		}
		*nit = make_node( SGS_SFT_CLSINH, SFTC_AT, NULL, NULL );
		nit = &(*nit)->next;
		SFTC_NEXT;
	}
	if( !SFTC_IS( '{' ) )
	{
		SFTC_PRINTERR( "Expected '{' after (inherited) class name" );
		goto fail;
	}
	SFTC_NEXT;
	
	while( !SFTC_IS( '}' ) )
	{
		sgs_FTNode* nn;
		if( SFTC_ISKEY( "global" ) )
		{
			SFTC_NEXT;
			nn = parse_arglist( F, ';' );
			if( !nn )
				goto fail;
			nn->type = SGS_SFT_CLSGLOB;
			*nit = nn;
			nit = &(*nit)->next;
			SFTC_NEXT;
		}
		else if( SFTC_ISKEY( "function" ) )
		{
			nn = parse_function( F, 0, node->child->token );
			if( !nn )
				goto fail;
			*nit = nn;
			nit = &(*nit)->next;
		}
		else
		{
			SFTC_PRINTERR( "Unexpected token in class" );
			goto fail;
		}
	}
	
	SGS_FN_END;
	return node;
	
fail:
	if( node ) SFTC_DESTROY( node );
	SFTC_SETERR;
	SGS_FN_END;
	return NULL;
}

SFTRET parse_command( SFTC, int multi )
{
	sgs_FTNode *nargs = NULL;
	sgs_TokenList begin = SFTC_AT;

	SGS_FN_BEGIN;
	SFTC_NEXT;
	
	nargs = parse_explist( F, ';' );
	if( !nargs ) goto fail;
	SFTC_NEXT;
	
	if( multi )
	{
		/* one argument to one function call */
		sgs_FTNode *r = NULL, *n = NULL, *p = nargs->child;
		
		if( !p )
		{
			if( nargs ) SFTC_DESTROY( nargs );
			SGS_FN_END;
			return make_node( SGS_SFT_BLOCK, begin, NULL, NULL );
		}
		
		nargs->child = NULL;
		SFTC_DESTROY( nargs );
		
		while( p )
		{
			sgs_FTNode* nn = make_node( SGS_SFT_FCALL, begin, NULL,
				make_node( SGS_SFT_IDENT, begin, 
					make_node( SGS_SFT_EXPLIST, p->token, NULL, p ),
						NULL ) );
			sgs_FTNode* pp = p;
			p = p->next;
			pp->next = NULL;
			if( !r )
				r = n = nn;
			else
				n = n->next = nn;
		}
		
		return make_node( SGS_SFT_BLOCK, begin, NULL, r );
	}
	else
	{
		/* one function call */
		sgs_FTNode* nname = make_node( SGS_SFT_IDENT, begin, nargs, NULL );
		SGS_FN_END;
		return make_node( SGS_SFT_FCALL, begin, NULL, nname );
	}
	
fail:
	if( nargs ) SFTC_DESTROY( nargs );
	SFTC_SETERR;
	SGS_FN_END;
	return NULL;
}

SFTRET parse_stmt( SFTC )
{
	sgs_FTNode* node;

	SGS_FN_BEGIN;

	if( SFTC_IS(0) )
	{
		SFTC_UNEXP;
		goto fail;
	}

	/* IF / ELSE */
	if( SFTC_ISKEY( "if" ) ) { node = parse_if( F ); SGS_FN_END; return node; }
	else if( SFTC_ISKEY( "else" ) )
	{
		SFTC_PRINTERR( "Found 'else' without matching 'if'" );
		goto fail;
	}
	/* WHILE */
	else if( SFTC_ISKEY( "while" ) ) { node = parse_while( F ); SGS_FN_END; return node; }
	/* DO / WHILE */
	else if( SFTC_ISKEY( "do" ) ) { node = parse_dowhile( F ); SGS_FN_END; return node; }
	/* FOR */
	else if( SFTC_ISKEY( "for" ) ) { node = parse_for( F ); SGS_FN_END; return node; }
	/* FOREACH */
	else if( SFTC_ISKEY( "foreach" ) ) { node = parse_foreach( F ); SGS_FN_END; return node; }
	/* BREAK */
	else if( SFTC_ISKEY( "break" ) )
	{
		sgs_TokenList orig = SFTC_AT;
		SFTC_NEXT;

		if( SFTC_IS( SGS_ST_NUMINT ) )
		{
			sgs_Int blev;
			SGS_AS_INTEGER( blev, SFTC_AT + 1 );
			if( blev < 1 || blev > 255 )
			{
				SFTC_PRINTERR( "Invalid break level (can be between 1 and 255)" );
				goto fail;
			}
			SFTC_NEXT;
		}
		
		if( !SFTC_IS( ';' ) )
		{
			SFTC_UNEXP;
			goto fail;
		}
		SFTC_NEXT;

		node = make_node( SGS_SFT_BREAK, orig, NULL, NULL );

		SGS_FN_END;
		return node;
	}
	/* CONTINUE */
	else if( SFTC_ISKEY( "continue" ) )
	{
		sgs_TokenList orig = SFTC_AT;
		SFTC_NEXT;

		if( SFTC_IS( SGS_ST_NUMINT ) )
		{
			sgs_Int blev;
			SGS_AS_INTEGER( blev, SFTC_AT + 1 );
			if( blev < 1 || blev > 255 )
			{
				SFTC_PRINTERR( "Invalid continue level (can be between 1 and 255)" );
				goto fail;
			}
			SFTC_NEXT;
		}
		
		if( !SFTC_IS( ';' ) )
		{
			SFTC_UNEXP;
			goto fail;
		}
		SFTC_NEXT;

		node = make_node( SGS_SFT_CONT, orig, NULL, NULL );

		SGS_FN_END;
		return node;
	}
	/* FUNCTION */
	else if( SFTC_ISKEY( "function" ) ) { node = parse_function( F, 0, NULL ); SGS_FN_END; return node; }
	/* RETURN */
	else if( SFTC_ISKEY( "return" ) )
	{
		SFTC_NEXT;
		node = parse_explist( F, ';' );
		if( !node )
			goto fail;

		node->type = SGS_SFT_RETURN;

		SFTC_NEXT;
		SGS_FN_END;
		return node;
	}
	/* VAR / GLOBAL - reuse code in parse_exp */
	else if( SFTC_ISKEY( "var" ) || SFTC_ISKEY( "global" ) )
	{
		node = parse_exp( F, ";", 1 );
		if( !node )
			goto fail;
		if( !SFTC_IS( ';' ) )
		{
			SFTC_UNEXP;
			SFTC_DESTROY( node );
			goto fail;
		}

		SFTC_NEXT;
		SGS_FN_END;
		return node;
	}
	/* DEFERRED BLOCKS (scope destruction) */
	else if( SFTC_ISKEY( "defer" ) )
	{
		sgs_TokenList orig = SFTC_AT;
		SFTC_NEXT;
		node = parse_stmt( F );
		if( !node )
			goto fail;
		node = make_node( SGS_SFT_DEFER, orig, NULL, node );
		SGS_FN_END;
		return node;
	}
	/* COMMAND HELPERS */
#define NOT_FCALL ( !sgsT_Next( F->at ) || '(' != *sgsT_Next( F->at ) )
	/* CLASS */
	else if( SFTC_IS_ID( "class" ) && NOT_FCALL )
	{
		SFTC_NEXT;
		node = parse_class( F );
		if( !node )
			goto fail;
		SFTC_NEXT;
		SGS_FN_END;
		return node;
	}
	/* SIMPLE COMMANDS */
	else if((
		SFTC_IS_ID( "print" ) ||
		SFTC_IS_ID( "println" ) ||
		SFTC_IS_ID( "yield" )
		) && NOT_FCALL )
	{
		node = parse_command( F, 0 );
		SGS_FN_END;
		return node;
	}
	/* MULTIPLIED COMMANDS */
	else if( SFTC_IS_ID( "include" ) && NOT_FCALL )
	{
		node = parse_command( F, 1 );
		SGS_FN_END;
		return node;
	}
	/* BLOCK OF STATEMENTS */
	else if( SFTC_IS( SGS_ST_CBRKL ) )
	{
		SFTC_NEXT;
		node = parse_stmtlist( F, '}' );
		if( !node ) goto fail;

		SFTC_NEXT;
		SGS_FN_END;
		return node;
	}
	/* SEPARATED STATEMENTS */
	else
	{
		node = parse_explist( F, ';' );
		if( node )
		{
			SFTC_NEXT;
			SGS_FN_END;
			return node;
		}
		else
			goto fail;
	}

fail:
	SFTC_SETERR;
	SGS_FN_END;
	return NULL;
}

SFTRET parse_stmtlist( SFTC, char end )
{
	sgs_FTNode* stmtlist = make_node( SGS_SFT_BLOCK, NULL, NULL, NULL );
	sgs_FTNode* curstmt = NULL;

	SGS_FN_BEGIN;

	for(;;)
	{
		if( SFTC_IS( end ) )
		{
			break;
		}
		else if( SFTC_IS( 0 ) )
		{
			SFTC_UNEXP;
			SFTC_SETERR;
			goto fail;
		}
		else
		{
			sgs_FTNode* stmt = parse_stmt( F );
			if( curstmt )
				curstmt->next = stmt;
			else
				stmtlist->child = stmt;
			curstmt = stmt;
		}

		if( F->C->state & SGS_MUST_STOP )
			goto fail;
	}

	SGS_FN_END;
	return stmtlist;

fail:
	SFTC_DESTROY( stmtlist );
	SFTC_SETERR;
	SGS_FN_END;
	return NULL;
}

sgs_FTNode* sgsFT_Compile( SGS_CTX, sgs_TokenList tlist )
{
	sgs_FTNode* ret;
	FTComp F;
	{
		F.C = C;
		F.at = tlist;
		F.heap = F.heapend = _make_heap( C );
	}
	ret = parse_stmtlist( &F, 0 );
	if( !ret )
	{
		sgsFT_Destroy( C, F.heap );
		return NULL;
	}
	F.heap->child = ret;
	return F.heap;
}




#include <math.h>
#include <limits.h>

#define SGS_INTERNAL_STRINGTABLES

#define StkIdx sgs_StkIdx




#define _EL_BACKUP int oapi = ( C->state & SGS_STATE_INSIDE_API ) != 0
#define _EL_SETAPI(is) C->state = ( C->state & ~SGS_STATE_INSIDE_API ) | ( is ? SGS_STATE_INSIDE_API : 0 )
#define _EL_RESET _EL_SETAPI(oapi)


#define TYPENAME( type ) sgs_VarNames[ type ]


#define STK_UNITSIZE sizeof( sgs_Variable )


typedef union intreal_t
{
	sgs_Int i;
	sgs_Real r;
}
intreal_t;


/* to work with both insertion and removal algorithms, this function has the following rules:
- return the index of the first found item with the right size or just after the right size
- if all sizes are less than specified, return the size of the object pool
*/
static int32_t objpool_binary_search( SGS_SHCTX, uint32_t appsize )
{
	int32_t pmin = 0, pmax = S->objpool_size - 1;
	while( pmin <= pmax )
	{
		int32_t pos = ( pmax + pmin ) / 2;
		uint32_t ssize = S->objpool_data[ pos ].appsize;
		if( ssize == appsize )
			return pos;
		else if( ssize < appsize )
			pmin = pos + 1;
		else if( ssize > appsize )
			pmax = pos - 1;
	}
	return pmin;
}

static void var_free_object( SGS_CTX, sgs_VarObj* O )
{
	SGS_SHCTX_USE;
	if( O->is_iface )
	{
		sgs_VHTVar* p = S->ifacetable.vars;
		sgs_VHTVar* pend = p + S->ifacetable.size;
		while( p < pend )
		{
			if( p->val.type == SGS_VT_OBJECT && p->val.data.O == O )
			{
				sgs_Variable K = p->key;
				O->refcount = 2;
				sgs_vht_unset( &S->ifacetable, C, &K );
				break;
			}
			p++;
		}
	}
	if( O->prev ) O->prev->next = O->next;
	if( O->next ) O->next->prev = O->prev;
	if( S->objs == O )
		S->objs = O->next;
#if SGS_OBJPOOL_SIZE > 0
	if( O->appsize <= SGS_OBJPOOL_MAX_APPMEM )
	{
		int32_t pos = 0;
		if( S->objpool_size )
		{
			pos = objpool_binary_search( S, O->appsize );
			if( S->objpool_size < SGS_OBJPOOL_SIZE && pos < S->objpool_size )
			{
				memmove( S->objpool_data + pos + 1, S->objpool_data + pos,
					sizeof( sgs_ObjPoolItem ) * (size_t) ( S->objpool_size - pos ) );
			}
			if( pos >= SGS_OBJPOOL_SIZE )
				pos = SGS_OBJPOOL_SIZE - 1;
			if( S->objpool_size >= SGS_OBJPOOL_SIZE )
				sgs_Dealloc( S->objpool_data[ pos ].obj );
		}
		S->objpool_data[ pos ].obj = O;
		S->objpool_data[ pos ].appsize = O->appsize;
		if( S->objpool_size < SGS_OBJPOOL_SIZE )
			S->objpool_size++;
	}
	else
		sgs_Dealloc( O );
#else
	sgs_Dealloc( O );
#endif
	S->objcount--;
}

static void var_destruct_object( SGS_CTX, sgs_VarObj* O )
{
	if( O->iface->destruct )
	{
		int ret;
		_STACK_PREPARE;
		_STACK_PROTECT;
		ret = O->iface->destruct( C, O );
		_STACK_UNPROTECT;
		if( SGS_FAILED( ret ) )
			sgs_Msg( C, SGS_ERROR, "failed to call the destructor" );
	}
	if( O->metaobj )
	{
		sgs_ObjRelease( C, O->metaobj );
		O->metaobj = NULL;
	}
}
void sgsVM_VarDestroyObject( SGS_CTX, sgs_VarObj* O )
{
	var_destruct_object( C, O );
	var_free_object( C, O );
}

static void var_destroy_string( SGS_CTX, sgs_iStr* str )
{
#if SGS_STRINGTABLE_MAXLEN >= 0
	if( str->size <= SGS_STRINGTABLE_MAXLEN )
	{
		SGS_SHCTX_USE;
		sgs_VHTVar* p;
		sgs_Variable tmp;
		tmp.type = SGS_VT_STRING;
		tmp.data.S = str;
		p = sgs_vht_get( &S->stringtable, &tmp );
		if( p )
		{
#  if SGS_DEBUG && SGS_DEBUG_EXTRA
			if( p->key.data.S != str )
			{
				printf( "WATWATWAT: (%d) |my>| [%d,%d] %d,%s |st>| [%d,%d] %d,%s\n",
					(int) sgs_HashFunc( sgs_str_cstr( str ), str->size ),
					(int) str->refcount, (int) str->hash, (int) str->size, sgs_str_cstr( str ),
					(int) p->key.data.S->refcount, (int) p->key.data.S->hash, (int) p->key.data.S->size, sgs_str_cstr( p->key.data.S ) );
			}
#  endif
			sgs_BreakIf( p->key.data.S != str );
			str->refcount = 2; /* the 'less code' way to avoid double free */
			sgs_vht_unset( &S->stringtable, C, &tmp );
		}
	}
#endif
	sgs_Dealloc( str );
}

static void var_destroy_func( SGS_CTX, sgs_iFunc* F )
{
	sgs_Variable *var = (sgs_Variable*) sgs_func_consts( F ),
		*vend = SGS_ASSUME_ALIGNED( sgs_func_bytecode( F ), sgs_Variable );
	while( var < vend )
	{
		VAR_RELEASE( var );
		var++;
	}
	sgs_Dealloc( F->lineinfo );
	if( --F->sfuncname->refcount <= 0 ) var_destroy_string( C, F->sfuncname );
	if( --F->sfilename->refcount <= 0 ) var_destroy_string( C, F->sfilename );
	sgs_Dealloc( F );
}

void sgsVM_DestroyVar( SGS_CTX, sgs_Variable* p )
{
	switch( p->type )
	{
	case SGS_VT_STRING: var_destroy_string( C, p->data.S ); break;
	case SGS_VT_FUNC: var_destroy_func( C, p->data.F ); break;
	case SGS_VT_OBJECT: if( !C->shared->gcrun ) sgsVM_VarDestroyObject( C, p->data.O ); break;
	case SGS_VT_THREAD: sgsCTX_FreeState( p->data.T ); break;
	}
}


static void var_create_0str( SGS_CTX, sgs_Variable* out, uint32_t len )
{
	out->type = SGS_VT_STRING;
	out->data.S = sgs_Alloc_a( sgs_iStr, len + 1 );
	out->data.S->refcount = 1;
	out->data.S->size = len;
	out->data.S->hash = 0;
	memset( sgs_var_cstr( out ), 0, len + 1 );
}

void sgsVM_VarCreateString( SGS_CTX, sgs_Variable* out, const char* str, sgs_SizeVal len )
{
	sgs_Hash hash;
	uint32_t ulen;
	SGS_SHCTX_USE;
	sgs_BreakIf( !str && len );
	
	ulen = (uint32_t) len; /* WP: string limit */
	
	hash = sgs_HashFunc( str, ulen );
#if SGS_STRINGTABLE_MAXLEN >= 0
	if( ulen <= SGS_STRINGTABLE_MAXLEN )
	{
		sgs_VHTVar* var = sgs_vht_get_str( &S->stringtable, str, ulen, hash );
		if( var )
		{
			*out = var->key;
			out->data.S->refcount++;
			return;
		}
	}
#endif
	
	var_create_0str( C, out, ulen );
	memcpy( sgs_str_cstr( out->data.S ), str, ulen );
	out->data.S->hash = hash;
	
	if( ulen <= SGS_STRINGTABLE_MAXLEN )
	{
		sgs_vht_set( &S->stringtable, C, out, NULL );
		out->data.S->refcount--;
	}
}

static void var_create_str( SGS_CTX, sgs_Variable* out, const char* str, sgs_SizeVal len )
{
	sgsVM_VarCreateString( C, out, str, len );
}
static void var_create_cstr( SGS_CTX, sgs_Variable* out, const char* str )
{
	sgsVM_VarCreateString( C, out, str, (sgs_SizeVal) SGS_STRINGLENGTHFUNC(str) );
}

static void var_finalize_str( SGS_CTX, sgs_Variable* out )
{
	char* str;
	sgs_Hash hash;
	uint32_t ulen;
	SGS_SHCTX_USE;
	
	str = sgs_str_cstr( out->data.S );
	ulen = out->data.S->size;
	hash = sgs_HashFunc( str, ulen );
	
#if SGS_STRINGTABLE_MAXLEN >= 0
	if( ulen <= SGS_STRINGTABLE_MAXLEN )
	{
		sgs_VHTVar* var = sgs_vht_get_str( &S->stringtable, str, ulen, hash );
		if( var )
		{
			sgs_Dealloc( out->data.S ); /* avoid querying the string table here */
			*out = var->key;
			out->data.S->refcount++;
			return;
		}
	}
#endif
	
	out->data.S->hash = hash;
	
	if( ulen <= SGS_STRINGTABLE_MAXLEN )
	{
		sgs_vht_set( &S->stringtable, C, out, NULL );
		out->data.S->refcount--;
	}
}

static void var_create_obj( SGS_CTX, sgs_Variable* out, void* data, sgs_ObjInterface* iface, uint32_t xbytes )
{
	SGS_SHCTX_USE;
	sgs_VarObj* obj = NULL;
#if SGS_OBJPOOL_SIZE > 0
	if( xbytes <= SGS_OBJPOOL_MAX_APPMEM )
	{
		int32_t pos = objpool_binary_search( S, xbytes );
		if( pos < S->objpool_size && S->objpool_data[ pos ].appsize == xbytes )
		{
			obj = S->objpool_data[ pos ].obj;
			S->objpool_size--;
			if( pos < S->objpool_size )
			{
				memmove( S->objpool_data + pos, S->objpool_data + pos + 1,
					sizeof( sgs_ObjPoolItem ) * (size_t) ( S->objpool_size - pos ) );
			}
		}
	}
#endif
	if( !obj )
		obj = sgs_Alloc_a( sgs_VarObj, xbytes );
	obj->appsize = xbytes;
	obj->data = data;
	if( xbytes )
		obj->data = ((char*)obj) + sizeof( sgs_VarObj );
	obj->iface = iface;
	obj->redblue = S->redblue;
	obj->next = S->objs;
	obj->prev = NULL;
	obj->refcount = 1;
	if( obj->next ) /* ! */
		obj->next->prev = obj;
	obj->metaobj = NULL;
	obj->mm_enable = SGS_FALSE;
	obj->in_setindex = SGS_FALSE;
	obj->is_iface = SGS_FALSE;
	S->objcount++;
	S->objs = obj;
	
	out->type = SGS_VT_OBJECT;
	out->data.O = obj;
}


/*
	Call stack
*/

int sgsVM_PushStackFrame( SGS_CTX, sgs_Variable* func )
{
	sgs_StackFrame* F;
	
	if( C->sf_count >= SGS_MAX_CALL_STACK_SIZE )
	{
		sgs_Msg( C, SGS_ERROR, SGS_ERRMSG_CALLSTACKLIMIT );
		return 0;
	}
	
	F = sgsCTX_AllocFrame( C );
	C->sf_count++;
	
	if( func->type == SGS_VT_OBJECT && func->data.O->iface == sgsstd_closure_iface )
	{
		uint8_t* cl = (uint8_t*) func->data.O->data;
		F->clsrlist = SGS_ASSUME_ALIGNED( cl + sizeof(sgs_Variable) + sizeof(sgs_clsrcount_t), sgs_Closure* );
#if SGS_DEBUG && SGS_DEBUG_VALIDATE
		F->clsrcount = (uint8_t) *SGS_ASSUME_ALIGNED( cl + sizeof(sgs_Variable), sgs_clsrcount_t );
#endif
		F->clsrref = func->data.O;
		F->clsrref->refcount++; /* acquire closure */
		p_setvar_race( func, SGS_ASSUME_ALIGNED( cl, sgs_Variable ) );
		
		if( func->type == SGS_VT_FUNC )
		{
			int i;
			sgs_iFunc* fn = func->data.F;
			for( i = fn->inclsr; i < fn->numclsr; ++i )
			{
				if( --F->clsrlist[ i ]->refcount <= 0 )
				{
					VAR_RELEASE( &F->clsrlist[ i ]->var );
					sgs_Dealloc( F->clsrlist[ i ] );
				}
				sgs_Closure* nc = sgs_Alloc( sgs_Closure );
				nc->refcount = 1;
				nc->var.type = SGS_VT_NULL;
				F->clsrlist[ i ] = nc;
			}
		}
	}
	else
	{
		F->clsrref = NULL;
#if SGS_DEBUG && SGS_DEBUG_VALIDATE
		F->clsrlist = NULL;
		F->clsrcount = 0;
#endif
	}
	
	F->func = func;
	if( func->type == SGS_VT_FUNC )
	{
		sgs_iFunc* fn = func->data.F;
		F->iptr = sgs_func_bytecode( fn );
		F->cptr = sgs_func_consts( fn );
#if SGS_DEBUG && SGS_DEBUG_VALIDATE
		F->code = F->iptr;
		F->iend = F->iptr + sgs_func_instr_count( fn );
		/* WP: const limit */
		F->constcount = (int32_t) sgs_func_const_count( fn );
#endif
	}
	else
	{
		F->iptr = NULL;
		F->cptr = NULL;
#if SGS_DEBUG && SGS_DEBUG_VALIDATE
		F->code = NULL;
		F->iend = NULL;
		F->constcount = 0;
#endif
	}
	F->next = NULL;
	F->prev = C->sf_last;
	F->errsup = 0;
	F->flags = 0;
	F->nfname = NULL;
	if( C->sf_last )
	{
		F->errsup = C->sf_last->errsup;
		C->sf_last->next = F;
	}
	else
	{
		C->sf_first = F;
	}
	C->sf_last = F;
	
	if( C->hook_fn )
		C->hook_fn( C->hook_ctx, C, SGS_HOOK_ENTER );
	
	return 1;
}

static void vm_frame_pop( SGS_CTX )
{
	sgs_StackFrame* F = C->sf_last;
	
	if( C->hook_fn )
		C->hook_fn( C->hook_ctx, C, SGS_HOOK_EXIT );
	
	if( F->clsrref )
		sgs_ObjRelease( C, F->clsrref );
	
	/* F->func is on stack */
	C->sf_count--;
	if( F->prev )
		F->prev->next = NULL;
	C->sf_last = F->prev;
	if( C->sf_first == F )
		C->sf_first = NULL;
	sgsCTX_FreeFrame( F );
}


/*
	Stack management
*/

static SGS_INLINE sgs_Variable* stk_getpos( SGS_CTX, StkIdx stkid )
{
#if SGS_DEBUG && SGS_DEBUG_VALIDATE && SGS_DEBUG_EXTRA
	if( stkid < 0 ) sgs_BreakIf( -stkid > C->stack_top - C->stack_off )
	else            sgs_BreakIf( stkid >= C->stack_top - C->stack_off )
#endif
	if( stkid < 0 )	return C->stack_top + stkid;
	else			return C->stack_off + stkid;
}

static SGS_INLINE void stk_setvar( SGS_CTX, StkIdx stkid, sgs_Variable* var )
{
	sgs_Variable* vpos = stk_getpos( C, stkid );
	VAR_RELEASE( vpos );
	*vpos = *var;
	VAR_ACQUIRE( vpos );
}
static SGS_INLINE void stk_setvar_leave( SGS_CTX, StkIdx stkid, sgs_Variable* var )
{
	sgs_Variable* vpos = stk_getpos( C, stkid );
	VAR_RELEASE( vpos );
	*vpos = *var;
}

static void stk_rebase_pointers( SGS_CTX, sgs_Variable* to, sgs_Variable* from )
{
	sgs_StackFrame* sf = C->sf_first;
	while( sf )
	{
		if( sf->func )
			sf->func = ( sf->func - from ) + to;
		sf = sf->next;
	}
}

void fstk_resize( SGS_CTX, size_t nsz )
{
	/* StkIdx item stack limit */
	sgs_Variable* prevbase;
	ptrdiff_t stkoff, stkend;
	
	stkoff = ptr_sub( C->stack_off, C->stack_base );
	stkend = ptr_sub( C->stack_top, C->stack_base );
	prevbase = C->stack_base;
	
#if DEBUG_STACK /* dangling pointer / wrong makespace size test */
	
	C->stack_base = sgs_Alloc_n( sgs_Variable, nsz );
	memcpy( C->stack_base, prevbase, (size_t) stkend );
	sgs_Dealloc( prevbase );
	
#else
	
	nsz += C->stack_mem * 2; /* cheaper alternative to MAX( nsz, C->stack_mem * 2 ); */
	C->stack_base = (sgs_Variable*) sgs_Realloc( C, C->stack_base, sizeof( sgs_Variable ) * nsz );
	
#endif
	
	stk_rebase_pointers( C, C->stack_base, prevbase );
	C->stack_mem = (uint32_t) nsz;
	C->stack_off = SGS_ASSUME_ALIGNED( ptr_add( C->stack_base, stkoff ), sgs_Variable );
	C->stack_top = SGS_ASSUME_ALIGNED( ptr_add( C->stack_base, stkend ), sgs_Variable );
}

void fstk_push( SGS_CTX, sgs_Variable* vp ){ stk_push( C, vp ); }
void fstk_push_leave( SGS_CTX, sgs_Variable* vp ){ stk_push_leave( C, vp ); }
void fstk_push2( SGS_CTX, sgs_Variable* vp1, sgs_Variable* vp2 ){ stk_push2( C, vp1, vp2 ); }
void fstk_push_null( SGS_CTX ){ stk_push_null( C ); }
void fstk_umpush( SGS_CTX, sgs_Variable* vp, sgs_SizeVal cnt )
{
	sgs_Variable* srcp = vp;
	sgs_Variable* srcend = srcp + cnt;
	while( srcp != srcend )
	{
		VAR_ACQUIRE( srcp );
		*(C)->stack_top++ = *srcp++;
	}
}
void fstk_pop1( SGS_CTX ){ stk_pop1( C ); }

static void stk_push_nulls( SGS_CTX, StkIdx cnt )
{
	sgs_Variable* tgt;
	stk_makespace( C, cnt );
	tgt = C->stack_top + cnt;
	while( C->stack_top < tgt )
		(C->stack_top++)->type = SGS_VT_NULL;
}

static sgs_Variable* stk_insert_pos( SGS_CTX, StkIdx off )
{
	sgs_Variable *op, *p;
	stk_makespace( C, 1 );
	op = C->stack_off + off, p = C->stack_top;
	while( p != op )
	{
		*p = *(p-1);
		p--;
	}
	C->stack_top++;
	return op;
}

static sgs_Variable* stk_insert_posn( SGS_CTX, StkIdx off )
{
	sgs_Variable *op, *p;
	stk_makespace( C, 1 );
	op = C->stack_top - off, p = C->stack_top;
	while( p != op )
	{
		*p = *(p-1);
		p--;
	}
	C->stack_top++;
	return op;
}

void fstk_clean( SGS_CTX, sgs_Variable* from, sgs_Variable* to )
{
	ptrdiff_t oh;
	sgs_Variable *p = from, *pend = to;
	sgs_BreakIf( C->stack_top < to );
	sgs_BreakIf( to < from );
	sgs_BreakIf( from < C->stack_base );
	
	while( p < pend )
	{
		VAR_RELEASE( p );
		p++;
	}
	
	oh = ptr_sub( C->stack_top, to );
	ptr_dec( C->stack_top, ptr_sub( to, from ) );
	
	if( oh )
		memmove( from, to, (size_t) oh );
}

void fstk_pop( SGS_CTX, sgs_StkIdx num )
{
	sgs_Variable* to = C->stack_top - num;
	sgs_BreakIf( num < 0 || num > SGS_STACKFRAMESIZE );
	stk_popto( C, to );
}

static void stk_pop2( SGS_CTX )
{
	sgs_BreakIf( C->stack_top - C->stack_off < 1 );
	
	C->stack_top -= 2;
	VAR_RELEASE( C->stack_top );
	VAR_RELEASE( C->stack_top + 1 );
}

static void stk_deltasize( SGS_CTX, int diff )
{
	if( diff < 0 )
	{
		fstk_pop( C, -diff );
	}
	else
	{
		stk_push_nulls( C, diff );
	}
}
#define stk_resize_expected( C, expect, rvc ) stk_deltasize( C, expect - rvc )
#define stk_setsize( C, size ) stk_deltasize( C, size - SGS_STACKFRAMESIZE )
#define stk_downsize( C, size ) fstk_pop( C, SGS_STACKFRAMESIZE - size )
#define stk_downsize_keep( C, size, keep ) stk_popskip( C, SGS_STACKFRAMESIZE - size - keep, keep )

static void varr_reverse( sgs_Variable* beg, sgs_Variable* end )
{
	sgs_Variable tmp;
	end--;
	while( beg < end )
	{
		tmp = *beg;
		*beg = *end;
		*end = tmp;
		beg++;
		end--;
	}
}

static void stk_transpose( SGS_CTX, StkIdx first, StkIdx all )
{
	/* assuming:
		- first >= 0 (1+)
		- all > first (implies all > 1 -- 2+)
	*/
	varr_reverse( C->stack_top - all, C->stack_top - all + first );
	varr_reverse( C->stack_top - all, C->stack_top );
}

/* transfer top of stack to *out (which must not be on the stack), skip acquire/release */
#define SGS_STACK_TOP_TO_NONSTACK( out ) *(out) = *--C->stack_top


/*
	Meta-methods
*/

#define _push_metamethod( C, obj, name ) _push_metamethod_buf_( C, (obj)->metaobj, SGS_STRLITBUF( name ) )
#define _push_metamethod_buf( C, obj, name, namelen ) _push_metamethod_buf_( C, (obj)->metaobj, name, namelen )
static int _push_metamethod_buf_( SGS_CTX, sgs_VarObj* metaobj, const char* name, size_t namelen )
{
	int ret;
	sgs_Variable fvar, kvar, ovar = sgs_MakeObjPtrNoRef( metaobj );
	if( !metaobj )
		return 0;
	sgs_InitStringBuf( C, &kvar, name, (sgs_SizeVal) namelen );
	ret = sgs_GetIndex( C, ovar, kvar, &fvar, 0 );
	VAR_RELEASE( &kvar );
	if( ret )
	{
		stk_push_leave( C, &fvar );
		return 1;
	}
	return 0;
}


/*
	Conversions
*/


static sgs_Bool var_getbool( SGS_CTX, const sgs_Variable* var )
{
	switch( var->type )
	{
	case SGS_VT_NULL: return SGS_FALSE;
	case SGS_VT_BOOL: return var->data.B;
	case SGS_VT_INT: return var->data.I != 0;
	case SGS_VT_REAL: return var->data.R != 0;
	case SGS_VT_STRING: return !!var->data.S->size;
	case SGS_VT_FUNC: return SGS_TRUE;
	case SGS_VT_CFUNC: return SGS_TRUE;
	case SGS_VT_OBJECT:
		{
			sgs_VarObj* O = var->data.O;
			_STACK_PREPARE;
			if( O->mm_enable && _push_metamethod( C, O, "__tobool" ) )
			{
				sgs_SizeVal ssz = SGS_STACKFRAMESIZE - 1;
				sgs_PushObjectPtr( C, O );
				if( sgs_XThisCall( C, 0 ) > 0 && stk_gettop( C )->type == SGS_VT_BOOL )
				{
					sgs_Bool v = !!stk_gettop( C )->data.B;
					stk_downsize( C, ssz );
					return v;
				}
				stk_downsize( C, ssz );
			}
			if( O->iface->convert )
			{
				SGSRESULT ret = SGS_EINPROC;
				_STACK_PROTECT;
				if( C->sf_count < SGS_MAX_CALL_STACK_SIZE )
				{
					C->sf_count++;
					ret = O->iface->convert( C, O, SGS_VT_BOOL );
					C->sf_count--;
				}
				else
					sgs_Msg( C, SGS_ERROR, SGS_ERRMSG_CALLSTACKLIMIT );
				
				if( SGS_SUCCEEDED( ret ) && SGS_STACKFRAMESIZE >= 1 && stk_gettop( C )->type == SGS_VT_BOOL )
				{
					sgs_Bool v = !!stk_gettop( C )->data.B;
					_STACK_UNPROTECT;
					return v;
				}
				_STACK_UNPROTECT;
			}
			return SGS_TRUE;
		}
	case SGS_VT_PTR: return var->data.P != NULL;
	case SGS_VT_THREAD: return var->data.T != NULL;
	}
	return SGS_FALSE;
}

static sgs_Int var_getint( sgs_Variable* var )
{
	switch( var->type )
	{
	case SGS_VT_BOOL: return (sgs_Int) var->data.B;
	case SGS_VT_INT: return var->data.I;
	case SGS_VT_REAL: return (sgs_Int) var->data.R;
	case SGS_VT_STRING: return sgs_util_atoi( sgs_str_cstr( var->data.S ), var->data.S->size );
	case SGS_VT_PTR: return (sgs_Int) (intptr_t) var->data.P;
	case SGS_VT_THREAD: return (sgs_Int) (intptr_t) var->data.T;
	}
	return 0;
}

static sgs_Real var_getreal( sgs_Variable* var )
{
	switch( var->type )
	{
	case SGS_VT_BOOL: return (sgs_Real) var->data.B;
	case SGS_VT_INT: return (sgs_Real) var->data.I;
	case SGS_VT_REAL: return var->data.R;
	case SGS_VT_STRING: return sgs_util_atof( sgs_str_cstr( var->data.S ), var->data.S->size );
	case SGS_VT_PTR: return (sgs_Real) (intptr_t) var->data.P;
	case SGS_VT_THREAD: return (sgs_Real) (intptr_t) var->data.T;
	}
	return 0;
}

static void* var_getptr( sgs_Variable* var )
{
	switch( var->type )
	{
	case SGS_VT_BOOL: return (void*) (size_t) var->data.B;
	case SGS_VT_INT: return (void*) (size_t) var->data.I;
	case SGS_VT_REAL: return (void*) (size_t) var->data.R;
	case SGS_VT_STRING: return (void*) (size_t) sgs_str_cstr( var->data.S );
	case SGS_VT_OBJECT: return var->data.O->data;
	case SGS_VT_PTR: return var->data.P;
	case SGS_VT_THREAD: return var->data.T;
	}
	return NULL;
}

#define var_getfin( C, var ) \
	( ((var)->type == SGS_VT_THREAD && (var)->data.T->sf_last == NULL ) \
	||( (var)->type == SGS_VT_BOOL && (var)->data.B != 0 ) \
	||( (var)->type == SGS_VT_REAL && (var)->data.R <= C->wait_timer ) \
	||( (var)->type == SGS_VT_INT && (var)->data.I <= C->wait_timer ) \
	||( (var)->type == SGS_VT_OBJECT && var_getbool( C, var ) ))

#define var_initnull( v ) \
do{ sgs_Variable* __var = (v); __var->type = SGS_VT_NULL; }while(0)
#define var_initbool( v, value ) \
do{ sgs_Variable* __var = (v); __var->type = SGS_VT_BOOL; __var->data.B = value; }while(0)
#define var_initint( v, value ) \
do{ sgs_Variable* __var = (v); __var->type = SGS_VT_INT; __var->data.I = value; }while(0)
#define var_initreal( v, value ) \
do{ sgs_Variable* __var = (v); __var->type = SGS_VT_REAL; __var->data.R = value; }while(0)

#define var_setnull( C, v ) \
do{ sgs_Variable* var = (v); VAR_RELEASE( var ); var->type = SGS_VT_NULL; }while(0)
#define var_setbool( C, v, value ) \
do{ sgs_Variable* var = (v); if( var->type != SGS_VT_BOOL ) \
	{ VAR_RELEASE( var ); var->type = SGS_VT_BOOL; } var->data.B = value; }while(0)
#define var_setint( C, v, value ) \
do{ sgs_Variable* var = (v); if( var->type != SGS_VT_INT ) \
	{ VAR_RELEASE( var ); var->type = SGS_VT_INT; } var->data.I = value; }while(0)
#define var_setreal( C, v, value ) \
do{ sgs_Variable* var = (v); if( var->type != SGS_VT_REAL ) \
	{ VAR_RELEASE( var ); var->type = SGS_VT_REAL; } var->data.R = value; }while(0)


static void init_var_string( SGS_CTX, sgs_Variable* out, sgs_Variable* var )
{
	char buf[ 32 ];
	switch( var->type )
	{
	case SGS_VT_NULL: var_create_str( C, out, SGS_STRLITBUF( "null" ) ); break;
	case SGS_VT_BOOL:
		if( var->data.B ) var_create_str( C, out, SGS_STRLITBUF( "true" ) );
		else var_create_str( C, out, SGS_STRLITBUF( "false" ) );
		break;
	case SGS_VT_INT: sprintf( buf, "%" PRId64, var->data.I ); var_create_cstr( C, out, buf ); break;
	case SGS_VT_REAL: sprintf( buf, "%g", var->data.R ); var_create_cstr( C, out, buf ); break;
	case SGS_VT_FUNC: var_create_str( C, out, SGS_STRLITBUF( "function" ) ); break;
	case SGS_VT_CFUNC: var_create_str( C, out, SGS_STRLITBUF( "C function" ) ); break;
	case SGS_VT_OBJECT:
		{
			sgs_VarObj* O = var->data.O;
			_STACK_PREPARE;
			
			if( O->mm_enable && _push_metamethod( C, O, "__tostring" ) )
			{
				sgs_SizeVal ssz = SGS_STACKFRAMESIZE - 1;
				sgs_PushObjectPtr( C, O );
				if( sgs_XThisCall( C, 0 ) > 0 && stk_gettop( C )->type == SGS_VT_STRING )
				{
					*out = *stk_gettop( C );
					out->data.S->refcount++; /* cancel release from stack to transfer successfully */
					stk_downsize( C, ssz );
					break;
				}
				stk_downsize( C, ssz );
			}
			if( O->iface->convert )
			{
				SGSRESULT ret = SGS_EINPROC;
				_STACK_PROTECT;
				if( C->sf_count < SGS_MAX_CALL_STACK_SIZE )
				{
					C->sf_count++;
					ret = O->iface->convert( C, O, SGS_VT_STRING );
					C->sf_count--;
				}
				else
					sgs_Msg( C, SGS_ERROR, SGS_ERRMSG_CALLSTACKLIMIT );
				
				if( SGS_SUCCEEDED( ret ) && SGS_STACKFRAMESIZE >= 1 && stk_gettop( C )->type == SGS_VT_STRING )
				{
					*out = *stk_gettop( C );
					out->data.S->refcount++; /* cancel release from stack to transfer successfully */
					_STACK_UNPROTECT;
					break;
				}
				_STACK_UNPROTECT;
			}
			var_create_cstr( C, out, O->iface->name );
		}
		break;
	case SGS_VT_PTR:
		sprintf( buf, "ptr(%p)", var->data.P );
		var_create_cstr( C, out, buf );
		break;
	case SGS_VT_THREAD:
		sprintf( buf, "thread(%p)", var->data.T );
		var_create_cstr( C, out, buf );
		break;
	default: var_create_str( C, out, SGS_STRLITBUF( "<bad typeid>" ) ); break;
	}
	sgs_BreakIf( out->type != SGS_VT_STRING );
}


static void vm_convert_string( SGS_CTX, sgs_Variable* var )
{
	sgs_Variable out;
	if( var->type == SGS_VT_STRING )
		return;
	init_var_string( C, &out, var );
	VAR_RELEASE( var );
	*var = out;
}

static void vm_convert_stack_string( SGS_CTX, StkIdx item )
{
	sgs_Variable out;
	if( stk_getpos( C, item )->type == SGS_VT_STRING )
		return;
	init_var_string( C, &out, stk_getpos( C, item ) );
	stk_setvar_leave( C, item, &out );
}


/*
	VM mutation
*/

#define obj_gcmark( S, O ) { if( (O)->redblue != (S)->redblue ) obj_gcmark_do( S, O ); }
#define vm_gcmark( S, var ) { if( var->type == SGS_VT_OBJECT ) obj_gcmark( S, var->data.O ); }
static void obj_gcmark_do( SGS_SHCTX, sgs_VarObj* O )
{
	O->redblue = S->redblue;
	if( O->iface->gcmark )
	{
		SGS_CTX = S->ctx_root;
		_STACK_PREPARE;
		_STACK_PROTECT;
		O->iface->gcmark( C, O );
		_STACK_UNPROTECT;
	}
	if( O->metaobj )
	{
		obj_gcmark( S, O->metaobj );
	}
}

/*
	Object property / array accessor handling
*/

int sgs_specfn_apply( SGS_CTX )
{
	sgs_SizeVal i, asize;
	sgs_Variable v_func, v_this, v_args;
	SGSFN( "apply" );
	sgs_Method( C );
	/* method, test arg0 = proc, get arg0, get arg1, get array size @ arg2, get arg2, req. count = 3 */
	if( !sgs_LoadArgs( C, "@?p<vva<v.", &v_func, &v_this, &asize, &v_args ) )
		return 0;
	
	fstk_push2( C, &v_func, &v_this );
	for( i = 0; i < asize; ++i )
		sgs_PushNumIndex( C, v_args, i );
	return sgs_XThisCall( C, asize );
}


static int vm_getidx_builtin( SGS_CTX, sgs_Variable* outmaybe, sgs_Variable* obj, sgs_Variable* idx )
{
	int res;
	sgs_Int pos, size;
	if( obj->type == SGS_VT_STRING )
	{
		size = obj->data.S->size;
		fstk_push( C, idx );
		res = sgs_ParseInt( C, -1, &pos );
		fstk_pop1( C );
		if( !res )
		{
			sgs_Msg( C, SGS_WARNING, "Expected integer as string index" );
			return SGS_EINVAL;
		}
		if( pos >= size || pos < -size )
		{
			sgs_Msg( C, SGS_WARNING, "String index out of bounds" );
			return SGS_EBOUNDS;
		}
		pos = ( pos + size ) % size;
		sgsVM_VarCreateString( C, outmaybe, sgs_var_cstr( obj ) + pos, 1 );
		return 0;
	}

	sgs_Msg( C, SGS_WARNING, "Cannot index variable of type '%s'", TYPENAME( obj->type ) );
	return SGS_ENOTFND;
}

int sgsstd_end_on( SGS_CTX );
int sgsstd_co_resume( SGS_CTX );
int sgsstd_abort( SGS_CTX );
static int vm_getprop_builtin( SGS_CTX, sgs_Variable* outmaybe, sgs_Variable* obj, sgs_Variable* idx )
{
	if( idx->type == SGS_VT_STRING )
	{
		const char* prop = sgs_var_cstr( idx );
		
		switch( obj->type )
		{
		case SGS_VT_STRING:
			if( !strcmp( prop, "length" ) )
			{
				*outmaybe = sgs_MakeInt( obj->data.S->size );
				return 0;
			}
			break;
		case SGS_VT_FUNC:
		case SGS_VT_CFUNC:
			if( !strcmp( prop, "apply" ) )
			{
				*outmaybe = sgs_MakeCFunc( sgs_specfn_apply );
				return 0;
			}
			break;
		case SGS_VT_THREAD:
			if( !strcmp( prop, "was_aborted" ) )
			{
				int wa = ( obj->data.T->state & SGS_STATE_LASTFUNCABORT ) != 0 ||
					( obj->data.T->sf_last && ( obj->data.T->sf_last->flags & SGS_SF_ABORTED ) != 0 );
				*outmaybe = sgs_MakeBool( wa );
				return 0;
			}
			if( !strcmp( prop, "not_started" ) )
			{
				*outmaybe = sgs_MakeBool( obj->data.T->state & SGS_STATE_COROSTART );
				return 0;
			}
			if( !strcmp( prop, "running" ) )
			{
				*outmaybe = sgs_MakeBool( obj->data.T->sf_count );
				return 0;
			}
			if( !strcmp( prop, "can_resume" ) )
			{
				*outmaybe = sgs_MakeBool(
					( obj->data.T->state & SGS_STATE_COROSTART ) ||
					obj->data.T->sf_count
				);
				return 0;
			}
			if( !strcmp( prop, "end_on" ) )
			{
				*outmaybe = sgs_MakeCFunc( sgsstd_end_on );
				return 0;
			}
			if( !strcmp( prop, "resume" ) )
			{
				*outmaybe = sgs_MakeCFunc( sgsstd_co_resume );
				return 0;
			}
			if( !strcmp( prop, "abort" ) )
			{
				*outmaybe = sgs_MakeCFunc( sgsstd_abort );
				return 0;
			}
			break;
		}
		
		sgs_Msg( C, SGS_WARNING, "Property '%s' not found on "
			"object of type '%s'", prop, TYPENAME( obj->type ) );
	}
	
	sgs_Msg( C, SGS_WARNING, "Property of type '%s' not found on "
		"object of type '%s'", TYPENAME( idx->type ), TYPENAME( obj->type ) );
	return SGS_ENOTFND;
}



/* PREDECL */
static SGSRESULT vm_getprop( SGS_CTX, sgs_Variable* outmaybe, sgs_Variable* obj, sgs_Variable* idx, int isprop );

static SGSRESULT vm_runerr_getprop( SGS_CTX, SGSRESULT type, StkIdx origsize,
	sgs_Variable* outmaybe, sgs_Variable* obj, sgs_Variable* idx, int isprop )
{
	if( type == SGS_ENOTFND )
	{
		char* p;
		const char* err;
		sgs_VarObj* O = obj->data.O;
		sgs_Variable cidx = *idx;
		
		if( obj->type == SGS_VT_OBJECT && obj->data.O->metaobj )
		{
			sgs_Variable tmp;
			if( obj->data.O->mm_enable && _push_metamethod( C, O, "__getindex" ) )
			{
				sgs_SizeVal ssz = SGS_STACKFRAMESIZE - 1;
				sgs_PushObjectPtr( C, O );
				fstk_push( C, idx );
				if( sgs_XThisCall( C, 1 ) > 0 )
				{
					stk_downsize_keep( C, ssz, 1 );
					return 1;
				}
				stk_downsize( C, ssz );
			}
			
			tmp.type = SGS_VT_OBJECT;
			tmp.data.O = obj->data.O->metaobj;
			return vm_getprop( C, outmaybe, &tmp, &cidx, isprop );
		}
		
		err = isprop ? "Readable property not found" : "Cannot find readable value by index";
		fstk_push( C, &cidx );
		p = sgs_ToString( C, -1 );
		sgs_Msg( C, SGS_WARNING, "%s: \"%s\"", err, p );
	}
	else if( type == SGS_EBOUNDS )
	{
		sgs_Msg( C, SGS_WARNING, "Index out of bounds" );
	}
	else if( type == SGS_EINVAL )
	{
		sgs_Msg( C, SGS_WARNING, "Invalid value type used for %s read",
			isprop ? "property" : "index" );
	}
	else if( type == SGS_EINPROC )
	{
		sgs_Cntl( C, SGS_CNTL_ERRSUP, 0 ); /* fatal error */
		sgs_Msg( C, SGS_ERROR, "%s read process interrupted, possibly by infinite recursion",
			isprop ? "Property" : "Index" );
	}
	else if( type == SGS_ENOTSUP )
	{
		sgs_Msg( C, SGS_WARNING, "%s read operation not supported on the given variable type",
			isprop ? "Property" : "Index" );
	}
	else
	{
		sgs_Msg( C, SGS_WARNING, "Unknown error on %s read", isprop ? "property" : "index" );
	}
	
	fstk_pop( C, SGS_STACKFRAMESIZE - origsize );
	return type;
}
#define VM_GETPROP_ERR( type ) vm_runerr_getprop( C, type, origsize, outmaybe, obj, idx, isprop )

static SGSRESULT vm_runerr_setprop( SGS_CTX, SGSRESULT type, StkIdx origsize, sgs_Variable* idx, int isprop )
{
	if( type == SGS_ENOTFND )
	{
		char* p;
		const char* err = isprop ? "Writable property not found" : "Cannot find writable value by index";
		fstk_push( C, idx );
		p = sgs_ToString( C, -1 );
		sgs_Msg( C, SGS_WARNING, "%s: \"%s\"", err, p );
	}
	else if( type == SGS_EBOUNDS )
	{
		sgs_Msg( C, SGS_WARNING, "Index out of bounds" );
	}
	else if( type == SGS_EINVAL )
	{
		sgs_Msg( C, SGS_WARNING, "Invalid value type used for %s write",
			isprop ? "property" : "index" );
	}
	else if( type == SGS_EINPROC )
	{
		sgs_Cntl( C, SGS_CNTL_ERRSUP, 0 ); /* fatal error */
		sgs_Msg( C, SGS_ERROR, "%s write process interrupted, possibly by infinite recursion",
			isprop ? "Property" : "Index" );
	}
	else if( type == SGS_ENOTSUP )
	{
		sgs_Msg( C, SGS_WARNING, "%s write operation not supported on the given variable type",
			isprop ? "Property" : "Index" );
	}
	else
	{
		sgs_Msg( C, SGS_WARNING, "Unknown error on %s write", isprop ? "property" : "index" );
	}
	
	fstk_pop( C, SGS_STACKFRAMESIZE - origsize );
	return type;
}
#define VM_SETPROP_ERR( type ) vm_runerr_setprop( C, type, origsize, idx, isprop )


/* VM_GETPROP
- two output states:
-- 1 is returned, there is a value in the stack
-- 0 is returned, valmaybe receives data
- if error is returned, no value is available anywhere */
#define VM_GETPROP_RETTOP( ret, ptr ) \
	do{ if( !ret ){ stk_push_leave( C, (ptr) ); } }while(0)
#define VM_GETPROP_RETPTR( ret, ptr ) \
	do{ if( ret ){ *(ptr) = *--C->stack_top; } }while(0)

static SGSRESULT vm_getprop( SGS_CTX, sgs_Variable* outmaybe, sgs_Variable* obj, sgs_Variable* idx, int isprop )
{
	int ret = SGS_ENOTSUP, isobj = obj->type == SGS_VT_OBJECT;
	StkIdx origsize = SGS_STACKFRAMESIZE;
	
	if( isobj && obj->data.O->iface == sgsstd_dict_iface )
	{
		sgs_VHTable* ht = (sgs_VHTable*) obj->data.O->data;
		if( idx->type == SGS_VT_STRING )
		{
			sgs_VHTVar* var = sgs_vht_get( ht, idx );
			if( !var )
				return VM_GETPROP_ERR( SGS_ENOTFND );
			else
			{
				p_initvar( outmaybe, &var->val );
				return 0;
			}
		}
		else
		{
			fstk_push( C, idx );
			sgs_ToString( C, -1 );
			{
				sgs_VHTVar* var = sgs_vht_get( ht, stk_gettop( C ) );
				if( !var )
					return VM_GETPROP_ERR( SGS_ENOTFND );
				else
				{
					p_initvar( outmaybe, &var->val );
					fstk_pop1( C );
					return 0;
				}
			}
		}
	}
	else if( isobj && obj->data.O->iface == sgsstd_map_iface )
	{
		sgs_VHTVar* var;
		sgs_VHTable* ht = (sgs_VHTable*) obj->data.O->data;
		/* sgs_vht_get does not modify search key */
		var = sgs_vht_get( ht, idx );
		
		if( !var )
			return VM_GETPROP_ERR( SGS_ENOTFND );
		else
		{
			p_initvar( outmaybe, &var->val );
			return 0;
		}
	}
	else if( isobj && obj->data.O->iface == sgsstd_array_iface && !isprop )
	{
		sgsstd_array_header_t* hdr = (sgsstd_array_header_t*) obj->data.O->data;
		sgs_Variable* ptr = hdr->data;
		sgs_Int i = var_getint( idx );
		if( i < 0 || i >= hdr->size )
		{
			sgs_Msg( C, SGS_WARNING, "array index out of bounds" );
			return VM_GETPROP_ERR( SGS_EBOUNDS );
		}
		p_initvar( outmaybe, &ptr[ i ] );
		return 0;
	}
	else if( isobj && obj->data.O->iface->getindex )
	{
		sgs_VarObj* O = obj->data.O;
		_STACK_PREPARE;
		_EL_BACKUP;
		int arg = C->object_arg;
		
		if( C->sf_count >= SGS_MAX_CALL_STACK_SIZE )
			return SGS_EINPROC;
		C->sf_count++;
		
		_STACK_PROTECT;
		_EL_SETAPI(0);
		fstk_push( C, idx );
		C->object_arg = isprop;
		ret = O->iface->getindex( C, O );
		C->object_arg = arg;
		_EL_RESET;
		
		C->sf_count--;
		if( SGS_SUCCEEDED( ret ) && SGS_STACKFRAMESIZE >= 1 )
		{
			_STACK_UNPROTECT_SKIP( 1 );
			ret = 1;
		}
		else
		{
			_STACK_UNPROTECT;
			ret = SGS_ENOTFND;
		}
	}
	else
	{
		return isprop ?
			vm_getprop_builtin( C, outmaybe, obj, idx ) :
			vm_getidx_builtin( C, outmaybe, obj, idx );
	}
	
	if( SGS_FAILED( ret ) )
		return VM_GETPROP_ERR( ret );
	return ret;
}

static void vm_getprop_safe( SGS_CTX, sgs_Variable* out, sgs_Variable* obj, sgs_Variable* idx, int isprop )
{
	SGSRESULT res = vm_getprop( C, out, obj, idx, isprop );
	if( SGS_FAILED( res ) )
		return;
	VM_GETPROP_RETPTR( res, out );
}

static SGSRESULT vm_setprop( SGS_CTX, sgs_Variable* obj, sgs_Variable* idx, sgs_Variable* src, int isprop )
{
	int ret = 0;
	StkIdx origsize = SGS_STACKFRAMESIZE;
	
	if( isprop && idx->type != SGS_VT_INT && idx->type != SGS_VT_STRING )
	{
		ret = SGS_EINVAL;
	}
	else if( obj->type == SGS_VT_OBJECT && obj->data.O->metaobj &&
		obj->data.O->mm_enable && !obj->data.O->in_setindex &&
		_push_metamethod( C, obj->data.O, "__setindex" ) )
	{
		sgs_SizeVal ssz = SGS_STACKFRAMESIZE - 1;
		sgs_PushObjectPtr( C, obj->data.O );
		fstk_push2( C, idx, src );
		obj->data.O->in_setindex = SGS_TRUE; /* prevent recursion */
		sgs_ThisCall( C, 2, 1 );
		obj->data.O->in_setindex = SGS_FALSE;
		ret = sgs_GetBool( C, -1 );
		stk_downsize( C, ssz );
		if( !ret ) /* was not handled by metamethod */
			goto nextcase;
	}
	else if( obj->type == SGS_VT_OBJECT && obj->data.O->iface->setindex )
	{
nextcase:;
		int arg = C->object_arg;
		sgs_VarObj* O = obj->data.O;
		_STACK_PREPARE;
		_EL_BACKUP;
		
		if( C->sf_count >= SGS_MAX_CALL_STACK_SIZE )
			return SGS_EINPROC;
		C->sf_count++;
		
		_EL_SETAPI(0);
		_STACK_PROTECT;
		fstk_push2( C, idx, src );
		C->object_arg = isprop;
		ret = O->iface->setindex( C, O );
		C->object_arg = arg;
		_EL_RESET;
		
		C->sf_count--;
		_STACK_UNPROTECT;
	}
	else
		ret = SGS_ENOTSUP;
	
	if( SGS_FAILED( ret ) )
		return VM_SETPROP_ERR( ret );
	
	fstk_pop( C, SGS_STACKFRAMESIZE - origsize );
	return ret;
}


/*
	OPs
*/

static SGSBOOL vm_clone( SGS_CTX, sgs_Variable* var )
{
	/*
		strings are supposed to be immutable
		(even though C functions can accidentally
		or otherwise modify them with relative ease)
	*/
	if( var->type == SGS_VT_OBJECT )
	{
		int ret = SGS_ENOTFND;
		sgs_VarObj* O = var->data.O;
		if( O->mm_enable && _push_metamethod( C, O, "__clone" ) )
		{
			sgs_SizeVal ssz = SGS_STACKFRAMESIZE - 1;
			sgs_PushObjectPtr( C, O );
			if( sgs_XThisCall( C, 0 ) > 0 )
			{
				stk_downsize_keep( C, ssz, 1 );
				return SGS_TRUE;
			}
			stk_downsize( C, ssz );
		}
		if( O->iface->convert )
		{
			_STACK_PREPARE;
			_STACK_PROTECT;
			ret = O->iface->convert( C, O, SGS_CONVOP_CLONE );
			_STACK_UNPROTECT_SKIP( SGS_FAILED( ret ) ? 0 : 1 );
		}
		if( SGS_FAILED( ret ) )
		{
			sgs_Msg( C, SGS_ERROR, "failed to clone variable" );
			return SGS_FALSE;
		}
	}
	else
	{
		/* even though functions are immutable, they're also impossible to modify,
			thus there is little need for showing an error when trying to convert one,
			especially if it's a part of some object to be cloned */
		fstk_push( C, var );
	}
	return SGS_TRUE;
}

static SGSBOOL vm_op_concat_ex( SGS_CTX, StkIdx args )
{
	StkIdx i;
	uint32_t totsz = 0, curoff = 0;
	sgs_Variable N;
	if( args < 2 )
	{
		if( args < 1 )
			sgs_PushString( C, "" );
		return 1;
	}
	if( SGS_STACKFRAMESIZE < args )
		return 0;
	for( i = 1; i <= args; ++i )
	{
		vm_convert_stack_string( C, -i );
		totsz += stk_ptop( C, -i )->data.S->size;
	}
	var_create_0str( C, &N, totsz );
	for( i = args; i >= 1; --i )
	{
		sgs_Variable* var = stk_ptop( C, -i );
		memcpy( sgs_var_cstr( &N ) + curoff, sgs_var_cstr( var ), var->data.S->size );
		curoff += var->data.S->size;
	}
	var_finalize_str( C, &N );
	stk_setvar_leave( C, -args, &N );
	fstk_pop( C, args - 1 );
	return 1;
}

static void vm_op_concat( SGS_CTX, StkIdx out, sgs_Variable* A, sgs_Variable* B )
{
	sgs_Variable lA = *A, lB = *B;
	int ssz = SGS_STACKFRAMESIZE;
	fstk_push2( C, &lA, &lB );
	vm_op_concat_ex( C, 2 );
	stk_setlvar( C, out, stk_gettop( C ) );
	fstk_pop( C, SGS_STACKFRAMESIZE - ssz );
}

static SGSBOOL vm_op_negate( SGS_CTX, sgs_Variable* out, sgs_Variable* A )
{
	sgs_Variable lA = *A;
	VAR_ACQUIRE( &lA );
	VAR_RELEASE( out );
	switch( lA.type )
	{
	case SGS_VT_NULL: /* guaranteed to be NULL after release */ break;
	case SGS_VT_BOOL: var_initint( out, -lA.data.B ); break;
	case SGS_VT_INT: var_initint( out, -lA.data.I ); break;
	case SGS_VT_REAL: var_initreal( out, -lA.data.R ); break;
	case SGS_VT_OBJECT:
		{
			int ret = SGS_ENOTFND;
			sgs_VarObj* O = lA.data.O;
			/* WP: stack limit */
			if( O->mm_enable && _push_metamethod( C, O, "__negate" ) )
			{
				sgs_SizeVal ssz = SGS_STACKFRAMESIZE - 1;
				sgs_PushObjectPtr( C, O );
				if( sgs_XThisCall( C, 0 ) > 0 )
				{
					SGS_STACK_TOP_TO_NONSTACK( out );
					stk_downsize( C, ssz );
					goto done;
				}
				stk_downsize( C, ssz );
			}
			if( O->iface->expr )
			{
				int arg = C->object_arg;
				_STACK_PREPARE;
				_STACK_PROTECT;
				fstk_push( C, A );
				C->object_arg = SGS_EOP_NEGATE;
				ret = O->iface->expr( C, O );
				C->object_arg = arg;
				if( SGS_SUCCEEDED( ret ) && SGS_STACKFRAMESIZE >= 1 )
				{
					SGS_STACK_TOP_TO_NONSTACK( out );
				}
				_STACK_UNPROTECT;
			}
			if( SGS_FAILED( ret ) )
			{
				sgs_Msg( C, SGS_ERROR, "Given object does not support negation." );
				/* guaranteed to be NULL after release */
			}
			VAR_RELEASE( &lA );
			return SGS_SUCCEEDED( ret );
		}
	default:
		sgs_Msg( C, SGS_WARNING, "Negating variable of type %s is not supported.", TYPENAME( lA.type ) );
		VAR_RELEASE( &lA );
		return SGS_FALSE;
	}
	
done:
	VAR_RELEASE( &lA );
	return SGS_TRUE;
}

static void vm_op_boolinv( SGS_CTX, int16_t out, sgs_Variable *A )
{
	int val = !var_getbool( C, A );
	var_setbool( C, C->stack_off + out, val );
}

static void vm_op_invert( SGS_CTX, int16_t out, sgs_Variable *A )
{
	sgs_Int val = ~var_getint( A );
	var_setint( C, C->stack_off + out, val );
}

static void vm_op_incdec( SGS_CTX, sgs_Variable* out, sgs_Variable *A, int diff )
{
	switch( A->type )
	{
	case SGS_VT_INT: var_setint( C, out, A->data.I + diff ); break;
	case SGS_VT_REAL: var_setreal( C, out, A->data.R + diff ); break;
	default:
		var_setnull( C, out );
		sgs_Msg( C, SGS_ERROR, "Cannot %screment non-numeric variables!", diff > 0 ? "in" : "de" );
		break;
	}
}


#define ARITH_OP_ADD	SGS_EOP_ADD
#define ARITH_OP_SUB	SGS_EOP_SUB
#define ARITH_OP_MUL	SGS_EOP_MUL
#define ARITH_OP_DIV	SGS_EOP_DIV
#define ARITH_OP_MOD	SGS_EOP_MOD
static const char* mm_arith_ops[] =
{
	"__add",
	"__sub",
	"__mul",
	"__div",
	"__mod",
};
static SGSBOOL vm_arith_op( SGS_CTX, sgs_Variable* out, sgs_Variable* a, sgs_Variable* b, int op )
{
	if( a->type == SGS_VT_REAL && b->type == SGS_VT_REAL )
	{
		sgs_Real A = a->data.R, B = b->data.R, R;
		switch( op ){
			case ARITH_OP_ADD: R = A + B; break;
			case ARITH_OP_SUB: R = A - B; break;
			case ARITH_OP_MUL: R = A * B; break;
			case ARITH_OP_DIV: if( B == 0 ) goto div0err; R = A / B; break;
			case ARITH_OP_MOD: if( B == 0 ) goto div0err; R = fmod( A, B ); break;
			default: R = 0; break;
		}
		var_setreal( C, out, R );
		return SGS_TRUE;
	}
	if( a->type == SGS_VT_INT && b->type == SGS_VT_INT )
	{
		sgs_Int A = a->data.I, B = b->data.I, R;
		switch( op ){
			case ARITH_OP_ADD: R = A + B; break;
			case ARITH_OP_SUB: R = A - B; break;
			case ARITH_OP_MUL: R = A * B; break;
			case ARITH_OP_DIV: if( B == 0 ) goto div0err;
				var_setreal( C, out, ((sgs_Real) A) / ((sgs_Real) B) ); return SGS_TRUE; break;
			case ARITH_OP_MOD: if( B == 0 ) goto div0err; R = A % B; break;
			default: R = 0; break;
		}
		var_setint( C, out, R );
		return SGS_TRUE;
	}
	
	if( a->type == SGS_VT_OBJECT || b->type == SGS_VT_OBJECT )
	{
		int ret;
		sgs_Variable lA = *a, lB = *b;
		VAR_ACQUIRE( &lA );
		VAR_ACQUIRE( &lB );
		/* WP: stack limit */
		
		if( a->type == SGS_VT_OBJECT && a->data.O->mm_enable &&
			_push_metamethod_buf( C, a->data.O, mm_arith_ops[ op ], 5 ) )
		{
			sgs_SizeVal ssz = SGS_STACKFRAMESIZE - 1;
			stk_makespace( C, 3 );
			*C->stack_top++ = *a;
			*C->stack_top++ = *a;
			*C->stack_top++ = *b;
			(*a->data.pRC) += 2;
			(*b->data.pRC) += 1;
			if( sgs_XThisCall( C, 2 ) > 0 )
			{
				SGS_STACK_TOP_TO_NONSTACK( out );
				stk_downsize( C, ssz );
				VAR_RELEASE( &lA );
				VAR_RELEASE( &lB );
				return 1;
			}
			stk_downsize( C, ssz );
		}
		
		if( b->type == SGS_VT_OBJECT && b->data.O->mm_enable &&
			_push_metamethod_buf( C, b->data.O, mm_arith_ops[ op ], 5 ) )
		{
			sgs_SizeVal ssz = SGS_STACKFRAMESIZE - 1;
			stk_makespace( C, 3 );
			*C->stack_top++ = *b;
			*C->stack_top++ = *a;
			*C->stack_top++ = *b;
			(*a->data.pRC) += 1;
			(*b->data.pRC) += 2;
			if( sgs_XThisCall( C, 2 ) > 0 )
			{
				SGS_STACK_TOP_TO_NONSTACK( out );
				stk_downsize( C, ssz );
				VAR_RELEASE( &lA );
				VAR_RELEASE( &lB );
				return 1;
			}
			stk_downsize( C, ssz );
		}
		
		if( a->type == SGS_VT_OBJECT && a->data.O->iface->expr )
		{
			int arg = C->object_arg;
			sgs_VarObj* O = a->data.O;
			_STACK_PREPARE;
			_STACK_PROTECT;
			fstk_push2( C, a, b );
			C->object_arg = op;
			ret = O->iface->expr( C, O );
			C->object_arg = arg;
			ret = SGS_SUCCEEDED( ret ) && SGS_STACKFRAMESIZE >= 1;
			if( ret )
			{
				SGS_STACK_TOP_TO_NONSTACK( out );
				_STACK_UNPROTECT;
				VAR_RELEASE( &lA );
				VAR_RELEASE( &lB );
				return SGS_TRUE;
			}
			else
			{
				_STACK_UNPROTECT;
			}
		}
		
		if( b->type == SGS_VT_OBJECT && b->data.O->iface->expr )
		{
			int arg = C->object_arg;
			sgs_VarObj* O = b->data.O;
			_STACK_PREPARE;
			_STACK_PROTECT;
			fstk_push2( C, a, b );
			C->object_arg = op;
			ret = O->iface->expr( C, O );
			C->object_arg = arg;
			ret = SGS_SUCCEEDED( ret ) && SGS_STACKFRAMESIZE >= 1;
			if( ret )
			{
				SGS_STACK_TOP_TO_NONSTACK( out );
				_STACK_UNPROTECT;
				VAR_RELEASE( &lA );
				VAR_RELEASE( &lB );
				return SGS_TRUE;
			}
			else
			{
				_STACK_UNPROTECT;
			}
		}
		
		VAR_RELEASE( &lA );
		VAR_RELEASE( &lB );
		goto fail;
	}
	
	/* if either variable is of a basic callable type */
	if( a->type == SGS_VT_FUNC || a->type == SGS_VT_CFUNC ||
		b->type == SGS_VT_FUNC || b->type == SGS_VT_CFUNC ||
		a->type == SGS_VT_THREAD || b->type == SGS_VT_THREAD ||
		a->type == SGS_VT_PTR  || b->type == SGS_VT_PTR   )
		goto fail;
	
	/* if either are REAL or STRING */
	if( a->type == SGS_VT_REAL || b->type == SGS_VT_STRING ||
		a->type == SGS_VT_STRING || b->type == SGS_VT_REAL )
	{
		sgs_Real A = var_getreal( a ), B = var_getreal( b ), R;
		switch( op ){
			case ARITH_OP_ADD: R = A + B; break;
			case ARITH_OP_SUB: R = A - B; break;
			case ARITH_OP_MUL: R = A * B; break;
			case ARITH_OP_DIV: if( B == 0 ) goto div0err; R = A / B; break;
			case ARITH_OP_MOD: if( B == 0 ) goto div0err; R = fmod( A, B ); break;
			default: R = 0; break;
		}
		var_setreal( C, out, R );
	}
	else
	{
		sgs_Int A = var_getint( a ), B = var_getint( b ), R;
		switch( op ){
			case ARITH_OP_ADD: R = A + B; break;
			case ARITH_OP_SUB: R = A - B; break;
			case ARITH_OP_MUL: R = A * B; break;
			case ARITH_OP_DIV: if( B == 0 ) goto div0err;
				var_setreal( C, out, ((sgs_Real) A) / ((sgs_Real) B) ); return SGS_TRUE;
			case ARITH_OP_MOD: if( B == 0 ) goto div0err; R = A % B; break;
			default: R = 0; break;
		}
		var_setint( C, out, R );
	}
	return SGS_TRUE;
	
div0err:
	VAR_RELEASE( out );
	sgs_Msg( C, SGS_ERROR, "Division by 0" );
	return SGS_FALSE;
fail:
	VAR_RELEASE( out );
	sgs_Msg( C, SGS_ERROR, "Specified arithmetic operation is not supported on the given set of arguments" );
	return SGS_FALSE;
}


#define VAR_IOP( pfx, op ) \
static void vm_op_##pfx( SGS_CTX, int16_t out, sgs_Variable* a, sgs_Variable* b ) { \
	sgs_Int A = var_getint( a ); \
	sgs_Int B = var_getint( b ); \
	var_setint( C, C->stack_off + out, A op B ); \
	}

VAR_IOP( and, & )
VAR_IOP( or, | )
VAR_IOP( xor, ^ )
VAR_IOP( lsh, << )
VAR_IOP( rsh, >> )


/* returns 0 if equal, >0 if A is bigger, <0 if B is bigger */
#define _SGS_SIGNDIFF( a, b ) ((a)==(b)?0:((a)<(b)?-1:1))
static int vm_compare( SGS_CTX, sgs_Variable* a, sgs_Variable* b )
{
	const uint32_t ta = a->type, tb = b->type;
	
	/* both are INT */
	if( ta == SGS_VT_INT && tb == SGS_VT_INT ) return _SGS_SIGNDIFF( a->data.I, b->data.I );
	/* both are REAL */
	if( ta == SGS_VT_REAL && tb == SGS_VT_REAL ) return _SGS_SIGNDIFF( a->data.R, b->data.R );
	
	/* either is OBJECT */
	if( ta == SGS_VT_OBJECT || tb == SGS_VT_OBJECT )
	{
		int ret = SGS_ENOTSUP, suc;
		sgs_Real out = _SGS_SIGNDIFF( ta, tb );
		sgs_Variable lA = *a, lB = *b;
		VAR_ACQUIRE( &lA );
		VAR_ACQUIRE( &lB );
		
		if( ta == SGS_VT_OBJECT && a->data.O->mm_enable &&
			_push_metamethod( C, a->data.O, "__compare" ) )
		{
			sgs_SizeVal ssz = SGS_STACKFRAMESIZE - 1;
			stk_makespace( C, 3 );
			*C->stack_top++ = *a;
			*C->stack_top++ = *a;
			*C->stack_top++ = *b;
			(*a->data.pRC) += 2;
			(*b->data.pRC) += 1;
			if( sgs_XThisCall( C, 2 ) > 0 )
			{
				out = var_getreal( stk_gettop( C ) );
				stk_downsize( C, ssz );
				VAR_RELEASE( &lA );
				VAR_RELEASE( &lB );
				return _SGS_SIGNDIFF( out, 0 );
			}
			stk_downsize( C, ssz );
		}
		
		if( tb == SGS_VT_OBJECT && b->data.O->mm_enable &&
			_push_metamethod( C, b->data.O, "__compare" ) )
		{
			sgs_SizeVal ssz = SGS_STACKFRAMESIZE - 1;
			stk_makespace( C, 3 );
			*C->stack_top++ = *b;
			*C->stack_top++ = *a;
			*C->stack_top++ = *b;
			(*a->data.pRC) += 1;
			(*b->data.pRC) += 2;
			if( sgs_XThisCall( C, 2 ) > 0 )
			{
				out = var_getreal( stk_gettop( C ) );
				stk_downsize( C, ssz );
				VAR_RELEASE( &lA );
				VAR_RELEASE( &lB );
				return _SGS_SIGNDIFF( out, 0 );
			}
			stk_downsize( C, ssz );
		}
		
		if( ta == SGS_VT_OBJECT && a->data.O->iface->expr )
		{
			int arg = C->object_arg;
			sgs_VarObj* O = a->data.O;
			_STACK_PREPARE;
			_STACK_PROTECT;
			fstk_push2( C, a, b );
			C->object_arg = SGS_EOP_COMPARE;
			ret = O->iface->expr( C, O );
			C->object_arg = arg;
			suc = SGS_SUCCEEDED( ret ) && SGS_STACKFRAMESIZE >= 1;
			if( suc )
				out = var_getreal( stk_gettop( C ) );
			_STACK_UNPROTECT;
			if( suc )
			{
				VAR_RELEASE( &lA );
				VAR_RELEASE( &lB );
				return _SGS_SIGNDIFF( out, 0 );
			}
		}
		
		if( tb == SGS_VT_OBJECT && b->data.O->iface->expr )
		{
			int arg = C->object_arg;
			sgs_VarObj* O = b->data.O;
			_STACK_PREPARE;
			_STACK_PROTECT;
			fstk_push2( C, a, b );
			C->object_arg = SGS_EOP_COMPARE;
			ret = O->iface->expr( C, O );
			C->object_arg = arg;
			suc = SGS_SUCCEEDED( ret ) && SGS_STACKFRAMESIZE >= 1;
			if( suc )
				out = var_getreal( stk_gettop( C ) );
			_STACK_UNPROTECT;
			if( suc )
			{
				VAR_RELEASE( &lA );
				VAR_RELEASE( &lB );
				return _SGS_SIGNDIFF( out, 0 );
			}
		}
		
		VAR_RELEASE( &lA );
		VAR_RELEASE( &lB );
		/* fallback: check for equality */
		if( ta == tb )
			return _SGS_SIGNDIFF( a->data.O, b->data.O );
		else
			return _SGS_SIGNDIFF( ta, tb );
	}
	
	/* both are FUNC/CFUNC */
	if( ( ta == SGS_VT_FUNC || ta == SGS_VT_CFUNC ) &&
		( tb == SGS_VT_FUNC || tb == SGS_VT_CFUNC ) )
	{
		if( ta != tb )
			return _SGS_SIGNDIFF( ta, tb );
		if( ta == SGS_VT_FUNC )
			return _SGS_SIGNDIFF( a->data.F, b->data.F );
		else
			return _SGS_SIGNDIFF( (void*)a->data.C, (void*)b->data.C );
	}
	
	/* either is STRING */
	if( ta == SGS_VT_STRING || tb == SGS_VT_STRING )
	{
		/* other is NULL */
		if( ta == SGS_VT_NULL || tb == SGS_VT_NULL )
			return _SGS_SIGNDIFF( ta, tb );
		
		ptrdiff_t out;
		sgs_Variable A = *a, B = *b;
		fstk_push2( C, &A, &B );
		vm_convert_stack_string( C, -2 );
		vm_convert_stack_string( C, -1 );
		a = stk_getpos( C, -2 );
		b = stk_getpos( C, -1 );
#if SGS_STRINGTABLE_MAXLEN >= 0x7fffffff
		if( a->data.S == b->data.S )
		{
			stk_pop2( C );
			return 0;
		}
#endif
		out = memcmp( sgs_var_cstr( a ), sgs_var_cstr( b ), SGS_MIN( a->data.S->size, b->data.S->size ) );
		if( out == 0 && a->data.S->size != b->data.S->size )
			out = (ptrdiff_t) ( a->data.S->size - b->data.S->size );
		stk_pop2( C );
		return _SGS_SIGNDIFF( out, 0 );
	}
	
	/* default comparison */
	{
		sgs_Real ra = var_getreal( a );
		sgs_Real rb = var_getreal( b );
		return _SGS_SIGNDIFF( ra, rb );
	}
}


static int vm_forprep( SGS_CTX, StkIdx outiter, sgs_Variable* obj )
{
	int ret = SGS_FALSE;
	sgs_VarObj* O = obj->data.O;
	
	VAR_RELEASE( stk_poff( C, outiter ) );
	
	if( obj->type != SGS_VT_OBJECT )
	{
		sgs_Msg( C, SGS_ERROR, "Variable of type '%s' "
			"doesn't have an iterator", TYPENAME( obj->type ) );
		return SGS_FALSE;
	}
	
	if( O->iface->convert )
	{
		_STACK_PREPARE;
		_STACK_PROTECT;
		ret = SGS_SUCCEEDED( O->iface->convert( C, O, SGS_CONVOP_TOITER ) ) && SGS_STACKFRAMESIZE >= 1;
		_STACK_UNPROTECT_SKIP( ret );
		if( ret )
		{
			stk_setlvar_leave( C, outiter, stk_gettop( C ) );
			stk_pop1nr( C );
		}
	}
	
	if( ret == SGS_FALSE )
	{
		sgs_Msg( C, SGS_ERROR, "Object '%s' doesn't have an iterator", obj->data.O->iface->name );
		return SGS_FALSE;
	}
	return SGS_TRUE;
}

static SGSBOOL vm_fornext( SGS_CTX, StkIdx outkey, StkIdx outval, sgs_Variable* iter )
{
	StkIdx expargs = 0;
	int flags = 0, ret = SGS_ENOTSUP;
	sgs_VarObj* O = iter->data.O;
	_STACK_PREPARE;
	
	if( iter->type != SGS_VT_OBJECT )
	{
		sgs_Msg( C, SGS_ERROR, "iterator is not an object" );
		return SGS_FALSE;
	}
	
	if( O->iface == sgsstd_array_iter_iface )
	{
		/* = sgsstd_array_iter_getnext */
		sgsstd_array_iter_t* it = (sgsstd_array_iter_t*) O->data;
		sgsstd_array_header_t* hdr = (sgsstd_array_header_t*) it->ref.data.O->data;
		if( it->size != hdr->size )
		{
			sgs_Msg( C, SGS_ERROR, "array changed size during iteration" );
			return SGS_FALSE;
		}
		else if( outkey < 0 && outval < 0 )
		{
			it->off++;
			return it->off < it->size;
		}
		else
		{
			if( outkey >= 0 )
				var_setint( C, stk_poff( C, outkey ), it->off );
			if( outval >= 0 )
				stk_setlvar( C, outval, hdr->data + it->off );
			return SGS_TRUE;
		}
	}
	
	if( outkey >= 0 ){ flags |= SGS_GETNEXT_KEY; expargs++; }
	if( outval >= 0 ){ flags |= SGS_GETNEXT_VALUE; expargs++; }
	
	_STACK_PROTECT;
	if( O->iface->getnext )
		ret = O->iface->getnext( C, O, flags );
	if( SGS_FAILED( ret ) || SGS_STACKFRAMESIZE < expargs )
	{
		_STACK_UNPROTECT;
		if( flags == 0 )
			sgs_Msg( C, SGS_ERROR, "failed to advance iterator" );
		else
			sgs_Msg( C, SGS_ERROR, "failed to retrieve data from iterator" );
		return SGS_FALSE;
	}
	_STACK_UNPROTECT_SKIP( expargs );
	
	if( !flags )
		return ret > 0;
	else
	{
		if( outkey >= 0 ) stk_setlvar( C, outkey, stk_getpos( C, -2 + (outval<0) ) );
		if( outval >= 0 ) stk_setlvar( C, outval, stk_gettop( C ) );
		fstk_pop( C, expargs );
	}
	
	return SGS_TRUE;
}


static void vm_make_class( SGS_CTX, int outpos, sgs_Variable* name, sgs_Variable* inhname )
{
	int ret;
	sgs_Variable cls, inhcls;
	
	sgs_BreakIf( name->type != SGS_VT_STRING );
	
	sgs_PushString( C, "__name" );
	fstk_push( C, name );
	ret = sgsSTD_MakeDict( C, &cls, 2 );
	SGS_UNUSED( ret );
	sgs_BreakIf( ret != SGS_TRUE );
	sgs_RegSymbol( C, NULL, sgs_var_cstr( name ), cls );
	
	if( inhname )
	{
		sgs_BreakIf( inhname->type != SGS_VT_STRING );
		if( sgs_GetGlobal( C, *inhname, &inhcls ) && inhcls.type == SGS_VT_OBJECT )
		{
			sgs_ObjSetMetaObj( C, cls.data.O, inhcls.data.O );
			sgs_ObjSetMetaMethodEnable( cls.data.O, 1 );
		}
		else
		{
			sgs_Msg( C, SGS_ERROR, "Class '%s' cannot inherit from '%s' - no such global object",
				sgs_var_cstr( name ), sgs_var_cstr( inhname ) );
		}
	}
	
	stk_setvar_leave( C, outpos, &cls );
}

static void vm_ctor( SGS_CTX, sgs_Variable* inst, sgs_Variable* ctorfunc, int argstart, int argend )
{
	int i;
	if( ctorfunc )
		fstk_push( C, ctorfunc );
	fstk_push( C, inst );
	stk_makespace( C, argend - argstart );
	for( i = argstart; i < argend; ++i )
	{
		sgs_Variable* ptr = stk_poff( C, i );
		VAR_ACQUIRE( ptr );
		*C->stack_top++ = *ptr;
	}
	sgs_ThisCall( C, argend - argstart, 0 );
}

static void vm_make_new( SGS_CTX, int outcls, int lastarg )
{
	sgs_Variable inst, clscopy = *stk_poff( C, outcls );
	if( clscopy.type == SGS_VT_FUNC || clscopy.type == SGS_VT_CFUNC )
	{
		sgsSTD_MakeDict( C, &inst, 0 );
		vm_ctor( C, &inst, &clscopy, outcls + 1, lastarg + 1 );
		stk_setvar_leave( C, outcls, &inst );
	}
	else if( clscopy.type == SGS_VT_OBJECT )
	{
		sgsSTD_MakeDict( C, &inst, 0 );
		sgs_ObjSetMetaObj( C, inst.data.O, clscopy.data.O );
		sgs_ObjSetMetaMethodEnable( inst.data.O, 1 );
		/* call the constructor */
		if( sgs_PushProperty( C, clscopy, "__construct" ) )
		{
			vm_ctor( C, &inst, NULL, outcls + 1, lastarg + 1 );
		}
		else sgs_Pop( C, 1 );
		
		stk_setvar_leave( C, outcls, &inst );
	}
	else
	{
		sgs_Msg( C, SGS_ERROR, "new: expected object" );
	}
}

static int vm_min_closure_count( sgs_Variable* vp )
{
	if( vp->type == SGS_VT_FUNC )
	{
		return vp->data.F->numclsr;
	}
	else
	{
		return 0;
	}
}


#define SGS_EXEC_PAUSED 0x1000
static int vm_exec( SGS_CTX );


/*
	Call the virtual machine.
	Args must equal the number of arguments pushed before the function
	
	- function expects the following items on stack: [this][args]
	- before call, stack is set up to display only the [args] part
	- [this] can be uncovered with sgs_Method
	- arguments are transposed to form [invextraargs][this][reqargs]
	- return value count is checked against the active range at the moment of return
	- upon return, this function replaces [this][args] with [expect]
*/
static int vm_call( SGS_CTX, int args, int gotthis, int* outrvc, int can_reenter )
{
	sgs_Variable* pfunc = C->stack_top - args - gotthis - 1;
	ptrdiff_t stkcallbase, stkoff = SGS_STACK_PRESERVE( C, C->stack_off );
	int rvc = 0, ret = 1, allowed;
	
	gotthis = !!gotthis;
	stkcallbase = ( C->stack_top - C->stack_base ) - args - gotthis;
	
	if( pfunc->type == SGS_VT_OBJECT && pfunc->data.O->mm_enable )
	{
		sgs_Variable objfunc, fncopy = *pfunc;
		sgs_PushString( C, "__call" );
		rvc = sgs_GetIndex( C, fncopy, *stk_gettop( C ), &objfunc, 0 );
		fstk_pop1( C );
		if( SGS_SUCCEEDED( rvc ) )
		{
			// set up metamethod call instead of object call
			if( !gotthis )
			{
				sgs_InsertVariable( C, -1 - args, sgs_MakeNull() );
				gotthis = 1;
			}
			sgs_InsertVariable( C, -2 - args, fncopy );
			args++;
			p_setvar_leave( C->stack_top - args - gotthis - 1, &objfunc );
		}
		pfunc = C->stack_top - args - gotthis - 1;
	}
	
	sgs_BreakIf( SGS_STACKFRAMESIZE < args + gotthis );
	allowed = sgsVM_PushStackFrame( C, pfunc );
	
	if( allowed )
	{
		sgs_StackFrame* sf = C->sf_last;
		/* WP (x4): stack size limit */
		sf->argbeg = (StkIdx) stkcallbase;
		sf->stkoff = (StkIdx) stkoff;
		/* WP: argument count limit */
		sf->argcount = (uint8_t) args;
		sf->flags = gotthis ? SGS_SF_METHOD : 0;
		
		switch( pfunc->type )
		{
		case SGS_VT_CFUNC: {
			C->stack_off = C->stack_top - args;
			SGS_PERFEVENT( ps_precall_end( (unsigned) -1, 1, FUNCTYPE_CFN ) );
			rvc = (*pfunc->data.C)( C );
			if( rvc > SGS_STACKFRAMESIZE )
			{
				sgs_Msg( C, SGS_ERROR, "Function returned more variables than there was on the stack" );
				rvc = 0;
				ret = 0;
			}
			if( rvc < 0 )
			{
				sgs_Msg( C, SGS_ERROR, "The function could not be called" );
				rvc = 0;
				ret = 0;
			}
		} break;
		case SGS_VT_FUNC: {
			sgs_iFunc* F = pfunc->data.F;
			int argend, stkargs, expargs = F->numargs + F->gotthis;
			
			/* if <this> was expected but wasn't properly passed, insert a NULL in its place */
			if( F->gotthis && !gotthis )
			{
				stk_insert_posn( C, args )->type = SGS_VT_NULL;
				sf->flags |= SGS_SF_METHOD;
				gotthis = SGS_TRUE;
			}
			C->stack_off = C->stack_top - args - F->gotthis;
			/* if <this> wasn't expected but was passed, ignore it */
			
			argend = (int) ( C->stack_top - C->stack_base );
			SGS_UNUSED( argend );
			
			/* fix argument stack */
			stkargs = args + F->gotthis;
			if( stkargs > expargs )
			{
				int first = F->numargs + gotthis;
				int all = args + gotthis;
				stk_transpose( C, first, all );
				C->stack_off += all - first;
				stk_push_nulls( C, F->numtmp );
			}
			else
				stk_push_nulls( C, F->numtmp + expargs - stkargs );
			
			{
#if SGS_DEBUG && SGS_DEBUG_VALIDATE && SGS_DEBUG_EXTRA
				/*
				printf("fcall begin=%d end=%d count=%d(%d) inthis=%s expthis=%s\n",
					stkcallbase, sf->argend, sf->argcount, args,
					SGS_HAS_FLAG( sf->flags, SGS_SF_METHOD ) ? "Y" : "N",
					SGS_HAS_FLAG( sf->flags, SGS_SF_HASTHIS ) ? "Y" : "N" );
				*/
				sgs_BreakIf( SGS_SF_ARG_COUNT(sf) != argend - stkcallbase );
#endif
				SGS_PERFEVENT( ps_precall_end( F->numargs, F->gotthis, FUNCTYPE_SGS ) );
				if( can_reenter )
				{
					sf->flags |= SGS_SF_REENTER;
					return -2;
				}
				
				rvc = vm_exec( C );
				if( rvc & SGS_EXEC_PAUSED )
				{
					rvc &= ~SGS_EXEC_PAUSED;
					if( outrvc )
						*outrvc = rvc;
					return 1;
				}
			}
		} break;
		case SGS_VT_OBJECT: {
			sgs_VarObj* O = pfunc->data.O;
			
			C->stack_off = C->stack_top - args;
			rvc = SGS_ENOTSUP;
			SGS_PERFEVENT( ps_precall_end( (unsigned) -1, 1, FUNCTYPE_OBJ ) );
			if( O->iface->call )
				rvc = O->iface->call( C, O );
			
			if( rvc > SGS_STACKFRAMESIZE )
			{
				sgs_Msg( C, SGS_ERROR, "Object returned more variables than there was on the stack" );
				rvc = 0;
				ret = 0;
			}
			if( rvc < 0 )
			{
				sgs_Msg( C, SGS_ERROR, "The object could not be called" );
				rvc = 0;
				ret = 0;
			}
		} break;
		default: {
			/* try to find some name to add to the error */
			if( sf->nfname == NULL && sf->prev )
				sf->nfname = sf->prev->nfname;
			sgs_Msg( C, SGS_ERROR, "Variable of type '%s' "
				"cannot be called", TYPENAME( pfunc->type ) );
			ret = 0;
		} break;
		}
		
		C->state &= ~SGS_STATE_LASTFUNCABORT;
		if( ret && sf->flags & SGS_SF_ABORTED )
		{
			C->state |= SGS_STATE_LASTFUNCABORT;
		}
		vm_frame_pop( C );
	}
	else
	{
		C->state &= ~SGS_STATE_LASTFUNCABORT;
	}
	
	/* remove all stack items before the returned ones */
	fstk_clean( C, C->stack_base + stkcallbase - 1, C->stack_top - rvc );
	C->stack_off = SGS_STACK_RESTORE( C, stkoff );
	
	if( outrvc )
		*outrvc = rvc;
	C->num_last_returned = rvc;
	
	return ret;
}

static void vm_postcall( SGS_CTX, int rvc )
{
	sgs_StackFrame* sf = C->sf_last;
	sgs_StkIdx stkcallbase = sf->argbeg;
	sgs_StkIdx stkoff = sf->stkoff;
	
	vm_frame_pop( C );
	
	/* remove all stack items before the returned ones */
	fstk_clean( C, C->stack_base + stkcallbase - 1, C->stack_top - rvc );
	C->stack_off = SGS_STACK_RESTORE( C, stkoff );
	
	C->num_last_returned = rvc;
	if( C->sf_last )
	{
		sgs_instr_t I;
		int expect, args_from;
		sf = C->sf_last; /* current function change */
		
		if( *sf->iptr == SGS_SI_RETN )
		{
			/* next instr = return 0, this can mean one of two things:
			- call is the last instruction in the function
			- sgs_Abort was called
			... in both cases we do not care about the return or expected values
			*/
			fstk_pop( C, rvc );
		}
		else
		{
			I = *(sf->iptr-1);
			sgs_BreakIf( SGS_INSTR_GET_OP( I ) != SGS_SI_CALL );
			expect = SGS_INSTR_GET_A( I );
			args_from = SGS_INSTR_GET_B( I ) & 0xff;
			stk_resize_expected( C, expect, rvc );
			
			if( expect )
			{
				int i;
				for( i = expect - 1; i >= 0; --i )
					stk_setlvar( C, args_from + i, C->stack_top - expect + i );
				fstk_pop( C, expect );
			}
		}
	}
}


#if SGS_DEBUG && SGS_DEBUG_VALIDATE
static SGS_INLINE sgs_Variable* const_getvar( sgs_Variable* consts, int count, int off )
{
	sgs_BreakIf( off < 0 || off >= count );
	return consts + off;
}

static SGS_INLINE sgs_Closure* clsr_get( sgs_Closure** clsrlist, int count, int off )
{
	sgs_BreakIf( off < 0 || off >= count );
	return clsrlist[ off ];
}
#endif

/*
	Main VM execution loop
*/
static int vm_exec( SGS_CTX )
{
	sgs_StackFrame* SF;
	int32_t ret;
	
#if SGS_DEBUG && SGS_DEBUG_VALIDATE
#  define RESVAR( v ) ( SGS_CONSTVAR(v) ? const_getvar( SF->cptr, SF->constcount, SGS_CONSTDEC(v) ) : stk_poff( C, (v) ) )
#  define CLSR( v ) clsr_get( SF->clsrlist, SF->clsrcount, (v) )
#else
#  define RESVAR( v ) ( SGS_CONSTVAR(v) ? ( SF->cptr + SGS_CONSTDEC(v) ) : stk_poff( C, (v) ) )
#  define CLSR( v ) (SF->clsrlist[v])
#endif
	
#define pp SF->iptr
#if SGS_DEBUG && SGS_DEBUG_VALIDATE
#  define pend SF->iend
#endif
#define instr SGS_INSTR_GET_OP(I)
#define argA SGS_INSTR_GET_A(I)
#define argB SGS_INSTR_GET_B(I)
#define argC SGS_INSTR_GET_C(I)
#define argE SGS_INSTR_GET_E(I)
	
restart_loop:
	SF = C->sf_last;
	ret = 0;
	
#if SGS_DEBUG && SGS_DEBUG_INSTR
	{
		const char *name, *file;
		sgs_StackFrameInfo( C, SF, &name, &file, NULL );
		printf( ">>>\n\t'%s' in %s @ %p [sz:%d]\n>>>\n", name, file, SF, (int) SGS_STACKFRAMESIZE );
	}
#endif
	
	for(;;)
	{
		if( C->hook_fn )
			C->hook_fn( C->hook_ctx, C, SGS_HOOK_STEP );
		
		const sgs_instr_t I = *SF->iptr++;
		
#if SGS_DEBUG
#  if SGS_DEBUG_INSTR
		printf( "*** [at 0x%04X] %s A=%d B=%d C=%d E=%d [sz:%d] ***\n",
			pp - 1 - SF->code, sgs_OpNames[ instr ],
			(int)(argA), (int)(argB), (int)(argC), (int)(argE),
			(int) SGS_STACKFRAMESIZE );
#  endif
#  if SGS_DEBUG_STATE
		sgsVM_StackDump( C );
#  endif
#endif
		SGS_UNUSED( sgs_ErrNames );
		SGS_UNUSED( sgs_OpNames );
		
		switch( instr )
		{
		case SGS_SI_NOP: break;

		case SGS_SI_PUSH:
		{
			sgs_Variable var = *RESVAR( argB );
			stk_push( C, &var );
			break;
		}
		
		case SGS_SI_INT:
		{
			switch( argC )
			{
			case SGS_INT_ERRSUP_INC: SF->errsup++; break;
			case SGS_INT_ERRSUP_DEC: SF->errsup--; break;
			case SGS_INT_RESET_WAIT_TIMER: C->wait_timer = 0; break;
			}
			break;
		}

		case SGS_SI_RET1:
		{
#if SGS_DEBUG && SGS_DEBUG_STATE
			/* TODO restore memcheck */
			sgsVM_StackDump( C );
#endif
			if( SF->flags & SGS_SF_REENTER )
			{
				/* move argument, if expected, to prev_stack_frame->func
				(=location of the expected argument list)
				.. then pop stack frame */
				sgs_instr_t PCI;
				int i, expect;
				sgs_StackFrame* psf = SF->prev;
				sgs_Variable* pvar = RESVAR( argC );
				
				PCI = *(psf->iptr-1);
				sgs_BreakIf( SGS_INSTR_GET_OP( PCI ) != SGS_SI_CALL );
				expect = SGS_INSTR_GET_A( PCI );
				
				/* remove all stack items before the returned ones */
				C->stack_off = SGS_STACK_RESTORE( C, SF->stkoff ); /* ---- STACK CHANGE ---- */
				if( expect )
				{
					int argstart = SGS_INSTR_GET_B( PCI ) & 0xff;
					sgs_Variable* pdst = stk_poff( C, argstart );
					p_setvar( pdst, pvar );
					for( i = 1; i < expect; ++i )
					{
						VAR_RELEASE( &pdst[ i ] );
					}
				}
				stk_popto( C, SF->func );
				C->num_last_returned = 1;
				
				vm_frame_pop( C );
				
				goto restart_loop;
			}
			else
			{
				sgs_Variable var = *RESVAR( argC );
				stk_push( C, &var );
				C->state &= ~SGS_STATE_LASTFUNCPAUSE;
				return 1;
			}
		}
		case SGS_SI_RETN:
		{
			int rvc = argA;
#if SGS_DEBUG && SGS_DEBUG_STATE
			/* TODO restore memcheck */
			sgsVM_StackDump( C );
#endif
			if( SF->flags & SGS_SF_REENTER )
			{
				vm_postcall( C, rvc );
				goto restart_loop;
			}
			C->state &= ~SGS_STATE_LASTFUNCPAUSE;
			return rvc;
		}

		case SGS_SI_JUMP:
		{
			int16_t off = (int16_t) argE;
			pp += off;
			sgs_BreakIf( pp+1 > pend || pp+1 < SF->code );
			break;
		}

		case SGS_SI_JMPT:
		{
			int16_t off = (int16_t) argE;
			sgs_BreakIf( pp+1 + off > pend || pp+1 + off < SF->code );
			if( var_getbool( C, RESVAR( argC ) ) )
				pp += off;
			break;
		}
		case SGS_SI_JMPF:
		{
			int16_t off = (int16_t) argE;
			sgs_BreakIf( pp+1 + off > pend || pp+1 + off < SF->code );
			if( !var_getbool( C, RESVAR( argC ) ) )
				pp += off;
			break;
		}
		case SGS_SI_JMPN:
		{
			int16_t off = (int16_t) argE;
			sgs_BreakIf( pp+1 + off > pend || pp+1 + off < SF->code );
			if( RESVAR( argC )->type == SGS_VT_NULL )
				pp += off;
			break;
		}

		case SGS_SI_CALL:
		SGS_PERFEVENT( ps_precall_begin(
			argC - (argB & 0xff) - (( argB & 0x100 ) != 0),
			( argB & 0x100 ) != 0, CALLSRC_SCRIPT ) );
		{
			/* argB & 0xff .. argC = func [this] [args] */
			int i, rvc = 0, expect = argA, args_from = argB & 0xff, args_to = argC;
			int gotthis = ( argB & 0x100 ) != 0;
			
			stk_makespace( C, args_to - args_from + 1 );
			for( i = args_from; i <= args_to; ++i )
			{
				*C->stack_top++ = C->stack_off[ i ];
				VAR_ACQUIRE( C->stack_off + i );
			}
			
			if( vm_call( C, args_to - args_from - gotthis, gotthis, &rvc, 1 ) == -2 )
			{
				goto restart_loop;
			}
			
			if( C->sf_last->flags & SGS_SF_PAUSED )
			{
				ret = rvc;
				goto paused;
			}
			
			stk_resize_expected( C, expect, rvc );
			
			if( expect )
			{
				for( i = expect - 1; i >= 0; --i )
					stk_setlvar( C, args_from + i, C->stack_top - expect + i );
				fstk_pop( C, expect );
			}
			
			break;
		}

		case SGS_SI_FORPREP: vm_forprep( C, argA, RESVAR( argB ) ); break;
		case SGS_SI_FORLOAD:
			vm_fornext( C,
				argB < 0x100 ? (int)argB : -1,
				argC < 0x100 ? (int)argC : -1,
				stk_poff( C, argA ) );
			break;
		case SGS_SI_FORJUMP:
		{
			int off = argE;
			sgs_BreakIf( pp+1 > pend || pp+1 < SF->code );
			if( vm_fornext( C, -1, -1, RESVAR( argC ) ) < 1 )
				pp += off;
			break;
		}

#define ARGS_2 sgs_Variable p2 = *RESVAR( argB );
#define ARGS_3 sgs_Variable p2 = *RESVAR( argB ), p3 = *RESVAR( argC );
#define SETOP sgs_Variable p1 = *stk_poff( C, argA );
#define GETOP sgs_Variable p1; p1.type = SGS_VT_NULL;
#define WRITEGET stk_setlvar_leave( C, argA, &p1 );
#define WRITEGETC stk_setlvar_leave( C, argC, &p1 );
		case SGS_SI_LOADCONST: { stk_setlvar( C, argC, SF->cptr + argE ); break; }
		case SGS_SI_GETVAR: { ARGS_2; GETOP; sgsSTD_GlobalGet( C, &p1, &p2 ); WRITEGET; break; }
		case SGS_SI_SETVAR: { ARGS_3; sgsSTD_GlobalSet( C, &p2, &p3 ); break; }
		case SGS_SI_GETPROP: { ARGS_3; GETOP; vm_getprop_safe( C, &p1, &p2, &p3, SGS_TRUE ); WRITEGET; break; }
		case SGS_SI_SETPROP: { ARGS_3; SETOP; vm_setprop( C, &p1, &p2, &p3, SGS_TRUE ); break; }
		case SGS_SI_GETINDEX: { ARGS_3; GETOP; vm_getprop_safe( C, &p1, &p2, &p3, SGS_FALSE ); WRITEGET; break; }
		case SGS_SI_SETINDEX: { ARGS_3; SETOP; vm_setprop( C, &p1, &p2, &p3, SGS_FALSE ); break; }
		case SGS_SI_ARRPUSH: {
			sgsstd_array_header_t* hdr = (sgsstd_array_header_t*) stk_poff( C, argA )->data.O->data;
			sgsstd_array_insert_p( C, hdr, hdr->size, RESVAR( argC ) ); break; }
		
		case SGS_SI_CLSRINFO: break;
		case SGS_SI_MAKECLSR: {
			sgs_Variable out;
			sgs_Closure** clsrlist;
			int i = 0, minclsr = vm_min_closure_count( RESVAR( argB ) );
			if( minclsr < argC )
				minclsr = argC;
			/* WP: range not affected by conversion */
			clsrlist = sgsSTD_MakeClosure( C, &out, RESVAR( argB ), (uint32_t) minclsr );
			stk_setvar_leave( C, argA, &out );
			if( argC )
			{
				for( i = 0; i < argC; ++i )
				{
					int varid;
					sgs_instr_t mdi = pp[ i / 3 ];
					switch( i % 3 )
					{
					case 0: varid = SGS_INSTR_GET_A( mdi ); break;
					case 1: varid = SGS_INSTR_GET_B( mdi ); break;
					case 2: varid = SGS_INSTR_GET_C( mdi ); break;
					}
					*clsrlist = CLSR( varid );
					(*clsrlist)->refcount++;
					clsrlist++;
				}
				pp += ( argC + 2 ) / 3;
			}
			for( ; i < minclsr; ++i )
			{
				*clsrlist = sgs_Alloc( sgs_Closure );
				(*clsrlist)->refcount = 1;
				(*clsrlist)->var.type = SGS_VT_NULL;
				clsrlist++;
			}
		} break;
		case SGS_SI_GETCLSR: stk_setlvar( C, argA, &CLSR( argB )->var ); break;
		case SGS_SI_SETCLSR: { sgs_Variable *p3 = RESVAR( argC ), *cv = &CLSR( argB )->var;
			VAR_RELEASE( cv ); *cv = *p3; VAR_ACQUIRE( RESVAR( argC ) ); } break;

		case SGS_SI_SET: { stk_setlvar( C, argA, RESVAR( argB ) ); break; }
		case SGS_SI_MCONCAT: { vm_op_concat_ex( C, argB ); stk_setlvar_leave( C, argA, stk_gettop( C ) ); stk_pop1nr( C ); break; }
		case SGS_SI_CONCAT: { ARGS_3; vm_op_concat( C, argA, &p2, &p3 ); break; }
		case SGS_SI_NEGATE: { ARGS_2; GETOP; vm_op_negate( C, &p1, &p2 ); WRITEGET; break; }
		case SGS_SI_BOOL_INV: { ARGS_2; vm_op_boolinv( C, (int16_t) argA, &p2 ); break; }
		case SGS_SI_INVERT: { ARGS_2; vm_op_invert( C, (int16_t) argA, &p2 ); break; }

		case SGS_SI_INC: { ARGS_2; GETOP; vm_op_incdec( C, &p1, &p2, +1 ); WRITEGET; break; }
		case SGS_SI_DEC: { ARGS_2; GETOP; vm_op_incdec( C, &p1, &p2, -1 ); WRITEGET; break; }
		case SGS_SI_ADD: { ARGS_3;
			if( p2.type == SGS_VT_REAL && p3.type == SGS_VT_REAL ){ GETOP; var_setreal( C, &p1, p2.data.R + p3.data.R ); WRITEGET; break; }
			if( p2.type == SGS_VT_INT && p3.type == SGS_VT_INT ){ GETOP; var_setint( C, &p1, p2.data.I + p3.data.I ); WRITEGET; break; }
			GETOP; vm_arith_op( C, &p1, &p2, &p3, ARITH_OP_ADD ); WRITEGET; break; }
		case SGS_SI_SUB: { ARGS_3;
			if( p2.type == SGS_VT_REAL && p3.type == SGS_VT_REAL ){ GETOP; var_setreal( C, &p1, p2.data.R - p3.data.R ); WRITEGET; break; }
			if( p2.type == SGS_VT_INT && p3.type == SGS_VT_INT ){ GETOP; var_setint( C, &p1, p2.data.I - p3.data.I ); WRITEGET; break; }
			GETOP; vm_arith_op( C, &p1, &p2, &p3, ARITH_OP_SUB ); WRITEGET; break; }
		case SGS_SI_MUL: { ARGS_3;
			if( p2.type == SGS_VT_REAL && p3.type == SGS_VT_REAL ){ GETOP; var_setreal( C, &p1, p2.data.R * p3.data.R ); WRITEGET; break; }
			if( p2.type == SGS_VT_INT && p3.type == SGS_VT_INT ){ GETOP; var_setint( C, &p1, p2.data.I * p3.data.I ); WRITEGET; break; }
			GETOP; vm_arith_op( C, &p1, &p2, &p3, ARITH_OP_MUL ); WRITEGET; break; }
		case SGS_SI_DIV: { ARGS_3; GETOP; vm_arith_op( C, &p1, &p2, &p3, ARITH_OP_DIV ); WRITEGET; break; }
		case SGS_SI_MOD: { ARGS_3; GETOP; vm_arith_op( C, &p1, &p2, &p3, ARITH_OP_MOD ); WRITEGET; break; }

		case SGS_SI_AND: { ARGS_3; vm_op_and( C, (int16_t) argA, &p2, &p3 ); break; }
		case SGS_SI_OR: { ARGS_3; vm_op_or( C, (int16_t) argA, &p2, &p3 ); break; }
		case SGS_SI_XOR: { ARGS_3; vm_op_xor( C, (int16_t) argA, &p2, &p3 ); break; }
		case SGS_SI_LSH: { ARGS_3; vm_op_lsh( C, (int16_t) argA, &p2, &p3 ); break; }
		case SGS_SI_RSH: { ARGS_3; vm_op_rsh( C, (int16_t) argA, &p2, &p3 ); break; }

#define STRICTLY_EQUAL( val ) if( p2.type != p3.type || ( p2.type == SGS_VT_OBJECT && \
								p2.data.O->iface != p3.data.O->iface ) ) { GETOP; var_setbool( C, &p1, val ); WRITEGET; break; }
#define VCOMPARE( op ) { int cr = vm_compare( C, &p2, &p3 ) op 0; GETOP; var_setbool( C, &p1, cr ); WRITEGET; }
		case SGS_SI_SEQ: { ARGS_3; STRICTLY_EQUAL( SGS_FALSE );
			if( p2.type == SGS_VT_INT && p3.type == SGS_VT_INT ){ var_setbool( C, stk_poff( C, argA ), p2.data.I == p3.data.I ); break; }
			if( p2.type == SGS_VT_REAL && p3.type == SGS_VT_REAL ){ var_setbool( C, stk_poff( C, argA ), p2.data.R == p3.data.R ); break; }
			VCOMPARE( == ); break; }
		case SGS_SI_SNEQ: { ARGS_3; STRICTLY_EQUAL( SGS_TRUE );
			if( p2.type == SGS_VT_INT && p3.type == SGS_VT_INT ){ var_setbool( C, stk_poff( C, argA ), p2.data.I != p3.data.I ); break; }
			if( p2.type == SGS_VT_REAL && p3.type == SGS_VT_REAL ){ var_setbool( C, stk_poff( C, argA ), p2.data.R != p3.data.R ); break; }
			VCOMPARE( != ); break; }
		case SGS_SI_EQ: { ARGS_3;
			if( p2.type == SGS_VT_INT && p3.type == SGS_VT_INT ){ var_setbool( C, stk_poff( C, argA ), p2.data.I == p3.data.I ); break; }
			if( p2.type == SGS_VT_REAL && p3.type == SGS_VT_REAL ){ var_setbool( C, stk_poff( C, argA ), p2.data.R == p3.data.R ); break; }
			VCOMPARE( == ); break; }
		case SGS_SI_NEQ: { ARGS_3;
			if( p2.type == SGS_VT_INT && p3.type == SGS_VT_INT ){ var_setbool( C, stk_poff( C, argA ), p2.data.I != p3.data.I ); break; }
			if( p2.type == SGS_VT_REAL && p3.type == SGS_VT_REAL ){ var_setbool( C, stk_poff( C, argA ), p2.data.R != p3.data.R ); break; }
			VCOMPARE( != ); break; }
		case SGS_SI_LT: { ARGS_3;
			if( p2.type == SGS_VT_INT && p3.type == SGS_VT_INT ){ var_setbool( C, stk_poff( C, argA ), p2.data.I < p3.data.I ); break; }
			if( p2.type == SGS_VT_REAL && p3.type == SGS_VT_REAL ){ var_setbool( C, stk_poff( C, argA ), p2.data.R < p3.data.R ); break; }
			VCOMPARE( < ); break; }
		case SGS_SI_GTE: { ARGS_3;
			if( p2.type == SGS_VT_INT && p3.type == SGS_VT_INT ){ var_setbool( C, stk_poff( C, argA ), p2.data.I >= p3.data.I ); break; }
			if( p2.type == SGS_VT_REAL && p3.type == SGS_VT_REAL ){ var_setbool( C, stk_poff( C, argA ), p2.data.R >= p3.data.R ); break; }
			VCOMPARE( >= ); break; }
		case SGS_SI_GT: { ARGS_3;
			if( p2.type == SGS_VT_INT && p3.type == SGS_VT_INT ){ var_setbool( C, stk_poff( C, argA ), p2.data.I > p3.data.I ); break; }
			if( p2.type == SGS_VT_REAL && p3.type == SGS_VT_REAL ){ var_setbool( C, stk_poff( C, argA ), p2.data.R > p3.data.R ); break; }
			VCOMPARE( > ); break; }
		case SGS_SI_LTE: { ARGS_3;
			if( p2.type == SGS_VT_INT && p3.type == SGS_VT_INT ){ var_setbool( C, stk_poff( C, argA ), p2.data.I <= p3.data.I ); break; }
			if( p2.type == SGS_VT_REAL && p3.type == SGS_VT_REAL ){ var_setbool( C, stk_poff( C, argA ), p2.data.R <= p3.data.R ); break; }
			VCOMPARE( <= ); break; }
		case SGS_SI_RAWCMP: { ARGS_3; GETOP; var_setint( C, &p1, vm_compare( C, &p2, &p3 ) ); WRITEGET; break; }
			
		case SGS_SI_ARRAY: { GETOP; sgsSTD_MakeArray( C, &p1, -argE ); WRITEGETC; break; }
		case SGS_SI_DICT: { GETOP; sgsSTD_MakeDict( C, &p1, 0 ); WRITEGETC; break; }
		case SGS_SI_MAP: { GETOP; sgsSTD_MakeMap( C, &p1, 0 ); WRITEGETC; break; }
		case SGS_SI_CLASS: { vm_make_class( C, argA, RESVAR( argB ), argC != argA ? RESVAR( argC ) : NULL ); break; }
		case SGS_SI_NEW: { vm_make_new( C, argB, argC ); break; }
		case SGS_SI_RSYM: { ARGS_3; sgs_Variable symtbl = sgs_Registry( C, SGS_REG_SYM );
			sgs_SetIndex( C, symtbl, p2, p3, SGS_FALSE ); sgs_SetIndex( C, symtbl, p3, p2, SGS_FALSE ); break; }
			
		case SGS_SI_COTRT: if( var_getfin( C, RESVAR( argB ) ) ){ GETOP; var_setbool( C, &p1, 1 ); WRITEGET; } break;
		case SGS_SI_COTRF: if( !var_getfin( C, RESVAR( argB ) ) ){ GETOP; var_setbool( C, &p1, 0 ); WRITEGET; } break;
		case SGS_SI_COABORT:
			if( var_getbool( C, RESVAR( argC ) ) )
			{
				int16_t off = (int16_t) argE;
				const sgs_instr_t* p = pp + 1 + off, *pstop = pp;
				sgs_BreakIf( p > pstop || p < SF->code );
				while( p < pstop )
				{
					sgs_instr_t cI = *p;
					int op = SGS_INSTR_GET_OP( cI );
					if( op == SGS_SI_COTRT || op == SGS_SI_COTRF )
					{
						int cIargB = SGS_INSTR_GET_B( cI );
						sgs_Variable* var = RESVAR( cIargB );
						if( var->type == SGS_VT_THREAD )
							sgs_Abort( var->data.T );
					}
					p++;
				}
			}
			break;
		case SGS_SI_YLDJMP:
			if( var_getbool( C, RESVAR( argC ) ) == SGS_FALSE )
			{
				ret = 0;
				goto paused;
			}
			break;
#undef VCOMPARE
#undef STRICTLY_EQUAL
#undef ARGS_2
#undef ARGS_3
#undef RESVAR
#undef argA
#undef argB
#undef argC
#undef argE
#undef instr
#undef pend
#undef pp

		default:
			sgs_Msg( C, SGS_ERROR, "Illegal instruction executed: 0x%08X", I );
			break;
		}
	}
	
paused:
	C->sf_last->flags |= SGS_SF_PAUSED;
	C->state |= SGS_STATE_LASTFUNCPAUSE;
	if( C->hook_fn )
		C->hook_fn( C->hook_ctx, C, SGS_HOOK_PAUSE );
	return ret | SGS_EXEC_PAUSED;
}


/* INTERNAL INERFACE */

SGSBOOL sgs_ResumeStateRet( SGS_CTX, int args, int* outrvc )
{
	sgs_instr_t I;
	int op, rvc = 0;
	if( C->sf_last == NULL || ( C->sf_last->flags & SGS_SF_PAUSED ) == 0 )
		return SGS_FALSE; /* already running, may not return the expected data */
	if( C->sf_first->flags & SGS_SF_ABORTED )
	{
		while( C->sf_last )
			vm_frame_pop( C );
		C->state |= SGS_STATE_LASTFUNCABORT;
		if( outrvc )
			*outrvc = 0;
		return SGS_TRUE;
	}
	
	/* TODO validate state corruption */
	I = *(C->sf_last->iptr-1);
	op = SGS_INSTR_GET_OP( I );
	sgs_BreakIf( op != SGS_SI_CALL && op != SGS_SI_YLDJMP );
	if( op == SGS_SI_CALL )
	{
		int i, expect = SGS_INSTR_GET_A( I );
		int args_from = SGS_INSTR_GET_B( I ) & 0xff;
		stk_resize_expected( C, expect, args );
		
		if( expect )
		{
			for( i = expect - 1; i >= 0; --i )
				stk_setlvar( C, args_from + i, C->stack_top - expect + i );
			fstk_pop( C, expect );
		}
	}
	else if( op == SGS_SI_YLDJMP )
	{
		sgs_StackFrame* sf = C->sf_last;
		int16_t off = (int16_t) SGS_INSTR_GET_E( I );
		sf->iptr += off;
		sgs_BreakIf( sf->iptr+1 > sf->iend || sf->iptr+1 < sf->code );
	}
	
	C->sf_last->flags &= (uint8_t)~SGS_SF_PAUSED;
	
	if( C->hook_fn )
		C->hook_fn( C->hook_ctx, C, SGS_HOOK_CONT );
	
	rvc = vm_exec( C );
	if( ( rvc & SGS_EXEC_PAUSED ) == 0 )
		vm_postcall( C, rvc );
	if( outrvc )
		*outrvc = rvc & ~SGS_EXEC_PAUSED;
	
	return SGS_TRUE;
}

static size_t funct_size( const sgs_iFunc* f )
{
	size_t sz = f->size + sizeof( sgs_iStr ) * 2 + f->sfuncname->size + f->sfilename->size;
	const sgs_Variable* beg = SGS_ASSUME_ALIGNED_CONST( sgs_func_c_consts( f ), sgs_Variable );
	const sgs_Variable* end = SGS_ASSUME_ALIGNED_CONST( sgs_func_c_bytecode( f ), sgs_Variable );
	while( beg < end )
		sz += sgsVM_VarSize( beg++ );
	return sz;
}

size_t sgsVM_VarSize( const sgs_Variable* var )
{
	size_t out;
	if( !var )
		return 0;

	out = sizeof( sgs_Variable );
	switch( var->type )
	{
	case SGS_VT_FUNC: out += funct_size( var->data.F ); break;
	/* case SGS_VT_OBJECT: break; */
	case SGS_VT_STRING: out += var->data.S->size + sizeof( sgs_iStr ); break;
	}
	return out;
}

void sgsVM_VarDump( const sgs_Variable* var )
{
	/* WP: variable size limit */
	printf( "%s (size:%d)", TYPENAME( var->type ), (uint32_t) sgsVM_VarSize( var ) );
	switch( var->type )
	{
	case SGS_VT_NULL: break;
	case SGS_VT_BOOL: printf( " = %s", var->data.B ? "True" : "False" ); break;
	case SGS_VT_INT: printf( " = %" PRId64, var->data.I ); break;
	case SGS_VT_REAL: printf( " = %f", var->data.R ); break;
	case SGS_VT_STRING: printf( " [rc:%" PRId32"] = \"", var->data.S->refcount );
		sgs_print_safe( stdout, sgs_var_cstr( var ), SGS_MIN( var->data.S->size, 16 ) );
		printf( var->data.S->size > 16 ? "...\"" : "\"" ); break;
	case SGS_VT_FUNC: printf( " [rc:%" PRId32"]", var->data.F->refcount ); break;
	case SGS_VT_CFUNC: printf( " = %p", (void*)(size_t) var->data.C ); break;
	case SGS_VT_OBJECT: printf( " [rc:%" PRId32"] = %p", var->data.O->refcount, var->data.O ); break;
	case SGS_VT_PTR: printf( " = %p", var->data.P ); break;
	case SGS_VT_THREAD: printf( " [rc:%" PRId32"] = %p", var->data.T->refcount, var->data.T ); break;
	}
}

void sgsVM_StackDump( SGS_CTX )
{
	ptrdiff_t i, stksz = C->stack_top - C->stack_base;
	/* WP: stack limit */
	printf( "STACK (size=%d, bytes=%d/%d)--\n", (int) stksz, (int)( stksz * (ptrdiff_t) STK_UNITSIZE ), (int)( C->stack_mem * STK_UNITSIZE ) );
	for( i = 0; i < stksz; ++i )
	{
		sgs_Variable* var = C->stack_base + i;
		if( var == C->stack_off )
			printf( "-- offset --\n" );
		printf( "  " ); sgsVM_VarDump( var ); printf( "\n" );
	}
	printf( "--\n" );
}



/* ---- The core interface ---- */

static void sgs_StackIdxError( SGS_CTX, sgs_StkIdx item )
{
	sgs_Msg( C, SGS_ERROR, "invalid stack index - %d (abs = %d, stack size = %d)",
		(int) item, (int) stk_absindex( C, item ), (int) SGS_STACKFRAMESIZE );
}

/*
	
	OBJECT CREATION
	
*/

void sgs_InitStringBuf( SGS_CTX, sgs_Variable* out, const char* str, sgs_SizeVal size )
{
	sgs_BreakIf( !str && size && "sgs_InitStringBuf: str = NULL" );
	var_create_str( C, out, str, size );
}

void sgs_InitString( SGS_CTX, sgs_Variable* out, const char* str )
{
	size_t sz;
	sgs_BreakIf( !str && "sgs_InitString: str = NULL" );
	sz = SGS_STRINGLENGTHFUNC(str);
	sgs_BreakIf( sz > 0x7fffffff && "sgs_InitString: size exceeded" );
	/* WP: error detection */
	var_create_str( C, out, str, (sgs_SizeVal) sz );
}

void sgs_InitObjectPtr( sgs_Variable* out, sgs_VarObj* obj )
{
	out->type = SGS_VT_OBJECT;
	out->data.O = obj;
	VAR_ACQUIRE( out );
}

void sgs_InitThreadPtr( sgs_Variable* out, sgs_Context* T )
{
	out->type = SGS_VT_THREAD;
	out->data.T = T;
	VAR_ACQUIRE( out );
}


SGSONE sgs_CreateObject( SGS_CTX, sgs_Variable* out, void* data, sgs_ObjInterface* iface )
{
	sgs_Variable var;
	var_create_obj( C, &var, data, iface, 0 );
	if( out )
		*out = var;
	else
		stk_push_leave( C, &var );
	return 1;
}

void* sgs_CreateObjectIPA( SGS_CTX, sgs_Variable* out, uint32_t added, sgs_ObjInterface* iface )
{
	sgs_Variable var;
	var_create_obj( C, &var, NULL, iface, added );
	if( out )
		*out = var;
	else
		stk_push_leave( C, &var );
	return var.data.O->data;
}

SGSONE sgs_CreateArray( SGS_CTX, sgs_Variable* out, sgs_SizeVal numitems )
{
	sgs_Variable var;
	var.type = SGS_VT_NULL;
	sgsSTD_MakeArray( C, &var, numitems );
	if( out )
		*out = var;
	else
		stk_push_leave( C, &var );
	return 1;
}

SGSONE sgs_CreateDict( SGS_CTX, sgs_Variable* out, sgs_SizeVal numitems )
{
	sgs_Variable var;
	var.type = SGS_VT_NULL;
	sgsSTD_MakeDict( C, &var, numitems );
	if( out )
		*out = var;
	else
		stk_push_leave( C, &var );
	return 1;
}

SGSONE sgs_CreateMap( SGS_CTX, sgs_Variable* out, sgs_SizeVal numitems )
{
	sgs_Variable var;
	var.type = SGS_VT_NULL;
	sgsSTD_MakeMap( C, &var, numitems );
	if( out )
		*out = var;
	else
		stk_push_leave( C, &var );
	return 1;
}


/*

	STACK & SUB-ITEMS

*/

SGSONE sgs_PushNulls( SGS_CTX, int count )
{
	stk_push_nulls( C, count );
	return 1;
}

SGSONE sgs_PushBool( SGS_CTX, sgs_Bool value )
{
	sgs_Variable var;
	var.type = SGS_VT_BOOL;
	var.data.B = value ? 1 : 0;
	stk_push_leave( C, &var );
	return 1;
}

SGSONE sgs_PushInt( SGS_CTX, sgs_Int value )
{
	sgs_Variable var;
	var.type = SGS_VT_INT;
	var.data.I = value;
	stk_push_leave( C, &var );
	return 1;
}

SGSONE sgs_PushReal( SGS_CTX, sgs_Real value )
{
	sgs_Variable var;
	var.type = SGS_VT_REAL;
	var.data.R = value;
	stk_push_leave( C, &var );
	return 1;
}

SGSONE sgs_PushStringBuf( SGS_CTX, const char* str, sgs_SizeVal size )
{
	sgs_Variable var;
	sgs_BreakIf( !str && size && "sgs_PushStringBuf: str = NULL" );
	var_create_str( C, &var, str, size );
	stk_push_leave( C, &var );
	return 1;
}

SGSONE sgs_PushString( SGS_CTX, const char* str )
{
	size_t sz;
	sgs_Variable var;
	sgs_BreakIf( !str && "sgs_PushString: str = NULL" );
	sz = SGS_STRINGLENGTHFUNC(str);
	sgs_BreakIf( sz > 0x7fffffff && "sgs_PushString: size exceeded" );
	/* WP: error detection */
	var_create_str( C, &var, str, (sgs_SizeVal) sz );
	stk_push_leave( C, &var );
	return 1;
}

SGSONE sgs_PushCFunc( SGS_CTX, sgs_CFunc func )
{
	sgs_Variable var;
	var.type = SGS_VT_CFUNC;
	var.data.C = func;
	stk_push_leave( C, &var );
	return 1;
}

SGSONE sgs_PushPtr( SGS_CTX, void* ptr )
{
	sgs_Variable var;
	var.type = SGS_VT_PTR;
	var.data.P = ptr;
	stk_push_leave( C, &var );
	return 1;
}

SGSONE sgs_PushObjectPtr( SGS_CTX, sgs_VarObj* obj )
{
	sgs_Variable var;
	var.type = SGS_VT_OBJECT;
	var.data.O = obj;
	fstk_push( C, &var ); /* a new reference must be created */
	return 1;
}

SGSONE sgs_PushThreadPtr( SGS_CTX, sgs_Context* T )
{
	sgs_Variable var;
	var.type = SGS_VT_THREAD;
	var.data.T = T;
	fstk_push( C, &var ); /* a new reference must be created */
	return 1;
}


SGSONE sgs_PushVariable( SGS_CTX, sgs_Variable var )
{
	fstk_push( C, &var );
	return 1;
}

void sgs_StoreVariable( SGS_CTX, sgs_Variable* out )
{
	*out = sgs_StackItem( C, -1 );
	stk_pop1nr( C );
}

void sgs_SetStackItem( SGS_CTX, StkIdx item, sgs_Variable val )
{
	if( sgs_IsValidIndex( C, item ) == SGS_FALSE )
	{
		sgs_StackIdxError( C, item );
		return;
	}
	stk_setvar( C, item, &val );
}

void sgs_InsertVariable( SGS_CTX, sgs_StkIdx pos, sgs_Variable val )
{
	if( pos > sgs_StackSize( C ) || pos < -sgs_StackSize( C ) - 1 )
	{
		sgs_Msg( C, SGS_ERROR, "sgs_InsertVariable: invalid index - %d (stack size = %d)",
			(int) pos, (int) SGS_STACKFRAMESIZE );
	}
	else
	{
		if( pos < 0 )
			pos = sgs_StackSize( C ) + 1 + pos;
		sgs_Variable* vp = stk_insert_pos( C, pos );
		*vp = val;
		VAR_ACQUIRE( vp );
	}
}


/* string generation */
char* sgs_PushStringAlloc( SGS_CTX, sgs_SizeVal size )
{
	sgs_Variable var;
	var_create_0str( C, &var, (uint32_t) size );
	stk_push_leave( C, &var );
	return sgs_var_cstr( &var );
}

char* sgs_InitStringAlloc( SGS_CTX, sgs_Variable* var, sgs_SizeVal size )
{
	var_create_0str( C, var, (uint32_t) size );
	return sgs_var_cstr( var );
}

void sgs_FinalizeStringAlloc( SGS_CTX, sgs_StkIdx item )
{
	sgs_Variable var = sgs_StackItem( C, item );
	sgs_FinalizeStringAllocP( C, &var );
	*stk_getpos( C, item ) = var;
}

void sgs_FinalizeStringAllocP( SGS_CTX, sgs_Variable* var )
{
	if( var->type != SGS_VT_STRING )
		sgs_Msg( C, SGS_APIERR, "sgs_FinalizeStringAlloc: string required" );
	else
		var_finalize_str( C, var );
}



SGSBOOL sgs_GetIndex( SGS_CTX, sgs_Variable obj, sgs_Variable idx, sgs_Variable* out, int isprop )
{
	int ret;
	_EL_BACKUP;
	_EL_SETAPI(1);
	ret = vm_getprop( C, out, &obj, &idx, isprop );
	if( SGS_SUCCEEDED( ret ) )
		VM_GETPROP_RETPTR( ret, out );
	else
		var_initnull( out );
	_EL_RESET;
	return SGS_SUCCEEDED( ret );
}

SGSBOOL sgs_SetIndex( SGS_CTX, sgs_Variable obj, sgs_Variable idx, sgs_Variable val, int isprop )
{
	int ret;
	_EL_BACKUP;
	_EL_SETAPI(1);
	ret = vm_setprop( C, &obj, &idx, &val, isprop );
	_EL_RESET;
	return SGS_SUCCEEDED( ret );
}

SGSBOOL sgs_PushIndex( SGS_CTX, sgs_Variable obj, sgs_Variable idx, int isprop )
{
	int ret;
	sgs_Variable tmp;
	_EL_BACKUP;
	_EL_SETAPI(1);
	ret = vm_getprop( C, &tmp, &obj, &idx, isprop );
	if( SGS_SUCCEEDED( ret ) )
		VM_GETPROP_RETTOP( ret, &tmp );
	else
		fstk_push_null( C );
	_EL_RESET;
	return SGS_SUCCEEDED( ret );
}


SGSBOOL sgs_PushProperty( SGS_CTX, sgs_Variable obj, const char* name )
{
	int ret;
	sgs_PushString( C, name );
	ret = sgs_PushIndex( C, obj, *stk_gettop( C ), SGS_TRUE );
	stk_popskip( C, 1, 1 );
	return ret;
}

SGSBOOL sgs_SetProperty( SGS_CTX, sgs_Variable obj, const char* name, sgs_Variable val )
{
	int ret;
	sgs_PushString( C, name );
	ret = sgs_SetIndex( C, obj, *stk_gettop( C ), val, SGS_TRUE );
	fstk_pop1( C );
	return ret;
}

SGSBOOL sgs_PushNumIndex( SGS_CTX, sgs_Variable obj, sgs_Int idx )
{
	sgs_Variable ivar;
	ivar.type = SGS_VT_INT;
	ivar.data.I = idx;
	return sgs_PushIndex( C, obj, ivar, SGS_FALSE );
}

SGSBOOL sgs_SetNumIndex( SGS_CTX, sgs_Variable obj, sgs_Int idx, sgs_Variable val )
{
	sgs_Variable ivar;
	ivar.type = SGS_VT_INT;
	ivar.data.I = idx;
	return sgs_SetIndex( C, obj, ivar, val, SGS_FALSE );
}


SGSBOOL sgs_GetGlobal( SGS_CTX, sgs_Variable idx, sgs_Variable* out )
{
	int ret;
	_EL_BACKUP;
	out->type = SGS_VT_NULL;
	_EL_SETAPI(1);
	ret = sgsSTD_GlobalGet( C, out, &idx );
	_EL_RESET;
	return ret;
}

SGSBOOL sgs_SetGlobal( SGS_CTX, sgs_Variable idx, sgs_Variable val )
{
	int ret;
	sgs_ObjAcquire( C, C->_G );
	sgs_Variable vO; vO.type = SGS_VT_OBJECT; vO.data.O = C->_G;
	ret = sgs_SetIndex( C, vO, idx, val, 0 );
	sgs_ObjRelease( C, C->_G );
	return ret;
}

SGSBOOL sgs_PushGlobalByName( SGS_CTX, const char* name )
{
	int ret;
	sgs_Variable val;
	ret = sgs_GetGlobalByName( C, name, &val );
	stk_push_leave( C, &val );
	return ret;
}

SGSBOOL sgs_GetGlobalByName( SGS_CTX, const char* name, sgs_Variable* out )
{
	int ret;
	sgs_Variable str;
	sgs_InitString( C, &str, name );
	ret = sgs_GetGlobal( C, str, out );
	sgs_Release( C, &str );
	return ret;
}

void sgs_SetGlobalByName( SGS_CTX, const char* name, sgs_Variable val )
{
	int ret;
	sgs_Variable str;
	sgs_InitString( C, &str, name );
	ret = sgs_SetGlobal( C, str, val );
	sgs_Release( C, &str );
	sgs_BreakIf( ret == SGS_FALSE );
	SGS_UNUSED( ret );
}


sgs_Variable sgs_Registry( SGS_CTX, int subtype )
{
	sgs_Variable out;
	out.type = SGS_VT_OBJECT;
	switch( subtype )
	{
	case SGS_REG_ROOT:
		out.data.O = C->shared->_R;
		break;
	case SGS_REG_SYM:
		out.data.O = C->shared->_SYM;
		break;
	case SGS_REG_INC:
		out.data.O = C->shared->_INC;
		break;
	default:
		out.type = SGS_VT_NULL;
		sgs_Msg( C, SGS_APIERR, "sgs_Registry: invalid subtype (%d)", subtype );
		break;
	}
	return out;
}

void sgs_GetEnv( SGS_CTX, sgs_Variable* out )
{
	sgs_InitObjectPtr( out, C->_G );
}

void sgs_SetEnv( SGS_CTX, sgs_Variable var )
{
	if( var.type != SGS_VT_OBJECT ||
		( var.data.O->iface != sgsstd_dict_iface && var.data.O->iface != sgsstd_map_iface ) )
	{
		sgs_Msg( C, SGS_APIERR, "sgs_SetEnv: argument not dict/map" );
		return;
	}
	VAR_ACQUIRE( &var );
	sgs_ObjRelease( C, C->_G );
	C->_G = var.data.O;
}

void sgs_PushEnv( SGS_CTX )
{
	sgs_PushObjectPtr( C, C->_G );
}


/*
	o = offset (isprop=true) (sizeval)
	p = property (isprop=true) (null-terminated string)
	s = super-secret property (isprop=true) (sizeval, buffer)
	i = index (isprop=false) (sizeval)
	k = key (isprop=false) (null-terminated string)
	n = null-including key (isprop=false) (sizeval, buffer)
*/

static SGSBOOL sgs_parse_path_key( SGS_CTX, const char* fn, size_t at,
	va_list* pargs, char a, sgs_Variable* pkey, int* pisprop )
{
	sgs_SizeVal S = -1;
	char* P = NULL;
	
	if( a == 'o' )
	{
		*pisprop = 1;
		S = va_arg( *pargs, int );
	}
	else if( a == 'p' )
	{
		*pisprop = 1;
		P = va_arg( *pargs, char* );
		if( !P )
			goto nullptrerr;
	}
	else if( a == 's' )
	{
		*pisprop = 1;
		S = va_arg( *pargs, int );
		P = va_arg( *pargs, char* );
		if( !P )
			goto nullptrerr;
	}
	else if( a == 'i' )
	{
		*pisprop = 0;
		S = va_arg( *pargs, int );
	}
	else if( a == 'k' )
	{
		*pisprop = 0;
		P = va_arg( *pargs, char* );
		if( !P )
			goto nullptrerr;
	}
	else if( a == 'n' )
	{
		*pisprop = 0;
		S = va_arg( *pargs, int );
		P = va_arg( *pargs, char* );
		if( !P )
			goto nullptrerr;
	}
	else
	{
		sgs_Msg( C, SGS_APIERR, "%s: (pos. %d) unrecognized character '%c'", fn, (int) at, a );
		return SGS_FALSE;
	}
	
	if( P )
	{
		if( S >= 0 )
			sgs_InitStringBuf( C, pkey, P, S );
		else
			sgs_InitString( C, pkey, P );
	}
	else if( S >= 0 )
		*pkey = sgs_MakeInt( S );
	else
	{
		sgs_Msg( C, SGS_INTERR, "%s: (pos. %d) internal path parsing error", fn, (int) at );
		return SGS_FALSE;
	}
	return SGS_TRUE;
	
nullptrerr:
	sgs_Msg( C, SGS_APIERR, "%s: (pos. %d) [%c] = null string pointer passed", fn, (int) at, a );
	return SGS_FALSE;
}

static SGSBOOL sgs_PushPathBuf( SGS_CTX, const char* fn,
	sgs_Variable var, const char* path, size_t plen, va_list* pargs )
{
	int ret = SGS_SUCCESS;
	size_t i = 0;
	fstk_push( C, &var );
	while( i < plen )
	{
		sgs_Variable key;
		int prop = -1;
		char a = path[ i++ ];
		
		if( sgs_parse_path_key( C, fn, i, pargs, a, &key, &prop ) == SGS_FALSE )
			return SGS_FALSE;
		
		ret = sgs_PushIndex( C, *stk_gettop( C ), key, prop );
		VAR_RELEASE( &key );
		if( ret == SGS_FALSE )
			return ret;
		stk_popskip( C, 1, 1 );
	}
	return SGS_TRUE;
}

SGSBOOL sgs_PushPath( SGS_CTX, sgs_Variable var, const char* path, ... )
{
	int ret;
	StkIdx ssz = SGS_STACKFRAMESIZE;
	va_list args;
	va_start( args, path );
	ret = sgs_PushPathBuf( C, "sgs_PushPath", var, path, strlen(path), &args );
	if( ret )
	{
		stk_popskip( C, SGS_STACKFRAMESIZE - ssz - 1, 1 );
	}
	else
	{
		fstk_pop( C, SGS_STACKFRAMESIZE - ssz );
	}
	va_end( args );
	return ret;
}

SGSBOOL sgs_StorePath( SGS_CTX, sgs_Variable var, sgs_Variable val, const char* path, ... )
{
	int ret;
	size_t len = strlen( path );
	StkIdx ssz = SGS_STACKFRAMESIZE;
	va_list args;
	if( !*path )
	{
		sgs_Msg( C, SGS_APIERR, "sgs_StorePath: expected non-empty path" );
		return SGS_FALSE;
	}
	va_start( args, path );
	ret = sgs_PushPathBuf( C, "sgs_StorePath", var, path, len - 1, &args );
	if( ret )
	{
		sgs_Variable key;
		int prop = -1;
		char a = path[ len - 1 ];
		
		if( ( ret = sgs_parse_path_key( C, "sgs_StorePath", len - 1, &args, a, &key, &prop ) ) == SGS_FALSE )
			goto fail;
		
		ret = sgs_SetIndex( C, *stk_gettop( C ), key, val, prop );
		VAR_RELEASE( &key );
		if( ret == SGS_FALSE )
			goto fail;
		ssz--;
	}
fail:
	va_end( args );
	fstk_pop( C, SGS_STACKFRAMESIZE - ssz );
	return ret;
}

/*
	argument unpacking:
	n - null (not sure why but I had a free letter here)
	b - boolean
	c,w,l,q,i,I - integers (int8,int16,int32,int64 x2,int)
	f,d,r - floats (reals) (float32,float64 x2)
	s,m - strings (string,string+size)
	p - function (callable, actually; p stands for "procedure";
		returns a SGSBOOL always, useful only for optional arguments)
	a,t,h,o,* - objects (array,dict,map,specific iface,any iface)
	& - pointer (void*)
	y - thread (sgs_Context*)
	v - any variable (returns sgs_Variable, checks if valid index or non-null if strict)
	x - custom check/return function
	? - check only, no writeback
	! - strict check (no conversions)
	-,+ - signed/unsigned (changes checked ranges and return value signs)
	| - optional arguments follow
	# - check ranges
	^ - clamp ranges
	~ - ignore ranges
	< - move argument pointer 1 item back
	> - move argument pointer 1 item forward
	@ - treat as method (1st arg = this, argnum -= 1)
	. - check if there are no more arguments
*/

int sgs_ArgErrorExt( SGS_CTX, int argid, int gotthis, const char* expect, const char* expfx )
{
	const char* got = sgs_StackSize( C ) <= argid ? "nothing" : 
		sgs_CodeString( SGS_CODE_VT, (int) sgs_ItemType( C, argid ) );
	if( argid == 0 && gotthis )
		return sgs_Msg( C, SGS_WARNING, "'this' - expected %s%s, got %s", expfx, expect, got );
	else
		return sgs_Msg( C, SGS_WARNING, "argument %d - expected %s%s, got %s",
			argid + !gotthis, expfx, expect, got );
}

int sgs_ArgError( SGS_CTX, int argid, int gotthis, int expect, int is_strict )
{
	return sgs_ArgErrorExt( C, argid, gotthis,
		sgs_CodeString( SGS_CODE_VT, expect ), is_strict ? "strict " : "" );
}

#define argerr sgs_ArgError
#define argerrx sgs_ArgErrorExt

SGSBOOL sgs_LoadArgsExtVA( SGS_CTX, int from, const char* cmd, va_list* args )
{
	int opt = 0;
	int strict = 0;
	int range = 0; /* <0: clamp, >0: check */
	int isig = 1;
	int nowrite = 0;
	int method = 0;
	while( *cmd )
	{
		switch( *cmd )
		{
		case '|': opt = 1; break;
		case '#': range = 1; break;
		case '^': range = -1; break;
		case '~': range = 0; break;
		case '-': isig = 1; break;
		case '+': isig = 0; break;
		case '!': strict = 1; break;
		case '?': nowrite = 1; break;
		case '<': from--;
			if( from < 0 )
			{
				return sgs_Msg( C, SGS_APIERR, "sgs_LoadArgs*: cannot move argument pointer before 0" );
			}
			break;
		case '>': from++; break;
		case '@': method = 1; break;
		case '.':
			if( from < sgs_StackSize( C ) )
			{
				return sgs_Msg( C, SGS_WARNING, "function expects exactly %d arguments, %d given",
					from - method, sgs_StackSize( C ) - method );
			}
			break;
		
		case 'n':
			{
				if( sgs_ItemType( C, from ) != SGS_VT_NULL )
				{
					argerr( C, from, method, SGS_VT_NULL, 0 );
					return opt;
				}
				
				if( !nowrite )
				{
					*va_arg( *args, SGSBOOL* ) = 1;
				}
			}
			strict = 0; nowrite = 0; from++; break;
		
		case 'b':
			{
				sgs_Bool b;
				
				if( opt && sgs_ItemType( C, from ) == SGS_VT_NULL )
				{
					if( !nowrite )
						va_arg( *args, sgs_Bool* );
					break;
				}
				
				if( !sgs_ParseBool( C, from, &b ) ||
					( strict && sgs_ItemType( C, from ) != SGS_VT_BOOL ) )
				{
					argerr( C, from, method, SGS_VT_BOOL, strict );
					return opt;
				}
				
				if( !nowrite )
				{
					*va_arg( *args, sgs_Bool* ) = b;
				}
			}
			strict = 0; nowrite = 0; from++; break;
		
		case 'c': case 'w': case 'l': case 'q': case 'i': case 'I':
			{
				sgs_Int i, imin = INT64_MIN, imax = INT64_MAX;
				
				if( opt && sgs_ItemType( C, from ) == SGS_VT_NULL )
				{
					if( !nowrite )
					{
						switch( *cmd )
						{
						case 'c': va_arg( *args, uint8_t* ); break;
						case 'w': va_arg( *args, uint16_t* ); break;
						case 'l': va_arg( *args, uint32_t* ); break;
						case 'q': va_arg( *args, uint64_t* ); break;
						case 'i': va_arg( *args, sgs_Int* ); break;
						case 'I': va_arg( *args, int* ); break;
						}
					}
					break;
				}
				
				if( range )
				{
					if( *cmd == 'c' && isig ){ imin = INT8_MIN; imax = INT8_MAX; }
					else if( *cmd == 'c' && !isig ){ imin = 0; imax = UINT8_MAX; }
					else if( *cmd == 'w' && isig ){ imin = INT16_MIN; imax = INT16_MAX; }
					else if( *cmd == 'w' && !isig ){ imin = 0; imax = UINT16_MAX; }
					else if( *cmd == 'l' && isig ){ imin = INT32_MIN; imax = INT32_MAX; }
					else if( *cmd == 'l' && !isig ){ imin = 0; imax = UINT32_MAX; }
					else if( *cmd == 'I' && isig ){ imin = INT_MIN; imax = INT_MAX; }
					else if( *cmd == 'I' && !isig ){ imin = 0; imax = UINT_MAX; }
				}
				
				if( !sgs_ParseInt( C, from, &i ) ||
					( strict && sgs_ItemType( C, from ) != SGS_VT_INT ) )
				{
					argerr( C, from, method, SGS_VT_INT, strict );
					return opt;
				}
				
				if( range > 0 && ( i < imin || i > imax ) )
				{
					sgs_Msg( C, SGS_WARNING, "integer argument %d (%" PRId64 ") out of range [%"
						PRId64 ",%" PRId64 "]", from, i, imin, imax );
					return opt;
				}
				else if( range < 0 )
				{
					if( i < imin ) i = imin;
					else if( i > imax ) i = imax;
				}
				
				if( !nowrite )
				{
					switch( *cmd )
					{
					case 'c': *va_arg( *args, uint8_t* ) = (uint8_t) i; break;
					case 'w': *va_arg( *args, uint16_t* ) = (uint16_t) i; break;
					case 'l': *va_arg( *args, uint32_t* ) = (uint32_t) i; break;
					case 'q': *va_arg( *args, uint64_t* ) = (uint64_t) i; break;
					case 'i': *va_arg( *args, sgs_Int* ) = i; break;
					case 'I': *va_arg( *args, int* ) = (int) i; break;
					}
				}
			}
			strict = 0; nowrite = 0; from++; break;
			
		case 'f': case 'd': case 'r':
			{
				sgs_Real r;
				
				if( opt && sgs_ItemType( C, from ) == SGS_VT_NULL )
				{
					if( !nowrite )
					{
						switch( *cmd )
						{
						case 'f': va_arg( *args, float* ); break;
						case 'd': va_arg( *args, double* ); break;
						case 'r': va_arg( *args, sgs_Real* ); break;
						}
					}
					break;
				}
				
				if( !sgs_ParseReal( C, from, &r ) ||
					( strict && sgs_ItemType( C, from ) != SGS_VT_REAL ) )
				{
					argerr( C, from, method, SGS_VT_REAL, strict );
					return opt;
				}
				
				if( !nowrite )
				{
					switch( *cmd )
					{
					case 'f': *va_arg( *args, float* ) = (float) r; break;
					case 'd': *va_arg( *args, double* ) = r; break;
					case 'r': *va_arg( *args, sgs_Real* ) = r; break;
					}
				}
			}
			strict = 0; nowrite = 0; from++; break;
			
		case 's': case 'm':
			{
				char* str;
				sgs_SizeVal sz;
				
				if( opt && sgs_ItemType( C, from ) == SGS_VT_NULL )
				{
					if( !nowrite )
					{
						va_arg( *args, char** );
						if( *cmd == 'm' )
							va_arg( *args, sgs_SizeVal* );
					}
					break;
				}
				
				if( ( strict && sgs_ItemType( C, from ) != SGS_VT_STRING ) ||
					!sgs_ParseString( C, from, &str, &sz ) )
				{
					argerr( C, from, method, SGS_VT_STRING, strict );
					return opt;
				}
				
				if( !nowrite )
				{
					*va_arg( *args, char** ) = str;
					if( *cmd == 'm' )
						*va_arg( *args, sgs_SizeVal* ) = sz;
				}
			}
			strict = 0; nowrite = 0; from++; break;
		
		case 'p':
			{
				if( opt && sgs_ItemType( C, from ) == SGS_VT_NULL )
				{
					if( !nowrite )
						va_arg( *args, SGSBOOL* );
					break;
				}
				
				if( !sgs_IsCallable( C, from ) )
				{
					argerrx( C, from, method, "callable", "" );
					return opt;
				}
				
				if( !nowrite )
				{
					*va_arg( *args, SGSBOOL* ) = 1;
				}
			}
			strict = 0; nowrite = 0; from++; break;
			
		case 'y':
			{
				if( opt && sgs_ItemType( C, from ) == SGS_VT_NULL )
				{
					if( !nowrite )
						*va_arg( *args, sgs_Context** ) = NULL;
					break;
				}
				
				if( sgs_ItemType( C, from ) != SGS_VT_THREAD )
				{
					argerrx( C, from, method, "thread", "" );
					return opt;
				}
				
				if( !nowrite )
				{
					*va_arg( *args, sgs_Context** ) = sgs_StackItem( C, from ).data.T;
				}
			}
			strict = 0; nowrite = 0; from++; break;
			
		case 'a': case 't': case 'h': case 'o': case '*':
			{
				sgs_ObjInterface* ifc = NULL;
				const char* ostr = "custom object";
				
				if( *cmd == 'a' ){ ifc = sgsstd_array_iface; ostr = "array"; }
				if( *cmd == 't' ){ ifc = sgsstd_dict_iface; ostr = "dict"; }
				if( *cmd == 'h' ){ ifc = sgsstd_map_iface; ostr = "map"; }
				if( *cmd == 'o' ) ifc = va_arg( *args, sgs_ObjInterface* );
				if( *cmd == '*' ) ostr = "object";
				
				if( opt && sgs_ItemType( C, from ) == SGS_VT_NULL )
				{
					if( !nowrite )
					{
						switch( *cmd )
						{
						case 'a': va_arg( *args, sgs_SizeVal* ); break;
						case 't':
						case 'h': va_arg( *args, sgs_SizeVal* ); break;
						case 'o': va_arg( *args, void** ); break;
						case '*': va_arg( *args, sgs_ObjInterface** ); break;
						}
					}
					break;
				}
				
				if( sgs_ItemType( C, from ) != SGS_VT_OBJECT ||
					( ifc != NULL && !sgs_IsObject( C, from, ifc ) ) )
				{
					argerrx( C, from, method, ostr, "" );
					return opt;
				}
				
				if( !nowrite )
				{
					sgs_VarObj* O = sgs_GetObjectStruct( C, from );
					switch( *cmd )
					{
					case 'a': *va_arg( *args, sgs_SizeVal* ) = 
						((sgsstd_array_header_t*) O->data)->size; break;
					case 't':
					case 'h': *va_arg( *args, sgs_SizeVal* ) =
						sgs_vht_size( ((sgs_VHTable*) O->data) ); break;
					case 'o': *va_arg( *args, void** ) = O->data; break;
					case '*': *va_arg( *args, sgs_ObjInterface** ) = O->iface; break;
					}
				}
			}
			strict = 0; nowrite = 0; from++; break;
			
		case '&':
			{
				void* ptr;
				
				if( opt && sgs_ItemType( C, from ) == SGS_VT_NULL )
				{
					if( !nowrite )
						va_arg( *args, void** );
					break;
				}
				
				if( !sgs_ParsePtr( C, from, &ptr ) )
				{
					argerrx( C, from, method, "pointer", "" );
					return opt;
				}
				
				if( !nowrite )
				{
					*va_arg( *args, void** ) = ptr;
				}
			}
			strict = 0; nowrite = 0; from++; break;
			
		case 'v':
			{
				if( opt && sgs_ItemType( C, from ) == SGS_VT_NULL )
				{
					if( !nowrite )
						va_arg( *args, sgs_Variable* );
					break;
				}
				
				if( from >= sgs_StackSize( C ) ||
					( strict && sgs_ItemType( C, from ) == SGS_VT_NULL ) )
				{
					argerrx( C, from, method, strict ? "non-null" : "any", "" );
					return opt;
				}
				
				if( !nowrite )
				{
					*va_arg( *args, sgs_Variable* ) = sgs_StackItem( C, from );
				}
			}
			strict = 0; nowrite = 0; from++; break;
			
		case 'x':
			{
				sgs_ArgCheckFunc acf = va_arg( *args, sgs_ArgCheckFunc );
				int flags = 0;
				
				if( strict )     flags |= SGS_LOADARG_STRICT;
				if( !nowrite )   flags |= SGS_LOADARG_WRITE;
				if( opt )        flags |= SGS_LOADARG_OPTIONAL;
				if( isig )       flags |= SGS_LOADARG_INTSIGN;
				if( range > 0 )  flags |= SGS_LOADARG_INTRANGE;
				if( range < 0 )  flags |= SGS_LOADARG_INTCLAMP;
				
				if( !acf( C, from, args, flags ) )
				{
					return opt;
				}
			}
			strict = 0; nowrite = 0; from++; break;
			
		case ' ': case '\t': case '\n': case '\r':
			break;
			
		default:
			return sgs_Msg( C, SGS_APIERR, "sgs_LoadArgs*: unrecognized character" );
			
		}
		if( opt && from >= sgs_StackSize( C ) && cmd[1] != '<' )
			break;
		cmd++;
	}
	return 1;
}

SGSBOOL sgs_LoadArgsExt( SGS_CTX, int from, const char* cmd, ... )
{
	SGSBOOL ret;
	va_list args;
	va_start( args, cmd );
	ret = sgs_LoadArgsExtVA( C, from, cmd, &args );
	va_end( args );
	return ret;
}

SGSBOOL sgs_LoadArgs( SGS_CTX, const char* cmd, ... )
{
	SGSBOOL ret;
	va_list args;
	va_start( args, cmd );
	ret = sgs_LoadArgsExtVA( C, 0, cmd, &args );
	va_end( args );
	return ret;
}

SGSBOOL sgs_ParseMethod( SGS_CTX, sgs_ObjInterface* iface, void** ptrout, const char* name )
{
	int method_call = sgs_Method( C );
	SGSFN( name );
	if( !sgs_IsObject( C, 0, iface ) )
	{
		sgs_ArgErrorExt( C, 0, method_call, iface->name, "" );
		return SGS_FALSE;
	}
	*ptrout = p_objdata( stk_poff( C, 0 ) );
	sgs_ForceHideThis( C );
	return SGS_TRUE;
}


int sgs_ArgCheck_Object( SGS_CTX, int argid, va_list* args, int flags )
{
	uint32_t ity;
	sgs_VarObj** out = NULL;
	if( flags & SGS_LOADARG_WRITE )
		out = va_arg( *args, sgs_VarObj** );
	
	ity = sgs_ItemType( C, argid );
	if( ity == SGS_VT_OBJECT || ( !( flags & SGS_LOADARG_STRICT ) && ity == SGS_VT_NULL ) )
	{
		if( out )
			*out = ity != SGS_VT_NULL ? sgs_GetObjectStruct( C, argid ) : NULL;
		return 1;
	}
	if( flags & SGS_LOADARG_OPTIONAL )
		return 1;
	return sgs_ArgError( C, argid, 0, SGS_VT_OBJECT, !!( flags & SGS_LOADARG_STRICT ) );
}


void sgs_Pop( SGS_CTX, StkIdx count )
{
	if( SGS_STACKFRAMESIZE < count || count < 0 )
	{
		sgs_Msg( C, SGS_APIERR, "sgs_Pop: invalid count - %d (stack size = %d)",
			(int) count, (int) SGS_STACKFRAMESIZE );
		return;
	}
	fstk_pop( C, count );
}

void sgs_PopSkip( SGS_CTX, StkIdx count, StkIdx skip )
{
	if( SGS_STACKFRAMESIZE < count + skip || count < 0 || skip < 0 )
	{
		sgs_Msg( C, SGS_APIERR, "sgs_PopSkip: invalid counts - skip:%d, pop:%d (stack size = %d)",
			(int) skip, (int) count, (int) SGS_STACKFRAMESIZE );
		return;
	}
	stk_popskip( C, count, skip );
}

StkIdx sgs_StackSize( SGS_CTX )
{
	return SGS_STACKFRAMESIZE;
}

void sgs_SetStackSize( SGS_CTX, StkIdx size )
{
	if( size < 0 )
	{
		sgs_Msg( C, SGS_APIERR, "sgs_SetStackSize: size (%d) cannot be negative", (int) size );
		return;
	}
	stk_setsize( C, size );
}

void sgs_SetDeltaSize( SGS_CTX, sgs_StkIdx diff )
{
	StkIdx tgtsize = SGS_STACKFRAMESIZE + diff;
	if( tgtsize < 0 )
	{
		sgs_Msg( C, SGS_APIERR, "sgs_SetDeltaSize: resulting size (%d) "
			"cannot be negative (diff = %d)", (int) tgtsize, (int) diff );
		return;
	}
	stk_deltasize( C, diff );
}

SGSRESULT sgs_AdjustStack( SGS_CTX, int expected, int ret )
{
	if( SGS_SUCCEEDED( ret ) )
		stk_deltasize( C, expected - ret );
	return ret;
}

StkIdx sgs_AbsIndex( SGS_CTX, StkIdx item )
{
	return stk_absindex( C, item );
}

SGSBOOL sgs_IsValidIndex( SGS_CTX, StkIdx item )
{
	item = stk_absindex( C, item );
	return ( item >= 0 && item < SGS_STACKFRAMESIZE );
}

sgs_Variable sgs_OptStackItem( SGS_CTX, sgs_StkIdx item )
{
	if( sgs_IsValidIndex( C, item ) == 0 )
	{
		sgs_Variable var;
		var.type = SGS_VT_NULL;
		return var;
	}
	return *stk_getpos( C, item );
}

sgs_Variable sgs_StackItem( SGS_CTX, sgs_StkIdx item )
{
	if( sgs_IsValidIndex( C, item ) == 0 )
	{
		sgs_Variable var;
		sgs_StackIdxError( C, item );
		var.type = SGS_VT_NULL;
		return var;
	}
	return *stk_getpos( C, item );
}

SGSBOOL sgs_GetStackItem( SGS_CTX, sgs_StkIdx item, sgs_Variable* out )
{
	if( !sgs_IsValidIndex( C, item ) )
	{
		out->type = SGS_VT_NULL;
		return SGS_FALSE;
	}
	*out = *stk_getpos( C, item );
	VAR_ACQUIRE( out );
	return SGS_TRUE;
}

uint32_t sgs_ItemType( SGS_CTX, StkIdx item )
{
	if( !sgs_IsValidIndex( C, item ) )
		return 0;
	return stk_getpos( C, item )->type;
}


void sgs_Assign( SGS_CTX, sgs_Variable* var_to, sgs_Variable* var_from )
{
	if( var_to == var_from )
		return;
	VAR_RELEASE( var_to );
	*var_to = *var_from;
	VAR_ACQUIRE( var_to );
}

void sgs_ArithOp( SGS_CTX, sgs_Variable* out, sgs_Variable* A, sgs_Variable* B, int op )
{
	if( op == SGS_EOP_NEGATE )
	{
		vm_op_negate( C, out, A ); /* throws an error */
		return;
	}
	if( op < SGS_EOP_ADD || op > SGS_EOP_MOD )
	{
		sgs_Msg( C, SGS_APIERR, "sgs_ArithOp: invalid operation ID (%d) "
			"specified (add/sub/mul/div/mod/negate allowed)", op );
		VAR_RELEASE( out ); /* out = null */
		return;
	}
	vm_arith_op( C, out, A, B, op ); /* throws an error */
}

void sgs_IncDec( SGS_CTX, sgs_Variable* out, sgs_Variable* A, int inc )
{
	vm_op_incdec( C, out, A, inc ? 1 : -1 );
}


/*

	OPERATIONS

*/

int sgs_XFCall( SGS_CTX, int args, int gotthis )
{
	int rvc = 0;
	if( args < 0 )
	{
		sgs_Msg( C, SGS_APIERR, "sgs_XFCall: negative argument count (%d)", args );
		return 0;
	}
	if( SGS_STACKFRAMESIZE < args + ( gotthis ? 2 : 1 ) )
	{
		sgs_Msg( C, SGS_APIERR, "sgs_XFCall: not enough items in stack (need: %d, got: %d)",
			args + ( gotthis ? 2 : 1 ), (int) SGS_STACKFRAMESIZE );
		return 0;
	}
	/* fix up closure */
	{
		sgs_Variable out, *vp = stk_ptop( C, - args - ( gotthis ? 2 : 1 ) );
		if( vp->type == SGS_VT_FUNC && vp->data.F->numclsr != 0 )
		{
			int i;
			sgs_Closure** clsrlist = sgsSTD_MakeClosure( C, &out, vp, vp->data.F->numclsr );
			for( i = 0; i < vp->data.F->numclsr; ++i )
			{
				sgs_Closure* nc = sgs_Alloc( sgs_Closure );
				nc->refcount = 1;
				nc->var.type = SGS_VT_NULL;
				*clsrlist++ = nc;
			}
			p_setvar_leave( vp, &out );
		}
	}
	SGS_PERFEVENT( ps_precall_begin( args, gotthis, CALLSRC_NATIVE ) );
	vm_call( C, args, gotthis, &rvc, 0 );
	return rvc;
}

SGSBOOL sgs_GlobalCall( SGS_CTX, const char* name, int args, int expect )
{
	sgs_Variable v_func;
	if( !sgs_GetGlobalByName( C, name, &v_func ) )
		return SGS_FALSE;
	sgs_InsertVariable( C, -args - 1, v_func );
	sgs_Release( C, &v_func );
	sgs_Call( C, args, expect );
	return SGS_TRUE;
}

void sgs_TypeOf( SGS_CTX, sgs_Variable var )
{
	const char* ty = "ERROR";
	
	switch( var.type )
	{
	case SGS_VT_NULL:   ty = "null"; break;
	case SGS_VT_BOOL:   ty = "bool"; break;
	case SGS_VT_INT:    ty = "int"; break;
	case SGS_VT_REAL:   ty = "real"; break;
	case SGS_VT_STRING: ty = "string"; break;
	case SGS_VT_CFUNC:  ty = "cfunc"; break;
	case SGS_VT_FUNC:   ty = "func"; break;
	case SGS_VT_OBJECT:
		{
			sgs_VarObj* O = var.data.O;
			if( O->mm_enable && _push_metamethod( C, O, "__typeof" ) )
			{
				sgs_SizeVal ssz = SGS_STACKFRAMESIZE - 1;
				sgs_PushObjectPtr( C, O );
				if( sgs_XThisCall( C, 0 ) > 0 && stk_gettop( C )->type == SGS_VT_STRING )
				{
					stk_downsize_keep( C, ssz, 1 );
					return;
				}
				stk_downsize( C, ssz );
			}
			ty = O->iface->name ? O->iface->name : "object";
		}
		break;
	case SGS_VT_PTR:    ty = "pointer"; break;
	case SGS_VT_THREAD: ty = "thread"; break;
	}
	
	sgs_PushString( C, ty );
}

void sgs_DumpVar( SGS_CTX, sgs_Variable var, int maxdepth )
{
	if( maxdepth <= 0 )
	{
		sgs_PushString( C, "..." );
		return;
	}
	
	switch( var.type )
	{
	case SGS_VT_NULL: sgs_PushString( C, "null" ); break;
	case SGS_VT_BOOL: sgs_PushString( C, var.data.B ? "bool (true)" : "bool (false)" ); break;
	case SGS_VT_INT: { char buf[ 32 ];
		sprintf( buf, "int (%" PRId64 ")", var.data.I );
		sgs_PushString( C, buf ); } break;
	case SGS_VT_REAL: { char buf[ 32 ];
		sprintf( buf, "real (%g)", var.data.R );
		sgs_PushString( C, buf ); } break;
	case SGS_VT_STRING:
		{
			char buf[ 532 ];
			char* bptr = buf;
			char* bend = buf + 512;
			char* source = sgs_var_cstr( &var );
			uint32_t len = var.data.S->size;
			char* srcend = source + len;
			sprintf( buf, "string [%" PRId32"] \"", len );
			bptr += strlen( buf );
			while( source < srcend && bptr < bend )
			{
				if( *source == ' ' || sgs_isgraph( (int)*source ) )
					*bptr++ = *source++;
				else
				{
					static const char* hexdigs = "0123456789ABCDEF";
					bptr[ 0 ] = '\\';
					bptr[ 1 ] = 'x';
					bptr[ 2 ] = hexdigs[ (*source & 0xf0) >> 4 ];
					bptr[ 3 ] = hexdigs[ *source & 0xf ];
					bptr += 4;
					source++;
				}
			}
			if( source != srcend )
			{
				*bptr++ = '.';
				*bptr++ = '.';
				*bptr++ = '.';
			}
			*bptr++ = '"';
			/* WP: string limit */
			sgs_PushStringBuf( C, buf, (sgs_SizeVal) ( bptr - buf ) );
		}
		break;
	case SGS_VT_FUNC:
		{
			sgs_MemBuf mb = sgs_membuf_create();
			sgs_iFunc* F = var.data.F;
			
			const char* str1 = F->sfuncname->size ? "SGS function '" : "SGS function <anonymous>";
			const char* str2 = F->sfuncname->size ? "' defined at " : " defined at ";
			const char* str3 = "'";
			
			sgs_membuf_appbuf( &mb, C, str1, strlen(str1) );
			if( F->sfuncname->size )
				sgs_membuf_appbuf( &mb, C, sgs_str_cstr( F->sfuncname ), F->sfuncname->size );
			if( F->sfilename->size )
			{
				char lnbuf[ 32 ];
				sgs_membuf_appbuf( &mb, C, str2, strlen(str2) );
				sgs_membuf_appbuf( &mb, C, sgs_str_cstr( F->sfilename ), F->sfilename->size );
				sprintf( lnbuf, ":%d", (int) F->linenum );
				sgs_membuf_appbuf( &mb, C, lnbuf, strlen(lnbuf) );
			}
			else if( F->sfuncname->size )
				sgs_membuf_appbuf( &mb, C, str3, strlen(str3) );
			
			/* WP: various limits */
			sgs_PushStringBuf( C, mb.ptr, (sgs_SizeVal) mb.size );
			sgs_membuf_destroy( &mb, C );
		}
		break;
	case SGS_VT_CFUNC:
		{
			char buf[ 32 ];
			sprintf( buf, "C function (%p)", var.data.C );
			sgs_PushString( C, buf );
		}
		break;
	case SGS_VT_OBJECT:
		{
			char buf[ 256 ];
			int q = 0;
			_STACK_PREPARE;
			sgs_VarObj* obj = var.data.O;
			
			if( obj->iface->dump )
			{
				_STACK_PROTECT;
				q = SGS_SUCCEEDED( obj->iface->dump( C, obj, maxdepth - 1 ) );
				_STACK_UNPROTECT_SKIP( q );
			}
			
			if( !q )
			{
				snprintf( buf, 255, "object (%p) [%" PRId32"] %s", (void*) obj, obj->refcount,
					obj->iface->name ? obj->iface->name : "<unnamed>" );
				buf[ 255 ] = 0;
				sgs_PushString( C, buf );
			}
			else
				sgs_ToString( C, -1 );
		}
		break;
	case SGS_VT_PTR:
		{
			char buf[ 32 ];
			sprintf( buf, "pointer (%p)", var.data.P );
			sgs_PushString( C, buf );
		}
		break;
	case SGS_VT_THREAD:
		{
			char buf[ 32 ];
			sprintf( buf, "thread (%p)", var.data.T );
			sgs_PushString( C, buf );
		}
		break;
	default:
		{
			char buf[ 32 ];
			sprintf( buf, "<invalid-type> (%d)", (int) var.type );
			sgs_BreakIf( "Invalid variable type in sgs_DumpVar!" );
			sgs_PushString( C, buf );
		}
		break;
	}
}

static void sgsVM_GCExecute( SGS_SHCTX )
{
	sgs_Variable *vbeg, *vend;
	sgs_VarObj* p;
	
	S->redblue = !S->redblue;
	S->gcrun = SGS_TRUE;
	
	SGS_CTX = S->state_list;
	/* shared state */
	/* - interfaces */
	if( S->array_iface )
	{
		obj_gcmark( S, S->array_iface );
	}
	/* - registry */
	sgsSTD_RegistryGC( S );
	
	while( C )
	{
		/* -- MARK -- */
		/* STACK */
		vbeg = C->stack_base; vend = C->stack_top;
		while( vbeg < vend )
		{
			vm_gcmark( S, vbeg );
			vbeg++;
		}
		/* FRAMES */
		{
			sgs_StackFrame* sf = C->sf_first;
			while( sf )
			{
				if( sf->clsrref )
					obj_gcmark( S, sf->clsrref );
				sf = sf->next;
			}
		}
		
		/* GLOBALS */
		sgsSTD_GlobalGC( C );
		/* THREADS */
		sgsSTD_ThreadsGC( C );
		
		C = C->next;
	}
	
	/* -- SWEEP -- */
	C = S->state_list; // any context is good enough here
	C->refcount++;
	/* destruct objects */
	p = S->objs;
	while( p ){
		sgs_VarObj* pn = p->next;
		if( p->redblue != S->redblue )
			var_destruct_object( C, p );
		p = pn;
	}
	
	/* free variables */
	p = S->objs;
	while( p ){
		sgs_VarObj* pn = p->next;
		if( p->redblue != S->redblue )
			var_free_object( C, p );
		p = pn;
	}
	
	C->refcount--;
	S->gcrun = SGS_FALSE;
}

void sgs_GCExecute( SGS_CTX )
{
	SGS_SHCTX_USE;
	sgsVM_GCExecute( S );
}


const char* sgs_DebugDumpVarExt( SGS_CTX, sgs_Variable var, int maxdepth )
{
	if( maxdepth < 0 )
	{
		fstk_push( C, &var );
		return sgs_ToString( C, -1 );
	}
	else
	{
		sgs_DumpVar( C, var, maxdepth );
		return sgs_GetStringPtr( C, -1 );
	}
}


void sgs_PadString( SGS_CTX )
{
	const char* padding = "  ";
	const uint32_t padsize = 2;
	
	if( sgs_StackSize( C ) < 1 )
	{
		sgs_Msg( C, SGS_APIERR, "sgs_PadString: stack is empty" );
		return;
	}
	else
	{
		uint32_t i, allocsize;
		char* ostr, *ostre;
		const char* cstr;
		sgs_Variable* var = stk_getpos( C, -1 );
		if( var->type != SGS_VT_STRING )
		{
			sgs_Msg( C, SGS_APIERR, "sgs_PadString: need string at top of stack" );
			return;
		}
		cstr = sgs_var_cstr( var );
		for( i = 0; cstr[ i ]; )
			if( cstr[ i ] == '\n' ) i++; else cstr++;
		allocsize = var->data.S->size + i * padsize;
		if( allocsize > 0x7fffffff )
			allocsize = 0x7fffffff;
		
		/* WP: unimportant */
		sgs_PushStringAlloc( C, (sgs_SizeVal) allocsize );
		cstr = sgs_var_cstr( stk_getpos( C, -2 ) );
		ostr = sgs_var_cstr( stk_getpos( C, -1 ) );
		ostre = ostr + allocsize;
		while( *cstr && ostr < ostre )
		{
			*ostr++ = *cstr;
			if( *cstr == '\n' )
			{
				const char* ppd = padding;
				while( *ppd && ostr < ostre )
					*ostr++ = *ppd++;
			}
			cstr++;
		}
		sgs_PopSkip( C, 1, 1 );
		sgs_FinalizeStringAlloc( C, -1 );
	}
}

void sgs_ToPrintSafeString( SGS_CTX )
{
	sgs_MemBuf mb = sgs_membuf_create();
	char* buf = NULL;
	sgs_SizeVal size = 0, i;
	if( sgs_StackSize( C ) < 1 )
	{
		sgs_Msg( C, SGS_APIERR, "sgs_PadString: stack is empty" );
		return;
	}
	buf = sgs_ToStringBuf( C, -1, &size );
	for( i = 0; i < size; ++i )
	{
		if( sgs_isgraph( (int)buf[ i ] ) || buf[ i ] == ' ' )
			sgs_membuf_appchr( &mb, C, buf[ i ] );
		else
		{
			char chrs[32];
			sprintf( chrs, "\\x%02X", (int) (unsigned char) buf[ i ] );
			sgs_membuf_appbuf( &mb, C, chrs, 4 );
		}
	}
	sgs_Pop( C, 1 );
	if( mb.size > 0x7fffffff )
		mb.size = 0x7fffffff;
	/* WP: unimportant */
	sgs_PushStringBuf( C, mb.ptr, (sgs_SizeVal) mb.size );
	sgs_membuf_destroy( &mb, C );
}

void sgs_StringConcat( SGS_CTX, StkIdx args )
{
	if( vm_op_concat_ex( C, args ) == SGS_FALSE )
	{
		sgs_Msg( C, SGS_APIERR, "sgs_StringConcat: not enough items on stack (need: %d, have: %d)",
			(int) args, (int) SGS_STACKFRAMESIZE );
	}
}

void sgs_CloneItem( SGS_CTX, sgs_Variable var )
{
	if( vm_clone( C, &var ) == SGS_FALSE )
		fstk_push_null( C );
}

int sgs_Compare( SGS_CTX, sgs_Variable* v1, sgs_Variable* v2 )
{
	return vm_compare( C, v1, v2 );
}

SGSBOOL sgs_EqualTypes( sgs_Variable* v1, sgs_Variable* v2 )
{
	return v1->type == v2->type && ( v1->type != SGS_VT_OBJECT
		|| v1->data.O->iface == v2->data.O->iface );
}


/*
	
	CONVERSION / RETRIEVAL
	
*/
/* pointer versions */
sgs_Bool sgs_GetBoolP( SGS_CTX, sgs_Variable* var ){ return var_getbool( C, var ); }
sgs_Int sgs_GetIntP( SGS_CTX, sgs_Variable* var ){ return var_getint( var ); }
sgs_Real sgs_GetRealP( SGS_CTX, sgs_Variable* var ){ return var_getreal( var ); }
void* sgs_GetPtrP( SGS_CTX, sgs_Variable* var ){ return var_getptr( var ); }

char* sgs_ToStringBufP( SGS_CTX, sgs_Variable* var, sgs_SizeVal* outsize )
{
	vm_convert_string( C, var );
	if( outsize )
		/* WP: string limit */
		*outsize = (sgs_SizeVal) var->data.S->size;
	return sgs_var_cstr( var );
}

char* sgs_ToStringBufFastP( SGS_CTX, sgs_Variable* var, sgs_SizeVal* outsize )
{
	if( var->type == SGS_VT_OBJECT )
	{
		sgs_TypeOf( C, *var );
		sgs_StoreVariable( C, var );
	}
	return sgs_ToStringBufP( C, var, outsize );
}


SGSBOOL sgs_IsObjectP( sgs_Variable* var, sgs_ObjInterface* iface )
{
	return var->type == SGS_VT_OBJECT && var->data.O->iface == iface;
}

SGSBOOL sgs_IsCallableP( sgs_Variable* var )
{
	uint32_t ty = var->type;
	if( ty == SGS_VT_FUNC || ty == SGS_VT_CFUNC )
		return 1;
	if( ty == SGS_VT_OBJECT && var->data.O->iface->call )
		return 1;
	return 0;
}

#define SGS_IS_SYSTEM_TYPE( ty ) ((ty) == SGS_VT_NULL || (ty) == SGS_VT_FUNC || \
	(ty) == SGS_VT_CFUNC || (ty) == SGS_VT_PTR || (ty) == SGS_VT_THREAD)

SGSBOOL sgs_ParseBoolP( SGS_CTX, sgs_Variable* var, sgs_Bool* out )
{
	uint32_t ty = var->type;
	if( SGS_IS_SYSTEM_TYPE( ty ) || ty == SGS_VT_STRING )
		return SGS_FALSE;
	if( out )
		*out = sgs_GetBoolP( C, var );
	return SGS_TRUE;
}

SGSBOOL sgs_ParseIntP( SGS_CTX, sgs_Variable* var, sgs_Int* out )
{
	sgs_Int i;
	if( SGS_IS_SYSTEM_TYPE( var->type ) )
		return SGS_FALSE;
	if( var->type == SGS_VT_STRING )
	{
		intreal_t OIR;
		const char* ostr = sgs_var_cstr( var );
		const char* str = ostr;
		int res = sgs_util_strtonum( &str, str + var->data.S->size, &OIR.i, &OIR.r );
		
		if( str == ostr )    return SGS_FALSE;
		if( res == 1 )       i = OIR.i;
		else if( res == 2 )  i = (sgs_Int) OIR.r;
		else                 return SGS_FALSE;
	}
	else
		i = sgs_GetIntP( C, var );
	if( out )
		*out = i;
	return SGS_TRUE;
}

SGSBOOL sgs_ParseRealP( SGS_CTX, sgs_Variable* var, sgs_Real* out )
{
	sgs_Real r;
	if( SGS_IS_SYSTEM_TYPE( var->type ) )
		return SGS_FALSE;
	if( var->type == SGS_VT_STRING )
	{
		intreal_t OIR;
		const char* ostr = sgs_var_cstr( var );
		const char* str = ostr;
		int res = sgs_util_strtonum( &str, str + var->data.S->size, &OIR.i, &OIR.r );
		
		if( str == ostr )    return SGS_FALSE;
		if( res == 1 )       r = (sgs_Real) OIR.i;
		else if( res == 2 )  r = OIR.r;
		else                 return SGS_FALSE;
	}
	else
		r = sgs_GetRealP( C, var );
	if( out )
		*out = r;
	return SGS_TRUE;
}

SGSBOOL sgs_ParseStringP( SGS_CTX, sgs_Variable* var, char** out, sgs_SizeVal* size )
{
	char* str;
	uint32_t ty;
	ty = var->type;
	if( SGS_IS_SYSTEM_TYPE( ty ) )
		return SGS_FALSE;
	str = sgs_ToStringBufP( C, var, size );
	if( out )
		*out = str;
	return str != NULL;
}

SGSBOOL sgs_ParseObjectPtrP( SGS_CTX, sgs_Variable* var,
	sgs_ObjInterface* iface, sgs_VarObj** out, int strict )
{
	if( !strict && var->type == SGS_VT_NULL )
	{
		if( *out )
			sgs_ObjRelease( C, *out );
		*out = NULL;
		return SGS_TRUE;
	}
	if( sgs_IsObjectP( var, iface ) )
	{
		if( *out )
			sgs_ObjRelease( C, *out );
		*out = sgs_GetObjectStructP( var );
		sgs_ObjAcquire( C, *out );
		return SGS_TRUE;
	}
	return SGS_FALSE;
}

SGSBOOL sgs_ParsePtrP( SGS_CTX, sgs_Variable* var, void** out )
{
	if( var->type != SGS_VT_NULL && var->type != SGS_VT_PTR )
		return SGS_FALSE;
	if( out )
		*out = var->type != SGS_VT_NULL ? var->data.P : NULL;
	return SGS_TRUE;
}


/* index versions */
sgs_Bool sgs_GetBool( SGS_CTX, StkIdx item )
{
	sgs_Variable var = sgs_OptStackItem( C, item );
	return var_getbool( C, &var );
}

sgs_Int sgs_GetInt( SGS_CTX, StkIdx item )
{
	sgs_Variable var = sgs_OptStackItem( C, item );
	return var_getint( &var );
}

sgs_Real sgs_GetReal( SGS_CTX, StkIdx item )
{
	sgs_Variable var = sgs_OptStackItem( C, item );
	return var_getreal( &var );
}

void* sgs_GetPtr( SGS_CTX, StkIdx item )
{
	sgs_Variable var = sgs_OptStackItem( C, item );
	return var_getptr( &var );
}


char* sgs_ToStringBuf( SGS_CTX, StkIdx item, sgs_SizeVal* outsize )
{
	sgs_Variable* var;
	if( !sgs_IsValidIndex( C, item ) )
		return NULL;
	vm_convert_stack_string( C, item );
	var = stk_getpos( C, item );
	if( outsize )
		/* WP: string limit */
		*outsize = (sgs_SizeVal) var->data.S->size;
	return sgs_var_cstr( var );
}

char* sgs_ToStringBufFast( SGS_CTX, StkIdx item, sgs_SizeVal* outsize )
{
	if( !sgs_IsValidIndex( C, item ) )
		return NULL;
	item = stk_absindex( C, item );
	if( stk_getpos( C, item )->type == SGS_VT_OBJECT )
	{
		sgs_TypeOf( C, *stk_getpos( C, item ) );
		sgs_SetStackItem( C, item, *stk_gettop( C ) );
		sgs_Pop( C, 1 );
	}
	return sgs_ToStringBuf( C, item, outsize );
}


SGSBOOL sgs_IsObject( SGS_CTX, StkIdx item, sgs_ObjInterface* iface )
{
	sgs_Variable var = sgs_OptStackItem( C, item );
	return var.type == SGS_VT_OBJECT && var.data.O->iface == iface;
}

SGSBOOL sgs_IsCallable( SGS_CTX, StkIdx item )
{
	uint32_t ty = sgs_ItemType( C, item );
	
	if( ty == SGS_VT_FUNC || ty == SGS_VT_CFUNC )
		return 1;
	if( ty == SGS_VT_OBJECT && sgs_GetObjectIface( C, item )->call )
		return 1;
	return 0;
}

SGSBOOL sgs_ParseBool( SGS_CTX, StkIdx item, sgs_Bool* out )
{
	sgs_Variable var = sgs_OptStackItem( C, item );
	return sgs_ParseBoolP( C, &var, out );
}

SGSBOOL sgs_ParseInt( SGS_CTX, StkIdx item, sgs_Int* out )
{
	sgs_Variable var = sgs_OptStackItem( C, item );
	return sgs_ParseIntP( C, &var, out );
}

SGSBOOL sgs_ParseReal( SGS_CTX, StkIdx item, sgs_Real* out )
{
	sgs_Variable var = sgs_OptStackItem( C, item );
	return sgs_ParseRealP( C, &var, out );
}

SGSBOOL sgs_ParseString( SGS_CTX, StkIdx item, char** out, sgs_SizeVal* size )
{
	char* str;
	uint32_t ty;
	if( !sgs_IsValidIndex( C, item ) )
		return SGS_FALSE;
	ty = sgs_ItemType( C, item );
	if( ty == SGS_VT_NULL || ty == SGS_VT_FUNC || ty == SGS_VT_CFUNC )
		return SGS_FALSE;
	str = sgs_ToStringBuf( C, item, size );
	if( out )
		*out = str;
	return str != NULL;
}

SGSBOOL sgs_ParseObjectPtr( SGS_CTX, sgs_StkIdx item,
	sgs_ObjInterface* iface, sgs_VarObj** out, int strict )
{
	sgs_Variable var = sgs_OptStackItem( C, item );
	return sgs_ParseObjectPtrP( C, &var, iface, out, strict );
}

SGSBOOL sgs_ParsePtr( SGS_CTX, StkIdx item, void** out )
{
	sgs_Variable var = sgs_OptStackItem( C, item );
	return sgs_ParsePtrP( C, &var, out );
}


sgs_Bool sgs_GlobalBool( SGS_CTX, const char* name )
{
	sgs_Bool v;
	sgs_PushGlobalByName( C, name );
	v = var_getbool( C, stk_gettop( C ) );
	fstk_pop1( C );
	return v;
}

sgs_Int sgs_GlobalInt( SGS_CTX, const char* name )
{
	sgs_Int v;
	sgs_PushGlobalByName( C, name );
	v = var_getint( stk_gettop( C ) );
	fstk_pop1( C );
	return v;
}

sgs_Real sgs_GlobalReal( SGS_CTX, const char* name )
{
	sgs_Real v;
	sgs_PushGlobalByName( C, name );
	v = var_getreal( stk_gettop( C ) );
	fstk_pop1( C );
	return v;
}

char* sgs_GlobalStringBuf( SGS_CTX, const char* name, sgs_SizeVal* outsize )
{
	char* buf;
	sgs_PushGlobalByName( C, name );
	if( !sgs_ParseString( C, -1, &buf, outsize ) )
	{
		fstk_pop1( C );
		return NULL;
	}
	return buf;
}


SGSBOOL sgs_PushIterator( SGS_CTX, sgs_Variable var )
{
	fstk_push_null( C );
	return vm_forprep( C, stk_absindex( C, -1 ), &var );
}

SGSBOOL sgs_GetIterator( SGS_CTX, sgs_Variable var, sgs_Variable* out )
{
	int ret;
	fstk_push_null( C );
	ret = vm_forprep( C, stk_absindex( C, -1 ), &var );
	sgs_StoreVariable( C, out );
	return ret;
}

SGSBOOL sgs_IterAdvance( SGS_CTX, sgs_Variable var )
{
	return vm_fornext( C, -1, -1, &var );
}

void sgs_IterPushData( SGS_CTX, sgs_Variable var, int key, int value )
{
	StkIdx idkey, idval;
	if( !key && !value )
		return;
	if( key )
	{
		idkey = stk_size( C );
		fstk_push_null( C );
	}
	else idkey = -1;
	if( value )
	{
		idval = stk_size( C );
		fstk_push_null( C );
	}
	else idval = -1;
	vm_fornext( C, idkey, idval, &var );
}

void sgs_IterGetData( SGS_CTX, sgs_Variable var, sgs_Variable* key, sgs_Variable* value )
{
	if( !key && !value )
		return;
	if( key ) fstk_push_null( C );
	if( value ) fstk_push_null( C );
	vm_fornext( C, key ? stk_topindex( C, value ? -2 : -1 ) : -1, value ? stk_topindex( C, -1 ) : -1, &var );
	if( value ) sgs_StoreVariable( C, value );
	if( key ) sgs_StoreVariable( C, key );
}


SGSBOOL sgs_IsNumericString( const char* str, sgs_SizeVal size )
{
	intreal_t out;
	const char* ostr = str;
	return sgs_util_strtonum( &str, str + size, &out.i, &out.r ) != 0 && str != ostr;
}


/*

	EXTENSION UTILITIES

*/

SGSBOOL sgs_Method( SGS_CTX )
{
	if( C->sf_last && SGS_HAS_FLAG( C->sf_last->flags, SGS_SF_METHOD ) &&
		!SGS_HAS_FLAG( C->sf_last->flags, SGS_SF_HASTHIS ) )
	{
		C->stack_off--;
		C->sf_last->flags |= SGS_SF_HASTHIS;
		return SGS_TRUE;
	}
	return SGS_FALSE;
}

SGSBOOL sgs_HideThis( SGS_CTX )
{
	if( C->sf_last && SGS_HAS_FLAG( C->sf_last->flags, (SGS_SF_METHOD|SGS_SF_HASTHIS) ) )
	{
		C->stack_off++;
		/* WP: implicit conversion to int */
		C->sf_last->flags &= (uint8_t) ~SGS_SF_HASTHIS;
		return SGS_TRUE;
	}
	return SGS_FALSE;
}

SGSBOOL sgs_ForceHideThis( SGS_CTX )
{
	if( !C->sf_last )
		return SGS_FALSE;
	if( SGS_HAS_FLAG( C->sf_last->flags, SGS_SF_METHOD ) )
		return sgs_HideThis( C );
	if( SGS_STACKFRAMESIZE < 1 )
		return SGS_FALSE;
	C->stack_off++;
	C->sf_last->flags = ( C->sf_last->flags | SGS_SF_METHOD ) & (uint8_t) (~SGS_SF_HASTHIS);
	return SGS_TRUE;
}

int sgs_ObjectArg( SGS_CTX )
{
	return C->object_arg;
}


void sgs_Acquire( SGS_CTX, sgs_Variable* var )
{
	SGS_UNUSED( C );
	VAR_ACQUIRE( var );
}

void sgs_AcquireArray( SGS_CTX, sgs_Variable* var, sgs_SizeVal count )
{
	sgs_Variable* vend = var + count;
	while( var < vend )
		sgs_Acquire( C, var++ );
}

void sgs_Release( SGS_CTX, sgs_Variable* var )
{
	VAR_RELEASE( var );
}

void sgs_ReleaseArray( SGS_CTX, sgs_Variable* var, sgs_SizeVal count )
{
	sgs_Variable* vend = var + count;
	while( var < vend )
		sgs_Release( C, var++ );
}

void sgs_GCMark( SGS_CTX, sgs_Variable* var )
{
	SGS_SHCTX_USE;
	vm_gcmark( S, var );
}

void sgs_GCMarkArray( SGS_CTX, sgs_Variable* var, sgs_SizeVal count )
{
	SGS_SHCTX_USE;
	sgs_Variable* vend = var + count;
	while( var < vend )
	{
		vm_gcmark( S, var );
		var++;
	}
}

void sgs_ObjAcquire( SGS_CTX, sgs_VarObj* obj )
{
	SGS_UNUSED( C );
	obj->refcount++;
}

void sgs_ObjRelease( SGS_CTX, sgs_VarObj* obj )
{
	if( --obj->refcount <= 0 && !C->shared->gcrun )
		sgsVM_VarDestroyObject( C, obj );
}

void sgs_ObjGCMark( SGS_CTX, sgs_VarObj* obj )
{
	SGS_SHCTX_USE;
	obj_gcmark( S, obj );
}

void sgs_ObjAssign( SGS_CTX, sgs_VarObj** dest, sgs_VarObj* src )
{
	if( src )
		sgs_ObjAcquire( C, src );
	if( *dest )
		sgs_ObjRelease( C, *dest );
	*dest = src;
}

void sgs_ObjCallDtor( SGS_CTX, sgs_VarObj* obj )
{
	var_destruct_object( C, obj );
}

void sgs_ObjSetMetaObj( SGS_CTX, sgs_VarObj* obj, sgs_VarObj* metaobj )
{
	sgs_VarObj* chk = metaobj;
	while( chk )
	{
		sgs_BreakIf( chk == obj && "sgs_ObjSetMetaObj: loop detected" );
		chk = chk->metaobj;
	}
	
	if( metaobj )
		sgs_ObjAcquire( C, metaobj );
	if( obj->metaobj )
		sgs_ObjRelease( C, obj->metaobj );
	obj->metaobj = metaobj;
}

sgs_VarObj* sgs_ObjGetMetaObj( sgs_VarObj* obj )
{
	return obj->metaobj;
}

void sgs_ObjSetMetaMethodEnable( sgs_VarObj* obj, SGSBOOL enable )
{
	obj->mm_enable = !!enable;
}

SGSBOOL sgs_ObjGetMetaMethodEnable( sgs_VarObj* obj )
{
	return obj->mm_enable;
}


#define DBLCHK( what, fval )\
	sgs_BreakIf( what );\
	if( what ) return fval;


#define _OBJPREP_P( ret ) \
	DBLCHK( var->type != SGS_VT_OBJECT, ret )

char* sgs_GetStringPtrP( sgs_Variable* var )
{
	DBLCHK( var->type != SGS_VT_STRING, NULL )
	return sgs_var_cstr( var );
}

sgs_SizeVal sgs_GetStringSizeP( sgs_Variable* var )
{
	DBLCHK( var->type != SGS_VT_STRING, 0 )
	/* WP: string limit */
	return (sgs_SizeVal) var->data.S->size;
}

sgs_VarObj* sgs_GetObjectStructP( sgs_Variable* var )
{
	_OBJPREP_P( NULL );
	return var->data.O;
}

void* sgs_GetObjectDataP( sgs_Variable* var )
{
	_OBJPREP_P( NULL );
	return var->data.O->data;
}

sgs_ObjInterface* sgs_GetObjectIfaceP( sgs_Variable* var )
{
	_OBJPREP_P( NULL );
	return var->data.O->iface;
}

void sgs_SetObjectDataP( sgs_Variable* var, void* data )
{
	_OBJPREP_P( ; );
	var->data.O->data = data;
}

void sgs_SetObjectIfaceP( sgs_Variable* var, sgs_ObjInterface* iface )
{
	_OBJPREP_P( ; );
	var->data.O->iface = iface;
}


#define _OBJPREP( ret ) \
	sgs_Variable var = sgs_OptStackItem( C, item ); \
	DBLCHK( var.type != SGS_VT_OBJECT, ret )

char* sgs_GetStringPtr( SGS_CTX, StkIdx item )
{
	sgs_Variable var = sgs_OptStackItem( C, item );
	DBLCHK( var.type != SGS_VT_STRING, NULL )
	return sgs_var_cstr( &var );
}

sgs_SizeVal sgs_GetStringSize( SGS_CTX, StkIdx item )
{
	sgs_Variable var = sgs_OptStackItem( C, item );
	DBLCHK( var.type != SGS_VT_STRING, 0 )
	/* WP: string limit */
	return (sgs_SizeVal) var.data.S->size;
}

sgs_VarObj* sgs_GetObjectStruct( SGS_CTX, StkIdx item )
{
	_OBJPREP( NULL );
	return var.data.O;
}

void* sgs_GetObjectData( SGS_CTX, StkIdx item )
{
	_OBJPREP( NULL );
	return var.data.O->data;
}

sgs_ObjInterface* sgs_GetObjectIface( SGS_CTX, StkIdx item )
{
	_OBJPREP( NULL );
	return var.data.O->iface;
}

void sgs_SetObjectData( SGS_CTX, StkIdx item, void* data )
{
	_OBJPREP( ; );
	var.data.O->data = data;
}

void sgs_SetObjectIface( SGS_CTX, StkIdx item, sgs_ObjInterface* iface )
{
	_OBJPREP( ; );
	var.data.O->iface = iface;
}

#undef DBLCHK


#undef StkIdx


#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

#define MAX(a,b) ((a)>(b)?(a):(b))


#define RX_NEED_DEFAULT_MEMFUNC


#pragma once

#ifdef __cplusplus
extern "C" {
#endif

#include <stddef.h>


#define RXSUCCESS 0
#define RXEINMOD  -1 /* invalid modifier */
#define RXEPART   -2 /* partial (sub-)expression */
#define RXEUNEXP  -3 /* unexpected character */
#define RXERANGE  -4 /* invalid range (min > max) */
#define RXELIMIT  -5 /* too many digits */
#define RXEEMPTY  -6 /* expression is effectively empty */
#define RXENOREF  -7 /* the specified backreference cannot be used here */

#define RX_ALLMODS "mis"

#ifndef RX_STRLENGTHFUNC
#define RX_STRLENGTHFUNC( str ) strlen( str )
#endif


typedef void* (*srx_MemFunc)
(
	void* /* userdata */,
	void* /* ptr */,
	size_t /* size */
);

#ifdef RX_NEED_DEFAULT_MEMFUNC
static void* srx_DefaultMemFunc( void* userdata, void* ptr, size_t size )
{
	(void) userdata;
	if( size )
		return realloc( ptr, size );
	free( ptr );
	return NULL;
}
#endif

typedef char RX_Char;

typedef struct _srx_Context srx_Context;


srx_Context* srx_CreateExt( const RX_Char* str, size_t strsize, const RX_Char* mods, int* errnpos, srx_MemFunc memfn, void* memctx );
#define srx_Create( str, mods ) srx_CreateExt( str, RX_STRLENGTHFUNC(str), mods, NULL, srx_DefaultMemFunc, NULL )
int srx_Destroy( srx_Context* R );
void srx_DumpToStdout( srx_Context* R );

int srx_MatchExt( srx_Context* R, const RX_Char* str, size_t size, size_t offset );
#define srx_Match( R, str, off ) srx_MatchExt( R, str, RX_STRLENGTHFUNC(str), off )
int srx_GetCaptureCount( srx_Context* R );
int srx_GetCaptured( srx_Context* R, int which, size_t* pbeg, size_t* pend );
int srx_GetCapturedPtrs( srx_Context* R, int which, const RX_Char** pbeg, const RX_Char** pend );

RX_Char* srx_ReplaceExt( srx_Context* R, const RX_Char* str, size_t strsize, const RX_Char* rep, size_t repsize, size_t* outsize );
#define srx_Replace( R, str, rep ) srx_ReplaceExt( R, str, RX_STRLENGTHFUNC(str), rep, RX_STRLENGTHFUNC(rep), NULL )
void srx_FreeReplaced( srx_Context* R, RX_Char* repstr );


#ifdef __cplusplus
}
#endif




#define RX_MAX_CAPTURES 10


#define RX_MALLOC( bytes ) R->memfn( R->memctx, NULL, bytes )
#define RX_ALLOC_N( what, N ) (what*) R->memfn( R->memctx, NULL, sizeof( what ) * ((size_t)(N)) )
#define RX_ALLOC( what ) RX_ALLOC_N( what, 1 )
#define RX_FREE( ptr ) R->memfn( R->memctx, ptr, 0 )

#define RX_IS_ALPHA( x ) rx_isalpha( x )
#define RX_EQUALIZE( x ) rx_tolower( x )


#define RIT_MATCH  1 /* matching */
#define RIT_RANGE  2
#define RIT_SPCBEG 3
#define RIT_SPCEND 4
#define RIT_BKREF  5
#define RIT_EITHER 11 /* control */
#define RIT_SUBEXP 12

#define RIF_LAZY   0x01
#define RIF_INVERT 0x02

#define RCF_MULTILINE 0x01 /* ^/$ matches beginning/end of line too */
#define RCF_CASELESS  0x02 /* pre-equalized case for match/range */
#define RCF_DOTALL    0x04 /* "." is compiled as "[^]" instead of "[^\r\n]" */

#ifndef RXLOG
#define RXLOG 0
#endif

#if RXLOG
#define RXLOGINFO( x ) x
#else
#define RXLOGINFO( x )
#endif
#define RX_LOGLIM( str, strend, off ) (int)((strend)-(str)<(off)?(strend)-(str):(off)), str


static int rx_isalpha( RX_Char c )
{
	return ( c >= 'a' && c <= 'z' )
	    || ( c >= 'A' && c <= 'Z' );
}

static RX_Char rx_tolower( RX_Char c )
{
	if( c >= 'A' && c <= 'Z' )
		return (RX_Char)( c - 'A' + 'a' );
	return c;
}


typedef struct _regex_item regex_item;
struct _regex_item
{
	/* structure */
	regex_item* prev;
	regex_item* next;
	regex_item* ch, *ch2;
	regex_item* pos;
	
	RX_Char* range;
	int count;
	
	int type, flags;
	RX_Char a;
	int min, max;
	
	/* match state */
	const RX_Char *matchbeg, *matchend;
	int counter;
};

struct _srx_Context
{
	/* structure */
	regex_item*    root;
	int            flags;
	
	/* memory */
	srx_MemFunc    memfn;
	void*          memctx;
	
	/* captures */
	regex_item*    caps[ RX_MAX_CAPTURES ];
	int            numcaps;
	
	/* temporary data */
	const RX_Char* string;
	const RX_Char* stringend;
};

typedef struct _match_ctx
{
	const RX_Char* string;
	const RX_Char* stringend;
	regex_item*    item;
	srx_Context*   R;
}
match_ctx;


static int regex_test( const RX_Char* str, match_ctx* ctx );


static int regex_match_once( match_ctx* ctx )
{
	int i;
	regex_item* item = ctx->item;
	const RX_Char* str = item->matchend;
	RXLOGINFO( printf( "type %d char %d('%c') action at %p (%.*s)\n",
		item->type, (int) item->a, item->a, str, RX_LOGLIM(str,ctx->stringend,5) ) );
	switch( item->type )
	{
	case RIT_MATCH:
		if( str >= ctx->stringend )
			break;
		{
			RX_Char ch = *str;
			if( ctx->R->flags & RCF_CASELESS )
				ch = RX_EQUALIZE( *str );
			if( ch == item->a )
			{
				item->matchend++;
				return 1;
			}
		}
		break;
	case RIT_RANGE:
		if( str >= ctx->stringend )
			break;
		{
			RX_Char ch = *str;
			int inv = ( item->flags & RIF_INVERT ) != 0, inrange = 0;
			if( ctx->R->flags & RCF_CASELESS )
				ch = RX_EQUALIZE( *str );
			for( i = 0; i < item->count*2; i += 2 )
			{
				if( ch >= item->range[i] && ch <= item->range[i+1] )
				{
					inrange = 1;
					break;
				}
			}
			if( inrange ^ inv )
			{
				item->matchend++;
				return 1;
			}
		}
		break;
	case RIT_SPCBEG:
		if( ctx->R->flags & RCF_MULTILINE && item->matchend < ctx->stringend && ( *item->matchend == '\n' || *item->matchend == '\r' ) )
		{
			if( *item->matchend == '\r' && item->matchend[1] == '\n' )
				item->matchend++;
			item->matchend++;
			item->matchbeg = item->matchend;
			return 1;
		}
		return ctx->string == item->matchend;
	case RIT_SPCEND:
		if( ctx->R->flags & RCF_MULTILINE && item->matchend < ctx->stringend && ( *item->matchend == '\n' || *item->matchend == '\r' ) )
		{
			return 1;
		}
		return str >= ctx->stringend;
	case RIT_BKREF:
		{
			regex_item* cap = ctx->R->caps[ (int) item->a ];
			ptrdiff_t len = cap->matchend - cap->matchbeg;
			ptrdiff_t len2 = ctx->stringend - str;
			if( len2 >= len && memcmp( cap->matchbeg, str, (size_t) len ) == 0 )
			{
				item->matchend += len;
				return 1;
			}
		}
		break;
	case RIT_SUBEXP:
		{
			match_ctx cc;
			{
				cc.string = ctx->string;
				cc.stringend = ctx->stringend;
				cc.item = item->pos ? item->pos : item->ch;
				cc.R = ctx->R;
			}
			if( regex_test( str, &cc ) )
			{
				regex_item* p = item->ch;
				while( p->next )
					p = p->next;
				item->pos = NULL;
				item->matchend = p->matchend;
				return 1;
			}
		}
		break;
	}
	return 0;
}

static int regex_match_many( match_ctx* ctx )
{
	/* returns whether matched */
	regex_item* item = ctx->item;
	item->matchend = item->matchbeg;
	if( item->type == RIT_EITHER )
	{
		regex_item* chi = item->counter ? item->ch2 : item->ch;
		match_ctx cc;
		{
			cc.string = ctx->string;
			cc.stringend = ctx->stringend;
			cc.item = chi;
			cc.R = ctx->R;
		}
		if( regex_test( item->matchbeg, &cc ) )
		{
			regex_item* p = chi;
			while( p->next )
				p = p->next;
			item->matchend = p->matchend;
			return 1;
		}
		return 0;
	}
	else
	{
		int i;
		for( i = 0; i < item->counter; ++i )
		{
			if( item->matchend >= ctx->stringend && item->type != RIT_SPCEND && item->type != RIT_EITHER && item->type != RIT_SUBEXP )
			{
				item->counter = item->flags & RIF_LAZY ? item->max : i;
				RXLOGINFO( printf( "stopped while matching, counter = %d, %d between %d and %d?\n", item->counter, i, item->min, item->max ) );
				return i >= item->min && i <= item->max;
			}
			if( !regex_match_once( ctx ) )
			{
				item->counter = item->flags & RIF_LAZY ? item->max : i;
				RXLOGINFO( printf( "did not match, counter reset to %d\n", item->counter ) );
				return i >= item->min && i <= item->max;
			}
			RXLOGINFO( else printf( "matched\n" ) );
		}
		return 1;
	}
}

static void regex_full_reset( regex_item* p );
static void regex_reset_one( regex_item* p )
{
	if( p->ch ) regex_full_reset( p->ch );
	if( p->ch2 ) regex_full_reset( p->ch2 );
	p->pos = p->ch;
	p->matchbeg = p->matchend = NULL;
	p->counter = p->flags & RIF_LAZY ? p->min : p->max;
}
static void regex_full_reset( regex_item* p )
{
	while( p )
	{
		regex_reset_one( p );
		p = p->next;
	}
}

static regex_item* regex_lastch( regex_item* item )
{
	regex_item* p = item->ch;
	while( p && p->next )
		p = p->next;
	return p;
}

static int regex_subexp_backtrack( regex_item* item )
{
	int chgh = 0;
	regex_item* p = item->pos ? item->pos : regex_lastch( item );
	
	while( p )
	{
		RXLOGINFO( printf( "backtracker at type %d char %d\n", p->type, (int) p->a ) );
		if( chgh && p->type == RIT_SUBEXP && regex_subexp_backtrack( p ) )
			break;
		else if( p->flags & RIF_LAZY )
		{
			p->counter++;
			if( p->counter <= p->max )
				break;
		}
		else
		{
			p->counter--;
			if( p->counter >= p->min )
				break;
		}
		RXLOGINFO( printf( "subexp backtrack - reset current, move back\n" ) );
		regex_reset_one( p );
		p = p->prev;
		chgh = 1;
	}
	
	RXLOGINFO( printf( "subexp backtrack - %s\n", p ? "success" : "failure" ) );
	RXLOGINFO( if( p ) printf( "subexp-backtracked to type %d ctr=%d min=%d max=%d\n", p->type, p->counter, p->min, p->max ) );
	
	return !!p;
}

static int regex_test( const RX_Char* str, match_ctx* ctx )
{
	regex_item* p = ctx->item;
	p->matchbeg = str;
	
	for(;;)
	{
		int res;
		match_ctx cc;
		{
			cc.string = ctx->string;
			cc.stringend = ctx->stringend;
			cc.item = p;
			cc.R = ctx->R;
		}
		RXLOGINFO( printf( "match_many: item %p type %d at position %p (%.*s)\n",
			(void*) p, p->type, p->matchbeg, RX_LOGLIM(p->matchbeg,ctx->stringend,5) ) );
		res = regex_match_many( &cc );
		if( res )
		{
			p = p->next;
			if( !p )
			{
				RXLOGINFO( printf( "test of subexp %p SUCCEEDED\n", (void*) ctx->item ) );
				return 1;
			}
			RXLOGINFO( printf( "moving on to type %d action\n", p->type ) );
			p->matchbeg = p->prev->matchend;
		}
		else
		{
			int chgh = 0;
			while( p )
			{
				if( chgh && p->type == RIT_SUBEXP && regex_subexp_backtrack( p ) )
					break;
				else if( p->flags & RIF_LAZY )
				{
					p->counter++;
					if( p->counter <= p->max )
						break;
				}
				else
				{
					p->counter--;
					if( p->counter >= p->min )
						break;
				}
				RXLOGINFO( printf( "backtrack, reset current\n" ) );
				regex_reset_one( p );
				p = p->prev;
				chgh = 1;
			}
			if( !p )
			{
				RXLOGINFO( printf( "test of subexp %p BT-ENDED\n", (void*) ctx->item ) );
				return 0;
			}
		}
	}
}

static int regex_test_start( const RX_Char* str, match_ctx* ctx )
{
	regex_item* p = ctx->item;
	RXLOGINFO( printf( "test start - counter reset\n" ) );
	regex_reset_one( p );
	return regex_test( str, ctx );
}


/*
	mapping:
	- [^a-zA-Z] ... RIT_RANGE, optional RIF_INVERT
	- "." ... empty RIT_RANGE + RIF_INVERT
	- "\s" and others ... predefined RIT_RANGE with optional RIF_INVERT
	- "|" ... RIT_EITHER
	- "(..)" ... RIT_SUBEXP
	- "?" ... range = [0,1]
	- "*" ... range = [0,INT_MAX]
	- "+" ... range = [1,INT_MAX]
	- "{1,5}" ... range = [1,5] (other ranges mapped similarly)
	- "^" ... RIT_SPCBEG
	- "$" ... RIT_SPCEND
	- "\1" ... RIT_BKREF
*/

static void regex_free_item( srx_Context* R, regex_item* item );
static void regex_dealloc_item( srx_Context* R, regex_item* item )
{
	if( item->range )
		RX_FREE( item->range );
	if( item->ch ) regex_free_item( R, item->ch );
	if( item->ch2 ) regex_free_item( R, item->ch2 );
	RX_FREE( item );
}

static void regex_free_item( srx_Context* R, regex_item* item )
{
	regex_item *p, *c;
	if( !item )
		return;
	p = item->prev;
	while( p )
	{
		c = p;
		p = p->prev;
		regex_dealloc_item( R, c );
	}
	p = item->next;
	while( p )
	{
		c = p;
		p = p->next;
		regex_dealloc_item( R, c );
	}
	regex_dealloc_item( R, item );
}

static void regex_level( regex_item** pitem )
{
	/* TODO: balanced/non-(pseudo-)binary leveling */
	regex_item* item = *pitem;
	while( item )
	{
		if( item->type == RIT_EITHER )
		{
			regex_item* next = item->next;
			regex_level( &next );
			
			if( item->prev )
			{
				item->prev->next = NULL;
				item->prev = NULL;
			}
			if( item->next )
			{
				item->next->prev = NULL;
				item->next = NULL;
			}
			
			item->ch = *pitem;
			item->ch2 = next;
			
			*pitem = item;
			return;
		}
		item = item->next;
	}
}

static int regex_real_compile( srx_Context* R, int* cel, const RX_Char** pstr, const RX_Char* pend, int sub, regex_item** out )
{
#define _RX_ALLOC_NODE( ty ) \
	item = RX_ALLOC( regex_item ); \
	memset( item, 0, sizeof(*item) ); \
	if( citem ) \
	{ \
		citem->next = item; \
		item->prev = citem; \
	} \
	item->type = ty; \
	item->min = 1; \
	item->max = 1;

#define _RXE( err ) for(;;){ error = err; goto fail; }
	
	const RX_Char* s = *pstr;
	regex_item* item = NULL, *citem = NULL;
	int error = 0;
	while( s < pend )
	{
		if( sub && *s == ')' )
			break;
		switch( *s )
		{
		case '[':
			{
				const RX_Char* sc;
				int inv = 0, cnt = 0;
				RX_Char* ri;
				s++;
				if( *s == '^' )
				{
					inv = 1;
					s++;
				}
				sc = s;
				if( *sc == ']' )
				{
					sc++;
					cnt++;
				}
				while( *sc && *sc != ']' )
				{
					if( *sc == '-' && sc > s && sc[1] != 0 && sc[1] != ']' )
						sc++;
					else
						cnt++;
					sc++;
				}
				if( !*sc )
					_RXE( RXEPART );
				_RX_ALLOC_NODE( RIT_RANGE );
				if( inv )
					item->flags |= RIF_INVERT;
				item->range = ri = RX_ALLOC_N( RX_Char, cnt * 2 );
				item->count = cnt;
				sc = s;
				if( *sc == ']' )
				{
					sc++;
					ri[0] = ri[1] = *sc;
					ri += 2;
				}
				while( *sc && *sc != ']' )
				{
					if( *sc == '-' && sc > s && sc[1] != 0 && sc[1] != ']' )
					{
						if( ri > item->range )
							*(ri-1) = sc[1];
						sc++;
					}
					else
					{
						ri[0] = ri[1] = *sc;
						ri += 2;
					}
					sc++;
				}
				s = sc;
				if( *s == ']' )
					s++;
				if( R->flags & RCF_CASELESS )
				{
					int i;
					ri = item->range;
					for( i = 0; i < cnt * 2; i += 2 )
					{
						RX_Char A = ri[ i ], B = ri[ i + 1 ];
						if( RX_IS_ALPHA( A ) && RX_IS_ALPHA( B ) )
						{
							ri[ i ] = RX_EQUALIZE( A );
							ri[ i + 1 ] = RX_EQUALIZE( B );
						}
					}
				}
			}
			break;
		case ']':
			_RXE( RXEUNEXP );
		case '(':
			{
				int r, cap = R->numcaps < RX_MAX_CAPTURES ? 1 : -1;
				_RX_ALLOC_NODE( RIT_SUBEXP );
				if( cap >= 0 )
				{
					cap = R->numcaps++;
					R->caps[ cap ] = item;
				}
				s++;
				r = regex_real_compile( R, cel, &s, pend, 1, &item->ch );
				if( r )
					_RXE( r );
				item->pos = item->ch;
				if( *s != ')' )
					_RXE( RXEUNEXP );
				if( cap >= 0 )
					cel[ cap ] = 1;
				s++;
			}
			break;
		case ')':
			_RXE( RXEUNEXP );
		case '{':
		case '?':
		case '*':
		case '+':
			if( s > *pstr && ( *(s-1) == '}' || *(s-1) == '?' || *(s-1) == '*' || *(s-1) == '+' ) )
			{
				if( *s == '?' )
					item->flags |= RIF_LAZY;
				else
					_RXE( RXEUNEXP );
			}
			else if( item && ( item->type == RIT_MATCH || item->type == RIT_RANGE || item->type == RIT_BKREF || item->type == RIT_SUBEXP ) )
			{
				int min = 1, max = 1;
				if( *s == '{' )
				{
					int ctr;
					s++;
					if( !isdigit( *s ) )
						_RXE( RXEUNEXP );
					min = 0;
					ctr = 8;
					while( isdigit( *s ) && ctr > 0 )
					{
						min = min * 10 + *s++ - '0';
						ctr--;
					}
					if( isdigit( *s ) && ctr == 0 )
						_RXE( RXELIMIT );
					if( *s == ',' )
					{
						if( !isdigit(s[1]) )
							_RXE( RXEUNEXP );
						s++;
						max = 0;
						ctr = 8;
						while( isdigit( *s ) && ctr > 0 )
						{
							max = max * 10 + *s++ - '0';
							ctr--;
						}
						if( isdigit( *s ) && ctr == 0 )
							_RXE( RXELIMIT );
						if( min > max )
							_RXE( RXERANGE );
					}
					else
						max = min;
					if( *s != '}' )
						_RXE( RXEUNEXP );
				}
				else if( *s == '?' ){ min = 0; max = 1; }
				else if( *s == '*' ){ min = 0; max = INT_MAX - 1; }
				else if( *s == '+' ){ min = 1; max = INT_MAX - 1; }
				item->min = min;
				item->max = max;
			}
			else
				_RXE( RXEUNEXP );
			s++;
			break;
		case '}':
			_RXE( RXEUNEXP );
		case '|':
			if( !citem )
				_RXE( RXEUNEXP );
			_RX_ALLOC_NODE( RIT_EITHER );
			item->min = 0;
			item->max = 1;
			item->flags |= RIF_LAZY;
			s++;
			break;
		case '^':
			_RX_ALLOC_NODE( RIT_SPCBEG );
			s++;
			break;
		case '$':
			_RX_ALLOC_NODE( RIT_SPCEND );
			s++;
			break;
		case '\\':
			if( s[1] )
			{
				s++;
				if( *s == '.' )
				{
					_RX_ALLOC_NODE( RIT_MATCH );
					item->a = *s++;
					break;
				}
				else if( isdigit( *s ) )
				{
					int dig = *s++ - '0';
					if( dig == 0 || dig >= RX_MAX_CAPTURES || !cel[ dig ] )
						_RXE( RXENOREF );
					_RX_ALLOC_NODE( RIT_BKREF );
					item->a = (RX_Char) dig;
					break;
				}
				else if( *s == 'd' || *s == 'D' )
				{
					_RX_ALLOC_NODE( RIT_RANGE );
					item->range = RX_ALLOC_N( RX_Char, 2 );
					item->count = 1;
					item->range[0] = '0';
					item->range[1] = '9';
					if( *s == 'D' )
						item->flags |= RIF_INVERT;
					s++;
					break;
				}
				else if( *s == 'h' || *s == 'H' )
				{
					_RX_ALLOC_NODE( RIT_RANGE );
					item->range = RX_ALLOC_N( RX_Char, 2 * 2 );
					item->count = 2;
					item->range[0] = item->range[1] = '\t';
					item->range[2] = item->range[3] = ' ';
					if( *s == 'H' )
						item->flags |= RIF_INVERT;
					s++;
					break;
				}
				else if( *s == 'v' || *s == 'V' )
				{
					_RX_ALLOC_NODE( RIT_RANGE );
					item->range = RX_ALLOC_N( RX_Char, 2 );
					item->count = 1;
					item->range[0] = 0x0A;
					item->range[1] = 0x0D;
					if( *s == 'V' )
						item->flags |= RIF_INVERT;
					s++;
					break;
				}
				else if( *s == 's' || *s == 'S' )
				{
					_RX_ALLOC_NODE( RIT_RANGE );
					item->range = RX_ALLOC_N( RX_Char, 2 * 2 );
					item->count = 2;
					item->range[0] = 0x09;
					item->range[1] = 0x0D;
					item->range[2] = item->range[3] = ' ';
					if( *s == 'S' )
						item->flags |= RIF_INVERT;
					s++;
					break;
				}
				else if( *s == 'w' || *s == 'W' )
				{
					_RX_ALLOC_NODE( RIT_RANGE );
					item->range = RX_ALLOC_N( RX_Char, 2 * 4 );
					item->count = 4;
					item->range[0] = 'a'; item->range[1] = 'z';
					item->range[2] = 'A'; item->range[3] = 'Z';
					item->range[4] = '0'; item->range[5] = '9';
					item->range[6] = item->range[7] = '_';
					if( *s == 'W' )
						item->flags |= RIF_INVERT;
					s++;
					break;
				}
				/* TODO: more character classes */
			}
			else
				_RXE( RXEPART );
		default:
			if( *s == '.' )
			{
				_RX_ALLOC_NODE( RIT_RANGE );
				if( !( R->flags & RCF_DOTALL ) )
				{
					item->range = RX_ALLOC_N( RX_Char, 2 * 2 );
					item->range[0] = item->range[1] = '\n';
					item->range[2] = item->range[3] = '\r';
					item->count = 2;
				}
				item->flags |= RIF_INVERT;
			}
			else
			{
				_RX_ALLOC_NODE( RIT_MATCH );
				item->a = *s;
				if( R->flags & RCF_CASELESS && RX_IS_ALPHA( item->a ) )
					item->a = RX_EQUALIZE( item->a );
			}
			s++;
			break;
		}
		citem = item;
	}
	if( !item )
		_RXE( RXEEMPTY );
	if( item->type == RIT_EITHER )
		_RXE( RXEPART );
	*pstr = s;
	while( item->prev )
		item = item->prev;
	regex_level( &item );
	*out = item;
	return RXSUCCESS;
fail:
	regex_free_item( R, item );
	return (int)( ( error & 0xf ) | ( ( s - R->string ) << 4 ) );
}

/*
	#### srx_CreateExt ####
*/
srx_Context* srx_CreateExt( const RX_Char* str, size_t strsize, const RX_Char* mods, int* errnpos, srx_MemFunc memfn, void* memctx )
{
	int flags = 0, err, cel[ RX_MAX_CAPTURES ];
	srx_Context* R = NULL;
	if( mods )
	{
		const RX_Char* modbegin = mods;
		while( *mods )
		{
			switch( *mods )
			{
			case 'm': flags |= RCF_MULTILINE; break;
			case 'i': flags |= RCF_CASELESS; break;
			case 's': flags |= RCF_DOTALL; break;
			default:
				err = ( RXEINMOD & 0xf ) | ( (int)( mods - modbegin ) << 4 );
				goto fail;
			}
			mods++;
		}
	}
	
	if( !memfn )
		memfn = srx_DefaultMemFunc;
	
	R = (srx_Context*) memfn( memctx, NULL, sizeof(srx_Context) );
	memset( R, 0, sizeof(*R) );
	memset( cel, 0, sizeof(cel) );
	R->memfn = memfn;
	R->memctx = memctx;
	R->string = str;
	R->stringend = str + strsize;
	R->flags = flags;
	R->numcaps = 1;
	
	err = regex_real_compile( R, cel, &str, str + strsize, 0, &R->root );
	
	if( err )
	{
		memfn( memctx, R, 0 );
		R = NULL;
	}
	else
	{
		regex_item* item = RX_ALLOC( regex_item );
		memset( item, 0, sizeof(*item) );
		item->type = RIT_SUBEXP;
		item->min = 1;
		item->max = 1;
		item->pos = item->ch = R->root;
		R->caps[ 0 ] = R->root = item;
	}
fail:
	if( errnpos )
	{
		unsigned uerr = (unsigned) err;
		errnpos[0] = (int)( uerr ? ( uerr & 0xf ) | 0xfffffff0 : 0 );
		errnpos[1] = (int)( ( uerr & 0xfffffff0 ) >> 4 );
	}
	RXLOGINFO( if( R ) srx_DumpToStdout(R) );
	return R;
}

/*
	#### srx_Destroy ####
*/
int srx_Destroy( srx_Context* R )
{
	if( R )
	{
		srx_MemFunc memfn = R->memfn;
		void* memctx = R->memctx;
		if( R->root )
			regex_free_item( R, R->root );
		memfn( memctx, R, 0 );
	}
	return !!R;
}


static void regex_dump_list( regex_item* items, int lev );
static void regex_dump_item( regex_item* item, int lev )
{
	const char* types[] =
	{
		"-", "MATCH (1)", "RANGE (2)", "SPCBEG (3)", "SPCEND (4)", "BKREF (5)", "-", "-", "-", "-",
		"-", "EITHER (11)", "SUBEXP (12)", "-"
	};
	
	int l = lev;
	while( l --> 0 )
		printf( "- " );
	printf( "%s", types[ item->type ] );
	if( item->flags & RIF_INVERT ) printf( " INV" );
	if( item->flags & RIF_LAZY ) printf( " LAZY" );
	switch( item->type )
	{
	case RIT_MATCH: printf( " char %d", (int) item->a ); break;
	case RIT_RANGE:
		for( l = 0; l < item->count; ++l )
		{
			if( l > 0 )
				printf( "," );
			printf( " %d - %d", (int) item->range[l*2], (int) item->range[l*2+1] );
		}
		break;
	case RIT_BKREF: printf( " #%d", (int) item->a ); break;
	}
	printf( " (%d to %d) (0x%p => 0x%p)\n", item->min, item->max, item->matchbeg, item->matchend );
	
	if( item->ch )
	{
		regex_dump_list( item->ch, lev + 1 );
		if( item->ch2 )
		{
			int l2 = lev;
			while( l2 --> 0 )
				printf( "- " );
			printf( "--|\n" );
			regex_dump_list( item->ch2, lev + 1 );
		}
	}
}
static void regex_dump_list( regex_item* items, int lev )
{
	while( items )
	{
		regex_dump_item( items, lev );
		items = items->next;
	}
}

/*
	#### srx_DumpToStdout ####
*/
void srx_DumpToStdout( srx_Context* R )
{
	regex_dump_list( R->root, 0 );
}

/*
	#### srx_Match ####
*/
int srx_MatchExt( srx_Context* R, const RX_Char* str, size_t size, size_t offset )
{
	int ret;
	const RX_Char* strend = str + size;
	match_ctx ctx;
	{
		ctx.string = str;
		ctx.stringend = strend;
		ctx.item = R->root;
		ctx.R = R;
	}
	R->string = str;
	if( offset > size )
		return 0;
	str += offset;
	while( str < strend )
	{
		ret = regex_test_start( str, &ctx );
		if( ret < 0 )
			return 0;
		if( ret > 0 )
			return 1;
		str++;
	}
	return 0;
}

/*
	#### srx_GetCaptureCount ####
*/
int srx_GetCaptureCount( srx_Context* R )
{
	return R->numcaps;
}

/*
	#### srx_GetCaptured ####
*/
int srx_GetCaptured( srx_Context* R, int which, size_t* pbeg, size_t* pend )
{
	const RX_Char* a, *b;
	if( srx_GetCapturedPtrs( R, which, &a, &b ) )
	{
		if( pbeg ) *pbeg = (size_t)( a - R->string );
		if( pend ) *pend = (size_t)( b - R->string );
		return 1;
	}
	return 0;
}

/*
	#### srx_GetCapturedPtrs ####
*/
int srx_GetCapturedPtrs( srx_Context* R, int which, const RX_Char** pbeg, const RX_Char** pend )
{
	if( which < 0 || which >= R->numcaps )
		return 0;
	if( R->caps[ which ] == NULL )
		return 0;
	if( pbeg ) *pbeg = R->caps[ which ]->matchbeg;
	if( pend ) *pend = R->caps[ which ]->matchend;
	return 1;
}

/*
	#### srx_ReplaceExt ####
*/
RX_Char* srx_ReplaceExt( srx_Context* R, const RX_Char* str, size_t strsize, const RX_Char* rep, size_t repsize, size_t* outsize )
{
	RX_Char* out = "";
	const RX_Char *from = str, *fromend = str + strsize, *repend = rep + repsize;
	size_t size = 0, mem = 0;
	
#define SR_CHKSZ( szext ) \
	if( (ptrdiff_t)( mem - size ) < (ptrdiff_t)(szext) ) \
	{ \
		size_t nsz = MAX( mem * 2, size + (size_t)(szext) ); \
		RX_Char* nmem = RX_ALLOC_N( RX_Char, nsz + 1 ); \
		if( mem ) \
		{ \
			memcpy( nmem, out, size + 1 ); /* copy with \0 */ \
			RX_FREE( out ); \
		} \
		out = nmem; \
		mem = nsz; \
	}
#define SR_ADDBUF( from, to ) \
	SR_CHKSZ( to - from ) \
	memcpy( out + size, from, (size_t)( to - from ) ); \
	size += (size_t)( to - from );
	
	while( from < fromend )
	{
		const RX_Char* ofp = NULL, *ep = NULL, *rp;
		if( !srx_MatchExt( R, from, (size_t)( fromend - from ), 0 ) )
			break;
		srx_GetCapturedPtrs( R, 0, &ofp, &ep );
		SR_ADDBUF( from, ofp );
		
		rp = rep;
		while( rp < repend )
		{
			RX_Char rc = *rp;
			if( ( rc == '\\' || rc == '$' ) && rp + 1 < repend )
			{
				if( isdigit( rp[1] ) )
				{
					int dig = rp[1] - '0';
					const RX_Char *brp, *erp;
					if( srx_GetCapturedPtrs( R, dig, &brp, &erp ) )
					{
						SR_ADDBUF( brp, erp );
					}
					rp += 2;
					continue;
				}
				else if( rp[1] == rc )
				{
					rp++;
				}
			}
			SR_ADDBUF( rp, rp + 1 );
			rp++;
		}
		
		if( from == ep )
			from++;
		else
			from = ep;
	}
	
	SR_ADDBUF( from, fromend );
	if( outsize )
		*outsize = size;
	{
		char nul[1] = {0};
		SR_ADDBUF( nul, &nul[1] );
	}
	return out;
}

/*
	#### srx_FreeReplaced ####
*/
void srx_FreeReplaced( srx_Context* R, RX_Char* repstr )
{
	RX_FREE( repstr );
}





#define SGS_DEBUG_SERIALIZE 0
#if SGS_DEBUG_SERIALIZE
#  define SRLZ_DEBUG( x ) x
#else
#  define SRLZ_DEBUG( x )
#endif


static int _serialize_function( SGS_CTX, sgs_iFunc* func, sgs_MemBuf* out )
{
	if( sgsBC_Func2Buf( C, func, out ) == SGS_FALSE )
	{
		sgs_Msg( C, SGS_INTERR, "failed to serialize function: error in data" );
		return 0;
	}
	return 1;
}

static int _unserialize_function( SGS_CTX, const char* buf, size_t sz, sgs_Variable* outfn )
{
	const char* err;
	if( sgsBC_ValidateHeader( buf, sz ) < SGS_HEADER_SIZE )
	{
		sgs_Msg( C, SGS_WARNING, "failed to unserialize function: incomplete data" );
		return 0;
	}
	err = sgsBC_Buf2Func( C, "<anonymous>", buf, sz, outfn );
	if( err )
		return sgs_Msg( C, SGS_WARNING, "failed to unserialize function: %s", err );
	return 1;
}


static SGSBOOL sgs__thread_serialize( SGS_CTX, sgs_Context* ctx, sgs_MemBuf* outbuf, sgs_MemBuf* argarray )
{
	sgs_MemBuf buf = sgs_membuf_create();
	sgs_StackFrame* sf;
	
	SRLZ_DEBUG( printf( "SRLZ thread_serialize\n" ) );
	
#define _WRITE32( x ) { int32_t _tmp = (int32_t)(x); sgs_membuf_appbuf( &buf, C, &_tmp, 4 ); }
#define _WRITE8( x ) { sgs_membuf_appchr( &buf, C, (char)(x) ); }
	
	/* failure condition: cannot serialize self */
	if( C == ctx )
		return 0;
	/* failure condition: C functions in stack frame */
	sf = ctx->sf_first;
	while( sf )
	{
		if( sf->iptr == NULL )
			return 0;
		sf = sf->next;
	}
	
	/* variables: _G */
	SRLZ_DEBUG( printf( "SRLZ thread _G\n" ) );
	{
		sgs_Variable v_obj; v_obj.type = SGS_VT_OBJECT; v_obj.data.O = ctx->_G;
		sgs_Serialize( C, v_obj );
	}
	/* variables: stack */
	SRLZ_DEBUG( printf( "SRLZ thread STACK\n" ) );
	{
		sgs_Variable* p = ctx->stack_base;
		sf = ctx->sf_first;
		while( p != ctx->stack_top )
		{
			/* if stack position equals frame function position and ..
			.. frame has a closure ref., serialize closure instead of its func. */
			int hasclsr = 0;
			if( sf && sf->func == p )
			{
				SRLZ_DEBUG( printf( "SRLZ thread found function at %d\n", (int)( sf->func - ctx->stack_base ) ) );
				if( sf->clsrref )
				{
					SRLZ_DEBUG( printf( "^ it has closure\n" ) );
					sgs_Variable vobj;
					vobj.type = SGS_VT_OBJECT;
					vobj.data.O = sf->clsrref;
					sgs_Serialize( C, vobj );
					hasclsr = 1;
				}
				/* next frames won't appear on stack before this position */
				sf = sf->next;
			}
			if( !hasclsr )
				sgs_Serialize( C, *p );
			p++;
		}
	}
	
	SRLZ_DEBUG( printf( "SRLZ thread CORE\n" ) );
	_WRITE32( 0x5C057A7E ); /* Serialized COroutine STATE */
	/* POD: main context */
	_WRITE32( ctx->minlev );
	_WRITE32( ctx->apilev );
	_WRITE32( ctx->last_errno );
	_WRITE32( ctx->state );
	_WRITE32( ctx->stack_top - ctx->stack_base );
	_WRITE32( ctx->stack_off - ctx->stack_base );
	_WRITE32( ctx->stack_mem );
	_WRITE32( ctx->sf_count );
	_WRITE32( ctx->num_last_returned );
	/* stack frames */
	SRLZ_DEBUG( printf( "SRLZ thread FRAMES\n" ) );
	sf = ctx->sf_first;
	while( sf )
	{
		/* 'code' will be taken from function */
		if( sf->func->type == SGS_VT_FUNC )
		{
			sgs_iFunc* F = sf->func->data.F;
			_WRITE32( sf->func - ctx->stack_base );
			_WRITE32( sf->iptr - sgs_func_bytecode( F ) );
			_WRITE32( sgs_func_instr_count( F ) ); /* - for validation */
			_WRITE32( sgs_func_const_count( F ) ); /* - for validation */
		}
		else
		{
			SRLZ_DEBUG( printf( "SRLZ THREAD NOT VT_FUNC\n" ) );
			goto fail;
		}
		/* 'cptr' will be taken from function */
		/* 'nfname' is irrelevant for non-native functions */
		/* 'prev', 'next', 'cached' are system pointers */
		_WRITE32( sf->argbeg );
		_WRITE32( sf->stkoff );
		_WRITE32( sf->errsup );
		_WRITE8( sf->argcount );
		_WRITE8( sf->flags );
		sf = sf->next;
	}
	
	SRLZ_DEBUG( printf( "SRLZ emit thread\n" ) );
	sgs_membuf_appchr( outbuf, C, 'T' );
	if( argarray )
	{
		uint32_t argcount = (uint32_t)( 1 /* _G */
			+ ( ctx->stack_top - ctx->stack_base ) /* stack */ );
		sgs_membuf_appbuf( outbuf, C, &argcount, 4 );
		sgs_membuf_appbuf( outbuf, C, argarray->ptr + argarray->size - argcount * 4, argcount * 4 );
		sgs_membuf_erase( argarray, argarray->size - argcount * 4, argarray->size );
	}
	
	sgs_membuf_appbuf( outbuf, C, buf.ptr, buf.size );
	sgs_membuf_destroy( &buf, C );
	
#undef _WRITE32
#undef _WRITE8
	return 1;
fail:
	SRLZ_DEBUG( printf( "SRLZ serialize_thread ERROR\n" ) );
	sgs_membuf_destroy( &buf, C );
	return 0;
}

static int sgs__thread_unserialize( SGS_CTX, sgs_Context** pT, char** pbuf, char* bufend )
{
	char *buf = *pbuf;
	sgs_Context* ctx = sgsCTX_ForkState( C, SGS_FALSE );
	
	SRLZ_DEBUG( printf( "USRZ thread_unserialize\n" ) );
	
#define _READ32( x ) { if( buf + 4 > bufend ) goto fail; memcpy( &x, buf, 4 ); buf += 4; }
#define _READ8( x ) { if( buf + 1 > bufend ) goto fail; x = (uint8_t) *buf++; }
	
	{
		int32_t i, tag, stacklen, stackoff, sfnum;
		
		SRLZ_DEBUG( printf( "USRZ thread CORE\n" ) );
		_READ32( tag );
		if( tag != 0x5C057A7E )
		{
			SRLZ_DEBUG( printf( "USRZ THREAD NOT 5C057A7E ([s]erialized [co]routine [state])\n" ) );
			goto fail;
		}
		
		/* POD: context */
		_READ32( ctx->minlev );
		_READ32( ctx->apilev );
		_READ32( ctx->last_errno );
		_READ32( ctx->state );
		_READ32( stacklen );
		_READ32( stackoff );
		_READ32( ctx->stack_mem );
		_READ32( sfnum );
		_READ32( ctx->num_last_returned );
		
		/* variables: _G */
		SRLZ_DEBUG( printf( "USRZ thread _G\n" ) );
		ctx->_G = sgs_StackItem( C, 0 ).data.O;
		sgs_ObjAcquire( ctx, ctx->_G );
		
		/* variables: stack */
		SRLZ_DEBUG( printf( "USRZ thread STACK\n" ) );
		sgs_BreakIf( ctx->stack_top != ctx->stack_base );
		for( i = 0; i < stacklen; ++i )
			sgs_PushVariable( ctx, sgs_StackItem( C, 1 + i ) );
		sgs_BreakIf( ctx->stack_top != ctx->stack_base + stacklen );
		if( stackoff > stacklen )
		{
			SRLZ_DEBUG( printf( "USRZ THREAD stackoff > stacklen\n" ) );
			goto fail;
		}
		ctx->stack_off = ctx->stack_base + stackoff;
		
		/* stack frames */
		SRLZ_DEBUG( printf( "USRZ thread FRAMES\n" ) );
		for( i = 0; i < sfnum; ++i )
		{
			sgs_StackFrame* sf;
			int32_t funcoff, iptrpos, iendpos, ccount;
			
			/* POD: stack frames */
			_READ32( funcoff );
			SRLZ_DEBUG( printf( "USRZ thread frame funcoff = %d\n", (int) funcoff ) );
			if( funcoff < 0 || funcoff >= stacklen )
			{
				SRLZ_DEBUG( printf( "USRZ THREAD funcoff out of bounds [0;stacklen)\n" ) );
				goto fail;
			}
			
			if( !sgsVM_PushStackFrame( ctx, ctx->stack_base + funcoff ) )
			{
				SRLZ_DEBUG( printf( "USRZ THREAD PushStackFrame failed\n" ) );
				goto fail;
			}
			sf = ctx->sf_last;
			
			/* 'code' will be taken from function */
			if( sf->func->type == SGS_VT_FUNC )
			{
				sgs_iFunc* F = sf->func->data.F;
				_READ32( iptrpos );
				_READ32( iendpos ); /* - for validation */
				if( iendpos != sgs_func_instr_count( F ) )
				{
					SRLZ_DEBUG( printf( "USRZ THREAD code size mismatch\n" ) );
					goto fail;
				}
				sf->iptr = sgs_func_bytecode( F ) + iptrpos;
				_READ32( ccount ); /* - for validation */
				if( ccount != sgs_func_const_count( F ) )
				{
					SRLZ_DEBUG( printf( "USRZ THREAD constant count mismatch\n" ) );
					goto fail;
				}
			}
			else
			{
				SRLZ_DEBUG( printf( "USRZ THREAD function NOT VT_FUNC (type=%d)\n", sf->func->type ) );
				SRLZ_DEBUG( if( sf->func->type == SGS_VT_OBJECT )
					printf( "^ iface = %s\n", sf->func->data.O->iface->name ) );
				goto fail;
			}
			
			/* 'cptr' will be taken from function */
			/* 'nfname' is irrelevant for non-native functions */
			/* 'prev', 'next', 'cached' are system pointers */
			_READ32( sf->argbeg );
			_READ32( sf->stkoff );
			_READ32( sf->errsup );
			_READ8( sf->argcount );
			_READ8( sf->flags );
		}
	}
	
#undef _READ32
#undef _READ8
	
	/* finalize */
	*pbuf = buf;
	sgs_BreakIf( ctx->refcount != 0 );
	*pT = ctx;
	SRLZ_DEBUG( printf( "USRZ thread_unserialize DONE\n" ) );
	return 1;
fail:
	SRLZ_DEBUG( printf( "USRZ thread_unserialize ERROR\n" ) );
	sgsCTX_FreeState( ctx );
	return 0;
}


typedef struct sgs_serialize1_data
{
	int mode;
	int ret;
	sgs_MemBuf data;
}
sgs_serialize1_data;

void sgs_SerializeInt_V1( SGS_CTX, sgs_Variable var )
{
	void* prev_serialize_state = C->serialize_state;
	sgs_serialize1_data SD = { 1, 1, sgs_membuf_create() }, *pSD;
	int ep = !C->serialize_state || *(int*)C->serialize_state != 1;
	
	if( ep )
	{
		C->serialize_state = &SD;
	}
	pSD = (sgs_serialize1_data*) C->serialize_state;
	
	if( var.type == SGS_VT_OBJECT || var.type == SGS_VT_CFUNC ||
		var.type == SGS_VT_FUNC || var.type == SGS_VT_THREAD )
	{
		sgs_Variable sym = sgs_MakeNull();
		if( sgs_GetSymbol( C, var, &sym ) && sym.type == SGS_VT_STRING )
		{
			sgs_SerializeInt_V1( C, sym );
			sgs_membuf_appchr( &pSD->data, C, 'S' );
			sgs_Release( C, &sym );
			goto end;
		}
		sgs_Release( C, &sym );
	}
	
	if( var.type == SGS_VT_OBJECT )
	{
		int parg = C->object_arg;
		sgs_VarObj* O = var.data.O;
		_STACK_PREPARE;
		if( !O->iface->serialize )
		{
			sgs_Msg( C, SGS_ERROR, "cannot serialize object of type '%s'", O->iface->name );
			pSD->ret = SGS_FALSE;
			goto end;
		}
		_STACK_PROTECT;
		C->object_arg = 1;
		pSD->ret = SGS_SUCCEEDED( O->iface->serialize( C, O ) );
		C->object_arg = parg;
		_STACK_UNPROTECT;
		goto end;
	}
	else if( var.type == SGS_VT_CFUNC )
	{
		sgs_Msg( C, SGS_ERROR, "serialization mode 1 does not support C function serialization" );
		pSD->ret = SGS_FALSE;
		goto end;
	}
	else if( var.type == SGS_VT_PTR )
	{
		sgs_Msg( C, SGS_ERROR, "serialization mode 1 does not support pointer serialization" );
		pSD->ret = SGS_FALSE;
		goto end;
	}
	else if( var.type == SGS_VT_THREAD )
	{
		sgs_Msg( C, SGS_ERROR, "serialization mode 1 does not support thread serialization" );
		pSD->ret = SGS_FALSE;
		goto end;
	}
	else
	{
		char pb[2];
		{
			pb[0] = 'P';
			/* WP: basetype limited to bits 0-7, sign interpretation does not matter */
			pb[1] = (char) var.type;
		}
		sgs_membuf_appbuf( &pSD->data, C, pb, 2 );
		switch( var.type )
		{
		case SGS_VT_NULL: break;
		/* WP: var.data.B uses only bit 0 */
		case SGS_VT_BOOL: { char b = (char) var.data.B; sgs_membuf_appbuf( &pSD->data, C, &b, 1 ); } break;
		case SGS_VT_INT: sgs_membuf_appbuf( &pSD->data, C, &var.data.I, sizeof( sgs_Int ) ); break;
		case SGS_VT_REAL: sgs_membuf_appbuf( &pSD->data, C, &var.data.R, sizeof( sgs_Real ) ); break;
		case SGS_VT_STRING:
			sgs_membuf_appbuf( &pSD->data, C, &var.data.S->size, 4 );
			sgs_membuf_appbuf( &pSD->data, C, sgs_var_cstr( &var ), var.data.S->size );
			break;
		case SGS_VT_FUNC:
			{
				size_t szbefore = pSD->data.size;
				if( !_serialize_function( C, var.data.F, &pSD->data ) )
				{
					sgs_Msg( C, SGS_INTERR, "sgs_Serialize: failed to serialize function "
						"(ptr = %p, name = %s, file = %s)", var.data.F,
						sgs_str_cstr( var.data.F->sfuncname ), sgs_str_cstr( var.data.F->sfilename ) );
					pSD->ret = SGS_FALSE;
					goto end;
				}
				else
				{
					uint32_t szdiff = (uint32_t) ( pSD->data.size - szbefore );
					sgs_membuf_insbuf( &pSD->data, C, szbefore, &szdiff, sizeof(szdiff) );
				}
			}
			break;
		default:
			sgs_Msg( C, SGS_INTERR, "sgs_Serialize: unknown memory error" );
			pSD->ret = SGS_FALSE;
			goto end;
		}
	}

end:
	if( ep )
	{
		if( SD.ret )
		{
			if( SD.data.size > 0x7fffffff )
			{
				sgs_Msg( C, SGS_ERROR, "serialized string too long" );
				sgs_PushNull( C );
			}
			else
			{
				/* WP: added error condition */
				sgs_PushStringBuf( C, SD.data.ptr, (sgs_SizeVal) SD.data.size );
			}
		}
		else
		{
			sgs_PushNull( C );
		}
		sgs_membuf_destroy( &SD.data, C );
		C->serialize_state = prev_serialize_state;
	}
}

static void sgs_SerializeObjectInt_V1( SGS_CTX, sgs_StkIdx args, const char* func, size_t fnsize )
{
	sgs_serialize1_data* pSD = (sgs_serialize1_data*) C->serialize_state;
	char pb[7] = { 'C', 0, 0, 0, 0, 0, 0 };
	{
		/* WP: they were pointless */
		pb[1] = (char)((args)&0xff);
		pb[2] = (char)((args>>8)&0xff);
		pb[3] = (char)((args>>16)&0xff);
		pb[4] = (char)((args>>24)&0xff);
		/* WP: have error condition + sign interpretation doesn't matter */
		pb[5] = (char) fnsize;
	}
	
	sgs_membuf_appbuf( &pSD->data, C, pb, 6 );
	sgs_membuf_appbuf( &pSD->data, C, func, fnsize );
	sgs_membuf_appbuf( &pSD->data, C, pb + 6, 1 );
}

static void sgs_SerializeObjIndexInt_V1( SGS_CTX, int isprop )
{
	sgs_serialize1_data* pSD = (sgs_serialize1_data*) C->serialize_state;
	sgs_membuf_appchr( &pSD->data, C, isprop ? '.' : '[' );
}

#define sgs_unserr_incomp( C ) sgs_Msg( C, SGS_WARNING, "failed to unserialize: incomplete data [L%d]", __LINE__ )
#define sgs_unserr_error( C ) sgs_Msg( C, SGS_WARNING, "failed to unserialize: error in data [L%d]", __LINE__ )
#define sgs_unserr_objcall( C ) sgs_Msg( C, SGS_WARNING, "failed to unserialize: could not create object from function [L%d]", __LINE__ )
#define sgs_unserr_symfail( C ) sgs_Msg( C, SGS_WARNING, "failed to unserialize: could not map name to symbol [L%d]", __LINE__ )

SGSBOOL sgs_UnserializeInt_V1( SGS_CTX, char* str, char* strend )
{
	while( str < strend )
	{
		char c = *str++;
		if( c == 'P' )
		{
			if( str >= strend )
				return sgs_unserr_incomp( C );
			c = *str++;
			switch( c )
			{
			case SGS_VT_NULL: sgs_PushNull( C ); break;
			case SGS_VT_BOOL:
				if( str >= strend )
					return sgs_unserr_incomp( C );
				sgs_PushBool( C, *str++ );
				break;
			case SGS_VT_INT:
				if( str >= strend-7 )
					return sgs_unserr_incomp( C );
				else
				{
					sgs_Int val;
					SGS_AS_INTEGER( val, str );
					sgs_PushInt( C, val );
				}
				str += 8;
				break;
			case SGS_VT_REAL:
				if( str >= strend-7 )
					return sgs_unserr_incomp( C );
				else
				{
					sgs_Real val;
					SGS_AS_REAL( val, str );
					sgs_PushReal( C, val );
				}
				str += 8;
				break;
			case SGS_VT_STRING:
				{
					sgs_SizeVal strsz;
					if( str >= strend-3 )
						return sgs_unserr_incomp( C );
					SGS_AS_INT32( strsz, str );
					str += 4;
					if( str > strend - strsz )
						return sgs_unserr_incomp( C );
					sgs_PushStringBuf( C, str, strsz );
					str += strsz;
				}
				break;
			case SGS_VT_FUNC:
				{
					sgs_Variable tmp;
					sgs_SizeVal bcsz;
					if( str >= strend-3 )
						return sgs_unserr_incomp( C );
					SGS_AS_INT32( bcsz, str );
					str += 4;
					if( str > strend - bcsz )
						return sgs_unserr_incomp( C );
					/* WP: conversion does not affect values */
					if( !_unserialize_function( C, str, (size_t) bcsz, &tmp ) )
						return SGS_FALSE; /* error already printed */
					fstk_push_leave( C, &tmp );
					str += bcsz;
				}
				break;
			default:
				return sgs_unserr_error( C );
			}
		}
		else if( c == 'C' )
		{
			int32_t argc;
			int fnsz, ret;
			if( str >= strend-4 )
				return sgs_unserr_incomp( C );
			SGS_AS_INT32( argc, str );
			str += 4;
			fnsz = *str++ + 1;
			if( str > strend - fnsz )
				return sgs_unserr_incomp( C );
			ret = sgs_GlobalCall( C, str, argc, 1 );
			if( SGS_FAILED( ret ) )
				return sgs_unserr_objcall( C );
			str += fnsz;
		}
		else if( c == 'T' )
		{
			sgs_Context* T = NULL;
			if( !sgs__thread_unserialize( C, &T, &str, strend ) )
				return sgs_unserr_incomp( C );
			sgs_PushThreadPtr( C, T );
		}
		else if( c == 'S' )
		{
			sgs_Variable sym;
			if( !sgs_GetSymbol( C, sgs_StackItem( C, -1 ), &sym ) )
			{
				return sgs_unserr_symfail( C );
			}
			sgs_Pop( C, 1 );
			fstk_push_leave( C, &sym );
		}
		else if( c == '.' || c == '[' )
		{
			sgs_SetIndex( C, sgs_StackItem( C, -3 ), sgs_StackItem( C, -2 ),
				sgs_StackItem( C, -1 ), c == '.' );
			sgs_Pop( C, 2 );
		}
		else
		{
			return sgs_unserr_error( C );
		}
	}
	return SGS_TRUE;
}


typedef struct sgs_serialize2_data
{
	int mode;
	int ret;
	sgs_VHTable servartable;
	sgs_MemBuf argarray;
	sgs_VarObj* curObj;
	sgs_MemBuf data;
	int32_t metaObjArg;
}
sgs_serialize2_data;

#define SGS_VTSPC_CLOSURE 0xf0

static void srlz_mode2_addvar( SGS_CTX, sgs_serialize2_data* pSD, sgs_Variable* pvar )
{
	sgs_Variable idxvar;
	uint32_t argidx = (uint32_t) sgs_vht_size( &pSD->servartable );
	idxvar.type = SGS_VT_INT;
	idxvar.data.I = argidx;
	sgs_vht_set( &pSD->servartable, C, pvar, &idxvar );
	sgs_membuf_appbuf( &pSD->argarray, C, &argidx, sizeof(argidx) );
}

void sgs_SerializeInt_V2( SGS_CTX, sgs_Variable var )
{
	void* prev_serialize_state = C->serialize_state;
	sgs_serialize2_data SD, *pSD;
	int ep = !C->serialize_state || *(int*)C->serialize_state != 2;
	
	if( ep )
	{
		SRLZ_DEBUG( printf( "SRLZ == mode 2 START ==\n") );
		
		SD.mode = 2;
		SD.ret = 1;
		sgs_vht_init( &SD.servartable, C, 64, 64 );
		SD.argarray = sgs_membuf_create();
		SD.curObj = NULL;
		SD.data = sgs_membuf_create();
		C->serialize_state = &SD;
	}
	pSD = (sgs_serialize2_data*) C->serialize_state;
	
	SRLZ_DEBUG( printf( "SRLZ var type=%d\n", (int) var.type ) );
	SRLZ_DEBUG( if( var.type == SGS_VT_OBJECT ){
		printf( "- object type: %s\n", var.data.O->iface->name ); } );
	
	/* if variable was repeated, push the index of it on argument stack */
	{
		sgs_VHTVar* vv = sgs_vht_get( &pSD->servartable, &var );
		if( vv )
		{
			uint32_t argidx = (uint32_t) ( vv - pSD->servartable.vars );
			
			SRLZ_DEBUG( printf( "SRLZ repeated variable\n" ) );
			
			sgs_membuf_appbuf( &pSD->argarray, C, &argidx, sizeof(argidx) );
			goto end;
		}
	}
	
	if( var.type == SGS_VT_OBJECT || var.type == SGS_VT_CFUNC ||
		var.type == SGS_VT_FUNC || var.type == SGS_VT_THREAD )
	{
		sgs_Variable sym = sgs_MakeNull();
		if( sgs_GetSymbol( C, var, &sym ) && sym.type == SGS_VT_STRING )
		{
			SRLZ_DEBUG( printf( "SRLZ new symbol\n" ) );
			
			sgs_SerializeInt_V2( C, sym );
			if( pSD->argarray.size < 4 )
			{
				/* error likely to be already printed */
				pSD->ret = SGS_FALSE;
				goto end;
			}
			
			SRLZ_DEBUG( printf( "SRLZ emit symbol\n" ) );
			
			sgs_membuf_appchr( &pSD->data, C, 'S' );
			sgs_membuf_appbuf( &pSD->data, C, pSD->argarray.ptr + pSD->argarray.size - 4, 4 );
			sgs_membuf_erase( &pSD->argarray, pSD->argarray.size - 4, pSD->argarray.size );
			sgs_Release( C, &sym );
			
			goto addvar;
		}
		sgs_Release( C, &sym );
	}
	
	if( var.type == SGS_VTSPC_CLOSURE )
	{
		sgs_Closure* cl = (sgs_Closure*) var.data.P;
		
		SRLZ_DEBUG( printf( "SRLZ new closure\n" ) );
		
		sgs_SerializeInt_V2( C, cl->var );
		if( pSD->argarray.size < 4 )
		{
			/* error likely to be already printed */
			pSD->ret = SGS_FALSE;
			goto end;
		}
		
		SRLZ_DEBUG( printf( "SRLZ emit closure\n" ) );
		
		sgs_membuf_appchr( &pSD->data, C, 'C' );
		sgs_membuf_appbuf( &pSD->data, C, pSD->argarray.ptr + pSD->argarray.size - 4, 4 );
		sgs_membuf_erase( &pSD->argarray, pSD->argarray.size - 4, pSD->argarray.size );
	}
	else if( var.type == SGS_VT_THREAD )
	{
		SRLZ_DEBUG( printf( "SRLZ new thread\n" ) );
		
		if( !sgs__thread_serialize( C, var.data.T, &pSD->data, &pSD->argarray ) )
		{
			sgs_Msg( C, SGS_ERROR, "failed to serialize thread" );
			pSD->ret = SGS_FALSE;
			goto end;
		}
	}
	else if( var.type == SGS_VT_OBJECT )
	{
		int parg = C->object_arg;
		int32_t mo_arg = -1, prev_mo_arg = pSD->metaObjArg;
		sgs_VarObj* O = var.data.O, *MO = sgs_ObjGetMetaObj( var.data.O );
		sgs_VarObj* prevObj = pSD->curObj;
		_STACK_PREPARE;
		
		if( O->iface == sgsstd_closure_iface )
		{
			/* serialize closure (data=sgs_Variable func,sgs_clsrcount_t n,sgs_Closure*[n]) */
			char* ptr;
			sgs_clsrcount_t i, count;
			sgs_Closure** clsrlist;
			
			SRLZ_DEBUG( printf( "SRLZ new obj/closure\n" ) );
			
			ptr = (char*) O->data;
			count = *SGS_ASSUME_ALIGNED( ptr + sizeof(sgs_Variable), sgs_clsrcount_t );
			clsrlist = SGS_ASSUME_ALIGNED( ptr + sizeof(sgs_Variable) + sizeof(sgs_clsrcount_t), sgs_Closure* );
			
			/* write closure */
			{
				int32_t args = (int32_t) count;
				char pb[5] = { 'Q', 0 };
				{
					/* WP: they were pointless */
					pb[1] = (char)((args)&0xff);
					pb[2] = (char)((args>>8)&0xff);
					pb[3] = (char)((args>>16)&0xff);
					pb[4] = (char)((args>>24)&0xff);
				}
				
				SRLZ_DEBUG( printf( "SRLZ emit obj/closure (%d args)\n", (int) args ) );
				
				sgs_membuf_appbuf( &pSD->data, C, pb, 5 );
			}
			srlz_mode2_addvar( C, pSD, &var );
			
			SRLZ_DEBUG( printf( "SRLZ obj/closure > func\n" ) );
			sgs_SerializeInt_V2( C, *(sgs_Variable*) O->data );
			
			SRLZ_DEBUG( printf( "SRLZ emit obj/closure [f]unc\n" ) );
			sgs_membuf_appchr( &pSD->data, C, 'f' );
			sgs_membuf_appbuf( &pSD->data, C, pSD->argarray.ptr + pSD->argarray.size - 8, 8 );
			sgs_membuf_erase( &pSD->argarray, pSD->argarray.size - 4, pSD->argarray.size );
			
			for( i = 0; i < count; ++i )
			{
				uint32_t idx = (uint32_t) i;
				sgs_Variable tmp;
				
				SRLZ_DEBUG( printf( "SRLZ obj/closure > closure %d\n", (int) i ) );
				
				tmp.type = SGS_VTSPC_CLOSURE;
				tmp.data.P = clsrlist[ i ];
				sgs_SerializeInt_V2( C, tmp );
				
				SRLZ_DEBUG( printf( "SRLZ emit obj/closure [<] argument variable\n" ) );
				
				sgs_membuf_appchr( &pSD->data, C, '<' );
				sgs_membuf_appbuf( &pSD->data, C, &idx, sizeof(idx) );
				sgs_membuf_appbuf( &pSD->data, C, pSD->argarray.ptr + pSD->argarray.size - 8, 8 );
				sgs_membuf_erase( &pSD->argarray, pSD->argarray.size - 4, pSD->argarray.size );
			}
			
			goto end;
		}
		else
		{
			SRLZ_DEBUG( printf( "SRLZ new object (%p/%p/%s)\n", O->data, O->iface, O->iface->name ) );
			
			if( MO )
			{
				size_t origsize = pSD->argarray.size;
				(void) origsize;
				
				SRLZ_DEBUG( printf( "SRLZ obj metaobj\n" ) );
				
				sgs_SerializeInt_V2( C, sgs_MakeObjPtrNoRef( MO ) );
				/* save position of target object */
				SGS_AS_INT32( mo_arg, pSD->argarray.ptr + pSD->argarray.size - 4 );
				/* remove the meta-object from arguments */
				sgs_membuf_erase( &pSD->argarray, pSD->argarray.size - 4, pSD->argarray.size );
				/* check if argument array size was not changed */
				sgs_BreakIf( origsize != pSD->argarray.size );
			}
			
			if( !O->iface->serialize )
			{
				SRLZ_DEBUG( printf( "- no callback\n" ) );
				
				sgs_Msg( C, SGS_WARNING, "Cannot serialize object of type '%s'", O->iface->name );
				pSD->ret = SGS_FALSE;
				goto end;
			}
			pSD->curObj = O;
			_STACK_PROTECT;
			C->object_arg = 2;
			pSD->metaObjArg = mo_arg;
			pSD->ret = SGS_SUCCEEDED( O->iface->serialize( C, O ) );
			pSD->metaObjArg = prev_mo_arg;
			C->object_arg = parg;
			_STACK_UNPROTECT;
			pSD->curObj = prevObj;
			/* variable was added in SerializeObject */
			goto end;
		}
	}
	else if( var.type == SGS_VT_CFUNC )
	{
		SRLZ_DEBUG( printf( "SRLZ cfunc BAD\n" ) );
		sgs_Msg( C, SGS_WARNING, "Cannot serialize C functions" );
		pSD->ret = SGS_FALSE;
		goto end;
	}
	else if( var.type == SGS_VT_PTR )
	{
		SRLZ_DEBUG( printf( "SRLZ ptr BAD\n" ) );
		sgs_Msg( C, SGS_WARNING, "Cannot serialize pointers" );
		pSD->ret = SGS_FALSE;
		goto end;
	}
	else
	{
		char pb[2];
		{
			pb[0] = 'P';
			/* WP: basetype limited to bits 0-7, sign interpretation does not matter */
			pb[1] = (char) var.type;
		}
		
		SRLZ_DEBUG( printf( "SRLZ new/emit basic type (%d)\n", (int) var.type ) );
		
		sgs_membuf_appbuf( &pSD->data, C, pb, 2 );
		switch( var.type )
		{
		case SGS_VT_NULL: break;
		/* WP: var.data.B uses only bit 0 */
		case SGS_VT_BOOL: { char b = (char) var.data.B; sgs_membuf_appbuf( &pSD->data, C, &b, 1 ); } break;
		case SGS_VT_INT: sgs_membuf_appbuf( &pSD->data, C, &var.data.I, sizeof( sgs_Int ) ); break;
		case SGS_VT_REAL: sgs_membuf_appbuf( &pSD->data, C, &var.data.R, sizeof( sgs_Real ) ); break;
		case SGS_VT_STRING:
			sgs_membuf_appbuf( &pSD->data, C, &var.data.S->size, 4 );
			sgs_membuf_appbuf( &pSD->data, C, sgs_var_cstr( &var ), var.data.S->size );
			break;
		case SGS_VT_FUNC:
			{
				size_t szbefore = pSD->data.size;
				if( !_serialize_function( C, var.data.F, &pSD->data ) )
				{
					pSD->ret = SGS_FALSE;
					goto end;
				}
				else
				{
					uint32_t szdiff = (uint32_t) ( pSD->data.size - szbefore );
					sgs_membuf_insbuf( &pSD->data, C, szbefore, &szdiff, sizeof(szdiff) );
				}
			}
			break;
		default:
			sgs_Msg( C, SGS_ERROR, "sgs_Serialize: unknown memory error" );
			pSD->ret = SGS_FALSE;
			goto end;
		}
	}
	
addvar:
	srlz_mode2_addvar( C, pSD, &var );
end:
	if( ep )
	{
		SRLZ_DEBUG( printf( "SRLZ -- mode 2 AT END --\n" ) );
		
		if( SD.argarray.size == 0 )
			SD.ret = 0;
		else
		{
			SRLZ_DEBUG( printf( "SRLZ emit return\n" ) );
			
			sgs_membuf_appchr( &SD.data, C, 'R' );
			sgs_membuf_appbuf( &SD.data, C, SD.argarray.ptr + SD.argarray.size - 4, 4 );
		}
		if( SD.ret )
		{
			if( SD.data.size > 0x7fffffff )
			{
				sgs_Msg( C, SGS_ERROR, "serialized string too long" );
				SD.ret = SGS_FALSE;
			}
			else
			{
				/* WP: added error condition */
				sgs_PushStringBuf( C, SD.data.ptr, (sgs_SizeVal) SD.data.size );
			}
		}
		if( SD.ret == SGS_FALSE )
		{
			sgs_PushNull( C );
		}
		sgs_vht_free( &SD.servartable, C );
		sgs_membuf_destroy( &SD.argarray, C );
		sgs_membuf_destroy( &SD.data, C );
		C->serialize_state = prev_serialize_state;
	}
}

void sgs_SerializeObjectInt_V2( SGS_CTX, sgs_StkIdx args, const char* func, size_t fnsize )
{
	size_t argsize;
	sgs_serialize2_data* pSD = (sgs_serialize2_data*) C->serialize_state;
	
	if( args < 0 || (size_t) args > pSD->argarray.size / sizeof(sgs_StkIdx) )
	{
		sgs_Msg( C, SGS_APIERR, "sgs_SerializeObject: specified "
			"more arguments than there are serialized items" );
		return;
	}
	
	{
		char pb[11] = { 'O', 0 };
		{
			/* WP: they were pointless */
			pb[1] = (char)((args)&0xff);
			pb[2] = (char)((args>>8)&0xff);
			pb[3] = (char)((args>>16)&0xff);
			pb[4] = (char)((args>>24)&0xff);
			/* WP: have error condition + sign interpretation doesn't matter */
			pb[5] = (char) fnsize;
			pb[6] = (char) sgs_ObjGetMetaMethodEnable( pSD->curObj );
			memcpy( &pb[7], &pSD->metaObjArg, 4 );
		}
		
		sgs_membuf_appbuf( &pSD->data, C, pb, 11 );
	}
	
	/* write args */
	argsize = sizeof(sgs_StkIdx) * (size_t) args; /* WP: added error condition */
	sgs_membuf_appbuf( &pSD->data, C, pSD->argarray.ptr + pSD->argarray.size - argsize, argsize );
	/* pop written args */
	sgs_membuf_erase( &pSD->argarray, pSD->argarray.size - argsize, pSD->argarray.size );
	/* write function name */
	sgs_membuf_appbuf( &pSD->data, C, func, fnsize );
	sgs_membuf_appchr( &pSD->data, C, '\0' );
	/* register the object variable */
	{
		sgs_Variable objvar;
		objvar.type = SGS_VT_OBJECT;
		objvar.data.O = pSD->curObj;
		srlz_mode2_addvar( C, pSD, &objvar );
	}
}

static void sgs_SerializeObjIndexInt_V2( SGS_CTX, int isprop )
{
	sgs_serialize2_data* pSD = (sgs_serialize2_data*) C->serialize_state;
	
	if( 3 > pSD->argarray.size / sizeof(sgs_StkIdx) )
	{
		sgs_Msg( C, SGS_APIERR, "sgs_SerializeObjIndex: less than 3 serialized items found" );
		return;
	}
	/* expecting 3 arguments: object, key, value */
	sgs_membuf_appchr( &pSD->data, C, isprop ? '.' : '[' );
	sgs_membuf_appbuf( &pSD->data, C, pSD->argarray.ptr + pSD->argarray.size - 4 * 3, 4 * 3 );
	/* remove only 2 of them */
	sgs_membuf_erase( &pSD->argarray, pSD->argarray.size - 4 * 2, pSD->argarray.size );
}

SGSBOOL sgs_UnserializeInt_V2( SGS_CTX, char* str, char* strend )
{
	int32_t retpos = -1;
	sgs_Variable var;
	SGSBOOL res = SGS_FALSE;
	sgs_MemBuf mb = sgs_membuf_create();
	_STACK_PREPARE;
	
	SRLZ_DEBUG( printf( "USRZ --- mode 2 START ---\n" ) );
	
	_STACK_PROTECT;
	while( str < strend )
	{
		char c = *str++;
		if( c == 'P' )
		{
			SRLZ_DEBUG( printf( "USRZ found [P]lain variable\n" ) );
			
			if( str >= strend && !sgs_unserr_incomp( C ) )
				goto fail;
			c = *str++;
			SRLZ_DEBUG( printf( "- type %d\n", c ) );
			switch( c )
			{
			case SGS_VT_NULL: var.type = SGS_VT_NULL; break;
			case SGS_VT_BOOL:
				if( str >= strend && !sgs_unserr_incomp( C ) )
					goto fail;
				var.type = SGS_VT_BOOL;
				var.data.B = *str++ != 0;
				break;
			case SGS_VT_INT:
				if( str >= strend-7 && !sgs_unserr_incomp( C ) )
					goto fail;
				else
				{
					sgs_Int val;
					SGS_AS_INTEGER( val, str );
					var.type = SGS_VT_INT;
					var.data.I = val;
				}
				str += 8;
				break;
			case SGS_VT_REAL:
				if( str >= strend-7 && !sgs_unserr_incomp( C ) )
					goto fail;
				else
				{
					sgs_Real val;
					SGS_AS_REAL( val, str );
					var.type = SGS_VT_REAL;
					var.data.R = val;
				}
				str += 8;
				break;
			case SGS_VT_STRING:
				{
					sgs_SizeVal strsz;
					if( str >= strend-3 && !sgs_unserr_incomp( C ) )
						goto fail;
					SGS_AS_INT32( strsz, str );
					str += 4;
					if( str > strend - strsz && !sgs_unserr_incomp( C ) )
						goto fail;
					sgs_InitStringBuf( C, &var, str, strsz );
					str += strsz;
				}
				break;
			case SGS_VT_FUNC:
				{
					sgs_SizeVal bcsz;
					if( str >= strend-3 && !sgs_unserr_incomp( C ) )
						goto fail;
					SGS_AS_INT32( bcsz, str );
					str += 4;
					if( str > strend - bcsz && !sgs_unserr_incomp( C ) )
						goto fail;
					/* WP: conversion does not affect values */
					if( !_unserialize_function( C, str, (size_t) bcsz, &var ) )
						goto fail; /* error already printed */
					str += bcsz;
				}
				break;
			default:
				sgs_unserr_error( C );
				goto fail;
			}
		}
		else if( c == 'O' )
		{
			sgs_StkIdx subsz;
			int32_t i, pos, argc, mo_arg;
			int fnsz, mm_enable, ret;
			
			SRLZ_DEBUG( printf( "USRZ found [O]bject\n" ) );
			
			if( str > strend-5 && !sgs_unserr_incomp( C ) )
				goto fail;
			SGS_AS_INT32( argc, str );
			SRLZ_DEBUG( printf( "- %d args\n", argc ) );
			str += 4;
			fnsz = *str++ + 1;
			mm_enable = *str++ != 0;
			SGS_AS_INT32( mo_arg, str );
			str += 4;
			for( i = 0; i < argc; ++i )
			{
				if( str > strend-4 && !sgs_unserr_incomp( C ) )
					goto fail;
				SGS_AS_INT32( pos, str );
				str += 4;
				if( pos < 0 || (size_t) pos >= mb.size / sizeof(sgs_Variable) )
				{
					sgs_unserr_error( C );
					goto fail;
				}
				sgs_PushVariable( C, (SGS_ASSUME_ALIGNED( mb.ptr, sgs_Variable ))[ pos ] );
			}
			if( str > strend - fnsz && !sgs_unserr_incomp( C ) )
				goto fail;
			
			SRLZ_DEBUG( printf( "- args=%d mmenable=%s func=%s\n", argc, mm_enable ? "Y" : "n", str ) );
			
			subsz = sgs_StackSize( C ) - argc;
			ret = SGS_SUCCEEDED( sgs_GlobalCall( C, str, argc, 1 ) );
			if( ret == SGS_FALSE || sgs_StackSize( C ) - subsz < 1 || sgs_ItemType( C, -1 ) != SGS_VT_OBJECT )
			{
				sgs_unserr_objcall( C );
				goto fail;
			}
			sgs_GetStackItem( C, -1, &var );
			sgs_ObjSetMetaMethodEnable( var.data.O, mm_enable );
			if( mo_arg < 0 )
				sgs_ObjSetMetaObj( C, var.data.O, NULL );
			else /* if( c == '3' ) */
			{
				sgs_Variable* mov;
				/* add meta-object */
				if( mo_arg < 0 || mo_arg >= (int32_t) ( mb.size / sizeof( sgs_Variable ) ) )
				{
					sgs_unserr_error( C );
					goto fail;
				}
				mov = &(SGS_ASSUME_ALIGNED( mb.ptr, sgs_Variable ))[ mo_arg ];
				if( mov->type != SGS_VT_OBJECT && !sgs_unserr_error( C ) )
					goto fail;
				if( mov->data.O == var.data.O && !sgs_unserr_error( C ) )
					goto fail;
				sgs_ObjSetMetaObj( C, var.data.O, mov->data.O );
			}
			sgs_SetStackSize( C, subsz );
			str += fnsz;
		}
		else if( c == 'C' )
		{
			/* single closure (sgs_Closure* in P) */
			sgs_Closure* clsr;
			int32_t pos;
			
			SRLZ_DEBUG( printf( "USRZ found single [C]losure\n" ) );
			
			if( str > strend-4 && !sgs_unserr_incomp( C ) )
				goto fail;
			SGS_AS_INT32( pos, str );
			
			SRLZ_DEBUG( printf( "- var=%d\n", (int) pos ) );
			
			str += 4;
			clsr = sgs_Alloc( sgs_Closure );
			clsr->refcount = 0;
			clsr->var = (SGS_ASSUME_ALIGNED( mb.ptr, sgs_Variable ))[ pos ];
			sgs_Acquire( C, &clsr->var );
			var.type = SGS_VTSPC_CLOSURE;
			var.data.P = clsr;
		}
		else if( c == 'Q' )
		{
			/* closure object */
			int32_t argc;
			sgs_Variable funcvar;
			
			SRLZ_DEBUG( printf( "USRZ found [Q] (closure object)\n" ) );
			
			if( str > strend-4 && !sgs_unserr_incomp( C ) )
				goto fail;
			SGS_AS_INT32( argc, str );
			str += 4;
			
			SRLZ_DEBUG( printf( "- %d args\n", argc ) );
			
			/* create the closure object */
			funcvar.type = SGS_VT_NULL;
			sgsSTD_MakeClosure( C, &var, &funcvar, (size_t) argc );
		}
		else if( c == 'f' )
		{
			sgs_Variable *cvp, *fvp;
			int32_t pos, func;
			
			SRLZ_DEBUG( printf( "USRZ found [f] (closure function)\n" ) );
			
			if( str > strend-8 && !sgs_unserr_incomp( C ) )
				goto fail;
			SGS_AS_INT32( pos, str ); str += 4;
			SGS_AS_INT32( func, str ); str += 4;
			
			SRLZ_DEBUG( printf( "- closure=%d func=%d\n", pos, func ) );
			
			if( pos < 0 || (size_t) pos >= mb.size / sizeof(sgs_Variable) ||
				func < 0 || (size_t) func >= mb.size / sizeof(sgs_Variable) )
			{
				sgs_unserr_error( C );
				goto fail;
			}
			
			cvp = &(SGS_ASSUME_ALIGNED( mb.ptr, sgs_Variable ))[ pos ];
			fvp = &(SGS_ASSUME_ALIGNED( mb.ptr, sgs_Variable ))[ func ];
			if( cvp->type != SGS_VT_OBJECT || cvp->data.O->iface != sgsstd_closure_iface )
			{
				sgs_unserr_error( C );
				goto fail;
			}
			*(sgs_Variable*)cvp->data.O->data = *fvp;
			VAR_ACQUIRE( fvp );
			continue;
		}
		else if( c == '<' )
		{
			sgs_Variable *cvp, *avp;
			int32_t which, cvpos, avpos;
			
			SRLZ_DEBUG( printf( "USRZ found [<] (closure variable)\n" ) );
			
			if( str > strend-12 && !sgs_unserr_incomp( C ) )
				goto fail;
			SGS_AS_INT32( which, str ); str += 4;
			SGS_AS_INT32( cvpos, str ); str += 4;
			SGS_AS_INT32( avpos, str ); str += 4;
			
			SRLZ_DEBUG( printf( "- closure=%d arg=%d avpos=%d\n", cvpos, which, avpos ) );
			
			if( cvpos < 0 || (size_t) cvpos >= mb.size / sizeof(sgs_Variable) ||
				avpos < 0 || (size_t) avpos >= mb.size / sizeof(sgs_Variable) )
			{
				sgs_unserr_error( C );
				goto fail;
			}
			
			cvp = &(SGS_ASSUME_ALIGNED( mb.ptr, sgs_Variable ))[ cvpos ];
			avp = &(SGS_ASSUME_ALIGNED( mb.ptr, sgs_Variable ))[ avpos ];
			if( cvp->type != SGS_VT_OBJECT || cvp->data.O->iface != sgsstd_closure_iface ||
				avp->type != SGS_VTSPC_CLOSURE )
			{
				sgs_unserr_error( C );
				goto fail;
			}
			
			{
				char* ptr = (char*) cvp->data.O->data;
				sgs_clsrcount_t count = *SGS_ASSUME_ALIGNED( ptr + sizeof(sgs_Variable), sgs_clsrcount_t );
				sgs_Closure** clsrlist = SGS_ASSUME_ALIGNED( ptr + sizeof(sgs_Variable) + sizeof(sgs_clsrcount_t), sgs_Closure* );
				if( which < 0 || (sgs_clsrcount_t) which >= count )
				{
					sgs_unserr_error( C );
					goto fail;
				}
				clsrlist[ which ] = (sgs_Closure*) avp->data.P;
				clsrlist[ which ]->refcount++;
			}
			continue;
		}
		else if( c == 'T' )
		{
			int32_t i, pos, argc;
			sgs_Context* T = NULL;
			
			SRLZ_DEBUG( printf( "USRZ found [T]hread\n" ) );
			
			if( str > strend-5 && !sgs_unserr_incomp( C ) )
				goto fail;
			SGS_AS_INT32( argc, str );
			SRLZ_DEBUG( printf( "- %d args\n", argc ) );
			str += 4;
			for( i = 0; i < argc; ++i )
			{
				if( str > strend-4 && !sgs_unserr_incomp( C ) )
					goto fail;
				SGS_AS_INT32( pos, str );
				
				SRLZ_DEBUG( printf( "- arg %d = %d\n", (int) i, (int) pos ) );
				
				str += 4;
				if( pos < 0 || (size_t) pos >= mb.size / sizeof(sgs_Variable) )
				{
					sgs_unserr_error( C );
					goto fail;
				}
				sgs_PushVariable( C, (SGS_ASSUME_ALIGNED( mb.ptr, sgs_Variable ))[ pos ] );
			}
			if( !sgs__thread_unserialize( C, &T, &str, strend ) && !sgs_unserr_incomp( C ) )
				goto fail;
			sgs_Pop( C, argc );
			sgs_InitThreadPtr( &var, T );
		}
		else if( c == 'S' )
		{
			int32_t pos;
			
			SRLZ_DEBUG( printf( "USRZ found [S]ymbol\n" ) );
			
			if( str > strend-4 && !sgs_unserr_incomp( C ) )
				goto fail;
			SGS_AS_INT32( pos, str );
			str += 4;
			
			SRLZ_DEBUG( printf( "- var=%d\n", (int) pos ) );
			
			if( pos < 0 || pos >= (int32_t) ( mb.size / sizeof( sgs_Variable ) ) )
			{
				sgs_unserr_error( C );
				goto fail;
			}
			if( !sgs_GetSymbol( C, (SGS_ASSUME_ALIGNED( mb.ptr, sgs_Variable ))[ pos ], &var ) )
			{
				sgs_unserr_symfail( C );
				goto fail;
			}
		}
		else if( c == '.' || c == '[' )
		{
			int32_t pobj, pkey, pval, pend = (int32_t) ( mb.size / sizeof( sgs_Variable ) );
			sgs_Variable* varlist = SGS_ASSUME_ALIGNED( mb.ptr, sgs_Variable );
			
			SRLZ_DEBUG( printf( "USRZ found [%c] (post-%s)\n", c, c == '.' ? "properties" : "indices" ) );
			
			if( str > strend-12 && !sgs_unserr_incomp( C ) )
				goto fail;
			SGS_AS_INT32( pobj, str );
			SGS_AS_INT32( pkey, str + 4 );
			SGS_AS_INT32( pval, str + 8 );
			str += 12;
			if( pobj < 0 || pobj >= pend ||
				pkey < 0 || pkey >= pend ||
				pval < 0 || pval >= pend )
			{
				sgs_unserr_error( C );
				goto fail;
			}
			sgs_SetIndex( C, varlist[ pobj ], varlist[ pkey ], varlist[ pval ], c == '.' );
			continue;
		}
		else if( c == 'R' )
		{
			SRLZ_DEBUG( printf( "USRZ found [R]eturn value\n" ) );
			
			if( str > strend-4 && !sgs_unserr_incomp( C ) )
				goto fail;
			SGS_AS_INT32( retpos, str );
			str += 4;
			
			SRLZ_DEBUG( printf( "- var=%d\n", (int) retpos ) );
			
			if( retpos < 0 || retpos >= (int32_t) ( mb.size / sizeof( sgs_Variable ) ) )
			{
				sgs_unserr_error( C );
				goto fail;
			}
			continue;
		}
		else
		{
			sgs_unserr_error( C );
			goto fail;
		}
		SRLZ_DEBUG( printf( "^ new var %d (type=%d)\n", (int)( mb.size / sizeof(var) ), var.type ) );
		SRLZ_DEBUG( if(var.type == SGS_VT_OBJECT) printf( "^ iface = %s\n", var.data.O->iface->name ) );
		sgs_membuf_appbuf( &mb, C, &var, sizeof(var) );
	}
	
	SRLZ_DEBUG( printf( "USRZ === mode 2 END ===\n" ) );
	if( mb.size )
	{
		if( retpos >= 0 )
			sgs_PushVariable( C, (SGS_ASSUME_ALIGNED( mb.ptr, sgs_Variable ))[ retpos ] );
		else
			sgs_PushVariable( C, *SGS_ASSUME_ALIGNED( mb.ptr + mb.size - sizeof(sgs_Variable), sgs_Variable ) );
	}
	res = mb.size != 0;
fail:
	_STACK_UNPROTECT_SKIP( res );
	{
		sgs_Variable* ptr = SGS_ASSUME_ALIGNED( mb.ptr, sgs_Variable );
		sgs_Variable* pend = SGS_ASSUME_ALIGNED( mb.ptr + mb.size, sgs_Variable );
		while( ptr < pend )
		{
			sgs_Release( C, ptr++ );
		}
	}
	sgs_membuf_destroy( &mb, C );
	return res;
}


/* Serialization mode 3 - text
	- [...] - array
	- {...} - dict
	- map{...} - map
	- <ident>(...) - custom object
*/
typedef struct s3callinfo
{
	sgs_Variable func_name; /* string */
	int32_t arg_offset;
	int32_t arg_count;
}
s3callinfo;

typedef struct sgs_serialize3_data
{
	int mode;
	/* serialized variable set, also maps objects to call info, offset at value.data.I */
	sgs_VHTable servartable;
	/* s3callinfo[], contains function call info */
	sgs_MemBuf callinfo;
	/* int32_t[], contains function call arguments, first after offset: int count */
	sgs_MemBuf callargs;
	/* sgs_Variable[], contains arguments for sgs_SerializeObject */
	sgs_MemBuf argarray;
	/* the current serialized object */
	sgs_VarObj* curObj;
	/* output byte array */
	sgs_MemBuf data;
}
sgs_serialize3_data;

#define sgson_tab( buf, C, depth, tab, tablen ) if( tab ){ \
	int _i = depth; \
	sgs_membuf_appchr( buf, C, '\n' ); \
	while( _i-- > 0 ) sgs_membuf_appbuf( buf, C, tab, (size_t) tablen ); }

#define sgs_tohex( c ) ("0123456789ABCDEF"[ (c) & 0x0f ])

#define sgs_isid( c ) ( sgs_isalnum( c ) || (c) == '_' )

static int sgson_encode_var( SGS_CTX, sgs_serialize3_data* data,
	int depth, const char* tab, sgs_SizeVal tablen )
{
	sgs_MemBuf* buf = &data->data;
	sgs_Variable var = sgs_StackItem( C, -1 );
	switch( var.type )
	{
	case SGS_VT_NULL:
		sgs_membuf_appbuf( buf, C, "null", 4 );
		return 1;
	case SGS_VT_BOOL:
		sgs_membuf_appbuf( buf, C, var.data.B ? "true" : "false", var.data.B ? 4 : 5 );
		return 1;
	case SGS_VT_INT:
		{
			char tmp[ 64 ];
			sprintf( tmp, "%" PRId64, var.data.I );
			sgs_membuf_appbuf( buf, C, tmp, strlen( tmp ) );
			return 1;
		}
	case SGS_VT_REAL:
		{
			char tmp[ 64 ];
			sprintf( tmp, "%g", var.data.R );
			sgs_membuf_appbuf( buf, C, tmp, strlen( tmp ) );
			return 1;
		}
	case SGS_VT_STRING:
		{
			char* str = sgs_GetStringPtr( C, -1 );
			char* frm = str, *end = str + sgs_GetStringSize( C, -1 );
			sgs_membuf_appchr( buf, C, '"' );
			{
				while( str < end )
				{
					if( *str == '"' || *str == '\\' )
					{
						char pp[2];
						{
							pp[0] = '\\';
							pp[1] = *str;
						}
						if( str != frm )
							sgs_membuf_appbuf( buf, C, frm, (size_t) ( str - frm ) );
						sgs_membuf_appbuf( buf, C, pp, 2 );
						frm = str + 1;
					}
					else if( *str < 0x20 || *str == 0x7f )
					{
						size_t len = 2;
						char pp[4];
						pp[0] = '\\';
						if( *str == '\n' ){ pp[1] = 'n'; }
						else if( *str == '\r' ){ pp[1] = 'r'; }
						else if( *str == '\t' ){ pp[1] = 't'; }
						else
						{
							pp[1] = 'x';
							pp[2] = sgs_tohex( *str >> 4 );
							pp[3] = sgs_tohex( *str );
							len = 4;
						}
						if( str != frm )
							sgs_membuf_appbuf( buf, C, frm, (size_t) ( str - frm ) );
						sgs_membuf_appbuf( buf, C, pp, len );
						frm = str + 1;
					}
					str++;
				}
				if( str != frm )
					sgs_membuf_appbuf( buf, C, frm, (size_t) ( str - frm ) );
			}
			sgs_membuf_appchr( buf, C, '"' );
			return 1;
		}
	case SGS_VT_FUNC:
	case SGS_VT_CFUNC:
	case SGS_VT_PTR:
	case SGS_VT_THREAD:
	case SGS_VT_OBJECT:
		/* if one of these exists, it has call info */
		{
			int32_t call_info_index, *args, i;
			s3callinfo* ci;
			const char* fn_name;
			size_t fn_size;
			
			sgs_VHTVar* vv = sgs_vht_get( &data->servartable, &var );
			sgs_BreakIf( vv == NULL );
			
			if( vv->val.type != SGS_VT_INT )
			{
				/* error probably already printed */
				sgs_membuf_appbuf( buf, C, "null", 4 );
				return 1;
			}
			
			call_info_index = (int32_t) vv->val.data.I;
			ci = &(SGS_ASSUME_ALIGNED( data->callinfo.ptr, s3callinfo ))[ call_info_index ];
			args = &(SGS_ASSUME_ALIGNED( data->callargs.ptr, int32_t ))[ ci->arg_offset ];
			fn_name = sgs_var_cstr( &ci->func_name );
			fn_size = (size_t) ci->func_name.data.S->size;
			
			if( fn_size == 5 && !strcmp( fn_name, "array" ) )
			{
				sgs_membuf_appchr( buf, C, '[' );
				depth++;
				for( i = 0; i < ci->arg_count; ++i )
				{
					if( i )
						sgs_membuf_appchr( buf, C, ',' );
					sgson_tab( buf, C, depth, tab, tablen );
					
					sgs_PushVariable( C, data->servartable.vars[ args[ i ] ].key );
					if( !sgson_encode_var( C, data, depth, tab, tablen ) )
						return 0;
					sgs_Pop( C, 1 );
				}
				depth--;
				if( ci->arg_count )
					sgson_tab( buf, C, depth, tab, tablen );
				sgs_membuf_appchr( buf, C, ']' );
			}
			else if( ( fn_size == 4 && !strcmp( fn_name, "dict" ) )
				|| ( fn_size == 3 && !strcmp( fn_name, "map" ) ) )
			{
				if( fn_size == 3 ) /* if map */
					sgs_membuf_appbuf( buf, C, fn_name, fn_size );
				sgs_membuf_appchr( buf, C, '{' );
				depth++;
				for( i = 0; i < ci->arg_count; i += 2 )
				{
					if( i )
						sgs_membuf_appchr( buf, C, ',' );
					sgson_tab( buf, C, depth, tab, tablen );
					
					/* key */
					{
						sgs_Variable key = data->servartable.vars[ args[ i ] ].key;
						int wrotekey = 0;
						/* small identifier optimization */
						if( key.type == SGS_VT_STRING )
						{
							char* str = sgs_GetStringPtrP( &key );
							char* end = str + sgs_GetStringSizeP( &key );
							if( end - str <= 32 && end - str > 0 && ( *str == '_' || sgs_isalpha( *str ) ) )
							{
								char* cc = str + 1;
								while( cc < end )
								{
									if( !sgs_isalnum( *cc ) && *cc != '_' )
										break;
									cc++;
								}
								if( cc == end )
								{
									/* only small identifiers */
									sgs_membuf_appbuf( buf, C, str, (size_t)( end - str ) );
									wrotekey = 1;
								}
							}
						}
						
						if( wrotekey == 0 )
						{
							if( key.type != SGS_VT_STRING )
								sgs_membuf_appchr( buf, C, '[' );
							
							sgs_PushVariable( C, key );
							if( !sgson_encode_var( C, data, depth, tab, tablen ) )
								return 0;
							sgs_Pop( C, 1 );
							
							if( key.type != SGS_VT_STRING )
								sgs_membuf_appchr( buf, C, ']' );
						}
					}
					
					/* = */
					if( tab )
						sgs_membuf_appbuf( buf, C, " = ", 3 );
					else
						sgs_membuf_appchr( buf, C, '=' );
					
					/* value */
					sgs_PushVariable( C, data->servartable.vars[ args[ i + 1 ] ].key );
					if( !sgson_encode_var( C, data, depth, tab, tablen ) )
						return 0;
					sgs_Pop( C, 1 );
				}
				depth--;
				if( ci->arg_count )
					sgson_tab( buf, C, depth, tab, tablen );
				sgs_membuf_appchr( buf, C, '}' );
			}
			else /* function call */
			{
				sgs_membuf_appbuf( buf, C, fn_name, fn_size );
				sgs_membuf_appchr( buf, C, '(' );
				for( i = 0; i < ci->arg_count; ++i )
				{
					if( i )
						sgs_membuf_appbuf( buf, C, ", ", 2 );
					
					sgs_PushVariable( C, data->servartable.vars[ args[ i ] ].key );
					if( !sgson_encode_var( C, data, depth, tab, tablen ) )
						return 0;
					sgs_Pop( C, 1 );
				}
				sgs_membuf_appchr( buf, C, ')' );
			}
			return 1;
		}
	}
	return 0;
}

void sgs_SerializeInt_V3( SGS_CTX, sgs_Variable var, const char* tab, sgs_SizeVal tablen )
{
	int ret = SGS_TRUE;
	void* prev_serialize_state = C->serialize_state;
	sgs_serialize3_data SD, *pSD;
	int ep = !C->serialize_state || *(int*)C->serialize_state != 3;
	
	if( ep )
	{
		SD.mode = 3;
		sgs_vht_init( &SD.servartable, C, 64, 64 );
		SD.callinfo = sgs_membuf_create();
		SD.callargs = sgs_membuf_create();
		SD.argarray = sgs_membuf_create();
		SD.data = sgs_membuf_create();
		SD.curObj = NULL;
		C->serialize_state = &SD;
	}
	pSD = (sgs_serialize3_data*) C->serialize_state;
	
	/* SYMBOLS */
	if( var.type == SGS_VT_OBJECT || var.type == SGS_VT_CFUNC ||
		var.type == SGS_VT_FUNC || var.type == SGS_VT_THREAD || var.type == SGS_VT_PTR )
	{
		int32_t argidx;
		sgs_VHTVar* vv = sgs_vht_get( &pSD->servartable, &var );
		if( vv )
		{
			argidx = (int32_t) ( vv - pSD->servartable.vars );
			sgs_membuf_appbuf( &pSD->argarray, C, &argidx, sizeof(argidx) );
			goto fail;
		}
		else
		{
			sgs_Variable sym = sgs_MakeNull();
			if( sgs_GetSymbol( C, var, &sym ) && sym.type == SGS_VT_STRING )
			{
				int32_t call_info_offset = (int32_t) ( pSD->callinfo.size / sizeof(s3callinfo) );
				int32_t call_args_offset = (int32_t) ( pSD->callargs.size / 4 );
				s3callinfo ci = { sgs_MakeNull(), call_args_offset, 1 };
				sgs_InitString( C, &ci.func_name, "sym_get" );
				
				/* this is expected to succeed (type=string) */
				sgs_SerializeInt_V3( C, sym, tab, tablen );
				sgs_Release( C, &sym );
				
				/* append sym_get call argument */
				sgs_membuf_appbuf( &pSD->callargs, C, pSD->argarray.ptr + pSD->argarray.size - 4, 4 );
				/* pop argument */
				sgs_membuf_erase( &pSD->argarray, pSD->argarray.size - sizeof(argidx), pSD->argarray.size );
				/* append sym_get call info */
				sgs_membuf_appbuf( &pSD->callinfo, C, &ci, sizeof(ci) );
				
				/* create variable resolve */
				sgs_Variable idxvar;
				argidx = sgs_vht_size( &pSD->servartable );
				idxvar.type = SGS_VT_INT;
				idxvar.data.I = call_info_offset;
				sgs_vht_set( &pSD->servartable, C, &var, &idxvar );
				
				/* push new argument */
				sgs_membuf_appbuf( &pSD->argarray, C, &argidx, sizeof(argidx) );
				goto fail;
			}
			sgs_Release( C, &sym );
		}
	}
	
	/* SPECIAL TYPES */
	if( var.type == SGS_VT_THREAD )
	{
		sgs_Msg( C, SGS_WARNING, "serialization mode 3 (SGSON text)"
			" does not support thread serialization" );
		ret = 0;
		goto fail;
	}
	else if( var.type == SGS_VT_FUNC || var.type == SGS_VT_CFUNC )
	{
		sgs_Msg( C, SGS_WARNING, "serialization mode 3 (SGSON text)"
			" does not support function serialization" );
		ret = 0;
		goto fail;
	}
	else if( var.type == SGS_VT_PTR )
	{
		sgs_Msg( C, SGS_WARNING, "serialization mode 3 (SGSON text)"
			" does not support pointer serialization" );
		ret = 0;
		goto fail;
	}
	
	/* OBJECTS */
	if( var.type == SGS_VT_OBJECT )
	{
		sgs_VarObj* O = var.data.O;
		sgs_VarObj* prevObj = pSD->curObj;
		_STACK_PREPARE;
		if( !O->iface->serialize )
		{
			sgs_Msg( C, SGS_WARNING, "cannot serialize object of type '%s'", O->iface->name );
			var = sgs_MakeNull();
		}
		else
		{
			int parg = C->object_arg;
			pSD->curObj = O;
			_STACK_PROTECT;
			C->object_arg = 3;
			ret = SGS_SUCCEEDED( O->iface->serialize( C, O ) );
			C->object_arg = parg;
			_STACK_UNPROTECT;
			pSD->curObj = prevObj;
			if( ret == SGS_FALSE )
			{
				sgs_Msg( C, SGS_ERROR, "failed to serialize object of type '%s'", O->iface->name );
				goto fail;
			}
		}
	}
	else /* null, bool, int, real, string */
	{
		int32_t argidx;
		sgs_VHTVar* vv = sgs_vht_get( &pSD->servartable, &var );
		if( vv )
			argidx = (int32_t) ( vv - pSD->servartable.vars );
		else
		{
			sgs_Variable val = sgs_MakeNull();
			argidx = sgs_vht_size( &pSD->servartable );
			sgs_vht_set( &pSD->servartable, C, &var, &val );
		}
		sgs_membuf_appbuf( &pSD->argarray, C, &argidx, sizeof(argidx) );
	}
	
fail:
	if( ep )
	{
		/* serialize the variable tree */
		if( ret )
		{
			sgs_PushVariable( C, var );
			ret = sgson_encode_var( C, &SD, 0, tab, tablen );
			if( SD.data.size > 0x7fffffff )
			{
				ret = 0;
				sgs_Msg( C, SGS_WARNING, "generated more string data than allowed to store" );
			}
			sgs_Pop( C, 1 );
			if( ret )
				sgs_PushStringBuf( C, SD.data.ptr, (sgs_SizeVal) SD.data.size );
		}
		if( !ret )
			sgs_PushNull( C );
		
		/* free the serialization state */
		sgs_vht_free( &SD.servartable, C );
		sgs_membuf_destroy( &SD.argarray, C );
		{
			s3callinfo* ci = SGS_ASSUME_ALIGNED( SD.callinfo.ptr, s3callinfo );
			s3callinfo* ciend = SGS_ASSUME_ALIGNED( SD.callinfo.ptr + SD.callinfo.size, s3callinfo );
			while( ci < ciend )
			{
				sgs_Release( C, &ci->func_name );
				ci++;
			}
		}
		sgs_membuf_destroy( &SD.callinfo, C );
		sgs_membuf_destroy( &SD.callargs, C );
		sgs_membuf_destroy( &SD.data, C );
		C->serialize_state = prev_serialize_state;
	}
}

static void sgs_SerializeObjectInt_V3( SGS_CTX, sgs_StkIdx args, const char* func, size_t fnsize )
{
	size_t argsize;
	sgs_VHTVar* vv;
	sgs_Variable V;
	int32_t argidx;
	sgs_serialize3_data* pSD = (sgs_serialize3_data*) C->serialize_state;
	
	V.type = SGS_VT_OBJECT;
	V.data.O = pSD->curObj;
	if( args < 0 || (size_t) args > pSD->argarray.size / sizeof(int32_t) )
	{
		sgs_Variable idxvar = sgs_MakeNull();
		sgs_Msg( C, SGS_APIERR, "sgs_SerializeObject: specified "
			"more arguments than there are serialized items" );
		sgs_vht_set( &pSD->servartable, C, &V, &idxvar );
		return;
	}
	/* WP: added error condition */
	argsize = sizeof(int32_t) * (size_t) args;
	
	vv = sgs_vht_get( &pSD->servartable, &V );
	if( vv )
		argidx = (int32_t) ( vv - pSD->servartable.vars );
	else
	{
		int32_t call_info_offset = (int32_t) ( pSD->callinfo.size / sizeof(s3callinfo) );
		int32_t call_args_offset = (int32_t) ( pSD->callargs.size / 4 );
		s3callinfo ci = { sgs_MakeNull(), call_args_offset, args };
		sgs_InitStringBuf( C, &ci.func_name, func, (sgs_SizeVal) fnsize );
		
		/* append call arguments */
		sgs_membuf_appbuf( &pSD->callargs, C, pSD->argarray.ptr + pSD->argarray.size - argsize, argsize );
		/* append call info */
		sgs_membuf_appbuf( &pSD->callinfo, C, &ci, sizeof(ci) );
		
		/* create variable resolve */
		sgs_Variable idxvar;
		argidx = sgs_vht_size( &pSD->servartable );
		idxvar.type = SGS_VT_INT;
		idxvar.data.I = call_info_offset;
		sgs_vht_set( &pSD->servartable, C, &V, &idxvar );
	}
	/* pop arguments */
	sgs_membuf_erase( &pSD->argarray, pSD->argarray.size - argsize, pSD->argarray.size );
	/* append object as new argument */
	sgs_membuf_appbuf( &pSD->argarray, C, &argidx, sizeof(argidx) );
}

static int sgs_UnserializeInt_V3( SGS_CTX, char* str, char* strend )
{
	int res;
	sgs_MemBuf stack = sgs_membuf_create();
	sgs_membuf_appchr( &stack, C, 0 );
	res = !sgson_parse( C, &stack, str, (sgs_SizeVal) ( strend - str ) );
	sgs_membuf_destroy( &stack, C );
	return res;
}


void sgs_SerializeExt( SGS_CTX, sgs_Variable var, int mode )
{
	if( mode == SGS_SERIALIZE_DEFAULT )
		mode = C->serialize_state ? *(int*) C->serialize_state : 2;
	
	if( mode == 3 )
		sgs_SerializeInt_V3( C, var, NULL, 0 );
	else if( mode == 2 )
		sgs_SerializeInt_V2( C, var );
	else if( mode == 1 )
		sgs_SerializeInt_V1( C, var );
	else
	{
		sgs_PushNull( C );
		sgs_Msg( C, SGS_APIERR, "sgs_SerializeExt: bad mode (%d)", mode );
	}
}

void sgs_SerializeObject( SGS_CTX, sgs_StkIdx args, const char* func )
{
	int mode;
	size_t fnsize = strlen( func );
	if( !C->serialize_state )
	{
		sgs_Msg( C, SGS_APIERR, "sgs_SerializeObject: called outside the serialization process" );
		return;
	}
	if( fnsize >= 255 )
	{
		sgs_Msg( C, SGS_APIERR, "sgs_SerializeObject: function name length exceeds 255" );
		return;
	}
	mode = *(int*) C->serialize_state;
	
	if( mode == 3 )
		sgs_SerializeObjectInt_V3( C, args, func, fnsize );
	else if( mode == 2 )
		sgs_SerializeObjectInt_V2( C, args, func, fnsize );
	else if( mode == 1 )
		sgs_SerializeObjectInt_V1( C, args, func, fnsize );
	else
	{
		sgs_Msg( C, SGS_APIERR, "sgs_SerializeObjectExt: bad mode (%d)", mode );
	}
}

void sgs_SerializeObjIndex( SGS_CTX, sgs_Variable key, sgs_Variable val, int isprop )
{
	int mode;
	if( !C->serialize_state )
	{
		sgs_Msg( C, SGS_APIERR, "sgs_SerializeObject: called outside the serialization process" );
		return;
	}
	mode = *(int*) C->serialize_state;
	
	if( mode == 1 || mode == 2 )
	{
		sgs_Serialize( C, key );
		sgs_Serialize( C, val );
	}
	
	if( mode == 3 )
		sgs_Msg( C, SGS_APIERR, "sgs_SerializeObjIndex: mode 3 is not supported" );
	else if( mode == 2 )
		sgs_SerializeObjIndexInt_V2( C, isprop );
	else if( mode == 1 )
		sgs_SerializeObjIndexInt_V1( C, isprop );
	else
	{
		sgs_Msg( C, SGS_APIERR, "sgs_SerializeObjectExt: bad mode (%d)", mode );
	}
}


SGSBOOL sgs_UnserializeExt( SGS_CTX, sgs_Variable var, int mode )
{
	SGSRESULT res = 0;
	char* str = NULL, *strend;
	sgs_SizeVal size = 0;
	_STACK_PREPARE;
	sgs_PushVariable( C, var );
	if( !sgs_ParseString( C, -1, &str, &size ) || !size )
	{
		sgs_Msg( C, SGS_APIERR, "sgs_Unserialize: variable does not resolve to a non-empty string" );
		sgs_Pop( C, 1 );
		sgs_PushNull( C );
		return 0;
	}
	
	strend = str + size;
	
	if( mode == SGS_SERIALIZE_DEFAULT )
		mode = C->serialize_state ? *(int*) C->serialize_state : 2;
	
	_STACK_PROTECT;
	if( mode == 3 )
		res = sgs_UnserializeInt_V3( C, str, strend );
	else if( mode == 2 )
		res = sgs_UnserializeInt_V2( C, str, strend );
	else if( mode == 1 )
		res = sgs_UnserializeInt_V1( C, str, strend );
	else
	{
		sgs_Msg( C, SGS_APIERR, "sgs_UnserializeExt: bad mode (%d)", mode );
	}
	_STACK_UNPROTECT_SKIP( res );
	return res;
}



/* SGS object notation [SGSON] */

void sgs_SerializeSGSON( SGS_CTX, sgs_Variable var, const char* tab )
{
	sgs_SizeVal tablen = tab ? (sgs_SizeVal) SGS_STRINGLENGTHFUNC( tab ) : 0;
	sgs_SerializeInt_V3( C, var, tab, tablen );
}


static void sgson_skipws( const char** p, const char* end )
{
	const char* pos = *p;
	while( pos < end )
	{
		if( *pos != ' ' && *pos != '\t' &&
			*pos != '\n' && *pos != '\r' )
			break;
		pos++;
	}
	*p = pos;
}

#define SGSON_STK_TOP stack->ptr[ stack->size - 1 ]
#define SGSON_STK_POP sgs_membuf_resize( stack, C, stack->size - 1 )
#define SGSON_STK_PUSH( what ) sgs_membuf_appchr( stack, C, what )

const char* sgson_parse( SGS_CTX, sgs_MemBuf* stack, const char* buf, sgs_SizeVal size )
{
	int stk = sgs_StackSize( C );
	const char* pos = buf, *end = buf + size;
	for(;;)
	{
		int push = 0;
		sgson_skipws( &pos, end );
		if( pos >= end )
			break;

		if( SGSON_STK_TOP == '{' &&
			*pos != '"' &&
			*pos != '\'' &&
			*pos != '_' &&
			*pos != '[' &&
			!sgs_isalpha( *pos ) &&
			*pos != '}' )
			return pos;

		if( SGSON_STK_TOP == 0 && sgs_StackSize( C ) > stk )
			return pos;

		if( *pos == '{' )
		{
			SGSON_STK_PUSH( '{' );
			sgs_CreateDict( C, NULL, 0 );
		}
		else if( *pos == '}' )
		{
			if( SGSON_STK_TOP != '{' )
				return pos;
			SGSON_STK_POP;
			push = 1;
		}
		else if( *pos == '[' )
		{
			if( SGSON_STK_TOP == '{' )
			{
				/* non-string key marker (for maps) */
				SGSON_STK_TOP = 'M';
			}
			else
			{
				SGSON_STK_PUSH( '[' );
				sgs_CreateArray( C, NULL, 0 );
			}
		}
		else if( *pos == ']' )
		{
			if( SGSON_STK_TOP == 'M' )
			{
				SGSON_STK_TOP = '=';
				pos++;
				sgson_skipws( &pos, end );
				if( *pos != '=' )
					return pos;
			}
			else
			{
				if( SGSON_STK_TOP != '[' )
					return pos;
				SGSON_STK_POP;
				push = 1;
			}
		}
		else if( *pos == ')' )
		{
			sgs_Variable func;
			if( SGSON_STK_TOP != '(' )
				return pos;
			SGSON_STK_POP;
			push = 1;
			
			/* search for marker */
			sgs_SizeVal i = sgs_StackSize( C ) - 1;
			while( i >= 0 && sgs_StackItem( C, i ).type != 255 )
				i--;
			sgs_BreakIf( i < 0 );
			
			/* replace function name with global function */
			sgs_GetGlobal( C, sgs_StackItem( C, i + 1 ), &func );
			sgs_SetStackItem( C, i + 1, func );
			VAR_RELEASE( &func );
			
			/* call the function */
			sgs_Call( C, sgs_StackSize( C ) - ( i + 2 ), 1 );
			
			/* clean up - pop marker, skip returned value */
			sgs_PopSkip( C, 1, 1 );
		}
		else if( *pos == '"' || *pos == '\'' )
		{
			char sc = *pos;
			const char* beg = ++pos;
			sgs_MemBuf str = sgs_membuf_create();
			while( pos < end && *pos != sc )
			{
				uint8_t cc = (uint8_t) *pos;
				if( cc <= 0x1f || cc == 0x7f )
				{
					sgs_membuf_destroy( &str, C );
					return pos;
				}
				if( *pos == '\\' )
				{
					pos++;
					switch( *pos )
					{
					case '"':
					case '\'':
					case '\\':
						sgs_membuf_appchr( &str, C, *pos );
						break;
					case 'n': sgs_membuf_appchr( &str, C, '\n' ); break;
					case 'r': sgs_membuf_appchr( &str, C, '\r' ); break;
					case 't': sgs_membuf_appchr( &str, C, '\t' ); break;
					case 'x':
						{
							uint8_t hex[ 2 ];
							uint8_t chr;
							pos++;
							if( !sgs_hexchar( pos[0] ) ){ goto strfail; }
							if( !sgs_hexchar( pos[1] ) ){ pos++; goto strfail; }
							hex[ 0 ] = (uint8_t) ( sgs_gethex( pos[0] ) );
							hex[ 1 ] = (uint8_t) ( sgs_gethex( pos[1] ) );
							pos++;
							chr = (uint8_t) ( ( hex[0] << 4 ) | hex[1] );
							sgs_membuf_appchr( &str, C, (char) chr );
						}
						break;
					default:
					strfail:
						sgs_membuf_destroy( &str, C );
						return pos;
					}
				}
				else
					sgs_membuf_appchr( &str, C, *pos );
				pos++;
			}
			if( pos >= end || str.size > 0x7fffffff )
			{
				sgs_membuf_destroy( &str, C );
				return beg;
			}
			sgs_PushStringBuf( C, str.ptr, (sgs_SizeVal) str.size );
			sgs_membuf_destroy( &str, C );
			if( SGSON_STK_TOP == '{' )
			{
				SGSON_STK_TOP = '=';
				pos++;
				sgson_skipws( &pos, end );
				if( *pos != '=' )
					return pos;
			}
			else
			{
				push = 1;
			}
		}
		else if( sgs_decchar( *pos ) || *pos == '-' )
		{
			sgs_Int outi;
			sgs_Real outf;
			int type = sgs_util_strtonum( &pos, end, &outi, &outf );
			if( type == 1 )
			{
				sgs_PushInt( C, outi );
				push = 1;
			}
			else if( type == 2 )
			{
				sgs_PushReal( C, outf );
				push = 1;
			}
			else
				return pos;
			pos--;
		}
		else if( *pos == '_' || sgs_isalpha( *pos ) )
		{
			if( end - pos >= 4 &&
				( pos[4] == '\0' || sgs_isoneof( pos[4], ")}]=, \n\r\t" ) ) &&
				memcmp( pos, "null", 4 ) == 0 )
			{
				if( SGSON_STK_TOP == '{' )
					return pos; /* no keywords as keys outside [..] */
				sgs_PushNull( C );
				pos += 4 - 1;
				push = 1;
			}
			else if( end - pos >= 4 &&
				( pos[4] == '\0' || sgs_isoneof( pos[4], ")}]=, \n\r\t" ) ) &&
				memcmp( pos, "true", 4 ) == 0 )
			{
				if( SGSON_STK_TOP == '{' )
					return pos; /* no keywords as keys outside [..] */
				sgs_PushBool( C, SGS_TRUE );
				pos += 4 - 1;
				push = 1;
			}
			else if( end - pos >= 5 &&
				( pos[5] == '\0' || sgs_isoneof( pos[5], ")}]=, \n\r\t" ) ) &&
				memcmp( pos, "false", 5 ) == 0 )
			{
				if( SGSON_STK_TOP == '{' )
					return pos; /* no keywords as keys outside [..] */
				sgs_PushBool( C, SGS_FALSE );
				pos += 5 - 1;
				push = 1;
			}
			else if( SGSON_STK_TOP == '{' ) /* identifiers as keys */
			{
				const char* idend = pos;
				while( sgs_isid( *idend ) )
					idend++;
				if( idend - pos > 255 )
				{
					return pos;
				}
				
				sgs_PushStringBuf( C, pos, (sgs_SizeVal)( idend - pos ) );
				pos = idend;
				SGSON_STK_TOP = '=';
				sgson_skipws( &pos, end );
				if( *pos != '=' )
					return pos;
			}
			else /* identifiers as function names */
			{
				const char* idstart = pos;
				const char* idend = pos;
				while( sgs_isid( *idend ) )
					idend++;
				if( idend - idstart > 255 )
					return pos; /* func. name size exceeds 255 chars */
				
				pos = idend;
				sgson_skipws( &pos, end );
				if( pos[0] != '(' && pos[0] != '{' )
					return pos; /* no opening char. */
				
				if( pos[0] == '(' )
				{
					sgs_Variable marker = { 255 };
					SGSON_STK_PUSH( '(' );
					sgs_PushVariable( C, marker ); /* marker for beginning of function */
					sgs_PushStringBuf( C, idstart, (sgs_SizeVal)( idend - idstart ) );
				}
				else if( idend - idstart == 3 && memcmp( idstart, "map", 3 ) == 0 )
				{
					SGSON_STK_PUSH( '{' );
					sgs_CreateMap( C, NULL, 0 );
				}
				else return pos;
			}
		}
		else
			return pos;

		if( push )
		{
			if( SGSON_STK_TOP == '[' || SGSON_STK_TOP == '=' || SGSON_STK_TOP == '(' )
			{
				int revchr = SGSON_STK_TOP == '[' ? ']' : '}';
				if( SGSON_STK_TOP == '(' )
					revchr = ')';
				
				pos++;
				sgson_skipws( &pos, end );
				if( pos >= end )
					break;
				if( *pos != ',' && *pos != revchr ) return pos;
				if( *pos != ',' )
					pos--;
			}
			if( SGSON_STK_TOP == '[' )
			{
				sgs_ArrayPush( C, sgs_StackItem( C, -2 ), 1 );
			}
			if( SGSON_STK_TOP == '=' )
			{
				sgs_SetIndex( C, sgs_StackItem( C, -3 ), sgs_StackItem( C, -2 ), sgs_StackItem( C, -1 ), SGS_FALSE );
				sgs_Pop( C, 2 );
				SGSON_STK_TOP = '{';
			}
		}
		pos++;
	}
/*	printf( "%d, %.*s, %d\n", stack->size, stack->size, stack->ptr, sgs_StackSize(C)-stk ); */
	return sgs_StackSize( C ) > stk && stack->size == 1 ? NULL : buf;
}

void sgs_UnserializeSGSONExt( SGS_CTX, const char* str, size_t size )
{
	const char* ret = NULL;
	sgs_MemBuf stack = sgs_membuf_create();
	sgs_SizeVal stksize = sgs_StackSize( C );
	sgs_membuf_appchr( &stack, C, 0 );
	ret = sgson_parse( C, &stack, str, (sgs_SizeVal) size );
	sgs_membuf_destroy( &stack, C );
	if( ret )
	{
		sgs_PushNull( C );
		sgs_Msg( C, SGS_ERROR, "failed to parse SGSON (position %d, %.8s...", ret - str, ret );
	}
	sgs_PopSkip( C, sgs_StackSize( C ) - stksize - 1, 1 );
}



/* bytecode serialization */

#define SGSNOMINDEC( cnt ) (D->end - D->buf < (ptrdiff_t)(cnt))

#define esi16( x ) ( (((x)&0xff)<<8) | (((x)>>8)&0xff) )

#define esi32( x ) (\
	(((x)&0xff)<<24) | (((x)&0xff00)<<8) |\
	(((x)>>8)&0xff00) | (((x)>>24)&0xff) )

typedef struct decoder_s
{
	SGS_CTX;
	const char* buf, *start, *end;
	char convend;
	const char* filename;
	size_t filename_len;
}
decoder_t;

static void esi16_array( uint16_t* data, unsigned cnt )
{
	unsigned i;
	for( i = 0; i < cnt; ++i )
	{
		data[ i ] = (uint16_t) esi16( data[ i ] );
	}
}

static void esi32_array( uint32_t* data, unsigned cnt )
{
	unsigned i;
	for( i = 0; i < cnt; ++i )
	{
		data[ i ] = (uint32_t) esi32( data[ i ] );
	}
}


/*
	i32 size
	byte[size] data
*/
static void bc_write_sgsstring( sgs_iStr* S, SGS_CTX, sgs_MemBuf* outbuf )
{
	sgs_membuf_appbuf( outbuf, C, &S->size, sizeof( int32_t ) );
	sgs_membuf_appbuf( outbuf, C, sgs_str_cstr( S ), S->size );
}

static const char* bc_read_sgsstring( decoder_t* D, sgs_Variable* var )
{
	const char* buf = D->buf;
	int32_t len;
	
	if( SGSNOMINDEC( 4 ) )
		return "data error (expected string length)";
	
	SGS_AS_INT32( len, buf );
	if( D->convend )
		len = esi32( len );
	buf += 4;
	
	if( SGSNOMINDEC( len ) )
		return "data error (expected string bytes)";
	
	sgsVM_VarCreateString( D->C, var, buf, len );
	D->buf = buf + len;
	
	return NULL;
}


/*
	byte type
	--if type = NULL:
	--if type = BOOL:
	byte value
	--if type = INT:
	integer value
	--if type = REAL:
	real value
	--if type = STRING:
	stringdata data
	--if type = FUNC:
	funcdata data
*/
static int bc_write_sgsfunc( sgs_iFunc* F, SGS_CTX, sgs_MemBuf* outbuf );
static int bc_write_var( sgs_Variable* var, SGS_CTX, sgs_MemBuf* outbuf )
{
	uint8_t vt = (uint8_t) var->type;
	/* WP: don't care about the sign when serializing bitfield */
	sgs_membuf_appchr( outbuf, C, (char) vt );
	switch( vt )
	{
	case SGS_VT_NULL: break;
	/* WP: var->data.B can only store 0/1 */
	case SGS_VT_BOOL: sgs_membuf_appchr( outbuf, C, (char) var->data.B ); break;
	case SGS_VT_INT: sgs_membuf_appbuf( outbuf, C, &var->data.I, sizeof( sgs_Int ) ); break;
	case SGS_VT_REAL: sgs_membuf_appbuf( outbuf, C, &var->data.R, sizeof( sgs_Real ) ); break;
	case SGS_VT_STRING: bc_write_sgsstring( var->data.S, C, outbuf ); break;
	case SGS_VT_FUNC: if( !bc_write_sgsfunc( var->data.F, C, outbuf ) ) return 0; break;
	default:
		return 0;
	}
	return 1;
}

static const char* bc_read_sgsfunc( decoder_t* D, sgs_Variable* var );
static const char* bc_read_var( decoder_t* D, sgs_Variable* var )
{
	const char* ret = NULL;
	uint8_t vt;
	
	if( SGSNOMINDEC( 1 ) )
		return "data error (expected type)";
	
	vt = (uint8_t) *D->buf++;
	var->type = SGS_VT_NULL;
	switch( vt )
	{
	case SGS_VT_NULL: var->type = vt; break;
	case SGS_VT_BOOL:
		if( SGSNOMINDEC( 1 ) )
			return "data error (expected value)";
		
		var->type = vt;
		var->data.B = *D->buf++ ? 1 : 0;
		break;
		
	case SGS_VT_INT:
		if( SGSNOMINDEC( sizeof( sgs_Int ) ) )
			return "data error (expected value)";
		
		var->type = vt;
		SGS_AS_INTEGER( var->data.I, D->buf );
		D->buf += sizeof( sgs_Int );
		break;
		
	case SGS_VT_REAL:
		if( SGSNOMINDEC( sizeof( sgs_Real ) ) )
			return "data error (expected value)";
		
		var->type = vt;
		SGS_AS_REAL( var->data.R, D->buf );
		D->buf += sizeof( sgs_Real );
		break;
		
	case SGS_VT_STRING:
		ret = bc_read_sgsstring( D, var );
		if( ret == NULL )
			var->type = SGS_VT_STRING;
		break;
		
	case SGS_VT_FUNC:
		ret = bc_read_sgsfunc( D, var );
		break;
		
	default:
		return "invalid variable type found";
	}
	return ret;
}


/*
	var[cnt] varlist
*/
static int bc_write_varlist( sgs_Variable* vlist, SGS_CTX, int cnt, sgs_MemBuf* outbuf )
{
	int i;
	for( i = 0; i < cnt; ++i )
	{
		if( !bc_write_var( vlist + i, C, outbuf ) )
			return 0;
	}
	return 1;
}

static const char* bc_read_varlist( decoder_t* D, sgs_Variable* vlist, int cnt )
{
	int i;
	for( i = 0; i < cnt; ++i )
	{
		const char* ret = bc_read_var( D, vlist + i );
		if( ret )
		{
			cnt = i;
			for( i = 0; i < cnt; ++i )
				sgs_Release( D->C, vlist + i );
			return ret;
		}
	}
	return NULL;
}


/*
	i16 constcount
	i16 instrcount
	byte gotthis
	byte numargs
	byte numtmp
	byte numclsr
	byte inclsr
	i16 linenum
	i16[instrcount] lineinfo
	i32 funcname_size
	byte[funcname_size] funcname
	varlist consts
	instr[instrcount] instrs
*/
static int bc_write_sgsfunc( sgs_iFunc* F, SGS_CTX, sgs_MemBuf* outbuf )
{
	uint32_t size = F->sfuncname->size;
	uint16_t cc, ic;
	uint8_t gntc[5] = { F->gotthis, F->numargs, F->numtmp, F->numclsr, F->inclsr };
	
	/* WP: const/instruction limits */
	cc = (uint16_t) sgs_func_const_count( F );
	ic = (uint16_t) sgs_func_instr_count( F );

	sgs_membuf_appbuf( outbuf, C, &cc, sizeof( cc ) );
	sgs_membuf_appbuf( outbuf, C, &ic, sizeof( ic ) );
	sgs_membuf_appbuf( outbuf, C, gntc, 5 );
	sgs_membuf_appbuf( outbuf, C, &F->linenum, sizeof( sgs_LineNum ) );
	sgs_membuf_appbuf( outbuf, C, sgs_func_lineinfo( F ), sizeof( sgs_LineNum ) * ic );
	sgs_membuf_appbuf( outbuf, C, &size, sizeof( size ) );
	sgs_membuf_appbuf( outbuf, C, sgs_str_cstr( F->sfuncname ), F->sfuncname->size );

	if( !bc_write_varlist( sgs_func_consts( F ), C, cc, outbuf ) )
		return 0;

	sgs_membuf_appbuf( outbuf, C, sgs_func_bytecode( F ), sizeof( sgs_instr_t ) * ic );
	return 1;
}

static const char* bc_read_sgsfunc( decoder_t* D, sgs_Variable* var )
{
	sgs_Variable strvar;
	sgs_iFunc* F = NULL;
	uint32_t coff, ioff, size, fnsize;
	uint16_t cc, ic;
	const char* ret = "data error (expected fn. data)";
	SGS_CTX = D->C;
	
	if( SGSNOMINDEC( 10 ) )
		goto fail;
	
	SGS_AS_UINT16( cc, D->buf );
	SGS_AS_UINT16( ic, D->buf + 2 );
	
	if( D->convend )
	{
		/* WP: int promotion will not affect the result */
		cc = (uint16_t) esi16( cc );
		ic = (uint16_t) esi16( ic );
	}
	
	/* basic tests to avoid allocating too much memory */
	if( SGSNOMINDEC( 10 + (ptrdiff_t) ( cc + ic * sizeof(sgs_LineNum) ) ) )
		goto fail;
	
	/* WP: const/instruction limits */
	ioff = (uint32_t) sizeof( sgs_Variable ) * cc;
	coff = (uint32_t) sizeof( sgs_instr_t ) * ic;
	size = ioff + coff;
	
	F = sgs_Alloc_a( sgs_iFunc, size );
	F->refcount = 1;
	F->size = size;
	F->instr_off = ioff;
	SGS_AS_UINT8( F->gotthis, D->buf + 4 );
	SGS_AS_UINT8( F->numargs, D->buf + 5 );
	SGS_AS_UINT8( F->numtmp, D->buf + 6 );
	SGS_AS_UINT8( F->numclsr, D->buf + 7 );
	SGS_AS_UINT8( F->inclsr, D->buf + 8 );
	SGS_AS_INT16( F->linenum, D->buf + 9 );
	if( D->convend )
		F->linenum = (sgs_LineNum) esi16( F->linenum );
	F->lineinfo = sgs_Alloc_n( sgs_LineNum, ic );
	F->sfuncname = NULL;
	F->sfilename = NULL;
	D->buf += 11;
	
	ret = "data error (expected fn. line numbers)";
	if( SGSNOMINDEC( sizeof( sgs_LineNum ) * ic ) )
		goto fail;
	
	memcpy( F->lineinfo, D->buf, sizeof( sgs_LineNum ) * ic );
	D->buf += sizeof( sgs_LineNum ) * ic;
	if( D->convend )
		esi16_array( (uint16_t*) F->lineinfo, ic );
	
	ret = "data error (expected fn. name)";
	if( SGSNOMINDEC( 4 ) )
		goto fail;
	SGS_AS_UINT32( fnsize, D->buf ); D->buf += 4;
	if( D->convend )
		fnsize = (uint32_t) esi32( fnsize );
	if( SGSNOMINDEC( fnsize ) )
		goto fail;
	/* WP: string limit */
	memcpy( sgs_InitStringAlloc( C, &strvar, (sgs_SizeVal) fnsize ), D->buf, fnsize );
	sgs_FinalizeStringAllocP( C, &strvar );
	F->sfuncname = strvar.data.S;
	D->buf += fnsize;
	
	/* WP: string limit */
	sgs_InitStringBuf( C, &strvar, D->filename, (sgs_SizeVal) D->filename_len );
	F->sfilename = strvar.data.S;
	
	/* the main data */
	ret = bc_read_varlist( D, sgs_func_consts( F ), cc );
	if( ret )
		goto fail;
	
	ret = "data error (expected fn. instructions)";
	if( SGSNOMINDEC( coff ) )
		goto fail;
	memcpy( sgs_func_bytecode( F ), D->buf, coff );
	if( D->convend )
		esi32_array( sgs_func_bytecode( F ), coff / sizeof( sgs_instr_t ) );
	D->buf += coff;

	var->data.F = F;
	var->type = SGS_VT_FUNC;
	return NULL;

fail:
	if( F )
	{
		/* everything is allocated together, between error jumps */
		sgs_Dealloc( F->lineinfo );
		strvar.type = SGS_VT_STRING;
		if( F->sfuncname )
		{
			strvar.data.S = F->sfuncname;
			sgs_Release( C, &strvar );
		}
		if( F->sfilename )
		{
			strvar.data.S = F->sfilename;
			sgs_Release( C, &strvar );
		}
		sgs_Dealloc( F );
	}
	return ret;
}


/*
	-- header -- 14 bytes --
	seq "SGS\0"
	byte version_major
	byte version_minor
	byte version_incr
	byte integer_size
	byte real_size
	byte flags
	u32 filesize
	-- header end --
	function main
*/
int sgsBC_Func2Buf( SGS_CTX, sgs_iFunc* func, sgs_MemBuf* outbuf )
{
	size_t origobsize = outbuf->size;
	char header_bytes[ 14 ] =
	{
		'S', 'G', 'S', 0,
		SGS_VERSION_MAJOR,
		SGS_VERSION_MINOR,
		SGS_VERSION_INCR,
		sizeof( sgs_Int ),
		sizeof( sgs_Real ),
		0,
		0, 0, 0, 0
	};
	header_bytes[ 9 ] = ( O32_HOST_ORDER == O32_LITTLE_ENDIAN ) ? SGSBC_FLAG_LITTLE_ENDIAN : 0;
	sgs_membuf_reserve( outbuf, C, origobsize + 1000 );
	sgs_membuf_appbuf( outbuf, C, header_bytes, 14 );
	
	{
		uint32_t sz;
		int ret = bc_write_sgsfunc( func, C, outbuf );
		if( !ret )
			return 0;
		sz = (uint32_t)( outbuf->size - origobsize );
		memcpy( outbuf->ptr + origobsize + 10, &sz, sizeof(sz) );
		return 1;
	}
}

const char* sgsBC_Buf2Func( SGS_CTX, const char* fn, const char* buf, size_t size, sgs_Variable* outfunc )
{
	char flags;
	uint32_t sz;
	
	if( size < 22 )
		return "data error (expected fn. header)";
	
	flags = buf[ 9 ];
	SGS_AS_UINT32( sz, buf + 10 );
	
	decoder_t Dstorage, *D;
	D = &Dstorage; /* macro compatibility */
	{
		D->C = C;
		D->buf = NULL;
		D->start = buf;
		D->end = buf + size;
		D->convend = ( O32_HOST_ORDER == O32_LITTLE_ENDIAN ) !=
			( ( flags & SGSBC_FLAG_LITTLE_ENDIAN ) != 0 );
		D->filename = fn;
		D->filename_len = strlen( fn );
	}
	
	if( D->convend )
		sz = esi32( sz );
	if( (size_t) sz != size )
		return "data error (fn. data size mismatch)";
	
	D->buf = buf + 14;
	return bc_read_sgsfunc( D, outfunc );
}

int sgsBC_ValidateHeader( const char* buf, size_t size )
{
	int i;
	char validate_bytes[ 9 ] =
	{
		'S', 'G', 'S', 0,
		SGS_VERSION_MAJOR,
		SGS_VERSION_MINOR,
		SGS_VERSION_INCR,
		sizeof( sgs_Int ),
		sizeof( sgs_Real )
	};

	if( size < SGS_MIN_BC_SIZE )
		return -1;
	for( i = 0; i < 9; ++i )
	{
		if( buf[ i ] != validate_bytes[ i ] )
			return i;
	}
	return SGS_HEADER_SIZE;
}




#include <stdio.h>
#include <errno.h>



#ifndef STDLIB_WARN
#  define STDLIB_WARN( warn ) return sgs_Msg( C, SGS_WARNING, warn );
#endif
#ifndef STDLIB_ERR
#  define STDLIB_ERR( err ) return sgs_Msg( C, SGS_ERROR, err );
#endif


static int sgsstd_expectnum( SGS_CTX, sgs_StkIdx n )
{
	sgs_StkIdx ssz = stk_size( C );
	if( n != ssz )
		return sgs_Msg( C, SGS_WARNING, "function expects exactly %d arguments"
			", %d given", n, ssz );
	return 1;
}


/* Containers */


/*
	ARRAY
*/

#define SGSARR_UNIT sizeof( sgs_Variable )
#define SGSARR_HDR sgsstd_array_header_t* hdr = (sgsstd_array_header_t*) data
#define SGSARR_HDR_OI sgsstd_array_header_t* hdr = (sgsstd_array_header_t*) obj->data
#define SGSARR_HDRSIZE sizeof(sgsstd_array_header_t)
#define SGSARR_ALLOCSIZE( cnt ) ((size_t)(cnt)*SGSARR_UNIT)
#define SGSARR_PTR( base ) (((sgsstd_array_header_t*)base)->data)

static void sgsstd_array_reserve( SGS_CTX, sgsstd_array_header_t* hdr, sgs_SizeVal size )
{
	if( size <= hdr->mem )
		return;

	hdr->data = (sgs_Variable*) sgs_Realloc( C, hdr->data, SGSARR_ALLOCSIZE( size ) );
	hdr->mem = size;
}

static void sgsstd_array_clear( SGS_CTX, sgsstd_array_header_t* hdr )
{
	sgs_ReleaseArray( C, SGSARR_PTR( hdr ), hdr->size );
	hdr->size = 0;
}

/* off = offset in stack to start inserting from */
static void sgsstd_array_insert( SGS_CTX, sgsstd_array_header_t* hdr, sgs_SizeVal pos, sgs_StkIdx off )
{
	int i;
	sgs_StkIdx cnt = stk_size( C ) - off;
	sgs_SizeVal nsz = hdr->size + cnt;
	sgs_Variable* ptr = SGSARR_PTR( hdr );
	
	if( !cnt ) return;
	
	if( nsz > hdr->mem )
	{
		sgsstd_array_reserve( C, hdr, SGS_MAX( nsz, hdr->mem * 2 ) );
		ptr = SGSARR_PTR( hdr );
	}
	if( pos < hdr->size )
		memmove( ptr + pos + cnt, ptr + pos, SGSARR_ALLOCSIZE( hdr->size - pos ) );
	for( i = off; i < stk_size( C ); ++i )
	{
		sgs_Variable* var = ptr + pos + i - off;
		sgs_GetStackItem( C, i, var );
	}
	hdr->size = nsz;
}

void sgsstd_array_insert_p( SGS_CTX, sgsstd_array_header_t* hdr, sgs_SizeVal pos, sgs_Variable* var )
{
	sgs_SizeVal nsz = hdr->size + 1;
	sgs_Variable* ptr = SGSARR_PTR( hdr );
	
	if( nsz > hdr->mem )
	{
		sgsstd_array_reserve( C, hdr, SGS_MAX( nsz, hdr->mem * 2 ) );
		ptr = SGSARR_PTR( hdr );
	}
	if( pos < hdr->size )
		memmove( ptr + pos + 1, ptr + pos, SGSARR_ALLOCSIZE( hdr->size - pos ) );
	
	ptr[ pos ] = *var;
	sgs_Acquire( C, var );
	
	hdr->size = nsz;
}

static void sgsstd_array_erase( SGS_CTX, sgsstd_array_header_t* hdr, sgs_SizeVal from, sgs_SizeVal to )
{
	sgs_SizeVal cnt = to - from + 1, to1 = to + 1;
	sgs_Variable* ptr = SGSARR_PTR( hdr );
	
	sgs_BreakIf( from < 0 || from >= hdr->size || to < 0 || to >= hdr->size || from > to );
	
	sgs_ReleaseArray( C, ptr + from, to - from + 1 );
	if( to1 < hdr->size )
		memmove( ptr + from, ptr + to1, SGSARR_ALLOCSIZE( hdr->size - to1 ) );
	hdr->size -= cnt;
}


#define SGSARR_IHDR( name ) \
	sgsstd_array_header_t* hdr; \
	if( !SGS_PARSE_METHOD( C, sgsstd_array_iface, hdr, array, name ) ) return 0; \
	SGS_UNUSED( hdr );


static int sgsstd_arrayI_push( SGS_CTX )
{
	SGSARR_IHDR( push );
	sgsstd_array_insert( C, hdr, hdr->size, 0 );
	
	SGS_RETURN_THIS( C );
}
static int sgsstd_arrayI_pop( SGS_CTX )
{
	sgs_Variable* ptr;
	SGSARR_IHDR( pop );
	ptr = SGSARR_PTR( hdr );
	if( !hdr->size )
		STDLIB_WARN( "array is empty, cannot pop" );
	
	hdr->size--;
	fstk_push_leave( C, &ptr[ hdr->size ] );
	return 1;
}
static int sgsstd_arrayI_shift( SGS_CTX )
{
	sgs_Variable* ptr;
	SGSARR_IHDR( shift );
	ptr = SGSARR_PTR( hdr );
	if( !hdr->size )
		STDLIB_WARN( "array is empty, cannot shift" );
	
	fstk_push( C, &ptr[ 0 ] );
	sgsstd_array_erase( C, hdr, 0, 0 );
	return 1;
}
static int sgsstd_arrayI_unshift( SGS_CTX )
{
	SGSARR_IHDR( unshift );
	sgsstd_array_insert( C, hdr, 0, 0 );
	
	SGS_RETURN_THIS( C );
}

static int sgsstd_arrayI_insert( SGS_CTX )
{
	sgs_Int at;
	SGSARR_IHDR( insert );
	
	if( !sgs_LoadArgs( C, "i?v", &at ) )
		return 0;
	
	if( at < 0 )
		at += hdr->size + 1;
	if( at < 0 || at > hdr->size )
		STDLIB_WARN( "index out of bounds" )
	
	sgsstd_array_insert( C, hdr, (sgs_SizeVal) at, 1 );
	
	SGS_RETURN_THIS( C );
}
static int sgsstd_arrayI_erase( SGS_CTX )
{
	int cnt = stk_size( C );
	sgs_Int at, at2;
	SGSARR_IHDR( erase );
	
	if( !sgs_LoadArgs( C, "i|i", &at, &at2 ) )
		return 0;
	
	if( at < 0 )
		at += hdr->size;
	if( at < 0 || at >= hdr->size )
		STDLIB_WARN( "index out of bounds" )
	
	if( cnt == 1 )
		at2 = at;
	else
	{
		if( at2 < 0 )
			at2 += hdr->size;
		if( at2 < 0 || at2 >= hdr->size )
			STDLIB_WARN( "index out of bounds" )
		
		if( at2 < at )
			STDLIB_WARN( "after resolving,"
				" index #1 must be smaller or equal than index #2" )
	}
	
	sgsstd_array_erase( C, hdr, (sgs_SizeVal) at, (sgs_SizeVal) at2 );
	
	SGS_RETURN_THIS( C );
}
static int sgsstd_arrayI_part( SGS_CTX )
{
	sgs_SizeVal from, max = 0x7fffffff, to, i;
	SGSARR_IHDR( part );
	
	if( !sgs_LoadArgs( C, "l|l", &from, &max ) )
		return 0;
	
	if( max < 0 )
		STDLIB_WARN( "argument 2 (count) cannot be negative" )
	
	if( from < 0 )
		from += hdr->size;
	to = from + max;
	if( to < from )
		to = hdr->size;
	
	sgs_CreateArray( C, NULL, 0 );
	if( from < hdr->size && 0 < to )
	{
		from = SGS_MAX( from, 0 );
		to = SGS_MIN( to, hdr->size );
		
		if( from < to )
		{
			sgs_Variable *psrc, *pdst;
			sgs_SizeVal cnt = to - from;
			sgsstd_array_header_t* nhdr = (sgsstd_array_header_t*) sgs_GetObjectData( C, -1 );
			sgsstd_array_reserve( C, nhdr, to - from );
			nhdr->size = cnt;
			psrc = SGSARR_PTR( hdr ) + from;
			pdst = SGSARR_PTR( nhdr );
			for( i = 0; i < cnt; ++i )
			{
				pdst[ i ] = psrc[ i ];
				sgs_Acquire( C, pdst + i );
			}
		}
	}
	
	return 1;
}

static int sgsstd_arrayI_clear( SGS_CTX )
{
	SGSARR_IHDR( clear );
	sgsstd_array_clear( C, hdr );
	
	SGS_RETURN_THIS( C );
}

static int sgsstd_arrayI_reverse( SGS_CTX )
{
	SGSARR_IHDR( reverse );
	
	/* emit a warning if any arguments are passed */
	if( !sgs_LoadArgs( C, "." ) )
		return 0;
	
	{
		sgs_Variable tmp;
		sgs_Variable* P = SGSARR_PTR( hdr );
		sgs_SizeVal i, j, hsz = hdr->size / 2;
		for( i = 0, j = hdr->size - 1; i < hsz; ++i, --j )
		{
			tmp = P[ i ];
			P[ i ] = P[ j ];
			P[ j ] = tmp;
		}
	}
	
	SGS_RETURN_THIS( C );
}

static void sgsstd_array_adjust( SGS_CTX, sgsstd_array_header_t* hdr, sgs_SizeVal cnt )
{
	while( hdr->size > cnt )
	{
		sgs_Release( C, SGSARR_PTR( hdr ) + hdr->size - 1 );
		hdr->size--;
	}
	while( hdr->size < cnt )
	{
		( SGSARR_PTR( hdr ) + (hdr->size++) )->type = SGS_VT_NULL;
	}
}
static int sgsstd_arrayI_resize( SGS_CTX )
{
	sgs_SizeVal sz;
	SGSARR_IHDR( resize );
	
	if( !sgs_LoadArgs( C, "l", &sz ) )
		return 0;
	
	if( sz < 0 )
		STDLIB_WARN( "argument 1 (size) must be bigger than or equal to 0" )
	
	sgsstd_array_reserve( C, hdr, sz );
	sgsstd_array_adjust( C, hdr, sz );
	
	SGS_RETURN_THIS( C );
}
static int sgsstd_arrayI_reserve( SGS_CTX )
{
	sgs_SizeVal sz;
	SGSARR_IHDR( reserve );
	
	if( !sgs_LoadArgs( C, "l", &sz ) )
		return 0;
	
	if( sz < 0 )
		STDLIB_WARN( "argument 1 (size) must be bigger than or equal to 0" )
	
	sgsstd_array_reserve( C, hdr, sz );
	
	SGS_RETURN_THIS( C );
}

static SGS_INLINE int sgsarrcomp_basic( const void* p1, const void* p2, void* userdata )
{
	SGS_CTX = (sgs_Context*) userdata;
	sgs_Variable v1 = *(const sgs_Variable*) p1;
	sgs_Variable v2 = *(const sgs_Variable*) p2;
	return sgs_Compare( C, &v1, &v2 );
}
static SGS_INLINE int sgsarrcomp_basic_rev( const void* p1, const void* p2, void* userdata )
{ return sgsarrcomp_basic( p2, p1, userdata ); }
static int sgsstd_arrayI_sort( SGS_CTX )
{
	int rev = 0;
	SGSARR_IHDR( sort );
	
	if( !sgs_LoadArgs( C, "|b", &rev ) )
		return 0;
	
	/* WP: array limit */
	sgs_quicksort( SGSARR_PTR( hdr ), (size_t) hdr->size, sizeof( sgs_Variable ),
		rev ? sgsarrcomp_basic_rev : sgsarrcomp_basic, C );
	
	SGS_RETURN_THIS( C );
}

typedef struct sgsarrcomp_cl2_s
{
	SGS_CTX;
	sgs_Variable sortfunc;
}
sgsarrcomp_cl2;
static SGS_INLINE int sgsarrcomp_custom( const void* p1, const void* p2, void* userdata )
{
	sgsarrcomp_cl2* u = (sgsarrcomp_cl2*) userdata;
	SGS_CTX = u->C;
	stk_push3( C, &u->sortfunc, (const sgs_Variable*) p1, (const sgs_Variable*) p2 );
	sgs_Call( C, 2, 1 );
	{
		sgs_Real r = sgs_GetReal( C, -1 );
		fstk_pop1( C );
		return r == 0 ? 0 : ( r < 0 ? -1 : 1 );
	}
}
static SGS_INLINE int sgsarrcomp_custom_rev( const void* p1, const void* p2, void* userdata )
{ return sgsarrcomp_custom( p2, p1, userdata ); }
static int sgsstd_arrayI_sort_custom( SGS_CTX )
{
	int rev = 0;
	sgsarrcomp_cl2 u;
	SGSARR_IHDR( sort_custom );
	
	u.C = C;
	u.sortfunc.type = SGS_VT_NULL;
	
	if( !sgs_LoadArgs( C, "?p<v|b", &u.sortfunc, &rev ) )
		return 0;
	
	/* WP: array limit */
	sgs_quicksort( SGSARR_PTR( hdr ), (size_t) hdr->size,
		sizeof( sgs_Variable ), rev ? sgsarrcomp_custom_rev : sgsarrcomp_custom, &u );
	
	SGS_RETURN_THIS( C );
}

typedef struct sgsarr_smi_s
{
	sgs_Real value;
	sgs_SizeVal pos;
}
sgsarr_smi;
static SGS_INLINE int sgsarrcomp_smi( const void* p1, const void* p2, void* userdata )
{
	const sgsarr_smi *v1 = (const sgsarr_smi*) p1;
	const sgsarr_smi *v2 = (const sgsarr_smi*) p2;
	if( v1->value < v2->value )
		return -1;
	return v1->value > v2->value ? 1 : 0;
}
static SGS_INLINE int sgsarrcomp_smi_rev( const void* p1, const void* p2, void* userdata )
{ return sgsarrcomp_smi( p2, p1, userdata ); }
static int sgsstd_arrayI_sort_mapped( SGS_CTX )
{
	sgs_Variable arr;
	sgs_SizeVal i, asize = 0;
	int rev = 0;
	
	SGSARR_IHDR( sort_mapped );
	if( !sgs_LoadArgs( C, "a<v|b", &asize, &arr, &rev ) )
		return 0;
	
	if( asize != hdr->size )
		STDLIB_WARN( "array sizes must match" )
	
	{
		/* WP: array limit */
		sgsarr_smi* smis = sgs_Alloc_n( sgsarr_smi, (size_t) asize );
		for( i = 0; i < asize; ++i )
		{
			if( sgs_PushNumIndex( C, arr, i ) == SGS_FALSE )
			{
				sgs_Dealloc( smis );
				STDLIB_WARN( "error in mapping array" )
			}
			smis[ i ].value = sgs_GetReal( C, -1 );
			smis[ i ].pos = i;
			fstk_pop1( C );
		}
		sgs_quicksort( smis, (size_t) asize, sizeof( sgsarr_smi ),
			rev ? sgsarrcomp_smi_rev : sgsarrcomp_smi, NULL );
		
		{
			sgs_Variable *p1, *p2;
			p1 = SGSARR_PTR( hdr );
			p2 = sgs_Alloc_n( sgs_Variable, (size_t) hdr->mem );
			memcpy( p2, p1, SGSARR_ALLOCSIZE( hdr->mem ) );
			for( i = 0; i < asize; ++i )
				p1[ i ] = p2[ smis[ i ].pos ];
			sgs_Dealloc( p2 );
		}
		
		sgs_Dealloc( smis );
		
		SGS_RETURN_THIS( C );
	}
}

static int sgsstd_arrayI_find( SGS_CTX )
{
	sgs_Variable comp;
	int strict = SGS_FALSE;
	sgs_SizeVal off = 0;
	
	SGSARR_IHDR( find );
	if( !sgs_LoadArgs( C, "v|bl", &comp, &strict, &off ) )
		return 0;
	
	while( off < hdr->size )
	{
		sgs_Variable* p = SGSARR_PTR( hdr ) + off;
		if( ( !strict || sgs_EqualTypes( p, &comp ) )
			&& sgs_Compare( C, p, &comp ) == 0 )
		{
			sgs_PushInt( C, off );
			return 1;
		}
		off++;
	}
	return 0;
}

static int sgsstd_arrayI_remove( SGS_CTX )
{
	sgs_Variable comp;
	int strict = SGS_FALSE, all = SGS_FALSE, rmvd = 0;
	sgs_SizeVal off = 0;
	
	SGSARR_IHDR( remove );
	if( !sgs_LoadArgs( C, "v|bbl", &comp, &strict, &all, &off ) )
		return 0;
	
	while( off < hdr->size )
	{
		sgs_Variable* p = SGSARR_PTR( hdr ) + off;
		if( ( !strict || sgs_EqualTypes( p, &comp ) )
			&& sgs_Compare( C, p, &comp ) == 0 )
		{
			sgsstd_array_erase( C, hdr, off, off );
			rmvd++;
			if( !all )
				break;
		}
		else
			off++;
	}
	sgs_PushInt( C, rmvd );
	return 1;
}

static int _in_array( SGS_CTX, void* data, sgs_Variable* var, int strconv )
{
	sgs_SizeVal off = 0;
	SGSARR_HDR;
	
	if( !strconv )
	{
		while( off < hdr->size )
		{
			sgs_Variable* cur = SGSARR_PTR( hdr ) + off;
			if( sgs_EqualTypes( var, cur ) && sgs_Compare( C, var, cur ) == 0 )
				return SGS_TRUE;
			off++;
		}
		return SGS_FALSE;
	}
	else
	{
		int found = 0;
		sgs_Variable A, B;
		
		A = *var;
		sgs_Acquire( C, &A );
		sgs_ToStringP( C, &A );
		while( off < hdr->size )
		{
			B = SGSARR_PTR( hdr )[ off ];
			sgs_Acquire( C, &B );
			sgs_ToStringP( C, &B );
			
			found = sgs_EqualTypes( &A, &B ) && sgs_Compare( C, &A, &B ) == 0;
			sgs_Release( C, &B );
			if( found )
				break;
			off++;
		}
		sgs_Release( C, &A );
		return found;
	}
}

static int sgsstd_arrayI_unique( SGS_CTX )
{
	int strconv = SGS_FALSE;
	sgs_SizeVal off = 0, asz = 0;
	void* nadata;
	
	SGSARR_IHDR( unique );
	if( !sgs_LoadArgs( C, "|b", &strconv ) )
		return 0;
	
	sgs_CreateArray( C, NULL, 0 );
	nadata = sgs_GetObjectData( C, -1 );
	while( off < hdr->size )
	{
		sgs_Variable* var = SGSARR_PTR( hdr ) + off;
		if( !_in_array( C, nadata, var, strconv ) )
		{
			sgsstd_array_insert_p( C, (sgsstd_array_header_t*) nadata, asz, var );
			asz++;
		}
		off++;
	}
	
	return 1;
}

static int sgsstd_arrayI_random( SGS_CTX )
{
	sgs_SizeVal num, asz = 0;
	sgsstd_array_header_t* nadata;
	
	SGSARR_IHDR( random );
	if( !sgs_LoadArgs( C, "l", &num ) )
		return 0;
	
	if( num < 0 )
		STDLIB_WARN( "argument 1 (count) cannot be negative" )
	
	sgs_CreateArray( C, NULL, 0 );
	nadata = (sgsstd_array_header_t*) sgs_GetObjectData( C, -1 );
	sgsstd_array_reserve( C, nadata, num );
	while( num-- )
	{
		sgsstd_array_insert_p( C, nadata, asz, SGSARR_PTR( hdr ) + ( rand() % hdr->size ) );
		asz++;
	}
	
	return 1;
}

static int sgsstd_arrayI_shuffle( SGS_CTX )
{
	sgs_Variable tmp;
	sgs_SizeVal i, j;
	
	SGSARR_IHDR( shuffle );
	if( !sgs_LoadArgs( C, "." ) )
		return 0;
	
	for( i = hdr->size - 1; i >= 1; i-- )
	{
		j = rand() % ( i + 1 );
		tmp = SGSARR_PTR( hdr )[ i ];
		SGSARR_PTR( hdr )[ i ] = SGSARR_PTR( hdr )[ j ];
		SGSARR_PTR( hdr )[ j ] = tmp;
	}
	
	SGS_RETURN_THIS( C );
}


static int sgsstd_array( SGS_CTX );
static sgs_RegFuncConst array_iface_fconsts[] =
{
	{ "push", sgsstd_arrayI_push },
	{ "pop", sgsstd_arrayI_pop },
	{ "shift", sgsstd_arrayI_shift },
	{ "unshift", sgsstd_arrayI_unshift },
	{ "insert", sgsstd_arrayI_insert },
	{ "erase", sgsstd_arrayI_erase },
	{ "part", sgsstd_arrayI_part },
	{ "clear", sgsstd_arrayI_clear },
	{ "reverse", sgsstd_arrayI_reverse },
	{ "resize", sgsstd_arrayI_resize },
	{ "reserve", sgsstd_arrayI_reserve },
	{ "sort", sgsstd_arrayI_sort },
	{ "sort_custom", sgsstd_arrayI_sort_custom },
	{ "sort_mapped", sgsstd_arrayI_sort_mapped },
	{ "find", sgsstd_arrayI_find },
	{ "remove", sgsstd_arrayI_remove },
	{ "unique", sgsstd_arrayI_unique },
	{ "random", sgsstd_arrayI_random },
	{ "shuffle", sgsstd_arrayI_shuffle },
	{ "__call", sgsstd_array },
};

static int sgsstd_array_iface_gen( SGS_CTX )
{
	sgs_CreateDict( C, NULL, 0 );
	sgs_StoreFuncConsts( C, *stk_gettop( C ), array_iface_fconsts,
		SGS_ARRAY_SIZE(array_iface_fconsts), "array." );
	sgs_ObjSetMetaMethodEnable( p_obj( stk_gettop( C ) ), 1 );
	return 1;
}

static int sgsstd_array_getprop( SGS_CTX, void* data )
{
	char* name;
	SGSARR_HDR;
	if( sgs_ParseString( C, 0, &name, NULL ) )
	{
		if( 0 == strcmp( name, "size" ) )
		{
			sgs_PushInt( C, hdr->size );
			return SGS_SUCCESS;
		}
		else if( 0 == strcmp( name, "capacity" ) )
		{
			sgs_PushInt( C, hdr->mem );
			return SGS_SUCCESS;
		}
		else if( 0 == strcmp( name, "first" ) )
		{
			if( hdr->size )
			{
				fstk_push( C, &SGSARR_PTR( hdr )[ 0 ] );
			}
			else
			{
				fstk_push_null( C );
				sgs_Msg( C, SGS_WARNING, "array is empty, cannot get first item" );
			}
			return SGS_SUCCESS;
		}
		else if( 0 == strcmp( name, "last" ) )
		{
			if( hdr->size )
			{
				fstk_push( C, &SGSARR_PTR( hdr )[ hdr->size - 1 ] );
			}
			else
			{
				fstk_push_null( C );
				sgs_Msg( C, SGS_WARNING, "array is empty, cannot get last item" );
			}
			return SGS_SUCCESS;
		}
	}
	return SGS_ENOTFND;
}

int sgsstd_array_getindex( SGS_ARGS_GETINDEXFUNC )
{
	if( C->object_arg )
		return sgsstd_array_getprop( C, obj->data );
	else
	{
		SGSARR_HDR_OI;
		sgs_Variable* ptr = SGSARR_PTR( hdr );
		sgs_Int i = sgs_GetInt( C, 0 );
		if( i < 0 || i >= hdr->size )
		{
			sgs_Msg( C, SGS_WARNING, "array index out of bounds" );
			return SGS_EBOUNDS;
		}
		fstk_push( C, &ptr[ i ] );
		return SGS_SUCCESS;
	}
}

static int sgsstd_array_setindex( SGS_ARGS_SETINDEXFUNC )
{
	if( C->object_arg )
		return SGS_ENOTSUP;
	else
	{
		SGSARR_HDR_OI;
		sgs_Variable* ptr = SGSARR_PTR( hdr );
		sgs_Int i = sgs_GetInt( C, 0 );
		if( i < 0 || i >= hdr->size )
		{
			sgs_Msg( C, SGS_WARNING, "array index out of bounds" );
			return SGS_EBOUNDS;
		}
		sgs_Release( C, ptr + i );
		sgs_StoreVariable( C, ptr + i );
		return SGS_SUCCESS;
	}
}

static int sgsstd_array_dump( SGS_CTX, sgs_VarObj* obj, int depth )
{
	char bfr[ 32 ];
	int i, ssz = stk_size( C );
	SGSARR_HDR_OI;
	sprintf( bfr, "array (%" PRId32")\n[", hdr->size );
	sgs_PushString( C, bfr );
	if( depth )
	{
		if( hdr->size )
		{
			for( i = 0; i < hdr->size; ++i )
			{
				sgs_PushString( C, "\n" );
				sgs_DumpVar( C, SGSARR_PTR( hdr )[ i ], depth );
			}
			sgs_StringConcat( C, hdr->size * 2 );
			sgs_PadString( C );
		}
	}
	else
	{
		sgs_PushString( C, "\n..." );
		sgs_PadString( C );
	}
	sgs_PushString( C, "\n]" );
	sgs_StringConcat( C, stk_size( C ) - ssz );
	return SGS_SUCCESS;
}

static int sgsstd_array_gcmark( SGS_CTX, sgs_VarObj* obj )
{
	SGSARR_HDR_OI;
	sgs_GCMarkArray( C, SGSARR_PTR( hdr ), hdr->size );
	return SGS_SUCCESS;
}

/* iterator */

static int sgsstd_array_iter_destruct( SGS_CTX, sgs_VarObj* obj )
{
	sgs_Release( C, &((sgsstd_array_iter_t*) obj->data)->ref );
	return SGS_SUCCESS;
}

static int sgsstd_array_iter_getnext( SGS_CTX, sgs_VarObj* obj, int mask )
{
	sgsstd_array_iter_t* iter = (sgsstd_array_iter_t*) obj->data;
	sgsstd_array_header_t* hdr = (sgsstd_array_header_t*) iter->ref.data.O->data;
	if( iter->size != hdr->size )
	{
		sgs_Msg( C, SGS_ERROR, "array changed size during iteration" );
		return SGS_EINPROC;
	}
	
	if( !mask )
	{
		iter->off++;
		return iter->off < iter->size;
	}
	else
	{
		if( mask & SGS_GETNEXT_KEY )
			sgs_PushInt( C, iter->off );
		if( mask & SGS_GETNEXT_VALUE )
		{
			fstk_push( C, &SGSARR_PTR( hdr )[ iter->off ] );
		}
		return SGS_SUCCESS;
	}
}

static int sgsstd_array_iter_gcmark( SGS_CTX, sgs_VarObj* obj )
{
	sgsstd_array_iter_t* iter = (sgsstd_array_iter_t*) obj->data;
	sgs_GCMark( C, &iter->ref );
	return SGS_SUCCESS;
}

SGS_APIFUNC sgs_ObjInterface sgsstd_array_iter_iface[1] =
{{
	"array_iterator",
	sgsstd_array_iter_destruct, sgsstd_array_iter_gcmark,
	NULL, NULL,
	NULL, NULL, NULL, sgsstd_array_iter_getnext,
	NULL, NULL
}};

static int sgsstd_array_convert( SGS_CTX, sgs_VarObj* obj, int type )
{
	SGSARR_HDR_OI;
	if( type == SGS_CONVOP_TOITER )
	{
		sgsstd_array_iter_t* iter = (sgsstd_array_iter_t*)
			sgs_CreateObjectIPA( C, NULL, sizeof(*iter), sgsstd_array_iter_iface );
		
		sgs_InitObjectPtr( &iter->ref, obj ); /* acquires ref */
		iter->size = hdr->size;
		iter->off = -1;
		
		return SGS_SUCCESS;
	}
	else if( type == SGS_VT_BOOL )
	{
		sgs_PushBool( C, !!hdr->size );
		return SGS_SUCCESS;
	}
	else if( type == SGS_VT_STRING )
	{
		sgs_Variable* var = SGSARR_PTR( obj->data );
		sgs_Variable* vend = var + hdr->size;
		
		sgs_PushString( C, "[" );
		while( var < vend )
		{
			fstk_push( C, var );
			sgs_ToStringFast( C, -1 );
			var++;
			if( var < vend )
				sgs_PushString( C, "," );
		}
		sgs_PushString( C, "]" );
		sgs_StringConcat( C, hdr->size * 2 + 1 + !hdr->size );
		return SGS_SUCCESS;
	}
	else if( type == SGS_CONVOP_CLONE )
	{
		sgsstd_array_header_t* hdr2 = (sgsstd_array_header_t*)
			sgs_CreateObjectIPA( C, NULL, sizeof( sgsstd_array_header_t ), sgsstd_array_iface );
		memcpy( hdr2, hdr, sizeof( sgsstd_array_header_t ) );
		hdr2->data = sgs_Alloc_n( sgs_Variable, (size_t) hdr->mem );
		memcpy( hdr2->data, hdr->data, SGSARR_ALLOCSIZE( hdr->mem ) );
		{
			sgs_Variable* ptr = SGSARR_PTR( hdr );
			sgs_Variable* pend = ptr + hdr->size;
			while( ptr < pend )
			{
				sgs_Acquire( C, ptr );
				ptr++;
			}
		}
		
		sgs_ObjSetMetaObj( C, p_obj( stk_ptop( C, -1 ) ), C->shared->array_iface );
		
		return SGS_SUCCESS;
	}
	return SGS_ENOTSUP;
}

static int sgsstd_array_serialize( SGS_CTX, sgs_VarObj* obj )
{
	sgs_Variable* pos, *posend;
	SGSARR_HDR_OI;
	pos = SGSARR_PTR( hdr );
	posend = pos + hdr->size;
	if( C->object_arg == 2 )
	{
		sgs_Variable sv;
		int32_t i = 0;
		sv.type = SGS_VT_INT;
		sv.data.I = hdr->size;
		sgs_Serialize( C, sv );
		sgs_SerializeObject( C, 1, "array_sized" );
		while( pos < posend )
		{
			sv.data.I = i++;
			sgs_SerializeObjIndex( C, sv, *pos++, SGS_FALSE );
		}
	}
	else
	{
		while( pos < posend )
		{
			sgs_Serialize( C, *pos++ );
		}
		sgs_SerializeObject( C, hdr->size, "array" );
	}
	return SGS_SUCCESS;
}

static int sgsstd_array_destruct( SGS_CTX, sgs_VarObj* obj )
{
	SGSARR_HDR_OI;
	sgsstd_array_clear( C, hdr );
	sgs_Dealloc( hdr->data );
	return 0;
}

SGS_APIFUNC sgs_ObjInterface sgsstd_array_iface[1] =
{{
	"array",
	sgsstd_array_destruct, sgsstd_array_gcmark,
	sgsstd_array_getindex, sgsstd_array_setindex,
	sgsstd_array_convert, sgsstd_array_serialize, sgsstd_array_dump, NULL,
	NULL, NULL
}};

static int sgsstd_array( SGS_CTX )
{
	/* first argument of __call metamethod = self */
	int i = 1, objcnt = stk_size( C ) - 1;
	void* data = sgs_Malloc( C, SGSARR_ALLOCSIZE( objcnt ) );
	sgs_Variable *p, *pend;
	sgsstd_array_header_t* hdr;
	
	SGSFN( "array" );
	hdr = (sgsstd_array_header_t*)
		sgs_CreateObjectIPA( C, NULL, sizeof( sgsstd_array_header_t ), sgsstd_array_iface );
	hdr->size = objcnt;
	hdr->mem = objcnt;
	p = hdr->data = (sgs_Variable*) data;
	pend = p + objcnt;
	while( p < pend )
		sgs_GetStackItem( C, i++, p++ );
	
	sgs_ObjSetMetaObj( C, p_obj( stk_ptop( C, -1 ) ), C->shared->array_iface );
	
	return 1;
}

static int sgsstd_array_sized( SGS_CTX )
{
	sgs_SizeVal size = 0;
	sgs_Variable *p, *pend;
	sgsstd_array_header_t* hdr;
	void* data;
	
	SGSFN( "array_sized" );
	if( !sgs_LoadArgs( C, "l", &size ) )
		return 0;
	
	data = sgs_Malloc( C, SGSARR_ALLOCSIZE( size ) );
	hdr = (sgsstd_array_header_t*)
		sgs_CreateObjectIPA( C, NULL, sizeof( sgsstd_array_header_t ), sgsstd_array_iface );
	hdr->size = size;
	hdr->mem = size;
	p = hdr->data = (sgs_Variable*) data;
	pend = p + size;
	while( p < pend )
		(p++)->type = SGS_VT_NULL;
	
	sgs_ObjSetMetaObj( C, p_obj( stk_ptop( C, -1 ) ), C->shared->array_iface );
	
	return 1;
}


/*
	VHT containers
*/

typedef
struct _DictHdr
{
	sgs_VHTable ht;
}
DictHdr;

#define HTHDR DictHdr* dh = (DictHdr*) obj->data; sgs_VHTable* ht = &dh->ht;

static int sgsstd_vht_serialize( SGS_CTX, sgs_VarObj* obj, const char* initfn )
{
	sgs_VHTVar *pair, *pend;
	HTHDR;
	pair = ht->vars;
	pend = ht->vars + sgs_vht_size( ht );
	if( C->object_arg == 2 )
	{
		sgs_SerializeObject( C, 0, initfn );
		while( pair < pend )
		{
			sgs_SerializeObjIndex( C, pair->key, pair->val, SGS_FALSE );
			pair++;
		}
	}
	else
	{
		while( pair < pend )
		{
			sgs_Serialize( C, pair->key );
			sgs_Serialize( C, pair->val );
			pair++;
		}
		sgs_SerializeObject( C, sgs_vht_size( ht ) * 2, initfn );
	}
	return SGS_SUCCESS;
}

static int sgsstd_vht_dump( SGS_CTX, sgs_VarObj* obj, int depth, const char* name )
{
	int ssz;
	char bfr[ 32 ];
	sgs_VHTVar *pair, *pend;
	HTHDR;
	pair = ht->vars;
	pend = ht->vars + sgs_vht_size( ht );
	ssz = stk_size( C );
	sprintf( bfr, "%s (%" PRId32")\n{", name, sgs_vht_size( ht ) );
	sgs_PushString( C, bfr );
	if( depth )
	{
		if( sgs_vht_size( ht ) )
		{
			while( pair < pend )
			{
				sgs_PushString( C, "\n" );
				if( pair->key.type == SGS_VT_STRING )
				{
					fstk_push( C, &pair->key );
					sgs_ToPrintSafeString( C );
				}
				else
				{
					sgs_DumpVar( C, pair->key, depth );
				}
				sgs_PushString( C, " = " );
				sgs_DumpVar( C, pair->val, depth );
				pair++;
			}
			/* WP: stack limit */
			sgs_StringConcat( C, (sgs_StkIdx) ( pend - ht->vars ) * 4 );
			sgs_PadString( C );
		}
	}
	else
	{
		sgs_PushString( C, "\n..." );
		sgs_PadString( C );
	}
	sgs_PushString( C, "\n}" );
	sgs_StringConcat( C, stk_size( C ) - ssz );
	return SGS_SUCCESS;
}


/*
	DICT
*/

static DictHdr* mkdict( SGS_CTX, sgs_Variable* out )
{
	sgs_ObjInterface* iface = sgsstd_dict_iface;
	SGS_IF_DLL( ;,
	if( !iface )
		iface = sgsstd_dict_iface;
	)
	DictHdr* dh = (DictHdr*) sgs_CreateObjectIPA( C, out, sizeof( DictHdr ), iface );
	sgs_vht_init( &dh->ht, C, 4, 4 );
	return dh;
}


static int sgsstd_dict_destruct( SGS_CTX, sgs_VarObj* obj )
{
	HTHDR;
	sgs_vht_free( ht, C );
	return SGS_SUCCESS;
}

static int sgsstd_dict_gcmark( SGS_CTX, sgs_VarObj* obj )
{
	sgs_VHTVar *pair, *pend;
	HTHDR;
	pair = ht->vars;
	pend = ht->vars + sgs_vht_size( ht );
	while( pair < pend )
	{
		sgs_GCMark( C, &pair->key );
		sgs_GCMark( C, &pair->val );
		pair++;
	}
	return SGS_SUCCESS;
}

static int sgsstd_dict_dump( SGS_CTX, sgs_VarObj* obj, int depth )
{
	return sgsstd_vht_dump( C, obj, depth, "dict" );
}

/* ref'd in sgs_proc.c */ int sgsstd_dict_getindex( SGS_ARGS_GETINDEXFUNC )
{
	sgs_VHTVar* pair = NULL;
	HTHDR;
	
	if( sgs_ParseString( C, 0, NULL, NULL ) )
	{
		pair = sgs_vht_get( ht, stk_poff( C, 0 ) );
		if( !pair )
			return SGS_ENOTFND;
		
		fstk_push( C, &pair->val );
		return SGS_SUCCESS;
	}
	return SGS_EINVAL;
}

static int sgsstd_dict_setindex( SGS_ARGS_SETINDEXFUNC )
{
	HTHDR;
	if( sgs_ParseString( C, 0, NULL, NULL ) )
	{
		sgs_vht_set( ht, C, stk_poff( C, 0 ), stk_poff( C, 1 ) );
		return SGS_SUCCESS;
	}
	return SGS_EINVAL;
}

/* iterator */

typedef struct sgsstd_dict_iter_s
{
	sgs_Variable ref;
	int32_t size;
	int32_t off;
}
sgsstd_dict_iter_t;

static int sgsstd_dict_iter_destruct( SGS_CTX, sgs_VarObj* obj )
{
	sgs_Release( C, &((sgsstd_dict_iter_t*) obj->data)->ref );
	return SGS_SUCCESS;
}

static int sgsstd_dict_iter_getnext( SGS_CTX, sgs_VarObj* obj, int flags )
{
	sgsstd_dict_iter_t* iter = (sgsstd_dict_iter_t*) obj->data;
	sgs_VHTable* ht = (sgs_VHTable*) iter->ref.data.O->data;
	if( iter->size != sgs_vht_size( ht ) )
		return SGS_EINVAL;
	
	if( !flags )
	{
		iter->off++;
		return iter->off < iter->size;
	}
	else
	{
		if( flags & SGS_GETNEXT_KEY )
		{
			fstk_push( C, &ht->vars[ iter->off ].key );
		}
		if( flags & SGS_GETNEXT_VALUE )
		{
			fstk_push( C, &ht->vars[ iter->off ].val );
		}
		return SGS_SUCCESS;
	}
}

static int sgsstd_dict_iter_gcmark( SGS_CTX, sgs_VarObj* obj )
{
	sgsstd_dict_iter_t* iter = (sgsstd_dict_iter_t*) obj->data;
	sgs_GCMark( C, &iter->ref );
	return SGS_SUCCESS;
}

SGS_APIFUNC sgs_ObjInterface sgsstd_dict_iter_iface[1] =
{{
	"dict_iterator",
	sgsstd_dict_iter_destruct, sgsstd_dict_iter_gcmark,
	NULL, NULL,
	NULL, NULL, NULL, sgsstd_dict_iter_getnext,
	NULL, NULL
}};


static int sgsstd_dict_convert( SGS_CTX, sgs_VarObj* obj, int type )
{
	HTHDR;
	if( type == SGS_CONVOP_TOITER )
	{
		sgsstd_dict_iter_t* iter = (sgsstd_dict_iter_t*)
			sgs_CreateObjectIPA( C, NULL, sizeof(*iter), sgsstd_dict_iter_iface );
		
		sgs_InitObjectPtr( &iter->ref, obj ); /* acquires ref */
		iter->size = sgs_vht_size( ht );
		iter->off = -1;
		
		return SGS_SUCCESS;
	}
	else if( type == SGS_VT_BOOL )
	{
		sgs_PushBool( C, sgs_vht_size( ht ) != 0 );
		return SGS_SUCCESS;
	}
	else if( type == SGS_VT_STRING )
	{
		sgs_VHTVar *pair = ht->vars, *pend = ht->vars + sgs_vht_size( ht );
		int cnt = 0;
		sgs_PushString( C, "{" );
		while( pair < pend )
		{
			if( cnt )
				sgs_PushString( C, "," );
			fstk_push( C, &pair->key );
			sgs_PushString( C, "=" );
			fstk_push( C, &pair->val );
			sgs_ToStringFast( C, -1 );
			cnt++;
			pair++;
		}
		sgs_PushString( C, "}" );
		sgs_StringConcat( C, cnt * 4 + 1 + !cnt );
		return SGS_SUCCESS;
	}
	else if( type == SGS_CONVOP_CLONE )
	{
		int i, htsize = sgs_vht_size( ht );
		DictHdr* ndh = mkdict( C, NULL );
		for( i = 0; i < htsize; ++i )
		{
			sgs_vht_set( &ndh->ht, C, &ht->vars[ i ].key, &ht->vars[ i ].val );
		}
		return SGS_SUCCESS;
	}
	return SGS_ENOTSUP;
}

static int sgsstd_dict_serialize( SGS_CTX, sgs_VarObj* obj )
{
	return sgsstd_vht_serialize( C, obj, "dict" );
}

SGS_APIFUNC sgs_ObjInterface sgsstd_dict_iface[1] =
{{
	"dict",
	sgsstd_dict_destruct, sgsstd_dict_gcmark,
	sgsstd_dict_getindex, sgsstd_dict_setindex,
	sgsstd_dict_convert, sgsstd_dict_serialize, sgsstd_dict_dump, NULL,
	NULL, NULL
}};

static int sgsstd_dict( SGS_CTX )
{
	DictHdr* dh;
	sgs_VHTable* ht;
	int i, objcnt = stk_size( C );
	
	SGSFN( "dict" );
	
	if( objcnt % 2 != 0 )
		STDLIB_WARN( "function expects 0 or an even number of arguments" )
	
	dh = mkdict( C, NULL );
	ht = &dh->ht;
	
	for( i = 0; i < objcnt; i += 2 )
	{
		sgs_ToString( C, i );
		sgs_vht_set( ht, C, stk_poff( C, i ), stk_poff( C, i + 1 ) );
	}
	
	return 1;
}


/* MAP */

static DictHdr* mkmap( SGS_CTX, sgs_Variable* out )
{
	DictHdr* dh = (DictHdr*) sgs_CreateObjectIPA( C, out, sizeof( DictHdr ), sgsstd_map_iface );
	sgs_vht_init( &dh->ht, C, 4, 4 );
	return dh;
}

#define sgsstd_map_destruct sgsstd_dict_destruct
#define sgsstd_map_gcmark sgsstd_dict_gcmark

static int sgsstd_map_dump( SGS_CTX, sgs_VarObj* obj, int depth )
{
	return sgsstd_vht_dump( C, obj, depth, "map" );
}

static int sgsstd_map_serialize( SGS_CTX, sgs_VarObj* obj )
{
	return sgsstd_vht_serialize( C, obj, "map" );
}

static int sgsstd_map_convert( SGS_CTX, sgs_VarObj* obj, int type )
{
	HTHDR;
	if( type == SGS_CONVOP_TOITER )
	{
		sgsstd_dict_iter_t* iter = (sgsstd_dict_iter_t*)
			sgs_CreateObjectIPA( C, NULL, sizeof(*iter), sgsstd_dict_iter_iface );
		
		sgs_InitObjectPtr( &iter->ref, obj ); /* acquires ref */
		iter->size = sgs_vht_size( ht );
		iter->off = -1;
		
		return SGS_SUCCESS;
	}
	else if( type == SGS_VT_BOOL )
	{
		sgs_PushBool( C, sgs_vht_size( ht ) != 0 );
		return SGS_SUCCESS;
	}
	else if( type == SGS_VT_STRING )
	{
		sgs_VHTVar *pair = ht->vars, *pend = ht->vars + sgs_vht_size( ht );
		int cnt = 0;
		sgs_PushString( C, "[map]{" );
		while( pair < pend )
		{
			if( cnt )
				sgs_PushString( C, "," );
			fstk_push( C, &pair->key );
			sgs_ToStringFast( C, -1 );
			sgs_PushString( C, "=" );
			fstk_push( C, &pair->val );
			sgs_ToStringFast( C, -1 );
			cnt++;
			pair++;
		}
		sgs_PushString( C, "}" );
		sgs_StringConcat( C, cnt * 4 + 1 + !cnt );
		return SGS_SUCCESS;
	}
	else if( type == SGS_CONVOP_CLONE )
	{
		int i, htsize = sgs_vht_size( ht );
		DictHdr* ndh = mkmap( C, NULL );
		for( i = 0; i < htsize; ++i )
		{
			sgs_vht_set( &ndh->ht, C, &ht->vars[ i ].key, &ht->vars[ i ].val );
		}
		return SGS_SUCCESS;
	}
	return SGS_ENOTSUP;
}

/* copy in sgs_proc.c */ static int sgsstd_map_getindex( SGS_ARGS_GETINDEXFUNC )
{
	sgs_VHTVar* pair = NULL;
	HTHDR;
	
	pair = sgs_vht_get( ht, stk_poff( C, 0 ) );
	if( !pair )
		return SGS_ENOTFND;
	
	fstk_push( C, &pair->val );
	return SGS_SUCCESS;
}

static int sgsstd_map_setindex( SGS_ARGS_SETINDEXFUNC )
{
	HTHDR;
	sgs_vht_set( ht, C, stk_poff( C, 0 ), stk_poff( C, 1 ) );
	return SGS_SUCCESS;
}

SGS_APIFUNC sgs_ObjInterface sgsstd_map_iface[1] =
{{
	"map",
	sgsstd_map_destruct, sgsstd_map_gcmark,
	sgsstd_map_getindex, sgsstd_map_setindex,
	sgsstd_map_convert, sgsstd_map_serialize, sgsstd_map_dump, NULL,
	NULL, NULL
}};

static int sgsstd_map( SGS_CTX )
{
	DictHdr* dh;
	sgs_VHTable* ht;
	int i, objcnt = stk_size( C );
	
	SGSFN( "map" );
	
	if( objcnt % 2 != 0 )
		STDLIB_WARN( "function expects 0 or an even number of arguments" )
	
	dh = mkmap( C, NULL );
	ht = &dh->ht;
	
	for( i = 0; i < objcnt; i += 2 )
	{
		sgs_vht_set( ht, C, stk_poff( C, i ), stk_poff( C, i + 1 ) );
	}
	
	return 1;
}


static int sgsstd_class( SGS_CTX )
{
	sgs_VarObj *obj1, *obj2;
	SGSFN( "class" );
	if( !sgs_LoadArgs( C, "!x!x", sgs_ArgCheck_Object, &obj1, sgs_ArgCheck_Object, &obj2 ) )
		return 0;
	sgs_ObjSetMetaObj( C, obj1, obj2 );
	sgs_ObjSetMetaMethodEnable( obj1, SGS_TRUE );
	sgs_SetStackSize( C, 1 );
	return 1;
}


/*
	closure memory layout:
	- sgs_Variable: function
	- int32: closure count
	- sgs_Closure* x ^^: closures
*/

static int sgsstd_closure_destruct( SGS_CTX, sgs_VarObj* obj )
{
	uint8_t* cl = (uint8_t*) obj->data;
	sgs_clsrcount_t i, cc = *SGS_ASSUME_ALIGNED( cl + sizeof(sgs_Variable), sgs_clsrcount_t );
	sgs_Closure** cls = SGS_ASSUME_ALIGNED( cl + sizeof(sgs_Variable) + sizeof(sgs_clsrcount_t), sgs_Closure* );
	
	sgs_Release( C, SGS_ASSUME_ALIGNED( cl, sgs_Variable ) );
	
	for( i = 0; i < cc; ++i )
	{
		if( --cls[ i ]->refcount < 1 )
		{
			sgs_Release( C, &cls[ i ]->var );
			sgs_Dealloc( cls[i] );
		}
	}
	
	return SGS_SUCCESS;
}

static int sgsstd_closure_getindex( SGS_ARGS_GETINDEXFUNC )
{
	char* str;
	if( sgs_ParseString( C, 0, &str, NULL ) )
	{
		if( !strcmp( str, "apply" ) )
		{
			sgs_PushCFunc( C, sgs_specfn_apply );
			return SGS_SUCCESS;
		}
	}
	return SGS_ENOTFND;
}

static int sgsstd_closure_call( SGS_CTX, sgs_VarObj* obj )
{
	sgs_BreakIf( "implemented elsewhere" );
	return 0;
}

static int sgsstd_closure_gcmark( SGS_CTX, sgs_VarObj* obj )
{
	uint8_t* cl = (uint8_t*) obj->data;
	sgs_clsrcount_t i, cc = *SGS_ASSUME_ALIGNED( cl + sizeof(sgs_Variable), sgs_clsrcount_t );
	sgs_Closure** cls = SGS_ASSUME_ALIGNED( cl + sizeof(sgs_Variable) + sizeof(sgs_clsrcount_t), sgs_Closure* );
	
	sgs_GCMark( C, SGS_ASSUME_ALIGNED( cl, sgs_Variable ) );
	
	for( i = 0; i < cc; ++i )
	{
		sgs_GCMark( C, &cls[ i ]->var );
	}
	
	return SGS_SUCCESS;
}

static int sgsstd_closure_dump( SGS_CTX, sgs_VarObj* obj, int depth )
{
	uint8_t* cl = (uint8_t*) obj->data;
	sgs_StkIdx ssz;
	sgs_clsrcount_t i, cc = *SGS_ASSUME_ALIGNED( cl + sizeof(sgs_Variable), sgs_clsrcount_t );
	sgs_Closure** cls = SGS_ASSUME_ALIGNED( cl + sizeof(sgs_Variable) + sizeof(sgs_clsrcount_t), sgs_Closure* );
	
	sgs_PushString( C, "closure\n{" );
	
	ssz = stk_size( C );
	sgs_PushString( C, "\nfunc: " );
	sgs_DumpVar( C, *SGS_ASSUME_ALIGNED( cl, sgs_Variable ), depth ); /* function */
	for( i = 0; i < cc; ++i )
	{
		char intro[ 64 ];
		sprintf( intro, "\n#%d (rc=%" PRId32"): ", (int) i, cls[ i ]->refcount );
		sgs_PushString( C, intro );
		sgs_DumpVar( C, cls[ i ]->var, depth );
	}
	sgs_StringConcat( C, stk_size( C ) - ssz );
	sgs_PadString( C );
	
	sgs_PushString( C, "\n}" );
	sgs_StringConcat( C, 3 );
	return SGS_SUCCESS;
}

SGS_APIFUNC sgs_ObjInterface sgsstd_closure_iface[1] =
{{
	"closure",
	sgsstd_closure_destruct, sgsstd_closure_gcmark,
	sgsstd_closure_getindex, NULL,
	NULL, NULL, sgsstd_closure_dump, NULL,
	sgsstd_closure_call, NULL
}};

sgs_Closure** sgsSTD_MakeClosure( SGS_CTX, sgs_Variable* out, sgs_Variable* func, sgs_clsrcount_t clc )
{
	/* WP: range not affected by conversion */
	uint32_t clsz = (uint32_t) ( sizeof(sgs_Closure*) * clc );
	uint32_t memsz = clsz + (uint32_t) ( sizeof(sgs_Variable) + sizeof(clc) );
	uint8_t* cl = (uint8_t*) sgs_CreateObjectIPA( C, out, memsz, sgsstd_closure_iface );
	
	memcpy( cl, func, sizeof(sgs_Variable) );
	sgs_Acquire( C, func );
	
	memcpy( cl + sizeof(sgs_Variable), &clc, sizeof(clc) );
	
	return SGS_ASSUME_ALIGNED( cl + sizeof(sgs_Variable) + sizeof(clc), sgs_Closure* );
}


/* UTILITIES */

static int sgsstd_array_filter( SGS_CTX )
{
	SGSBOOL cset = 0, use;
	sgs_SizeVal asz, off = 0, nasz = 0;
	sgs_Variable v_func;
	void *data;
	sgsstd_array_header_t* nadata;
	
	p_initnull( &v_func );
	
	SGSFN( "array_filter" );
	if( !sgs_LoadArgs( C, "a|p<v", &asz, &cset, &v_func ) )
		return 0;
	
	sgs_CreateArray( C, NULL, 0 );
	data = sgs_GetObjectData( C, 0 );
	nadata = (sgsstd_array_header_t*) sgs_GetObjectData( C, -1 );
	
	{
		SGSARR_HDR;
		
		while( off < asz )
		{
			if( cset )
			{
				fstk_push( C, &v_func );
			}
			fstk_push( C, &SGSARR_PTR( hdr )[ off ] );
			if( cset )
			{
				sgs_PushInt( C, off );
				sgs_Call( C, 2, 1 );
			}
			use = sgs_GetBool( C, -1 );
			fstk_pop1( C );
			fstk_push( C, &SGSARR_PTR( hdr )[ off ] );
			if( use )
			{
				sgsstd_array_insert( C, nadata, nasz, stk_size( C ) - 1 );
				nasz++;
			}
			fstk_pop1( C );
			
			off++;
		}
	}
	
	return 1;
}

static int sgsstd_array_process( SGS_CTX )
{
	sgs_SizeVal asz, off = 0;
	sgs_Variable v_func;
	void *data;
	
	SGSFN( "array_process" );
	if( !sgs_LoadArgs( C, "a?p<v", &asz, &v_func ) )
		return 0;
	
	data = sgs_GetObjectData( C, 0 );
	{
		SGSARR_HDR;
		
		while( off < asz )
		{
			fstk_push2( C, &v_func, &SGSARR_PTR( hdr )[ off ] );
			sgs_PushInt( C, off );
			sgs_Call( C, 2, 1 );
			sgs_SetIndex( C, *stk_poff( C, 0 ), sgs_MakeInt( off ), *stk_gettop( C ), SGS_FALSE );
			fstk_pop1( C );
			off++;
		}
	}
	
	sgs_SetStackSize( C, 1 );
	return 1;
}

static int _sgsstd_ht_filter( SGS_CTX, int usemap )
{
	SGSBOOL cset = 0, use;
	sgs_Variable v_func, v_dest;
	
	p_initnull( &v_func );
	
	if( !sgs_LoadArgs( C, usemap ? "?h|p<v" : "?t|p<v", &cset, &v_func ) )
		return 0;
	
	if( usemap )
		sgs_CreateMap( C, NULL, 0 );
	else
		sgs_CreateDict( C, NULL, 0 );
	v_dest = *stk_gettop( C );
	sgs_PushIterator( C, *stk_poff( C, 0 ) );
	while( sgs_IterAdvance( C, *stk_gettop( C ) ) > 0 )
	{
		sgs_IterPushData( C, *stk_gettop( C ), 1, 1 );
		if( cset )
		{
			fstk_push( C, &v_func );
			sgs_PushItem( C, -2 );
			sgs_PushItem( C, -4 );
			sgs_Call( C, 2, 1 );
		}
		use = sgs_GetBool( C, -1 );
		if( cset )
			fstk_pop1( C );
		if( use )
		{
			/* src-dict, ... dest-dict, iterator, key, value */
			sgs_SetIndex( C, v_dest, *stk_ptop( C, -2 ), *stk_gettop( C ), SGS_FALSE );
			fstk_pop1( C );
		}
		sgs_Pop( C, use ? 1 : 2 );
	}
	fstk_pop1( C );
	return 1;
}

static int sgsstd_dict_filter( SGS_CTX )
{
	SGSFN( "dict_filter" );
	return _sgsstd_ht_filter( C, 0 );
}

static int sgsstd_map_filter( SGS_CTX )
{
	SGSFN( "map_filter" );
	return _sgsstd_ht_filter( C, 1 );
}

static int sgsstd_map_process( SGS_CTX )
{
	sgs_Variable v_func;
	SGSFN( "map_process" );
	if( !sgs_LoadArgs( C, "?v?p<v", &v_func ) )
		return 0;
	
	sgs_PushIterator( C, *stk_poff( C, 0 ) );
	while( sgs_IterAdvance( C, *stk_gettop( C ) ) > 0 )
	{
		sgs_IterPushData( C, *stk_gettop( C ), 1, 1 );
		sgs_PushItem( C, -2 );
		sgs_InsertVariable( C, -3, v_func );
		sgs_Call( C, 2, 1 );
		/* src-dict, callable, ... iterator, key, proc.val. */
		sgs_SetIndex( C, *stk_poff( C, 0 ), *stk_ptop( C, -2 ), *stk_gettop( C ), SGS_FALSE );
		sgs_Pop( C, 2 );
	}
	
	sgs_SetStackSize( C, 1 );
	return 1;
}


static int sgsstd_dict_size( SGS_CTX )
{
	sgs_SizeVal size;
	
	SGSFN( "dict_size" );
	if( !sgs_LoadArgs( C, "t.", &size ) )
		return 0;
	
	sgs_PushInt( C, size );
	return 1;
}

static int sgsstd_map_size( SGS_CTX )
{
	sgs_SizeVal size;
	
	SGSFN( "map_size" );
	if( !sgs_LoadArgs( C, "h.", &size ) )
		return 0;
	
	sgs_PushInt( C, size );
	return 1;
}

static int sgsstd_isset( SGS_CTX )
{
	int ret;
	int32_t oldapilev;
	SGSFN( "isset" );
	if( !sgs_LoadArgs( C, "?v?v." ) )
		return 0;
	
	oldapilev = sgs_Cntl( C, SGS_CNTL_APILEV, SGS_ERROR );
	ret = sgs_PushIndex( C, *stk_poff( C, 0 ), *stk_poff( C, 1 ), SGS_FALSE );
	sgs_Cntl( C, SGS_CNTL_APILEV, oldapilev );
	sgs_PushBool( C, ret );
	return 1;
}

static int sgsstd_unset( SGS_CTX )
{
	DictHdr* dh;
	
	SGSFN( "unset" );
	
	if( !sgs_IsObject( C, 0, sgsstd_dict_iface ) &&
		!sgs_IsObject( C, 0, sgsstd_map_iface ) )
		return sgs_ArgErrorExt( C, 0, 0, "dict / map", "" );
	
	dh = (DictHdr*) sgs_GetObjectData( C, 0 );
	
	if( sgs_IsObject( C, 0, sgsstd_dict_iface ) )
	{
		if( !sgs_LoadArgs( C, ">?m." ) )
			return 0;
	}
	else
	{
		if( !sgs_LoadArgs( C, ">?v." ) )
			return 0;
	}
	sgs_vht_unset( &dh->ht, C, stk_poff( C, 1 ) );
	
	return 0;
}

static int sgsstd_clone( SGS_CTX )
{
	SGSFN( "clone" );
	
	if( !sgs_LoadArgs( C, "?v." ) )
		return 0;
	
	sgs_CloneItem( C, *stk_poff( C, 0 ) );
	return 1;
}


static int _foreach_lister( SGS_CTX, int vnk )
{
	if( !sgs_LoadArgs( C, ">." ) )
		return 0;
	
	if( sgs_PushIterator( C, *stk_poff( C, 0 ) ) == SGS_FALSE )
		return sgs_ArgErrorExt( C, 0, 0, "iterable", "" );
	
	sgs_CreateArray( C, NULL, 0 );
	/* arg1, arg1 iter, output */
	while( sgs_IterAdvance( C, *stk_poff( C, 1 ) ) > 0 )
	{
		sgs_IterPushData( C, *stk_poff( C, 1 ), !vnk, vnk );
		sgs_ArrayPush( C, *stk_poff( C, 2 ), 1 );
	}
	return 1;
}
static int sgsstd_get_keys( SGS_CTX )
{
	SGSFN( "get_keys" );
	return _foreach_lister( C, 0 );
}
static int sgsstd_get_values( SGS_CTX )
{
	SGSFN( "get_values" );
	return _foreach_lister( C, 1 );
}

static int sgsstd_get_concat( SGS_CTX )
{
	int i, ssz = stk_size( C );
	SGSFN( "get_concat" );
	if( ssz < 2 )
	{
		return sgs_Msg( C, SGS_WARNING,
			"function expects at least 2 arguments, got %d",
			stk_size( C ) );
	}
	
	sgs_CreateArray( C, NULL, 0 );
	for( i = 0; i < ssz; ++i )
	{
		if( sgs_PushIterator( C, sgs_StackItem( C, i ) ) == SGS_FALSE )
			return sgs_ArgErrorExt( C, i, 0, "iterable", "" );
		while( sgs_IterAdvance( C, *stk_gettop( C ) ) > 0 )
		{
			/* ..., output, arg2 iter */
			sgs_IterPushData( C, *stk_gettop( C ), SGS_FALSE, SGS_TRUE );
			sgs_ArrayPush( C, sgs_StackItem( C, -3 ), 1 );
		}
		fstk_pop1( C );
	}
	return 1;
}

static int sgsstd__get_merged__common( SGS_CTX, sgs_SizeVal ssz )
{
	sgs_SizeVal i;
	sgs_Variable v_dest = *stk_gettop( C );
	for( i = 0; i < ssz; ++i )
	{
		if( sgs_PushIterator( C, sgs_StackItem( C, i ) ) == SGS_FALSE )
			return sgs_ArgErrorExt( C, i, 0, "iterable", "" );
		while( sgs_IterAdvance( C, *stk_gettop( C ) ) > 0 )
		{
			/* ..., output, arg2 iter */
			sgs_IterPushData( C, *stk_gettop( C ), SGS_TRUE, SGS_TRUE );
			sgs_SetIndex( C, v_dest, sgs_StackItem( C, -2 ), *stk_gettop( C ), SGS_FALSE );
			sgs_Pop( C, 2 );
		}
		fstk_pop1( C );
	}
	return 1;
}

static int sgsstd_get_merged( SGS_CTX )
{
	sgs_SizeVal ssz = stk_size( C );
	SGSFN( "get_merged" );
	if( ssz < 2 )
	{
		return sgs_Msg( C, SGS_WARNING,
			"function expects at least 2 arguments, got %d",
			stk_size( C ) );
	}
	
	sgs_CreateDict( C, NULL, 0 );
	return sgsstd__get_merged__common( C, ssz );
}

static int sgsstd_get_merged_map( SGS_CTX )
{
	sgs_SizeVal ssz = stk_size( C );
	SGSFN( "get_merged_map" );
	if( ssz < 2 )
	{
		return sgs_Msg( C, SGS_WARNING,
			"function expects at least 2 arguments, got %d",
			stk_size( C ) );
	}
	
	sgs_CreateMap( C, NULL, 0 );
	return sgsstd__get_merged__common( C, ssz );
}

static int sgsstd_get_iterator( SGS_CTX )
{
	SGSFN( "get_iterator" );
	if( !sgs_LoadArgs( C, "?!v" ) )
		return 0;
	sgs_PushIterator( C, *stk_poff( C, 0 ) );
	return 1;
}

static int sgsstd_iter_advance( SGS_CTX )
{
	int ret;
	SGSFN( "iter_advance" );
	if( !sgs_LoadArgs( C, "?!v" ) )
		return 0;
	ret = sgs_IterAdvance( C, *stk_poff( C, 0 ) );
	if( SGS_FAILED( ret ) )
		STDLIB_WARN( "failed to advance iterator" )
	sgs_PushBool( C, ret != 0 );
	return 1;
}

static int sgsstd_iter_getdata( SGS_CTX )
{
	sgs_Bool pushkey = 0, pushval = 1;
	SGSFN( "iter_getdata" );
	if( !sgs_LoadArgs( C, "?!v|bb", &pushkey, &pushval ) )
		return 0;
	if( pushkey + pushval == 0 )
		STDLIB_WARN( "no data requested from iterator" );
	sgs_IterPushData( C, *stk_poff( C, 0 ), pushkey, pushval );
	return pushkey + pushval;
}


static int sgsstd_tobool( SGS_CTX )
{
	SGSFN( "tobool" );
	return sgs_PushBool( C, sgs_GetBool( C, 0 ) );
}

static int sgsstd_toint( SGS_CTX )
{
	SGSFN( "toint" );
	return sgs_PushInt( C, sgs_GetInt( C, 0 ) );
}

static int sgsstd_toreal( SGS_CTX )
{
	SGSFN( "toreal" );
	return sgs_PushReal( C, sgs_GetReal( C, 0 ) );
}

static int sgsstd_tostring( SGS_CTX )
{
	SGSFN( "tostring" );
	sgs_SetStackSize( C, 1 );
	sgs_ToString( C, 0 );
	return 1;
}

static int sgsstd_toptr( SGS_CTX )
{
	SGSFN( "toptr" );
	return sgs_PushPtr( C, sgs_GetPtr( C, 0 ) );
}


static int sgsstd_parseint( SGS_CTX )
{
	sgs_Int i;
	SGSFN( "parseint" );
	if( sgs_ParseInt( C, 0, &i ) )
		sgs_PushInt( C, i );
	else
		fstk_push_null( C );
	return 1;
}

static int sgsstd_parsereal( SGS_CTX )
{
	sgs_Real r;
	SGSFN( "parsereal" );
	if( sgs_ParseReal( C, 0, &r ) )
		sgs_PushReal( C, r );
	else
		fstk_push_null( C );
	return 1;
}


static int sgsstd_typeof( SGS_CTX )
{
	SGSFN( "typeof" );
	if( !sgsstd_expectnum( C, 1 ) )
		return 0;
	sgs_TypeOf( C, *stk_poff( C, 0 ) );
	return 1;
}

static int sgsstd_typeid( SGS_CTX )
{
	SGSFN( "typeid" );
	if( !sgsstd_expectnum( C, 1 ) )
		return 0;
	sgs_PushInt( C, sgs_ItemType( C, 0 ) );
	return 1;
}

static int sgsstd_typeptr( SGS_CTX )
{
	SGSFN( "typeptr" );
	if( !sgs_LoadArgs( C, ">." ) )
		return 0;
	if( sgs_ItemType( C, 0 ) == SGS_VT_OBJECT )
		sgs_PushPtr( C, sgs_GetObjectIface( C, 0 ) );
	else
		sgs_PushPtr( C, NULL );
	return 1;
}

static int sgsstd_typeptr_by_name( SGS_CTX )
{
	char* typenm;
	SGSFN( "typeptr_by_name" );
	if( !sgs_LoadArgs( C, "s", &typenm ) )
		return 0;
	sgs_PushPtr( C, sgs_FindType( C, typenm ) );
	return 1;
}

static int sgsstd_is_numeric( SGS_CTX )
{
	int res;
	uint32_t ty;
	
	SGSFN( "is_numeric" );
	if( !sgsstd_expectnum( C, 1 ) )
		return 0;
	
	ty = sgs_ItemType( C, 0 );
	if( ty == SGS_VT_NULL || ty == SGS_VT_FUNC ||
		ty == SGS_VT_CFUNC || ty == SGS_VT_OBJECT )
		res = SGS_FALSE;
	else
		res = ty != SGS_VT_STRING || sgs_IsNumericString(
			sgs_GetStringPtr( C, 0 ), sgs_GetStringSize( C, 0 ) );
	
	sgs_PushBool( C, res );
	return 1;
}

static int sgsstd_is_callable( SGS_CTX )
{
	SGSFN( "is_callable" );
	if( !sgsstd_expectnum( C, 1 ) )
		return 0;
	
	sgs_PushBool( C, sgs_IsCallable( C, 0 ) );
	return 1;
}

static int sgsstd_is_array( SGS_CTX )
{
	SGSFN( "is_array" );
	sgs_PushBool( C, sgs_IsObject( C, 0, sgsstd_array_iface ) );
	return 1;
}

static int sgsstd_is_dict( SGS_CTX )
{
	SGSFN( "is_dict" );
	sgs_PushBool( C, sgs_IsObject( C, 0, sgsstd_dict_iface ) );
	return 1;
}

static int sgsstd_is_map( SGS_CTX )
{
	SGSFN( "is_map" );
	sgs_PushBool( C, sgs_IsObject( C, 0, sgsstd_map_iface ) );
	return 1;
}


/* I/O */

static int sgsstd_print( SGS_CTX )
{
	sgs_StkIdx i, ssz;
	SGSBASEFN( "print" );
	ssz = stk_size( C );
	for( i = 0; i < ssz; ++i )
	{
		sgs_SizeVal size;
		char* buf = sgs_ToStringBuf( C, i, &size );
		/* WP: string limit */
		sgs_Write( C, buf, (size_t) size );
	}
	return 0;
}

static int sgsstd_println( SGS_CTX )
{
	SGSFN( "println" );
	sgsstd_print( C );
	sgs_Write( C, "\n", 1 );
	return 0;
}

static int sgsstd_printlns( SGS_CTX )
{
	sgs_StkIdx i, ssz;
	ssz = stk_size( C );
	
	SGSFN( "printlns" );
	
	for( i = 0; i < ssz; ++i )
	{
		sgs_SizeVal size;
		char* buf = sgs_ToStringBuf( C, i, &size );
		/* WP: string limit */
		sgs_Write( C, buf, (size_t) size );
		sgs_Write( C, "\n", 1 );
	}
	return 0;
}

static int sgsstd_errprint( SGS_CTX )
{
	sgs_StkIdx i, ssz;
	SGSBASEFN( "errprint" );
	ssz = stk_size( C );
	for( i = 0; i < ssz; ++i )
	{
		sgs_SizeVal size;
		char* buf = sgs_ToStringBuf( C, i, &size );
		/* WP: string limit */
		sgs_ErrWrite( C, buf, (size_t) size );
	}
	return 0;
}

static int sgsstd_errprintln( SGS_CTX )
{
	SGSFN( "errprintln" );
	sgsstd_errprint( C );
	sgs_ErrWrite( C, "\n", 1 );
	return 0;
}

static int sgsstd_errprintlns( SGS_CTX )
{
	sgs_StkIdx i, ssz;
	ssz = stk_size( C );
	
	SGSFN( "errprintlns" );
	
	for( i = 0; i < ssz; ++i )
	{
		sgs_SizeVal size;
		char* buf = sgs_ToStringBuf( C, i, &size );
		/* WP: string limit */
		sgs_ErrWrite( C, buf, (size_t) size );
		sgs_ErrWrite( C, "\n", 1 );
	}
	return 0;
}

static int sgsstd_printvar( SGS_CTX )
{
	sgs_StkIdx i, ssz;
	ssz = stk_size( C );
	
	SGSFN( "printvar" );
	
	for( i = 0; i < ssz; ++i )
	{
		sgs_DumpVar( C, sgs_StackItem( C, i ), 5 );
		{
			sgs_SizeVal bsz;
			char* buf = sgs_ToStringBuf( C, -1, &bsz );
			/* WP: string limit */
			sgs_Write( C, buf, (size_t) bsz );
			sgs_Write( C, "\n", 1 );
		}
		fstk_pop1( C );
	}
	return 0;
}
static int sgsstd_printvar_ext( SGS_CTX )
{
	sgs_Int depth = 5;
	
	SGSFN( "printvar_ext" );
	
	if( !sgs_LoadArgs( C, ">|i.", &depth ) )
		return 0;
	
	sgs_DumpVar( C, *stk_poff( C, 0 ), (int) depth );
	{
		sgs_SizeVal bsz;
		char* buf = sgs_ToStringBuf( C, -1, &bsz );
		/* WP: string limit */
		sgs_Write( C, buf, (size_t) bsz );
		sgs_Write( C, "\n", 1 );
	}
	return 0;
}

static int sgsstd_read_stdin( SGS_CTX )
{
	sgs_MemBuf B;
	char bfr[ 1024 ];
	int all = 0;
	
	SGSFN( "read_stdin" );
	
	if( !sgs_LoadArgs( C, "|b", &all ) )
		return 0;
	
	B = sgs_membuf_create();
	while( fgets( bfr, 1024, stdin ) )
	{
		size_t len = strlen( bfr );
		sgs_membuf_appbuf( &B, C, bfr, len );
		if( len && !all && bfr[ len - 1 ] == '\n' )
		{
			B.size--;
			break;
		}
	}
	if( B.size > 0x7fffffff )
	{
		sgs_membuf_destroy( &B, C );
		STDLIB_WARN( "read more bytes than allowed to store" );
	}
	/* WP: error condition */
	sgs_PushStringBuf( C, B.ptr, (sgs_SizeVal) B.size );
	sgs_membuf_destroy( &B, C );
	return 1;
}


/* OS */

static int sgsstd_ftime( SGS_CTX )
{
	SGSFN( "ftime" );
	sgs_PushReal( C, sgs_GetTime() );
	return 1;
}


/* utils */

static int sgsstd_rand( SGS_CTX )
{
	SGSFN( "rand" );
	sgs_PushInt( C, rand() );
	return 1;
}

static int sgsstd_randf( SGS_CTX )
{
	SGSFN( "randf" );
	sgs_PushReal( C, (double) rand() / (double) RAND_MAX );
	return 1;
}

static int sgsstd_srand( SGS_CTX )
{
	uint32_t s;
	SGSFN( "srand" );
	if( !sgs_LoadArgs( C, "+l", &s ) )
		return 0;
	srand( s );
	return 0;
}

static int sgsstd_hash_fnv( SGS_CTX )
{
	uint8_t* buf;
	sgs_SizeVal i, bufsize;
	uint32_t hv = 2166136261u;
	sgs_Bool as_hexstr = 0;
	SGSFN( "hash_fnv" );
	if( !sgs_LoadArgs( C, "m|b", &buf, &bufsize, &as_hexstr ) )
		return 0;
	for( i = 0; i < bufsize; ++i )
	{
		hv ^= buf[ i ];
		hv *= 16777619u;
	}
	if( as_hexstr )
	{
		char hexstr[ 9 ] = {0};
		sprintf( hexstr, "%08x", hv );
		sgs_PushStringBuf( C, hexstr, 8 );
	}
	else
		sgs_PushInt( C, hv );
	return 1;
}


/* internal utils */

static int sgsstd_va_get_args( SGS_CTX )
{
	uint8_t i, xac, pcnt, inexp;
	sgs_StackFrame* sf;
	SGSFN( "va_get_args" );
	if( !C->sf_last || !C->sf_last->prev )
		STDLIB_WARN( "not called from function" )
	sf = C->sf_last->prev;
	/* WP: argument count limit */
	
	/* accepted arguments */
	inexp = SGS_SF_ARGC_EXPECTED( sf );
	pcnt = SGS_MIN( sf->argcount, inexp );
	stk_mpush( C, &C->stack_base[ sf->argbeg + SGS_SF_ARG_COUNT( sf ) - pcnt ], pcnt );
	/* extra arguments */
	if( sf->argcount > inexp )
	{
		xac = (uint8_t)( sf->argcount - inexp );
		for( i = 0; i < xac; ++i )
		{
			sgs_Variable tmp = C->stack_base[ sf->argbeg + xac - 1 - i ];
			fstk_push( C, &tmp );
		}
	}
	sgs_CreateArray( C, NULL, sf->argcount );
	return 1;
}

static int sgsstd_va_get_arg( SGS_CTX )
{
	sgs_Int argnum;
	uint8_t i, xac, pcnt, inexp;
	sgs_StackFrame* sf;
	SGSFN( "va_get_arg" );
	if( !C->sf_last || !C->sf_last->prev )
		STDLIB_WARN( "not called from function" )
	if( !sgs_LoadArgs( C, "i", &argnum ) )
		return 0;
	sf = C->sf_last->prev;
	if( argnum < 0 || argnum >= sf->argcount )
		STDLIB_WARN( "argument ID out of bounds" )
	
	/* WP: argument count limit */
	i = (uint8_t) argnum;
	
	/* accepted arguments */
	inexp = SGS_SF_ARGC_EXPECTED( sf );
	pcnt = SGS_MIN( sf->argcount, inexp );
	if( i < pcnt )
	{
		sgs_Variable tmp = C->stack_base[ sf->argbeg + SGS_SF_ARG_COUNT( sf ) - pcnt + i ];
		fstk_push( C, &tmp );
	}
	else
	/* extra arguments */
	if( sf->argcount > inexp )
	{
		sgs_Variable tpv;
		i = (uint8_t)( i - pcnt );
		xac = (uint8_t)( sf->argcount - inexp );
		tpv = C->stack_base[ sf->argbeg + xac - 1 - i ];
		fstk_push( C, &tpv );
	}
	else
		fstk_push_null( C );
	
	return 1;
}

static int sgsstd_va_arg_count( SGS_CTX )
{
	SGSFN( "va_arg_count" );
	if( !C->sf_last || !C->sf_last->prev )
		STDLIB_WARN( "not called from function" )
	sgs_PushInt( C, C->sf_last->prev->argcount );
	return 1;
}


static int sgsstd_metaobj_set( SGS_CTX )
{
	sgs_VarObj *obj, *metaobj;
	SGSFN( "metaobj_set" );
	if( !sgs_LoadArgs( C, "!xx", sgs_ArgCheck_Object, &obj, sgs_ArgCheck_Object, &metaobj ) )
		return 0;
	sgs_VarObj* chk = metaobj;
	while( chk )
	{
		if( chk == obj )
			STDLIB_WARN( "loop detected" );
		chk = chk->metaobj;
	}
	sgs_ObjSetMetaObj( C, obj, metaobj );
	sgs_SetStackSize( C, 1 );
	return 1;
}

static int sgsstd_metaobj_get( SGS_CTX )
{
	sgs_VarObj* obj;
	SGSFN( "metaobj_get" );
	if( !sgs_LoadArgs( C, "!x", sgs_ArgCheck_Object, &obj ) )
		return 0;
	obj = sgs_ObjGetMetaObj( obj );
	if( obj )
	{
		sgs_PushObjectPtr( C, obj );
		return 1;
	}
	return 0;
}

static int sgsstd_metamethods_enable( SGS_CTX )
{
	sgs_VarObj* obj;
	sgs_Bool enable;
	SGSFN( "metamethods_enable" );
	if( !sgs_LoadArgs( C, "!xb", sgs_ArgCheck_Object, &obj, &enable ) )
		return 0;
	sgs_ObjSetMetaMethodEnable( obj, enable );
	sgs_SetStackSize( C, 1 );
	return 1;
}

static int sgsstd_metamethods_test( SGS_CTX )
{
	sgs_VarObj* obj;
	SGSFN( "metamethods_test" );
	if( !sgs_LoadArgs( C, "!x", sgs_ArgCheck_Object, &obj ) )
		return 0;
	sgs_PushBool( C, sgs_ObjGetMetaMethodEnable( obj ) );
	return 1;
}

static int sgsstd_mm_getindex_router( SGS_CTX )
{
	sgs_Variable func, movar;
	SGSFN( "mm_getindex_router" );
	
	if( stk_size( C ) != 1 ) goto fail;
	if( !sgs_Method( C ) || sgs_ItemType( C, 0 ) != SGS_VT_OBJECT ) goto fail;
	if( !( movar.data.O = sgs_ObjGetMetaObj( sgs_GetObjectStruct( C, 0 ) ) ) ) goto fail;
	movar.type = SGS_VT_OBJECT;
	
	sgs_PushString( C, "__get_" );
	sgs_PushItem( C, 1 );
	sgs_StringConcat( C, 2 );
	if( sgs_GetIndex( C, movar, *stk_gettop( C ), &func, SGS_FALSE ) == SGS_FALSE ) goto fail;
	
	sgs_SetStackSize( C, 1 ); /* this */
	sgs_InsertVariable( C, 0, func );
	sgs_Release( C, &func );
	sgs_ThisCall( C, 0, 1 );
	return 1;
	
fail:
	return 0;
}

static int sgsstd_mm_setindex_router( SGS_CTX )
{
	sgs_Variable func, movar;
	SGSFN( "mm_setindex_router" );
	
	if( stk_size( C ) != 2 ) goto fail;
	if( !sgs_Method( C ) || sgs_ItemType( C, 0 ) != SGS_VT_OBJECT ) goto fail;
	if( !( movar.data.O = sgs_ObjGetMetaObj( sgs_GetObjectStruct( C, 0 ) ) ) ) goto fail;
	movar.type = SGS_VT_OBJECT;
	
	sgs_PushString( C, "__set_" );
	sgs_PushItem( C, 1 );
	sgs_StringConcat( C, 2 );
	if( sgs_GetIndex( C, movar, *stk_gettop( C ), &func, SGS_FALSE ) == SGS_FALSE ) goto fail;
	
	sgs_SetStackSize( C, 3 ); /* this, arg:key[0], arg:value[1] */
	sgs_PopSkip( C, 1, 1 ); /* this, arg:value[1] */
	sgs_InsertVariable( C, 0, func );
	sgs_Release( C, &func );
	sgs_ThisCall( C, 1, 0 );
	return sgs_GetBool( C, -1 );
	
fail:
	return 0;
}



static int sgsstd_event_getindex( SGS_ARGS_GETINDEXFUNC )
{
	SGS_BEGIN_INDEXFUNC
		SGS_CASE( "signaled" ) return sgs_PushBool( C, obj->data != NULL );
	SGS_END_INDEXFUNC;
}

static int sgsstd_event_setindex( SGS_ARGS_SETINDEXFUNC )
{
	SGS_BEGIN_INDEXFUNC
		SGS_CASE( "signaled" )
		{
			sgs_Bool val = SGS_FALSE;
			if( sgs_ParseBool( C, 1, &val ) == SGS_FALSE )
				return SGS_EINVAL;
			obj->data = val ? obj : NULL;
			return SGS_SUCCESS;
		}
	SGS_END_INDEXFUNC;
}

static int sgsstd_event_convert( SGS_CTX, sgs_VarObj* obj, int type )
{
	if( type == SGS_VT_BOOL )
	{
		return sgs_PushBool( C, obj->data != NULL );
	}
	return SGS_ENOTSUP;
}

static int sgsstd_event_serialize( SGS_CTX, sgs_VarObj* obj )
{
	sgs_Serialize( C, sgs_MakeBool( obj->data != NULL ) );
	sgs_SerializeObject( C, 1, "event" );
	return SGS_SUCCESS;
}

static int sgsstd_event_dump( SGS_CTX, sgs_VarObj* obj, int maxdepth )
{
	char bfr[ 32 ];
	SGS_UNUSED( maxdepth );
	sprintf( bfr, "event(signaled=%s)", obj->data ? "true" : "false" );
	return sgs_PushString( C, bfr );
}

SGS_APIFUNC sgs_ObjInterface sgsstd_event_iface[1] =
{{
	"event",
	NULL, NULL,
	sgsstd_event_getindex, sgsstd_event_setindex,
	sgsstd_event_convert, sgsstd_event_serialize, sgsstd_event_dump, NULL,
	NULL, NULL
}};

SGSONE sgs_CreateEvent( SGS_CTX, sgs_Variable* out )
{
	return sgs_CreateObject( C, out, NULL, sgsstd_event_iface );
}

SGSBOOL sgs_EventState( SGS_CTX, sgs_Variable evt, int state )
{
	SGSBOOL origstate;
	if( !sgs_IsObjectP( &evt, sgsstd_event_iface ) )
		return sgs_Msg( C, SGS_APIERR, "sgs_EventState: specified variable is not of 'event' type" );
	origstate = p_objdata( &evt ) != NULL;
	if( state != SGS_QUERY )
		sgs_SetObjectDataP( &evt, state ? evt.data.O : NULL );
	return origstate;
}

static int sgsstd_event( SGS_CTX )
{
	sgs_Bool val = SGS_FALSE;
	SGSFN( "event" );
	if( !sgs_LoadArgs( C, "|b", &val ) )
		return 0;
	sgs_CreateEvent( C, NULL );
	if( val )
		sgs_EventState( C, *stk_gettop( C ), SGS_TRUE );
	return 1;
}

static void sgs__create_pooled_event( SGS_CTX, sgs_Variable* out, sgs_Variable dict, sgs_Variable key, int val )
{
	sgs_Variable evar;
	if( out ? sgs_GetIndex( C, dict, key, out, SGS_FALSE ) : sgs_PushIndex( C, dict, key, SGS_FALSE ) )
		return;
	sgs_CreateEvent( C, out );
	evar = out ? *out : *stk_gettop( C );
	if( val )
		sgs_EventState( C, evar, SGS_TRUE );
	sgs_SetIndex( C, dict, key, evar, SGS_FALSE );
}

SGSONE sgs_CreatePooledEventBuf( SGS_CTX, sgs_Variable* out, sgs_Variable dict, const char* str, sgs_SizeVal size )
{
	sgs_Variable key;
	sgs_InitStringBuf( C, &key, str, size );
	sgs__create_pooled_event( C, out, dict, key, 0 );
	sgs_Release( C, &key );
	return 1;
}

static int sgsstd_pooled_event( SGS_CTX )
{
	sgs_Bool val = SGS_FALSE;
	SGSFN( "pooled_event" );
	if( !sgs_LoadArgs( C, "?t?s|b", &val ) )
		return 0;
	sgs__create_pooled_event( C, NULL, *stk_poff( C, 0 ), *stk_poff( C, 1 ), val );
	return 1;
}

static void sgs__check_threadendtbl( SGS_CTX )
{
	sgs_Variable endtbl;
	if( C->_E )
		return;
	
	sgsSTD_MakeMap( C, &endtbl, 0 );
	C->_E = endtbl.data.O;
}

void sgs_EndOn( SGS_CTX, sgs_Variable ev, int enable )
{
	/* if we're trying to disable an end event and ..
	.. table doesn't exist - it's already disabled */
	if( enable )
		sgs__check_threadendtbl( C );
	if( C->_E )
	{
		sgs_VHTable* ht = &((DictHdr*)C->_E->data)->ht;
		if( enable )
		{
			sgs_Variable val;
			p_initnull( &val );
			sgs_vht_set( ht, C, &ev, &val );
		}
		else
		{
			sgs_vht_unset( ht, C, &ev );
		}
	}
}

int sgsstd_end_on( SGS_CTX )
{
	sgs_Bool enable = SGS_TRUE;
	sgs_Context* which = C;
	SGSFN( "end_on" );
	if( sgs_Method( C ) )
	{
		if( !sgs_LoadArgs( C, "@y", &which ) )
			return 0;
		sgs_HideThis( C );
	}
	if( !sgs_LoadArgs( C, "?*|b", &enable ) )
		return 0;
	
	sgs_EndOn( which, *stk_poff( C, 0 ), enable );
	return 0;
}



static int sgsstd_co_create( SGS_CTX )
{
	sgs_Context* T;
	SGSFN( "co_create" );
	if( !sgs_LoadArgs( C, "?p." ) )
		return 0;
	
	T = sgsCTX_ForkState( C, 0 );
	fstk_push( T, stk_poff( C, 0 ) );
	T->state |= SGS_STATE_COROSTART;
	sgs_BreakIf( T->refcount != 0 );
	return sgs_PushThreadPtr( C, T );
}

int sgsstd_co_resume( SGS_CTX )
{
	sgs_Context* T = NULL;
	sgs_StkIdx ssz;
	int rvc = 0;
	
	SGSFN( "co_resume" );
	sgs_Method( C );
	if( !sgs_LoadArgs( C, "@y", &T ) )
		return 0;
	sgs_ForceHideThis( C );
	
	if( ( T->state & SGS_STATE_COROSTART ) == 0 && T->sf_last == NULL )
	{
		STDLIB_WARN( "coroutine is finished, cannot resume" );
	}
	
	ssz = stk_size( C );
	
	if( C->hook_fn )
		C->hook_fn( C->hook_ctx, C, SGS_HOOK_PAUSE );
	
	if( T->sf_last )
	{
		stk_mpush( T, stk_poff( C, 0 ), ssz );
		if( !sgs_ResumeStateRet( T, ssz, &rvc ) )
		{
			if( C->hook_fn )
				C->hook_fn( C->hook_ctx, C, SGS_HOOK_CONT );
			STDLIB_WARN( "failed to resume coroutine" );
		}
	}
	else if( T->state & SGS_STATE_COROSTART )
	{
		T->state &= ~SGS_STATE_COROSTART;
		stk_mpush( T, stk_poff( C, 0 ), ssz );
		rvc = sgs_XCall( T, ssz );
	}
	/* else - already handled */
	
	if( C->hook_fn )
		C->hook_fn( C->hook_ctx, C, SGS_HOOK_CONT );
	
	stk_mpush( C, stk_ptop( T, -rvc ), rvc );
	
	return rvc;
}

int sgsstd_abort( SGS_CTX )
{
	sgs_SizeVal i, ssz, abc = 0;
	SGSFN( "abort" );
	sgs_Method( C );
	ssz = stk_size( C );
	if( ssz == 0 )
		return sgs_PushBool( C, sgs_Abort( C ) );
	
	for( i = 0; i < ssz; ++i )
	{
		sgs_Context* T = NULL;
		if( !sgs_LoadArgsExt( C, i, "y", &T ) )
			return 0;
		abc += sgs_Abort( T );
	}
	
	sgs_PushInt( C, abc );
	return 1;
}

sgs_Context* sgs_TopmostContext( SGS_CTX )
{
	while( C->parent )
		C = C->parent;
	return C;
}

void sgs_CreateSubthread( sgs_Context* parent_thread, SGS_CTX,
	sgs_Variable* out, sgs_StkIdx args, int gotthis )
{
	sgs_Real waittime = 0;
	sgs_StkIdx num = args + !!gotthis + 1;
	sgs_Context* co_ctx;
	
	/* call the function */
	co_ctx = sgsCTX_ForkState( parent_thread, 0 );
	stk_mpush( co_ctx, stk_ptop( C, -num ), num );
	sgs_FCall( co_ctx, args, 1, gotthis );
	
	waittime = sgs_GetReal( co_ctx, -1 );
	sgs_Pop( co_ctx, 1 );
	
	/* register thread if not done */
	if( co_ctx->sf_last && ( co_ctx->sf_last->flags & SGS_SF_PAUSED ) )
	{
		co_ctx->parent = parent_thread;
		co_ctx->st_next = parent_thread->subthreads;
		co_ctx->st_timeout = waittime;
		parent_thread->subthreads = co_ctx;
		
		co_ctx->refcount++; /* acquire */
	}
	if( out )
		sgs_InitThreadPtr( out, co_ctx );
	else
		sgs_PushThreadPtr( C, co_ctx );
}

static int sgsstd_thread_create( SGS_CTX )
{
	SGSFN( "thread_create" );
	if( !sgs_LoadArgs( C, "?p?v" ) )
		return 0;
	
	sgs_CreateSubthread( sgs_TopmostContext( C ), C, NULL, stk_size( C ) - 2, 1 );
	return 1;
}

static int sgsstd_subthread_create( SGS_CTX )
{
	SGSFN( "subthread_create" );
	if( !sgs_LoadArgs( C, "?p?v" ) )
		return 0;
	
	sgs_CreateSubthread( C, C, NULL, stk_size( C ) - 2, 1 );
	return 1;
}

static int sgs__anyevent( SGS_CTX )
{
	if( C->_E )
	{
		sgs_VHTIdx i;
		sgs_VHTable* ht = &((DictHdr*)C->_E->data)->ht;
		for( i = 0; i < ht->size; ++i )
		{
			sgs_VHTVar* v = &ht->vars[ i ];
			if( sgs_GetBoolP( C, &v->key ) )
				return 1;
		}
	}
	return 0;
}

#define THREAD_RELEASE( T ) if(--(T)->refcount <= 0){ sgsCTX_FreeState( T ); }

int sgs_ProcessSubthreads( SGS_CTX, sgs_Real dt )
{
	int numleft = 0;
	sgs_Context** pst;
	
	C->wait_timer += dt;
	for( pst = &C->subthreads; *pst != NULL; )
	{
		sgs_Context* thctx = *pst;
		
		sgs_ProcessSubthreads( thctx, dt );
		thctx->st_timeout -= dt;
		thctx->tm_accum += dt;
		
		if( sgs__anyevent( thctx ) )
		{
			sgs_Abort( thctx );
		}
		else if( thctx->st_timeout <= 0 )
		{
			sgs_PushReal( thctx, thctx->tm_accum );
			thctx->tm_accum = 0;
			sgs_ResumeStateExp( thctx, 1, 1 );
			thctx->st_timeout = sgs_GetReal( thctx, -1 );
			fstk_pop1( thctx );
		}
		
		if( thctx->sf_last == NULL || ( thctx->sf_first->flags & SGS_SF_ABORTED ) )
		{
			*pst = thctx->st_next;
			thctx->parent = NULL;
			thctx->st_next = NULL;
			THREAD_RELEASE( thctx );
		}
		else
		{
			pst = &(*pst)->st_next;
			numleft++;
		}
	}
	return numleft;
}

void sgsSTD_ThreadsFree( SGS_CTX )
{
	sgs_Context* thctx = C->subthreads, *next;
	C->subthreads = NULL;
	for( ; thctx != NULL; thctx = next )
	{
		next = thctx->st_next;
		thctx->parent = NULL; /* don't want next 'if' to be triggered for 'thctx' */
		THREAD_RELEASE( thctx );
	}
	if( C->parent )
	{
		sgs_Context* PC = C->parent, **pst;
		for( pst = &PC->subthreads; *pst != NULL; pst = &(*pst)->st_next )
		{
			if( *pst == C )
			{
				*pst = C->st_next;
				break;
			}
		}
		C->refcount--;
		if( PC->refcount <= 0 )
			sgsCTX_FreeState( PC );
	}
	if( C->_E )
	{
		sgs_ObjRelease( C, C->_E );
		C->_E = NULL;
	}
}

void sgsSTD_ThreadsGC( SGS_CTX )
{
	if( C->_E )
	{
		sgs_ObjGCMark( C, C->_E );
	}
}

static int sgsstd_process_threads( SGS_CTX )
{
	sgs_Real dt = 0;
	SGSFN( "process_threads" );
	if( !sgs_LoadArgs( C, "|r", &dt ) )
		return 0;
	
	return sgs_PushInt( C, sgs_ProcessSubthreads( C, dt ) );
}

static int sgsstd_yield( SGS_CTX )
{
	SGSFN( "yield" );
	if( sgs_PauseState( C ) == SGS_FALSE )
		STDLIB_WARN( "cannot yield with C functions in stack" );
	return stk_size( C );
}



struct pcall_printinfo
{
	sgs_MsgFunc pfn;
	void* pctx;
	sgs_Variable handler;
	int depth;
};

static void sgsstd_pcall_print( void* data, SGS_CTX, int type, const char* message )
{
	int ret = 0;
	struct pcall_printinfo* P = (struct pcall_printinfo*) data;
	P->depth++;
	
	if( P->depth > 1 )
		ret = type; /* don't handle errors thrown inside handlers */
	else if( P->handler.type != SGS_VT_NULL )
	{
		fstk_push( C, &P->handler );
		sgs_PushInt( C, type );
		sgs_PushString( C, message );
		sgs_Call( C, 2, 1 );
		if( sgs_Cntl( C, SGS_CNTL_GET_ABORT, 0 ) )
			sgs_Abort( C );
		ret = (int) sgs_GetInt( C, -1 );
		fstk_pop1( C );
	}
	
	if( ret > 0 )
		P->pfn( P->pctx, C, ret, message );
	
	P->depth--;
}

static int sgsstd_pcall( SGS_CTX )
{
	struct pcall_printinfo P;
	int b = 0;
	
	SGSFN( "pcall" );
	
	if( !sgs_LoadArgs( C, "?p|p", &b ) )
		return 0;
	
	P.pfn = C->msg_fn;
	P.pctx = C->msg_ctx;
	P.handler.type = SGS_VT_NULL;
	P.depth = 0;
	if( b )
		sgs_GetStackItem( C, 1, &P.handler );
	
	C->msg_fn = sgsstd_pcall_print;
	C->msg_ctx = &P;
	
	sgs_PushItem( C, 0 );
	sgs_Call( C, 0, 0 );
	
	C->msg_fn = P.pfn;
	C->msg_ctx = P.pctx;
	if( b )
		sgs_Release( C, &P.handler );
	
	return 0;
}

static int sgsstd_assert( SGS_CTX )
{
	char* str = NULL;
	
	SGSFN( "assert" );
	
	if( !sgs_LoadArgs( C, "?v|s", &str ) )
		return 0;
	
	SGSFN( NULL );
	if( !sgs_GetBool( C, 0 ) )
		sgs_Msg( C, SGS_ERROR, !str ? "assertion failed" :
			"assertion failed: %s", str );
	return 0;
}

static int sgsstd_sym_register( SGS_CTX )
{
	char* str = NULL;
	sgs_Variable var;
	
	SGSFN( "sym_register" );
	if( !sgs_LoadArgs( C, "sv", &str, &var ) )
		return 0;
	
	sgs_RegSymbol( C, "", str, var );
	return 0;
}

static int sgsstd_sym_get( SGS_CTX )
{
	sgs_Variable var, sym;
	SGSFN( "sym_get" );
	if( !sgs_LoadArgs( C, "v", &var ) )
		return 0;
	
	if( !sgs_GetSymbol( C, var, &sym ) )
		return sgs_Msg( C, SGS_WARNING, "symbol not found" );
	fstk_push_leave( C, &sym );
	return 1;
}

static int sgsstd_eval( SGS_CTX )
{
	int rvc;
	char* str;
	sgs_SizeVal size;
	
	SGSFN( "eval" );
	
	if( !sgs_LoadArgs( C, "m", &str, &size ) )
		return 0;
	
	/* WP: string limit */
	rvc = sgs_EvalBuffer( C, str, (size_t) size );
	return SGS_MAX( rvc, 0 );
}

static int sgsstd_eval_file( SGS_CTX )
{
	int rvc;
	char* str;
	
	SGSFN( "eval_file" );
	
	if( !sgs_LoadArgs( C, "s", &str ) )
		return 0;
	
	rvc = sgs_EvalFile( C, str );
	if( rvc == SGS_ENOTFND )
		STDLIB_WARN( "file not found" )
	return SGS_MAX( rvc, 0 );
}


static void _sgsstd_compile_pfn( void* data, SGS_CTX, int type, const char* msg )
{
	sgs_Variable* pvar = (sgs_Variable*) data;
	
	fstk_push( C, pvar );
	
	sgs_PushString( C, "type" );
	sgs_PushInt( C, type );
	sgs_PushString( C, "msg" );
	sgs_PushString( C, msg );
	sgs_CreateDict( C, NULL, 4 );
	
	sgs_ArrayPush( C, sgs_StackItem( C, -2 ), 1 );
	fstk_pop1( C );
}

static int sgsstd_compile_sgs( SGS_CTX )
{
	int ret;
	char* buf = NULL, *outbuf = NULL;
	sgs_SizeVal size = 0;
	size_t outsize = 0;
	sgs_Variable var;
	
	sgs_MsgFunc oldpfn;
	void* oldpctx;
	
	SGSFN( "compile_sgs" );
	
	if( !sgs_LoadArgs( C, "m", &buf, &size ) )
		return 0;
	
	sgs_CreateArray( C, NULL, 0 );
	sgs_GetStackItem( C, -1, &var );
	fstk_pop1( C );
	
	oldpfn = C->msg_fn;
	oldpctx = C->msg_ctx;
	sgs_SetMsgFunc( C, _sgsstd_compile_pfn, &var );
	SGSFN( NULL );
	
	/* WP: string limit */
	ret = sgs_Compile( C, buf, (size_t) size, &outbuf, &outsize );
	
	SGSFN( "compile_sgs" );
	C->msg_fn = oldpfn;
	C->msg_ctx = oldpctx;
	
	if( ret < 0 )
	{
		fstk_push_null( C );
	}
	else
	{
		if( outsize <= 0x7fffffff )
			sgs_PushStringBuf( C, outbuf, (sgs_SizeVal) outsize );
		else
		{
			fstk_push_null( C );
			sgs_Msg( C, SGS_WARNING, "size of compiled code is bigger than allowed to store" );
		}
		sgs_Dealloc( outbuf );
	}
	fstk_push_leave( C, &var );
	
	return 2;
}


static SGSBOOL sgsstd__inclib( SGS_CTX, const char* name, int override )
{
	char buf[ 16 ];
	sgs_Variable reginc = sgs_Registry( C, SGS_REG_INC );
	
	sprintf( buf, ":%.14s", name );
	if( !override && sgs_PushProperty( C, reginc, buf ) )
	{
		fstk_pop1( C );
		return SGS_TRUE;
	}
	fstk_pop1( C );
	
#if !SGS_NO_STDLIB
	if( strcmp( name, "fmt" ) == 0 ) sgs_LoadLib_Fmt( C );
	else if( strcmp( name, "io" ) == 0 ) sgs_LoadLib_IO( C );
	else if( strcmp( name, "math" ) == 0 ) sgs_LoadLib_Math( C );
	else if( strcmp( name, "os" ) == 0 ) sgs_LoadLib_OS( C );
	else if( strcmp( name, "re" ) == 0 ) sgs_LoadLib_RE( C );
	else if( strcmp( name, "string" ) == 0 ) sgs_LoadLib_String( C );
	else return SGS_FALSE;
#endif
	
	sgs_SetProperty( C, reginc, buf, sgs_MakeBool( SGS_TRUE ) );
	return SGS_TRUE;
}

static int sgsstd__chkinc( SGS_CTX, int argid )
{
	sgs_Variable val;
	if( sgs_GetIndex( C, sgs_Registry( C, SGS_REG_INC ),
		sgs_StackItem( C, argid ), &val, SGS_FALSE ) )
	{
		sgs_Release( C, &val );
		return SGS_TRUE;
	}
	return SGS_FALSE;
}

static void sgsstd__setinc( SGS_CTX, int argid )
{
	sgs_SetIndex( C, sgs_Registry( C, SGS_REG_INC ), sgs_StackItem( C, argid ),
		sgs_MakeBool( SGS_TRUE ), SGS_FALSE );
}

static int sgsstd_include_library( SGS_CTX )
{
	int ret, over = SGS_FALSE;
	char* str;
	
	SGSBASEFN( "include_library" );
	
	if( !sgs_LoadArgs( C, "s|b", &str, &over ) )
		return 0;
	
	ret = sgsstd__inclib( C, str, over );
	
	if( ret == SGS_FALSE )
		STDLIB_WARN( "library not found" )
	sgs_PushBool( C, ret );
	return 1;
}

static int sgsstd_include_file( SGS_CTX )
{
	int ret, over = SGS_FALSE;
	char* str;
	
	SGSBASEFN( "include_file" );
	
	if( !sgs_LoadArgs( C, "s|b", &str, &over ) )
		return 0;
	
	if( !over && sgsstd__chkinc( C, 0 ) )
		return 1;
	
	ret = sgs_ExecFile( C, str );
	if( ret == SGS_ENOTFND )
		return sgs_Msg( C, SGS_WARNING, "file '%s' was not found", str );
	if( ret == SGS_SUCCESS )
		sgsstd__setinc( C, 0 );
	sgs_PushBool( C, ret == SGS_SUCCESS );
	return 1;
}

static int sgsstd_include_shared( SGS_CTX )
{
	char* fnstr;
	int ret, over = SGS_FALSE;
	sgs_CFunc func;
	
	SGSBASEFN( "include_shared" );
	
	if( !sgs_LoadArgs( C, "s|b", &fnstr, &over ) )
		return 0;
	
	if( !over && sgsstd__chkinc( C, 0 ) )
		return 1;
	
	ret = sgsXPC_GetProcAddress( fnstr, SGS_LIB_ENTRY_POINT, (void**) &func );
	if( ret != 0 )
	{
		if( ret == SGS_XPC_NOFILE )
			return sgs_Msg( C, SGS_WARNING, "file '%s' was not found", fnstr );
		else if( ret == SGS_XPC_NOPROC )
			return sgs_Msg( C, SGS_WARNING, "procedure '" SGS_LIB_ENTRY_POINT "' was not found" );
		else if( ret == SGS_XPC_NOTSUP )
			STDLIB_WARN( "feature is not supported on this platform" )
		else STDLIB_WARN( "unknown error occured" )
	}
	
	ret = func( C );
	if( ret != SGS_SUCCESS )
		STDLIB_WARN( "module failed to initialize" )
	
	sgsstd__setinc( C, 0 );
	sgs_PushBool( C, SGS_TRUE );
	return 1;
}

static int _push_curdir( SGS_CTX )
{
	const char* file, *fend;
	sgs_StackFrame* sf;
	
	sf = sgs_GetFramePtr( C, NULL, 1 )->prev;
	if( !sf )
		return 0;
	
	sgs_StackFrameInfo( C, sf, NULL, &file, NULL );
	if( file )
	{
		fend = file + strlen( file );
		while( fend > file && *fend != '/' && *fend != '\\' )
			fend--;
		if( fend == file )
		{
#ifdef _WIN32
			sgs_PushString( C, "." );
#else
			if( *file == '/' )
				sgs_PushString( C, "" );
			else
				sgs_PushString( C, "." );
#endif
		}
		else
		{
			ptrdiff_t len = fend - file;
			if( sizeof(ptrdiff_t) > 4 && len > 0x7fffffff )
				return 0;
			else
				/* WP: error condition */
				sgs_PushStringBuf( C, file, (sgs_SizeVal) len );
		}
		return 1;
	}
	
	return 0;
}

static int _push_procdir( SGS_CTX )
{
	char* mfn = sgsXPC_GetModuleFileName();
	if( mfn )
	{
		char* mfnend = mfn + strlen( mfn );
		while( mfnend > mfn && *mfnend != '/' && *mfnend != '\\' )
			mfnend--;
		if( ((size_t)( mfnend - mfn )) > 0x7fffffff )
		{
			free( mfn );
			return 0;
		}
		/* WP: added error condition */
		sgs_PushStringBuf( C, mfn, (sgs_SizeVal)( mfnend - mfn ) );
		free( mfn );
		return 1;
	}
	else
		return 0;
}

static int _find_includable_file( SGS_CTX, sgs_MemBuf* tmp, char* ps,
	size_t pssize, char* fn, size_t fnsize, char* dn, size_t dnsize, char* pd, size_t pdsize )
{
	SGS_SHCTX_USE;
	if( ( fnsize > 2 && *fn == '.' && ( fn[1] == '/' || fn[1] == '\\' ) ) ||
#ifdef _WIN32
		( fnsize > 2 && fn[1] == ':' ) )
#else
		( fnsize > 1 && *fn == '/' ) )
#endif
	{
		sgs_membuf_setstrbuf( tmp, C, fn, fnsize );
		
		sgs_ScriptFSData fsd = {0};
		fsd.filename = tmp->ptr;
		if( SGS_SUCCEEDED( S->sfs_fn( S->sfs_ctx, C, SGS_SFS_FILE_EXISTS, &fsd ) ) )
			return 1;
	}
	else
	{
		char* pse = ps + pssize;
		char* psc = ps;
		while( ps <= pse )
		{
			if( ps == pse || *ps == ';' )
			{
				sgs_membuf_resize( tmp, C, 0 );
				while( psc < ps )
				{
					if( *psc == '?' )
						sgs_membuf_appbuf( tmp, C, fn, fnsize );
					else if( *psc == '|' )
					{
						if( dn )
							sgs_membuf_appbuf( tmp, C, dn, dnsize );
						else
						{
							psc = ps;
							goto notthispath;
						}
					}
					else if( *psc == '@' )
					{
						if( pd )
							sgs_membuf_appbuf( tmp, C, pd, pdsize );
						else
						{
							psc = ps;
							goto notthispath;
						}
					}
					else
						sgs_membuf_appchr( tmp, C, *psc );
					psc++;
				}
				sgs_membuf_appchr( tmp, C, 0 );
				
				{
					sgs_ScriptFSData fsd = {0};
					fsd.filename = tmp->ptr;
					if( SGS_SUCCEEDED( S->sfs_fn( S->sfs_ctx, C, SGS_SFS_FILE_EXISTS, &fsd ) ) )
						return 1;
				}
notthispath:
				psc++;
			}
			ps++;
		}
	}
	return 0;
}

static int sgsstd_find_include_file( SGS_CTX )
{
	int ret;
	char* fnstr, *dnstr = NULL, *pdstr = NULL, *ps = NULL;
	sgs_SizeVal fnsize, dnsize = 0, pdsize = 0, pssize = 0;
	sgs_MemBuf mb = sgs_membuf_create();
	
	SGSFN( "find_include_file" );
	
	if( !sgs_LoadArgs( C, "m", &fnstr, &fnsize ) )
		return 0;
	
	ret = sgs_PushGlobalByName( C, "SGS_PATH" );
	if( ret != SGS_SUCCESS ||
		( ps = sgs_ToStringBuf( C, -1, &pssize ) ) == NULL )
	{
		ps = SGS_INCLUDE_PATH;
		pssize = (sgs_SizeVal) strlen( ps );
	}
	
	if( _push_curdir( C ) )
	{
		dnstr = p_cstr( stk_gettop( C ) );
		dnsize = p_strsz( stk_gettop( C ) );
	}
	if( _push_procdir( C ) )
	{
		pdstr = p_cstr( stk_gettop( C ) );
		pdsize = p_strsz( stk_gettop( C ) );
	}
	/* WP: string limit */
	ret = _find_includable_file( C, &mb, ps, (size_t) pssize, fnstr, (size_t) fnsize, dnstr, (size_t) dnsize, pdstr, (size_t) pdsize );
	if( ret == 0 || mb.size == 0 )
	{
		sgs_membuf_destroy( &mb, C );
		return 0;
	}
	
	sgs_PushStringBuf( C, mb.ptr, (sgs_SizeVal) mb.size );
	sgs_membuf_destroy( &mb, C );
	return 1;
}

static int sgsstd_include( SGS_CTX )
{
	char* fnstr, *dnstr = NULL, *pdstr = NULL;
	sgs_SizeVal fnsize, dnsize = 0, pdsize = 0;
	int over = 0, ret;
	
	SGSFN( "include" );
	
	if( !sgs_LoadArgs( C, "m|b", &fnstr, &fnsize, &over ) )
		return 0;
	
	ret = sgsstd__inclib( C, fnstr, over );
	if( ret )
		goto success;
	else
	{
		char* ps;
		sgs_SizeVal pssize;
		sgs_CFunc func;
		sgs_MemBuf mb = sgs_membuf_create();
		
		ret = sgs_PushGlobalByName( C, "SGS_PATH" );
		if( ret == SGS_FALSE ||
			( ps = sgs_ToStringBuf( C, -1, &pssize ) ) == NULL )
		{
			ps = SGS_INCLUDE_PATH;
			pssize = (sgs_SizeVal) strlen( ps );
		}
		
		if( _push_curdir( C ) )
		{
			dnstr = p_cstr( stk_gettop( C ) );
			dnsize = p_strsz( stk_gettop( C ) );
		}
		if( _push_procdir( C ) )
		{
			pdstr = p_cstr( stk_gettop( C ) );
			pdsize = p_strsz( stk_gettop( C ) );
		}
		/* WP: string limit */
		ret = _find_includable_file( C, &mb, ps, (size_t) pssize, fnstr, (size_t) fnsize, dnstr, (size_t) dnsize, pdstr, (size_t) pdsize );
		if( ret == 0 || mb.size == 0 )
		{
			sgs_membuf_destroy( &mb, C );
			return sgs_Msg( C, SGS_WARNING, "could not find '%.*s' "
				"with include path '%.*s'", fnsize, fnstr, pssize, ps );
		}
		
		sgs_PushString( C, mb.ptr );
		// copy found path to replace arg0
		p_setvar( stk_poff( C, 0 ), stk_gettop( C ) );
		if( !over && sgsstd__chkinc( C, 0 ) )
		{
			sgs_membuf_destroy( &mb, C );
			goto success;
		}
		
		sgs_PushString( C, " - include" );
		sgs_StringConcat( C, 2 );
		SGSFN( p_cstr( stk_gettop( C ) ) );
		ret = sgs_ExecFile( C, mb.ptr );
		SGSFN( "include" );
		if( ret == SGS_SUCCESS )
		{
			sgs_membuf_destroy( &mb, C );
			goto success;
		}
		else if( ret == SGS_ECOMP || ret == SGS_EINVAL )
		{
			/* compilation / parsing error, already printed */
			sgs_membuf_destroy( &mb, C );
			sgs_PushBool( C, 0 );
			return 1;
		}
		/* SGS_ENOTSUP - binary detected */
		
		ret = sgsXPC_GetProcAddress( mb.ptr, SGS_LIB_ENTRY_POINT, (void**) &func );
		if( SGS_SUCCEEDED( ret ) )
		{
			ret = func( C );
			if( SGS_SUCCEEDED( ret ) )
			{
				sgs_membuf_destroy( &mb, C );
				goto success;
			}
		}
		else
		{
			sgs_membuf_destroy( &mb, C );
			return sgs_Msg( C, SGS_ERROR, "failed to load native module '%.*s'%s", fnsize, fnstr,
				ret == SGS_XPC_NOPROC ? " (library was loaded but '" SGS_LIB_ENTRY_POINT
					"' function was not found)" : "" );
		}
		
		sgs_membuf_destroy( &mb, C );
	}
	
	sgs_Msg( C, SGS_WARNING, "could not load '%.*s'", fnsize, fnstr );
	sgs_PushBool( C, 0 );
	return 1;
	
success:
	sgsstd__setinc( C, 0 );
	sgs_PushBool( C, 1 );
	return 1;
}

static int sgsstd_import_cfunc( SGS_CTX )
{
	char* fnstr, *pnstr;
	int ret;
	sgs_CFunc func;
	
	SGSFN( "import_cfunc" );
	
	if( !sgs_LoadArgs( C, "ss", &fnstr, &pnstr ) )
		return 0;
	
	ret = sgsXPC_GetProcAddress( fnstr, pnstr, (void**) &func );
	if( ret != 0 )
	{
		if( ret == SGS_XPC_NOFILE )
			return sgs_Msg( C, SGS_WARNING, "file '%s' was not found", fnstr );
		else if( ret == SGS_XPC_NOPROC )
			return sgs_Msg( C, SGS_WARNING, "procedure '%s' was not found", pnstr );
		else if( ret == SGS_XPC_NOTSUP )
			STDLIB_WARN( "feature is not supported on this platform" )
		else STDLIB_WARN( "unknown error occured" )
	}
	
	sgs_PushCFunc( C, func );
	return 1;
}

static int sgsstd_sys_curfile( SGS_CTX )
{
	const char* file;
	sgs_StackFrame* sf;
	
	SGSFN( "sys_curfile" );
	
	if( stk_size( C ) )
		STDLIB_WARN( "function expects 0 arguments" )
	
	sf = sgs_GetFramePtr( C, NULL, 1 )->prev;
	if( !sf )
		return 0;
	
	sgs_StackFrameInfo( C, sf, NULL, &file, NULL );
	if( file )
	{
		sgs_PushString( C, file );
		return 1;
	}
	return 0;
}

static int sgsstd_sys_curfiledir( SGS_CTX )
{
	SGSFN( "sys_curfiledir" );
	
	if( stk_size( C ) )
		STDLIB_WARN( "function expects 0 arguments" )
	
	return _push_curdir( C );
}

static int sgsstd_sys_curprocfile( SGS_CTX )
{
	char* path;
	SGSFN( "sys_curprocfile" );
	
	if( stk_size( C ) )
		STDLIB_WARN( "function expects 0 arguments" )
	
	path = sgsXPC_GetModuleFileName();
	sgs_Errno( C, path != NULL );
	if( path )
	{
		sgs_PushString( C, path );
		free( path );
		return 1;
	}
	else
		return 0;
}

static int sgsstd_sys_curprocdir( SGS_CTX )
{
	SGSFN( "sys_curprocdir" );
	
	if( stk_size( C ) )
		STDLIB_WARN( "function expects 0 arguments" )
	
	return _push_procdir( C );
}

static int sgsstd_multiply_path_ext_lists( SGS_CTX )
{
	char *pp, *ss, *prefixes, *osfx, *suffixes = SGS_INCLUDE_FILEPATHS( "" ), *joinstr = "/";
	size_t joinstrlen;
	SGSFN( "multiply_path_ext_lists" );
	
	if( !sgs_LoadArgs( C, "s|ss", &prefixes, &joinstr, &suffixes ) )
		return 0;
	
	joinstrlen = strlen( joinstr );
	
	sgs_CreateArray( C, NULL, 0 );
	osfx = suffixes;
	pp = prefixes;
	for(;;)
	{
		if( *pp == ';' || *pp == '\0' )
		{
			ss = suffixes = osfx;
			for(;;)
			{
				if( *ss == ';' || *ss == '\0' )
				{
					char* tmp;
					/* WP: assume no wrap-arounds on string iteration */
					size_t pplen = (size_t) ( pp - prefixes );
					size_t sslen = (size_t) ( ss - suffixes );
					if( pplen + sslen + joinstrlen > 0x7fffffff )
						STDLIB_WARN( "generated path size is bigger than allowed to store" )
					/* WP: error condition */
					tmp = sgs_PushStringAlloc( C, (sgs_SizeVal)( pplen + sslen + joinstrlen ) );
					memcpy( tmp, prefixes, pplen );
					memcpy( tmp + pplen, joinstr, joinstrlen );
					memcpy( tmp + pplen + joinstrlen, suffixes, sslen );
					sgs_FinalizeStringAlloc( C, -1 );
					sgs_ArrayPush( C, sgs_StackItem( C, -2 ), 1 );
					if( !*ss )
						break;
					suffixes = ++ss;
				}
				else
					ss += !!*ss;
			}
			if( !*pp )
				break;
			prefixes = ++pp;
		}
		else
			pp += !!*pp;
	}
	
	return 1;
}

static int sgsstd_sys_backtrace( SGS_CTX )
{
	sgs_Bool as_errinfo = 0;
	SGSFN( "sys_backtrace" );
	
	if( !sgs_LoadArgs( C, "|b", &as_errinfo ) )
		return 0;
	
	if( as_errinfo )
		sgs_PushErrorInfo( C, SGS_ERRORINFO_STACK, 0, NULL );
	else
	{
		sgs_StkIdx sz = stk_size( C );
		sgs_StackFrame* p = sgs_GetFramePtr( C, NULL, SGS_FALSE );
		while( p != NULL )
		{
			const char* file, *name;
			int ln;
			if( !p->next && !p->iptr )
				break;
			sgs_StackFrameInfo( C, p, &name, &file, &ln );
			
			sgs_PushString( C, "func" );
			sgs_PushString( C, name );
			sgs_PushString( C, "line" );
			if( ln )
				sgs_PushInt( C, ln );
			else
				fstk_push_null( C );
			sgs_PushString( C, "file" );
			sgs_PushString( C, file );
			
			sgs_CreateDict( C, NULL, 6 );
			
			p = p->next;
		}
		sgs_CreateArray( C, NULL, stk_size( C ) - sz );
	}
	return 1;
}

static int sgsstd_sys_msg( SGS_CTX )
{
	char* errmsg;
	sgs_Int errcode;
	
	SGSFN( "sys_msg" );
	
	if( !sgs_LoadArgs( C, "is", &errcode, &errmsg ) )
		return 0;
	
	SGSFN( NULL );
	
	sgs_Msg( C, (int) errcode, "%s", errmsg );
	return 0;
}

static int sgsstd__msgwrapper( SGS_CTX, const char* fn, int code )
{
	char* msg;
	SGSFN( fn );
	if( sgs_LoadArgs( C, "s", &msg ) )
	{
		SGSFN( NULL );
		sgs_Msg( C, code, "%s", msg );
	}
	return 0;
}

static int sgsstd_INFO( SGS_CTX ){ return sgsstd__msgwrapper( C, "INFO", SGS_INFO ); }
static int sgsstd_WARNING( SGS_CTX ){ return sgsstd__msgwrapper( C, "WARNING", SGS_WARNING ); }
static int sgsstd_ERROR( SGS_CTX ){ return sgsstd__msgwrapper( C, "ERROR", SGS_ERROR ); }

static int sgsstd_app_abort( SGS_CTX )
{
	SGSFN( "app_abort" );
	abort();
	return 0;
}
static int sgsstd_app_exit( SGS_CTX )
{
	sgs_Int ret = 0;
	
	SGSFN( "app_exit" );
	if( !sgs_LoadArgs( C, "|i", &ret ) )
		return 0;
	
	exit( (int) ret );
	return 0;
}

static int sgsstd_sys_replevel( SGS_CTX )
{
	int lev = C->minlev;
	
	SGSFN( "sys_replevel" );
	
	if( stk_size( C ) )
	{
		sgs_Int i;
		if( !sgs_LoadArgs( C, "i", &i ) )
			return 0;
		C->minlev = (int) i;
		return 0;
	}
	sgs_PushInt( C, lev );
	return 1;
}

static int sgsstd_sys_stat( SGS_CTX )
{
	sgs_Int type;
	
	SGSFN( "sys_stat" );
	
	if( !sgs_LoadArgs( C, "i", &type ) )
		return 0;
	
	sgs_PushInt( C, sgs_Stat( C, (int) type ) );
	return 1;
}

static int sgsstd_errno( SGS_CTX )
{
	int retstr = 0;
	
	SGSFN( "errno" );
	
	if( !sgs_LoadArgs( C, "|b", &retstr ) )
		return 0;
	
	if( retstr )
		sgs_PushString( C, strerror( C->last_errno ) );
	else
		sgs_PushInt( C, C->last_errno );
	
	return 1;
}

static int sgsstd_errno_string( SGS_CTX )
{
	sgs_Int e;
	
	SGSFN( "errno_string" );
	
	if( !sgs_LoadArgs( C, "i", &e ) )
		return 0;
	
	sgs_PushString( C, strerror( (int) e ) );
	
	return 1;
}


#define AE( x ) #x, (const char*) x
static const char* errno_key_table[] =
{
	AE( E2BIG ), AE( EACCES ), AE( EAGAIN ), AE( EBADF ), AE( EBUSY ),
	AE( ECHILD ), AE( EDEADLK ), AE( EDOM ), AE( EEXIST ), AE( EFAULT ),
	AE( EFBIG ), AE( EILSEQ ), AE( EINTR ), AE( EINVAL ), AE( EIO ),
	AE( EISDIR ), AE( EMFILE ), AE( EMLINK ), AE( ENAMETOOLONG ), AE( ENFILE ),
	AE( ENODEV ), AE( ENOENT ), AE( ENOEXEC ), AE( ENOLCK ), AE( ENOMEM ),
	AE( ENOSPC ), AE( ENOSYS ), AE( ENOTDIR ), AE( ENOTEMPTY ), AE( ENOTTY ),
	AE( ENXIO ), AE( EPERM ), AE( EPIPE ), AE( ERANGE ), AE( EROFS ),
	AE( ESPIPE ), AE( ESRCH ), AE( EXDEV ),
	
	NULL,
};
#undef AE

static int sgsstd_errno_value( SGS_CTX )
{
	const char** ekt = errno_key_table;
	char* str;
	
	SGSFN( "errno_value" );
	
	if( !sgs_LoadArgs( C, "s", &str ) )
		return 0;
	
	while( *ekt )
	{
		if( strcmp( *ekt, str ) == 0 )
		{
			sgs_PushInt( C, (int) (size_t) ekt[1] );
			return 1;
		}
		ekt += 2;
	}
	
	sgs_Msg( C, SGS_ERROR, "this errno value is unsupported" );
	return 0;
}

static int sgsstd_dumpvar( SGS_CTX )
{
	int i, ssz, rc = 0;
	ssz = stk_size( C );
	
	SGSFN( "dumpvar" );
	
	for( i = 0; i < ssz; ++i )
	{
		sgs_DumpVar( C, sgs_StackItem( C, i ), 5 );
		sgs_PushString( C, "\n" );
		rc += 2;
	}
	if( rc )
	{
		sgs_StringConcat( C, rc );
		return 1;
	}
	return 0;
}
static int sgsstd_dumpvar_ext( SGS_CTX )
{
	sgs_Int depth = 5;
	
	SGSFN( "dumpvar_ext" );
	
	if( !sgs_LoadArgs( C, ">|i.", &depth ) )
		return 0;
	
	sgs_DumpVar( C, *stk_poff( C, 0 ), (int) depth );
	return 1;
}

static int sgsstd_gc_collect( SGS_CTX )
{
	ptrdiff_t orvc;
	
	SGSFN( "gc_collect" );
	
	orvc = sgs_Stat( C, SGS_STAT_OBJCOUNT );
	sgs_GCExecute( C );
	sgs_PushInt( C, orvc - sgs_Stat( C, SGS_STAT_OBJCOUNT ) );
	return 1;
}


static int sgsstd_serialize( SGS_CTX )
{
	sgs_Int which = 2;
	
	SGSFN( "serialize" );
	if( !sgs_LoadArgs( C, "?v|i", &which ) )
		return 0;
	
	if( which != 1 && which != 2 && which != 3 )
		return sgs_Msg( C, SGS_ERROR, "bad serialization mode" );
	
	sgs_SerializeExt( C, *stk_poff( C, 0 ), (int) which );
	return 1;
}


static int sgsstd_unserialize( SGS_CTX )
{
	sgs_Int which = 2;
	
	SGSFN( "unserialize" );
	if( !sgs_LoadArgs( C, "?s|i", &which ) )
		return 0;
	
	if( which != 1 && which != 2 && which != 3 )
		return sgs_Msg( C, SGS_ERROR, "bad serialization mode" );
	
	return sgs_UnserializeExt( C, *stk_poff( C, 0 ), (int) which );
}


static int sgsstd_sgson_encode( SGS_CTX )
{
	const char* tab = NULL;
	sgs_SizeVal tablen = 0;
	
	SGSFN( "sgson_encode" );
	if( !sgs_LoadArgs( C, "?v|m", &tab, &tablen ) )
		return 0;
	
	sgs_SerializeInt_V3( C, *stk_poff( C, 0 ), tab, tablen );
	return 1;
}

static int sgsstd_sgson_decode( SGS_CTX )
{
	char* str;
	sgs_SizeVal size;
	
	SGSFN( "sgson_decode" );
	if( !sgs_LoadArgs( C, "m", &str, &size ) )
		return 0;
	
	{
		const char* ret = NULL;
		sgs_MemBuf stack = sgs_membuf_create();
		sgs_membuf_appchr( &stack, C, 0 );
		ret = sgson_parse( C, &stack, str, size );
		sgs_membuf_destroy( &stack, C );
		if( ret )
		{
			return sgs_Msg( C, SGS_WARNING, "failed to parse SGSON (position %d, %.8s...", ret - str, ret );
		}
		return 1;
	}
}


/* register all */
#ifndef STDLIB_FN
#  define STDLIB_FN( x ) { #x, sgsstd_##x }
#endif
static sgs_RegFuncConst regfuncs[] =
{
	/* containers */
	/* STDLIB_FN( array ), -- object */ STDLIB_FN( array_sized ), STDLIB_FN( dict ),
	STDLIB_FN( map ), { "class", sgsstd_class },
	STDLIB_FN( array_filter ), STDLIB_FN( array_process ),
	STDLIB_FN( dict_filter ), STDLIB_FN( map_filter ), STDLIB_FN( map_process ),
	STDLIB_FN( dict_size ), STDLIB_FN( map_size ), STDLIB_FN( isset ), STDLIB_FN( unset ), STDLIB_FN( clone ),
	STDLIB_FN( get_keys ), STDLIB_FN( get_values ), STDLIB_FN( get_concat ),
	STDLIB_FN( get_merged ), STDLIB_FN( get_merged_map ),
	STDLIB_FN( get_iterator ), STDLIB_FN( iter_advance ), STDLIB_FN( iter_getdata ),
	/* types */
	STDLIB_FN( tobool ), STDLIB_FN( toint ), STDLIB_FN( toreal ), STDLIB_FN( tostring ), STDLIB_FN( toptr ),
	STDLIB_FN( parseint ), STDLIB_FN( parsereal ),
	STDLIB_FN( typeof ), STDLIB_FN( typeid ), STDLIB_FN( typeptr ), STDLIB_FN( typeptr_by_name ),
	STDLIB_FN( is_numeric ), STDLIB_FN( is_callable ),
	STDLIB_FN( is_array ), STDLIB_FN( is_dict ), STDLIB_FN( is_map ),
	/* I/O */
	STDLIB_FN( print ), STDLIB_FN( println ), STDLIB_FN( printlns ),
	STDLIB_FN( errprint ), STDLIB_FN( errprintln ), STDLIB_FN( errprintlns ),
	STDLIB_FN( printvar ), STDLIB_FN( printvar_ext ),
	STDLIB_FN( read_stdin ),
	/* OS */
	STDLIB_FN( ftime ),
	/* utils */
	STDLIB_FN( rand ), STDLIB_FN( randf ), STDLIB_FN( srand ), STDLIB_FN( hash_fnv ),
	/* internal utils */
	STDLIB_FN( va_get_args ), STDLIB_FN( va_get_arg ), STDLIB_FN( va_arg_count ),
	{ "sys_apply", sgs_specfn_apply },
	STDLIB_FN( metaobj_set ), STDLIB_FN( metaobj_get ), STDLIB_FN( metamethods_enable ), STDLIB_FN( metamethods_test ),
	STDLIB_FN( mm_getindex_router ), STDLIB_FN( mm_setindex_router ),
	STDLIB_FN( event ), STDLIB_FN( pooled_event ), STDLIB_FN( end_on ),
	STDLIB_FN( co_create ), STDLIB_FN( co_resume ),
	STDLIB_FN( thread_create ), STDLIB_FN( subthread_create ), STDLIB_FN( abort ),
	STDLIB_FN( process_threads ),
	STDLIB_FN( yield ),
	STDLIB_FN( pcall ), STDLIB_FN( assert ),
	STDLIB_FN( sym_register ), STDLIB_FN( sym_get ),
	STDLIB_FN( eval ), STDLIB_FN( eval_file ), STDLIB_FN( compile_sgs ),
	STDLIB_FN( include_library ), STDLIB_FN( include_file ),
	STDLIB_FN( include_shared ), STDLIB_FN( import_cfunc ),
	STDLIB_FN( find_include_file ),
	STDLIB_FN( include ),
	STDLIB_FN( sys_curfile ), STDLIB_FN( sys_curfiledir ), STDLIB_FN( sys_curprocfile ), STDLIB_FN( sys_curprocdir ),
	STDLIB_FN( multiply_path_ext_lists ),
	STDLIB_FN( sys_backtrace ), STDLIB_FN( sys_msg ), STDLIB_FN( INFO ), STDLIB_FN( WARNING ), STDLIB_FN( ERROR ),
	STDLIB_FN( app_abort ), STDLIB_FN( app_exit ),
	STDLIB_FN( sys_replevel ), STDLIB_FN( sys_stat ),
	STDLIB_FN( errno ), STDLIB_FN( errno_string ), STDLIB_FN( errno_value ),
	STDLIB_FN( dumpvar ), STDLIB_FN( dumpvar_ext ),
	STDLIB_FN( gc_collect ),
	STDLIB_FN( serialize ), STDLIB_FN( unserialize ),
	STDLIB_FN( sgson_encode ), STDLIB_FN( sgson_decode ),
};

static const sgs_RegIntConst regiconsts[] =
{
	{ "SGS_INFO", SGS_INFO },
	{ "SGS_WARNING", SGS_WARNING },
	{ "SGS_ERROR", SGS_ERROR },
	{ "SGS_APIERR", SGS_APIERR },
	
	{ "VT_NULL", SGS_VT_NULL },
	{ "VT_BOOL", SGS_VT_BOOL },
	{ "VT_INT", SGS_VT_INT },
	{ "VT_REAL", SGS_VT_REAL },
	{ "VT_STRING", SGS_VT_STRING },
	{ "VT_FUNC", SGS_VT_FUNC },
	{ "VT_CFUNC", SGS_VT_CFUNC },
	{ "VT_OBJECT", SGS_VT_OBJECT },
	{ "VT_PTR", SGS_VT_PTR },
	{ "VT_THREAD", SGS_VT_THREAD },
	
	{ "RAND_MAX", RAND_MAX },
};


void sgsSTD_PostInit( SGS_CTX )
{
	sgs_PushEnv( C );
	sgs_RegSymbol( C, "", "_G", *stk_gettop( C ) );
	fstk_pop1( C );
	
	sgs_RegIntConsts( C, regiconsts, SGS_ARRAY_SIZE( regiconsts ) );
	sgs_RegFuncConsts( C, regfuncs, SGS_ARRAY_SIZE( regfuncs ) );
	
	sgs_PushInterface( C, sgsstd_array_iface_gen );
	C->shared->array_iface = stk_gettop( C )->data.O;
	sgs_ObjAcquire( C, C->shared->array_iface );
	sgs_SetGlobalByName( C, "array", *stk_gettop( C ) );
	sgs_RegSymbol( C, "", "array", *stk_gettop( C ) );
	fstk_pop1( C );
	
	sgs_PushString( C, SGS_INCLUDE_PATH );
	sgs_SetGlobalByName( C, "SGS_PATH", *stk_gettop( C ) );
	fstk_pop1( C );
	
	sgs_RegisterType( C, "array", sgsstd_array_iface );
	sgs_RegisterType( C, "array_iterator", sgsstd_array_iter_iface );
	sgs_RegisterType( C, "dict", sgsstd_dict_iface );
	sgs_RegisterType( C, "dict_iterator", sgsstd_dict_iter_iface );
	sgs_RegisterType( C, "map", sgsstd_map_iface );
	sgs_RegisterType( C, "closure", sgsstd_closure_iface );
}

SGSBOOL sgsSTD_MakeArray( SGS_CTX, sgs_Variable* out, sgs_SizeVal cnt )
{
	sgs_StkIdx i = 0, ssz = stk_size( C );
	sgs_BreakIf( out == NULL ); /* CreateObjectIPA modifies the stack otherwise */
	
	if( ssz < cnt )
	{
		sgs_Msg( C, SGS_APIERR, "sgs_CreateArray: not enough items on stack (need at least %d, got %d)",
			(int) cnt, (int) ssz );
		return SGS_FALSE;
	}
	else
	{
		int reserve = cnt < 0;
		sgs_Variable *p, *pend;
		void* data;
		if( reserve )
			cnt = -cnt;
		data = sgs_Malloc( C, SGSARR_ALLOCSIZE( cnt ) );
		sgsstd_array_header_t* hdr = (sgsstd_array_header_t*) sgs_CreateObjectIPA( C,
			out, sizeof( sgsstd_array_header_t ), sgsstd_array_iface );
		
		hdr->mem = cnt;
		p = hdr->data = (sgs_Variable*) data;
		if( reserve )
		{
			hdr->size = 0;
			pend = p + cnt;
			while( p < pend )
				(p++)->type = SGS_VT_NULL;
		}
		else
		{
			hdr->size = cnt;
			pend = p + cnt;
			while( p < pend )
				sgs_GetStackItem( C, i++ - cnt, p++ );
			sgs_Pop( C, cnt );
		}
		
		sgs_ObjSetMetaObj( C, p_obj( out ), C->shared->array_iface );
		
		return SGS_TRUE;
	}
}

SGSBOOL sgsSTD_MakeDict( SGS_CTX, sgs_Variable* out, sgs_SizeVal cnt )
{
	DictHdr* dh;
	sgs_VHTable* ht;
	sgs_StkIdx i, ssz = stk_size( C );
	
	if( cnt % 2 != 0 )
	{
		sgs_Msg( C, SGS_APIERR, "sgs_CreateDict: specified item count not even (multiple of 2 required, got %d)",
			(int) cnt );
		return SGS_FALSE;
	}
	if( cnt > ssz )
	{
		sgs_Msg( C, SGS_APIERR, "sgs_CreateDict: not enough items on stack (need at least %d, got %d)",
			(int) cnt, (int) ssz );
		return SGS_FALSE;
	}
	
	sgs_BreakIf( out == NULL );
	dh = mkdict( C, out );
	ht = &dh->ht;
	
	for( i = 0; i < cnt; i += 2 )
	{
		sgs_ToString( C, i - cnt );
		sgs_vht_set( ht, C, (C->stack_top+i-cnt), (C->stack_top+i+1-cnt) );
	}
	
	sgs_Pop( C, cnt );
	return SGS_TRUE;
}

SGSBOOL sgsSTD_MakeMap( SGS_CTX, sgs_Variable* out, sgs_SizeVal cnt )
{
	DictHdr* dh;
	sgs_VHTable* ht;
	sgs_StkIdx i, ssz = stk_size( C );
	
	if( cnt % 2 != 0 )
	{
		sgs_Msg( C, SGS_APIERR, "sgs_CreateMap: specified item count not even (multiple of 2 required, got %d)",
			(int) cnt );
		return SGS_FALSE;
	}
	if( cnt > ssz )
	{
		sgs_Msg( C, SGS_APIERR, "sgs_CreateMap: not enough items on stack (need at least %d, got %d)",
			(int) cnt, (int) ssz );
		return SGS_FALSE;
	}
	
	dh = mkmap( C, out );
	ht = &dh->ht;
	
	for( i = 0; i < cnt; i += 2 )
	{
		sgs_vht_set( ht, C, (C->stack_top+i-cnt), (C->stack_top+i+1-cnt) );
	}
	
	sgs_Pop( C, cnt );
	return SGS_TRUE;
}


#define RLBP S->_R
#define SYMP S->_SYM
#define INCP S->_INC

void sgsSTD_RegistryInit( SGS_CTX )
{
	SGS_SHCTX_USE;
	sgs_Variable var, subtbl, subkey;
	sgsSTD_MakeMap( C, &var, 0 );
	/* symbol table */
	{
		sgs_InitString( C, &subkey, "$sym" );
		sgsSTD_MakeMap( C, &subtbl, 0 );
		sgs_SetIndex( C, var, subkey, subtbl, SGS_TRUE );
		sgs_Release( C, &subkey );
		SYMP = subtbl.data.O;
	}
	/* include table */
	{
		sgs_InitString( C, &subkey, "$inc" );
		sgsSTD_MakeMap( C, &subtbl, 0 );
		sgs_SetIndex( C, var, subkey, subtbl, SGS_TRUE );
		sgs_Release( C, &subkey );
		INCP = subtbl.data.O;
	}
	RLBP = var.data.O;
}

void sgsSTD_RegistryFree( SGS_CTX )
{
	SGS_SHCTX_USE;
	
	/* include table */
	if( INCP )
	{
		sgs_ObjRelease( C, INCP );
		INCP = NULL;
	}
	
	/* symbol table */
	if( SYMP )
	{
		sgs_ObjRelease( C, SYMP );
		SYMP = NULL;
	}
	
	/* registry */
	if( RLBP )
	{
		sgs_ObjRelease( C, RLBP );
		RLBP = NULL;
	}
}

void sgsSTD_RegistryGC( SGS_SHCTX )
{
	if( RLBP )
	{
		sgs_ObjGCMark( S->ctx_root, RLBP );
	}
	if( INCP )
	{
		sgs_ObjGCMark( S->ctx_root, INCP );
	}
	if( SYMP )
	{
		sgs_ObjGCMark( S->ctx_root, SYMP );
	}
}

void sgsSTD_RegistryIter( SGS_CTX, int subtype, sgs_VHTVar** outp, sgs_VHTVar** outpend )
{
	SGS_SHCTX_USE;
	sgs_VarObj* obj = NULL;
	switch( subtype )
	{
	case SGS_REG_ROOT: obj = RLBP; break;
	case SGS_REG_SYM: obj = SYMP; break;
	case SGS_REG_INC: obj = INCP; break;
	}
	sgs_BreakIf( !obj );
	{
		HTHDR;
		*outp = ht->vars;
		*outpend = ht->vars + sgs_vht_size( ht );
	}
}


#define GLBP C->_G

void sgsSTD_GlobalInit( SGS_CTX )
{
	sgs_Variable var;
	sgsSTD_MakeMap( C, &var, 0 );
	GLBP = var.data.O;
}

void sgsSTD_GlobalFree( SGS_CTX )
{
	sgs_ObjRelease( C, GLBP );
	GLBP = NULL;
}

SGSBOOL sgsSTD_GlobalGet( SGS_CTX, sgs_Variable* out, sgs_Variable* idx )
{
	const char* name;
	sgs_VHTVar* pair;
	sgs_VarObj* obj = GLBP;
	HTHDR;
	
	/* `out` is expected to point at an initialized variable, which could be same as `idx` */
	
	if( idx->type != SGS_VT_STRING )
		return SGS_FALSE;
	
	name = sgs_var_cstr( idx );
	
	if( idx->data.S->size == 2 && name[0] == '_' )
	{
		if( name[1] == 'G' )
		{
			sgs_Variable tmp;
			sgs_InitObjectPtr( &tmp, obj );
			VAR_RELEASE( out );
			*out = tmp;
			return SGS_TRUE;
		}
		
		if( name[1] == 'R' )
		{
			SGS_SHCTX_USE;
			VAR_RELEASE( out );
			sgs_InitObjectPtr( out, RLBP );
			return SGS_TRUE;
		}
		
		if( name[1] == 'T' )
		{
			sgs_Variable tmp;
			sgs_InitThreadPtr( &tmp, C );
			VAR_RELEASE( out );
			*out = tmp;
			return SGS_TRUE;
		}
		
		if( name[1] == 'F' )
		{
			if( C->sf_last )
			{
				p_setvar( out, C->sf_last->func );
			}
			else
			{
				VAR_RELEASE( out );
			}
			return SGS_TRUE;
		}
	}
	
	if( obj->mm_enable )
	{
		int ret;
		sgs_Variable obv, tmp;
		obv.type = SGS_VT_OBJECT;
		obv.data.O = obj;
		ret = sgs_GetIndex( C, obv, *idx, &tmp, 0 );
		VAR_RELEASE( out );
		*out = tmp;
		return ret;
	}
	
	if( ( pair = sgs_vht_get( ht, idx ) ) != NULL )
	{
		VAR_RELEASE( out );
		*out = pair->val;
		VAR_ACQUIRE( out );
		return SGS_TRUE;
	}
	
	sgs_Msg( C, SGS_WARNING, "variable '%s' was not found", sgs_str_cstr( idx->data.S ) );
	VAR_RELEASE( out );
	return SGS_FALSE;
}

SGSBOOL sgsSTD_GlobalSet( SGS_CTX, sgs_Variable* idx, sgs_Variable* val )
{
	const char* name;
	sgs_VarObj* obj = GLBP;
	HTHDR;
	
	if( idx->type != SGS_VT_STRING )
		return SGS_FALSE;
	
	name = sgs_var_cstr( idx );
	
	if( idx->data.S->size == 2 && name[0] == '_' )
	{
		if( name[1] == 'G' )
		{
			if( val->type != SGS_VT_OBJECT ||
				( val->data.O->iface != sgsstd_dict_iface && val->data.O->iface != sgsstd_map_iface ) )
			{
				sgs_Msg( C, SGS_ERROR, "_G only accepts 'map'/'dict' values" );
				return SGS_FALSE;
			}
			sgs_SetEnv( C, *val );
			return SGS_TRUE;
		}
		
		if( name[1] == 'R' ||
			name[1] == 'T' ||
			name[1] == 'F' )
		{
			sgs_Msg( C, SGS_WARNING, "cannot change %s", name );
			return SGS_FALSE;
		}
	}
	
	if( obj->mm_enable )
	{
		sgs_Variable obv;
		obv.type = SGS_VT_OBJECT;
		obv.data.O = obj;
		return sgs_SetIndex( C, obv, *idx, *val, 0 );
	}
	
	sgs_vht_set( ht, C, idx, val );
	return SGS_TRUE;
}

void sgsSTD_GlobalGC( SGS_CTX )
{
	sgs_VarObj* obj = GLBP;
	if( obj )
	{
		sgs_ObjGCMark( C, obj );
	}
}

void sgsSTD_GlobalIter( SGS_CTX, sgs_VHTVar** outp, sgs_VHTVar** outpend )
{
	sgs_VarObj* obj = GLBP;
	sgs_BreakIf( !obj );
	{
		HTHDR;
		*outp = ht->vars;
		*outpend = ht->vars + sgs_vht_size( ht );
	}
}



void sgs_RegSymbol( SGS_CTX, const char* prefix, const char* name, sgs_Variable var )
{
	sgs_Variable str, symtbl = sgs_Registry( C, SGS_REG_SYM );
	
	if( prefix == NULL )
		prefix = "";
	if( name == NULL )
		name = "";
	sgs_BreakIf( *prefix == '\0' && *name == '\0' );
	
	if( *prefix )
		sgs_PushString( C, prefix );
	if( *name )
		sgs_PushString( C, name );
	if( *prefix && *name )
		sgs_StringConcat( C, 2 );
	
	str = *stk_gettop( C );
	sgs_SetIndex( C, symtbl, str, var, SGS_FALSE );
	sgs_SetIndex( C, symtbl, var, str, SGS_FALSE );
	fstk_pop1( C );
}

SGSBOOL sgs_GetSymbol( SGS_CTX, sgs_Variable key, sgs_Variable* out )
{
	sgs_Variable symtbl = sgs_Registry( C, SGS_REG_SYM );
	return sgs_GetIndex( C, symtbl, key, out, SGS_FALSE );
}

void sgs_RegFuncConstsExt( SGS_CTX, const sgs_RegFuncConst* list, int size, const char* prefix )
{
	while( size-- )
	{
		sgs_Variable v_func;
		if( !list->name )
			break;
		v_func = sgs_MakeCFunc( list->value );
		sgs_SetGlobalByName( C, list->name, v_func );
		/* put into symbol table */
		if( prefix )
			sgs_RegSymbol( C, prefix, list->name, v_func );
		list++;
	}
}

void sgs_RegIntConsts( SGS_CTX, const sgs_RegIntConst* list, int size )
{
	while( size-- )
	{
		if( !list->name )
			break;
		sgs_SetGlobalByName( C, list->name, sgs_MakeInt( list->value ) );
		list++;
	}
}

void sgs_RegRealConsts( SGS_CTX, const sgs_RegRealConst* list, int size )
{
	while( size-- )
	{
		if( !list->name )
			break;
		sgs_SetGlobalByName( C, list->name, sgs_MakeReal( list->value ) );
		list++;
	}
}


void sgs_StoreFuncConsts( SGS_CTX, sgs_Variable var, const sgs_RegFuncConst* list, int size, const char* prefix )
{
	sgs_Variable key;
	while( size-- )
	{
		sgs_Variable v_func;
		if( !list->name )
			break;
		v_func = sgs_MakeCFunc( list->value );
		sgs_InitString( C, &key, list->name );
		sgs_SetIndex( C, var, key, v_func, 1 );
		sgs_Release( C, &key );
		if( prefix )
			sgs_RegSymbol( C, prefix, list->name, v_func );
		list++;
	}
}

void sgs_StoreIntConsts( SGS_CTX, sgs_Variable var, const sgs_RegIntConst* list, int size )
{
	sgs_Variable key;
	while( size-- )
	{
		if( !list->name )
			break;
		sgs_InitString( C, &key, list->name );
		sgs_SetIndex( C, var, key, sgs_MakeInt( list->value ), 1 );
		sgs_Release( C, &key );
		list++;
	}
}

void sgs_StoreRealConsts( SGS_CTX, sgs_Variable var, const sgs_RegRealConst* list, int size )
{
	sgs_Variable key;
	while( size-- )
	{
		if( !list->name )
			break;
		sgs_InitString( C, &key, list->name );
		sgs_SetIndex( C, var, key, sgs_MakeReal( list->value ), 1 );
		sgs_Release( C, &key );
		list++;
	}
}


SGSBOOL sgs_IncludeExt( SGS_CTX, const char* name, const char* searchpath )
{
	int ret = 0, sz, sz0;
	int pathrep = 0;
	
	sz0 = stk_size( C );
	if( searchpath )
	{
		pathrep = sgs_PushGlobalByName( C, "SGS_PATH" ) ? 1 : 2;
		sgs_PushString( C, searchpath );
		sgs_SetGlobalByName( C, "SGS_PATH", *stk_gettop( C ) );
		fstk_pop1( C );
	}
	
	sz = stk_size( C );
	sgs_PushCFunc( C, sgsstd_include );
	sgs_PushString( C, name );
	sgs_Call( C, 1, 1 );
	ret = sgs_GetBool( C, -1 );
	sgs_SetStackSize( C, sz );
	
	if( pathrep == 1 )
		sgs_SetGlobalByName( C, "SGS_PATH", *stk_gettop( C ) );
	else if( pathrep == 2 )
	{
		sgs_PushEnv( C );
		sgs_PushString( C, "SGS_PATH" );
		sgs_Unset( C, sgs_StackItem( C, -2 ), *stk_gettop( C ) );
	}
	
	sgs_SetStackSize( C, sz0 );
	return ret;
}


SGSBOOL sgs_IsArray( sgs_Variable var )
{
	return sgs_IsObjectP( &var, sgsstd_array_iface );
}

sgs_SizeVal sgs_ArraySize( sgs_Variable var )
{
	if( !sgs_IsObjectP( &var, sgsstd_array_iface ) )
		return -1; /* no error */
	return ((sgsstd_array_header_t*)p_objdata( &var ))->size;
}

void sgs_ArrayPush( SGS_CTX, sgs_Variable var, sgs_StkIdx count )
{
	if( !sgs_IsObjectP( &var, sgsstd_array_iface ) )
	{
		sgs_Msg( C, SGS_APIERR, "sgs_ArrayPush: variable is not an array" );
		return;
	}
	if( stk_size( C ) < count )
	{
		sgs_Msg( C, SGS_APIERR, "sgs_ArrayPush: too few items on stack "
			"(need: %d, got: %d)", (int) count, (int) stk_size( C ) );
		return;
	}
	if( count )
	{
		sgsstd_array_header_t* hdr = (sgsstd_array_header_t*) p_objdata( &var );
		sgsstd_array_insert( C, hdr, hdr->size, stk_size( C ) - count );
		sgs_Pop( C, count );
	}
}

void sgs_ArrayPop( SGS_CTX, sgs_Variable var, sgs_StkIdx count, SGSBOOL ret )
{
	if( !sgs_IsObjectP( &var, sgsstd_array_iface ) )
	{
		sgs_Msg( C, SGS_APIERR, "sgs_ArrayPush: variable is not an array" );
		return;
	}
	if( count )
	{
		sgsstd_array_header_t* hdr = (sgsstd_array_header_t*) p_objdata( &var );
		if( hdr->size < count )
		{
			sgs_Msg( C, SGS_APIERR, "sgs_ArrayPush: too few items on stack "
				"(need: %d, got: %d)", (int) count, (int) stk_size( C ) );
			return;
		}
		if( ret )
		{
			stk_mpush( C, &SGSARR_PTR( hdr )[ hdr->size - count ], count );
		}
		sgsstd_array_erase( C, hdr, hdr->size - count, hdr->size - 1 );
	}
}

void sgs_ArrayErase( SGS_CTX, sgs_Variable var, sgs_StkIdx at, sgs_StkIdx count )
{
	if( !sgs_IsObjectP( &var, sgsstd_array_iface ) )
	{
		sgs_Msg( C, SGS_APIERR, "sgs_ArrayErase: variable is not an array" );
		return;
	}
	if( count )
	{
		sgsstd_array_header_t* hdr = (sgsstd_array_header_t*) p_objdata( &var );
		if( at < 0 || at > hdr->size || at + count > hdr->size )
		{
			sgs_Msg( C, SGS_APIERR, "sgs_ArrayErase: invalid range (erasing: %d - %d, size: %d)",
				(int) at, (int) ( at + count - 1 ), (int) hdr->size );
			return;
		}
		sgsstd_array_erase( C, hdr, at, at + count - 1 );
	}
}

sgs_SizeVal sgs_ArrayFind( SGS_CTX, sgs_Variable var, sgs_Variable what )
{
	if( !sgs_IsObjectP( &var, sgsstd_array_iface ) )
	{
		sgs_Msg( C, SGS_APIERR, "sgs_ArrayFind: variable is not an array" );
		return -1;
	}
	//
	{
		sgs_SizeVal off = 0;
		sgsstd_array_header_t* hdr = (sgsstd_array_header_t*) p_objdata( &var );
		while( off < hdr->size )
		{
			sgs_Variable* p = SGSARR_PTR( hdr ) + off;
			if( sgs_EqualTypes( p, &what ) &&
				sgs_Compare( C, p, &what ) == 0 )
			{
				return off;
			}
			off++;
		}
		return -1;
	}
}

sgs_SizeVal sgs_ArrayRemove( SGS_CTX, sgs_Variable var, sgs_Variable what, SGSBOOL all )
{
	if( !sgs_IsObjectP( &var, sgsstd_array_iface ) )
	{
		sgs_Msg( C, SGS_APIERR, "sgs_ArrayRemove: variable is not an array" );
		return 0;
	}
	//
	{
		sgs_SizeVal off = 0, rmvd = 0;
		sgsstd_array_header_t* hdr = (sgsstd_array_header_t*) p_objdata( &var );
		while( off < hdr->size )
		{
			sgs_Variable* p = SGSARR_PTR( hdr ) + off;
			if( sgs_EqualTypes( p, &what ) &&
				sgs_Compare( C, p, &what ) == 0 )
			{
				sgsstd_array_erase( C, hdr, off, off );
				rmvd++;
				if( all == SGS_FALSE )
					break;
			}
			off++;
		}
		return rmvd;
	}
}

SGSBOOL sgs_IsDict( sgs_Variable var )
{
	return sgs_IsObjectP( &var, sgsstd_dict_iface );
}

SGSBOOL sgs_IsMap( sgs_Variable var )
{
	return sgs_IsObjectP( &var, sgsstd_map_iface );
}

SGSBOOL sgs_Unset( SGS_CTX, sgs_Variable var, sgs_Variable key )
{
	if( !sgs_IsObjectP( &var, sgsstd_dict_iface ) &&
		!sgs_IsObjectP( &var, sgsstd_map_iface ) )
	{
		sgs_Msg( C, SGS_APIERR, "sgs_Unset: variable is not dict/map" );
		return SGS_FALSE;
	}
	//
	{
		sgs_VHTable* T = &((DictHdr*)p_objdata( &var ))->ht;
		sgs_SizeVal sz = T->size;
		sgs_vht_unset( T, C, &key );
		return T->size < sz;
	}
}



#include <time.h>
#include <math.h>
#if !defined( _MSC_VER ) && !defined( __GNUC__ )
#  include <assert.h>
#endif
#include <limits.h>
#if __APPLE__
#  include <sys/time.h>
#endif




void sgs_BreakIfFunc( const char* code, const char* file, int line )
{
	fprintf( stderr, "\n== Error detected: \"%s\", file: %s, line %d ==\n", code, file, line );
#if defined( _MSC_VER )
	__debugbreak();
#elif defined( __GNUC__ )
#  if SGS_ARCH_X86
	__asm__( "int $3" );
#  else
	__builtin_trap();
#  endif
#else
	assert( 0 );
#endif
}


int sgs_isoneof( int chr, const char* str )
{
	while( *str )
	{
		if( *str == chr )
			return 1;
		str++;
	}
	return 0;
}

int sgs_hexchar( int c )
{
	return ( (c) >= '0' && (c) <= '9' ) ||
	( (c) >= 'a' && (c) <= 'f' ) || ( (c) >= 'A' && (c) <= 'F' );
}

int sgs_gethex( int c )
{
	return ( (c) >= '0' && (c) <= '9' ) ? ( (c) - '0' ) :
	( ( (c) >= 'a' && (c) <= 'f' ) ? ( (c) - 'a' + 10 ) : ( (c) - 'A' + 10 ) );
}

int sgs_tolower( int c )
{
	return c >= 'A' && c <= 'Z' ? c - 'A' + 'a' : c;
}


void sgs_print_safe( FILE* fp, const char* buf, size_t size )
{
	size_t i;
	for( i = 0; i < size; ++i )
	{
		if( sgs_isgraph( buf[ i ] ) || buf[ i ] == ' ' )
			fputc( buf[ i ], fp );
		else
			fprintf( fp, "\\x%02X", (int) (unsigned char) buf[ i ] );
	}
}


sgs_MemBuf sgs_membuf_create( void )
{
	sgs_MemBuf sb = { NULL, 0, 0 };
	return sb;
}

void sgs_membuf_destroy( sgs_MemBuf* sb, SGS_CTX )
{
	if( sb->ptr )
		sgs_Dealloc( sb->ptr );
	sb->ptr = NULL;
}

sgs_MemBuf sgs_membuf_partial( char* ch, size_t size )
{
	sgs_MemBuf sb;
	sb.ptr = ch;
	sb.size = size;
	sb.mem = size;
	return sb;
}

void sgs_membuf_reserve( sgs_MemBuf* mb, SGS_CTX, size_t size )
{
	if( size <= mb->mem )
		return;

	mb->mem = size;
	mb->ptr = (char*) sgs_Realloc( C, mb->ptr, size );
}

void sgs_membuf_resize( sgs_MemBuf* mb, SGS_CTX, size_t size )
{
	sgs_membuf_reserve( mb, C, size );
	mb->size = size;
}

void sgs_membuf_resize_opt( sgs_MemBuf* mb, SGS_CTX, size_t size )
{
	if( size > mb->mem )
		sgs_membuf_reserve( mb, C, mb->mem * 2 < size ? size : mb->mem * 2 );
	if( size > mb->size )
		mb->size = size;
}

void sgs_membuf_insbuf( sgs_MemBuf* mb, SGS_CTX, size_t pos, const void* buf, size_t size )
{
	sgs_membuf_reserve( mb, C, mb->mem < mb->size + size ? SGS_MAX( mb->mem * 2, mb->size + size ) : 0 );
	memmove( mb->ptr + pos + size, mb->ptr + pos, mb->size - pos );
	memcpy( mb->ptr + pos, buf, size );
	mb->size += size;
}

void sgs_membuf_erase( sgs_MemBuf* mb, size_t from, size_t to )
{
	sgs_BreakIf( from > mb->size );
	sgs_BreakIf( to > mb->size );
	sgs_BreakIf( from > to );
	if( mb->size - to > 0 )
		memmove( mb->ptr + from, mb->ptr + to, mb->size - to );
	mb->size -= to - from;
}

void sgs_membuf_appbuf( sgs_MemBuf* mb, SGS_CTX, const void* buf, size_t size )
{
	sgs_membuf_reserve( mb, C, mb->mem < mb->size + size ? SGS_MAX( mb->mem * 2, mb->size + size ) : 0 );
	memcpy( mb->ptr + mb->size, buf, size );
	mb->size += size;
}


sgs_Hash sgs_HashFunc( const char* str, size_t size )
{
	size_t i, adv = size / 127 + 1;
	sgs_Hash h = 2166136261u;
	for( i = 0; i < size; i += adv )
	{
		h ^= (sgs_Hash) (uint8_t) str[ i ];
		h *= 16777619u;
	}
	return h;
}

sgs_Hash sgs_HashVar( const sgs_Variable* v )
{
	size_t size;
	switch( v->type )
	{
	/* special */
	case SGS_VT_NULL: return 0;
	case SGS_VT_BOOL: return ( v->data.B != 0 );
	case SGS_VT_STRING: return v->data.S->hash;
	/* data */
	case SGS_VT_INT: size = sizeof( sgs_Int ); break;
	case SGS_VT_REAL: size = sizeof( sgs_Real ); break;
	case SGS_VT_FUNC:
	case SGS_VT_CFUNC:
	case SGS_VT_OBJECT:
	case SGS_VT_PTR:
	case SGS_VT_THREAD:
		size = sizeof( void* ); break;
	default:
		return 0;
	}
	return sgs_HashFunc( (const char*) &v->data, size );
}


// http://stackoverflow.com/a/5694432 implementation 5
static int sgs_is_prime( size_t x )
{
	size_t i, o = 4;
	for( i = 5; 1; i += o )
	{
		size_t q = x / i;
		if (q < i)
			return 1;
		if (x == q * i)
			return 0;
		o ^= 6;
	}
	return 1;
}
static size_t sgs_next_prime( size_t x )
{
	switch( x )
	{
	case 0:
	case 1:
	case 2: return 2;
	case 3: return 3;
	case 4:
	case 5: return 5;
	}
	size_t k = x / 6;
	size_t i = x - 6 * k;
	size_t o = i < 2 ? 1 : 5;
	x = 6 * k + o;
	for( i = ( 3 + o ) / 2; !sgs_is_prime( x ); x += i )
		i ^= 6;
	return x;
}



static int equal_variables( sgs_Variable* v1, sgs_Variable* v2 )
{
	if( v1->type != v2->type )
		return 0;
	switch( v1->type )
	{
	case SGS_VT_NULL: return 1;
	case SGS_VT_BOOL: return v1->data.B == v2->data.B;
	case SGS_VT_INT: return v1->data.I == v2->data.I;
	case SGS_VT_REAL: return v1->data.R == v2->data.R;
	case SGS_VT_STRING:
#if SGS_STRINGTABLE_MAXLEN >= 0x7fffffff
		return v1->data.S == v2->data.S;
#else
		if( v1->data.S == v2->data.S ) return 1;
		return v1->data.S->size == v2->data.S->size &&
			memcmp( sgs_var_cstr( v1 ), sgs_var_cstr( v2 ), v1->data.S->size ) == 0;
#endif
	case SGS_VT_FUNC: return v1->data.F == v2->data.F;
	case SGS_VT_CFUNC: return v1->data.C == v2->data.C;
	case SGS_VT_OBJECT: return v1->data.O == v2->data.O;
	case SGS_VT_PTR: return v1->data.P == v2->data.P;
	case SGS_VT_THREAD: return v1->data.T == v2->data.T;
	}
	return v1->data.P == v2->data.P;
}


#define SGSCFG_VHT_PROBE_DIST 2
#define SGSCFG_VHT_ENABLE_ROBIN_HOOD_HASHING 1

#define SGS_VHT_PROBE_ADV( i, size ) i += SGSCFG_VHT_PROBE_DIST; if( i >= (size) ) i -= (size);
/****  ^ slightly faster on x86 than:  i = ( i + SGSCFG_VHT_PROBE_DIST ) % size;  ****/
#define SGS_VHT_PAIR_MEM_( T ) ((unsigned)(T)->pair_mem)
#define SGS_VHT_PROBE_LEN_( T, pos_idx, pos_hash ) (((SGS_VHT_PAIR_MEM_(T) + (unsigned)(pos_idx) - (pos_hash) % SGS_VHT_PAIR_MEM_(T)) % SGS_VHT_PAIR_MEM_(T))/SGSCFG_VHT_PROBE_DIST)

void sgs_vht_init( sgs_VHTable* T, SGS_CTX, sgs_VHTIdx initial_pair_mem, sgs_VHTIdx initial_var_mem )
{
	sgs_BreakIf( initial_pair_mem < 1 );
	sgs_BreakIf( initial_var_mem < 1 );
	
	T->pairs = sgs_Alloc_n( sgs_VHTIdx, (size_t) initial_pair_mem );
	T->pair_mem = initial_pair_mem;
	T->vars = sgs_Alloc_n( sgs_VHTVar, (size_t) initial_var_mem );
	T->var_mem = initial_var_mem;
	T->size = 0;
	T->num_rem = 0;
	
	memset( T->pairs, SGS_VHTIDX_EMPTY, sizeof(sgs_VHTIdx) * (size_t) initial_pair_mem );
}

void sgs_vht_free( sgs_VHTable* T, SGS_CTX )
{
	sgs_VHTVar* p = T->vars;
	sgs_VHTVar* pend = p + T->size;
	while( p < pend )
	{
		sgs_Release( C, &p->key );
		sgs_Release( C, &p->val );
		p++;
	}
	
	sgs_Dealloc( T->pairs );
	sgs_Dealloc( T->vars );
}

void sgs_vht_analyze( sgs_VHTable* T, sgs_VHTStats* io )
{
	unsigned numcols = 0, numused = 0, numempty = 0, numrem = 0, distsum = 0, worstdist = 0;
	
	sgs_VHTIdx* p = T->pairs;
	sgs_VHTIdx* pend = p + T->pair_mem;
	
	if( io->print )
	{
		printf( "Hash table %p [size=%d, pair_mem=%d var_mem=%d] ---\n", T,
			T->size, T->pair_mem, T->var_mem );
	}
	
	while( p < pend )
	{
		if( *p == SGS_VHTIDX_EMPTY )
			numempty++;
		else if( *p == SGS_VHTIDX_REMOVED )
			numrem++;
		else
		{
			sgs_VHTVar* v = T->vars + *p;
			unsigned dist = SGS_VHT_PROBE_LEN_( T, p - T->pairs, v->hash );
			
			numused++;
			distsum += dist + 1;
			if( worstdist < dist + 1 )
				worstdist = dist + 1;
			
			if( v->hash % (unsigned) T->pair_mem != ( p - T->pairs ) )
			{
				numcols++;
				if( io->print && io->print_cols )
				{
					printf( "collision: hash=0x%08X mod=%u loc=%u dist=%u\n",
						v->hash, v->hash % (unsigned) T->pair_mem, (unsigned) ( p - T->pairs ), dist );
				}
			}
		}
		++p;
	}
	
	if( io->print )
	{
		if( io->print_ubmp )
		{
			printf( "--- usage bitmap ---\n" );
			p = T->pairs;
			while( p < pend )
			{
				if( *p == SGS_VHTIDX_EMPTY )
					printf( " " );
				else if( *p == SGS_VHTIDX_REMOVED )
					printf( "r" );
				else
				{
					unsigned dist = SGS_VHT_PROBE_LEN_( T, p - T->pairs, T->vars[ *p ].hash );
					if( dist < 10 )
						printf( "%u", dist );
					else
						printf( "#" );
				}
				++p;
			}
			printf( "\n" );
		}
		
		printf( "--- summary ---\n" );
		printf( "# used: %u\n", numused );
		printf( "# empty: %u\n", numempty );
		printf( "# removed: %u\n", numrem );
		printf( "# collisions: %u\n", numcols );
		printf( "> average probe length: %.2f\n", numused != 0 ? (float) distsum / (float) numused : 0.0f );
		printf( "> worst probe length: %u\n", worstdist );
		{
			float fbkts = (float) T->pair_mem, fins = (float) numused, fcols = (float) numcols;
			printf( "%% collisions: %.2f%% (expected=%.2f%%)\n", fcols / fbkts * 100, (fins - fbkts * (1 - powf((fbkts-1)/fbkts, fins))) / fbkts * 100);
		}
		printf( "---\n" );
	}
	
	io->buckets = (unsigned) T->pair_mem;
	io->used = numused;
	io->empty = numempty;
	io->removed = numrem;
	io->collisions = numcols;
	io->worst_probe_length = worstdist;
	io->avg_probe_length = numused != 0 ? (float) distsum / (float) numused : 0.0f;
}

static void sgs_vht_rehash( sgs_VHTable* T, SGS_CTX, sgs_VHTIdx size )
{
	sgs_Hash h;
	sgs_VHTIdx i, si, sp, idx, *np;
	sgs_BreakIf( size < T->size );
	
	if( size < 4 )
		size = 4;
	size = (sgs_VHTIdx) sgs_next_prime( (size_t) size );
	if( size == T->pair_mem )
		return;
	
	np = sgs_Alloc_n( sgs_VHTIdx, (size_t) size );
	memset( np, SGS_VHTIDX_EMPTY, sizeof(sgs_VHTIdx) * (size_t) size );
	
#if 0
	printf( "rehash %d -> %d (size = %d, mem = %d kB)\n", T->pair_mem, size, T->size,
		(size * sizeof(sgs_VHTIdx) + T->var_mem * sizeof(sgs_VHTVar)) / 1024 );
#endif
	
	for( si = 0; si < T->pair_mem; ++si )
	{
		idx = T->pairs[ si ];
		if( idx >= 0 )
		{
			h = T->vars[ idx ].hash;
			sp = i = (sgs_VHTIdx)( h % (sgs_Hash) size );
			do
			{
				sgs_VHTIdx nidx = np[ i ];
				if( nidx == SGS_VHTIDX_EMPTY )
				{
					np[ i ] = idx;
					break;
				}
				SGS_VHT_PROBE_ADV( i, size );
			}
			while( i != sp );
		}
	}
	
	sgs_Dealloc( T->pairs );
	T->pairs = np;
	T->pair_mem = size;
	T->num_rem = 0;
}

static void sgs_vht_reserve( sgs_VHTable* T, SGS_CTX, sgs_VHTIdx size )
{
	sgs_VHTVar* p;
	
	sgs_BreakIf( size < T->size );
	
	if( size == T->var_mem )
		return;
	if( size < 4 )
		size = 4;
	
#if 0
	printf( "reserve %d -> %d (size = %d, mem = %d kB)\n", T->var_mem, size, T->size,
		(T->pair_mem * sizeof(sgs_VHTIdx) + size * sizeof(sgs_VHTVar)) / 1024 );
#endif
	
	/* WP: hash table limit */
	p = sgs_Alloc_n( sgs_VHTVar, (size_t) size );
	memcpy( p, T->vars, sizeof(sgs_VHTVar) * (size_t) T->size );
	sgs_Dealloc( T->vars );
	T->vars = p;
	T->var_mem = size;
}

sgs_VHTIdx sgs_vht_pair_id( sgs_VHTable* T, sgs_Variable* K, sgs_Hash hash )
{
	sgs_VHTIdx i, sp = (sgs_VHTIdx)( hash % (sgs_Hash) T->pair_mem );
	i = sp;
	do
	{
		sgs_VHTIdx idx = T->pairs[ i ];
		if( idx == SGS_VHTIDX_EMPTY )
			break;
		if( idx != SGS_VHTIDX_REMOVED && equal_variables( K, &T->vars[ idx ].key ) )
			return i;
		SGS_VHT_PROBE_ADV( i, T->pair_mem );
	}
	while( i != sp );
	return -1;
}

sgs_VHTVar* sgs_vht_get( sgs_VHTable* T, sgs_Variable* K )
{
	sgs_VHTIdx i = sgs_vht_pair_id( T, K, sgs_HashVar( K ) );
	if( i >= 0 )
		return T->vars + T->pairs[ i ];
	else
		return NULL;
}

sgs_VHTVar* sgs_vht_get_str( sgs_VHTable* T, const char* str, uint32_t size, sgs_Hash hash )
{
	sgs_VHTIdx i, sp = (sgs_VHTIdx)( hash % (sgs_Hash) T->pair_mem );
	i = sp;
	do
	{
		sgs_VHTIdx idx = T->pairs[ i ];
		if( idx == SGS_VHTIDX_EMPTY )
			return NULL;
		else if( idx != SGS_VHTIDX_REMOVED )
		{
			sgs_Variable* var = &T->vars[ idx ].key;
			if( var->type == SGS_VT_STRING )
			{
				sgs_iStr* S = var->data.S;
				if( S->size == size && memcmp( sgs_str_cstr( S ), str, size ) == 0 )
					return T->vars + idx;
			}
		}
		SGS_VHT_PROBE_ADV( i, T->pair_mem );
	}
	while( i != sp );
	return NULL;
}

sgs_VHTVar* sgs_vht_set( sgs_VHTable* T, SGS_CTX, sgs_Variable* K, sgs_Variable* V )
{
	sgs_Hash h = sgs_HashVar( K );
	sgs_VHTIdx sp, i = sgs_vht_pair_id( T, K, h );
	if( i >= 0 )
	{
		sgs_VHTVar* p = T->vars + T->pairs[ i ];
		if( V )
			sgs_Acquire( C, V );
		sgs_Release( C, &p->val );
		if( V )
			p->val = *V;
		else
			p->val.type = SGS_VT_NULL;
		return p;
	}
	else
	{
		unsigned curdist = 0; /* current probe length for item to be inserted */
		sgs_VHTIdx
			osize = T->size, /* original hash table item count */
			ipos = T->size; /* currently inserted item index */
		SGS_UNUSED( osize );
		
		/* prefer to rehash if too many removed (num_rem) items are found */
		if( T->size + T->num_rem + 1.0 >= T->pair_mem * 0.7 )
			sgs_vht_rehash( T, C, (sgs_VHTIdx) SGS_MAX( T->pair_mem * 1.5, T->size + 16 ) );
		if( T->size >= T->var_mem )
			sgs_vht_reserve( T, C, (sgs_VHTIdx) SGS_MAX( T->size * 1.5, T->size + 16 ) );
		
		{
			sgs_VHTVar* p = T->vars + T->size;
			p->key = *K;
			p->hash = h;
			sgs_Acquire( C, K );
			if( V )
			{
				p->val = *V;
				sgs_Acquire( C, V );
			}
			else
				p->val.type = SGS_VT_NULL;
		}
		
		sp = i = (sgs_VHTIdx)( h % (sgs_Hash) T->pair_mem );
		do
		{
			sgs_VHTIdx idx = T->pairs[ i ];
			if( idx == SGS_VHTIDX_EMPTY || idx == SGS_VHTIDX_REMOVED )
			{
				if( idx == SGS_VHTIDX_REMOVED )
					T->num_rem--;
				T->pairs[ i ] = ipos;
				T->size++;
				break;
			}
#if SGSCFG_VHT_ENABLE_ROBIN_HOOD_HASHING
			else
			{
				/* occupied, try Robin Hood Hashing */
				
				/* probe length for existing item */
				unsigned exdist = SGS_VHT_PROBE_LEN_( T, i, T->vars[ idx ].hash );
				
				if( exdist < curdist )
				{
					/* success, swap current / new items */
					T->pairs[ i ] = ipos;
					ipos = idx;
					curdist = exdist;
				}
			}
#endif
			SGS_VHT_PROBE_ADV( i, T->pair_mem );
			curdist++;
		}
		while( i != sp );
		
		sgs_BreakIf( T->size == osize );
		
		return T->vars + T->size - 1;
	}
}

void sgs_vht_unset( sgs_VHTable* T, SGS_CTX, sgs_Variable* K )
{
	sgs_Hash h = sgs_HashVar( K );
	sgs_VHTIdx i = sgs_vht_pair_id( T, K, h );
	if( i >= 0 )
	{
		sgs_VHTIdx idx = T->pairs[ i ];
		sgs_VHTVar* p = T->vars + idx;
		sgs_VHTVar bp = *p;
		
		T->pairs[ i ] = SGS_VHTIDX_REMOVED;
		
		T->num_rem++;
		T->size--;
		if( p < T->vars + T->size )
		{
			sgs_VHTVar* ep = T->vars + T->size;
			i = sgs_vht_pair_id( T, &ep->key, ep->hash );
			sgs_BreakIf( i == -1 );
			*p = *ep;
			T->pairs[ i ] = idx;
		}
		
		sgs_Release( C, &bp.key );
		sgs_Release( C, &bp.val );
	}
	
	if( T->num_rem > T->var_mem * 0.25 + 16 )
	{
		sgs_vht_reserve( T, C, (sgs_VHTIdx) ( T->size * 0.75 + T->var_mem * 0.25 ) );
		sgs_vht_rehash( T, C, (sgs_VHTIdx) ( T->size * 0.5 + T->var_mem * 0.5 ) );
	}
}


double sgs_GetTime()
{
#if __linux
	struct timespec ts;
	clock_gettime( CLOCK_MONOTONIC, &ts );
	return (double) ts.tv_sec + 0.000000001 * (double) ts.tv_nsec;
#elif __APPLE__
	struct timeval tv;
	gettimeofday( &tv, NULL );
	return (double) tv.tv_sec + 0.000001 * (double) tv.tv_usec;
#else
	clock_t clk = clock();
	return (double)( clk ) / (double)( CLOCKS_PER_SEC );
#endif
}


/* string -> number conversion */

typedef const char CCH;

static int strtonum_hex( CCH** at, CCH* end, sgs_Int* outi )
{
	sgs_Int val = 0;
	CCH* str = *at + 2;
	while( str < end && sgs_hexchar( *str ) )
	{
		val *= 16;
		val += sgs_gethex( *str );
		str++;
	}
	*at = str;
	*outi = val;
	return 1;
}

static int strtonum_oct( CCH** at, CCH* end, sgs_Int* outi )
{
	sgs_Int val = 0;
	CCH* str = *at + 2;
	while( str < end && sgs_octchar( *str ) )
	{
		val *= 8;
		val += sgs_getoct( *str );
		str++;
	}
	*at = str;
	*outi = val;
	return 1;
}

static int strtonum_bin( CCH** at, CCH* end, sgs_Int* outi )
{
	sgs_Int val = 0;
	CCH* str = *at + 2;
	while( str < end && sgs_binchar( *str ) )
	{
		val *= 2;
		val += sgs_getbin( *str );
		str++;
	}
	*at = str;
	*outi = val;
	return 1;
}

static int strtonum_real( CCH** at, CCH* end, sgs_Real* outf )
{
	sgs_Real val = 0;
	sgs_Real vsign = 1;
	CCH* str = *at, *teststr;
	
	if( *str == '+' ) str++;
	else if( *str == '-' ){ vsign = -1; str++; }
	
	teststr = str;
	while( str < end && sgs_decchar( *str ) )
	{
		val *= 10;
		val += sgs_getdec( *str );
		str++;
	}
	if( str == teststr )
		return 0;
	if( str >= end )
		goto done;
	if( *str == '.' )
	{
		sgs_Real mult = 1.0;
		str++;
		while( str < end && sgs_decchar( *str ) )
		{
			mult /= 10;
			val += sgs_getdec( *str ) * mult;
			str++;
		}
	}
	if( str < end && ( *str == 'e' || *str == 'E' ) )
	{
		sgs_Real sign, e = 0;
		str++;
		if( str >= end || ( *str != '+' && *str != '-' ) )
			goto done;
		sign = *str++ == '-' ? -1 : 1;
		while( str < end && sgs_decchar( *str ) )
		{
			e *= 10;
			e += sgs_getdec( *str );
			str++;
		}
		val *= pow( 10, e * sign );
	}
	
done:
	*outf = val * vsign;
	*at = str;
	return 2;
}

static int strtonum_dec( CCH** at, CCH* end, sgs_Int* outi, sgs_Real* outf )
{
	CCH* str = *at, *teststr;
	if( *str == '+' || *str == '-' ) str++;
	teststr = str;
	while( str < end && sgs_decchar( *str ) )
		str++;
	if( str == teststr )
		return 0;
	if( str < end && ( *str == '.' || *str == 'E' || *str == 'e' ) )
		return strtonum_real( at, end, outf );
	else
	{
		sgs_Int val = 0;
		int invsign = 0;
		
		str = *at;
		if( *str == '+' ) str++;
		else if( *str == '-' ){ invsign = 1; str++; }
		
		while( str < end && sgs_decchar( *str ) )
		{
			val *= 10;
			val += sgs_getdec( *str );
			str++;
		}
		if( invsign ) val = -val;
		*outi = val;
		*at = str;
		return 1;
	}
}

int sgs_util_strtonum( CCH** at, CCH* end, sgs_Int* outi, sgs_Real* outf )
{
	CCH* str = *at;
	if( str >= end )
		return 0;
	if( end - str >= 3 && *str == '0' )
	{
		if( str[1] == 'x' ) return strtonum_hex( at, end, outi );
		else if( str[1] == 'o' ) return strtonum_oct( at, end, outi );
		else if( str[1] == 'b' ) return strtonum_bin( at, end, outi );
	}
	return strtonum_dec( at, end, outi, outf );
}


sgs_Int sgs_util_atoi( const char* str, size_t len )
{
	sgs_Int vi = 0;
	sgs_Real vr = 0;
	const char* p = str;
	int ret = sgs_util_strtonum( &p, str + len, &vi, &vr );
	if( p == str ) return 0;
	if( ret == 1 ) return vi;
	else if( ret == 2 ) return (sgs_Int) vr;
	else return 0;
}

sgs_Real sgs_util_atof( const char* str, size_t len )
{
	sgs_Int vi = 0;
	sgs_Real vr = 0;
	const char* p = str;
	int ret = sgs_util_strtonum( &p, str + len, &vi, &vr );
	if( p == str ) return 0;
	if( ret == 1 ) return (sgs_Real) vi;
	else if( ret == 2 ) return vr;
	else return 0;
}



/**** BEGIN CUSTOM QSORT CODE ****/

/*******************************************************************************
*
*  Author:  Remi Dufour - remi.dufour@gmail.com
*  ! code is modified !, for original, refer to:
*    http://www.codeproject.com/Articles/426706/A-simple-portable-yet-efficient-Quicksort-implemen
*  Date:    July 23rd, 2012
*
*  Name:        Quicksort
*
*  Description: This is a well-known sorting algorithm developed by C. A. R. 
*               Hoare. It is a comparison sort and in this implementation,
*               is not a stable sort.
*
*  Note:        This is public-domain C implementation written from
*               scratch.  Use it at your own risk.
*
*******************************************************************************/

/* Insertion sort threshold shift
 *
 * This macro defines the threshold shift (power of 2) at which the insertion
 * sort algorithm replaces the Quicksort.  A zero threshold shift disables the
 * insertion sort completely.
 *
 * The value is optimized for Linux and MacOS on the Intel x86 platform.
 */
#ifndef INSERTION_SORT_THRESHOLD_SHIFT
# if defined( __APPLE__ ) && defined( __MACH__ )
#  define INSERTION_SORT_THRESHOLD_SHIFT 0
# else
#  define INSERTION_SORT_THRESHOLD_SHIFT 2
# endif
#endif

/* Macro SWAP
 *
 * Swaps the elements of two arrays.
 *
 * The length of the swap is determined by the value of "SIZE".  While both
 * arrays can't overlap, the case in which both pointers are the same works.
 */
#define SWAP(A,B,SIZE)                               \
	{                                                \
		register char       *a_byte = A;             \
		register char       *b_byte = B;             \
		register const char *a_end  = a_byte + SIZE; \
		while (a_byte < a_end)                       \
		{                                            \
			register const char swap_byte = *b_byte; \
			*b_byte++ = *a_byte;                     \
			*a_byte++ = swap_byte;                   \
		}                                            \
	}

/* Macro SWAP_NEXT
 *
 * Swaps the elements of an array with its next value.
 *
 * The length of the swap is determined by the value of "size".  This macro
 * must be used at the beginning of a scope and "A" shouldn't be an expression.
 */
#define SWAP_NEXT(A,SIZE)                                 \
	register char       *a_byte = A;                      \
	register const char *a_end  = A + SIZE;               \
	while (a_byte < a_end)                                \
	{                                                     \
		register const char swap_byte = *(a_byte + SIZE); \
		*(a_byte + SIZE) = *a_byte;                       \
		*a_byte++ = swap_byte;                            \
	}

void sgs_quicksort( void *array, size_t length, size_t size,
	int(*compare)(const void *, const void *, void*), void* userdata)
{
	struct stackframe
	{
		void *left;
		void *right;
	} stack[CHAR_BIT * sizeof(void *)];

	/* Recursion level */
	struct stackframe *recursion = stack;

#if INSERTION_SORT_THRESHOLD_SHIFT != 0
	/* Insertion sort threshold */
	const ptrdiff_t threshold = (ptrdiff_t) size << INSERTION_SORT_THRESHOLD_SHIFT;
#endif
	
	if( length <= 1 )
		return;

	/* Assign the first recursion level of the sorting */
	recursion->left = array;
	recursion->right = (char *)array + size * (length - 1);

	do
	{
		/* Partition the array */
		register char *idx = (char*) recursion->left;
		register char *right = (char*) recursion->right;
		char          *left  = idx;

		/* Assigning store to the left */
		register char *store = idx;

		/* Pop the stack */
		--recursion;

		/* Determine a pivot (in the middle) and move it to the end */
		/* @modification@ changed the left address to something that works */
		SWAP(left + ((size_t)((right - left) >> 1) / size * size),right,size)

		/* From left to right */
		while (idx < right)
		{
			/* If item is smaller than pivot */
			if (compare(right, idx, userdata) > 0)
			{
				/* Swap item and store */
				SWAP(idx,store,size)

				/* We increment store */
				store += size;
			}

			idx += size;
		}

	    /* Move the pivot to its final place */
		SWAP(right,store,size)

/* Performs a recursion to the left */
#define RECURSE_LEFT                     \
	if (left < store - size)             \
	{                                    \
		(++recursion)->left = left;      \
		recursion->right = store - size; \
	}

/* Performs a recursion to the right */
#define RECURSE_RIGHT                       \
	if (store + size < right)               \
	{                                       \
		(++recursion)->left = store + size; \
		recursion->right = right;           \
	}

/* Insertion sort inner-loop */
#define INSERTION_SORT_LOOP(LEFT)                                 \
	{                                                             \
		register char *trail = idx - size;                        \
		while (trail >= LEFT && compare(trail, trail + size, userdata) > 0) \
		{                                                         \
			SWAP_NEXT(trail,size)                                 \
			trail -= size;                                        \
		}                                                         \
	}

/* Performs insertion sort left of the pivot */
#define INSERTION_SORT_LEFT                                \
	for (idx = left + size; idx < store; idx +=size)       \
		INSERTION_SORT_LOOP(left)

/* Performs insertion sort right of the pivot */
#define INSERTION_SORT_RIGHT                                        \
	for (idx = store + (size << 1); idx <= right; idx +=size)       \
		INSERTION_SORT_LOOP(store + size)

/* Sorts to the left */
#if INSERTION_SORT_THRESHOLD_SHIFT == 0
# define SORT_LEFT RECURSE_LEFT
#else
# define SORT_LEFT                 \
	if (store - left <= threshold) \
	{                              \
		INSERTION_SORT_LEFT        \
	}                              \
	else                           \
	{                              \
		RECURSE_LEFT               \
	}
#endif

/* Sorts to the right */
#if INSERTION_SORT_THRESHOLD_SHIFT == 0
# define SORT_RIGHT RECURSE_RIGHT
#else
# define SORT_RIGHT                 \
	if (right - store <= threshold) \
	{                               \
		INSERTION_SORT_RIGHT        \
	}                               \
	else                            \
	{                               \
		RECURSE_RIGHT               \
	}
#endif

		/* Recurse into the smaller partition first */
		if (store - left < right - store)
		{
		/* Left side is smaller */
			SORT_RIGHT
			SORT_LEFT

			continue;
		}

		/* Right side is smaller */
		SORT_LEFT
		SORT_RIGHT

#undef RECURSE_LEFT
#undef RECURSE_RIGHT
#undef INSERTION_SORT_LOOP
#undef INSERTION_SORT_LEFT
#undef INSERTION_SORT_RIGHT
#undef SORT_LEFT
#undef SORT_RIGHT
	}
	while (recursion >= stack);
}

#undef INSERTION_SORT_THRESHOLD_SHIFT
#undef SWAP
#undef SWAP_NEXT

/**** END CUSTOM QSORT CODE ****/


#define U8NFL( x ) ((x&0xC0)!=0x80)

int sgs_utf8_decode( char* buf, size_t size, uint32_t* outchar )
{
	char c;
	if( size == 0 )
		return 0;
	
	c = *buf;
	if( !( c & 0x80 ) )
	{
		*outchar = (uint32_t) c;
		return 1;
	}
	
	if( ( c & 0xE0 ) == 0xC0 )
	{
		if( size < 2 || U8NFL( buf[1] ) )
			return - (int) SGS_MIN(size,2);
		*outchar = (uint32_t) ( ( ((int)(buf[0]&0x1f)) << 6 ) | ((int)(buf[1]&0x3f)) );
		return 2;
	}
	
	if( ( c & 0xF0 ) == 0xE0 )
	{
		if( size < 3 || U8NFL( buf[1] ) || U8NFL( buf[2] ) )
			return - (int) SGS_MIN(size,3);
		*outchar = (uint32_t) ( ( ((int)(buf[0]&0x0f)) << 12 ) | ( ((int)(buf[1]&0x3f)) << 6 )
			| ((int)(buf[2]&0x3f)) );
		return 3;
	}
	
	if( ( c & 0xF8 ) == 0xF0 )
	{
		if( size < 4 || U8NFL( buf[1] ) || U8NFL( buf[2] ) || U8NFL( buf[3] ) )
			return - (int) SGS_MIN(size,4);
		*outchar = (uint32_t) ( ( ((int)(buf[0]&0x07)) << 18 ) | ( ((int)(buf[1]&0x3f)) << 12 )
				| ( ((int)(buf[2]&0x3f)) << 6 ) | ((int)(buf[3]&0x3f)) );
		return 4;
	}
	
	return -1;
}

int sgs_utf8_encode( uint32_t ch, char* out )
{
	if( ch <= 0x7f )
	{
		*out = (char) ch;
		return 1;
	}
	if( ch <= 0x7ff )
	{
		out[ 0 ] = (char)( 0xc0 | ( ( ch >> 6 ) & 0x1f ) );
		out[ 1 ] = (char)( 0x80 | ( ch & 0x3f ) );
		return 2;
	}
	if( ch <= 0xffff )
	{
		out[ 0 ] = (char)( 0xe0 | ( ( ch >> 12 ) & 0x0f ) );
		out[ 1 ] = (char)( 0x80 | ( ( ch >> 6 ) & 0x3f ) );
		out[ 2 ] = (char)( 0x80 | ( ch & 0x3f ) );
		return 3;
	}
	if( ch <= 0x10ffff )
	{
		out[ 0 ] = (char)( 0xf0 | ( ( ch >> 18 ) & 0x07 ) );
		out[ 1 ] = (char)( 0x80 | ( ( ch >> 12 ) & 0x3f ) );
		out[ 2 ] = (char)( 0x80 | ( ( ch >> 6 ) & 0x3f ) );
		out[ 3 ] = (char)( 0x80 | ( ch & 0x3f ) );
		return 4;
	}

	return 0;
}



#include <math.h>




int sgsT_IsKeyword( sgs_TokenList tok, const char* text )
{
	return *tok == SGS_ST_KEYWORD && tok[ 1 ] == strlen( text ) &&
		memcmp( (const char*) tok + 2, text, tok[ 1 ] ) == 0;
}

int sgsT_IsIdent( sgs_TokenList tok, const char* text )
{
	return *tok == SGS_ST_IDENT && tok[ 1 ] == strlen( text ) &&
		memcmp( (const char*) tok + 2, text, tok[ 1 ] ) == 0;
}


static SGS_INLINE int detectline( const char* code, int32_t at )
{
	return code[ at ] == '\r' || ( code[ at ] == '\n' && ( at == 0 || code[ at - 1 ] != '\r' ) );
}

static void skipcomment( SGS_CTX, sgs_LineNum* line, const char* code, int32_t* at, int32_t length )
{
	int32_t i = *at + 1;
	if( code[ i ] == '/' )
	{
		i++;
		while( i < length && code[ i ] != '\n' && code[ i ] != '\r' )
			i++;
		if( i + 1 < length && code[ i ] == '\r' && code[ i + 1 ] == '\n' )
			i++;
		(*line)++;
		*at = i;
	}
	else
	{
		sgs_LineNum init = *line;
		i++;
		while( i < length )
		{
			if( detectline( code, i ) )
				(*line)++;
			if( code[ i ] == '/' && i > 0 && code[ i - 1 ] == '*' )
				break;
			else
				i++;
		}
		if( i == length )
		{
			sgs_Msg( C, SGS_ERROR, "[line %d] Comment has no end", init );
			*at = i - 1;
		}
		else
			*at = i;
	}
}

static int32_t string_inplace_fix( char* str, int32_t len )
{
	char *ipos = str, *opos = str, *iend = str + len;
	while( ipos < iend )
	{
		if( *ipos == '\\' )
		{
			ipos++;
			/* assumption that there's always a character after '\' */
			if( *ipos >= '0' && *ipos <= '7' )
			{
				int oct = *ipos++ - '0';
				if( ipos < iend && *ipos >= '0' && *ipos <= '7' ){ oct *= 8; oct += *ipos++ - '0'; }
				if( ipos < iend && *ipos >= '0' && *ipos <= '7' ){ oct *= 8; oct += *ipos++ - '0'; }
				ipos--;
				if( oct > 0xffff ) *opos++ = (char)( oct >> 8 );
				if( oct > 0xff ) *opos++ = (char)( oct >> 4 );
				*opos = (char) oct;
			}
			else
			{
				switch( *ipos )
				{
				case 'a': *opos = '\a'; break;
				case 'b': *opos = '\b'; break;
				case 'f': *opos = '\f'; break;
				case 'n': *opos = '\n'; break;
				case 'r': *opos = '\r'; break;
				case 't': *opos = '\t'; break;
				case 'v': *opos = '\v'; break;
				case 'x':
					if( ipos + 2 < iend && sgs_hexchar( ipos[1] ) && sgs_hexchar( ipos[2] ) )
					{
						*opos = (char)( ( sgs_gethex( ipos[1] ) << 4 ) | sgs_gethex( ipos[2] ) );
						ipos += 2;
						break;
					}
				/* ', ", \ too: */
				default: *opos = *ipos; break;
				}
			}
		}
		else
			*opos = *ipos;
		ipos++;
		opos++;
	}
	/* WP: returned string can only be shorter */
	return (int32_t) ( opos - str );
}


#define STRLIT_LEN(lit) (sizeof(lit)-1)
#define STRLIT_BUF(lit) lit, STRLIT_LEN(lit)

static int ident_equal( const char* ptr, int size, const char* what, int wlen )
{
	return size == wlen && memcmp( ptr, what, (size_t) size ) == 0;
}
static void readident( SGS_CTX, sgs_MemBuf* out, const char* code, int32_t* at, int32_t length )
{
	int32_t sz = 0;
	int32_t i = *at;
	int32_t pos_rev = (int32_t) out->size;
	sgs_membuf_appchr( out, C, SGS_ST_IDENT );
	sgs_membuf_appchr( out, C, 0 );
	while( i < length && ( sgs_isalnum( code[ i ] ) || code[ i ] == '_' ) )
	{
		sz++;
		if( sz < 256 )
			sgs_membuf_appchr( out, C, code[ i ] );
		else if( sz == 256 )
		{
			C->state |= SGS_HAS_ERRORS;
			sgs_Msg( C, SGS_ERROR, "[line %d] identifier too long", *at );
		}
		i++;
	}
	if( sz >= 255 ) sz = 255;
	out->ptr[ pos_rev + 1 ] = (char) sz;
	if( ident_equal( out->ptr + pos_rev + 2, sz, STRLIT_BUF("var") ) ||
		ident_equal( out->ptr + pos_rev + 2, sz, STRLIT_BUF("global") ) ||
		ident_equal( out->ptr + pos_rev + 2, sz, STRLIT_BUF("thread") ) ||
		ident_equal( out->ptr + pos_rev + 2, sz, STRLIT_BUF("subthread") ) ||
		ident_equal( out->ptr + pos_rev + 2, sz, STRLIT_BUF("sync") ) ||
		ident_equal( out->ptr + pos_rev + 2, sz, STRLIT_BUF("race") ) ||
		ident_equal( out->ptr + pos_rev + 2, sz, STRLIT_BUF("null") ) ||
		ident_equal( out->ptr + pos_rev + 2, sz, STRLIT_BUF("true") ) ||
		ident_equal( out->ptr + pos_rev + 2, sz, STRLIT_BUF("false") ) ||
		ident_equal( out->ptr + pos_rev + 2, sz, STRLIT_BUF("if") ) ||
		ident_equal( out->ptr + pos_rev + 2, sz, STRLIT_BUF("else") ) ||
		ident_equal( out->ptr + pos_rev + 2, sz, STRLIT_BUF("do") ) ||
		ident_equal( out->ptr + pos_rev + 2, sz, STRLIT_BUF("while") ) ||
		ident_equal( out->ptr + pos_rev + 2, sz, STRLIT_BUF("for") ) ||
		ident_equal( out->ptr + pos_rev + 2, sz, STRLIT_BUF("foreach") ) ||
		ident_equal( out->ptr + pos_rev + 2, sz, STRLIT_BUF("break") ) ||
		ident_equal( out->ptr + pos_rev + 2, sz, STRLIT_BUF("continue") ) ||
		ident_equal( out->ptr + pos_rev + 2, sz, STRLIT_BUF("function") ) ||
		ident_equal( out->ptr + pos_rev + 2, sz, STRLIT_BUF("use") ) ||
		ident_equal( out->ptr + pos_rev + 2, sz, STRLIT_BUF("return") ) ||
		ident_equal( out->ptr + pos_rev + 2, sz, STRLIT_BUF("this") ) ||
		ident_equal( out->ptr + pos_rev + 2, sz, STRLIT_BUF("new") ) ||
		ident_equal( out->ptr + pos_rev + 2, sz, STRLIT_BUF("defer") ) )
	{
		out->ptr[ pos_rev ] = SGS_ST_KEYWORD;
	}
	*at = --i;
}

static void readstring( SGS_CTX, sgs_MemBuf* out, sgs_LineNum* line, const char* code, int32_t* at, int32_t length )
{
	int32_t begln = *line;
	int32_t i = *at + 1;
	char endchr = code[ *at ];
	int escaped = 0;
	
	if( i + 1 < length && code[i] == endchr && code[i+1] == endchr )
	{
		/* string without escape code parsing */
		int32_t beg = i + 2;
		i = beg + 2;
		while( i < length )
		{
			char c = code[ i ];
			if( detectline( code, i ) )
				(*line)++;
			if( c == endchr && code[i-1] == endchr && code[i-2] == endchr )
			{
				int32_t size = i - beg - 2;
				sgs_membuf_appchr( out, C, SGS_ST_STRING );
				sgs_membuf_appbuf( out, C, (char*) &size, 4 );
				sgs_membuf_appbuf( out, C, code + beg, (size_t) size );
				*at = i;
				return;
			}
			i++;
		}
	}
	
	while( i < length )
	{
		char c = code[ i ];
		if( detectline( code, i ) )
			(*line)++;
		if( c == '\\' )
			escaped = !escaped;
		else if( c == endchr && !escaped )
		{
			int32_t size = i - *at - 1, newsize;
			size_t numpos = out->size + 1;
			sgs_membuf_appchr( out, C, SGS_ST_STRING );
			sgs_membuf_appbuf( out, C, (char*) &size, 4 );
			sgs_membuf_appbuf( out, C, code + *at + 1, (size_t) size );
			*at += size + 1;
			newsize = string_inplace_fix( out->ptr + numpos + 4, size );
			memcpy( out->ptr + numpos, &newsize, sizeof(newsize) );
			/* WP: size-newsize always non-negative */
			out->size -= (size_t)( size - newsize );
			return;
		}
		else
			escaped = 0;
		i++;
	}
	
	C->state |= SGS_MUST_STOP;
	sgs_Msg( C, SGS_ERROR, "[line %d] end of string not found", begln );
}

static const char* sgs_opchars = "=<>+-*/%?!~&|^.$@";
static const char* sgs_operators = "<=>;===;!==;==;!=;<=;>=;+=;-=;*=;/=;%=;&=;|=;^=;<<=;>>=;$=;..=;"
	"<<;>>;&&=;||=;?""?=;&&;||;?""?;..;<;>;=;++;--;+;-;*;/;%;&;|;^;.;$;!;~;@"; /* trigraphs detected */
static const sgs_TokenType sgs_optable[] =
{
	SGS_ST_OP_RWCMP, SGS_ST_OP_SEQ, SGS_ST_OP_SNEQ, SGS_ST_OP_EQ, SGS_ST_OP_NEQ, SGS_ST_OP_LEQ, SGS_ST_OP_GEQ,
	SGS_ST_OP_ADDEQ, SGS_ST_OP_SUBEQ, SGS_ST_OP_MULEQ, SGS_ST_OP_DIVEQ, SGS_ST_OP_MODEQ,
	SGS_ST_OP_ANDEQ, SGS_ST_OP_OREQ, SGS_ST_OP_XOREQ, SGS_ST_OP_LSHEQ, SGS_ST_OP_RSHEQ, SGS_ST_OP_CATEQ, SGS_ST_OP_CATEQ,
	SGS_ST_OP_LSH, SGS_ST_OP_RSH, SGS_ST_OP_BLAEQ, SGS_ST_OP_BLOEQ, SGS_ST_OP_NLOEQ, SGS_ST_OP_BLAND,
	SGS_ST_OP_BLOR, SGS_ST_OP_NLOR, SGS_ST_OP_CAT, SGS_ST_OP_LESS, SGS_ST_OP_GRTR, SGS_ST_OP_SET, SGS_ST_OP_INC, SGS_ST_OP_DEC,
	SGS_ST_OP_ADD, SGS_ST_OP_SUB, SGS_ST_OP_MUL, SGS_ST_OP_DIV, SGS_ST_OP_MOD, SGS_ST_OP_AND,
	SGS_ST_OP_OR, SGS_ST_OP_XOR, SGS_ST_OP_MMBR, SGS_ST_OP_CAT, SGS_ST_OP_NOT, SGS_ST_OP_INV, SGS_ST_OP_ERSUP
};
static const char sgs_opsep = ';';

static int op_oneof( const char* str, const char* test, char sep, int* outlen )
{
	const char* pstr = str;
	int passed = 0, equal = 0, which = 0, len = 0;
	
	do
	{
		if( *test == sep || *test == 0 )
		{
			if( passed == equal )
			{
				*outlen = len;
				return which;
			}
			if( *test == 0 )
				return -1;
			passed = 0;
			equal = 0;
			len = 0;
			pstr = str;
			which++;
		}
		else
		{
			len++;
			passed++;
			if( *pstr == *test )
				equal++;
			pstr += !!*pstr;
		}
	}
	while( *test++ );
	
	return -1;
}

static void readop( SGS_CTX, sgs_MemBuf* out, sgs_LineNum line, const char* code, int32_t* at, int32_t length )
{
	char opstr[ 4 ];
	int32_t opsize = 0, ropsize = 0, i = *at, whichop, len = -1;
	
	memset( opstr, 0, 4 );
	
	/* read in the operator */
	while( i < length && sgs_isoneof( code[ i ], sgs_opchars ) )
	{
		if( opsize < 3 )
			opstr[ opsize++ ] = code[ i ];
		ropsize++;
		i++;
	}
	
	/* test for various errors */
	if( ropsize > 3 ){ *at += ropsize; goto op_read_error; }
	whichop = op_oneof( opstr, sgs_operators, sgs_opsep, &len );
	*at += len - 1;
	if( whichop < 0 ) goto op_read_error;
	
	sgs_membuf_appchr( out, C, (char) sgs_optable[ whichop ] );
	return;
	
op_read_error:
	C->state |= SGS_HAS_ERRORS;
	sgs_Msg( C, SGS_ERROR, "[line %d] invalid operator found: \"%s%s\", size=%d", line, opstr, ropsize > 3 ? "..." : "", ropsize );
}

sgs_TokenList sgsT_Gen( SGS_CTX, const char* code, size_t length )
{
	int32_t i, ilen = (int32_t) length; /* WP: code limit */
	sgs_LineNum line = 1;
	sgs_MemBuf s = sgs_membuf_create();
	sgs_membuf_reserve( &s, C, SGS_TOKENLIST_PREALLOC );
	
	for( i = 0; i < ilen; ++i )
	{
		sgs_LineNum tokline = line;
		char fc = code[ i ];
		size_t isz = s.size;
		
		/* whitespace */
		if( detectline( code, i ) )
			line++;
		if( sgs_isoneof( fc, " \n\r\t" ) )
			continue;
		
		/* comment */
		if( fc == '/' && ( code[ i + 1 ] == '/'
					|| code[ i + 1 ] == '*' ) )   skipcomment( C, &line, code, &i, ilen );
		
		/* special symbol */
		else if( sgs_isoneof( fc, "()[]{},;:" ) ) sgs_membuf_appchr( &s, C, fc );
		
		/* identifier */
		else if( fc == '_' || sgs_isalpha( fc ) ) readident( C, &s, code, &i, ilen );
		
		/* number */
		else if( sgs_isdigit( fc ) )
		{
			sgs_Int vi = 0;
			sgs_Real vr = 0;
			const char* pos = code + i;
			int res = sgs_util_strtonum( &pos, code + length, &vi, &vr );
			if( res == 0 )
			{
				C->state |= SGS_HAS_ERRORS;
				sgs_Msg( C, SGS_ERROR, "[line %d] failed to parse numeric constant", line );
			}
			else if( res == 1 )
			{
				sgs_membuf_appchr( &s, C, SGS_ST_NUMINT );
				sgs_membuf_appbuf( &s, C, &vi, sizeof( vi ) );
			}
			else if( res == 2 )
			{
				sgs_membuf_appchr( &s, C, SGS_ST_NUMREAL );
				sgs_membuf_appbuf( &s, C, &vr, sizeof( vr ) );
			}
			else{ sgs_BreakIf( "Invalid return value from util_strtonum." ); }
			/* WP: code limit */
			i = (int32_t) ( pos - code );
			i--;
		}
		
		/* string */
		else if( fc == '\'' || fc == '\"' )	      readstring( C, &s, &line, code, &i, ilen );
		
		/* operator */
		else if( sgs_isoneof( fc, sgs_opchars ) ) readop( C, &s, line, code, &i, ilen );
		
		/* - unexpected symbol */
		else
		{
			C->state |= SGS_HAS_ERRORS;
			sgs_Msg( C, SGS_ERROR, "[line %d], unexpected symbol: %c", line, fc );
		}
		
		if( s.size != isz ) /* write a line only if successfully wrote something (a token) */
			sgs_membuf_appbuf( &s, C, &tokline, sizeof( tokline ) );
		
		if( C->state & SGS_MUST_STOP )
			break;
	}
	
	sgs_membuf_appchr( &s, C, SGS_ST_NULL );
	return (sgs_TokenList) s.ptr;
}

void sgsT_Free( SGS_CTX, sgs_TokenList tlist )
{
	sgs_MemBuf s = sgs_membuf_partial( (char*) tlist, 0 );
	sgs_membuf_destroy( &s, C );
}

sgs_TokenList sgsT_Next( sgs_TokenList tok )
{
	sgs_BreakIf( !tok );
	sgs_BreakIf( !*tok );
	
	switch( *tok )
	{
	case SGS_ST_IDENT:
	case SGS_ST_KEYWORD:
		return tok + tok[ 1 ] + 2 + sizeof( sgs_LineNum );
	case SGS_ST_NUMREAL:
	case SGS_ST_NUMINT:
		return tok + 9 + sizeof( sgs_LineNum );
	case SGS_ST_STRING:
		{
			int32_t len;
			SGS_ST_READINT( len, tok + 1 );
			return tok + 5 + sizeof( sgs_LineNum ) + len;
		}
	default:
		return tok + 1 + sizeof( sgs_LineNum );
	}
}

sgs_LineNum sgsT_LineNum( sgs_TokenList tok )
{
	if( !*tok )
		return -1;
	tok = sgsT_Next( tok );
	{
		sgs_LineNum ln;
		SGS_ST_READLN( ln, tok - 2 );
		return ln;
	}
}


size_t sgsT_ListSize( sgs_TokenList tlist )
{
	size_t i = 0;
	while( *tlist )
	{
		tlist = sgsT_Next( tlist );
		i++;
	}
	return i;
}

size_t sgsT_ListMemSize( sgs_TokenList tlist )
{
	sgs_TokenList last = tlist;
	while( *last )
		last = sgsT_Next( last );
	return (size_t) ( last - tlist + 1 );
}


static void tp_token( SGS_CTX, sgs_MemBuf* out, sgs_TokenList t )
{
	switch( *t )
	{
	case SGS_ST_RBRKL:
	case SGS_ST_RBRKR:
	case SGS_ST_SBRKL:
	case SGS_ST_SBRKR:
	case SGS_ST_CBRKL:
	case SGS_ST_CBRKR:
	case SGS_ST_ARGSEP:
	case SGS_ST_STSEP:
	case SGS_ST_PICKSEP:
		sgs_membuf_appchr( out, C, (char) *t );
		break;
	case SGS_ST_IDENT:
	case SGS_ST_KEYWORD:
		sgs_membuf_appbuf( out, C, t + 2, t[1] );
		break;
	case SGS_ST_NUMREAL:
		{
			sgs_Real val;
			char tmp[ 1024 ];
			SGS_AS_REAL( val, t+1 );
			sprintf( tmp, "%g", val );
			sgs_membuf_appbuf( out, C, tmp, strlen( tmp ) );
		}
		break;
	case SGS_ST_NUMINT:
		{
			sgs_Int val;
			char tmp[ 24 ];
			SGS_AS_INTEGER( val, t+1 );
			sprintf( tmp, "%" PRId64, val );
			sgs_membuf_appbuf( out, C, tmp, strlen( tmp ) );
		}
		break;
	case SGS_ST_STRING:
		{
			int32_t i, size;
			SGS_ST_READINT( size, t + 1 );
			sgs_TokenList buf = t + 5;
			for( i = 0; i < size; ++i )
			{
				if( sgs_isgraph( buf[ i ] ) || buf[ i ] == ' ' )
					sgs_membuf_appchr( out, C, (char) buf[ i ] );
				else
				{
					static const char* hexdigs = "0123456789ABCDEF";
					char tmp[ 4 ] = { '\\', 'x', 0, 0 };
					tmp[2] = hexdigs[ (buf[i] & 0xf0) >> 4 ];
					tmp[3] = hexdigs[ buf[i] & 0xf ];
					sgs_membuf_appbuf( out, C, tmp, 4 );
				}
			}
		}
		break;
#define OPR( op ) sgs_membuf_appbuf( out, C, STRLIT_BUF(op) )
	case SGS_ST_OP_RWCMP: OPR( "<=>" ); break;
	case SGS_ST_OP_SEQ: OPR( "===" ); break;
	case SGS_ST_OP_SNEQ: OPR( "!==" ); break;
	case SGS_ST_OP_EQ: OPR( "==" ); break;
	case SGS_ST_OP_NEQ: OPR( "!=" ); break;
	case SGS_ST_OP_LEQ: OPR( "<=" ); break;
	case SGS_ST_OP_GEQ: OPR( ">=" ); break;
	case SGS_ST_OP_ADDEQ: OPR( "+=" ); break;
	case SGS_ST_OP_SUBEQ: OPR( "-=" ); break;
	case SGS_ST_OP_MULEQ: OPR( "*=" ); break;
	case SGS_ST_OP_DIVEQ: OPR( "/=" ); break;
	case SGS_ST_OP_MODEQ: OPR( "%=" ); break;
	case SGS_ST_OP_ANDEQ: OPR( "&=" ); break;
	case SGS_ST_OP_OREQ: OPR( "|=" ); break;
	case SGS_ST_OP_XOREQ: OPR( "^=" ); break;
	case SGS_ST_OP_LSHEQ: OPR( "<<=" ); break;
	case SGS_ST_OP_RSHEQ: OPR( ">>=" ); break;
	case SGS_ST_OP_BLAEQ: OPR( "&&=" ); break;
	case SGS_ST_OP_BLOEQ: OPR( "||=" ); break;
	case SGS_ST_OP_CATEQ: OPR( "$=" ); break;
	case SGS_ST_OP_BLAND: OPR( "&&" ); break;
	case SGS_ST_OP_BLOR: OPR( "||" ); break;
	case SGS_ST_OP_LESS: OPR( "<" ); break;
	case SGS_ST_OP_GRTR: OPR( ">" ); break;
	case SGS_ST_OP_SET: OPR( "=" ); break;
	case SGS_ST_OP_ADD: OPR( "+" ); break;
	case SGS_ST_OP_SUB: OPR( "-" ); break;
	case SGS_ST_OP_MUL: OPR( "*" ); break;
	case SGS_ST_OP_DIV: OPR( "/" ); break;
	case SGS_ST_OP_MOD: OPR( "%" ); break;
	case SGS_ST_OP_AND: OPR( "&" ); break;
	case SGS_ST_OP_OR: OPR( "|" ); break;
	case SGS_ST_OP_XOR: OPR( "^" ); break;
	case SGS_ST_OP_LSH: OPR( "<<" ); break;
	case SGS_ST_OP_RSH: OPR( ">>" ); break;
	case SGS_ST_OP_MMBR: OPR( "." ); break;
	case SGS_ST_OP_CAT: OPR( "$" ); break;
	case SGS_ST_OP_NOT: OPR( "!" ); break;
	case SGS_ST_OP_INV: OPR( "~" ); break;
	case SGS_ST_OP_INC: OPR( "++" ); break;
	case SGS_ST_OP_DEC: OPR( "--" ); break;
#undef OPR
	default:
		sgs_membuf_appbuf( out, C, "<error>", 7 );
		break;
	}
}

static int tp_tt2i( sgs_TokenType t )
{
	/* 0 ident | 1 const | 2 punct | 3 op */
	if( SGS_ST_ISOP( t ) ) return 3;
	if( t == SGS_ST_IDENT || t == SGS_ST_KEYWORD ) return 0;
	if( t == SGS_ST_NUMREAL || t == SGS_ST_NUMINT || t == SGS_ST_STRING ) return 1;
	return 2;
}

/*
	kerning table:
	  I C P O
	I 1 1 0 +
	C 1 1 0 +
	P 0 0 0 +
	O + + + 1
*/
static void tp_kerning( SGS_CTX, sgs_MemBuf* out, sgs_TokenList t1, sgs_TokenList t2, int xs )
{
	static const int32_t mask = 0x8033;
	static const int32_t maskg = 0xf8bb;
	int32_t m;
	int ty1, ty2;
	
	if( !t1 || !t2 )
		return;
	
	m = xs ? maskg : mask;
	ty1 = tp_tt2i( *t1 );
	ty2 = tp_tt2i( *t2 );
	
	if( m & (1 << (ty1 + ty2 * 4)) )
		sgs_membuf_appchr( out, C, ' ' );
}

void sgsT_TokenString( SGS_CTX, sgs_MemBuf* out, sgs_TokenList tlist, sgs_TokenList tend, int xs )
{
	while( tlist < tend && *tlist )
	{
		sgs_TokenList t = tlist;
		tlist = sgsT_Next( t );
		
		tp_token( C, out, t );
		tp_kerning( C, out, t, tlist == tend ? NULL : tlist, xs );
	}
}


void sgsT_DumpToken( sgs_TokenList tok )
{
	switch( *tok )
	{
	case SGS_ST_RBRKL:
	case SGS_ST_RBRKR:
	case SGS_ST_SBRKL:
	case SGS_ST_SBRKR:
	case SGS_ST_CBRKL:
	case SGS_ST_CBRKR:
	case SGS_ST_ARGSEP:
	case SGS_ST_STSEP:
	case SGS_ST_PICKSEP:
		printf( "%c", *tok );
		break;
	case SGS_ST_IDENT:
		fwrite( "id(", 1, 3, stdout );
		fwrite( tok + 2, 1, tok[ 1 ], stdout );
		fwrite( ")", 1, 1, stdout );
		break;
	case SGS_ST_KEYWORD:
		fwrite( "[", 1, 1, stdout );
		fwrite( tok + 2, 1, tok[ 1 ], stdout );
		fwrite( "]", 1, 1, stdout );
		break;
	case SGS_ST_NUMREAL:
		{
			sgs_Real val;
			SGS_AS_REAL( val, tok + 1 );
			printf( "real(%f)", val );
		}
		break;
	case SGS_ST_NUMINT:
		{
			sgs_Int val;
			SGS_AS_INTEGER( val, tok + 1 );
			printf( "int(%" PRId64 ")", val );
		}
		break;
	case SGS_ST_STRING:
		{
			int32_t len;
			SGS_ST_READINT( len, tok + 1 );
			fwrite( "str(", 1, 4, stdout );
			sgs_print_safe( stdout, (const char*) tok + 5, (size_t) len );
			fwrite( ")", 1, 1, stdout );
		}
		break;
#define OPR( op ) printf( "%s", op );
	case SGS_ST_OP_RWCMP: OPR( "<=>" ); break;
	case SGS_ST_OP_SEQ: OPR( "===" ); break;
	case SGS_ST_OP_SNEQ: OPR( "!==" ); break;
	case SGS_ST_OP_EQ: OPR( "==" ); break;
	case SGS_ST_OP_NEQ: OPR( "!=" ); break;
	case SGS_ST_OP_LEQ: OPR( "<=" ); break;
	case SGS_ST_OP_GEQ: OPR( ">=" ); break;
	case SGS_ST_OP_ADDEQ: OPR( "+=" ); break;
	case SGS_ST_OP_SUBEQ: OPR( "-=" ); break;
	case SGS_ST_OP_MULEQ: OPR( "*=" ); break;
	case SGS_ST_OP_DIVEQ: OPR( "/=" ); break;
	case SGS_ST_OP_MODEQ: OPR( "%=" ); break;
	case SGS_ST_OP_ANDEQ: OPR( "&=" ); break;
	case SGS_ST_OP_OREQ: OPR( "|=" ); break;
	case SGS_ST_OP_XOREQ: OPR( "^=" ); break;
	case SGS_ST_OP_LSHEQ: OPR( "<<=" ); break;
	case SGS_ST_OP_RSHEQ: OPR( ">>=" ); break;
	case SGS_ST_OP_BLAEQ: OPR( "&&=" ); break;
	case SGS_ST_OP_BLOEQ: OPR( "||=" ); break;
	case SGS_ST_OP_CATEQ: OPR( "$=" ); break;
	case SGS_ST_OP_BLAND: OPR( "&&" ); break;
	case SGS_ST_OP_BLOR: OPR( "||" ); break;
	case SGS_ST_OP_LESS: OPR( "<" ); break;
	case SGS_ST_OP_GRTR: OPR( ">" ); break;
	case SGS_ST_OP_SET: OPR( "=" ); break;
	case SGS_ST_OP_ADD: OPR( "+" ); break;
	case SGS_ST_OP_SUB: OPR( "-" ); break;
	case SGS_ST_OP_MUL: OPR( "*" ); break;
	case SGS_ST_OP_DIV: OPR( "/" ); break;
	case SGS_ST_OP_MOD: OPR( "%" ); break;
	case SGS_ST_OP_AND: OPR( "&" ); break;
	case SGS_ST_OP_OR: OPR( "|" ); break;
	case SGS_ST_OP_XOR: OPR( "^" ); break;
	case SGS_ST_OP_LSH: OPR( "<<" ); break;
	case SGS_ST_OP_RSH: OPR( ">>" ); break;
	case SGS_ST_OP_MMBR: OPR( "." ); break;
	case SGS_ST_OP_ERSUP: OPR( "@" ); break;
	case SGS_ST_OP_CAT: OPR( "$" ); break;
	case SGS_ST_OP_NOT: OPR( "!" ); break;
	case SGS_ST_OP_INV: OPR( "~" ); break;
	case SGS_ST_OP_INC: OPR( "++" ); break;
	case SGS_ST_OP_DEC: OPR( "--" ); break;
#undef OPR
	default:
		fwrite( "<invalid>", 1, 9, stdout );
		break;
	}
}
void sgsT_DumpList( sgs_TokenList tlist, sgs_TokenList tend )
{
	printf( "\n" );
	while( tlist != tend && *tlist != 0 )
	{
		printf( "   " );
		sgsT_DumpToken( tlist );
		tlist = sgsT_Next( tlist );
	}
	printf( "\n\n" );
}


#include <stdlib.h>
#include <errno.h>

#ifdef _WIN32
#  define WIN32_LEAN_AND_MEAN
#  define NOSERVICE
#  define NOUSER
#  define NOWH
#  define NOMCX
#  define NOMINMAX
#  include <windows.h>
#  define SGS_MAX_PATH 4096
#else
#  include <string.h>
#  include <dlfcn.h>
#  include <sys/stat.h>
#  include <unistd.h>
#  ifndef PATH_MAX
#    define PATH_MAX 4096
#  endif
#  define SGS_MAX_PATH PATH_MAX
#  if defined(__APPLE__) && defined(__MACH__)
#    include <mach-o/dyld.h>
#  endif
#endif




int sgsXPC_GetProcAddress( const char* file, const char* proc, void** out )
{
#if SGS_WINAPP
	return SGS_XPC_NOTSUP;
#elif defined(WIN32)
	HMODULE mod;
	UINT pe;
	WCHAR widepath[ SGS_MAX_PATH + 1 ];
	WCHAR abspath[ SGS_MAX_PATH + 1 ];
	WCHAR* pathstr;
	int ret_i;
	DWORD ret_dw;
	
	ret_i = MultiByteToWideChar( CP_UTF8, 0, file, -1, widepath, SGS_MAX_PATH );
	if( ret_i == 0 )
		return SGS_XPC_NOFILE;
	widepath[ SGS_MAX_PATH ] = 0;
	
	pathstr = widepath;
	ret_dw = GetFullPathNameW( widepath, SGS_MAX_PATH, abspath, NULL );
	if( ret_dw > 0 && ret_dw < SGS_MAX_PATH )
	{
		abspath[ SGS_MAX_PATH ] = 0;
		pathstr = abspath;
	}
	
	pe = SetErrorMode( SEM_FAILCRITICALERRORS );
	mod = LoadLibraryW( pathstr );
	SetErrorMode( pe );
	ret_dw = GetLastError();
	if( !mod )
	{
		if( ret_dw == ERROR_MOD_NOT_FOUND )
			return SGS_XPC_NOFILE;
		if( ret_dw == ERROR_BAD_EXE_FORMAT )
			return SGS_XPC_NOTLIB;
		return SGS_XPC_LDFAIL;
	}
	
	*out = (void*) GetProcAddress( mod, proc );
	if( !*out )
		return SGS_XPC_NOPROC;
	
	return 0;
	
#else
	void* lib;
	char abspath[ SGS_MAX_PATH + 1 ];
	
	if( realpath( file, abspath ) )
	{
		abspath[ SGS_MAX_PATH ] = 0;
		file = abspath;
	}
	
	lib = dlopen( file, RTLD_NOW | RTLD_GLOBAL );
	if( !lib )
		return SGS_XPC_NOFILE;
	
	*out = (void*) dlsym( lib, proc );
	if( !*out )
		return SGS_XPC_NOPROC;
	
	return 0;
	
#endif
}


char* sgsXPC_GetCurrentDirectory()
{
#if SGS_WINAPP
	errno = ENOTSUP;
	return NULL;

#elif defined(WIN32)
	DWORD buf16_size, buf8_size;
	WCHAR* buf16;
	char* buf8;
	
	buf16_size = GetCurrentDirectoryW( 0, NULL );
	if( buf16_size == 0 )
	{
		errno = EACCES;
		return NULL;
	}
	buf16 = (WCHAR*) malloc( sizeof(WCHAR) * buf16_size );
	if( GetCurrentDirectoryW( buf16_size, buf16 ) == 0 )
	{
		free( buf16 );
		errno = EACCES;
		return NULL;
	}
	
	buf8_size = (DWORD) WideCharToMultiByte( CP_UTF8, 0, buf16, (int) buf16_size, NULL, 0, NULL, NULL );
	if( buf8_size == 0 )
	{
		free( buf16 );
		errno = EACCES;
		return NULL;
	}
	buf8 = (char*) malloc( buf8_size );
	if( WideCharToMultiByte( CP_UTF8, 0, buf16, (int) buf16_size, buf8, (int) buf8_size, NULL, NULL ) == 0 )
	{
		free( buf16 );
		free( buf8 );
		errno = EACCES;
		return NULL;
	}
	free( buf16 );
	
	return buf8;
	
#else
	char stack_buf[ SGS_MAX_PATH ];
	char* buf;
	size_t cur_size = SGS_MAX_PATH, max_size = SGS_MAX_PATH * 10;
	
	buf = getcwd( NULL, 0 );
	if( buf )
		return buf;
	
	errno = 0;
	buf = stack_buf;
	for(;;)
	{
		if( getcwd( buf, cur_size ) )
		{
			if( buf == stack_buf )
			{
				size_t len = strlen( buf ) + 1;
				buf = (char*) malloc( len );
				memcpy( buf, stack_buf, len );
			}
			return buf;
		}
		if( errno != ERANGE || cur_size >= max_size )
		{
			if( buf != stack_buf )
				free( buf );
			return NULL;
		}
		cur_size *= 2;
		if( buf == stack_buf )
			buf = malloc( cur_size );
		else
			buf = realloc( buf, cur_size );
		if( !buf )
		{
			errno = ENOMEM;
			return NULL;
		}
	}
	return NULL;
	
#endif
}

int sgsXPC_SetCurrentDirectory( char* path )
{
#if SGS_WINAPP
	errno = ENOTSUP;
	return -1;

#elif defined(WIN32)
	size_t path_len;
	int buf16_size;
	WCHAR stack_buf16[ SGS_MAX_PATH ];
	WCHAR* buf16;
	BOOL result;
	
	buf16 = stack_buf16;
	path_len = strlen( path );
	buf16_size = MultiByteToWideChar( CP_UTF8, 0, path, (int) path_len + 1, NULL, 0 );
	if( buf16_size == 0 )
	{
		errno = EACCES;
		return -1;
	}
	if( buf16_size > SGS_MAX_PATH )
		buf16 = (WCHAR*) malloc( sizeof(WCHAR) * (size_t) buf16_size );
	if( MultiByteToWideChar( CP_UTF8, 0, path, (int) path_len + 1, buf16, buf16_size ) == 0 )
	{
		if( buf16 != stack_buf16 )
			free( buf16 );
		errno = EACCES;
		return -1;
	}
	result = SetCurrentDirectoryW( buf16 );
	if( buf16 != stack_buf16 )
		free( buf16 );
	if( !result )
	{
		DWORD error;
		error = GetLastError();
		if( error == ERROR_ACCESS_DENIED )
			errno = EACCES;
		else
			errno = ENOENT;
	}
	return result ? 0 : -1;
	
#else
	return chdir( path );
	
#endif
}

char* sgsXPC_GetModuleFileName()
{
#if SGS_WINAPP
	errno = ENOTSUP;
	return NULL;
	
#elif defined(WIN32)
	WCHAR buf16[ 32768 ];
	DWORD buf16_size, buf8_size;
	char* buf8;
	
	buf16_size = GetModuleFileNameW( NULL, buf16, 32768 );
	if( buf16_size == 0 || buf16_size >= 32768 )
	{
		errno = EACCES;
		return NULL;
	}
	buf16[ buf16_size ] = 0;
	
	buf8_size = (DWORD) WideCharToMultiByte( CP_UTF8, 0, buf16, (int) buf16_size + 1, NULL, 0, NULL, NULL );
	if( buf8_size == 0 )
	{
		errno = EACCES;
		return NULL;
	}
	buf8 = (char*) malloc( buf8_size );
	if( WideCharToMultiByte( CP_UTF8, 0, buf16, (int) buf16_size + 1, buf8, (int) buf8_size, NULL, NULL ) == 0 )
	{
		free( buf8 );
		errno = EACCES;
		return NULL;
	}
	
	return buf8;
	
#elif defined(__APPLE__) && (__MACH__)
	char stack_buf[ SGS_MAX_PATH + 1 ];
	char* buf;
	uint32_t bufsize = SGS_MAX_PATH;
	errno = 0;
	if( _NSGetExecutablePath( stack_buf, &bufsize ) == 0 )
	{
		size_t len = strlen( stack_buf ) + 1;
		buf = (char*) malloc( len );
		if( buf )
			memcpy( buf, stack_buf, len );
		else
			errno = ENOMEM;
		return buf;
	}
	else
	{
		buf = (char*) malloc( bufsize + 1 );
		if( !buf )
		{
			errno = ENOMEM;
			return NULL;
		}
		if( _NSGetExecutablePath( buf, &bufsize ) != 0 )
		{
			errno = EACCES;
			free( buf );
			return NULL;
		}
		buf[ bufsize ] = 0;
		return buf;
	}
	
#else
	char stack_buf[ SGS_MAX_PATH + 1 ];
	char* buf;
	size_t cur_size = SGS_MAX_PATH, max_size = SGS_MAX_PATH * 10;
	
	errno = 0;
	buf = stack_buf;
	for(;;)
	{
		ssize_t res = readlink( "/proc/self/exe", buf, cur_size );
		if( res >= 0 && res < cur_size )
		{
			buf[ res ] = 0;
			if( buf == stack_buf )
			{
				size_t len = strlen( buf ) + 1;
				buf = (char*) malloc( len );
				memcpy( buf, stack_buf, len );
			}
			return buf;
		}
		if( errno != ERANGE || cur_size >= max_size )
		{
			if( buf != stack_buf )
				free( buf );
			return NULL;
		}
		cur_size *= 2;
		if( buf == stack_buf )
			buf = malloc( cur_size );
		else
			buf = realloc( buf, cur_size );
		if( !buf )
		{
			errno = ENOMEM;
			return NULL;
		}
	}
	return NULL;
	
#endif
}

